SystemOrganization addCategory: #'Graphics-Display Objects'!
SystemOrganization addCategory: #'Graphics-Files'!
SystemOrganization addCategory: #'Graphics-Primitives'!
SystemOrganization addCategory: #'Graphics-Text'!
SystemOrganization addCategory: #'Graphics-Transformations'!
SystemOrganization addCategory: #'Graphics-Fonts'!
(PackageInfo named: 'Graphics') postscript: '"below, add code to be run after the loading of this package"
StrikeFont installDejaVu'!

Object subclass: #AbstractFont
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Fonts'!

!AbstractFont commentStamp: '<historical>' prior: 0!
AbstractFont defines the generic interface that all fonts need to implement.!

!AbstractFont class methodsFor: 'as yet unclassified' stamp: 'nk 9/1/2004 11:41'!
emphasisStringFor: emphasisCode
	"Answer a translated string that represents the attributes given in emphasisCode."

	| emphases bit |
	emphasisCode = 0 ifTrue: [ ^'Normal' translated ].

	emphases := (IdentityDictionary new)
		at: 1 put: 'Bold' translated;
		at: 2 put: 'Italic' translated;
		at: 4 put: 'Underlined' translated;
		at: 8 put: 'Narrow' translated;
		at: 16 put: 'StruckOut' translated;
		yourself.

	bit := 1.
	^String streamContents: [ :s |
		[ bit < 32 ] whileTrue: [ | code |
			code := emphasisCode bitAnd: bit.
			code isZero ifFalse: [ s nextPutAll: (emphases at: code); space ].
			bit := bit bitShift: 1 ].
		s position isZero ifFalse: [ s skip: -1 ].
	]! !

!AbstractFont methodsFor: 'measuring' stamp: 'tak 1/11/2005 17:20'!
approxWidthOfText: aText
"Return the width of aText -- quickly, and a little bit dirty. Used by lists morphs containing Text objects to get a quick, fairly accurate measure of the width of a list item."

    | w |
    
    (aText isNil or: [aText size == 0 ])
        ifTrue:[^0].
       
    w := self
        widthOfString: aText asString.

     "If the text has no emphasis, just return the string size.  If it is empasized, 
    just approximate the width by adding about 20% to the width"   
    (((aText runLengthFor: 1) == aText size)
        and: [(aText emphasisAt: 1) == 0 ])
            ifTrue:[^w]
            ifFalse:[ ^w * 6 // 5 ]. ! !

!AbstractFont methodsFor: 'accessing' stamp: 'yo 1/7/2005 11:19'!
ascent

	self subclassResponsibility.
! !

!AbstractFont methodsFor: 'accessing' stamp: 'yo 1/7/2005 11:18'!
ascentOf: aCharacter

	^ self ascent.
! !

!AbstractFont methodsFor: 'accessing' stamp: 'nk 4/2/2004 11:06'!
baseKern
	^0! !

!AbstractFont methodsFor: 'accessing' stamp: 'yo 1/7/2005 11:18'!
basicAscentOf: aCharacter

	^ self ascent.
! !

!AbstractFont methodsFor: 'accessing' stamp: 'yo 1/7/2005 11:19'!
basicDescentOf: aCharacter

	^ self descent.
! !

!AbstractFont methodsFor: 'accessing' stamp: 'ar 5/19/2000 14:56'!
characterToGlyphMap
	"Return the character to glyph mapping table. If the table is not provided the character scanner will query the font directly for the width of each individual character."
	^nil! !

!AbstractFont methodsFor: 'accessing' stamp: 'nk 3/15/2004 18:57'!
derivativeFonts
	^#()! !

!AbstractFont methodsFor: 'accessing' stamp: 'yo 1/7/2005 11:20'!
descent

	self subclassResponsibility.
! !

!AbstractFont methodsFor: 'accessing' stamp: 'yo 1/7/2005 11:20'!
descentOf: aCharacter

	^ self descent.
! !

!AbstractFont methodsFor: 'displaying' stamp: 'ar 5/19/2000 14:59'!
displayString: aString on: aDisplayContext from: startIndex to: stopIndex at: aPoint kern: kernDelta
	"Draw the given string from startIndex to stopIndex 
	at aPoint on the (already prepared) display context."
	^self subclassResponsibility! !

!AbstractFont methodsFor: 'displaying' stamp: 'yo 1/7/2005 11:36'!
displayString: aString on: aDisplayContext from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: baselineY
	"Draw the given string from startIndex to stopIndex 
	at aPoint on the (already prepared) display context."
	^self subclassResponsibility! !

!AbstractFont methodsFor: 'accessing' stamp: 'nk 8/31/2004 09:15'!
familyName
	"Answer the name to be used as a key in the TextConstants dictionary."
	^self subclassResponsibility! !

!AbstractFont methodsFor: 'accessing' stamp: 'nk 8/31/2004 09:14'!
height
	"Answer the height of the receiver, total of maximum extents of 
	characters above and below the baseline."

	^self subclassResponsibility! !

!AbstractFont methodsFor: 'displaying' stamp: 'ar 5/19/2000 14:59'!
installOn: aDisplayContext foregroundColor: foregroundColor backgroundColor: backgroundColor
	"Install the receiver on the given DisplayContext (either BitBlt or Canvas) for further drawing operations."
	^self subclassResponsibility! !

!AbstractFont methodsFor: 'testing' stamp: 'yo 2/12/2007 19:34'!
isFontSet

	^ false.
! !

!AbstractFont methodsFor: 'accessing' stamp: 'nk 5/26/2003 09:45'!
isRegular
	^false! !

!AbstractFont methodsFor: 'testing' stamp: 'nk 6/25/2003 12:54'!
isTTCFont
	^false! !

!AbstractFont methodsFor: 'accessing' stamp: 'nk 8/31/2004 09:14'!
lineGrid
	"Answer the relative space between lines"

	^self subclassResponsibility! !

!AbstractFont methodsFor: 'accessing' stamp: 'nk 4/2/2004 11:33'!
pixelSize
	"Make sure that we don't return a Fraction"
	^ TextStyle pointsToPixels: self pointSize! !

!AbstractFont methodsFor: 'notifications' stamp: 'nk 4/2/2004 11:25'!
pixelsPerInchChanged
	"The definition of TextStyle class>>pixelsPerInch has changed. Do whatever is necessary."! !

!AbstractFont methodsFor: 'accessing' stamp: 'nk 4/1/2004 10:48'!
pointSize
	self subclassResponsibility.! !

!AbstractFont methodsFor: 'caching' stamp: 'nk 3/15/2004 18:47'!
releaseCachedState
	! !

!AbstractFont methodsFor: 'accessing' stamp: 'nk 7/11/2004 21:15'!
textStyle
	^ TextStyle actualTextStyles detect:
		[:aStyle | aStyle fontArray includes: self] ifNone: [ TextStyle fontArray: { self } ]! !

!AbstractFont methodsFor: 'accessing' stamp: 'nk 3/22/2004 15:15'!
textStyleName
	"Answer the name to be used as a key in the TextConstants dictionary."
	^self familyName! !

!AbstractFont methodsFor: 'measuring' stamp: 'ar 5/19/2000 14:58'!
widthOf: aCharacter
	"Return the width of the given character"
	^self subclassResponsibility! !

!AbstractFont methodsFor: 'measuring' stamp: 'ar 12/31/2001 14:25'!
widthOfString: aString
	aString ifNil:[^0].
	^self widthOfString: aString from: 1 to: aString size.
"
	TextStyle default defaultFont widthOfString: 'zort' 21
"! !

!AbstractFont methodsFor: 'measuring' stamp: 'ar 12/31/2001 00:54'!
widthOfString: aString from: startIndex to: stopIndex
	"Measure the length of the given string between start and stop index"
	| character resultX |
	resultX := 0.
	startIndex to: stopIndex do:[:i | 
		character := aString at: i.
		resultX := resultX + (self widthOf: character)].
	^resultX! !

!AbstractFont methodsFor: 'measuring' stamp: 'sps 3/23/2004 15:50'!
widthOfStringOrText: aStringOrText
    aStringOrText ifNil:[^0].
    ^aStringOrText isText
        ifTrue:[self approxWidthOfText: aStringOrText ]
        ifFalse:[self widthOfString: aStringOrText ] ! !

!AbstractFont methodsFor: 'accessing' stamp: 'ar 5/19/2000 14:57'!
xTable
	"Return the xTable for the font. The xTable defines the left x-value for each individual glyph in the receiver. If such a table is not provided, the character scanner will ask the font directly for the appropriate width of each individual character."
	^nil! !

Object subclass: #BDFFontReader
	instanceVariableNames: 'file properties'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Fonts'!

!BDFFontReader commentStamp: '<historical>' prior: 0!
I am a conversion utility for reading X11 Bitmap Distribution Format fonts.  My code is derived from the multilingual Squeak changeset written by OHSHIMA Yoshiki (ohshima@is.titech.ac.jp), although all support for fonts with more than 256 glyphs has been ripped out.  See http://www.is.titech.ac.jp/~ohshima/squeak/squeak-multilingual-e.html .

My class methods contain tools for fetching BDF source files from a well-known archive site, batch conversion to Squeak's .sf2 format, and installation of these fonts as TextStyles.  Also, the legal notices for the standard 75dpi fonts I process this way are included as "x11FontLegalNotices'.!

!BDFFontReader class methodsFor: 'file creation' stamp: 'nop 1/23/2000 19:00'!
convertFilesNamed: fileName toFamilyNamed: familyName inDirectoryNamed: dirName
		"BDFFontReader convertFilesNamed: 'helvR' toFamilyNamed: 'Helvetica' inDirectoryNamed: '' "

	"This utility converts X11 BDF font files to Squeak .sf2 StrikeFont files."

	"For this utility to work as is, the BDF files must be named 'familyNN.bdf',
	and must reside in the directory named by dirName (use '' for the current directory).
	The output StrikeFont files will be named familyNN.sf2, and will be placed in the
	current directory."

	| f allFontNames sizeChars dir |
	"Check for matching file names."
	dir := dirName isEmpty
		ifTrue: [FileDirectory default]
		ifFalse: [FileDirectory default directoryNamed: dirName].
	allFontNames := dir fileNamesMatching: fileName , '##.bdf'.
	allFontNames isEmpty ifTrue: [^ self error: 'No files found like ' , fileName , 'NN.bdf'].
	
	Utilities informUserDuring: [:info |
		allFontNames do: [:fname | 
			info value: 'Converting ', familyName, ' BDF file ', fname, ' to SF2 format'.
			sizeChars := (fname copyFrom: fileName size + 1 to: fname size) copyUpTo: $. .

			f := StrikeFont new readBDFFromFile: (dir fullNameFor: fname) name: familyName, sizeChars.
			f writeAsStrike2named: familyName, sizeChars, '.sf2'.
		].
	]! !

!BDFFontReader class methodsFor: 'resource download' stamp: 'nop 1/23/2000 18:43'!
convertX11FontsToStrike2  "BDFFontReader convertX11FontsToStrike2"
	"Given a set of standard X11 BDF font files (probably downloaded via BDFFontReader downloadFonts), produce .sf2 format fonts.  The source and destination directory is the current directory."

	"Charter currently tickles a bug in the BDF parser.  Skip it for now."
	"self convertFilesNamed: 'charR' toFamilyNamed: 'Charter' inDirectoryNamed: ''."

	self convertFilesNamed: 'courR' toFamilyNamed: 'Courier' inDirectoryNamed: ''.
	self convertFilesNamed: 'helvR' toFamilyNamed: 'Helvetica' inDirectoryNamed: ''.

	self convertFilesNamed: 'lubR' toFamilyNamed: 'LucidaBright' inDirectoryNamed: ''.
	self convertFilesNamed: 'luRS' toFamilyNamed: 'Lucida' inDirectoryNamed: ''.
	self convertFilesNamed: 'lutRS' toFamilyNamed: 'LucidaTypewriter' inDirectoryNamed: ''.

	self convertFilesNamed: 'ncenR' toFamilyNamed: 'NewCenturySchoolbook' inDirectoryNamed: ''.
	self convertFilesNamed: 'timR' toFamilyNamed: 'TimesRoman' inDirectoryNamed: ''.! !

!BDFFontReader class methodsFor: 'resource download' stamp: 'nop 2/11/2001 00:24'!
downloadFonts  "BDFFontReader downloadFonts"
	"Download a standard set of BDF sources from x.org.  
	The combined size of these source files is around 1.2M; after conversion 
	to .sf2 format they may be deleted."

	| heads tails filenames baseUrl basePath newUrl newPath document f |
	heads := #( 'charR' 'courR' 'helvR' 'lubR' 'luRS' 'lutRS' 'ncenR' 'timR' ).
	tails := #( '08' '10' '12' '14' '18' '24').

	filenames := OrderedCollection new.
	heads do: [:head |
		filenames addAll: (tails collect: [:tail | head , tail , '.bdf'])
	].

	baseUrl := Url absoluteFromText: 'http://ftp.x.org/pub/R6.4/xc/fonts/bdf/75dpi/'.
	basePath := baseUrl path.

	filenames do: [:filename |
		newUrl := baseUrl clone.
		newPath := OrderedCollection newFrom: basePath.

		newPath addLast: filename.
		newUrl path: newPath.

		Utilities informUser: 'Fetching ' , filename during: 
			[document := newUrl retrieveContents].

		f := CrLfFileStream newFileNamed: filename.
		f nextPutAll: document content.
		f close.
	].
! !

!BDFFontReader class methodsFor: 'documentation' stamp: 'nop 2/11/2001 00:22'!
gettingAndInstallingTheFonts

	"Download the 1.3M of BDF font source files from x.org:

		BDFFontReader downloadFonts.

	Convert them to .sf2 StrikeFont files:

		BDFFontReader convertX11FontsToStrike2.

	Install them into the system as TextStyles:

		BDFFontReader installX11Fonts.

	Read the legal notices in 'BDFFontReader x11FontLegalNotices' before
	redistributing images containing these fonts."! !

!BDFFontReader class methodsFor: 'resource download' stamp: 'nop 1/23/2000 18:44'!
installX11Fonts "BDFFontReader installX11Fonts"
	"Installs previously-converted .sf2 fonts into the TextConstants dictionary.  This makes them available as TextStyles everywhere in the image."

	| families fontArray textStyle |
	families := #( 'Courier' 'Helvetica' 'LucidaBright' 'Lucida' 'LucidaTypewriter' 'NewCenturySchoolbook' 'TimesRoman' ).

	families do: [:family |
		fontArray := StrikeFont readStrikeFont2Family: family.
		textStyle := TextStyle fontArray: fontArray.
		TextConstants at: family asSymbol put: textStyle.
	].
! !

!BDFFontReader class methodsFor: 'file creation' stamp: 'yo 5/25/2004 10:52'!
new

	^ self basicNew.
! !

!BDFFontReader class methodsFor: 'file creation' stamp: 'ar 10/25/2005 00:21'!
openFileNamed: fileName
	^self new openFileNamed: fileName! !

!BDFFontReader class methodsFor: 'documentation' stamp: 'nop 1/23/2000 18:30'!
x11FontLegalNotices

	^ 'The X11 BDF fonts contain copyright and license information as comments in the
font source code.  For the font family files "cour" (Courier), "helv" (Helvetica), "ncen" (New Century Schoolbook), and "tim" (Times Roman) the notice reads:

COMMENT  Copyright 1984-1989, 1994 Adobe Systems Incorporated.
COMMENT  Copyright 1988, 1994 Digital Equipment Corporation.
COMMENT
COMMENT  Adobe is a trademark of Adobe Systems Incorporated which may be
COMMENT  registered in certain jurisdictions.
COMMENT  Permission to use these trademarks is hereby granted only in
COMMENT  association with the images described in this file.
COMMENT
COMMENT  Permission to use, copy, modify, distribute and sell this software
COMMENT  and its documentation for any purpose and without fee is hereby
COMMENT  granted, provided that the above copyright notices appear in all
COMMENT  copies and that both those copyright notices and this permission
COMMENT  notice appear in supporting documentation, and that the names of
COMMENT  Adobe Systems and Digital Equipment Corporation not be used in
COMMENT  advertising or publicity pertaining to distribution of the software
COMMENT  without specific, written prior permission.  Adobe Systems and
COMMENT  Digital Equipment Corporation make no representations about the
COMMENT  suitability of this software for any purpose.  It is provided "as
COMMENT  is" without express or implied warranty.

For the font family files "char" (Charter), the notice reads:

COMMENT  Copyright 1988 Bitstream, Inc., Cambridge, Massachusetts, USA
COMMENT  Bitstream and Charter are registered trademarks of Bitstream, Inc.
COMMENT  
COMMENT  The names "Bitstream" and "Charter" are registered trademarks of
COMMENT  Bitstream, Inc.  Permission to use these trademarks is hereby
COMMENT  granted only in association with the images described in this file.
COMMENT  
COMMENT  Permission to use, copy, modify, and distribute this software and
COMMENT  its documentation for any purpose and without fee is hereby
COMMENT  granted, provided that the above copyright notice appear in all
COMMENT  copies and that both that copyright notice and this permission
COMMENT  notice appear in supporting documentation, and that the name of
COMMENT  Bitstream not be used in advertising or publicity pertaining to
COMMENT  distribution of the software without specific, written prior
COMMENT  permission.  Bitstream makes no representations about the
COMMENT  suitability of this software for any purpose.  It is provided "as
COMMENT  is" without express or implied warranty.
COMMENT  
COMMENT  BITSTREAM DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
COMMENT  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
COMMENT  NO EVENT SHALL BITSTREAM BE LIABLE FOR ANY SPECIAL, INDIRECT OR
COMMENT  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
COMMENT  OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
COMMENT  NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
COMMENT  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

For the font family files "lu" (Lucida), "lub" (Lucida Bright), and "lut" (Lucida Typewriter),
the notice reads:

COMMENT  (c) Copyright Bigelow & Holmes 1986, 1985. Lucida is a registered
COMMENT  trademark of Bigelow & Holmes. See LEGAL NOTICE file for terms 
COMMENT  of the license. 

The LEGAL NOTICE contains:

This is the LEGAL NOTICE pertaining to the Lucida fonts from Bigelow & Holmes:

	NOTICE TO USER: The source code, including the glyphs or icons 
	forming a par of the OPEN LOOK TM Graphic User Interface, on this 
	tape and in these files is copyrighted under U.S. and international
	laws. Sun Microsystems, Inc. of Mountain View, California owns
	the copyright and has design patents pending on many of the icons. 
	AT&T is the owner of the OPEN LOOK trademark associated with the
	materials on this tape. Users and possessors of this source code 
	are hereby granted a nonexclusive, royalty-free copyright and 
	design patent license to use this code in individual and 
	commercial software. A royalty-free, nonexclusive trademark
	license to refer to the code and output as "OPEN LOOK" compatible 
	is available from AT&T if, and only if, the appearance of the 
	icons or glyphs is not changed in any manner except as absolutely
	necessary to accommodate the standard resolution of the screen or
	other output device, the code and output is not changed except as 
	authorized herein, and the code and output is validated by AT&T. 
	Bigelow & Holmes is the owner of the Lucida (R) trademark for the
	fonts and bit-mapped images associated with the materials on this 
	tape. Users are granted a royalty-free, nonexclusive license to use
	the trademark only to identify the fonts and bit-mapped images if, 
	and only if, the fonts and bit-mapped images are not modified in any
	way by the user. 


	Any use of this source code must include, in the user documentation 
	and internal comments to the code, notices to the end user as  
	follows:


	(c) Copyright 1989 Sun Microsystems, Inc. Sun design patents
	pending in the U.S. and foreign countries. OPEN LOOK is a 
	trademark of AT&T. Used by written permission of the owners.


 	(c) Copyright Bigelow & Holmes 1986, 1985. Lucida is a registered 
	trademark of Bigelow & Holmes. Permission to use the Lucida 
	trademark is hereby granted only in association with the images 
	and fonts described in this file.



	SUN MICROSYSTEMS, INC., AT&T, AND BIGELOW & HOLMES 
	MAKE NO REPRESENTATIONS ABOUT THE SUITABILITY OF
 	THIS SOURCE CODE FOR ANY PURPOSE. IT IS PROVIDED "AS IS" 
	WITHOUT EXPRESS OR IMPLIED WARRANTY OF ANY KIND. 
	SUN  MICROSYSTEMS, INC., AT&T AND BIGELOW  & HOLMES, 
	SEVERALLY AND INDIVIDUALLY, DISCLAIM ALL WARRANTIES 
	WITH REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
	PARTICULAR PURPOSE. IN NO EVENT SHALL SUN MICROSYSTEMS,
	INC., AT&T OR BIGELOW & HOLMES BE LIABLE FOR ANY
	SPECIAL, INDIRECT, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
	OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA 	
	OR PROFITS, WHETHER IN AN ACTION OF  CONTRACT, NEGLIGENCE
	OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
	WITH THE USE OR PERFORMANCE OF THIS SOURCE CODE.

'.





! !

!BDFFontReader methodsFor: 'reading' stamp: 'nop 1/18/2000 19:45'!
errorFileFormat
	self error: 'malformed bdf format'! !

!BDFFontReader methodsFor: 'reading' stamp: 'nop 1/18/2000 19:46'!
errorUnsupported
	self error: 'unsupported bdf'! !

!BDFFontReader methodsFor: 'reading' stamp: 'ar 10/25/2005 00:35'!
getLine
	^file upTo: Character cr.! !

!BDFFontReader methodsFor: 'initialize' stamp: 'nop 1/18/2000 19:44'!
initialize
	properties := Dictionary new.! !

!BDFFontReader methodsFor: 'initialize' stamp: 'ar 10/25/2005 00:34'!
openFileNamed: fileName
	file := CrLfFileStream readOnlyFileNamed: fileName.! !

!BDFFontReader methodsFor: 'reading' stamp: 'ar 10/25/2005 00:35'!
read
	| xTable strikeWidth glyphs ascent descent minAscii maxAscii maxWidth chars charsNum height form encoding bbx array width blt lastAscii pointSize ret stream |
	form _ encoding _ bbx _ nil.
	self readAttributes.
	height _ Integer readFromString: ((properties at: #FONTBOUNDINGBOX) at: 2).
	ascent _ Integer readFromString: (properties at: 'FONT_ASCENT' asSymbol) first.
	descent _ Integer readFromString: (properties at: 'FONT_DESCENT' asSymbol) first.
	pointSize _ (Integer readFromString: (properties at: 'POINT_SIZE' asSymbol) first) // 10.
	
	maxWidth _ 0.
	minAscii _ 9999.
	strikeWidth _ 0.
	maxAscii _ 0.

	charsNum _ Integer readFromString: (properties at: #CHARS) first.
	chars _ Set new: charsNum.

	1 to: charsNum do: [:i |
		array _ self readOneCharacter.
		stream _ ReadStream on: array.
		form _ stream next.
		encoding _ stream next.
		bbx _ stream next.
		form ifNotNil: [
			width _ bbx at: 1.
			maxWidth _ maxWidth max: width.
			minAscii _ minAscii min: encoding.
			maxAscii _ maxAscii max: encoding.
			strikeWidth _ strikeWidth + width.
			chars add: array.
		].
	].

	chars _ chars asSortedCollection: [:x :y | (x at: 2) <= (y at: 2)].
	charsNum _ chars size. "undefined encodings make this different"

	charsNum > 256 ifTrue: [
		"it should be 94x94 charset, and should be fixed width font"
		strikeWidth _ 94*94*maxWidth.
		maxAscii _ 94*94.
		minAscii _ 0.
		xTable _ XTableForFixedFont new.
		xTable maxAscii: 94*94.
		xTable width: maxWidth.
	] ifFalse: [
		xTable _ (Array new: 258) atAllPut: 0.
	].

	glyphs _ Form extent: strikeWidth@height.
	blt _ BitBlt toForm: glyphs.
	lastAscii _ 0.
	
	charsNum > 256 ifTrue: [
		1 to: charsNum do: [:i |
			stream _ ReadStream on: (chars at: i).
			form _ stream next.
			encoding _ stream next.
			bbx _ stream next.
			encoding _ ((encoding // 256) - 33) * 94 + ((encoding \\ 256) - 33).
			blt copy: ((encoding * maxWidth)@0 extent: maxWidth@height)
				from: 0@0 in: form.
		].
	] ifFalse: [
		1 to: charsNum do: [:i |
			stream _ ReadStream on: (chars at: i).
			form _ stream next.
			encoding _ stream next.
			bbx _ stream next.
			lastAscii+1 to: encoding-1 do: [:a | xTable at: a+2 put: (xTable at: a+1)].
			blt copy: (((xTable at: encoding+1)@(ascent - (bbx at: 2) - (bbx at: 4)))
					extent: (bbx at: 1)@(bbx at: 2))
				from: 0@0 in: form.
			xTable at: encoding+2 put: (xTable at: encoding+1)+(bbx at: 1).
			lastAscii _ encoding.
		]
	].

	ret _ Array new: 8.
	ret at: 1 put: xTable.
	ret at: 2 put: glyphs.
	ret at: 3 put: minAscii.
	ret at: 4 put: maxAscii.
	ret at: 5 put: maxWidth.
	ret at: 6 put: ascent.
	ret at: 7 put: descent.
	ret at: 8 put: pointSize.
	^ret.
" ^{xTable. glyphs. minAscii. maxAscii. maxWidth. ascent. descent. pointSize}"
! !

!BDFFontReader methodsFor: 'reading' stamp: 'ar 10/25/2005 00:35'!
readAttributes
	| str a |
	"I don't handle double-quotes correctly, but it works"
	file reset.
	[file atEnd] whileFalse: [
		str := self getLine.
		(str beginsWith: 'STARTCHAR') ifTrue: [file skip: (0 - str size - 1). ^self].
		a := str substrings.
		properties at: a first asSymbol put: a allButFirst.
	].
	self error: 'file seems corrupted'.! !

!BDFFontReader methodsFor: 'reading' stamp: 'ar 10/25/2005 00:35'!
readChars
	| strikeWidth ascent descent minAscii maxAscii maxWidth chars charsNum height form encoding bbx array width pointSize stream |
	form _ encoding _ bbx _ nil.
	self readAttributes.
	height _ Integer readFromString: ((properties at: #FONTBOUNDINGBOX) at: 2).
	ascent _ Integer readFromString: (properties at: 'FONT_ASCENT' asSymbol) first.
	descent _ Integer readFromString: (properties at: 'FONT_DESCENT' asSymbol) first.
	pointSize _ (Integer readFromString: (properties at: 'POINT_SIZE' asSymbol) first) // 10.
	
	maxWidth _ 0.
	minAscii _ 9999.
	strikeWidth _ 0.
	maxAscii _ 0.

	charsNum _ Integer readFromString: (properties at: #CHARS) first.
	chars _ Set new: charsNum.

	1 to: charsNum do: [:i |
		array _ self readOneCharacter.
		stream _ ReadStream on: array.
		form _ stream next.
		encoding _ stream next.
		bbx _ stream next.
		form ifNotNil: [
			width _ bbx at: 1.
			maxWidth _ maxWidth max: width.
			minAscii _ minAscii min: encoding.
			maxAscii _ maxAscii max: encoding.
			strikeWidth _ strikeWidth + width.
			chars add: array.
		].
	].

	chars _ chars asSortedCollection: [:x :y | (x at: 2) <= (y at: 2)].

	^ chars.
! !

!BDFFontReader methodsFor: 'reading' stamp: 'ar 10/25/2005 00:35'!
readOneCharacter
	| str a encoding bbx form bits hi low pos |
	((str := self getLine) beginsWith: 'ENDFONT') ifTrue: [^ {nil. nil. nil}].
	(str beginsWith: 'STARTCHAR') ifFalse: [self errorFileFormat].
	((str := self getLine) beginsWith: 'ENCODING') ifFalse: [self errorFileFormat].
	encoding := Integer readFromString: str substrings second.
	(self getLine beginsWith: 'SWIDTH') ifFalse: [self errorFileFormat].
	(self getLine beginsWith: 'DWIDTH') ifFalse: [self errorFileFormat].
	
	((str := self getLine) beginsWith: 'BBX') ifFalse: [self errorFileFormat].
	a := str substrings.
	bbx := (2 to: 5) collect: [:i | Integer readFromString: (a at: i)].
	((str := self getLine) beginsWith: 'ATTRIBUTES') ifTrue: [str := self getLine].
	(str beginsWith: 'BITMAP') ifFalse: [self errorFileFormat].

	form := Form extent: (bbx at: 1)@(bbx at: 2).
	bits := form bits.
	pos := 0.
	1 to: (bbx at: 2) do: [:t |
		1 to: (((bbx at: 1) - 1) // 8 + 1) do: [:i |
			hi := (('0123456789ABCDEF' indexOf: (file next asUppercase)) - 1) bitShift: 4.
			low := ('0123456789ABCDEF' indexOf: (file next asUppercase)) - 1.
			
			bits byteAt: (pos+i) put: (hi+low).
		].
		file next ~= Character cr ifTrue: [self errorFileFormat].
		pos := pos + ((((bbx at: 1) // 32) + 1) * 4).
	].

	(self getLine beginsWith: 'ENDCHAR') ifFalse: [self errorFileFormat].

	encoding < 0 ifTrue: [^{nil. nil. nil}].
	^{form. encoding. bbx}.
	
	
	! !

Object subclass: #BitBlt
	instanceVariableNames: 'destForm sourceForm halftoneForm combinationRule destX destY width height sourceX sourceY clipX clipY clipWidth clipHeight colorMap'
	classVariableNames: 'CachedFontColorMaps ColorConvertingMaps'
	poolDictionaries: ''
	category: 'Graphics-Primitives'!

!BitBlt commentStamp: 'ar 8/4/2009 20:19' prior: 0!
WARNING: BitBlt's shape cannot be modified since WarpBlt relies on the exact layout.

I represent a block transfer (BLT) of pixels into a rectangle (destX, destY, width, height) of the destinationForm.  The source of pixels may be a similar rectangle (at sourceX, sourceY) in the sourceForm, or a constant color, currently called halftoneForm.  If both are specified, their pixel values are combined with a logical AND function prior to transfer.  In any case, the pixels from the source are combined with those of the destination by as specified by the combinationRule.

The combination rule whose value is 0 through 15 programs the transfer to produce 1 or 0 according to its 4-bit representation as follows:
	8:	if source is 0 and destination is 0
	4:	if source is 0 and destination is 1
	2:	if source is 1 and destination is 0
	1:	if source is 1 and destination is 1.
At each pixel the corresponding bits of the source and destination pixel values determine one of these conditions;  if the combination rule has a 1 in the corresponding bit position, then the new destination value will be 1, otherwise it will be zero.  Forms may be of different depths, see the comment in class Form.

In addition to the original 16 combination rules, this BitBlt supports
	16	fails (to simulate paint bits)
	17	fails (to simulate erase bits)
	18	sourceWord + destinationWord
	19	sourceWord - destinationWord
	20	rgbAdd: sourceWord with: destinationWord.  Sum of color components
	21	rgbSub: sourceWord with: destinationWord.  Difference of color components
	22	OLDrgbDiff: sourceWord with: destinationWord.  Sum of abs of differences in components
	23	OLDtallyIntoMap: destinationWord.  Tallies pixValues into a colorMap
			these old versions don't do bitwise dest clipping.  Use 32 and 33 now.
	24	alphaBlend: sourceWord with: destinationWord.  32-bit source and dest only
	25	pixPaint: sourceWord with: destinationWord.  Wherever the sourceForm is non-zero, it replaces the destination.  Can be used with a 1-bit source color mapped to (0, FFFFFFFF), and a fillColor to fill the dest with that color wherever the source is 1.
	26	pixMask: sourceWord with: destinationWord.  Like pixPaint, but fills with 0.
	27	rgbMax: sourceWord with: destinationWord.  Max of each color component.
	28	rgbMin: sourceWord with: destinationWord.  Min of each color component.
	29	rgbMin: sourceWord bitInvert32 with: destinationWord.  Min with (max-source)
	30	alphaBlendConst: sourceWord with: destinationWord.  alpha is an arg. works in 16 bits.
	31	alphaPaintConst: sourceWord with: destinationWord.  alpha is an arg. works in 16 bits.
	32	rgbDiff: sourceWord with: destinationWord.  Sum of abs of differences in components
	33	tallyIntoMap: destinationWord.  Tallies pixValues into a colorMap
	34	alphaBlendScaled: srcWord with: dstWord. Alpha blend of scaled srcWord and destWord.

The color specified by halftoneForm may be either a Color or a Pattern.   A Color is converted to a pixelValue for the depth of the destinationForm.  If a Pattern, BitBlt will simply interpret its bitmap as an array of Color pixelValues.  BitBlt aligns the first element of this array with the top scanline of the destinationForm, the second with the second, and so on, cycling through the color array as necessary.  Within each scan line the 32-bit value is repeated from left to right across the form.  If the value repeats on pixels boudaries, the effect will be a constant color;  if not, it will produce a halftone that repeats on 32-bit boundaries.

Any transfer specified is further clipped by the specified rectangle (clipX, clipY, clipWidth, clipHeight), and also by the bounds of the source and destination forms.
	To make a small Form repeat and fill a big form, use an InfiniteForm as the source.
	To write on a form and leave with both transparent and opapue areas, use a MaskedForm as the source.

Pixels from a source to a destination whose pixels have a different depth are converted based on the optional colorMap.  If colorMap is nil, then conversion to more bits is done by filling the new high-order bits with zero, and conversion to fewer bits is done by truncating the lost high-order bits.  

The colorMap, if specified, must be a either word array (ie Bitmap) with 2^n elements, where n is the pixel depth of the source, or a fully specified ColorMap which may contain a lookup table (ie Bitmap) and/or four separate masks and shifts which are applied to the pixels. For every source pixel, BitBlt will first perform masking and shifting and then index the lookup table, and select the corresponding pixelValue and mask it to the destination pixel size before storing.
	When blitting from a 32 or 16 bit deep Form to one 8 bits or less, the default is truncation.  This will produce very strange colors, since truncation of the high bits does not produce the nearest encoded color.  Supply a 512 long colorMap, and red, green, and blue will be shifted down to 3 bits each, and mapped.  The message copybits...stdColors will use the best map to the standard colors for destinations of depths 8, 4, 2 and 1.  Two other sized of colorMaps are allowed, 4096 (4 bits per color) and 32786 (five bits per color).
	Normal blits between 16 and 32 bit forms truncates or pads the colors automatically to provide the best preservation of colors.
	Colors can be remapped at the same depth.  Sometimes a Form is in terms of colors that are not the standard colors for this depth, for example in a GIF file.  Convert the Form to a MaskedForm and send colorMap: the list of colors that the picture is in terms of.  MaskedForm will use the colorMap when copying to the display or another Form. (Note also that a Form can be copied to itself, and transformed in the process, if a non-nil colorMap is supplied.)!

!BitBlt class methodsFor: 'examples' stamp: 'di 12/1/97 12:08'!
alphaBlendDemo
	"To run this demo, use...
		Display restoreAfter: [BitBlt alphaBlendDemo]	
	Displays 10 alphas, then lets you paint.  Option-Click to stop painting."

	"This code exhibits alpha blending in any display depth by performing
	the blend in an off-screen buffer with 32-bit pixels, and then copying
	the result back onto the screen with an appropriate color map. - tk 3/10/97"
	
	"This version uses a sliding buffer for painting that keeps pixels in 32 bits
	as long as they are in the buffer, so as not to lose info by converting down
	to display resolution and back up to 32 bits at each operation. - di 3/15/97"

	| brush buff dispToBuff buffToDisplay mapDto32 map32toD prevP p brushToBuff theta buffRect buffSize buffToBuff brushRect delta newBuffRect updateRect |  

	"compute color maps if needed"
	Display depth <= 8 ifTrue: [
		mapDto32 := Color cachedColormapFrom: Display depth to: 32.
		map32toD := Color cachedColormapFrom: 32 to: Display depth].

	"display 10 different alphas, across top of screen"
	buff := Form extent: 500@50 depth: 32.
	dispToBuff := BitBlt toForm: buff.
	dispToBuff colorMap: mapDto32.
	dispToBuff copyFrom: (50@10 extent: 500@50) in: Display to: 0@0.
	1 to: 10 do: [:i | dispToBuff fill: (50*(i-1)@0 extent: 50@50)
						fillColor: (Color red alpha: i/10)
						rule: Form blend].
	buffToDisplay := BitBlt toForm: Display.
	buffToDisplay colorMap: map32toD.
	buffToDisplay copyFrom: buff boundingBox in: buff to: 50@10.

	"Create a brush with radially varying alpha"
	brush := Form extent: 30@30 depth: 32.
	1 to: 5 do: 
		[:i | brush fillShape: (Form dotOfSize: brush width*(6-i)//5)
				fillColor: (Color red alpha: 0.02 * i - 0.01)
				at: brush extent // 2].

	"Now paint with the brush using alpha blending."
	buffSize := 100.
	buff := Form extent: brush extent + buffSize depth: 32.  "Travelling 32-bit buffer"
	dispToBuff := BitBlt toForm: buff.  "This is from Display to buff"
	dispToBuff colorMap: mapDto32.
	brushToBuff := BitBlt toForm: buff.  "This is from brush to buff"
	brushToBuff sourceForm: brush; sourceOrigin: 0@0.
	brushToBuff combinationRule: Form blend.
	buffToBuff := BitBlt toForm: buff.  "This is for slewing the buffer"

	[Sensor yellowButtonPressed] whileFalse:
		[prevP := nil.
		buffRect := Sensor cursorPoint - (buffSize // 2) extent: buff extent.
		dispToBuff copyFrom: buffRect in: Display to: 0@0.
		[Sensor redButtonPressed] whileTrue:
			["Here is the painting loop"
			p := Sensor cursorPoint - (brush extent // 2).
			(prevP == nil or: [prevP ~= p]) ifTrue:
				[prevP == nil ifTrue: [prevP := p].
				(p dist: prevP) > buffSize ifTrue:
					["Stroke too long to fit in buffer -- clip to buffer,
						and next time through will do more of it"
					theta := (p-prevP) theta.
					p := ((theta cos@theta sin) * buffSize asFloat + prevP) truncated].
				brushRect := p extent: brush extent.
				(buffRect containsRect: brushRect) ifFalse:
					["Brush is out of buffer region.  Scroll the buffer,
						and fill vacated regions from the display"
					delta := brushRect amountToTranslateWithin: buffRect.
					buffToBuff copyFrom: buff boundingBox in: buff to: delta.
					newBuffRect := buffRect translateBy: delta negated.
					(newBuffRect areasOutside: buffRect) do:
						[:r | dispToBuff copyFrom: r in: Display to: r origin - newBuffRect origin].
					buffRect := newBuffRect].

				"Interpolate from prevP to p..."
				brushToBuff drawFrom: prevP - buffRect origin
									to: p - buffRect origin
									withFirstPoint: false.

				"Update (only) the altered pixels of the destination"
				updateRect := (p min: prevP) corner: (p max: prevP) + brush extent.
				buffToDisplay copy: updateRect from: updateRect origin - buffRect origin in: buff.
				prevP := p]]]! !

!BitBlt class methodsFor: 'examples' stamp: 'di 12/1/97 12:09'!
antiAliasDemo 
	"To run this demo, use...
		Display restoreAfter: [BitBlt antiAliasDemo]
	Goes immediately into on-screen paint mode.  Option-Click to stop painting."

	"This code exhibits alpha blending in any display depth by performing
	the blend in an off-screen buffer with 32-bit pixels, and then copying
	the result back onto the screen with an appropriate color map. - tk 3/10/97"
	
	"This version uses a sliding buffer for painting that keeps pixels in 32 bits
	as long as they are in the buffer, so as not to lose info by converting down
	to display resolution and back up to 32 bits at each operation. - di 3/15/97"
	
	"This version also uses WarpBlt to paint into twice as large a buffer,
	and then use smoothing when reducing back down to the display.
	In fact this same routine will now work for 3x3 soothing as well.
	Remove the statements 'buff displayAt: 0@0' to hide the buffer. - di 3/19/97"

	| brush buff dispToBuff buffToDisplay mapDto32 map32toD prevP p brushToBuff theta buffRect buffSize buffToBuff brushRect delta newBuffRect updateRect scale p0 |  
	"compute color maps if needed"
	Display depth <= 8 ifTrue: [
		mapDto32 := Color cachedColormapFrom: Display depth to: 32.
		map32toD := Color cachedColormapFrom: 32 to: Display depth].

	"Create a brush with radially varying alpha"
	brush := Form extent: 3@3 depth: 32.
	brush fill: brush boundingBox fillColor: (Color red alpha: 0.05).
	brush fill: (1@1 extent: 1@1) fillColor: (Color red alpha: 0.2).

	scale := 2.  "Actual drawing happens at this magnification"
	"Scale brush up for painting in magnified buffer"
	brush := brush magnify: brush boundingBox by: scale.

	"Now paint with the brush using alpha blending."
	buffSize := 100.
	buff := Form extent: (brush extent + buffSize) * scale depth: 32.  "Travelling 32-bit buffer"
	dispToBuff := (WarpBlt toForm: buff)  "From Display to buff - magnify by 2"
		sourceForm: Display;
		colorMap: mapDto32;
		combinationRule: Form over.
	brushToBuff := (BitBlt toForm: buff)  "From brush to buff"
		sourceForm: brush;
		sourceOrigin: 0@0;
		combinationRule: Form blend.
	buffToDisplay := (WarpBlt toForm: Display)  "From buff to Display - shrink by 2"
		sourceForm: buff;
		colorMap: map32toD;
		cellSize: scale;  "...and use smoothing"
		combinationRule: Form over.
	buffToBuff := BitBlt toForm: buff.  "This is for slewing the buffer"

	[Sensor yellowButtonPressed] whileFalse:
		[prevP := nil.
		buffRect := Sensor cursorPoint - (buff extent // scale // 2) extent: buff extent // scale.
		p0 := (buff extent // 2) - (buffRect extent // 2).
		dispToBuff copyQuad: buffRect innerCorners toRect: buff boundingBox.
buff displayAt: 0@0.  "** remove to hide sliding buffer **"
		[Sensor redButtonPressed] whileTrue:
			["Here is the painting loop"
			p := Sensor cursorPoint - buffRect origin + p0.  "p, prevP are rel to buff origin"
			(prevP == nil or: [prevP ~= p]) ifTrue:
				[prevP == nil ifTrue: [prevP := p].
				(p dist: prevP) > (buffSize-1) ifTrue:
					["Stroke too long to fit in buffer -- clip to buffer,
						and next time through will do more of it"
					theta := (p-prevP) theta.
					p := ((theta cos@theta sin) * (buffSize-2) asFloat + prevP) truncated].
				brushRect := p extent: brush extent.
				((buff boundingBox insetBy: scale) containsRect: brushRect) ifFalse:
					["Brush is out of buffer region.  Scroll the buffer,
						and fill vacated regions from the display"
					delta := (brushRect amountToTranslateWithin: (buff boundingBox insetBy: scale)) // scale.
					buffToBuff copyFrom: buff boundingBox in: buff to: delta*scale.
					newBuffRect := buffRect translateBy: delta negated.
					p := p translateBy: delta*scale.
					prevP := prevP translateBy: delta*scale.
					(newBuffRect areasOutside: buffRect) do:
						[:r | dispToBuff copyQuad: r innerCorners toRect: (r origin - newBuffRect origin*scale extent: r extent*scale)].
					buffRect := newBuffRect].

				"Interpolate from prevP to p..."
				brushToBuff drawFrom: prevP to: p withFirstPoint: false.
buff displayAt: 0@0.  "** remove to hide sliding buffer **"

				"Update (only) the altered pixels of the destination"
				updateRect := (p min: prevP) corner: (p max: prevP) + brush extent.
				updateRect := updateRect origin // scale * scale
						corner: updateRect corner + scale // scale * scale.
				buffToDisplay copyQuad: updateRect innerCorners
							toRect: (updateRect origin // scale + buffRect origin
										extent: updateRect extent // scale).
				prevP := p]]]! !

!BitBlt class methodsFor: 'instance creation' stamp: 'ar 5/28/2000 12:04'!
asGrafPort
	"Return the GrafPort associated with the receiver"
	^GrafPort! !

!BitBlt class methodsFor: 'benchmarks' stamp: 'ar 4/24/2001 23:49'!
benchDiffsFrom: before to: afterwards
	"Given two outputs of BitBlt>>benchmark show the relative improvements."
	| old new log oldLine newLine oldVal newVal improvement |
	log := WriteStream on: String new.
	old := ReadStream on: before.
	new := ReadStream on: afterwards.
	[old atEnd or:[new atEnd]] whileFalse:[
		oldLine := old upTo: Character cr.
		newLine := new upTo: Character cr.
		(oldLine includes: Character tab) ifTrue:[
			oldLine := ReadStream on: oldLine.
			newLine := ReadStream on: newLine.
			Transcript cr; show: (oldLine upTo: Character tab); tab.
			log cr; nextPutAll: (newLine upTo: Character tab); tab.

			[oldLine skipSeparators. newLine skipSeparators.
			oldLine atEnd] whileFalse:[
				oldVal := Integer readFrom: oldLine.
				newVal := Integer readFrom: newLine.
				improvement := oldVal asFloat / newVal asFloat roundTo: 0.01.
				Transcript show: improvement printString; tab; tab.
				log print: improvement; tab; tab].
		] ifFalse:[
			Transcript cr; show: oldLine.
			log cr; nextPutAll: oldLine.
		].
	].
	^log contents! !

!BitBlt class methodsFor: 'benchmarks' stamp: 'ar 5/14/2001 23:31'!
benchmark		"BitBlt benchmark"
	"Run a benchmark on different combinations rules, source/destination depths and BitBlt modes. Note: This benchmark doesn't give you any 'absolute' value - it is intended only for benchmarking improvements in the bitblt code and nothing else.
	Attention: *this*may*take*a*while*"
	| bb source dest destRect log t |
	log := WriteStream on: String new.
	destRect := 0@0 extent: 600@600.
	"Form paint/Form over - the most common rules"
	#( 25 3 ) do:[:rule|
		Transcript cr; show:'---- Combination rule: ', rule printString,' ----'.
		log cr; nextPutAll:'---- Combination rule: ', rule printString,' ----'.
		#(1 2 4 8 16 32) do:[:destDepth|
			dest := nil.
			dest := Form extent: destRect extent depth: destDepth.
			Transcript cr.
			log cr.
			#(1 2 4 8 16 32) do:[:sourceDepth|
				Transcript cr; show: sourceDepth printString, ' => ', destDepth printString.
				log cr; nextPutAll: sourceDepth printString, ' => ', destDepth printString.
				source := nil. bb := nil.
				source := Form extent: destRect extent depth: sourceDepth.
				(source getCanvas) fillOval: dest boundingBox color: Color yellow borderWidth: 30 borderColor: Color black.
				bb := WarpBlt toForm: dest.
				bb sourceForm: source.
				bb sourceRect: source boundingBox.
				bb destRect: dest boundingBox.
				bb colorMap: (source colormapIfNeededFor: dest).
				bb combinationRule: rule.

				"Measure speed of copyBits"
				t := Time millisecondsToRun:[bb copyBits].
				Transcript tab; show: t printString.
				log tab; nextPutAll: t printString.

				bb sourceForm: source destRect: source boundingBox.

				"Measure speed of 1x1 warpBits"
				bb cellSize: 1.
				t := Time millisecondsToRun:[bb warpBits].
				Transcript tab; show: t printString.
				log tab; nextPutAll: t printString.

				"Measure speed of 2x2 warpBits"
				bb cellSize: 2.
				t := Time millisecondsToRun:[bb warpBits].
				Transcript tab; show: t printString.
				log tab; nextPutAll: t printString.

				"Measure speed of 3x3 warpBits"
				bb cellSize: 3.
				t := Time millisecondsToRun:[bb warpBits].
				Transcript tab; show: t printString.
				log tab; nextPutAll: t printString.
			].
		].
	].
	^log contents! !

!BitBlt class methodsFor: 'benchmarks' stamp: 'ar 5/14/2001 23:31'!
benchmark2		"BitBlt benchmark"
	"Run a benchmark on different combinations rules, source/destination depths and BitBlt modes. Note: This benchmark doesn't give you any 'absolute' value - it is intended only for benchmarking improvements in the bitblt code and nothing else.
	Attention: *this*may*take*a*while*"
	| bb source dest destRect log t |
	log := WriteStream on: String new.
	destRect := 0@0 extent: 600@600.
	"Form paint/Form over - the most common rules"
	#( 25 3 ) do:[:rule|
		Transcript cr; show:'---- Combination rule: ', rule printString,' ----'.
		log cr; nextPutAll:'---- Combination rule: ', rule printString,' ----'.
		#(1 2 4 8 16 32) do:[:destDepth|
			dest := nil.
			dest := Form extent: destRect extent depth: destDepth.
			Transcript cr.
			log cr.
			#(1 2 4 8 16 32) do:[:sourceDepth|
				Transcript cr; show: sourceDepth printString, ' => ', destDepth printString.
				log cr; nextPutAll: sourceDepth printString, ' => ', destDepth printString.
				source := nil. bb := nil.
				source := Form extent: destRect extent depth: sourceDepth.
				(source getCanvas) fillOval: dest boundingBox color: Color yellow borderWidth: 30 borderColor: Color black.
				bb := WarpBlt toForm: dest.
				bb sourceForm: source.
				bb sourceRect: source boundingBox.
				bb destRect: dest boundingBox.
				bb colorMap: (source colormapIfNeededFor: dest).
				bb combinationRule: rule.

				"Measure speed of copyBits"
				t := Time millisecondsToRun:[1 to: 10 do:[:i| bb copyBits]].
				Transcript tab; show: t printString.
				log tab; nextPutAll: t printString.

				bb sourceForm: source destRect: source boundingBox.

				"Measure speed of 1x1 warpBits"
				bb cellSize: 1.
				t := Time millisecondsToRun:[1 to: 4 do:[:i| bb warpBits]].
				Transcript tab; show: t printString.
				log tab; nextPutAll: t printString.

				"Measure speed of 2x2 warpBits"
				bb cellSize: 2.
				t := Time millisecondsToRun:[bb warpBits].
				Transcript tab; show: t printString.
				log tab; nextPutAll: t printString.

				"Measure speed of 3x3 warpBits"
				bb cellSize: 3.
				t := Time millisecondsToRun:[bb warpBits].
				Transcript tab; show: t printString.
				log tab; nextPutAll: t printString.
			].
		].
	].
	^log contents! !

!BitBlt class methodsFor: 'benchmarks' stamp: 'ar 4/26/2001 21:04'!
benchmark3		"BitBlt benchmark"
	"Run a benchmark on different combinations rules, source/destination depths and BitBlt modes. Note: This benchmark doesn't give you any 'absolute' value - it is intended only for benchmarking improvements in the bitblt code and nothing else.
	Attention: *this*may*take*a*while*"
	| bb source dest destRect log t |
	log := WriteStream on: String new.
	destRect := 0@0 extent: 600@600.
	"Form paint/Form over - the most common rules"
	#( 25 3 ) do:[:rule|
		Transcript cr; show:'---- Combination rule: ', rule printString,' ----'.
		log cr; nextPutAll:'---- Combination rule: ', rule printString,' ----'.
		#(1 2 4 8 16 32) do:[:destDepth|
			dest := nil.
			dest := Form extent: destRect extent depth: destDepth.
			Transcript cr.
			log cr.
			#(1 2 4 8 16 32) do:[:sourceDepth|
				Transcript cr; show: sourceDepth printString, ' => ', destDepth printString.
				log cr; nextPutAll: sourceDepth printString, ' => ', destDepth printString.
				source := nil. bb := nil.
				source := Form extent: destRect extent depth: sourceDepth.
				(source getCanvas) fillOval: dest boundingBox color: Color yellow borderWidth: 30 borderColor: Color black.
				bb := WarpBlt toForm: dest.
				bb sourceForm: source.
				bb sourceRect: source boundingBox.
				bb destRect: dest boundingBox.
				bb colorMap: (source colormapIfNeededFor: dest).
				bb combinationRule: rule.

				"Measure speed of copyBits"
				t := Time millisecondsToRun:[1 to: 10 do:[:i| bb copyBits]].
				Transcript tab; show: t printString.
				log tab; nextPutAll: t printString.

				bb sourceForm: source destRect: source boundingBox.

				"Measure speed of 1x1 warpBits"
				bb cellSize: 1.
				t := Time millisecondsToRun:[1 to: 4 do:[:i| bb warpBits]].
				Transcript tab; show: t printString.
				log tab; nextPutAll: t printString.

				"Measure speed of 2x2 warpBits"
				bb cellSize: 2.
				t := Time millisecondsToRun:[bb warpBits].
				Transcript tab; show: t printString.
				log tab; nextPutAll: t printString.

				"Measure speed of 3x3 warpBits"
				bb cellSize: 3.
				t := Time millisecondsToRun:[bb warpBits].
				Transcript tab; show: t printString.
				log tab; nextPutAll: t printString.
			].
		].
	].
	^log contents! !

!BitBlt class methodsFor: 'instance creation' stamp: 'di 3/2/98 12:53'!
bitPeekerFromForm: sourceForm
	"Answer an instance to be used extract individual pixels from the given Form. The destination for a 1x1 copyBits will be the low order bits of (bits at: 1)."
	| pixPerWord |
	pixPerWord := 32 // sourceForm depth.
	sourceForm unhibernate.
	^ self destForm: (Form extent: pixPerWord@1 depth: sourceForm depth)
	 	sourceForm: sourceForm
		halftoneForm: nil
		combinationRule: Form over
		destOrigin: (pixPerWord - 1)@0
		sourceOrigin: 0@0
		extent: 1@1
		clipRect: (0@0 extent: pixPerWord@1)
! !

!BitBlt class methodsFor: 'instance creation' stamp: 'di 3/2/98 12:53'!
bitPokerToForm: destForm
	"Answer an instance to be used for valueAt: aPoint put: pixValue.
	The source for a 1x1 copyBits will be the low order of (bits at: 1)"
	| pixPerWord |
	pixPerWord := 32//destForm depth.
	destForm unhibernate.
	^ self destForm: destForm
	 	sourceForm: (Form extent: pixPerWord@1 depth: destForm depth)
		halftoneForm: nil combinationRule: Form over
		destOrigin: 0@0 sourceOrigin: (pixPerWord-1)@0
		extent: 1@1 clipRect: (0@0 extent: destForm extent)
! !

!BitBlt class methodsFor: 'instance creation' stamp: 'ar 5/28/2000 12:00'!
current
	"Return the class currently to be used for BitBlt"
	^Display defaultBitBltClass! !

!BitBlt class methodsFor: 'instance creation' stamp: ''!
destForm: df sourceForm: sf fillColor: hf combinationRule: cr destOrigin: destOrigin sourceOrigin: sourceOrigin extent: extent clipRect: clipRect 
	"Answer an instance of me with values set according to the arguments."

	^ self new
		setDestForm: df
		sourceForm: sf
		fillColor: hf
		combinationRule: cr
		destOrigin: destOrigin
		sourceOrigin: sourceOrigin
		extent: extent
		clipRect: clipRect! !

!BitBlt class methodsFor: 'instance creation' stamp: ''!
destForm: df sourceForm: sf halftoneForm: hf combinationRule: cr destOrigin: destOrigin sourceOrigin: sourceOrigin extent: extent clipRect: clipRect 
	"Answer an instance of me with values set according to the arguments."

	^ self new
		setDestForm: df
		sourceForm: sf
		fillColor: hf
		combinationRule: cr
		destOrigin: destOrigin
		sourceOrigin: sourceOrigin
		extent: extent
		clipRect: clipRect! !

!BitBlt class methodsFor: 'private' stamp: 'jrm 2/21/2001 23:45'!
exampleAt: originPoint rule: rule fillColor: mask 
	"This builds a source and destination form and copies the source to the
	destination using the specifed rule and mask. It is called from the method
	named exampleOne. Only works with Display depth of 1"

	| s d border aBitBlt | 
	border:=Form extent: 32@32.
	border fillBlack.
	border fill: (1@1 extent: 30@30) fillColor: Color white.
	s := Form extent: 32@32.
	s fillWhite.
	s fillBlack: (7@7 corner: 25@25).
	d := Form extent: 32@32.
	d fillWhite.
	d fillBlack: (0@0 corner: 32@16).

	s displayOn: Display at: originPoint.
	border displayOn: Display at: originPoint rule: Form under.
	d displayOn: Display at: originPoint + (s width @0).
	border displayOn: Display at: originPoint + (s width @0) rule: Form under.

	d displayOn: Display at: originPoint + (s extent // (2 @ 1)). 
	aBitBlt := BitBlt
		destForm: Display
		sourceForm: s
		fillColor: mask
		combinationRule: rule
		destOrigin: originPoint + (s extent // (2 @ 1))
		sourceOrigin: 0 @ 0
		extent: s extent
		clipRect: Display computeBoundingBox.
	aBitBlt copyBits.
	border 
		displayOn: Display at: originPoint + (s extent // (2 @ 1))
		rule: Form under.

	"BitBlt exampleAt: 100@100 rule: 0 fillColor: nil"  ! !

!BitBlt class methodsFor: 'examples' stamp: 'ar 5/4/2001 16:02'!
exampleColorMap	"BitBlt exampleColorMap"
	"This example shows what one can do with the fixed part of a color map. The color map, as setup below, rotates the bits of a pixel all the way around. Thus you'll get a (sometime strange looking ;-) animation of colors which will end up exactly the way it looked at the beginning. The example is given to make you understand that the masks and shifts can be used for a lot more than simply color converting pixels. In this example, for instance, we use only two of the four independent shifters."
	| cc bb |
	cc := ColorMap masks: {
		1 << (Display depth-1). "mask out high bit of color component"
		1 << (Display depth-1) - 1. "mask all other bits"
		0.
		0}
		shifts: {
			1 - Display depth. "shift right to bottom most position"
			1. "shift all other pixels one bit left"
			0.
			0}.
	bb := BitBlt toForm: Display.
	bb 
		sourceForm: Display;
		combinationRule: 3;
		colorMap: cc.
	1 to: Display depth do:[:i|
		bb copyBits.
		Display forceDisplayUpdate.
	].
! !

!BitBlt class methodsFor: 'examples' stamp: 'dew 9/18/2001 02:30'!
exampleOne
	"This tests BitBlt by displaying the result of all sixteen combination rules that BitBlt is capable of using. (Please see the comment in BitBlt for the meaning of the combination rules). This only works at Display depth of 1. (Rule 15 does not work?)"
	| path displayDepth |

	displayDepth := Display depth.
	Display newDepth: 1.

	path := Path new.
	0 to: 3 do: [:i | 0 to: 3 do: [:j | path add: j * 100 @ (i * 75)]].
	Display fillWhite.
	path := path translateBy: 60 @ 40.
	1 to: 16 do: [:index | BitBlt
			exampleAt: (path at: index)
			rule: index - 1
			fillColor: nil].

	[Sensor anyButtonPressed] whileFalse: [].
	Display newDepth: displayDepth.

	"BitBlt exampleOne"! !

!BitBlt class methodsFor: 'examples' stamp: 'jrm 2/21/2001 23:43'!
exampleTwo
	"This is to test painting with a gray tone. It also tests that the seaming with gray patterns is correct in the microcode. Lets you paint for a while and then automatically stops. This only works at Depth of 1."
	| f aBitBlt displayDepth |
	"create a small black Form source as a brush. "
	displayDepth := Display depth.
	Display newDepth: 1.
	f := Form extent: 20 @ 20.
	f fillBlack.
	"create a BitBlt which will OR gray into the display. "
	aBitBlt := BitBlt
		destForm: Display
		sourceForm: f
		fillColor: Color gray
		combinationRule: Form over
		destOrigin: Sensor cursorPoint
		sourceOrigin: 0 @ 0
		extent: f extent
		clipRect: Display computeBoundingBox.
	"paint the gray Form on the screen for a while. "
	[Sensor anyButtonPressed] whileFalse: 
		[aBitBlt destOrigin: Sensor cursorPoint.
		aBitBlt copyBits].
	Display newDepth: displayDepth.
	"BitBlt exampleTwo"! !

!BitBlt class methodsFor: 'private' stamp: 'jmv 8/2/2009 21:37'!
recreateColorMaps
	CachedFontColorMaps _ ColorConvertingMaps _ nil! !

!BitBlt class methodsFor: 'instance creation' stamp: ''!
toForm: aForm
	^ self new setDestForm: aForm! !

!BitBlt methodsFor: 'copying' stamp: 'jmv 8/5/2009 22:09'!
basicDisplayString: aString from: startIndex to: stopIndex at: aPoint strikeFont: font kern: kernDelta

	destY := aPoint y.
	destX := aPoint x.

	"the following are not really needed, but theBitBlt primitive will fail if not set"
	sourceX ifNil: [sourceX := 100].
	width ifNil: [width := 100].

	self primDisplayString: aString from: startIndex to: stopIndex
			map: font characterToGlyphMap xTable: font xTable
			kern: kernDelta.
	^ destX@destY.
! !

!BitBlt methodsFor: 'private' stamp: 'hg 6/27/2000 12:27'!
cachedFontColormapFrom: sourceDepth to: destDepth

	| srcIndex map |
	CachedFontColorMaps class == Array 
		ifFalse: [CachedFontColorMaps := (1 to: 9) collect: [:i | Array new: 32]].
	srcIndex := sourceDepth.
	sourceDepth > 8 ifTrue: [srcIndex := 9].
	(map := (CachedFontColorMaps at: srcIndex) at: destDepth) ~~ nil ifTrue: [^ map].

	map := (Color cachedColormapFrom: sourceDepth to: destDepth) copy.
	(CachedFontColorMaps at: srcIndex) at: destDepth put: map.
	^ map
! !

!BitBlt methodsFor: 'accessing' stamp: 'ar 12/30/2001 20:31'!
clipBy: aRectangle
	| aPoint right bottom |
	right := clipX + clipWidth.
	bottom := clipY + clipHeight.
	aPoint := aRectangle origin.
	aPoint x > clipX ifTrue:[clipX := aPoint x].
	aPoint y > clipY ifTrue:[clipY := aPoint y].
	aPoint := aRectangle corner.
	aPoint x < right ifTrue:[right := aPoint x].
	aPoint y < bottom ifTrue:[bottom := aPoint y].
	clipWidth := right - clipX.
	clipHeight := bottom - clipY.
	clipWidth < 0 ifTrue:[clipWidth := 0].
	clipHeight < 0 ifTrue:[clipHeight := 0].! !

!BitBlt methodsFor: 'accessing' stamp: 'ar 12/30/2001 20:33'!
clipByX1: x1 y1: y1 x2: x2 y2: y2
	| right bottom |
	right := clipX + clipWidth.
	bottom := clipY + clipHeight.
	x1 > clipX ifTrue:[clipX := x1].
	y1 > clipY ifTrue:[clipY := y1].
	x2 < right ifTrue:[right := x2].
	y2 < bottom ifTrue:[bottom := y2].
	clipWidth := right - clipX.
	clipHeight := bottom - clipY.
	clipWidth < 0 ifTrue:[clipWidth := 0].
	clipHeight < 0 ifTrue:[clipHeight := 0].! !

!BitBlt methodsFor: 'accessing' stamp: 'ar 5/17/2000 18:58'!
clipHeight
	^clipHeight! !

!BitBlt methodsFor: 'accessing' stamp: ''!
clipHeight: anInteger 
	"Set the receiver's clipping area height to be the argument, anInteger."

	clipHeight := anInteger! !

!BitBlt methodsFor: 'private' stamp: 'ar 3/8/2003 00:34'!
clipRange
	"clip and adjust source origin and extent appropriately"
	"first in x"
	| sx sy dx dy bbW bbH |
	"fill in the lazy state if needed"
	destX ifNil:[destX := 0].
	destY ifNil:[destY := 0].
	width ifNil:[width := destForm width].
	height ifNil:[height := destForm height].
	sourceX ifNil:[sourceX := 0].
	sourceY ifNil:[sourceY := 0].
	clipX ifNil:[clipX := 0].
	clipY ifNil:[clipY := 0].
	clipWidth ifNil:[clipWidth := destForm width].
	clipHeight ifNil:[clipHeight := destForm height].

	destX >= clipX
		ifTrue: [sx := sourceX.
				dx := destX.
				bbW := width]
		ifFalse: [sx := sourceX + (clipX - destX).
				bbW := width - (clipX - destX).
				dx := clipX].
	(dx + bbW) > (clipX + clipWidth)
		ifTrue: [bbW := bbW - ((dx + bbW) - (clipX + clipWidth))].
	"then in y"
	destY >= clipY
		ifTrue: [sy := sourceY.
				dy := destY.
				bbH := height]
		ifFalse: [sy := sourceY + clipY - destY.
				bbH := height - (clipY - destY).
				dy := clipY].
	(dy + bbH) > (clipY + clipHeight)
		ifTrue: [bbH := bbH - ((dy + bbH) - (clipY + clipHeight))].
	sourceForm ifNotNil:[
		sx < 0
			ifTrue: [dx := dx - sx.
					bbW := bbW + sx.
					sx := 0].
		sx + bbW > sourceForm width
			ifTrue: [bbW := bbW - (sx + bbW - sourceForm width)].
		sy < 0
			ifTrue: [dy := dy - sy.
					bbH := bbH + sy.
					sy := 0].
		sy + bbH > sourceForm height
			ifTrue: [bbH := bbH - (sy + bbH - sourceForm height)].
	].
	(bbW <= 0 or:[bbH <= 0]) ifTrue:[
		sourceX := sourceY := destX := destY := clipX := clipY := width := height := 0.
		^true].
	(sx = sourceX 
		and:[sy = sourceY 
		and:[dx = destX 
		and:[dy = destY 
		and:[bbW = width 
		and:[bbH = height]]]]]) ifTrue:[^false].
	sourceX := sx.
	sourceY := sy.
	destX := dx.
	destY := dy.
	width := bbW.
	height := bbH.
	^true! !

!BitBlt methodsFor: 'accessing' stamp: ''!
clipRect
	"Answer the receiver's clipping area rectangle."

	^clipX @ clipY extent: clipWidth @ clipHeight! !

!BitBlt methodsFor: 'accessing' stamp: 'ar 10/4/2000 16:37'!
clipRect: aRectangle 
	"Set the receiver's clipping area rectangle to be the argument, aRectangle."

	clipX := aRectangle left truncated.
	clipY := aRectangle top truncated.
	clipWidth := aRectangle right truncated - clipX.
	clipHeight := aRectangle bottom truncated - clipY.! !

!BitBlt methodsFor: 'accessing' stamp: 'ar 5/17/2000 18:58'!
clipWidth
	^clipWidth! !

!BitBlt methodsFor: 'accessing' stamp: ''!
clipWidth: anInteger 
	"Set the receiver's clipping area width to be the argument, anInteger."

	clipWidth := anInteger! !

!BitBlt methodsFor: 'accessing' stamp: 'ar 5/17/2000 18:58'!
clipX
	^clipX! !

!BitBlt methodsFor: 'accessing' stamp: ''!
clipX: anInteger 
	"Set the receiver's clipping area top left x coordinate to be the argument, 
	anInteger."

	clipX := anInteger! !

!BitBlt methodsFor: 'accessing' stamp: 'ar 5/17/2000 18:58'!
clipY
	^clipY! !

!BitBlt methodsFor: 'accessing' stamp: ''!
clipY: anInteger 
	"Set the receiver's clipping area top left y coordinate to be the argument, 
	anInteger."

	clipY := anInteger! !

!BitBlt methodsFor: 'accessing' stamp: 'tk 8/15/2001 10:56'!
color
	"Return the current fill color as a Color.  
	 Gives the wrong answer if the halftoneForm is a complex pattern of more than one word."

	halftoneForm ifNil: [^ Color black].
	^ Color colorFromPixelValue: halftoneForm first depth: destForm depth! !

!BitBlt methodsFor: 'private' stamp: 'jmv 8/2/2009 21:36'!
colorConvertingMap: targetColor from: sourceDepth to: destDepth keepSubPixelAA: keepSubPix

	| srcIndex map mapsForSource mapsForSourceAndDest |
	ColorConvertingMaps class == Array 
		ifFalse: [ColorConvertingMaps _ (1 to: 10) collect: [:i | Array new: 32]].
		
	srcIndex _ sourceDepth.
	sourceDepth > 8 ifTrue: [ srcIndex _ keepSubPix ifTrue: [9] ifFalse: [10] ].
	mapsForSource _ ColorConvertingMaps at: srcIndex.
	(mapsForSourceAndDest _ mapsForSource at: destDepth) isNil ifTrue: [
		mapsForSourceAndDest _ mapsForSource at: destDepth put: Dictionary new ].
	
	map _ mapsForSourceAndDest at: targetColor ifAbsentPut: [
		Color computeColorConvertingMap: targetColor from: sourceDepth to: destDepth keepSubPixelAA: keepSubPix ].

	^ map! !

!BitBlt methodsFor: 'accessing' stamp: ''!
colorMap
	^ colorMap! !

!BitBlt methodsFor: 'accessing' stamp: 'ar 5/4/2001 15:45'!
colorMap: map
	"See last part of BitBlt comment. 6/18/96 tk"
	colorMap := map.! !

!BitBlt methodsFor: 'accessing' stamp: ''!
combinationRule: anInteger 
	"Set the receiver's combination rule to be the argument, anInteger, a 
	number in the range 0-15."

	combinationRule := anInteger! !

!BitBlt methodsFor: 'copying' stamp: ''!
copy: destRectangle from: sourcePt in: srcForm
	| destOrigin |
	sourceForm := srcForm.
	halftoneForm := nil.
	combinationRule := 3.  "store"
	destOrigin := destRectangle origin.
	destX := destOrigin x.
	destY := destOrigin y.
	sourceX := sourcePt x.
	sourceY := sourcePt y.
	width := destRectangle width.
	height := destRectangle height.
	self copyBits! !

!BitBlt methodsFor: 'copying' stamp: 'ar 5/14/2001 23:32'!
copy: destRectangle from: sourcePt in: srcForm fillColor: hf rule: rule
	"Specify a Color to fill, not a Form. 6/18/96 tk"  
	| destOrigin |
	sourceForm := srcForm.
	self fillColor: hf.	"sets halftoneForm"
	combinationRule := rule.
	destOrigin := destRectangle origin.
	destX := destOrigin x.
	destY := destOrigin y.
	sourceX := sourcePt x.
	sourceY := sourcePt y.
	width := destRectangle width.
	height := destRectangle height.
	srcForm == nil ifFalse:
		[colorMap := srcForm colormapIfNeededFor: destForm].
	^ self copyBits! !

!BitBlt methodsFor: 'copying' stamp: ''!
copy: destRectangle from: sourcePt in: srcForm halftoneForm: hf rule: rule 
	| destOrigin |
	sourceForm := srcForm.
	self fillColor: hf.		"sets halftoneForm"
	combinationRule := rule.
	destOrigin := destRectangle origin.
	destX := destOrigin x.
	destY := destOrigin y.
	sourceX := sourcePt x.
	sourceY := sourcePt y.
	width := destRectangle width.
	height := destRectangle height.
	self copyBits! !

!BitBlt methodsFor: 'copying' stamp: 'yo 3/15/2005 09:54'!
copyBits
	"Primitive. Perform the movement of bits from the source form to the 
	destination form. Fail if any variables are not of the right type (Integer, 
	Float, or Form) or if the combination rule is not implemented. 
	In addition to the original 16 combination rules, this BitBlt supports
	16	fail (to simulate paint)
	17	fail (to simulate mask)
	18	sourceWord + destinationWord
	19	sourceWord - destinationWord
	20	rgbAdd: sourceWord with: destinationWord
	21	rgbSub: sourceWord with: destinationWord
	22	rgbDiff: sourceWord with: destinationWord
	23	tallyIntoMap: destinationWord
	24	alphaBlend: sourceWord with: destinationWord
	25	pixPaint: sourceWord with: destinationWord
	26	pixMask: sourceWord with: destinationWord
	27	rgbMax: sourceWord with: destinationWord
	28	rgbMin: sourceWord with: destinationWord
	29	rgbMin: sourceWord bitInvert32 with: destinationWord
"
	<primitive: 'primitiveCopyBits' module: 'BitBltPlugin'>

	"Check for compressed source, destination or halftone forms"
	(combinationRule >= 30 and: [combinationRule <= 31]) ifTrue:
		["No alpha specified -- re-run with alpha = 1.0"
		^ self copyBitsTranslucent: 255].
	((sourceForm isForm) and: [sourceForm unhibernate])
		ifTrue: [^ self copyBits].
	((destForm isForm) and: [destForm unhibernate])
		ifTrue: [^ self copyBits].
	((halftoneForm isForm) and: [halftoneForm unhibernate])
		ifTrue: [^ self copyBits].

	"Check for unimplmented rules"
	combinationRule = Form oldPaint ifTrue: [^ self paintBits].
	combinationRule = Form oldErase1bitShape ifTrue: [^ self eraseBits].

	"Check if BitBlt doesn't support full color maps"
	(colorMap notNil and:[colorMap isColormap]) ifTrue:[
		colorMap := colorMap colors.
		^self copyBits].
	"Check if clipping gots us way out of range"
	self clipRange ifTrue:[self roundVariables. ^self copyBitsAgain].

	self error: 'Bad BitBlt arg (Fraction?); proceed to convert.'.
	"Convert all numeric parameters to integers and try again."
	self roundVariables.
	^ self copyBitsAgain! !

!BitBlt methodsFor: 'private' stamp: 'ar 2/2/2001 15:09'!
copyBitsAgain
	"Primitive. See BitBlt|copyBits, also a Primitive. Essential. See Object
	documentation whatIsAPrimitive."

	<primitive: 'primitiveCopyBits' module: 'BitBltPlugin'>
	self primitiveFailed! !

!BitBlt methodsFor: 'private' stamp: 'ar 10/25/1998 17:30'!
copyBitsFrom: x0 to: x1 at: y
	destX := x0.
	destY := y.
	sourceX := x0.
	width := (x1 - x0).
	self copyBits.! !

!BitBlt methodsFor: 'copying' stamp: 'ar 2/13/2001 21:12'!
copyBitsSimulated
	^Smalltalk at: #BitBltSimulation ifPresent:[:bb| bb copyBitsFrom: self].! !

!BitBlt methodsFor: 'copying' stamp: 'nk 4/17/2004 19:42'!
copyBitsTranslucent: factor
	"This entry point to BitBlt supplies an extra argument to specify translucency
	for operations 30 and 31.  The argument must be an integer between 0 and 255."

	<primitive: 'primitiveCopyBits' module: 'BitBltPlugin'>

	"Check for compressed source, destination or halftone forms"
	((sourceForm isForm) and: [sourceForm unhibernate])
		ifTrue: [^ self copyBitsTranslucent: factor].
	((destForm isForm) and: [destForm unhibernate])
		ifTrue: [^ self copyBitsTranslucent: factor].
	((halftoneForm isForm) and: [halftoneForm unhibernate])
		ifTrue: [^ self copyBitsTranslucent: factor].

	self primitiveFailed  "Later do nicer error recovery -- share copyBits recovery"! !

!BitBlt methodsFor: 'copying' stamp: 'ar 5/14/2001 23:32'!
copyForm: srcForm to: destPt rule: rule
	^ self copyForm: srcForm to: destPt rule: rule
		colorMap: (srcForm colormapIfNeededFor: destForm)! !

!BitBlt methodsFor: 'copying' stamp: ''!
copyForm: srcForm to: destPt rule: rule color: color
	sourceForm := srcForm.
	halftoneForm := color.
	combinationRule := rule.
	destX := destPt x + sourceForm offset x.
	destY := destPt y + sourceForm offset y.
	sourceX := 0.
	sourceY := 0.
	width := sourceForm width.
	height := sourceForm height.
	self copyBits! !

!BitBlt methodsFor: 'copying' stamp: 'di 7/17/97 10:04'!
copyForm: srcForm to: destPt rule: rule colorMap: map
	sourceForm := srcForm.
	halftoneForm := nil.
	combinationRule := rule.
	destX := destPt x + sourceForm offset x.
	destY := destPt y + sourceForm offset y.
	sourceX := 0.
	sourceY := 0.
	width := sourceForm width.
	height := sourceForm height.
	colorMap := map.
	self copyBits! !

!BitBlt methodsFor: 'copying' stamp: ''!
copyForm: srcForm to: destPt rule: rule fillColor: color
	sourceForm := srcForm.
	self fillColor: color.	"sets halftoneForm"
	combinationRule := rule.
	destX := destPt x + sourceForm offset x.
	destY := destPt y + sourceForm offset y.
	sourceX := 0.
	sourceY := 0.
	width := sourceForm width.
	height := sourceForm height.
	self copyBits! !

!BitBlt methodsFor: 'copying' stamp: 'ar 5/14/2001 23:32'!
copyFrom: sourceRectangle in: srcForm to: destPt
	| sourceOrigin |
	sourceForm := srcForm.
	halftoneForm := nil.
	combinationRule := 3.  "store"
	destX := destPt x.
	destY := destPt y.
	sourceOrigin := sourceRectangle origin.
	sourceX := sourceOrigin x.
	sourceY := sourceOrigin y.
	width := sourceRectangle width.
	height := sourceRectangle height.
	colorMap := srcForm colormapIfNeededFor: destForm.
	self copyBits! !

!BitBlt methodsFor: 'accessing' stamp: ''!
destForm
	^ destForm! !

!BitBlt methodsFor: 'accessing' stamp: ''!
destOrigin: aPoint 
	"Set the receiver's destination top left coordinates to be those of the 
	argument, aPoint."

	destX := aPoint x.
	destY := aPoint y! !

!BitBlt methodsFor: 'accessing' stamp: 'tk 3/19/97'!
destRect
	"The rectangle we are about to blit to or just blitted to.  "

	^ destX @ destY extent: width @ height! !

!BitBlt methodsFor: 'accessing' stamp: ''!
destRect: aRectangle 
	"Set the receiver's destination form top left coordinates to be the origin of 
	the argument, aRectangle, and set the width and height of the receiver's 
	destination form to be the width and height of aRectangle."

	destX := aRectangle left.
	destY := aRectangle top.
	width := aRectangle width.
	height := aRectangle height! !

!BitBlt methodsFor: 'accessing' stamp: ''!
destX: anInteger 
	"Set the top left x coordinate of the receiver's destination form to be the 
	argument, anInteger."

	destX := anInteger! !

!BitBlt methodsFor: 'accessing' stamp: ''!
destX: x destY: y width: w height: h
	"Combined init message saves 3 sends from DisplayScanner"
	destX := x.
	destY := y.
	width := w.
	height := h.! !

!BitBlt methodsFor: 'accessing' stamp: ''!
destY: anInteger 
	"Set the top left y coordinate of the receiver's destination form to be the 
	argument, anInteger."

	destY := anInteger! !

!BitBlt methodsFor: 'copying' stamp: 'ar 8/14/2009 20:26'!
displayGlyph: aForm at: aPoint left: leftX right: rightX font: aFont
	"Display a glyph in a multi-lingual font. Do 2 pass rendering if necessary.
	This happens when #installStrikeFont:foregroundColor:backgroundColor: sets rule 37 (rgbMul).
	the desired effect is to do two bitblt calls. The first one is with rule 37 and special colormap.
	The second one is rule 34, with a colormap for applying the requested foreground color.
	This two together do component alpha blending, i.e. alpha blend red, green and blue separatedly.
	This is needed for arbitrary color over abitrary background text with subpixel AA."

	| prevRule secondPassMap |
	self sourceForm: aForm.
	destX := aPoint x.
	destY := aPoint y.
	sourceX := leftX.
	sourceY := 0.
	width := rightX - leftX.
	height := aFont height.
	combinationRule = 37 ifTrue:[
		"We need to do a second pass. The colormap set is for use in the second pass."
		secondPassMap := colorMap.
		colorMap := sourceForm depth = destForm depth
			ifFalse: [ self cachedFontColormapFrom: sourceForm depth to: destForm depth ].
		self copyBits.
		prevRule := combinationRule.
		combinationRule := 20. "rgbAdd"
		colorMap := secondPassMap.
		self copyBits.
		combinationRule := prevRule.
	] ifFalse:[self copyBits].! !

!BitBlt methodsFor: 'text display' stamp: 'ar 10/24/2005 21:49'!
displayString: aString from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: baselineY font: aFont
	"Double dispatch into the font. This method is present so that other-than-bitblt entities can be used by CharacterScanner and friends to display text."
	^ aFont displayString: aString on: self from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: baselineY! !

!BitBlt methodsFor: 'text display' stamp: 'ar 10/25/2005 01:12'!
displayString: aString from: startIndex to: stopIndex at: aPoint kern: kernDelta font: aFont
	"Double dispatch into the font. This method is present so that other-than-bitblt entities can be used by CharacterScanner and friends to display text."
	^ aFont displayString: aString on: self from: startIndex to: stopIndex at: aPoint kern: kernDelta! !

!BitBlt methodsFor: 'copying' stamp: 'jmv 8/5/2009 22:23'!
displayString: aString from: startIndex to: stopIndex at: aPoint strikeFont: font kern: kernDelta
	"If required, do a second pass with new rule and colorMap.
	This happens when #installStrikeFont:foregroundColor:backgroundColor: sets rule 37 (rgbMul).
	the desired effect is to do two bitblt calls. The first one is with rule 37 and special colormap.
	The second one is rule 34, with a colormap for applying the requested foreground color.
	This two together do component alpha blending, i.e. alpha blend red, green and blue separatedly.
	This is needed for arbitrary color over abitrary background text with subpixel AA."

	| answer prevRule secondPassMap |
	"If combinationRule is rgbMul, we might need the special two-pass technique for component alpha blending.
	If not, do it simply"
	combinationRule = 37 "rgbMul" ifFalse: [
		^self basicDisplayString: aString from: startIndex to: stopIndex at: aPoint strikeFont: font kern: kernDelta ].
	
	"We need to do a second pass. The colormap set is for use in the second pass."
	secondPassMap := colorMap.
	colorMap := sourceForm depth ~= destForm depth
		ifTrue: [ self cachedFontColormapFrom: sourceForm depth to: destForm depth ].
	answer := self basicDisplayString: aString from: startIndex to: stopIndex at: aPoint strikeFont: font kern: kernDelta.
	secondPassMap ifNotNil: [
		prevRule := combinationRule.
		combinationRule := 20. "rgbAdd"
		colorMap := secondPassMap.
		self basicDisplayString: aString from: startIndex to: stopIndex at: aPoint strikeFont: font kern: kernDelta.
		combinationRule := prevRule ].
	^answer! !

!BitBlt methodsFor: 'line drawing' stamp: ''!
drawFrom: startPoint to: stopPoint 
	
	 ^ self drawFrom: startPoint to: stopPoint withFirstPoint: true! !

!BitBlt methodsFor: 'line drawing' stamp: '6/8/97 15:41 di'!
drawFrom: startPoint to: stopPoint withFirstPoint: drawFirstPoint
	"Draw a line whose end points are startPoint and stopPoint.
	The line is formed by repeatedly calling copyBits at every
	point along the line.  If drawFirstPoint is false, then omit
	the first point so as not to overstrike at line junctions."
	| offset point1 point2 forwards |
	"Always draw down, or at least left-to-right"
	forwards := (startPoint y = stopPoint y and: [startPoint x < stopPoint x])
				or: [startPoint y < stopPoint y].
	forwards
		ifTrue: [point1 := startPoint. point2 := stopPoint]
		ifFalse: [point1 := stopPoint. point2 := startPoint].
	sourceForm == nil ifTrue:
		[destX := point1 x.
		destY := point1 y]
		ifFalse:
		[width := sourceForm width.
		height := sourceForm height.
		offset := sourceForm offset.
		destX := (point1 x + offset x) rounded.
		destY := (point1 y + offset y) rounded].

	"Note that if not forwards, then the first point is the last and vice versa.
	We agree to always paint stopPoint, and to optionally paint startPoint."
	(drawFirstPoint or: [forwards == false  "ie this is stopPoint"])
		ifTrue: [self copyBits].
	self drawLoopX: (point2 x - point1 x) rounded 
				  Y: (point2 y - point1 y) rounded.
	(drawFirstPoint or: [forwards  "ie this is stopPoint"])
		ifTrue: [self copyBits].
! !

!BitBlt methodsFor: 'line drawing' stamp: 'ar 2/2/2001 15:09'!
drawLoopX: xDelta Y: yDelta 
	"Primitive. Implements the Bresenham plotting algorithm (IBM Systems
	Journal, Vol. 4 No. 1, 1965). It chooses a principal direction, and
	maintains a potential, P. When P's sign changes, it is time to move in
	the minor direction as well. This particular version does not write the
	first and last points, so that these can be called for as needed in client code.
	Optional. See Object documentation whatIsAPrimitive."
	| dx dy px py P |
	<primitive: 'primitiveDrawLoop' module: 'BitBltPlugin'>
	dx := xDelta sign.
	dy := yDelta sign.
	px := yDelta abs.
	py := xDelta abs.
	"self copyBits."
	py > px
		ifTrue: 
			["more horizontal"
			P := py // 2.
			1 to: py do: 
				[:i |
				destX := destX + dx.
				(P := P - px) < 0 ifTrue: 
						[destY := destY + dy.
						P := P + py].
				i < py ifTrue: [self copyBits]]]
		ifFalse: 
			["more vertical"
			P := px // 2.
			1 to: px do:
				[:i |
				destY := destY + dy.
				(P := P - py) < 0 ifTrue: 
						[destX := destX + dx.
						P := P + px].
				i < px ifTrue: [self copyBits]]]! !

!BitBlt methodsFor: 'private' stamp: ''!
eraseBits
	"Perform the erase operation, which puts 0's in the destination
	wherever the source (which is assumed to be just 1 bit deep)
	has a 1.  This requires the colorMap to be set in order to AND
	all 1's into the destFrom pixels regardless of their size."
	| oldMask oldMap |
	oldMask := halftoneForm.
	halftoneForm := nil.
	oldMap := colorMap.
	self colorMap: (Bitmap with: 0 with: 16rFFFFFFFF).
	combinationRule := Form erase.
	self copyBits. 		"Erase the dest wherever the source is 1"
	halftoneForm := oldMask.	"already converted to a Bitmap"
	colorMap := oldMap! !

!BitBlt methodsFor: 'copying' stamp: ''!
fill: destRect fillColor: grayForm rule: rule
	"Fill with a Color, not a Form. 6/18/96 tk"
	sourceForm := nil.
	self fillColor: grayForm.		"sets halftoneForm"
	combinationRule := rule.
	destX := destRect left.
	destY := destRect top.
	sourceX := 0.
	sourceY := 0.
	width := destRect width.
	height := destRect height.
	self copyBits! !

!BitBlt methodsFor: 'accessing' stamp: ''!
fillColor
	^ halftoneForm! !

!BitBlt methodsFor: 'accessing' stamp: 'ar 5/14/2001 23:25'!
fillColor: aColorOrPattern 
	"The destForm will be filled with this color or pattern of colors.  May be an old Color, a new type Color, a Bitmap (see BitBlt comment), a Pattern, or a Form.  6/18/96 tk"

	aColorOrPattern == nil ifTrue: [halftoneForm := nil. ^ self].
	destForm == nil ifTrue: [self error: 'Must set destForm first'].
	halftoneForm := destForm bitPatternFor: aColorOrPattern ! !

!BitBlt methodsFor: 'private' stamp: 'ar 5/26/2000 16:38'!
getPluginName
	"Private. Return the name of the plugin representing BitBlt.
	Used for dynamically switching between different BB representations only."
	^'BitBltPlugin'! !

!BitBlt methodsFor: 'accessing' stamp: 'tbn 9/14/2004 20:38'!
halftoneForm
	"Returns the receivers half tone form. See class commment."
	
	^halftoneForm! !

!BitBlt methodsFor: 'accessing' stamp: 'tbn 9/14/2004 20:39'!
halftoneForm: aBitmap
	"Sets the receivers half tone form. See class commment."
	
	halftoneForm := aBitmap
	
 ! !

!BitBlt methodsFor: 'accessing' stamp: ''!
height: anInteger 
	"Set the receiver's destination form height to be the argument, anInteger."

	height := anInteger! !

!BitBlt methodsFor: 'text display' stamp: 'ar 10/24/2005 21:48'!
installFont: aFont foregroundColor: foregroundColor backgroundColor: backgroundColor
	"Double dispatch into the font. This method is present so that other-than-bitblt entities can be used by CharacterScanner and friends to display text."
	^aFont installOn: self foregroundColor: foregroundColor backgroundColor: backgroundColor! !

!BitBlt methodsFor: 'private' stamp: 'jmv 8/5/2009 22:26'!
installStrikeFont: aStrikeFont foregroundColor: foregroundColor backgroundColor: backgroundColor
	| lastSourceDepth targetColor |
	sourceForm ifNotNil:[lastSourceDepth := sourceForm depth].
	sourceForm := aStrikeFont glyphs.

	"Ignore any halftone pattern since we use a color map approach here"
	halftoneForm := nil.
	sourceY := 0.
	height := aStrikeFont height.

	sourceForm depth = 1 ifTrue: [
		self combinationRule: Form paint.
		(colorMap notNil and:[lastSourceDepth = sourceForm depth]) ifFalse: [
			"Set up color map for a different source depth (color font)"
			"Uses caching for reasonable efficiency"
			colorMap := self cachedFontColormapFrom: sourceForm depth to: destForm depth.
			colorMap at: 1 put: (destForm pixelValueFor: backgroundColor)].
		colorMap at: 2 put: (destForm pixelValueFor: foregroundColor).
	]
	ifFalse: [
		(Preferences subPixelRenderFonts and: [ foregroundColor = Color black or: [ Preferences subPixelRenderColorFonts ]]) ifTrue: [
			destForm depth > 8 ifTrue: [
				"rgbMul is equivalent to component alpha blend if text is black (only faster, hehe)"
				self combinationRule: 37.		"RGBMul"
				(Preferences properAlphaForBlackText or: [ (foregroundColor = Color black) not ]) ifTrue: [
					"rgbMul / rgbAdd IS component alpha blend for any color of text (neat trick, eh!!)"
					"This colorMap is to be used on the second pass with rule 20 (rgbAdd)
					See #displayString:from:to:at:strikeFont:kern:"
					colorMap := self colorConvertingMap: foregroundColor from: sourceForm depth to: destForm depth keepSubPixelAA: true]]
			ifFalse: [
				self combinationRule: 25.		"Paint"
				targetColor := foregroundColor = Color black ifFalse: [ foregroundColor ].
				colorMap := self colorConvertingMap: targetColor from: sourceForm depth to: destForm depth keepSubPixelAA: true]
		]
		ifFalse: [
			"Do not use rule 34 for 16bpp display. TTCFont uses it, but it builds a glyphs cache for each color used!!"
			self combinationRule: (destForm depth = 32 ifTrue: [34 "alphaBlendScaled"] ifFalse: [25 "Paint"]).
			colorMap := self colorConvertingMap: foregroundColor from: sourceForm depth to: destForm depth keepSubPixelAA: false
		]
	].! !

!BitBlt methodsFor: 'private' stamp: 'yo 1/8/2005 09:12'!
installTTCFont: aTTCFont foregroundColor: foregroundColor backgroundColor: backgroundColor
	"Set up the parameters.  Since the glyphs in a TTCFont is 32bit depth form, it tries to use rule=34 to get better AA result if possible."

	((aTTCFont depth = 32)) ifTrue: [
		destForm depth <= 8 ifTrue: [
			self colorMap: (self cachedFontColormapFrom: aTTCFont depth to: destForm depth).
			self combinationRule: Form paint.
		] ifFalse: [
			self colorMap: nil.
			self combinationRule: 34.
		].
		halftoneForm := nil.
		sourceY := 0.
		height := aTTCFont height.
	].
! !

!BitBlt methodsFor: 'private' stamp: ''!
paintBits
	"Perform the paint operation, which requires two calls to BitBlt."
	| color oldMap saveRule |
	sourceForm depth = 1 ifFalse: 
		[^ self halt: 'paint operation is only defined for 1-bit deep sourceForms'].
	saveRule := combinationRule.
	color := halftoneForm.  halftoneForm := nil.
	oldMap := colorMap.
	"Map 1's to ALL ones, not just one"
	self colorMap: (Bitmap with: 0 with: 16rFFFFFFFF).
	combinationRule := Form erase.
	self copyBits. 		"Erase the dest wherever the source is 1"
	halftoneForm := color.
	combinationRule := Form under.
	self copyBits.	"then OR, with whatever color, into the hole"
	colorMap := oldMap.
	combinationRule := saveRule

" | dot |
dot := Form dotOfSize: 32.
((BitBlt destForm: Display
		sourceForm: dot
		fillColor: Color lightGray
		combinationRule: Form paint
		destOrigin: Sensor cursorPoint
		sourceOrigin: 0@0
		extent: dot extent
		clipRect: Display boundingBox)
		colorMap: (Bitmap with: 0 with: 16rFFFFFFFF)) copyBits"! !

!BitBlt methodsFor: 'copying' stamp: 'ar 3/1/2004 13:49'!
pixelAt: aPoint
	"Assumes this BitBlt has been set up specially (see the init message,
	BitBlt bitPeekerFromForm:.  Returns the pixel at aPoint."
	sourceX := aPoint x.
	sourceY := aPoint y.
	destForm unhibernate. "before poking"
	destForm bits at: 1 put: 0.  "Just to be sure"
	self copyBits.
	^ destForm bits at: 1! !

!BitBlt methodsFor: 'copying' stamp: 'ar 3/1/2004 13:49'!
pixelAt: aPoint put: pixelValue
	"Assumes this BitBlt has been set up specially (see the init message,
	BitBlt bitPokerToForm:.  Overwrites the pixel at aPoint."
	destX := aPoint x.
	destY := aPoint y.
	sourceForm unhibernate. "before poking"
	sourceForm bits at: 1 put: pixelValue.
	self copyBits
"
| bb |
bb := (BitBlt bitPokerToForm: Display).
[Sensor anyButtonPressed] whileFalse:
	[bb pixelAt: Sensor cursorPoint put: 55]
"! !

!BitBlt methodsFor: 'private' stamp: 'ar 8/15/2009 11:16'!
primDisplayString: aString from: startIndex to: stopIndex map: glyphMap xTable: xTable kern: kernDelta
	| ascii |
	<primitive:'primitiveDisplayString' module:'BitBltPlugin'>
	startIndex to: stopIndex do:[:charIndex|
		ascii := (aString at: charIndex) asciiValue.
		glyphMap ifNotNil:[ascii := glyphMap at: ascii+1].
		sourceX := xTable at: ascii + 1.
		width := (xTable at: ascii + 2) - sourceX.
		self copyBits.
		destX := destX + width + kernDelta.
	].! !

!BitBlt methodsFor: 'private' stamp: 'yo 3/15/2005 09:47'!
roundVariables

	| maxVal minVal |
	maxVal := SmallInteger maxVal.
	minVal := SmallInteger minVal.
	destX := destX asInteger min: maxVal max: minVal.
	destY := destY asInteger min: maxVal max: minVal.
	width := width asInteger min: maxVal max: minVal.
	height := height asInteger min: maxVal max: minVal.
	sourceX := sourceX asInteger min: maxVal max: minVal.
	sourceY := sourceY asInteger min: maxVal max: minVal.
	clipX := clipX asInteger min: maxVal max: minVal.
	clipY := clipY asInteger min: maxVal max: minVal.
	clipWidth := clipWidth asInteger min: maxVal max: minVal.
	clipHeight := clipHeight asInteger min: maxVal max: minVal.
! !

!BitBlt methodsFor: 'private' stamp: ''!
setDestForm: df
	| bb |
	bb := df boundingBox.
	destForm := df.
	clipX := bb left.
	clipY := bb top.
	clipWidth := bb width.
	clipHeight := bb height! !

!BitBlt methodsFor: 'private' stamp: 'ar 5/14/2001 23:32'!
setDestForm: df sourceForm: sf fillColor: hf combinationRule: cr destOrigin: destOrigin sourceOrigin: sourceOrigin extent: extent clipRect: clipRect

	| aPoint |
	destForm := df.
	sourceForm := sf.
	self fillColor: hf.	"sets halftoneForm"
	combinationRule := cr.
	destX := destOrigin x.
	destY := destOrigin y.
	sourceX := sourceOrigin x.
	sourceY := sourceOrigin y.
	width := extent x.
	height := extent y.
	aPoint := clipRect origin.
	clipX := aPoint x.
	clipY := aPoint y.
	aPoint := clipRect corner.
	clipWidth := aPoint x - clipX.
	clipHeight := aPoint y - clipY.
	sourceForm == nil ifFalse:
		[colorMap := sourceForm colormapIfNeededFor: destForm]! !

!BitBlt methodsFor: 'accessing' stamp: ''!
sourceForm

	^ sourceForm! !

!BitBlt methodsFor: 'accessing' stamp: ''!
sourceForm: aForm 
	"Set the receiver's source form to be the argument, aForm."

	sourceForm := aForm! !

!BitBlt methodsFor: 'accessing' stamp: ''!
sourceOrigin: aPoint 
	"Set the receiver's source form coordinates to be those of the argument, 
	aPoint."

	sourceX := aPoint x.
	sourceY := aPoint y! !

!BitBlt methodsFor: 'accessing' stamp: ''!
sourceRect: aRectangle 
	"Set the receiver's source form top left x and y, width and height to be 
	the top left coordinate and extent of the argument, aRectangle."

	sourceX := aRectangle left.
	sourceY := aRectangle top.
	width := aRectangle width.
	height := aRectangle height! !

!BitBlt methodsFor: 'accessing' stamp: ''!
sourceX: anInteger 
	"Set the receiver's source form top left x to be the argument, anInteger."

	sourceX := anInteger! !

!BitBlt methodsFor: 'accessing' stamp: ''!
sourceY: anInteger 
	"Set the receiver's source form top left y to be the argument, anInteger."

	sourceY := anInteger! !

!BitBlt methodsFor: 'accessing' stamp: 'ar 5/25/2000 19:39'!
tallyMap
	"Return the map used for tallying pixels"
	^colorMap! !

!BitBlt methodsFor: 'accessing' stamp: 'ar 5/25/2000 19:39'!
tallyMap: aBitmap
	"Install the map used for tallying pixels"
	colorMap := aBitmap! !

!BitBlt methodsFor: 'accessing' stamp: ''!
width: anInteger 
	"Set the receiver's destination form width to be the argument, anInteger."

	width := anInteger! !

BitBlt subclass: #Pen
	instanceVariableNames: 'location direction penDown'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Primitives'!

!Pen commentStamp: '<historical>' prior: 0!
My instances can scribble on the screen or some other Form, drawing and printing at any angle. Since I am a BitBlt, the drawing can be done with an arbitary source Form.
!

!Pen class methodsFor: 'examples' stamp: 'jm 5/6/1998 22:28'!
example
	"Draw a spiral with a pen that is 2 pixels wide."
	"Display restoreAfter: [Pen example]"

	| bic |
	bic := self new.
	bic defaultNib: 2.
	bic color: Color blue.
	bic combinationRule: Form over.
	1 to: 100 do: [:i | bic go: i*4. bic turn: 89].
! !

!Pen class methodsFor: 'tablet drawing examples' stamp: 'ar 5/14/2001 23:35'!
feltTip: width cellSize: cellSize
	"Warning: This example potentially uses a large amount of memory--it creates a Form with cellSize squared bits for every Display pixel."
	"In this example, all drawing is done into a large, monochrome Form and then scaled down onto the Display using smoothing. The larger the cell size, the more possible shades of gray can be generated, and the smoother the resulting line appears. A cell size of 8 yields 64 possible grays, while a cell size of 16 gives 256 levels, which is about the maximum number of grays that the human visual system can distinguish. The width parameter determines the maximum line thickness. Requires the optional tablet support primitives which may not be supported on all platforms. Works best in full screen mode. Shift-mouse to exit." 
	"Pen feltTip: 2.7 cellSize: 8"

	| tabletScale bitForm pen warp p srcR dstR nibSize startP r |
	tabletScale := self tabletScaleFactor.
	bitForm := Form extent: Display extent * cellSize depth: 1.
	pen := Pen newOnForm: bitForm.
	pen color: Color black.
	warp := (WarpBlt current toForm: Display)
		sourceForm: bitForm;
		colorMap: (bitForm colormapIfNeededFor: Display);
		cellSize: cellSize;
		combinationRule: Form over.
	Display fillColor: Color white.
	Display restoreAfter: [
		[Sensor shiftPressed and: [Sensor anyButtonPressed]] whileFalse: [
			p := (Sensor tabletPoint * cellSize * tabletScale) rounded.
			nibSize := (Sensor tabletPressure * (cellSize * width)) rounded.
		     nibSize > 0
				ifTrue: [
					pen squareNib: nibSize.
					startP := pen location.
					pen goto: p.
					r := startP rect: pen location.
					dstR := (r origin // cellSize) corner: ((r corner + nibSize + (cellSize - 1)) // cellSize).
					srcR := (dstR origin * cellSize) corner: (dstR corner * cellSize).
					warp copyQuad: srcR innerCorners toRect: dstR]
				ifFalse: [
					pen place: p]]].
! !

!Pen class methodsFor: 'tablet drawing examples' stamp: 'jm 4/13/1999 11:20'!
inkBrush
	"Similar to simplePressurePen, but this example uses the average of the recent pen pressure values. The effect is that of a Japanese ink brush that comes up gradually off the paper as the brush is lifted, causing end (and beginning) of each stroke to taper. Requires the optional tablet support primitives which may not be supported on all platforms. Works best in full screen mode. Shift-mouse to exit." 
	"Pen inkBrush"

	| tabletScale historyMSecs pressureHistory pen now currentPressure sum averagePressure p |
	tabletScale := self tabletScaleFactor.
	historyMSecs := 120.
	pressureHistory := OrderedCollection new.
	pen := Pen newOnForm: Display.
	pen color: Color black.
	Display fillColor: Color white.
	Display restoreAfter: [
		[Sensor shiftPressed and: [Sensor anyButtonPressed]] whileFalse: [
			"compute the average pressure over last historyMSecs milliseconds"
			now := Time millisecondClockValue.
			currentPressure := (20.0 * Sensor tabletPressure) rounded.
			pressureHistory addLast: (Array with: now with: currentPressure).
			[pressureHistory size > 0 and:
			 [(pressureHistory first first + historyMSecs) < now]]
				whileTrue: [pressureHistory removeFirst].  "prune old entries"
			sum := pressureHistory inject: 0 into: [:t :e | t + e last].
			averagePressure := sum // pressureHistory size.

			p := (Sensor tabletPoint * tabletScale) rounded.
		     averagePressure > 0
				ifTrue: [
					pen roundNib: averagePressure.
					pen goto: p]
				ifFalse: [
					pen place: p]]].
! !

!Pen class methodsFor: 'instance creation' stamp: ''!
new
	^ self newOnForm: Display! !

!Pen class methodsFor: 'instance creation' stamp: ''!
newOnForm: aForm
	| pen |
	pen := super new.
	pen setDestForm: aForm.
	pen sourceOrigin: 0@0.
	pen home.
	pen defaultNib: 1.
	pen north.
	pen down.
	^ pen! !

!Pen class methodsFor: 'tablet drawing examples' stamp: 'jm 4/13/1999 11:13'!
simplePressurePen
	"An example of using a pressure sensitive pen to control the thickness of the pen. This requires the optional tablet support primitives which may not be supported on all platforms. Works best in full screen mode. Shift-mouse to exit." 
	"Pen simplePressurePen"

	| tabletScale pen pressure p |
	tabletScale := self tabletScaleFactor.
	pen := Pen newOnForm: Display.
	pen color: Color black.
	Display fillColor: Color white.
	Display restoreAfter: [
		[Sensor shiftPressed and: [Sensor anyButtonPressed]] whileFalse: [
			p := (Sensor tabletPoint * tabletScale) rounded.
			pressure := (15.0 * Sensor tabletPressure) rounded.
		     pressure > 0
				ifTrue: [
					pen roundNib: pressure.
					pen goto: p]
				ifFalse: [
					pen place: p]]].
! !

!Pen class methodsFor: 'tablet drawing examples' stamp: 'jm 4/13/1999 11:12'!
tabletScaleFactor
	"Answer a Point that scales tablet coordinates to Display coordinates, where the full extent of the tablet maps to the extent of the entire Display."

	| tabletExtent |
	tabletExtent := Sensor tabletExtent.
	^ (Display width asFloat / tabletExtent x) @ (Display height asFloat / tabletExtent y)
! !

!Pen class methodsFor: 'tablet drawing examples' stamp: 'jm 4/12/1999 12:51'!
testMouseTracking
	"A very simple example of drawing using the mouse. Compare the tracking speed of this example with that of testTabletTracking. Mouse down to draw a stroke, shift-mouse to exit." 
	"Pen testMouseTracking"

	| pen p |
	pen := Pen newOnForm: Display.
	pen roundNib: 8.
	pen color: Color black.
	Display fillColor: Color white.
	Display restoreAfter: [
		[Sensor shiftPressed and: [Sensor anyButtonPressed]] whileFalse: [
			p := Sensor cursorPoint.
		     Sensor anyButtonPressed
				ifTrue: [pen goto: p]
				ifFalse: [
					pen color: Color random.
					pen place: p]]].
! !

!Pen class methodsFor: 'tablet drawing examples' stamp: 'jm 4/13/1999 11:21'!
testTabletTracking
	"A very simple example of drawing using the pen of a digitizing tablet such as a Wacom ArtZ tablet. This requires the optional tablet support primitives which may not be supported on all platforms. Compare the tracking speed of this example with that of testMouseTracking. On a Macintosh, the tablet primitives provide roughly 120 samples/second versus only 60 mouse samples/second, and the difference is noticable. Works best in full screen mode. Mouse down to draw a stroke, shift-mouse to exit." 
	"Pen testTabletTracking"

	| tabletScale pen p |
	tabletScale := self tabletScaleFactor.
	pen := Pen newOnForm: Display.
	pen roundNib: 8.
	pen color: Color black.
	Display fillColor: Color white.
	Display restoreAfter: [
		[Sensor shiftPressed and: [Sensor anyButtonPressed]] whileFalse: [
			p := (Sensor tabletPoint * tabletScale) rounded.
		     Sensor tabletPressure > 0
				ifTrue: [pen goto: p]
				ifFalse: [
					pen color: Color random.
					pen place: p]]].
! !

!Pen methodsFor: 'operations' stamp: 'sw 10/5/2002 03:17'!
arrowHead
	"Put an arrowhead on the previous pen stroke"
	" | pen | pen := Pen new. 20 timesRepeat: [pen turn: 360//20; go: 20; arrowHead]."

	penDown ifTrue:
		[self arrowHeadFrom: (direction degreeCos @ direction degreeSin) * -40 + location 
			to: location
			arrowSpec: (Preferences parameterAt: #arrowSpec ifAbsent: [5 @ 4])]! !

!Pen methodsFor: 'operations' stamp: 'sw 10/5/2002 02:29'!
arrowHeadForArrowSpec: anArrowSpec
	"Put an arrowhead on the previous pen stroke"
"
	 | pen aPoint |
	aPoint := Point fromUser.
	pen := Pen new.
	20 timesRepeat: [pen turn: 360//20; go: 20; arrowHeadForArrowSpec: aPoint].
"


	penDown ifTrue:
		[self arrowHeadFrom: (direction degreeCos @ direction degreeSin) * -40 + location 
			to: location
			arrowSpec: anArrowSpec]! !

!Pen methodsFor: 'operations' stamp: 'sw 10/5/2002 02:25'!
arrowHeadFrom: prevPt to: newPt arrowSpec: anArrowSpec
	"Put an arrowhead on the pen stroke from oldPt to newPt"

	| pm af myColor finalPt delta |
	myColor := self color.
	delta := newPt - prevPt.
	delta r <= 2 "pixels" ifTrue: [^ self].
	finalPt := newPt + (Point r: sourceForm width degrees: delta degrees).	"in same direction"
	pm := PolygonMorph vertices: (Array with: prevPt asIntegerPoint with: finalPt asIntegerPoint)  
		color: myColor  "not used"
		borderWidth: sourceForm width borderColor: myColor.
	pm makeOpen; makeForwardArrow.
	anArrowSpec ifNotNil: [pm arrowSpec: anArrowSpec].
	af := pm arrowForms first.
	"render it onto the destForm"
	(FormCanvas on: destForm "Display") stencil: af at: af offset + (1@1)
		color: myColor! !

!Pen methodsFor: 'operations' stamp: 'sw 10/5/2002 02:11'!
arrowHeadFrom: prevPt to: newPt forPlayer: aPlayer
	"Put an arrowhead on the pen stroke from oldPt to newPt"
	
	| aSpec |
	(aPlayer notNil and: [(aSpec := aPlayer costume renderedMorph valueOfProperty: #arrowSpec) notNil]) 
		ifFalse:
			[aSpec := Preferences parameterAt: #arrowSpec "may well be nil"].
	self arrowHeadFrom: prevPt to: newPt arrowSpec: aSpec! !

!Pen methodsFor: 'operations' stamp: 'jm 4/28/1998 03:40'!
color: aColorOrInteger
	"Set the pen to the given color or to a color chosen from a fixed set of colors."

	| count c |
	aColorOrInteger isInteger
		ifTrue: [
			destForm depth = 1 ifTrue: [^ self fillColor: Color black].
			count := 19.  "number of colors in color wheel"
			c := (Color red wheel: count) at: ((aColorOrInteger * 7) \\ count) + 1]
		ifFalse: [c := aColorOrInteger].  "assume aColorOrInteger is a Color"
	self fillColor: c.
! !

!Pen methodsFor: 'initialize-release' stamp: 'jm 4/28/1998 04:02'!
defaultNib: widthInteger 
	"Nib is the tip of a pen. This sets up the pen, with a nib of width widthInteger. You can also set the shape of the pen nib using:
		roundNib: widthInteger, or
		squareNib: widthInteger, or
		sourceForm: aForm"
"Example:
	| bic |
	bic := Pen new sourceForm: Cursor normal.
	bic combinationRule: Form paint; turn: 90.
	10 timesRepeat: [bic down; go: 3; up; go: 10]."

	self color: Color black.
	self squareNib: widthInteger.
! !

!Pen methodsFor: 'accessing' stamp: ''!
direction
	"Answer the receiver's current direction. 0 is towards the top of the
	screen."

	^direction! !

!Pen methodsFor: 'operations' stamp: ''!
down
	"Set the state of the receiver's pen to down (drawing)."

	penDown := true! !

!Pen methodsFor: 'geometric designs' stamp: 'di 6/11/1998 22:01'!
dragon: n  "Display restoreAfter: [Display fillWhite. Pen new dragon: 10]."
	"Display restoreAfter: [Display fillWhite. 1 to: 4 do:
				[:i | Pen new color: i; turn: 90*i; dragon: 10]]"
	"Draw a dragon curve of order n in the center of the screen."
	n = 0
		ifTrue: [self go: 5]
		ifFalse: [n > 0
				ifTrue: [self dragon: n - 1; turn: 90; dragon: 1 - n]
				ifFalse: [self dragon: -1 - n; turn: -90; dragon: 1 + n]]
! !

!Pen methodsFor: 'geometric designs' stamp: 'di 6/14/1998 13:42'!
filberts: n side: s   "Display restoreAfter: [Pen new filberts: 4 side: 5]"
	"Two Hilbert curve fragments form a Hilbert tile. Draw four interlocking 
	tiles of order n and sides length s."
	| n2 |
	Display fillWhite.
	n2 := 1 bitShift: n - 1.
	self up; go: 0 - n2 * s; down.
	1 to: 4 do: 
		[:i | 
		self fill: [:p |
				p hilbert: n side: s.
				p go: s.
				p hilbert: n side: s.
				p go: s.
				p up.
				p go: n2 - 1 * s.
				p turn: -90.
				p go: n2 * s.
				p turn: 180.
				p down]
			color: (Color perform: (#(yellow red green blue) at: i))]! !

!Pen methodsFor: 'operations' stamp: 'di 6/21/1998 09:37'!
fill: drawBlock color: color
	| region tileForm tilePen shape saveColor recorder |
	drawBlock value: (recorder := self as: PenPointRecorder).
	region := Rectangle encompassing: recorder points.
	tileForm := Form extent: region extent+6.
	tilePen := Pen newOnForm: tileForm.
	tilePen location: location-(region origin-3)
		direction: direction
		penDown: penDown.
	drawBlock value: tilePen.  "Draw the shape in B/W"
	saveColor := halftoneForm.
	drawBlock value: self.
	halftoneForm := saveColor.
	shape := (tileForm findShapeAroundSeedBlock: [:f | f borderWidth: 1]) reverse.
	shape copy: shape boundingBox from: tileForm to: 0@0 rule: Form erase.
	destForm fillShape: shape fillColor: color at: region origin-3! !

!Pen methodsFor: 'operations' stamp: ''!
go: distance 
	"Move the pen in its current direction a number of bits equal to the 
	argument, distance. If the pen is down, a line will be drawn using the 
	receiver's form source as the shape of the drawing brush."

	self goto: (direction degreeCos @ direction degreeSin) * distance + location! !

!Pen methodsFor: 'operations' stamp: 'di 11/4/97 20:11'!
goto: aPoint 
	"Move the receiver to position aPoint. If the pen is down, a line will be 
	drawn from the current position to the new one using the receiver's 
	form source as the shape of the drawing brush. The receiver's set 
	direction does not change."
	| old |
	old := location.
	location := aPoint.
	penDown ifTrue: [self drawFrom: old rounded
								to: location rounded]

	"NOTE:  This should be changed so it does NOT draw the first point, so as
	not to overstrike at line junctions.  At the same time, place should draw
	a single dot if the pen is down, as should down (put-pen-down) if it
	was not down before."! !

!Pen methodsFor: 'geometric designs' stamp: ''!
hilbert: n side: s 
	"Draw an nth level Hilbert curve with side length s in the center of the 
	screen. Write directly into the display's bitmap only. A Hilbert curve is 
	a space-filling curve."

	| a m |
	n = 0 ifTrue: [^self turn: 180].
	n > 0
		ifTrue: 
			[a := 90.
			m := n - 1]
		ifFalse: 
			[a := -90.
			m := n + 1].
	self turn: a.
	self hilbert: 0 - m side: s.
	self turn: a; go: s.
	self hilbert: m side: s.
	self turn: 0 - a; go: s; turn: 0 - a.
	self hilbert: m side: s.
	self go: s; turn: a.
	self hilbert: 0 - m side: s.
	self turn: a
	" 
	(Pen new) hilbert: 3 side: 8. 
	(Pen new sourceForm: Cursor wait) combinationRule: Form under; 
	hilbert: 3 side: 25.
	"! !

!Pen methodsFor: 'geometric designs' stamp: ''!
hilberts: n   "Display restoreAfter: [Display fillWhite.  Pen new hilberts: 5]"
	"Draws n levels of nested Hilbert curves"
	| s |
	self up; turn: 90; go: 128; down.
	1 to: n do: 
		[:i | 
		s := 256 bitShift: 0 - i.
		self defaultNib: n - i * 2 + 1.
		self color: i+1.
		self up; go: 0 - s / 2; turn: -90; go: s / 2; turn: 90; down.
		self hilbert: i side: s.
		self go: s.
		self hilbert: i side: s.
		self go: s]! !

!Pen methodsFor: 'operations' stamp: ''!
home
	"Place the receiver at the center of its frame."
	location := destForm boundingBox center! !

!Pen methodsFor: 'accessing' stamp: ''!
location
	"Answer where the receiver is currently located."

	^location! !

!Pen methodsFor: 'private' stamp: 'di 6/11/1998 16:09'!
location: aPoint direction: aFloat penDown: aBoolean
	location := aPoint.
	direction := aFloat.
	penDown := aBoolean! !

!Pen methodsFor: 'geometric designs' stamp: ''!
mandala: npoints
	"Display restoreAfter: [Pen new mandala: 30]"
	"On a circle of diameter d, place npoints number of points. Draw all 	possible connecting lines between the circumferential points."
	| l points d |
	Display fillWhite.
	d := Display height-50.
	l := 3.14 * d / npoints.
	self home; up; turn: -90; go: d // 2; turn: 90; go: 0 - l / 2; down.
	points := Array new: npoints.
	1 to: npoints do: 
		[:i | 
		points at: i put: location rounded.
		self go: l; turn: 360.0 / npoints].
	npoints // 2
		to: 1
		by: -1
		do: 
			[:i | 
			self color: i.
			1 to: npoints do: 
				[:j | 
				self place: (points at: j).
				self goto: (points at: j + i - 1 \\ npoints + 1)]]
! !

!Pen methodsFor: 'operations' stamp: ''!
north
	"Set the receiver's direction to facing toward the top of the display screen."

	direction := 270! !

!Pen methodsFor: 'operations' stamp: ''!
place: aPoint 
	"Set the receiver at position aPoint. No lines are drawn."

	location := aPoint! !

!Pen methodsFor: 'operations' stamp: 'ar 5/28/2000 12:10'!
print: str withFont: font
	"Print the given string in the given font at the current heading"
	| lineStart form charStart rowStart scale wasDown bb pix |
	scale := sourceForm width.
	wasDown := penDown.
	lineStart := location.
	str do:
		[:char |
		char = Character cr ifTrue:
			[self place: lineStart; up; turn: 90; go: font height*scale; turn: -90; down]
		ifFalse:
			[form := font characterFormAt: char.
			charStart := location.
wasDown ifTrue: [
			self up; turn: -90; go: font descent*scale; turn: 90; down.
			0 to: form height-1 do:
				[:y |
				rowStart := location.
				bb := BitBlt current bitPeekerFromForm: form.
				pix := RunArray newFrom:
					((0 to: form width-1) collect: [:x | bb pixelAt: x@y]).
				pix runs with: pix values do:
					[:run :value |
					value = 0
						ifTrue: [self up; go: run*scale; down]
						ifFalse: [self go: run*scale]].
				self place: rowStart; up; turn: 90; go: scale; turn: -90; down].
].
			self place: charStart; up; go: form width*scale; down].
			].
	wasDown ifFalse: [self up]
"
Display restoreAfter:
[Pen new squareNib: 2; color: Color red; turn: 45;
	print: 'The owl and the pussycat went to sea
in a beautiful pea green boat.' withFont: TextStyle defaultFont]
"! !

!Pen methodsFor: 'operations' stamp: 'sw 4/10/2003 22:37'!
putDotOfDiameter: aDiameter at: aPoint
	"Put a dot of the given size at the given point, using my colot"

	(FormCanvas on: destForm) 
			fillOval: (Rectangle center: aPoint extent: (aDiameter @ aDiameter))
			color: self color! !

!Pen methodsFor: 'initialize-release' stamp: 'jm 4/28/1998 04:03'!
roundNib: diameter
	"Makes this pen draw with a round dot of the given diameter."

	self sourceForm: (Form dotOfSize: diameter).
	combinationRule := Form paint.
! !

!Pen methodsFor: 'private' stamp: ''!
sourceForm: aForm
	(aForm depth = 1 and: [destForm depth > 1])
		ifTrue: ["Map 1-bit source to all ones for color mask"
				colorMap := Bitmap with: 0 with: 16rFFFFFFFF]
		ifFalse: [colorMap := nil].
	^ super sourceForm: aForm! !

!Pen methodsFor: 'geometric designs' stamp: 'jm 5/6/1998 22:26'!
spiral: n angle: a 
	"Draw a double squiral (see Papert, MindStorms), where each design is made
	by moving the receiver a distance of n after turning the amount + or -a."

	1 to: n do: 
		[:i | 
		self color: i * 2.
		self go: i; turn: a]
"
	Display restoreAfter: [
		Display fillWhite. Pen new spiral: 200 angle: 89; home; spiral: 200 angle: -89].
"! !

!Pen methodsFor: 'initialize-release' stamp: 'jm 4/28/1998 04:03'!
squareNib: widthInteger 
	"Makes this pen draw with a square nib of the given width."

	self sourceForm: (Form extent: widthInteger @widthInteger) fillBlack.
	self combinationRule: Form over.  "a bit faster than paint mode"
! !

!Pen methodsFor: 'operations' stamp: ''!
turn: degrees 
	"Change the direction that the receiver faces by an amount equal to the 
	argument, degrees."

	direction := direction + degrees! !

!Pen methodsFor: 'operations' stamp: ''!
up
	"Set the state of the receiver's pen to up (no drawing)."

	penDown := false! !

!Pen methodsFor: 'geometric designs' stamp: 'jm 8/1/97 11:42'!
web   "Display restoreAfter: [Pen new web]"
	"Draw pretty web-like patterns from the mouse movement on the screen.
	Press the mouse button to draw, option-click to exit.
	By Dan Ingalls and Mark Lentczner. "
	| history newPoint ancientPoint lastPoint filter color |
	"self erase."
	color := 1.
	[ true ] whileTrue:
		[ history := OrderedCollection new.
		Sensor waitButton.
		Sensor yellowButtonPressed ifTrue: [^ self].
		filter := lastPoint := Sensor mousePoint.
		20 timesRepeat: [ history addLast: lastPoint ].
		self color: (color := color + 1).
		[ Sensor redButtonPressed ] whileTrue: 
			[ newPoint := Sensor mousePoint.
			(newPoint = lastPoint) ifFalse:
				[ ancientPoint := history removeFirst.
				filter := filter * 4 + newPoint // 5.
				self place: filter.
				self goto: ancientPoint.
				lastPoint := newPoint.
				history addLast: filter ] ] ]! !

Pen subclass: #PenPointRecorder
	instanceVariableNames: 'points'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Primitives'!

!PenPointRecorder commentStamp: '<historical>' prior: 0!
This class is a special kind of Pen that instead of actually drawing lines records the destination points for those lines. These points can later be accessed through my accessing method #points.

This can be useful when determining the boundaries of a drawing session.

Example:

| pen |
pen _ PenPointRecorder new.
pen up; goto: 100@100; down; goto: 120@120.
Transcript cr;
	show: 'Bounding box for drawing: ';
	show: (Rectangle encompassing: pen points)

Implementation note: Shouldn't we override #drawFrom:to:withFirstPoint: instead, and what about #drawLoopX:Y:? Aren't we missing those calls?!

!PenPointRecorder methodsFor: 'line drawing' stamp: 'md 11/14/2003 16:56'!
drawFrom: p1 to: p2
	"Overridden to skip drawing but track bounds of the region traversed."

	points ifNil: [points := OrderedCollection with: p1].
	points addLast: p2! !

!PenPointRecorder methodsFor: 'accessing' stamp: 'di 6/21/1998 09:35'!
points
	^ points! !

BitBlt subclass: #WarpBlt
	instanceVariableNames: 'p1x p1y p1z p2x p2y p2z p3x p3y p3z p4x p4y p4z cellSize sourceRGBmap'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Primitives'!

!WarpBlt commentStamp: '<historical>' prior: 0!
WarpBlt is a little warp-drive added on to BitBlt.  It takes a quadrilateral as its source specification, while its destination is traversed and combined just like any other call to copyBits.

The source quadrilateral is specified as an array of points starting with the corner that wants to end up in the topLeft, and proceding to the successive points that want to follow CCW around the destination rectangle.  Note that in specifying a plain old rectangle source, its non topLeft points must be actual pixels, not outside by 1, as with rectangle bottmRight, eg.  See the method Rectangle asQuad.

WarpBlt does a fast job of rotation, reflection and scaling, and it can even produce a semblance of perspective.  Depth parameters are included for future improvements in this direction. but the primitve does not support this yet.!

!WarpBlt class methodsFor: 'instance creation' stamp: 'ar 5/28/2000 12:00'!
current
	"Return the class currently to be used for WarpBlt"
	^Display defaultWarpBltClass! !

!WarpBlt class methodsFor: 'form rotation' stamp: 'nk 4/17/2004 19:45'!
rotate: srcForm degrees: angleInDegrees center: aPoint scaleBy: scalePoint smoothing: cellSize
	"Rotate the given Form the given number of degrees about the given center and scale its width and height by x and y of the given scale point. Smooth using the given cell size, an integer between 1 and 3, where 1 means no smoothing. Return a pair where the first element is the rotated Form and the second is the position offset required to align the center of the rotated Form with that of the original. Note that the dimensions of the resulting Form generally differ from those of the original."

	| srcRect center radians dstOrigin dstCorner p dstRect inverseScale quad dstForm newCenter warpSrc |
	srcRect := srcForm boundingBox.
	center := srcRect center.
	radians := angleInDegrees degreesToRadians.
	dstOrigin := dstCorner := center.
	srcRect corners do: [:corner |
		"find the limits of a rectangle that just encloses the rotated
		 original; in general, this rectangle will be larger than the
		 original (e.g., consider a square rotated by 45 degrees)"
		p := ((corner - center) scaleBy: scalePoint) + center.
		p := (p rotateBy: radians about: center) rounded.
		dstOrigin := dstOrigin min: p.
		dstCorner := dstCorner max: p].

	"rotate the enclosing rectangle back to get the source quadrilateral"
	dstRect := dstOrigin corner: dstCorner.
	inverseScale := (1.0 / scalePoint x)@(1.0 / scalePoint y).
	quad := dstRect innerCorners collect: [:corner |
		p := corner rotateBy: radians negated about: center.
		((p - center) scaleBy: inverseScale) + center].

	"make a Form to hold the result and do the rotation"
	warpSrc := srcForm.
	(srcForm isColorForm)
		ifTrue: [
			cellSize > 1 | true "ar 12/27/2001: Always enable - else sketches won't work"
				ifTrue: [
					warpSrc := Form extent: srcForm extent depth: 16.
					srcForm displayOn: warpSrc.
					dstForm := Form extent: dstRect extent depth: 16]  "use 16-bit depth to allow smoothing"
				ifFalse: [
					dstForm := srcForm class extent: dstRect extent depth: srcForm depth]]
		ifFalse: [
			dstForm := srcForm class extent: dstRect extent depth: srcForm depth].

	(WarpBlt toForm: dstForm)
		sourceForm: warpSrc;
		colorMap: (warpSrc colormapIfNeededFor: dstForm);
		cellSize: cellSize;  "installs a new colormap if cellSize > 1"
		combinationRule: Form paint;
		copyQuad: quad toRect: dstForm boundingBox.

	(dstForm isColorForm) ifTrue: [dstForm colors: srcForm colors copy].
	newCenter := (center rotateBy: radians about: aPoint) truncated.
	^ Array with: dstForm with: dstRect origin + (newCenter - center)
! !

!WarpBlt class methodsFor: 'examples' stamp: ''!
test1   "Display restoreAfter: [WarpBlt test1]"
	"Demonstrates variable scale and rotate"
	| warp pts r1 p0 p ext |
	Utilities informUser: 'Choose a rectangle with interesting stuff'
		during: [r1 := Rectangle originFromUser: 50@50.
				Sensor waitNoButton].
	Utilities informUser: 'Now click down and up
and move the mouse around the dot'
		during: [p0 := Sensor waitClickButton.
				(Form dotOfSize: 8) displayAt: p0].
	warp := (self toForm: Display)
		clipRect: (0@0 extent: r1 extent*5);
		sourceForm: Display;
		combinationRule: Form over.
	[Sensor anyButtonPressed] whileFalse:
		[p := Sensor cursorPoint.
		pts := {r1 topLeft. r1 bottomLeft. r1 bottomRight. r1 topRight}
			collect: [:pt | pt rotateBy: (p-p0) theta about: r1 center].
		ext := (r1 extent*((p-p0) r / 20.0 max: 0.1)) asIntegerPoint.
		warp copyQuad: pts toRect: (r1 extent*5-ext//2 extent: ext)]! !

!WarpBlt class methodsFor: 'examples' stamp: 'sma 4/22/2000 20:23'!
test12   "Display restoreAfter: [WarpBlt test12]"
	"Just like test1, but comparing smooth to non-smooth warps"
	| warp pts r1 p0 p ext warp2 |
	Utilities informUser: 'Choose a rectangle with interesting stuff'
		during: [r1 := Rectangle originFromUser: 50@50.
				Sensor waitNoButton].
	Utilities informUser: 'Now click down and up
and move the mouse around the dot'
		during: [p0 := Sensor waitClickButton.
				(Form dotOfSize: 8) displayAt: p0].
	warp := (self toForm: Display)
		cellSize: 2;  "installs a colormap"
		clipRect: (0@0 extent: r1 extent*5);
		sourceForm: Display;
		combinationRule: Form over.
	warp2 := (self toForm: Display)
		clipRect: ((0@0 extent: r1 extent*5) translateBy: 250@0);
		sourceForm: Display;
		combinationRule: Form over.
	[Sensor anyButtonPressed] whileFalse:
		[p := Sensor cursorPoint.
		pts := {r1 topLeft. r1 bottomLeft. r1 bottomRight. r1 topRight}
			collect: [:pt | pt rotateBy: (p-p0) theta about: r1 center].
		ext := (r1 extent*((p-p0) r / 20.0 max: 0.1)) asIntegerPoint.
		warp copyQuad: pts toRect: (r1 extent*5-ext//2 extent: ext).
		warp2 copyQuad: pts toRect: ((r1 extent*5-ext//2 extent: ext) translateBy: 250@0).
		]! !

!WarpBlt class methodsFor: 'examples' stamp: ''!
test3   "Display restoreAfter: [WarpBlt test3]"

	"The Squeak Release Mandala - 9/23/96 di"

	"Move the mouse near the center of the square.
	Up and down affects shrink/grow
	Left and right affect rotation angle"
	| warp pts p0 p box map d t |
	box := 100@100 extent: 300@300.
	Display border: (box expandBy: 2) width: 2.

	"Make a color map that steps through the color space"
	map := (Display depth > 8
		ifTrue: ["RGB is a bit messy..."
				d := Display depth = 16 ifTrue: [5] ifFalse: [8].
				(1 to: 512) collect: [:i | t := i bitAnd: 511.
					((t bitAnd: 16r7) bitShift: d-3)
					+ ((t bitAnd: 16r38) bitShift: d-3*2)
					+ ((t bitAnd: 16r1C0) bitShift: d-3*3)]]
		ifFalse: ["otherwise simple"
				1 to: (1 bitShift: Display depth)])
			as: Bitmap.
	warp := (WarpBlt toForm: Display)
		clipRect: box;
		sourceForm: Display;
		colorMap: map;
		combinationRule: Form over.
	p0 := box center.
	[Sensor anyButtonPressed] whileFalse:
		[p := Sensor cursorPoint.
		pts := (box insetBy: p y - p0 y) innerCorners
			collect: [:pt | pt rotateBy: p x - p0 x / 50.0 about: p0].
		warp copyQuad: pts toRect: box]! !

!WarpBlt class methodsFor: 'examples' stamp: ''!
test4   "Display restoreAfter: [WarpBlt test4]"

	"The Squeak Release Mandala - 9/23/96 di
	This version does smoothing"

	"Move the mouse near the center ofhe square.
	Up and dn affects shrink/grow
	Left and right affect rotation angle"
	| warp pts p0 p box |
	box := 100@100 extent: 300@300.
	Display border: (box expandBy: 2) width: 2.

	warp := (WarpBlt toForm: Display)
		clipRect: box;
		sourceForm: Display;
		cellSize: 2;  "installs a colormap"
		combinationRule: Form over.
	p0 := box center.
	[Sensor anyButtonPressed] whileFalse:
		[p := Sensor cursorPoint.
		pts := (box insetBy: p y - p0 y) innerCorners
			collect: [:pt | pt rotateBy: p x - p0 x / 50.0 about: p0].
		warp copyQuad: pts toRect: box]! !

!WarpBlt class methodsFor: 'examples' stamp: ''!
test5   "Display restoreAfter: [WarpBlt test5]"
	"Demonstrates variable scale and rotate"
	| warp pts r1 p0 p |
	Utilities informUser: 'Choose a rectangle with interesting stuff'
		during: [r1 := Rectangle fromUser.
				Sensor waitNoButton].
	Utilities informUser: 'Now click down and up
and move the mouse around the dot'
		during: [p0 := Sensor waitClickButton.
				(Form dotOfSize: 8) displayAt: p0].
	warp := (self toForm: Display)
		cellSize: 1;
		sourceForm: Display;
		cellSize: 2;  "installs a colormap"
		combinationRule: Form over.
	[Sensor anyButtonPressed] whileFalse:
		[p := Sensor cursorPoint.
		pts := {r1 topLeft. r1 bottomLeft. r1 bottomRight. r1 topRight}
			collect: [:pt | pt rotateBy: (p-p0) theta about: r1 center].
		warp copyQuad: pts toRect: (r1 translateBy: r1 width@0)]! !

!WarpBlt class methodsFor: 'initialization' stamp: ''!
toForm: destinationForm
	"Default cell size is 1 (no pixel smoothing)"
	^ (super toForm: destinationForm) cellSize: 1! !

!WarpBlt methodsFor: 'setup' stamp: ''!
cellSize
	^ cellSize! !

!WarpBlt methodsFor: 'setup' stamp: 'jm 4/11/1999 12:00'!
cellSize: s
	cellSize := s.
	cellSize = 1 ifTrue: [^ self].
	colorMap := Color colorMapIfNeededFrom: 32 to: destForm depth.
! !

!WarpBlt methodsFor: 'primitives' stamp: ''!
copyQuad: pts toRect: destRect
	self sourceQuad: pts destRect: destRect.
	self warpBits! !

!WarpBlt methodsFor: 'primitives' stamp: ''!
deltaFrom: x1 to: x2 nSteps: n
	"Utility routine for computing Warp increments.
	x1 is starting pixel, x2 is ending pixel;  assumes n >= 1"
	| fixedPtOne |
	fixedPtOne := 16384.  "1.0 in fixed-pt representation"
	x2 > x1
		ifTrue: [^ x2 - x1 + fixedPtOne // (n+1) + 1]
		ifFalse: [x2 = x1 ifTrue: [^ 0].
				^ 0 - (x1 - x2 + fixedPtOne // (n+1) + 1)]! !

!WarpBlt methodsFor: 'smoothing' stamp: 'di 6/24/97 00:09'!
mixPix: pix sourceMap: sourceMap destMap: destMap
	"Average the pixels in array pix to produce a destination pixel.
	First average the RGB values either from the pixels directly,
	or as supplied in the sourceMap.  Then return either the resulting
	RGB value directly, or use it to index the destination color map." 
	| r g b rgb nPix bitsPerColor d |
	nPix := pix size.
	r := 0. g := 0. b := 0.
	1 to: nPix do:
		[:i |   "Sum R, G, B values for each pixel"
		rgb := sourceForm depth <= 8
				ifTrue: [sourceMap at: (pix at: i) + 1]
				ifFalse: [sourceForm depth = 32
						ifTrue: [pix at: i]
						ifFalse: [self rgbMap: (pix at: i) from: 5 to: 8]].
		r := r + ((rgb bitShift: -16) bitAnd: 16rFF).
		g := g + ((rgb bitShift: -8) bitAnd: 16rFF).
		b := b + ((rgb bitShift: 0) bitAnd: 16rFF)].
	destMap == nil
		ifTrue: [bitsPerColor := 3.  "just in case eg depth <= 8 and no map"
				destForm depth = 16 ifTrue: [bitsPerColor := 5].
				destForm depth = 32 ifTrue: [bitsPerColor := 8]]
		ifFalse: [destMap size = 512 ifTrue: [bitsPerColor := 3].
				destMap size = 4096 ifTrue: [bitsPerColor := 4].
				destMap size = 32768 ifTrue: [bitsPerColor := 5]].
	d := bitsPerColor - 8.
	rgb := ((r // nPix bitShift: d) bitShift: bitsPerColor*2)
		+ ((g // nPix bitShift: d) bitShift: bitsPerColor)
		+ ((b // nPix bitShift: d) bitShift: 0).
	destMap == nil
		ifTrue: [^ rgb]
		ifFalse: [^ destMap at: rgb+1]! !

!WarpBlt methodsFor: 'smoothing' stamp: 'di 6/24/97 00:08'!
rgbMap: sourcePixel from: nBitsIn to: nBitsOut
	"NOTE: This code is copied verbatim from BitBltSimulation so that it
	may be removed from the system"
	"Convert the given pixel value with nBitsIn bits for each color component to a pixel value with nBitsOut bits for each color component. Typical values for nBitsIn/nBitsOut are 3, 5, or 8."
	| mask d srcPix destPix |
	self inline: true.
	(d := nBitsOut - nBitsIn) > 0
		ifTrue:
			["Expand to more bits by zero-fill"
			mask := (1 << nBitsIn) - 1.  "Transfer mask"
			srcPix := sourcePixel << d.
			mask := mask << d.
			destPix := srcPix bitAnd: mask.
			mask := mask << nBitsOut.
			srcPix := srcPix << d.
			^ destPix + (srcPix bitAnd: mask)
				 	+ (srcPix << d bitAnd: mask << nBitsOut)]
		ifFalse:
			["Compress to fewer bits by truncation"
			d = 0 ifTrue: [^ sourcePixel].  "no compression"
			sourcePixel = 0 ifTrue: [^ sourcePixel].  "always map 0 (transparent) to 0"
			d := nBitsIn - nBitsOut.
			mask := (1 << nBitsOut) - 1.  "Transfer mask"
			srcPix := sourcePixel >> d.
			destPix := srcPix bitAnd: mask.
			mask := mask << nBitsOut.
			srcPix := srcPix >> d.
			destPix := destPix + (srcPix bitAnd: mask)
					+ (srcPix >> d bitAnd: mask << nBitsOut).
			destPix = 0 ifTrue: [^ 1].  "Dont fall into transparent by truncation"
			^ destPix]! !

!WarpBlt methodsFor: 'primitives' stamp: 'jm 5/2/1999 07:09'!
sourceForm: srcForm destRect: dstRectangle
	"Set up a WarpBlt from the entire source Form to the given destination rectangle."

	| w h |
	sourceForm := srcForm.
	sourceX := sourceY := 0.
	destX := dstRectangle left.
	destY := dstRectangle top.
	width := dstRectangle width.
	height := dstRectangle height.
	w := 16384 * (srcForm width - 1).
	h := 16384 * (srcForm height - 1).
	p1x := 0.
	p2x := 0.
	p3x := w.
	p4x := w.
	p1y := 0.
	p2y := h.
	p3y := h.
	p4y := 0.
	p1z := p2z := p3z := p4z := 16384.  "z-warp ignored for now"
! !

!WarpBlt methodsFor: 'primitives' stamp: ''!
sourceQuad: pts destRect: aRectangle
	| fixedPt1 |
	sourceX := sourceY := 0.
	self destRect: aRectangle.
	fixedPt1 := (pts at: 1) x isInteger ifTrue: [16384] ifFalse: [16384.0].
	p1x := (pts at: 1) x * fixedPt1.
	p2x := (pts at: 2) x * fixedPt1.
	p3x := (pts at: 3) x * fixedPt1.
	p4x := (pts at: 4) x * fixedPt1.
	p1y := (pts at: 1) y * fixedPt1.
	p2y := (pts at: 2) y * fixedPt1.
	p3y := (pts at: 3) y * fixedPt1.
	p4y := (pts at: 4) y * fixedPt1.
	p1z := p2z := p3z := p4z := 16384.  "z-warp ignored for now"
! !

!WarpBlt methodsFor: 'primitives' stamp: ''!
startFrom: x1 to: x2 offset: sumOfDeltas
	"Utility routine for computing Warp increments."
	x2 >= x1
		ifTrue: [^ x1]
		ifFalse: [^ x2 - sumOfDeltas]! !

!WarpBlt methodsFor: 'primitives' stamp: 'nk 11/3/2004 09:26'!
warpBits
	"Move those pixels!!"

	cellSize < 1 ifTrue: [ ^self error: 'cellSize must be >= 1' ].

	self warpBitsSmoothing: cellSize
		sourceMap: (sourceForm colormapIfNeededForDepth: 32).
! !

!WarpBlt methodsFor: 'system simulation' stamp: 'nk 11/3/2004 09:26'!
warpBitsSimulated
	"Simulate WarpBlt"

	cellSize < 1 ifTrue: [ ^self error: 'cellSize must be >= 1' ].

	self warpBitsSimulated: cellSize
		sourceMap: (sourceForm colormapIfNeededForDepth: 32).
! !

!WarpBlt methodsFor: 'system simulation' stamp: 'ar 2/13/2001 21:12'!
warpBitsSimulated: n sourceMap: sourceMap
	"Simulate WarpBlt"
	Smalltalk at: #BitBltSimulation ifPresent:[:bb| bb warpBitsFrom: self].! !

!WarpBlt methodsFor: 'primitives' stamp: 'nk 4/17/2004 19:50'!
warpBitsSmoothing: n sourceMap: sourceMap
	| deltaP12 deltaP43 pA pB deltaPAB sp fixedPtOne picker poker pix nSteps |
	<primitive: 'primitiveWarpBits' module: 'BitBltPlugin'>

	"Check for compressed source, destination or halftone forms"
	((sourceForm isForm) and: [sourceForm unhibernate])
		ifTrue: [^ self warpBitsSmoothing: n sourceMap: sourceMap].
	((destForm isForm) and: [destForm unhibernate])
		ifTrue: [^ self warpBitsSmoothing: n sourceMap: sourceMap].
	((halftoneForm isForm) and: [halftoneForm unhibernate])
		ifTrue: [^ self warpBitsSmoothing: n sourceMap: sourceMap].

	(width < 1) | (height < 1) ifTrue: [^ self].
	fixedPtOne := 16384.  "1.0 in fixed-pt representation"
	n > 1 ifTrue:
		[(destForm depth < 16 and: [colorMap == nil])
			ifTrue: ["color map is required to smooth non-RGB dest"
					^ self primitiveFail].
		pix := Array new: n*n].

	nSteps := height-1 max: 1.
	deltaP12 := (self deltaFrom: p1x to: p2x nSteps: nSteps)
			@ (self deltaFrom: p1y to: p2y nSteps: nSteps).
	pA := (self startFrom: p1x to: p2x offset: nSteps*deltaP12 x)
		@ (self startFrom: p1y to: p2y offset: nSteps*deltaP12 y).
	deltaP43 := (self deltaFrom: p4x to: p3x nSteps: nSteps)
			@ (self deltaFrom: p4y to: p3y nSteps: nSteps).
	pB := (self startFrom: p4x to: p3x offset: nSteps*deltaP43 x)
		@ (self startFrom: p4y to: p3y offset: nSteps*deltaP43 y).

	picker := BitBlt current bitPeekerFromForm: sourceForm.
	poker := BitBlt current bitPokerToForm: destForm.
	poker clipRect: self clipRect.
	nSteps := width-1 max: 1.
	destY to: destY+height-1 do:
		[:y |
		deltaPAB := (self deltaFrom: pA x to: pB x nSteps: nSteps)
				@ (self deltaFrom: pA y to: pB y nSteps: nSteps).
		sp := (self startFrom: pA x to: pB x offset: nSteps*deltaPAB x)
			@ (self startFrom: pA y to: pB y offset: nSteps*deltaPAB x).
		destX to: destX+width-1 do:
			[:x | 
			n = 1
			ifTrue:
				[poker pixelAt: x@y
						put: (picker pixelAt: sp // fixedPtOne asPoint)]
			ifFalse:
				[0 to: n-1 do:
					[:dx | 0 to: n-1 do:
						[:dy |
						pix at: dx*n+dy+1 put:
								(picker pixelAt: sp
									+ (deltaPAB*dx//n)
									+ (deltaP12*dy//n)
										// fixedPtOne asPoint)]].
				poker pixelAt: x@y put: (self mixPix: pix
										sourceMap: sourceMap
										destMap: colorMap)].
			sp := sp + deltaPAB].
		pA := pA + deltaP12.
		pB := pB + deltaP43]! !

Object subclass: #Color
	instanceVariableNames: 'rgb cachedDepth cachedBitPattern'
	classVariableNames: 'Black Blue BlueShift Brown CachedColormaps ColorChart ColorNames ComponentMask ComponentMax Cyan DarkGray Gray GrayToIndexMap Green GreenShift HalfComponentMask HighLightBitmaps IndexedColors LightBlue LightBrown LightCyan LightGray LightGreen LightMagenta LightOrange LightRed LightYellow Magenta MaskingMap Orange PaleBlue PaleBuff PaleGreen PaleMagenta PaleOrange PalePeach PaleRed PaleTan PaleYellow PureBlue PureCyan PureGreen PureMagenta PureRed PureYellow RandomStream Red RedShift TranslucentPatterns Transparent VeryDarkGray VeryLightGray VeryPaleRed VeryVeryDarkGray VeryVeryLightGray White Yellow'
	poolDictionaries: ''
	category: 'Graphics-Primitives'!

!Color commentStamp: '<historical>' prior: 0!
This class represents abstract color, regardless of the depth of bitmap it will be shown in.  At the very last moment a Color is converted to a pixelValue that depends on the depth of the actual Bitmap inside the Form it will be used with.  The supported depths (in bits) are 1, 2, 4, 8, 16, and 32.  The number of actual colors at these depths are: 2, 4, 16, 256, 32768, and 16 million.  (See comment in BitBlt.)  To change the depth of the Display and set how many colors you can see, execute: (Display newDepth: 8).  (See comment in DisplayMedium)
	Color is represented as the amount of light in red, green, and blue.  White is (1.0, 1.0, 1.0) and black is (0, 0, 0).  Pure red is (1.0, 0, 0).  These colors are "additive".  Think of Color's instance variables as:
	r	amount of red, a Float between 0.0 and 1.0.
	g	amount of green, a Float between 0.0 and 1.0.
	b	amount of blue, a Float between 0.0 and 1.0.
(But, in fact, the three are encoded as values from 0 to 1023 and combined in a single integer, rgb.  The user does not need to know this.)
	Many colors are named.  You find a color by name by sending a message to class Color, for example (Color lightBlue).  Also, (Color red: 0.2 green: 0.6 blue: 1.0) or (Color r: 0.2 g: 0.6 b: 1.0) creates a color. (see below)
	A color is essentially immutable.  Once you set red, green, and blue, you cannot change them.  Instead, create a new Color and use it.
	Applications such as contour maps and bar graphs will want to display one of a set of shades based on a number.  Convert the range of this number to an integer from 1 to N.  Then call (Color green lightShades: N) to get an Array of colors from white to green.  Use the Array messages at:, atPin:, or atWrap: to pull out the correct color from the array.  atPin: gives the first (or last) color if the index is out of range.  atWrap: wraps around to the other end if the index is out of range.
	Here are some fun things to run in when your screen has color:
		Pen new mandala: 30 diameter: Display height-100.
		Pen new web  "Draw with the mouse, opt-click to end"
		Display fillWhite.  Pen new hilberts: 5.
		Form toothpaste: 30  "Draw with mouse, opt-click to end"
You might also want to try the comment in
	Form>class>examples>tinyText...


Messages:
	mixed: proportion with: aColor	Answer this color mixed with the given color additively. The proportion, a number between 0.0 and 1.0, determines what what fraction of the receiver to use in the mix.

	+ 	add two colors
	- 	subtract two colors
	*	multiply the values of r, g, b by a number or an Array of factors.  ((Color named: #white) * 0.3) gives a darkish gray.  (aColor * #(0 0 0.9)) gives a color with slightly less blue.
	/	divide a color by a factor or an array of three factors.

	errorForDepth: d     How close the nearest color at this depth is to this abstract color.  Sum of the squares of the RGB differences, square rooted and normalized to 1.0.  Multiply by 100 to get percent.

	hue			Returns the hue of the color. On a wheel from 0 to 360 with pure red at 0 and again at 360.
	saturation	Returns the saturation of the color.  0.0 to 1.0
	brightness	Returns the brightness of the color.  0.0 to 1.0

	name    Look to see if this Color has a name.
	display	Show a swatch of this color tracking the cursor.

	lightShades: thisMany		An array of thisMany colors from white to the receiver. 
	darkShades: thisMany		An array of thisMany colors from black to the receiver.  Array is of length num.
	mix: color2 shades: thisMany		An array of thisMany colors from the receiver to color2.
	wheel: thisMany			An array of thisMany colors around the color wheel starting and ending at the receiver.

	pixelValueForDepth: d    Returns the bits that appear be in a Bitmap of this depth for this color.  Represents the nearest available color at this depth.  Normal users do not need to know which pixelValue is used for which color. 

Messages to Class Color.
	red: r green: g blue: b		Return a color with the given r, g, and b components.
	r: g: b:		Same as above, for fast typing.

 	hue: h saturation: s brightness: b		Create a color with the given hue, saturation, and brightness.

	pink
 	blue
	red ...	Many colors have messages that return an instance of Color.
	canUnderstand: #brown	  Returns true if #brown is a defined color.
	names		An OrderedCollection of the names of the colors.
	named: #notAllThatGray put: aColor    Add a new color to the list and create an access message and a class variable for it.
	fromUser	Shows the palette of colors available at this display depth.  Click anywhere to return the color you clicked on.

	hotColdShades: thisMany	An array of thisMany colors showing temperature from blue to red to white hot.

    stdColorsForDepth: d        An Array of colors available at this depth.  For 16 bit and 32 bits, returns a ColorGenerator.  It responds to at: with a Color for that index, simulating a very big Array. 

   colorFromPixelValue: value depth: d    Returns a Color whose bit pattern (inside a Bitmap) at this depth is the number specified.  Normal users do not need to use this.

(See also comments in these classes: Form, Bitmap, BitBlt, Pattern, MaskedForm.)!

!Color class methodsFor: 'named colors' stamp: ''!
black
	^Black! !

!Color class methodsFor: 'named colors' stamp: ''!
blue
	^Blue! !

!Color class methodsFor: 'named colors' stamp: ''!
brown
	^Brown! !

!Color class methodsFor: 'colormaps' stamp: 'jm 5/2/1999 07:24'!
cachedColormapFrom: sourceDepth to: destDepth
	"Return a cached colormap for mapping between the given depths. Always return a real colormap, not nil; this allows the client to get an identity colormap that can then be copied and modified to do color transformations."
	"Note: This method returns a shared, cached colormap to save time and space. Clients that need to modify a colormap returned by this method should make a copy and modify that!!"
	"Note: The colormap cache may be cleared by evaluating 'Color shutDown'."

	| srcIndex map |
	CachedColormaps class == Array ifFalse: [CachedColormaps := (1 to: 9) collect: [:i | Array new: 32]].
	srcIndex := sourceDepth.
	sourceDepth > 8 ifTrue: [srcIndex := 9].
	(map := (CachedColormaps at: srcIndex) at: destDepth) ~~ nil ifTrue: [^ map].

	map := self computeColormapFrom: sourceDepth to: destDepth.
	(CachedColormaps at: srcIndex) at: destDepth put: map.
	^ map
! !

!Color class methodsFor: 'instance creation' stamp: 'ar 4/10/2005 18:45'!
colorFrom: parm
	"Return an instantiated color from parm.  If parm is already a color, return it, else return the result of my performing it if it's a symbol or, if it is a list, it can either be an array of three numbers, which will be interpreted as RGB values, or a list of symbols, the first of which is sent to me and then the others of which are in turn sent to the prior result, thus allowing entries of the form #(blue darker).  Else just return the thing"

	| aColor firstParm |
	(parm isKindOf: Color) ifTrue: [^ parm].
	(parm isSymbol) ifTrue: [^ self perform: parm].
	(parm isString) ifTrue: [^ self fromString: parm].
	((parm isKindOf: SequenceableCollection) and: [parm size > 0])
		ifTrue:
			[firstParm := parm first.
			(firstParm isKindOf: Number) ifTrue:
				[^ self fromRgbTriplet: parm].
			aColor := self colorFrom: firstParm.
			parm doWithIndex:
				[:sym :ind | ind > 1 ifTrue:
					[aColor := aColor perform: sym]].
			^ aColor].
	^ parm

"
Color colorFrom: #(blue darker)
Color colorFrom: Color blue darker
Color colorFrom: #blue
Color colorFrom: #(0.0 0.0 1.0)
"! !

!Color class methodsFor: 'instance creation' stamp: 'tk 8/15/2001 11:03'!
colorFromPixelValue: p depth: d
	"Convert a pixel value for the given display depth into a color."
	"Details: For depths of 8 or less, the pixel value is simply looked up in a table. For greater depths, the color components are extracted and converted into a color."

	| r g b alpha |
	d = 8 ifTrue: [^ IndexedColors at: (p bitAnd: 16rFF) + 1].
	d = 4 ifTrue: [^ IndexedColors at: (p bitAnd: 16r0F) + 1].
	d = 2 ifTrue: [^ IndexedColors at: (p bitAnd: 16r03) + 1].
	d = 1 ifTrue: [^ IndexedColors at: (p bitAnd: 16r01) + 1].

	(d = 16) | (d = 15) ifTrue: [
		"five bits per component"
		r := (p bitShift: -10) bitAnd: 16r1F.
		g := (p bitShift: -5) bitAnd: 16r1F.
		b := p bitAnd: 16r1F.
		(r = 0 and: [g = 0]) ifTrue: [
			b = 0 ifTrue: [^Color transparent].
			b = 1 ifTrue: [^Color black]].
		^ Color r: r g: g b: b range: 31].

	d = 32 ifTrue: [
		"eight bits per component; 8 bits of alpha"
		r := (p bitShift: -16) bitAnd: 16rFF.
		g := (p bitShift: -8) bitAnd: 16rFF.
		b := p bitAnd: 16rFF.
		alpha := p bitShift: -24.
		alpha = 0 ifTrue: [^Color transparent].
		(r = 0 and: [g = 0 and: [b = 0]])  ifTrue: [^Color transparent].
		alpha < 255
			ifTrue: [^ (Color r: r g: g b: b range: 255) alpha: (alpha asFloat / 255.0)]
			ifFalse: [^ (Color r: r g: g b: b range: 255)]].

	d = 12 ifTrue: [
		"four bits per component"
		r := (p bitShift: -8) bitAnd: 16rF.
		g := (p bitShift: -4) bitAnd: 16rF.
		b := p bitAnd: 16rF.
		^ Color r: r g: g b: b range: 15].

	d = 9 ifTrue: [
		"three bits per component"
		r := (p bitShift: -6) bitAnd: 16r7.
		g := (p bitShift: -3) bitAnd: 16r7.
		b := p bitAnd: 16r7.
		^ Color r: r g: g b: b range: 7].

	self error: 'unknown pixel depth: ', d printString
! !

!Color class methodsFor: 'colormaps' stamp: ''!
colorMapIfNeededFrom: sourceDepth to: destDepth
	"Return a colormap for mapping between the given depths, or nil if no colormap is needed."
	"Note: This method returns a shared, cached colormap to save time and space. Clients that need to modify a colormap returned by this method should make a copy and modify that!!"

	sourceDepth = destDepth ifTrue: [^ nil].  "not needed if depths are the same"

	(sourceDepth >= 16) & (destDepth >= 16) ifTrue: [
		"mapping is done in BitBlt by zero-filling or truncating each color component"
		^ nil].

	^ Color cachedColormapFrom: sourceDepth to: destDepth
! !

!Color class methodsFor: 'other' stamp: ''!
colorNames
	"Return a collection of color names."

	^ ColorNames! !

!Color class methodsFor: 'examples' stamp: ''!
colorRampForDepth: depth extent: aPoint
	"Returns a form of the given size showing R, G, B, and gray ramps for the given depth. Useful for testing color conversions between different depths."
	"(Color colorRampForDepth: Display depth extent: 256@80) display"
	"(Color colorRampForDepth: 32 extent: 256@80) displayOn: Display at: 0@0 rule: Form paint"

	| f dx dy r |
	f := Form extent: aPoint depth: depth.
	dx := aPoint x // 256.
	dy := aPoint y // 4.
	0 to: 255 do: [:i |
		r := (dx * i)@0 extent: dx@dy.
		f fill: r fillColor: (Color r: i g: 0 b: 0 range: 255).
		r := r translateBy: 0@dy.
		f fill: r fillColor: (Color r: 0 g: i b: 0 range: 255).
		r := r translateBy: 0@dy.
		f fill: r fillColor: (Color r: 0 g: 0 b: i range: 255).
		r := r translateBy: 0@dy.
		f fill: r fillColor: (Color r: i g: i b: i range: 255)].
	^ f
! !

!Color class methodsFor: 'color from user' stamp: 'jm 1/19/1999 11:33'!
colorTest: depth extent: chartExtent colorMapper: colorMapper
	"Create a palette of colors sorted horizontally by hue and vertically by lightness. Useful for eyeballing the color gamut of the display, or for choosing a color interactively."
	"Note: It is slow to build this palette, so it should be cached for quick access."
	"(Color colorTest: 32 extent: 570@180 colorMapper: [:c | c]) display"
	"(Color colorTest: 32 extent: 570@180 colorMapper:
		[:c | Color
			r: (c red * 7) asInteger / 7
			g: (c green * 7) asInteger / 7
			b: (c blue * 3) asInteger / 3]) display"
	"(Color colorTest: 32 extent: 570@180 colorMapper:
		[:c | Color
			r: (c red * 5) asInteger / 5
			g: (c green * 5) asInteger / 5
			b: (c blue * 5) asInteger / 5]) display"
	"(Color colorTest: 32 extent: 570@180 colorMapper:
		[:c | Color
			r: (c red * 15) asInteger / 15
			g: (c green * 15) asInteger / 15
			b: (c blue * 15) asInteger / 15]) display"
	"(Color colorTest: 32 extent: 570@180 colorMapper:
		[:c | Color
			r: (c red * 31) asInteger / 31
			g: (c green * 31) asInteger / 31
			b: (c blue * 31) asInteger / 31]) display"

	| basicHue x y c startHue palette transHt vSteps transCaption grayWidth hSteps |
	palette := Form extent: chartExtent depth: depth.
	transCaption := "(DisplayText text: 'no color' asText textStyle: (TextConstants at: #ComicPlain)) form storeString"
		(Form extent: 34@9 depth: 1
			fromArray: #(0 0 256 0 256 0 3808663859 2147483648 2491688266 2147483648 2491688266 0 2491688266 0 2466486578 0 0 0)
			offset: 0@0).
	transHt := transCaption height.
	palette fillWhite: (0@0 extent: palette width@transHt).
	palette fillBlack: (0@transHt extent: palette width@1).
	transCaption displayOn: palette at: palette boundingBox topCenter - ((transCaption width // 2)@0).
	grayWidth := 10.
	startHue := 338.0.
	vSteps := palette height - transHt // 2.
	hSteps := palette width - grayWidth.
	x := 0.
	startHue to: startHue + 360.0 by: 360.0/hSteps do: [:h |
		basicHue := Color h: h asFloat s: 1.0 v: 1.0.
		y := transHt+1.
		0 to: vSteps do: [:n |
 			c := basicHue mixed: (n asFloat / vSteps asFloat) with: Color white.
			c := colorMapper value: c.
			palette fill: (x@y extent: 1@1) fillColor: c.
			y := y + 1].
		1 to: vSteps do: [:n |
 			c := Color black mixed: (n asFloat / vSteps asFloat) with: basicHue.
			c := colorMapper value: c.
			palette fill: (x@y extent: 1@1) fillColor: c.
			y := y + 1].
		x := x + 1].
	y := transHt + 1.
	1 to: vSteps * 2 do: [:n |
 		c := Color black mixed: (n asFloat / (vSteps*2) asFloat) with: Color white.
		c := colorMapper value: c.
		palette fill: (x@y extent: 10@1) fillColor: c.
		y := y + 1].
	^ palette
! !

!Color class methodsFor: 'colormaps' stamp: 'jmv 8/2/2009 21:32'!
computeColorConvertingMap: targetColor from: sourceDepth to: destDepth keepSubPixelAA: keepSubPix

	sourceDepth < 16 ifTrue: [
		"source is 1-, 2-, 4-, or 8-bit indexed color.
		Assumed not to include subpixelAA"
		^ self computeIndexedColorConvertingMap: targetColor from: sourceDepth to: destDepth
	] ifFalse: [
		"source is 16-bit or 32-bit RGB.
		Might include subpixelAA"
		^ self computeRGBColorConvertingMap: targetColor to: destDepth keepSubPixelAA: keepSubPix
	]! !

!Color class methodsFor: 'colormaps' stamp: 'jm 3/25/1999 19:48'!
computeColormapFrom: sourceDepth to: destDepth
	"Compute a colorMap for translating between the given depths. A colormap is a Bitmap whose entries contain the pixel values for the destination depth. Typical clients use cachedColormapFrom:to: instead."

	| map bitsPerColor |
	sourceDepth < 16 ifTrue: [
		"source is 1-, 2-, 4-, or 8-bit indexed color"
		map := (IndexedColors copyFrom: 1 to: (1 bitShift: sourceDepth))
					collect: [:c | c pixelValueForDepth: destDepth].
		map := map as: Bitmap.
	] ifFalse: [
		"source is 16-bit or 32-bit RGB"
		destDepth > 8
			ifTrue: [bitsPerColor := 5]  "retain maximum color resolution"
			ifFalse: [bitsPerColor := 4].
		map := self computeRGBColormapFor: destDepth bitsPerColor: bitsPerColor].

	"Note: zero is transparent except when source depth is one-bit deep"
	sourceDepth > 1 ifTrue: [map at: 1 put: 0].
	^ map
! !

!Color class methodsFor: 'colormaps' stamp: 'jmv 8/2/2009 21:31'!
computeIndexedColorConvertingMap: targetColor from: sourceDepth to: destDepth
	| map f c |
	
	map _ (IndexedColors copyFrom: 1 to: (1 bitShift: sourceDepth)) collect: [ :cc | 
		f _ 1.0 - (cc red + cc green + cc blue / 3.0 ).
		c _ targetColor notNil
			ifTrue: [
				destDepth = 32
					ifTrue: [ targetColor * f alpha: f]
					ifFalse: [ targetColor alphaMixed: f*1.5 with: Color white ]]
			ifFalse: [ cc ].
		destDepth = 32
			ifTrue: [ c pixelValueForDepth: destDepth]
			ifFalse: [
				f = 0.0
					ifTrue: [ 0 ]
					ifFalse: [ c pixelValueForDepth: destDepth ]]].
	map _ map as: Bitmap.
	^map! !

!Color class methodsFor: 'colormaps' stamp: 'jmv 8/2/2009 21:32'!
computeRGBColorConvertingMap: targetColor to: destDepth keepSubPixelAA: keepSubPix
	"Builds a colormap intended to convert from subpixelAA black values to targetColor values.
	keepSubPix
		ifTrue: [ Answer colors that also include subpixelAA ]
		ifFalse: [ 
			Take fullpixel luminance level. Apply it to targetColor.
			I.e. answer colors with NO subpixelAA ]"

	| mask map c bitsPerColor r g b f v |

	destDepth > 8
		ifTrue: [bitsPerColor _ 5]  "retain maximum color resolution"
		ifFalse: [bitsPerColor _ 4].
	"Usually a bit less is enough, but make it configurable"
	bitsPerColor _ bitsPerColor min: Preferences aaFontsColormapDepth.
	mask _ (1 bitShift: bitsPerColor) - 1.
	map _ Bitmap new: (1 bitShift: (3 * bitsPerColor)).
	0 to: map size - 1 do: [:i |
		r _ (i bitShift: 0 - (2 * bitsPerColor)) bitAnd: mask.
		g _ (i bitShift: 0 - bitsPerColor) bitAnd: mask.
		b _ (i bitShift: 0) bitAnd: mask.
		f _ 1.0 - (r + g + b / 3.0 / mask).
		c _ targetColor notNil
			ifTrue: [
				(keepSubPix and: [destDepth > 8]) ifTrue: [
						Color
							r: 1.0 - (r/mask) * targetColor red
							g: 1.0 - (g/mask) * targetColor green
							b: 1.0 - (b/mask) * targetColor blue ]
				ifFalse: [
					destDepth = 32
						ifTrue: [ targetColor * f alpha: f ]
						ifFalse: [ targetColor alphaMixed: f*1.5 with: Color white ]]]
			ifFalse: [ Color r: r g: g b: b range: mask].
		v _ destDepth = 32
			ifTrue: [ c pixelValueForDepth: destDepth]
			ifFalse: [
				f < 0.1
					ifTrue: [ 0 ]
					ifFalse: [ c pixelValueForDepth: destDepth ]].
		map at: i + 1 put: v ].
	^ map! !

!Color class methodsFor: 'colormaps' stamp: 'jm 12/4/97 15:25'!
computeRGBColormapFor: destDepth bitsPerColor: bitsPerColor
	"Compute a colorMap for translating from 16-bit or 32-bit RGB color to the given depth, using the given number of of bits per color component."

	| mask map c |
	(#(3 4 5) includes: bitsPerColor)
		ifFalse: [self error: 'BitBlt only supports 3, 4, or 5 bits per color component'].
	mask := (1 bitShift: bitsPerColor) - 1.
	map := Bitmap new: (1 bitShift: (3 * bitsPerColor)).
	0 to: map size - 1 do: [:i |
		c := Color
			r: ((i bitShift: 0 - (2 * bitsPerColor)) bitAnd: mask)
			g: ((i bitShift: 0 - bitsPerColor) bitAnd: mask)
			b: ((i bitShift: 0) bitAnd: mask)
			range: mask.
		map at: i + 1 put: (c pixelValueForDepth: destDepth)].

	map at: 1 put: (Color transparent pixelWordForDepth: destDepth).  "zero always transparent"
	^ map
! !

!Color class methodsFor: 'named colors' stamp: ''!
cyan
	^Cyan! !

!Color class methodsFor: 'named colors' stamp: ''!
darkGray
	^DarkGray! !

!Color class methodsFor: 'instance creation' stamp: 'mir 7/21/1999 11:54'!
fromArray: colorDef
	colorDef size == 3
			ifTrue: [^self r: (colorDef at: 1) g: (colorDef at: 2) b: (colorDef at: 3)].
	colorDef size == 0
			ifTrue: [^Color transparent].
	colorDef size == 4
			ifTrue: [^(TranslucentColor r: (colorDef at: 1) g: (colorDef at: 2) b: (colorDef at: 3)) alpha: (colorDef at: 4)].
	self error: 'Undefined color definition'! !

!Color class methodsFor: 'instance creation' stamp: 'sw 8/8/97 22:03'!
fromRgbTriplet: list
	^ self r: list first g: list second b: list last! !

!Color class methodsFor: 'instance creation' stamp: 'dvf 6/16/2000 17:48'!
fromString: aString
	"for HTML color spec: #FFCCAA or white/black"
	"Color fromString: '#FFCCAA'.
	 Color fromString: 'white'.
	 Color fromString: 'orange'"
	| aColorHex red green blue |
	aString isEmptyOrNil ifTrue: [^ Color white].
	aString first = $#
		ifTrue: [aColorHex := aString copyFrom: 2 to: aString size]
		ifFalse: [aColorHex := aString].
	[aColorHex size = 6
		ifTrue:
			[aColorHex := aColorHex asUppercase.
			red := ('16r', (aColorHex copyFrom: 1 to: 2)) asNumber/255.
			green := ('16r', (aColorHex copyFrom: 3 to: 4)) asNumber/255.
			blue := ('16r', (aColorHex copyFrom: 5 to: 6)) asNumber/255.
			^ self r: red g: green b: blue]]
	ifError: [:err :rcvr | "not a hex color triplet" ].
	
	"try to match aColorHex with known named colors"
	aColorHex := aColorHex asLowercase.

	^self perform: (ColorNames detect: [:i | i asString asLowercase = aColorHex]
		ifNone: [#white])! !

!Color class methodsFor: 'color from user' stamp: 'di 4/13/1999 14:30'!
fromUser
	"Displays a color palette of colors, waits for a mouse click, and returns the selected color. Any pixel on the Display can be chosen, not just those in the color palette."
	"Note: Since the color chart is cached, you may need to do 'ColorChart := nil' after changing the oldColorPaletteForDepth:extent: method."
	"Color fromUser"

	| d startPt save tr oldColor c here s |
	d := Display depth.
	((ColorChart == nil) or: [ColorChart depth ~= Display depth]) 
		ifTrue: [ColorChart := self oldColorPaletteForDepth: d extent: (2 * 144)@80].
	Sensor cursorPoint y < Display center y 
		ifTrue: [startPt := 0@(Display boundingBox bottom - ColorChart height)]
		ifFalse: [startPt := 0@0].

	save := Form fromDisplay: (startPt extent: ColorChart extent).
	ColorChart displayAt: startPt.
	tr := ColorChart extent - (50@19) corner: ColorChart extent.
	tr := tr translateBy: startPt.

	oldColor := nil.
	[Sensor anyButtonPressed] whileFalse: [
		c := Display colorAt: (here := Sensor cursorPoint).
		(tr containsPoint: here)
			ifFalse: [Display fill: (0@61+startPt extent: 20@19) fillColor: c]
			ifTrue: [
				c := Color transparent.
				Display fill: (0@61+startPt extent: 20@19) fillColor: Color white].
		c = oldColor ifFalse: [
			Display fillWhite: (20@61 + startPt extent: 135@19).
			c isTransparent
				ifTrue: [s := 'transparent']
				ifFalse: [s := c shortPrintString.
						s := s copyFrom: 7 to: s size - 1].
			s displayAt: 20@61 + startPt.
			oldColor := c]].
	save displayAt: startPt.
	Sensor waitNoButton.
	^ c
! !

!Color class methodsFor: 'named colors' stamp: ''!
gray
	^Gray! !

!Color class methodsFor: 'instance creation' stamp: 'jm 12/4/97 13:05'!
gray: brightness
	"Return a gray shade with the given brightness in the range [0.0..1.0]."

	^ self basicNew setRed: brightness green: brightness blue: brightness
! !

!Color class methodsFor: 'named colors' stamp: ''!
green
	^Green! !

!Color class methodsFor: 'instance creation' stamp: ''!
h: hue s: saturation v: brightness
	"Create a color with the given hue, saturation, and brightness. Hue is given as the angle in degrees of the color on the color circle where red is zero degrees. Saturation and brightness are numbers in [0.0..1.0] where larger values are more saturated or brighter colors. For example, (Color h: 0 s: 1 v: 1) is pure red."
	"Note: By convention, brightness is abbreviated 'v' to to avoid confusion with blue."

	^ self basicNew setHue: hue saturation: saturation brightness: brightness! !

!Color class methodsFor: 'instance creation' stamp: 'dew 3/19/2002 23:49'!
h: h s: s v: v alpha: alpha

	^ (self h: h s: s v: v) alpha: alpha! !

!Color class methodsFor: 'other' stamp: 'BG 3/16/2005 08:18'!
hex: aFloat
	"Return an hexadecimal two-digits string between 00 and FF
	for a float between 0.0 and 1.0"
	| str |
	str := ((aFloat * 255) asInteger printStringHex) asLowercase.
	str size = 1 ifTrue: [^'0',str] ifFalse: [^str]! !

!Color class methodsFor: 'examples' stamp: 'tk 6/19/96'!
hotColdShades: thisMany
	"An array of thisMany colors showing temperature from blue to red to white hot.  (Later improve this by swinging in hue.)  "
	"Color showColors: (Color hotColdShades: 25)"

	| n s1 s2 s3 s4 s5 |
	thisMany < 5 ifTrue: [^ self error: 'must be at least 5 shades'].
	n := thisMany // 5.
	s1 := self white mix: self yellow shades: (thisMany - (n*4)).
	s2 := self yellow mix: self red shades: n+1.
	s2 := s2 copyFrom: 2 to: n+1.
	s3 := self red mix: self green darker shades: n+1.
	s3 := s3 copyFrom: 2 to: n+1.
	s4 := self green darker mix: self blue shades: n+1.
	s4 := s4 copyFrom: 2 to: n+1.
	s5 := self blue mix: self black shades: n+1.
	s5 := s5 copyFrom: 2 to: n+1.
	^ s1, s2, s3, s4, s5
! !

!Color class methodsFor: 'other' stamp: ''!
indexedColors

	^ IndexedColors! !

!Color class methodsFor: 'class initialization' stamp: ''!
initialize
	"Color initialize"

	"Details: Externally, the red, green, and blue components of color
	are floats in the range [0.0..1.0]. Internally, they are represented
	as integers in the range [0..ComponentMask] packing into a
	small integer to save space and to allow fast hashing and
	equality testing.

	For a general description of color representations for computer
	graphics, including the relationship between the RGB and HSV
	color models used here, see Chapter 17 of Foley and van Dam,
	Fundamentals of Interactive Computer Graphics, Addison-Wesley,
	1982."

	ComponentMask := 1023.
	HalfComponentMask := 512.  "used to round up in integer calculations"
	ComponentMax := 1023.0.  "a Float used to normalize components"
	RedShift := 20.
	GreenShift := 10.
	BlueShift := 0.

	PureRed		 := self r: 1 g: 0 b: 0.
	PureGreen	 := self r: 0 g: 1 b: 0.
	PureBlue	 := self r: 0 g: 0 b: 1.
	PureYellow	 := self r: 1 g: 1 b: 0.
	PureCyan	 := self r: 0 g: 1 b: 1.
	PureMagenta := self r: 1 g: 0 b: 1.

	RandomStream := Random new.

	self initializeIndexedColors.
	self initializeGrayToIndexMap.
	self initializeNames.
	self initializeHighLights.
! !

!Color class methodsFor: 'class initialization' stamp: ''!
initializeGrayToIndexMap
	"Build an array of gray values available in the 8-bit colormap. This array is indexed by a gray level between black (1) and white (256) and returns the pixel value for the corresponding gray level."
	"Note: This method must be called after initializeIndexedColors, since it uses IndexedColors."
	"Color initializeGrayToIndexMap"

	| grayLevels grayIndices c distToClosest dist indexOfClosest |
	"record the level and index of each gray in the 8-bit color table"
	grayLevels := OrderedCollection new.
	grayIndices := OrderedCollection new.
	"Note: skip the first entry, which is reserved for transparent"
	2 to: IndexedColors size do: [:i |
		c := IndexedColors at: i.
		c saturation = 0.0 ifTrue: [  "c is a gray"
			grayLevels add: (c privateBlue) >> 2.  "top 8 bits; R, G, and B are the same"
			grayIndices add: i - 1]].  "pixel values are zero-based"
	grayLevels := grayLevels asArray.
	grayIndices := grayIndices asArray.

	"for each gray level in [0..255], select the closest match"
	GrayToIndexMap := ByteArray new: 256.
	0 to: 255 do: [:level |
		distToClosest := 10000.  "greater than distance to any real gray"
		1 to: grayLevels size do: [:i |
			dist := (level - (grayLevels at: i)) abs.
			dist < distToClosest ifTrue: [
				distToClosest := dist.
				indexOfClosest := grayIndices at: i]].
		GrayToIndexMap at: (level + 1) put: indexOfClosest].
! !

!Color class methodsFor: 'class initialization' stamp: 'tk 6/22/96'!
initializeHighLights
	"Create a set of Bitmaps for quickly reversing areas of the screen without converting colors. "
	"Color initializeHighLights"

	| t |
	t := Array new: 32.
	t at: 1 put: (Bitmap with: 16rFFFFFFFF).
	t at: 2 put: (Bitmap with: 16rFFFFFFFF).
	t at: 4 put: (Bitmap with: 16r55555555).
	t at: 8 put: (Bitmap with: 16r7070707).
	t at: 16 put: (Bitmap with: 16rFFFFFFFF).
	t at: 32 put: (Bitmap with: 16rFFFFFFFF).
	HighLightBitmaps := t.
! !

!Color class methodsFor: 'class initialization' stamp: ''!
initializeIndexedColors
	"Build an array of colors corresponding to the fixed colormap used
	 for display depths of 1, 2, 4, or 8 bits."
	"Color initializeIndexedColors"

	| a index grayVal |
	a := Array new: 256.

	"1-bit colors (monochrome)"
	a at: 1 put: (Color r: 1.0 g: 1.0 b: 1.0).		"white or transparent"
	a at: 2 put: (Color r: 0.0 g: 0.0 b: 0.0).	"black"

	"additional colors for 2-bit color"
	a at: 3 put: (Color r: 1.0 g: 1.0 b: 1.0).	"opaque white"
	a at: 4 put: (Color r: 0.5 g: 0.5 b: 0.5).	"1/2 gray"

	"additional colors for 4-bit color"
	a at:  5 put: (Color r: 1.0 g: 0.0 b: 0.0).	"red"
	a at:  6 put: (Color r: 0.0 g: 1.0 b: 0.0).	"green"
	a at:  7 put: (Color r: 0.0 g: 0.0 b: 1.0).	"blue"
	a at:  8 put: (Color r: 0.0 g: 1.0 b: 1.0).	"cyan"
	a at:  9 put: (Color r: 1.0 g: 1.0 b: 0.0).	"yellow"
	a at: 10 put: (Color r: 1.0 g: 0.0 b: 1.0).	"magenta"

	a at: 11 put: (Color r: 0.125 g: 0.125 b: 0.125).		"1/8 gray"
	a at: 12 put: (Color r: 0.25 g: 0.25 b: 0.25).		"2/8 gray"
	a at: 13 put: (Color r: 0.375 g: 0.375 b: 0.375).		"3/8 gray"
	a at: 14 put: (Color r: 0.625 g: 0.625 b: 0.625).		"5/8 gray"
	a at: 15 put: (Color r: 0.75 g: 0.75 b: 0.75).		"6/8 gray"
	a at: 16 put: (Color r: 0.875 g: 0.875 b: 0.875).		"7/8 gray"

	"additional colors for 8-bit color"
	"24 more shades of gray (1/32 increments but not repeating 1/8 increments)"
	index := 17.
	1 to: 31 do: [:v |
		(v \\ 4) = 0 ifFalse: [
			grayVal := v / 32.0.
			a at: index put: (Color r: grayVal g: grayVal b: grayVal).
			index := index + 1]].

	"The remainder of color table defines a color cube with six steps
	 for each primary color. Note that the corners of this cube repeat
	 previous colors, but this simplifies the mapping between RGB colors
	 and color map indices. This color cube spans indices 40 through 255
	 (indices 41-256 in this 1-based array)."
	0 to: 5 do: [:r |
		0 to: 5 do: [:g |
			0 to: 5 do: [:b |
				index := 41 + ((36 * r) + (6 * b) + g).
				index > 256 ifTrue: [
					self error: 'index out of range in color table compuation'].
				a at: index put: (Color r: r g: g b: b range: 5)]]].

	IndexedColors := a.
! !

!Color class methodsFor: 'class initialization' stamp: 'dwh 7/7/1999 23:57'!
initializeNames
	"Name some colors."
	"Color initializeNames"

	ColorNames := OrderedCollection new.
	self named: #black put: (Color r: 0 g: 0 b: 0).
	self named: #veryVeryDarkGray put: (Color r: 0.125 g: 0.125 b: 0.125).
	self named: #veryDarkGray put: (Color r: 0.25 g: 0.25 b: 0.25).
	self named: #darkGray put: (Color r: 0.375 g: 0.375 b: 0.375).
	self named: #gray put: (Color r: 0.5 g: 0.5 b: 0.5).
	self named: #lightGray put: (Color r: 0.625 g: 0.625 b: 0.625).
	self named: #veryLightGray put: (Color r: 0.75 g: 0.75 b: 0.75).
	self named: #veryVeryLightGray put: (Color r: 0.875 g: 0.875 b: 0.875).
	self named: #white put: (Color r: 1.0 g: 1.0 b: 1.0).
	self named: #red put: (Color r: 1.0 g: 0 b: 0).
	self named: #yellow put: (Color r: 1.0 g: 1.0 b: 0).
	self named: #green put: (Color r: 0 g: 1.0 b: 0).
	self named: #cyan put: (Color r: 0 g: 1.0 b: 1.0).
	self named: #blue put: (Color r: 0 g: 0 b: 1.0).
	self named: #magenta put: (Color r: 1.0 g: 0 b: 1.0).
	self named: #brown put: (Color r: 0.6 g: 0.2 b: 0).
	self named: #orange put: (Color r: 1.0 g: 0.6 b: 0).
	self named: #lightRed put: (Color r: 1.0 g: 0.8 b: 0.8).
	self named: #lightYellow put: (Color r: 1.0 g: 1.0 b: 0.8).
	self named: #lightGreen put: (Color r: 0.8 g: 1.0 b: 0.6).
	self named: #lightCyan put: (Color r: 0.4 g: 1.0 b: 1.0).
	self named: #lightBlue put: (Color r: 0.8 g: 1.0 b: 1.0).
	self named: #lightMagenta put: (Color r: 1.0 g: 0.8 b: 1.0).
	self named: #lightBrown put: (Color r: 1.0 g: 0.6 b: 0.2).
	self named: #lightOrange put: (Color r: 1.0 g: 0.8 b: 0.4).
	self named: #transparent put: (TranslucentColor new alpha: 0.0).
	self named: #paleBuff put: (Color r: 254 g: 250 b: 235 range: 255).
	self named: #paleBlue put: (Color r: 222 g: 249 b: 254 range: 255).
	self named: #paleYellow put: (Color r: 255 g: 255 b: 217 range: 255).
	self named: #paleGreen put: (Color r: 223 g: 255 b: 213 range: 255).
	self named: #paleRed put: (Color r: 255 g: 230 b: 230 range: 255).
	self named: #veryPaleRed put: (Color r: 255 g: 242 b: 242 range: 255).
	self named: #paleTan put: (Color r: 235 g: 224 b: 199 range: 255).
	self named: #paleMagenta put: (Color r: 255 g: 230 b: 255 range: 255).
	self named: #paleOrange put: (Color r: 253 g: 237 b: 215 range: 255).
	self named: #palePeach put: (Color r: 255 g: 237 b: 213 range: 255).

! !

!Color class methodsFor: 'class initialization' stamp: 'ar 2/16/2000 21:56'!
initializeTranslucentPatterns
	"Color initializeTranslucentPatterns"
	| mask bits pattern patternList |
	TranslucentPatterns := Array new: 8.
	#(1 2 4 8) do:[:d|
		patternList := Array new: 5.
		mask := (1 bitShift: d) - 1.
		bits := 2 * d.
		[bits >= 32] whileFalse: [
			mask := mask bitOr: (mask bitShift: bits).  "double the length of mask"
			bits := bits + bits].
		"0% pattern"
		pattern := Bitmap with: 0 with: 0.
		patternList at: 1 put: pattern.
		"25% pattern"
		pattern := Bitmap with: mask with: 0.
		patternList at: 2 put: pattern.
		"50% pattern"
		pattern := Bitmap with: mask with: mask bitInvert32.
		patternList at: 3 put: pattern.
		"75% pattern"
		pattern := Bitmap with: mask with: 16rFFFFFFFF.
		patternList at: 4 put: pattern.
		"100% pattern"
		pattern := Bitmap with: 16rFFFFFFFF with: 16rFFFFFFFF.
		patternList at: 5 put: pattern.
		TranslucentPatterns at: d put: patternList.
	].! !

!Color class methodsFor: 'named colors' stamp: ''!
lightBlue
	^LightBlue! !

!Color class methodsFor: 'named colors' stamp: ''!
lightBrown
	^LightBrown! !

!Color class methodsFor: 'named colors' stamp: ''!
lightCyan
	^LightCyan! !

!Color class methodsFor: 'named colors' stamp: ''!
lightGray
	^LightGray! !

!Color class methodsFor: 'named colors' stamp: ''!
lightGreen
	^LightGreen! !

!Color class methodsFor: 'named colors' stamp: ''!
lightMagenta
	^LightMagenta! !

!Color class methodsFor: 'named colors' stamp: ''!
lightOrange
	^LightOrange! !

!Color class methodsFor: 'named colors' stamp: ''!
lightRed
	^LightRed! !

!Color class methodsFor: 'named colors' stamp: ''!
lightYellow
	^LightYellow! !

!Color class methodsFor: 'named colors' stamp: ''!
magenta
	^Magenta! !

!Color class methodsFor: 'other' stamp: 'di 3/29/1999 13:33'!
maskingMap: depth
	"Return a color map that maps all colors except transparent to words of all ones. Used to create a mask for a Form whose transparent pixel value is zero. Cache the most recently used map."

	| sizeNeeded |
	depth <= 8
		ifTrue: [sizeNeeded := 1 bitShift: depth]
		ifFalse: [sizeNeeded := 4096].

	(MaskingMap == nil or: [MaskingMap size ~= sizeNeeded]) ifTrue:
		[MaskingMap := Bitmap new: sizeNeeded withAll: 16rFFFFFFFF.
		MaskingMap at: 1 put: 0.  "transparent"].

	^ MaskingMap
! !

!Color class methodsFor: 'class initialization' stamp: 'tk 6/13/96'!
named: newName put: aColor
	"Add a new color to the list and create an access message and a class variable for it.  The name should start with a lowercase letter.  (The class variable will start with an uppercase letter.)  (Color colorNames) returns a list of all color names.  "
	| str cap sym accessor csym |
	(aColor isKindOf: self) ifFalse: [^ self error: 'not a Color'].
	str := newName asString.
	sym := str asSymbol.
	cap := str capitalized.
	csym := cap asSymbol.
	(self class canUnderstand: sym) ifFalse: [
		"define access message"
		accessor := str, (String with: Character cr with: Character tab), 			'^', cap.
		self class compile: accessor
			classified: 'named colors'].
	(self classPool includesKey: csym) ifFalse: [
		self addClassVarName: cap].
	(ColorNames includes: sym) ifFalse: [
		ColorNames add: sym].
	^ self classPool at: csym put: aColor! !

!Color class methodsFor: 'instance creation' stamp: ''!
new

	^ self r: 0.0 g: 0.0 b: 0.0! !

!Color class methodsFor: 'color from user' stamp: 'di 4/13/1999 14:28'!
oldColorPaletteForDepth: depth extent: paletteExtent
	"Returns a form of the given size showing a color palette for the given depth."
	"(Color oldColorPaletteForDepth: Display depth extent: 720@100) display"

	| c p f nSteps rect w h q |
	f := Form extent: paletteExtent depth: depth.
	f fill: f boundingBox fillColor: Color white.
	nSteps := depth > 8 ifTrue: [12] ifFalse: [6].
	w := paletteExtent x // (nSteps * nSteps).
	h := paletteExtent y - 20 // nSteps.
	0 to: nSteps-1 do: [:r |
		0 to: nSteps-1 do: [:g |
			0 to: nSteps-1 do: [:b |
				c := Color r: r g: g b: b range: nSteps - 1.
				rect := ((r * nSteps * w) + (b * w)) @ (g * h) extent: w@(h + 1).
				f fill: rect fillColor: c]]].
	q := Quadrangle origin: paletteExtent - (50@19) corner: paletteExtent.
	q displayOn: f.
	'Trans.' displayOn: f at: q origin + (9@1).

	w := ((paletteExtent x - q width - 130) // 64) max: 1.
	p := paletteExtent x - q width - (64 * w) - 1 @ (paletteExtent y - 19).
	0 to: 63 do:
		[:v | c := Color r: v g: v b: v range: 63.
		f fill: ((v * w)@0 + p extent: (w + 1)@19) fillColor: c].
	^ f
! !

!Color class methodsFor: 'named colors' stamp: ''!
orange
	^Orange! !

!Color class methodsFor: 'named colors' stamp: 'dwh 7/7/1999 23:56'!
paleBlue
	^PaleBlue! !

!Color class methodsFor: 'named colors' stamp: 'dwh 7/7/1999 23:56'!
paleBuff
	^PaleBuff! !

!Color class methodsFor: 'named colors' stamp: 'dwh 7/7/1999 23:56'!
paleGreen
	^PaleGreen! !

!Color class methodsFor: 'named colors' stamp: 'dwh 7/7/1999 23:56'!
paleMagenta
	^PaleMagenta! !

!Color class methodsFor: 'named colors' stamp: 'dwh 7/7/1999 23:56'!
paleOrange
	^PaleOrange! !

!Color class methodsFor: 'named colors' stamp: 'dwh 7/7/1999 23:56'!
palePeach
	^PalePeach! !

!Color class methodsFor: 'named colors' stamp: 'dwh 7/7/1999 23:56'!
paleRed
	^PaleRed! !

!Color class methodsFor: 'named colors' stamp: 'dwh 7/7/1999 23:56'!
paleTan
	^PaleTan! !

!Color class methodsFor: 'named colors' stamp: 'dwh 7/7/1999 23:56'!
paleYellow
	^PaleYellow! !

!Color class methodsFor: 'other' stamp: ''!
pixelScreenForDepth: depth
	"Return a 50% stipple containing alternating pixels of all-zeros and all-ones to be used as a mask at the given depth."

	| mask bits |
	mask := (1 bitShift: depth) - 1.
	bits := 2 * depth.
	[bits >= 32] whileFalse: [
		mask := mask bitOr: (mask bitShift: bits).  "double the length of mask"
		bits := bits + bits].
	^ Bitmap with: mask with: mask bitInvert32
! !

!Color class methodsFor: 'other' stamp: ''!
quickHighLight: depth
	"Quickly return a Bitblt-ready raw colorValue for highlighting areas.  6/22/96 tk"

	^ HighLightBitmaps at: depth! !

!Color class methodsFor: 'instance creation' stamp: 'jm 12/4/97 13:04'!
r: r g: g b: b
	"Return a color with the given r, g, and b components in the range [0.0..1.0]."

	^ self basicNew setRed: r green: g blue: b
! !

!Color class methodsFor: 'instance creation' stamp: ''!
r: r g: g b: b alpha: alpha

	^ (self r: r g: g b: b) alpha: alpha! !

!Color class methodsFor: 'instance creation' stamp: ''!
r: r g: g b: b range: range
	"Return a color with the given r, g, and b components specified as integers in the range [0..r]. This avoids the floating point arithmetic in the red:green:blue: message and is thus a bit faster for certain applications (such as computing a sequence of colors for a palette)."

	^ self basicNew setRed: r green: g blue: b range: range! !

!Color class methodsFor: 'instance creation' stamp: ''!
random
	"Return a random color that isn't too dark or under-saturated."

	^ self basicNew
		setHue: (360.0 * RandomStream next)
		saturation: (0.3 + (RandomStream next * 0.7))
		brightness: (0.4 + (RandomStream next * 0.6))! !

!Color class methodsFor: 'named colors' stamp: ''!
red
	^Red! !

!Color class methodsFor: 'examples' stamp: ''!
showColorCube
	"Show a 12x12x12 color cube."
	"Color showColorCube"

	0 to: 11 do: [:r |
		0 to: 11 do: [:g |
			0 to: 11 do: [:b |	
				Display fill: (((r*60) + (b*5)) @ (g*5) extent: 5@5)
					fillColor: (Color r: r g: g b: b range: 11)]]].
! !

!Color class methodsFor: 'examples' stamp: ''!
showColors: colorList
	"Display the given collection of colors across the top of the Display."

	| w r |
	w := Display width // colorList size.
	r := 0@0 extent: w@((w min: 30) max: 10).
	colorList do: [:c |
		Display fill: r fillColor: c.
		r := r translateBy: w@0].
! !

!Color class methodsFor: 'examples' stamp: ''!
showHSVPalettes
	"Shows a palette of hues, varying the saturation and brightness for each one. Best results are with depths 16 and 32."
	"Color showHSVPalettes"

	| left top c |
	left := top := 0.
	0 to: 179 by: 15 do: [:h |
		0 to: 10 do: [:s |
			left := (h * 4) + (s * 4).
			0 to: 10 do: [:v |
				c := Color h: h s: s asFloat / 10.0 v: v asFloat / 10.0.
				top := (v * 4).
				Display fill: (left@top extent: 4@4) fillColor: c.

				c := Color h: h + 180 s: s asFloat / 10.0 v: v asFloat / 10.0.
				top := (v * 4) + 50.
				Display fill: (left@top extent: 4@4) fillColor: c]]].
! !

!Color class methodsFor: 'examples' stamp: ''!
showHuesInteractively
	"Shows a palette of hues at a (saturation, brightness) point determined by the mouse position. Click the mouse button to exit and return the selected (saturation, brightness) point."
	"Color showHuesInteractively"

	| p s v |
	[Sensor anyButtonPressed] whileFalse: [
		p := Sensor cursorPoint.
		s := p x asFloat / 300.0.
		v := p y asFloat / 300.0.
		self showColors: (self wheel: 12 saturation: s brightness: v)].
	^ (s min: 1.0) @ (v min: 1.0)! !

!Color class methodsFor: 'other' stamp: ''!
shutDown
	"Color shutDown"

	ColorChart := nil.		"Palette of colors for the user to pick from"
	CachedColormaps := nil.	"Maps to translate between color depths"
	MaskingMap := nil.		"Maps all colors except transparent to black for creating a mask"
! !

!Color class methodsFor: 'named colors' stamp: 'wod 5/24/1998 01:56'!
tan
	^  Color r: 0.8 g: 0.8 b: 0.5! !

!Color class methodsFor: 'other' stamp: 'ar 2/16/2000 21:56'!
translucentMaskFor: alphaValue depth: d
	"Return a pattern representing a mask usable for stipple transparency"
	^(TranslucentPatterns at: d) at: ((alphaValue min: 1.0 max: 0.0) * 4) rounded + 1! !

!Color class methodsFor: 'named colors' stamp: ''!
transparent
	^Transparent! !

!Color class methodsFor: 'named colors' stamp: ''!
veryDarkGray
	^VeryDarkGray! !

!Color class methodsFor: 'named colors' stamp: ''!
veryLightGray
	^VeryLightGray! !

!Color class methodsFor: 'named colors' stamp: 'dwh 7/7/1999 23:56'!
veryPaleRed
	^VeryPaleRed! !

!Color class methodsFor: 'named colors' stamp: ''!
veryVeryDarkGray
	^VeryVeryDarkGray! !

!Color class methodsFor: 'named colors' stamp: ''!
veryVeryLightGray
	^VeryVeryLightGray! !

!Color class methodsFor: 'examples' stamp: ''!
wheel: thisMany
	"Return a collection of thisMany colors evenly spaced around the color wheel."
	"Color showColors: (Color wheel: 12)"

	^ Color wheel: thisMany saturation: 0.9 brightness: 0.7
! !

!Color class methodsFor: 'examples' stamp: ''!
wheel: thisMany saturation: s brightness: v
	"Return a collection of thisMany colors evenly spaced around the color wheel, all of the given saturation and brightness."
	"Color showColors: (Color wheel: 12 saturation: 0.4 brightness: 1.0)"
	"Color showColors: (Color wheel: 12 saturation: 0.8 brightness: 0.5)"

	^ (Color h: 0.0 s: s v: v) wheel: thisMany
! !

!Color class methodsFor: 'named colors' stamp: ''!
white
	^White! !

!Color class methodsFor: 'named colors' stamp: ''!
yellow
	^Yellow! !

!Color methodsFor: 'transformations' stamp: 'fbs 2/3/2005 13:09'!
* aNumberOrArray
	"Answer this color with its RGB multiplied by the given number, or
	 multiply this color's RGB values by the corresponding entries in the
	given array."
	"(Color brown * 2) display"
	"(Color brown * #(1 0 1)) display"
	| multipliers |
	multipliers := aNumberOrArray isCollection
		ifTrue: [aNumberOrArray]
		ifFalse:
			[Array
				with: aNumberOrArray
				with: aNumberOrArray
				with: aNumberOrArray].

	^ Color basicNew
		setPrivateRed: (self privateRed * multipliers first) asInteger
		green: (self privateGreen * multipliers second) asInteger
		blue: (self privateBlue * multipliers third) asInteger.! !

!Color methodsFor: 'transformations' stamp: 'di 11/2/97 14:05'!
+ aColor
	"Answer this color mixed with the given color in an additive color space.  "
	"(Color blue + Color green) display"

	^ Color basicNew
		setPrivateRed: self privateRed + aColor privateRed
		green: self privateGreen + aColor privateGreen
		blue: self privateBlue + aColor  privateBlue
! !

!Color methodsFor: 'transformations' stamp: 'di 11/2/97 14:05'!
- aColor
	"Answer aColor is subtracted from the given color in an additive color space.  "
	"(Color white - Color red) display"

	^ Color basicNew
		setPrivateRed: self privateRed - aColor privateRed
		green: self privateGreen - aColor privateGreen
		blue: self privateBlue - aColor  privateBlue
! !

!Color methodsFor: 'transformations' stamp: 'di 11/2/97 14:07'!
/ aNumber
	"Answer this color with its RGB divided by the given number. "
	"(Color red / 2) display"

	^ Color basicNew
		setPrivateRed: (self privateRed / aNumber) asInteger
		green: (self privateGreen / aNumber) asInteger
		blue: (self privateBlue / aNumber) asInteger
! !

!Color methodsFor: 'equality' stamp: 'di 1/6/1999 20:26'!
= aColor
	"Return true if the receiver equals the given color. This method handles TranslucentColors, too."

	aColor isColor ifFalse: [^ false].
	^ aColor privateRGB = rgb and:
		[aColor privateAlpha = self privateAlpha]
! !

!Color methodsFor: 'Morphic menu' stamp: 'dgd 10/17/2003 12:10'!
addFillStyleMenuItems: aMenu hand: aHand from: aMorph
	"Add the items for changing the current fill style of the receiver"
	aMenu add: 'change color...' translated target: self selector: #changeColorIn:event: argument: aMorph! !

!Color methodsFor: 'transformations' stamp: 'dew 3/19/2002 23:50'!
adjustBrightness: brightness
	"Adjust the relative brightness of this color. (lowest value is 0.005 so that hue information is not lost)"

	^ Color
		h: self hue
		s: self saturation
		v: (self brightness + brightness min: 1.0 max: 0.005)
		alpha: self alpha! !

!Color methodsFor: 'transformations' stamp: 'dew 3/19/2002 23:51'!
adjustSaturation: saturation brightness: brightness
	"Adjust the relative saturation and brightness of this color. (lowest value is 0.005 so that hue information is not lost)"

	^ Color
		h: self hue
		s: (self saturation + saturation min: 1.0 max: 0.005)
		v: (self brightness + brightness min: 1.0 max: 0.005)
		alpha: self alpha! !

!Color methodsFor: 'access' stamp: ''!
alpha
	"Return the opacity ('alpha') value of opaque so that normal colors can be compared to TransparentColors."

	^ 1.0
! !

!Color methodsFor: 'transformations' stamp: 'sma 6/25/2000 15:36'!
alpha: alphaValue 
	"Answer a new Color with the given amount of opacity ('alpha')."

	alphaValue = 1.0
		ifFalse: [^ TranslucentColor basicNew setRgb: rgb alpha: alphaValue]! !

!Color methodsFor: 'transformations' stamp: 'tk 7/4/2000 11:55'!
alphaMixed: proportion with: aColor 
	"Answer this color mixed with the given color. The proportion, a number 
	between 0.0 and 1.0, determines what what fraction of the receiver to  
	use in the mix. For example, 0.9 would yield a color close to the  
	receiver. This method uses RGB interpolation; HSV interpolation can lead 
	to surprises.  Mixes the alphas (for transparency) also."

	| frac1 frac2 |
	frac1 := proportion asFloat min: 1.0 max: 0.0.
	frac2 := 1.0 - frac1.
	^ Color
		r: self red * frac1 + (aColor red * frac2)
		g: self green * frac1 + (aColor green * frac2)
		b: self blue * frac1 + (aColor blue * frac2)
		alpha: self alpha * frac1 + (aColor alpha * frac2)! !

!Color methodsFor: 'conversions' stamp: 'ar 11/2/1998 12:19'!
asColor
	"Convert the receiver into a color"
	^self! !

!Color methodsFor: 'conversions' stamp: 'TBn 6/15/2000 20:37'!
asColorref
	"Convert the receiver into a colorref"
	^(self red * 255) asInteger + ((self green * 255) asInteger << 8) + ((self green * 255) asInteger << 16)! !

!Color methodsFor: 'conversions' stamp: 'st 9/27/2004 13:42'!
asHTMLColor
	^ '#', (self class hex: self red), (self class hex: self green), (self class hex: self blue)! !

!Color methodsFor: 'conversions' stamp: 'sw 10/27/1999 10:51'!
asNontranslucentColor
	^ self! !

!Color methodsFor: 'transformations' stamp: 'RAA 6/2/2000 08:47'!
atLeastAsLuminentAs: aFloat

	| revisedColor |
	revisedColor := self.
	[revisedColor luminance < aFloat] whileTrue: [revisedColor := revisedColor slightlyLighter].
	^revisedColor
! !

!Color methodsFor: 'transformations' stamp: 'nk 3/8/2004 09:43'!
atMostAsLuminentAs: aFloat

	| revisedColor |
	revisedColor := self.
	[revisedColor luminance > aFloat] whileTrue: [revisedColor := revisedColor slightlyDarker].
	^revisedColor
! !

!Color methodsFor: 'private' stamp: ''!
attemptToMutateError
	"A color is immutable. Once a color's red, green, and blue have been initialized, you cannot change them. Instead, create a new Color and use it."

	self error: 'Color objects are immutable once created'
! !

!Color methodsFor: 'conversions' stamp: 'di 3/25/2000 10:13'!
balancedPatternForDepth: depth
	"A generalization of bitPatternForDepth: as it exists.  Generates a 2x2 stipple of color.
	The topLeft and bottomRight pixel are closest approx to this color"
	| pv1 pv2 mask1 mask2 pv3 c |
	(depth == cachedDepth and:[cachedBitPattern size = 2]) ifTrue: [^ cachedBitPattern].
	(depth between: 4 and: 16) ifFalse: [^ self bitPatternForDepth: depth].
	cachedDepth := depth.
	pv1 := self pixelValueForDepth: depth.
"
	Subtract error due to pv1 to get pv2.
	pv2 := (self - (err1 := (Color colorFromPixelValue: pv1 depth: depth) - self))
						pixelValueForDepth: depth.
	Subtract error due to 2 pv1's and pv2 to get pv3.
	pv3 := (self - err1 - err1 - ((Color colorFromPixelValue: pv2 depth: depth) - self))
						pixelValueForDepth: depth.
"
	"Above two statements computed faster by the following..."
	pv2 := (c := self - ((Color colorFromPixelValue: pv1 depth: depth) - self))
						pixelValueForDepth: depth.
	pv3 := (c + (c - (Color colorFromPixelValue: pv2 depth: depth)))
						pixelValueForDepth: depth.

	"Return to a 2-word bitmap that encodes a 2x2 stipple of the given pixelValues."
	mask1 := (#(- - -	
			16r01010101 - - -			"replicates every other 4 bits"
			16r00010001 - - - - - - -	"replicates every other 8 bits"
			16r00000001) at: depth).	"replicates every other 16 bits"
	mask2 := (#(- - -	
			16r10101010 - - -			"replicates the other 4 bits"
			16r01000100 - - - - - - -	"replicates the other 8 bits"
			16r00010000) at: depth).	"replicates the other 16 bits"
	^ cachedBitPattern := Bitmap with: (mask1*pv1) + (mask2*pv2) with: (mask1*pv3) + (mask2*pv1)! !

!Color methodsFor: 'queries' stamp: 'sw 9/27/2001 17:26'!
basicType
	"Answer a symbol representing the inherent type of the receiver"

	^ #Color! !

!Color methodsFor: 'conversions' stamp: 'hmm 4/25/2000 09:40'!
bitPatternForDepth: depth
	"Return a Bitmap, possibly containing a stipple pattern, that best represents this color at the given depth. BitBlt calls this method to convert colors into Bitmaps. The resulting Bitmap may be multiple words to represent a stipple pattern of several lines.  "
	"See also:	pixelValueAtDepth:	-- value for single pixel
				pixelWordAtDepth:	-- a 32-bit word filled with the pixel value"
	"Details: The pattern for the most recently requested depth is cached."
	"Note for depths > 2, there are stippled and non-stippled versions (generated with #balancedPatternForDepth: and #bitPatternForDepth:, respectively). The stippled versions don't work with the window bit caching of StandardSystemView, so we make sure that for these depths, only unstippled patterns are returned"

	(depth == cachedDepth and: [depth <= 2 or: [cachedBitPattern size = 1]]) ifTrue: [^ cachedBitPattern].
	cachedDepth := depth.

	depth > 2 ifTrue: [^ cachedBitPattern := Bitmap with: (self pixelWordForDepth: depth)].
	depth = 1 ifTrue: [^ cachedBitPattern := self halfTonePattern1].
	depth = 2 ifTrue: [^ cachedBitPattern := self halfTonePattern2].
! !

!Color methodsFor: 'transformations' stamp: 'dew 3/23/2002 01:38'!
blacker

	^ self alphaMixed: 0.8333 with: Color black
! !

!Color methodsFor: 'access' stamp: ''!
blue
	"Return the blue component of this color, a float in the range [0.0..1.0]."

	^ self privateBlue asFloat / ComponentMax! !

!Color methodsFor: 'access' stamp: ''!
brightness
	"Return the brightness of this color, a float in the range [0.0..1.0]."

	^ ((self privateRed max:
	    self privateGreen) max:
	    self privateBlue) asFloat / ComponentMax! !

!Color methodsFor: 'printing' stamp: 'MPW 1/1/1901 22:14'!
byteEncode: aStream

	aStream
		print: '(';
		print: self class name;
		print: ' r: ';
		write: (self red roundTo: 0.001);
		print: ' g: ';
		write: (self green roundTo: 0.001);
		print: ' b: ';
		write: (self blue roundTo: 0.001) ;
		print: ')'.
! !

!Color methodsFor: 'Morphic menu' stamp: 'ar 10/5/2000 18:50'!
changeColorIn: aMorph event: evt
	"Note: This is just a workaround to make sure we don't use the old color inst var"
	aMorph changeColorTarget: aMorph selector: #fillStyle: originalColor: self hand: evt hand! !

!Color methodsFor: 'conversions' stamp: ''!
closestPixelValue1
	"Return the nearest approximation to this color for a monochrome Form."

	"fast special cases"
	rgb = 0 ifTrue: [^ 1].  "black"
	rgb = 16r3FFFFFFF ifTrue: [^ 0].  "white"

	self luminance > 0.5
		ifTrue: [^ 0]  "white"
		ifFalse: [^ 1].  "black"
! !

!Color methodsFor: 'conversions' stamp: ''!
closestPixelValue2
	"Return the nearest approximation to this color for a 2-bit deep Form."

	| lum |
	"fast special cases"
	rgb = 0 ifTrue: [^ 1].  "black"
	rgb = 16r3FFFFFFF ifTrue: [^ 2].  "opaque white"

	lum := self luminance.
	lum < 0.2 ifTrue: [^ 1].  "black"
	lum > 0.6 ifTrue: [^ 2].  "opaque white"
	^ 3  "50% gray"
! !

!Color methodsFor: 'conversions' stamp: ''!
closestPixelValue4
	"Return the nearest approximation to this color for a 4-bit deep Form."

	| bIndex |
	"fast special cases"
	rgb = 0 ifTrue: [^ 1].  "black"
	rgb = 16r3FFFFFFF ifTrue: [^ 2].  "opaque white"

	rgb = PureRed privateRGB ifTrue: [^ 4].
	rgb = PureGreen privateRGB ifTrue: [^ 5].
	rgb = PureBlue privateRGB ifTrue: [^ 6].
	rgb = PureCyan privateRGB ifTrue: [^ 7].
	rgb = PureYellow privateRGB ifTrue: [^ 8].
	rgb = PureMagenta privateRGB ifTrue: [^ 9].

	bIndex := (self luminance * 8.0) rounded.  "bIndex in [0..8]"
	^ #(
		1	"black"
		10	"1/8 gray"
		11	"2/8 gray"
		12	"3/8 gray"
		3	"4/8 gray"
		13	"5/8 gray"
		14	"6/8 gray"
		15	"7/8 gray"
		2	"opaque white"
	) at: bIndex + 1.
! !

!Color methodsFor: 'conversions' stamp: ''!
closestPixelValue8
	"Return the nearest approximation to this color for an 8-bit deep Form."

	"fast special cases"
	rgb = 0 ifTrue: [^ 1].  "black"
	rgb = 16r3FFFFFFF ifTrue: [^ 255].  "white"

	self saturation < 0.2 ifTrue: [
		^ GrayToIndexMap at: (self privateGreen >> 2) + 1.  "nearest gray"
	] ifFalse: [
		"compute nearest entry in the color cube"
		^ 40 +
		  ((((self privateRed * 5) + HalfComponentMask) // ComponentMask) * 36) +
		  ((((self privateBlue * 5) + HalfComponentMask) // ComponentMask) * 6) +
		  (((self privateGreen * 5) + HalfComponentMask) // ComponentMask)].
! !

!Color methodsFor: 'other' stamp: 'sw 2/16/98 03:42'!
colorForInsets
	^ self! !

!Color methodsFor: 'transformations' stamp: 'dew 3/19/2002 23:54'!
dansDarker
	"Return a darker shade of the same color.
	An attempt to do better than the current darker method.
	(now obsolete, since darker has been changed to do this. -dew)"
	^ Color h: self hue s: self saturation
		v: (self brightness - 0.16 max: 0.0)! !

!Color methodsFor: 'groups of shades' stamp: 'tk 6/18/96'!
darkShades: thisMany
	"An array of thisMany colors from black to the receiver.  Array is of length num. Very useful for displaying color based on a variable in your program.  "
	"Color showColors: (Color red darkShades: 12)"

	^ self class black mix: self shades: thisMany
! !

!Color methodsFor: 'transformations' stamp: 'dew 3/4/2002 01:40'!
darker
	"Answer a darker shade of this color."

	^ self adjustBrightness: -0.08! !

!Color methodsFor: 'equality' stamp: 'di 9/27/2000 08:07'!
diff: theOther
	"Returns a number between 0.0 and 1.0"

	^ ((self privateRed - theOther privateRed) abs
		+ (self privateGreen - theOther privateGreen) abs
		+ (self privateBlue - theOther privateBlue) abs)
		/ 3.0 / ComponentMax! !

!Color methodsFor: 'other' stamp: 'tk 6/14/96'!
display
	"Show a swatch of this color tracking the cursor until the next mouseClick. "
	"Color red display"
	| f |
	f := Form extent: 40@20 depth: Display depth.
	f fillColor: self.
	Cursor blank showWhile:
		[f follow: [Sensor cursorPoint] while: [Sensor noButtonPressed]]! !

!Color methodsFor: 'conversions' stamp: 'di 9/2/97 20:21'!
dominantColor
	^ self! !

!Color methodsFor: 'transformations' stamp: 'dew 3/8/2002 00:13'!
duller

	^ self adjustSaturation: -0.03 brightness: -0.2! !

!Color methodsFor: 'private' stamp: ''!
flushCache
	"Flush my cached bit pattern."

	cachedDepth := nil.
	cachedBitPattern := nil.
! !

!Color methodsFor: 'access' stamp: ''!
green
	"Return the green component of this color, a float in the range [0.0..1.0]."

	^ self privateGreen asFloat / ComponentMax! !

!Color methodsFor: 'conversions' stamp: 'di 6/23/97 23:27'!
halfTonePattern1
	"Return a halftone-pattern to approximate luminance levels on 1-bit deep Forms."

	| lum |
	lum := self luminance.
	lum < 0.1 ifTrue: [^ Bitmap with: 16rFFFFFFFF]. "black"
	lum < 0.4 ifTrue: [^ Bitmap with: 16rBBBBBBBB with: 16rEEEEEEEE]. "dark gray"
	lum < 0.6 ifTrue: [^ Bitmap with: 16r55555555 with: 16rAAAAAAAA]. "medium gray"
	lum < 0.9 ifTrue: [^ Bitmap with: 16r44444444 with: 16r11111111]. "light gray"
	^ Bitmap with: 0  "1-bit white"
! !

!Color methodsFor: 'conversions' stamp: ''!
halfTonePattern2
	"Return a halftone-pattern to approximate luminance levels on 2-bit deep Forms."

	| lum |
	lum := self luminance.
	lum < 0.125 ifTrue: [^ Bitmap with: 16r55555555].  "black"
	lum < 0.25 ifTrue: [^ Bitmap with: 16r55555555 with: 16rDDDDDDDD].  "1/8 gray"
	lum < 0.375 ifTrue: [^ Bitmap with: 16rDDDDDDDD with: 16r77777777].  "2/8 gray"
	lum < 0.5 ifTrue: [^ Bitmap with: 16rFFFFFFFF with: 16r77777777].  "3/8 gray"
	lum < 0.625 ifTrue: [^ Bitmap with: 16rFFFFFFFF].  "4/8 gray"
	lum < 0.75 ifTrue: [^ Bitmap with: 16rFFFFFFFF with: 16rBBBBBBBB].  "5/8 gray"
	lum < 0.875 ifTrue: [^ Bitmap with: 16rEEEEEEEE with: 16rBBBBBBBB].  "6/8 gray"
	lum < 1.0 ifTrue: [^ Bitmap with: 16rAAAAAAAA with: 16rBBBBBBBB].  "7/8 gray"
	^ Bitmap with: 16rAAAAAAAA  "opaque white"

"handy expression for computing patterns for 2x2 tiles;
 set p to a string of 4 letters (e.g., 'wggw' for a gray-and-
 white checkerboard) and print the result of evaluating:
| p d w1 w2 |
p := 'wggw'.
d := Dictionary new.
d at: $b put: '01'.
d at: $w put: '10'.
d at: $g put: '11'.
w1 := (d at: (p at: 1)), (d at: (p at: 2)).
w1 := '2r', w1, w1, w1, w1, w1, w1, w1, w1, ' hex'.
w2 := (d at: (p at: 3)), (d at: (p at: 4)).
w2 := '2r', w2, w2, w2, w2, w2, w2, w2, w2, ' hex'.
Array with: (Compiler evaluate: w1) with: (Compiler evaluate: w2) 
"! !

!Color methodsFor: 'equality' stamp: ''!
hash

	^ rgb! !

!Color methodsFor: 'access' stamp: ''!
hue
	"Return the hue of this color, an angle in the range [0.0..360.0]."

	| r g b max min span h |
	r := self privateRed.
	g := self privateGreen.
	b := self privateBlue. 

	max := ((r max: g) max: b).
	min := ((r min: g) min: b).
	span := (max - min) asFloat.
	span = 0.0 ifTrue: [ ^ 0.0 ].

	r = max ifTrue: [
		h := ((g - b) asFloat / span) * 60.0.
	] ifFalse: [
		g = max
			ifTrue: [ h := 120.0 + (((b - r) asFloat / span) * 60.0). ]
			ifFalse: [ h := 240.0 + (((r - g) asFloat / span) * 60.0). ].
	].

	h < 0.0 ifTrue: [ h := 360.0 + h ].
	^ h! !

!Color methodsFor: 'conversions' stamp: 'tk 4/24/97'!
indexInMap: aColorMap
	"Return the index corresponding to this color in the given color map. RGB colors are truncated to 3-, 4-, or 5-bits per color component when indexing into such a colorMap.  "

	aColorMap size = 2 ifTrue: [^ (self pixelValueForDepth: 1) + 1].
	aColorMap size = 4 ifTrue: [^ (self pixelValueForDepth: 2) + 1].
	aColorMap size = 16 ifTrue: [^ (self pixelValueForDepth: 4) + 1].
	aColorMap size = 256 ifTrue: [^ (self pixelValueForDepth: 8) + 1].
	aColorMap size = 512 ifTrue: [^ (self pixelValueForDepth: 9) + 1].
	aColorMap size = 4096 ifTrue: [^ (self pixelValueForDepth: 12) + 1].
	aColorMap size = 32768 ifTrue: [^ (self pixelValueForDepth: 16) + 1].
	self error: 'unknown pixel depth'.
! !

!Color methodsFor: 'queries' stamp: 'ar 1/14/1999 15:27'!
isBitmapFill
	^false! !

!Color methodsFor: 'queries' stamp: 'ar 11/12/1998 19:43'!
isBlack
	"Return true if the receiver represents black"
	^rgb = 0! !

!Color methodsFor: 'queries' stamp: ''!
isColor

	^ true
! !

!Color methodsFor: 'queries' stamp: 'ar 6/18/1999 06:58'!
isGradientFill
	^false! !

!Color methodsFor: 'queries' stamp: 'ar 11/12/1998 19:44'!
isGray
	"Return true if the receiver represents a shade of gray"
	^(self privateRed = self privateGreen) and:[self privateRed = self privateBlue]! !

!Color methodsFor: 'queries' stamp: 'ar 4/20/2001 04:33'!
isOpaque
	^true! !

!Color methodsFor: 'queries' stamp: 'ar 6/18/1999 07:57'!
isOrientedFill
	"Return true if the receiver keeps an orientation (e.g., origin, direction, and normal)"
	^false! !

!Color methodsFor: 'self evaluating' stamp: 'sd 7/31/2005 21:46'!
isSelfEvaluating
	^ self class == Color! !

!Color methodsFor: 'queries' stamp: 'ar 11/7/1998 20:20'!
isSolidFill
	^true! !

!Color methodsFor: 'queries' stamp: 'di 12/30/1998 14:33'!
isTranslucent

	^ false
! !

!Color methodsFor: 'queries' stamp: 'di 1/3/1999 12:23'!
isTranslucentColor
	"This means: self isTranslucent, but isTransparent not"
	^ false! !

!Color methodsFor: 'queries' stamp: ''!
isTransparent

	^ false
! !

!Color methodsFor: 'groups of shades' stamp: 'tk 6/18/96'!
lightShades: thisMany
	"An array of thisMany colors from white to self. Very useful for displaying color based on a variable in your program.  "
	"Color showColors: (Color red lightShades: 12)"

	^ self class white mix: self shades: thisMany
! !

!Color methodsFor: 'transformations' stamp: 'dew 1/23/2002 20:19'!
lighter
	"Answer a lighter shade of this color."

	^ self adjustSaturation: -0.03 brightness: 0.08! !

!Color methodsFor: 'access' stamp: ''!
luminance
	"Return the luminance of this color, a brightness value weighted by the human eye's color sensitivity."

	^ ((299 * self privateRed) +
	   (587 * self privateGreen) +
	   (114 * self privateBlue)) / (1000 * ComponentMax)
! !

!Color methodsFor: 'conversions' stamp: 'bf 4/18/2001 16:25'!
makeForegroundColor
        "Make a foreground color contrasting with me"
        ^self luminance >= 0.5
                ifTrue: [Color black]
                ifFalse: [Color white]! !

!Color methodsFor: 'groups of shades' stamp: 'tk 6/18/96'!
mix: color2 shades: thisMany
	"Return an array of thisMany colors from self to color2. Very useful for displaying color based on a variable in your program.  "
	"Color showColors: (Color red mix: Color green shades: 12)"

	| redInc greenInc blueInc rr gg bb c out |
	thisMany = 1 ifTrue: [^ Array with: color2].
	redInc := color2 red - self red / (thisMany-1).
	greenInc := color2 green - self green / (thisMany-1).
	blueInc := color2 blue - self blue / (thisMany-1).
	rr := self red.  gg := self green.  bb := self blue.
	out := (1 to: thisMany) collect: [:num |
		c := Color r: rr g: gg b: bb.
		rr := rr + redInc.
		gg := gg + greenInc.
		bb := bb + blueInc.
		c].
	out at: out size put: color2.	"hide roundoff errors"
	^ out
! !

!Color methodsFor: 'transformations' stamp: 'tk 7/4/2000 12:00'!
mixed: proportion with: aColor 
	"Mix with another color and do not preserve transpareny.  Only use this for extracting the RGB value and mixing it.  All other callers should use instead: 
	aColor alphaMixed: proportion with: anotherColor
	"

	| frac1 frac2 |
	frac1 := proportion asFloat min: 1.0 max: 0.0.
	frac2 := 1.0 - frac1.
	^ Color
		r: self red * frac1 + (aColor red * frac2)
		g: self green * frac1 + (aColor green * frac2)
		b: self blue * frac1 + (aColor blue * frac2)! !

!Color methodsFor: 'transformations' stamp: 'dew 1/19/2002 01:29'!
muchDarker

	^ self alphaMixed: 0.5 with: Color black
! !

!Color methodsFor: 'transformations' stamp: 'tk 7/4/2000 12:07'!
muchLighter

	^ self alphaMixed: 0.233 with: Color white
! !

!Color methodsFor: 'other' stamp: 'jm 12/4/97 10:24'!
name
	"Return this color's name, or nil if it has no name. Only returns a name if it exactly matches the named color."

	ColorNames do:
		[:name | (Color perform: name) = self ifTrue: [^ name]].
	^ nil
! !

!Color methodsFor: 'transformations' stamp: 'ar 6/19/1999 00:36'!
negated
	"Return an RGB inverted color"
	^Color
		r: 1.0 - self red
		g: 1.0 - self green
		b: 1.0 - self blue! !

!Color methodsFor: 'transformations' stamp: 'di 9/27/2000 08:14'!
orColorUnlike: theOther
	"If this color is a lot like theOther, then return its complement, otherwide, return self"

	(self diff: theOther) < 0.3
		ifTrue: [^ theOther negated]
		ifFalse: [^ self]! !

!Color methodsFor: 'transformations' stamp: 'dew 3/4/2002 01:42'!
paler
	"Answer a paler shade of this color."

	^ self adjustSaturation: -0.09 brightness: 0.09
! !

!Color methodsFor: 'conversions' stamp: 'ar 5/15/2001 16:12'!
pixelValue32
	"Note: pixelWord not pixelValue so we include translucency"
	^self pixelWordForDepth: 32! !

!Color methodsFor: 'conversions' stamp: 'jm 1/26/2001 15:11'!
pixelValueForDepth: d
	"Returns an integer representing the bits that appear in a single pixel of this color in a Form of the given depth. The depth must be one of 1, 2, 4, 8, 16, or 32. Contrast with pixelWordForDepth: and bitPatternForDepth:, which return either a 32-bit word packed with the given pixel value or a multiple-word Bitmap containing a pattern. The inverse is the class message colorFromPixelValue:depth:"
	"Details: For depths of 8 or less, the result is a colorMap index. For depths of 16 and 32, it is a direct color value with 5 or 8 bits per color component."
	"Transparency: The pixel value zero is reserved for transparent. For depths greater than 8, black maps to the darkest possible blue."

	| rgbBlack val |
	d = 8 ifTrue: [^ self closestPixelValue8].  "common case"
	d < 8 ifTrue: [
		d = 4 ifTrue: [^ self closestPixelValue4].
		d = 2 ifTrue: [^ self closestPixelValue2].
		d = 1 ifTrue: [^ self closestPixelValue1]].

	rgbBlack := 1.  "closest black that is not transparent in RGB"

	d = 16 ifTrue: [
		"five bits per component; top bits ignored"
		val := (((rgb bitShift: -15) bitAnd: 16r7C00) bitOr:
			 ((rgb bitShift: -10) bitAnd: 16r03E0)) bitOr:
			 ((rgb bitShift: -5) bitAnd: 16r001F).
		^ val = 0 ifTrue: [rgbBlack] ifFalse: [val]].

	d = 32 ifTrue: [
		"eight bits per component; top 8 bits set to all ones (opaque alpha)"
		val := LargePositiveInteger new: 4.
		val at: 3 put: ((rgb bitShift: -22) bitAnd: 16rFF).
		val at: 2 put: ((rgb bitShift: -12) bitAnd: 16rFF).
		val at: 1 put: ((rgb bitShift: -2) bitAnd: 16rFF).
		val = 0 ifTrue: [val at: 1 put: 1].  "closest non-transparent black"
		val at: 4 put: 16rFF.  "opaque alpha"
		^ val].

	d = 12 ifTrue: [  "for indexing a color map with 4 bits per color component"
		val := (((rgb bitShift: -18) bitAnd: 16r0F00) bitOr:
			 ((rgb bitShift: -12) bitAnd: 16r00F0)) bitOr:
			 ((rgb bitShift: -6) bitAnd: 16r000F).
		^ val = 0 ifTrue: [rgbBlack] ifFalse: [val]].

	d = 9 ifTrue: [  "for indexing a color map with 3 bits per color component"
		val := (((rgb bitShift: -21) bitAnd: 16r01C0) bitOr:
			 ((rgb bitShift: -14) bitAnd: 16r0038)) bitOr:
			 ((rgb bitShift: -7) bitAnd: 16r0007).
		^ val = 0 ifTrue: [rgbBlack] ifFalse: [val]].

	self error: 'unknown pixel depth: ', d printString
! !

!Color methodsFor: 'conversions' stamp: 'di 11/30/1998 09:03'!
pixelWordFor: depth filledWith: pixelValue
	"Return to a 32-bit word that concatenates enough copies of the given pixel value to fill the word (i.e., 32/depth copies). Depth should be one of 1, 2, 4, 8, 16, or 32. The pixel value should be an integer in 0..2^depth-1."
	| halfword |
	depth = 32 ifTrue: [^ pixelValue].
	depth = 16
		ifTrue: [halfword := pixelValue]
		ifFalse: [halfword := pixelValue * 
					(#(16rFFFF				"replicates at every bit"
						16r5555 -			"replicates every 2 bits"
						16r1111 - - -			"replicates every 4 bits"
						16r0101) at: depth)	"replicates every 8 bits"].
	^ halfword bitOr: (halfword bitShift: 16)! !

!Color methodsFor: 'conversions' stamp: ''!
pixelWordForDepth: depth
	"Return to a 32-bit word that concatenates enough copies of the receiver's pixel value to fill the word (i.e., 32/depth copies). Depth should be one of 1, 2, 4, 8, 16, or 32. The pixel value should be an integer in 0..2^depth-1."

	| pixelValue |
	pixelValue := self pixelValueForDepth: depth.
	^ self pixelWordFor: depth filledWith: pixelValue
! !

!Color methodsFor: 'html' stamp: 'hmm 11/1/2006 23:28'!
printHtmlString
	"answer a string whose characters are the html representation  
	of the receiver"
	^ ((self red * 255) asInteger printStringBase: 16 length: 2 padded: true)
	, ((self green * 255) asInteger printStringBase: 16 length: 2 padded: true)
	, ((self blue * 255) asInteger printStringBase: 16 length: 2 padded: true)! !

!Color methodsFor: 'printing' stamp: 'bf 5/25/2000 16:52'!
printOn: aStream
	| name |
	(name := self name) ifNotNil:
		[^ aStream
			nextPutAll: 'Color ';
			nextPutAll: name].
	self storeOn: aStream.
! !

!Color methodsFor: 'private' stamp: ''!
privateAlpha
	"Private!! Return the raw alpha value for opaque. Used only for equality testing."

	^ 255! !

!Color methodsFor: 'private' stamp: ''!
privateBlue
	"Private!! Return the internal representation of my blue component."

	^ rgb bitAnd: ComponentMask! !

!Color methodsFor: 'private' stamp: ''!
privateGreen
	"Private!! Return the internal representation of my green component.
	Replaced >> by bitShift: 0 -. SqR!! 2/25/1999 23:08"

	^ (rgb bitShift: 0 - GreenShift) bitAnd: ComponentMask! !

!Color methodsFor: 'private' stamp: ''!
privateRGB
	"Private!! Return the internal representation of my RGB components."

	^ rgb
! !

!Color methodsFor: 'private' stamp: ''!
privateRed
	"Private!! Return the internal representation of my red component."

	^ (rgb bitShift: 0 - RedShift) bitAnd: ComponentMask! !

!Color methodsFor: 'other' stamp: 'ar 8/16/2001 12:47'!
raisedColor
	^ self! !

!Color methodsFor: 'access' stamp: ''!
red
	"Return the red component of this color, a float in the range [0.0..1.0]."

	^ self privateRed asFloat / ComponentMax! !

!Color methodsFor: 'other' stamp: 'jm 12/4/97 10:27'!
rgbTriplet
	"Color fromUser rgbTriplet"

	^ Array
		with: (self red roundTo: 0.01)
		with: (self green roundTo: 0.01)
		with: (self blue roundTo: 0.01)
! !

!Color methodsFor: 'access' stamp: ''!
saturation
	"Return the saturation of this color, a value between 0.0 and 1.0."

	| r g b max min |
	r := self privateRed.
	g := self privateGreen.
	b := self privateBlue. 

	max := min := r.
	g > max ifTrue: [max := g].
	b > max ifTrue: [max := b].
	g < min ifTrue: [min := g].
	b < min ifTrue: [min := b].

	max = 0
		ifTrue: [ ^ 0.0 ]
		ifFalse: [ ^ (max - min) asFloat / max asFloat ].
! !

!Color methodsFor: 'conversions' stamp: 'ar 1/14/1999 15:28'!
scaledPixelValue32
	"Return the alpha scaled pixel value for depth 32"
	^self pixelWordForDepth: 32! !

!Color methodsFor: 'private' stamp: ''!
setHue: hue saturation: saturation brightness: brightness
	"Initialize this color to the given hue, saturation, and brightness. See the comment in the instance creation method for details."

	| s v hf i f p q t | 
	s := (saturation asFloat max: 0.0) min: 1.0.
	v := (brightness asFloat max: 0.0) min: 1.0.

	"zero saturation yields gray with the given brightness"
	s = 0.0 ifTrue: [ ^ self setRed: v green: v blue: v ].

	hf := hue asFloat.
	(hf < 0.0 or: [hf >= 360.0])
		ifTrue: [hf := hf - ((hf quo: 360.0) asFloat * 360.0)].
	hf := hf / 60.0.
	i := hf asInteger.  "integer part of hue"
	f := hf fractionPart.         "fractional part of hue"
	p := (1.0 - s) * v.
	q := (1.0 - (s * f)) * v.
	t := (1.0 - (s * (1.0 - f))) * v.

	0 = i ifTrue: [ ^ self setRed: v green: t blue: p ].
	1 = i ifTrue: [ ^ self setRed: q green: v blue: p ].
	2 = i ifTrue: [ ^ self setRed: p green: v blue: t ].
	3 = i ifTrue: [ ^ self setRed: p green: q blue: v ].
	4 = i ifTrue: [ ^ self setRed: t green: p blue: v ].
	5 = i ifTrue: [ ^ self setRed: v green: p blue: q ].

	self error: 'implementation error'.
! !

!Color methodsFor: 'private' stamp: 'di 11/2/97 12:19'!
setPrivateRed: r green: g blue: b
	"Initialize this color's r, g, and b components to the given values in the range [0..ComponentMax].  Encoded in a single variable as 3 integers in [0..1023]."

	rgb == nil ifFalse: [self attemptToMutateError].
	rgb := ((r min: ComponentMask max: 0) bitShift: RedShift) +
		((g min: ComponentMask max: 0) bitShift: GreenShift) +
		 (b min: ComponentMask max: 0).
	cachedDepth := nil.
	cachedBitPattern := nil.
! !

!Color methodsFor: 'private' stamp: 'ls 9/24/1999 20:04'!
setRGB: rgb0
	rgb == nil ifFalse: [self attemptToMutateError].
	rgb := rgb0! !

!Color methodsFor: 'private' stamp: ''!
setRed: r green: g blue: b
	"Initialize this color's r, g, and b components to the given values in the range [0.0..1.0].  Encoded in a single variable as 3 integers in [0..1023]."

	rgb == nil ifFalse: [self attemptToMutateError].
	rgb :=
		(((r * ComponentMax) rounded bitAnd: ComponentMask) bitShift: RedShift) +
		(((g * ComponentMax) rounded bitAnd: ComponentMask) bitShift: GreenShift) +
		 ((b * ComponentMax) rounded bitAnd: ComponentMask).
	cachedDepth := nil.
	cachedBitPattern := nil.
! !

!Color methodsFor: 'private' stamp: ''!
setRed: r green: g blue: b range: range
	"Initialize this color's r, g, and b components to the given values in the range [0..r]."

	rgb == nil ifFalse: [self attemptToMutateError].
	rgb :=
		((((r * ComponentMask) // range) bitAnd: ComponentMask) bitShift: RedShift) +
		((((g * ComponentMask) // range) bitAnd: ComponentMask) bitShift: GreenShift) +
		 (((b * ComponentMask) // range) bitAnd: ComponentMask).
	cachedDepth := nil.
	cachedBitPattern := nil.
! !

!Color methodsFor: 'printing' stamp: ''!
shortPrintString
	"Return a short (but less precise) print string for use where space is tight."

	| s |
	s := WriteStream on: ''.
	s
		nextPutAll: '(' , self class name;
		nextPutAll: ' r: ';
		nextPutAll: (self red roundTo: 0.01) printString;
		nextPutAll: ' g: ';
		nextPutAll: (self green roundTo: 0.01) printString;
		nextPutAll: ' b: ';
		nextPutAll: (self blue roundTo: 0.01) printString;
		nextPutAll: ')'.
	^ s contents
! !

!Color methodsFor: 'transformations' stamp: 'dew 3/4/2002 01:43'!
slightlyDarker

	^ self adjustBrightness: -0.03
! !

!Color methodsFor: 'transformations' stamp: 'dew 3/4/2002 01:43'!
slightlyLighter

	^ self adjustSaturation: -0.01 brightness: 0.03! !

!Color methodsFor: 'transformations' stamp: 'dew 1/19/2002 01:25'!
slightlyWhiter

	^ self alphaMixed: 0.85 with: Color white
! !

!Color methodsFor: 'printing' stamp: 'mir 7/21/1999 11:41'!
storeArrayOn: aStream

	aStream nextPutAll: '#('.
	self storeArrayValuesOn: aStream.
	aStream nextPutAll: ') '
! !

!Color methodsFor: 'printing' stamp: 'mir 7/21/1999 11:41'!
storeArrayValuesOn: aStream

	(self red roundTo: 0.001) storeOn: aStream.
	aStream space.
	(self green roundTo: 0.001) storeOn: aStream.
	aStream space.
	(self blue roundTo: 0.001) storeOn: aStream.

! !

!Color methodsFor: 'printing' stamp: 'di 9/27/2000 13:34'!
storeOn: aStream

	aStream
		nextPutAll: '(' , self class name;
		nextPutAll: ' r: '; print: (self red roundTo: 0.001);
		nextPutAll: ' g: '; print: (self green roundTo: 0.001);
		nextPutAll: ' b: '; print: (self blue roundTo: 0.001);
		nextPutAll: ')'.
! !

!Color methodsFor: 'transformations' stamp: 'dew 3/4/2002 01:44'!
twiceDarker
	"Answer a significantly darker shade of this color."

	^ self adjustBrightness: -0.15! !

!Color methodsFor: 'transformations' stamp: 'dew 3/4/2002 01:45'!
twiceLighter
	"Answer a significantly lighter shade of this color."

	^ self adjustSaturation: -0.06 brightness: 0.15! !

!Color methodsFor: 'copying' stamp: 'tk 8/19/1998 16:12'!
veryDeepCopyWith: deepCopier
	"Return self.  I am immutable in the Morphic world.  Do not record me."! !

!Color methodsFor: 'transformations' stamp: 'tk 7/4/2000 12:07'!
veryMuchLighter

	^ self alphaMixed: 0.1165 with: Color white
! !

!Color methodsFor: 'groups of shades' stamp: 'di 10/23/2000 09:45'!
wheel: thisMany
	"An array of thisMany colors around the color wheel starting at self and ending all the way around the hue space just before self.  Array is of length thisMany.  Very useful for displaying color based on a variable in your program.  "

	| sat bri hue step c |
	sat := self saturation.
	bri := self brightness.
	hue := self hue.
	step := 360.0 / (thisMany max: 1).
	^ (1 to: thisMany) collect: [:num |
		c := Color h: hue s: sat v: bri.  "hue is taken mod 360"
		hue := hue + step.
		c].
"
(Color wheel: 8) withIndexDo: [:c :i | Display fill: (i*10@20 extent: 10@20) fillColor: c]
"! !

!Color methodsFor: 'transformations' stamp: 'dew 3/23/2002 01:38'!
whiter

	^ self alphaMixed: 0.8333 with: Color white
! !

Color subclass: #TranslucentColor
	instanceVariableNames: 'alpha'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Primitives'!

!TranslucentColor commentStamp: '<historical>' prior: 0!
A TranslucentColor behaves just like a normal color, except that it will pack its alpha value into the high byte of a 32-bit pixelValue.  This allows creating forms with translucency for use with the alpha blend function of BitBlt.  An alpha of zero is transparent, and 1.0 is opaque.!

!TranslucentColor methodsFor: 'accessing' stamp: ''!
alpha
	"Return my alpha value, a number between 0.0 and 1.0 where 0.0 is completely transparent and 1.0 is completely opaque."

	^ alpha asFloat / 255.0
! !

!TranslucentColor methodsFor: 'conversions' stamp: 'di 1/15/1999 11:44'!
alpha: alphaValue
	alphaValue = 1.0 ifTrue:
		[^ Color basicNew
			setPrivateRed: self privateRed
			green: self privateGreen
			blue: self privateBlue].
	^ super alpha: alphaValue! !

!TranslucentColor methodsFor: 'conversions' stamp: 'sw 10/27/1999 10:51'!
asNontranslucentColor
	^ self alpha: 1.0! !

!TranslucentColor methodsFor: 'conversions' stamp: 'di 3/25/2000 17:56'!
balancedPatternForDepth: depth
	"Return an appropriate bit pattern or stipple.  This will almost never be meaningful for tranlucentColors, except for the degenerate case of tranparency."

	alpha = 0 ifTrue: [^ Bitmap with: 0].
	^ super balancedPatternForDepth: depth! !

!TranslucentColor methodsFor: 'conversions' stamp: 'di 1/14/1999 20:05'!
bitPatternForDepth: depth
	"Return an appropriate bit pattern or stipple.  This will almost never be meaningful for tranlucentColors, except for the degenerate case of tranparency."

	alpha = 0 ifTrue: [^ Bitmap with: 0].
	^ super bitPatternForDepth: depth! !

!TranslucentColor methodsFor: 'object fileIn' stamp: 'RAA 12/20/2000 17:50'!
convertToCurrentVersion: varDict refStream: smartRefStrm
	
	"1/13/1999 -- old versions did not have alpha??"
	varDict at: 'alpha' ifAbsent: [^ Color transparent].

	^super convertToCurrentVersion: varDict refStream: smartRefStrm.
	! !

!TranslucentColor methodsFor: 'equality' stamp: ''!
hash

	^ rgb bitXor: alpha
! !

!TranslucentColor methodsFor: 'queries' stamp: 'ar 4/20/2001 04:33'!
isOpaque
	^alpha = 255! !

!TranslucentColor methodsFor: 'queries' stamp: 'di 12/30/1998 14:33'!
isTranslucent
	^ alpha < 255! !

!TranslucentColor methodsFor: 'queries' stamp: 'di 1/3/1999 12:22'!
isTranslucentColor
	"This means: self isTranslucent, but isTransparent not"
	^ alpha > 0! !

!TranslucentColor methodsFor: 'queries' stamp: 'di 12/30/1998 14:33'!
isTransparent
	^ alpha = 0! !

!TranslucentColor methodsFor: 'conversions' stamp: 'ar 5/27/2001 16:30'!
pixelValueForDepth: d
	"Return the pixel value for this color at the given depth. Translucency only works in RGB; this color will appear either opaque or transparent at all other depths."
	| basicPixelWord |
	alpha = 0 ifTrue: [^ 0].
	basicPixelWord := super pixelValueForDepth: d.
	d < 32
		ifTrue: [^ basicPixelWord]
		ifFalse: [^ (basicPixelWord bitAnd: 16rFFFFFF) bitOr: (alpha bitShift: 24)].
! !

!TranslucentColor methodsFor: 'conversions' stamp: 'di 1/6/1999 16:14'!
pixelWordForDepth: depth
	"Return the pixel value for this color at the given depth. Translucency only works in RGB; this color will appear either opaque or transparent at all other depths."

	| basicPixelWord |
	alpha = 0 ifTrue: [^ 0].
	basicPixelWord := super pixelWordForDepth: depth.
	depth < 32
		ifTrue: [^ basicPixelWord]
		ifFalse: [^ (basicPixelWord bitAnd: 16rFFFFFF) bitOr: (alpha bitShift: 24)].
! !

!TranslucentColor methodsFor: 'private' stamp: ''!
privateAlpha
	"Return my raw alpha value, an integer in the range 0..255. Used for fast equality testing."

	^ alpha
! !

!TranslucentColor methodsFor: 'conversions' stamp: 'ar 1/14/1999 15:30'!
scaledPixelValue32
	"Return the alpha scaled pixel value for depth 32"
	| pv32 a b g r |
	pv32 := super scaledPixelValue32.
	a := (self alpha * 255.0) rounded.
	b := (pv32 bitAnd: 255) * a // 256.
	g := ((pv32 bitShift: -8) bitAnd: 255) * a // 256.
	r := ((pv32 bitShift: -16) bitAnd: 255) * a // 256.
	^b + (g bitShift: 8) + (r bitShift: 16) + (a bitShift: 24)! !

!TranslucentColor methodsFor: 'private' stamp: ''!
setRgb: rgbValue alpha: alphaValue
	"Set the state of this translucent color. Alpha is represented internally by an integer in the range 0..255."

	rgb == nil ifFalse: [self attemptToMutateError].
	rgb := rgbValue.
	alpha := (255.0 * alphaValue) asInteger min: 255 max: 0.
! !

!TranslucentColor methodsFor: 'printing' stamp: 'mir 7/21/1999 11:43'!
storeArrayValuesOn: aStream

	self isTransparent ifTrue: [
		^ aStream space].
	super storeArrayValuesOn: aStream.

	aStream space.
	(self alpha roundTo: 0.001) storeOn: aStream.

! !

!TranslucentColor methodsFor: 'printing' stamp: 'di 9/27/2000 13:33'!
storeOn: aStream

	self isTransparent ifTrue: [^ aStream nextPutAll: '(Color transparent)'].
	super storeOn: aStream.
	aStream
		skip: -1;	  "get rid of trailing )"
		nextPutAll: ' alpha: ';
		print: (self alpha roundTo: 0.001);
		nextPutAll: ')'.
! !

Object subclass: #ColorMap
	instanceVariableNames: 'shifts masks colors'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Primitives'!

!ColorMap class methodsFor: 'instance creation' stamp: 'ar 2/22/2000 14:08'!
colors: colorArray
	^self new setShifts: nil masks: nil colors: colorArray! !

!ColorMap class methodsFor: 'instance creation' stamp: 'ar 5/27/2000 20:09'!
mapBitsFrom: srcBitMask to: dstBitMask
	"Return an array consisting of the shift and the mask for
	mapping component values out of srcBitMask and into dstBitMask.
	While this computation is somewhat complicated it eases the batch
	conversion of all the pixels in BitBlt."
	| srcBits dstBits srcLow srcHigh dstLow dstHigh bits mask shift |
	(srcBitMask = 0 or:[dstBitMask = 0]) ifTrue:[^#(0 0)]. "Zero mask and shift"
	"Compute low and high bit position for source and dest bit mask"
	srcLow := srcBitMask lowBit - 1.	srcHigh := srcBitMask highBit.
	dstLow := dstBitMask lowBit - 1.	dstHigh := dstBitMask highBit.
	"Compute the number of bits in source and dest bit mask"
	srcBits := srcHigh - srcLow.		dstBits := dstHigh - dstLow.
	"Compute the maximum number of bits we can transfer inbetween"
	bits := srcBits min: dstBits.
	"Compute the (unshifted) transfer mask"
	mask := (1 bitShift: bits) - 1.
	"Shift the transfer mask to the mask the highest n bits of srcBitMask"
	mask := mask bitShift: (srcHigh - bits).
	"Compute the delta shift so that the most significant bit of the
	source bit mask falls on the most significant bit of the dest bit mask.
	Note that delta is used for #bitShift: so
		shift > 0 : shift right
		shift < 0 : shift left
	e.g., if dstHigh > srcHigh we need to shift left and if dstHigh < srcHigh
	we need to shift right. This leads to:"
	shift := dstHigh - srcHigh.
	"And that's all we need"
	^Array with: shift with: mask! !

!ColorMap class methodsFor: 'instance creation' stamp: 'ar 5/27/2000 19:41'!
mappingFrom: srcBitMasks to: dstBitMasks
	"Return a color map mapping from the array of source bit masks
	to the array of dest bit masks."
	| shifts masks shiftAndMask |
	shifts := IntegerArray new: 4.
	masks := WordArray new: 4.
	1 to: 4 do:[:i|
		shiftAndMask := self mapBitsFrom: (srcBitMasks at: i) to: (dstBitMasks at: i).
		shifts at: i put: (shiftAndMask at: 1).
		masks at: i put: (shiftAndMask at: 2).
	].
	^self shifts: shifts masks: masks! !

!ColorMap class methodsFor: 'instance creation' stamp: 'ar 5/27/2000 20:08'!
mappingFromARGB: dstBitMasks
	"Return a ColorMap mapping from canonical ARGB space into dstBitMasks"
	^self mappingFrom: #(16rFF0000 16rFF00 16rFF 16rFF000000) to: dstBitMasks! !

!ColorMap class methodsFor: 'instance creation' stamp: 'ar 5/27/2000 20:08'!
mappingToARGB: srcBitMasks
	"Return a ColorMap mapping from srcBitMasks into canonical ARGB space"
	^self mappingFrom: srcBitMasks to: #(16rFF0000 16rFF00 16rFF 16rFF000000)! !

!ColorMap class methodsFor: 'instance creation' stamp: 'ar 5/4/2001 15:59'!
masks: maskArray shifts: shiftArray
	^self shifts: shiftArray masks: maskArray colors: nil.! !

!ColorMap class methodsFor: 'instance creation' stamp: 'ar 1/16/2000 16:02'!
shifts: shiftArray masks: maskArray
	^self shifts: shiftArray masks: maskArray colors: nil.! !

!ColorMap class methodsFor: 'instance creation' stamp: 'ar 1/16/2000 16:02'!
shifts: shiftArray masks: maskArray colors: colorArray
	^self new setShifts: shiftArray masks: maskArray colors: colorArray! !

!ColorMap methodsFor: 'comparing' stamp: 'tk 7/5/2001 21:59'!
= aColorMap
	"Return true if the receiver is equal to aColorMap"
	self species == aColorMap species ifFalse:[^false].
	self isIndexed == aColorMap isIndexed ifFalse:[^false].
	^self colors = aColorMap colors and:[
		self shifts = aColorMap shifts and:[
			self masks = aColorMap masks]]! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:54'!
alphaMask
	^masks at: 4! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:55'!
alphaMask: value
	masks at: 4 put: value! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:31'!
alphaShift
	^shifts at: 4! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:31'!
alphaShift: value
	shifts at: 4 put: value! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:39'!
at: index
	^colors at: index! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:39'!
at: index put: value
	^colors at: index put: value! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:31'!
blueMask
	^masks at: 3! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:31'!
blueMask: value
	masks at: 3 put: value! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:31'!
blueShift
	^shifts at: 3! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:32'!
blueShift: value
	shifts at: 3 put: value! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 2/10/2000 17:12'!
colors
	^colors! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:32'!
greenMask
	^masks at: 2! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:32'!
greenMask: value
	masks at: 2 put: value! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:32'!
greenShift
	^shifts at: 2! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:36'!
greenShift: value
	shifts at: 2 put: value.! !

!ColorMap methodsFor: 'comparing' stamp: 'ar 5/27/2000 19:29'!
hash
	"Hash is re-implemented because #= is re-implemented"
	^colors hash bitXor: (shifts hash bitXor: masks hash)! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 5/28/2000 22:08'!
inverseMap
	"Return the inverse map of the receiver"
	| newMasks newShifts |
	colors ifNotNil:[^self error:'Not yet implemented'].
	newMasks := WriteStream on: (Array new: 4).
	newShifts := WriteStream on: (Array new: 4).
	masks with: shifts do:[:mask :shift|
		newMasks nextPut: (mask bitShift: shift).
		newShifts nextPut: shift negated].
	^ColorMap
		shifts: newShifts contents
		masks: newMasks contents! !

!ColorMap methodsFor: 'testing' stamp: 'ar 5/25/2000 19:41'!
isColormap
	^true! !

!ColorMap methodsFor: 'testing' stamp: 'ar 5/27/2000 19:06'!
isFixed
	"Return true if the receiver does not use a lookup mechanism for pixel mapping"
	^self isIndexed not! !

!ColorMap methodsFor: 'testing' stamp: 'ar 5/27/2000 19:06'!
isIndexed
	"Return true if the receiver uses a lookup mechanism for pixel mapping"
	^colors notNil! !

!ColorMap methodsFor: 'pixel mapping' stamp: 'ar 5/15/2001 16:12'!
mapPixel: pixelValue
	"Perform a forward pixel mapping operation"
	| pv |
	(shifts == nil and:[masks == nil]) ifFalse:[
		pv := (((pixelValue bitAnd: self redMask) bitShift: self redShift) bitOr:
			((pixelValue bitAnd: self greenMask) bitShift: self greenShift)) bitOr:
			(((pixelValue bitAnd: self blueMask) bitShift: self blueShift) bitOr:
			((pixelValue bitAnd: self alphaMask) bitShift: self alphaShift)).
	] ifTrue:[pv := pixelValue].
	colors ifNotNil:[pv := colors at: pv].
	"Need to check for translucency else Form>>paint goes gaga"
	pv = 0 ifTrue:[pixelValue = 0 ifFalse:[pv := 1]].
	^pv! !

!ColorMap methodsFor: 'pixel mapping' stamp: 'ar 6/8/2000 20:36'!
mappingTo: aColorMap
	"Compute a new color map through the receiver and aColorMap.
	Both maps are assumed to be mappings into canonical ARGB space"
	| fixedMap |
	self = aColorMap ifTrue:[^nil]. "No mapping needed"
	aColorMap isIndexed ifTrue:[^nil]. "We can't compute mappings to an indexed map yet"
	fixedMap := self class mappingFrom: self rgbaBitMasks to: aColorMap rgbaBitMasks.
	self isIndexed ifFalse:[^fixedMap].
	"If the receiver is indexed then we need to map the colors as well"
	self flag: #untested.
	^ColorMap
		shifts: fixedMap shifts
		masks: fixedMap masks
		colors: (colors collect:[:pv| aColorMap pixelMap: pv]).
! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 5/27/2000 19:16'!
masks
	^masks! !

!ColorMap methodsFor: 'pixel mapping' stamp: 'ar 5/15/2001 16:12'!
pixelMap: pixelValue
	"Perform a reverse pixel mapping operation"
	| pv |
	colors == nil
		ifTrue:[pv := pixelValue]
		ifFalse:[pv := colors at: pixelValue].
	(shifts == nil and:[masks == nil]) 
		ifFalse:[pv := (((pv bitAnd: self redMask) bitShift: self redShift) bitOr: 
				((pv bitAnd: self greenMask) bitShift: self greenShift)) bitOr:
					(((pv bitAnd: self blueMask) bitShift: self blueShift) bitOr: 
						((pv bitAnd: self alphaMask) bitShift: self alphaShift))].
	"Need to check for translucency else Form>>paint goes gaga"
	pv = 0 ifTrue:[pixelValue = 0 ifFalse:[pv := 1]].
	^pv! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:36'!
redMask
	^masks at: 1! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:37'!
redMask: value
	masks at: 1 put: value! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:37'!
redShift
	^shifts at: 1! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 1/16/2000 15:37'!
redShift: value
	shifts at: 1 put: value! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 5/27/2000 20:48'!
rgbaBitMasks
	"Return the rgba bit masks for the receiver"
	^masks asArray with: shifts collect:[:m :s| m bitShift: s]! !

!ColorMap methodsFor: 'private' stamp: 'ar 2/22/2000 16:47'!
setShifts: shiftArray masks: maskArray colors: colorArray
	shiftArray ifNotNil:[shifts := shiftArray asIntegerArray].
	maskArray ifNotNil:[masks := maskArray asWordArray].
	colorArray ifNotNil:[colors := colorArray asWordArray].! !

!ColorMap methodsFor: 'accessing' stamp: 'ar 5/27/2000 19:16'!
shifts
	^shifts! !

Object subclass: #CornerRounder
	instanceVariableNames: 'cornerMasks cornerOverlays underBits'
	classVariableNames: 'CR0 CR1 CR2'
	poolDictionaries: ''
	category: 'Graphics-Display Objects'!

!CornerRounder commentStamp: '<historical>' prior: 0!
This class is a quick hack to support rounded corners in morphic.

Rather than produce rounded rectangles, it tweaks the display of corners.
Rather than work for any radius, it only supports a radius of 6.
Rather than work for any border width, it only supports widths 0, 1 and 2.
The corners, while apparently transparent, still behave opaquely to mouse clicks.

Worse than this, the approach relies on the ability to extract underlying bits from the canvas prior to display.  This ran afoul of top-down display, it seems, in SystemWindow spawnReframeHandle: (qv).  It will also make a postscript printer very unhappy.

But, hey, it's cute.!

!CornerRounder class methodsFor: 'all' stamp: 'di 6/28/1999 15:51'!
initialize  "CornerRounder initialize"

	CR0 := CR1 := self new
		masterMask:
			(Form extent: 6@6
				fromArray: #(2r1e26 2r111e26 2r1111e26 2r11111e26 2r11111e26 2r111111e26)
				offset: 0@0)
		masterOverlay:
			(Form extent: 6@6
				fromArray: #(2r1e26 2r110e26 2r1000e26 2r10000e26 2r10000e26 2r100000e26)
				offset: 0@0).
	CR2 := self new
		masterMask:
			(Form extent: 6@6
				fromArray: #(2r1e26 2r111e26 2r1111e26 2r11111e26 2r11111e26 2r111111e26)
				offset: 0@0)
		masterOverlay:
			(Form extent: 6@6
				fromArray: #(2r1e26 2r111e26 2r1111e26 2r11100e26 2r11000e26 2r111000e26)
				offset: 0@0).

! !

!CornerRounder class methodsFor: 'all' stamp: 'di 3/25/2000 11:12'!
rectWithinCornersOf: aRectangle
	"Return a single sub-rectangle that lies entirely inside corners
	that are made by me.
	Used to identify large regions of window that do not need to be redrawn."

	^ aRectangle insetBy: 0@6! !

!CornerRounder class methodsFor: 'all' stamp: 'ar 1/5/2002 17:24'!
roundCornersOf: aMorph on: aCanvas in: bounds displayBlock: displayBlock borderWidth: w corners: aList

	| rounder |
	rounder := CR0.
	w = 1 ifTrue: [rounder := CR1].
	w = 2 ifTrue: [rounder := CR2].
	rounder := rounder copy.
	rounder saveBitsUnderCornersOf: aMorph on: aCanvas in: bounds corners: aList.
	displayBlock value.
	rounder tweakCornersOf: aMorph on: aCanvas in: bounds borderWidth: w corners: aList! !

!CornerRounder methodsFor: 'all' stamp: 'di 6/24/1999 09:35'!
masterMask: maskForm masterOverlay: overlayForm

	cornerMasks := #(none left pi right) collect:
		[:dir | (maskForm rotateBy: dir centerAt: 0@0) offset: 0@0].
	cornerOverlays := #(none left pi right) collect:
		[:dir | (overlayForm rotateBy: dir centerAt: 0@0) offset: 0@0].
! !

!CornerRounder methodsFor: 'all' stamp: 'ar 1/5/2002 17:26'!
saveBitsUnderCornersOf: aMorph on: aCanvas in: bounds corners: cornerList

	| offset corner mask form corners rect |
	underBits := Array new: 4.
	corners := bounds corners.
	cornerList do:[:i|
		mask := cornerMasks at: i.
		corner := corners at: i.
		i = 1 ifTrue: [offset := 0@0].
		i = 2 ifTrue: [offset := 0@mask height negated].
		i = 3 ifTrue: [offset := mask extent negated].
		i = 4 ifTrue: [offset := mask width negated@0].
		rect := corner + offset extent: mask extent.
		(aCanvas isVisible: rect) ifTrue:[
			form := aCanvas contentsOfArea: rect.
			form copyBits: form boundingBox from: mask at: 0@0 clippingBox: form boundingBox rule: Form and fillColor: nil map: (Bitmap with: 16rFFFFFFFF with: 0).
			underBits at: i put: form]].
! !

!CornerRounder methodsFor: 'all' stamp: 'kfr 8/4/2003 23:28'!
tweakCornersOf: aMorph on: aCanvas in: bounds borderWidth: w corners: cornerList
	"This variant has a cornerList argument, to allow some corners to be rounded and others not"
	| offset corner saveBits fourColors mask outBits shadowColor corners |
	shadowColor := aCanvas shadowColor.
	aCanvas shadowColor: nil. "for tweaking it's essential"
	w > 0 ifTrue:[
			fourColors := shadowColor 
				ifNil:[aMorph borderStyle colorsAtCorners]
				ifNotNil:[Array new: 4 withAll: Color transparent]].
	mask := Form extent: cornerMasks first extent depth: aCanvas depth.
	corners := bounds corners.
	cornerList do:[:i|
		corner := corners at: i.
		saveBits := underBits at: i.
		saveBits ifNotNil:[
			i = 1 ifTrue: [offset := 0@0].
			i = 2 ifTrue: [offset := 0@saveBits height negated].
			i = 3 ifTrue: [offset := saveBits extent negated].
			i = 4 ifTrue: [offset := saveBits width negated@0].

			"Mask out corner area (painting saveBits won't clear if transparent)."
			mask copyBits: mask boundingBox from: (cornerMasks at: i) at: 0@0 clippingBox: mask boundingBox rule: Form over fillColor: nil map: (Bitmap with: 0 with: 16rFFFFFFFF).
			outBits := aCanvas contentsOfArea: (corner + offset extent: mask extent).
			mask displayOn: outBits at: 0@0 rule: Form and.
			"Paint back corner bits."
			saveBits displayOn: outBits at: 0@0 rule: Form paint.
			"Paint back corner bits."
			aCanvas drawImage: outBits at: corner + offset.

			w > 0 ifTrue:[
				
				aCanvas stencil: (cornerOverlays at: i) at: corner + offset
						color: (fourColors at: i)]]].
	aCanvas shadowColor: shadowColor. "restore shadow color"
! !

Object subclass: #DisplayObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Display Objects'!

!DisplayObject commentStamp: '<historical>' prior: 0!
The abstract protocol for most display primitives that are used by Views for presenting information on the screen.!

DisplayObject subclass: #DisplayMedium
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Display Objects'!

!DisplayMedium commentStamp: '<historical>' prior: 0!
I am a display object which can both paint myself on a medium (displayOn: messages), and can act as a medium myself. My chief subclass is Form.!

!DisplayMedium methodsFor: 'bordering' stamp: ''!
border: aRectangle width: borderWidth 
	"Paint a border whose rectangular area is defined by aRectangle. The 
	width of the border of each side is borderWidth. Uses black for 
	drawing the border."

	self border: aRectangle width: borderWidth fillColor: Color black.
! !

!DisplayMedium methodsFor: 'bordering' stamp: ''!
border: aRectangle width: borderWidth fillColor: aHalfTone 
	"Paint a border whose rectangular area is defined by aRectangle. The 
	width of the border of each side is borderWidth. Uses aHalfTone for 
	drawing the border."

	self border: aRectangle
		widthRectangle: 
			(Rectangle
				left: borderWidth
				right: borderWidth
				top: borderWidth
				bottom: borderWidth)
		rule: Form over
		fillColor: aHalfTone! !

!DisplayMedium methodsFor: 'bordering' stamp: ''!
border: aRectangle width: borderWidth rule: combinationRule fillColor: aHalfTone 
	"Paint a border whose rectangular area is defined by aRectangle. The 
	width of the border of each side is borderWidth. Uses aHalfTone for 
	drawing the border."

	self border: aRectangle
		widthRectangle: 
			(Rectangle
				left: borderWidth
				right: borderWidth
				top: borderWidth
				bottom: borderWidth)
		rule: combinationRule
		fillColor: aHalfTone! !

!DisplayMedium methodsFor: 'bordering' stamp: ''!
border: aRectangle widthRectangle: insets rule: combinationRule fillColor: aHalfTone
	"Paint a border whose rectangular area is defined by aRectangle. The 
	width of each edge of the border is determined by the four coordinates 
	of insets. Uses aHalfTone and combinationRule for drawing the border."

	(aRectangle areasOutside: (aRectangle insetBy: insets)) do:
		[:edgeStrip | self fill: edgeStrip rule: combinationRule fillColor: aHalfTone]! !

!DisplayMedium methodsFor: 'displaying' stamp: ''!
copyBits: sourceRect from: sourceForm at: destOrigin clippingBox: clipRect rule: rule fillColor: aForm 
	"Make up a BitBlt table and copy the bits."

	self subclassResponsibility! !

!DisplayMedium methodsFor: 'displaying' stamp: 'hmm 9/16/2000 21:27'!
deferUpdatesIn: aRectangle while: aBlock
	"DisplayScreen overrides with something more involved..."
	^aBlock value! !

!DisplayMedium methodsFor: 'displaying' stamp: ''!
drawLine: sourceForm from: beginPoint to: endPoint clippingBox: clipRect rule: anInteger fillColor: aForm 
	"Draw line by copying the argument, sourceForm, starting at location 
	beginPoint and ending at endPoint, clipped by the rectangle, clipRect. 
	The rule and mask for copying are the arguments anInteger and aForm."

	self subclassResponsibility! !

!DisplayMedium methodsFor: 'coloring' stamp: ''!
fill: aRectangle fillColor: aForm 
	"Replace a rectangular area of the receiver with the pattern described by 
	aForm according to the rule over."

	self fill: aRectangle rule: Form over fillColor: aForm! !

!DisplayMedium methodsFor: 'coloring' stamp: ''!
fill: aRectangle rule: anInteger fillColor: aForm 
	"Replace a rectangular area of the receiver with the pattern described by 
	aForm according to the rule anInteger."

	self subclassResponsibility! !

!DisplayMedium methodsFor: 'coloring' stamp: ''!
fillBlack
	"Set all bits in the receiver to black (ones)."

	self fill: self boundingBox fillColor: Color black! !

!DisplayMedium methodsFor: 'coloring' stamp: ''!
fillBlack: aRectangle 
	"Set all bits in the receiver's area defined by aRectangle to black (ones)."

	self fill: aRectangle rule: Form over fillColor: Color black! !

!DisplayMedium methodsFor: 'coloring' stamp: ''!
fillColor: aColor
	"Set all pixels in the receiver to the color.  Must be a correct color for this depth of medium.  TK 1 Jun 96"

	self fill: self boundingBox fillColor: aColor! !

!DisplayMedium methodsFor: 'coloring' stamp: ''!
fillGray
	"Set all bits in the receiver to gray."

	self fill: self boundingBox fillColor: Color gray! !

!DisplayMedium methodsFor: 'coloring' stamp: ''!
fillGray: aRectangle
	"Set all bits in the receiver's area defined by aRectangle to the gray mask."

	self fill: aRectangle rule: Form over fillColor: Color gray! !

!DisplayMedium methodsFor: 'coloring' stamp: ''!
fillShape: aShapeForm fillColor: aColor
	"Fill a region corresponding to 1 bits in aShapeForm with aColor"

	^ self fillShape: aShapeForm fillColor: aColor at: 0@0! !

!DisplayMedium methodsFor: 'coloring' stamp: 'ar 5/28/2000 12:06'!
fillShape: aShapeForm fillColor: aColor at: location
	"Fill a region corresponding to 1 bits in aShapeForm with aColor"

	((BitBlt current destForm: self sourceForm: aShapeForm fillColor: aColor
		combinationRule: Form paint
		destOrigin: location + aShapeForm offset sourceOrigin: 0@0
		extent: self extent clipRect: self boundingBox)
		colorMap: (Bitmap with: 0 with: 16rFFFFFFFF))
		copyBits! !

!DisplayMedium methodsFor: 'coloring' stamp: ''!
fillWhite
	"Set all bits in the form to white."

	self fill: self boundingBox fillColor: Color white.
! !

!DisplayMedium methodsFor: 'coloring' stamp: ''!
fillWhite: aRectangle
	"Set all bits in the receiver's area defined by aRectangle to white."

	self fill: aRectangle rule: Form over fillColor: Color white.
! !

!DisplayMedium methodsFor: 'coloring' stamp: ''!
fillWithColor: aColor
	"Fill the receiver's bounding box with the given color."

	self fill: self boundingBox fillColor: aColor.
! !

!DisplayMedium methodsFor: 'coloring' stamp: 'jm 6/18/1999 19:01'!
reverse
	"Change all the bits in the receiver that are white to black, and the ones 
	that are black to white."

	self fill: self boundingBox rule: Form reverse fillColor: (Color quickHighLight: self depth)! !

!DisplayMedium methodsFor: 'coloring' stamp: 'jm 6/18/1999 19:00'!
reverse: aRectangle
	"Change all the bits in the receiver's area that intersects with aRectangle 
	that are white to black, and the ones that are black to white."

	self fill: aRectangle rule: Form reverse fillColor: (Color quickHighLight: self depth)! !

!DisplayMedium methodsFor: 'coloring' stamp: ''!
reverse: aRectangle fillColor: aMask	
	"Change all the bits in the receiver's area that intersects with aRectangle 
	according to the mask. Black does not necessarily turn to white, rather it 
	changes with respect to the rule and the bit in a corresponding mask 
	location. Bound to give a surprise."

	self fill: aRectangle rule: Form reverse fillColor: aMask! !

DisplayMedium subclass: #Form
	instanceVariableNames: 'bits width height depth offset'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Display Objects'!

!Form commentStamp: 'ls 1/4/2004 17:16' prior: 0!
A rectangular array of pixels, used for holding images.  All pictures, including character images are Forms.  The depth of a Form is how many bits are used to specify the color at each pixel.  The actual bits are held in a Bitmap, whose internal structure is different at each depth.  Class Color allows you to deal with colors without knowing how they are actually encoded inside a Bitmap.
	  The supported depths (in bits) are 1, 2, 4, 8, 16, and 32.  The number of actual colors at these depths are: 2, 4, 16, 256, 32768, and 16 million.
	Forms are indexed starting at 0 instead of 1; thus, the top-left pixel of a Form has coordinates 0@0.
	Forms are combined using BitBlt.  See the comment in class BitBlt.  Forms that repeat many times to fill a large destination are InfiniteForms.

	colorAt: x@y		Returns the abstract Color at this location
	displayAt: x@y		shows this form on the screen
	displayOn: aMedium at: x@y	shows this form in a Window, a Form, or other DisplayMedium
	fillColor: aColor		Set all the pixels to the color.
	edit		launch an editor to change the bits of this form.
	pixelValueAt: x@y	The encoded color.  The encoding depends on the depth.
!

Form subclass: #ColorForm
	instanceVariableNames: 'colors cachedDepth cachedColormap'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Display Objects'!

!ColorForm commentStamp: '<historical>' prior: 0!
ColorForm is a normal Form plus a color map of up to 2^depth Colors. Typically, one reserves one entry in the color map for transparent. This allows 1, 3, 15, or 255 non-transparent colors in ColorForms of depths 1, 2, 4, and 8 bits per pixel. ColorForms don't support depths greater than 8 bits because that would require excessively large color maps with little real benefit, since 16-bit and 32-bit depths already support thousands and millions of colors.

ColorForms have several uses:
  1) Precise colors. You can have up to 256 true colors, instead being limited to the 8-bit color palette.
  2) Easy transparency. Just store (Color transparent) at the desired position in the color map.
  3) Cheap color remapping by changing the color map.

A color map is an Array of up to 2^depth Color objects. A Bitmap colorMap is automatically computed and cached for rapid display. Note that if you change the color map, you must resubmit it via the colors: method to flush this cache.

ColorForms can be a bit tricky. Note that:
  a) When you BitBlt from one ColorForm to another, you must remember to copy the color map of the source ColorForm to the destination ColorForm.
  b) A ColorForm's color map is an array of depth-independent Color objects. BitBlt requires a BitMap of actual pixel values, adjusted to the destination depth. These are different things!! ColorForms automatically maintain a cache of the BitBlt-style color map corresponding to the colors array for the last depth on which the ColorForm was displayed, so there should be little need for clients to work with BitBlt-style color maps.
  c) The default map for 8 bit depth has black in the first entry, not transparent.  Say (cform colors at: 1 put: Color transparent).
!

!ColorForm class methodsFor: 'as yet unclassified' stamp: 'nk 4/17/2004 19:44'!
mappingWhiteToTransparentFrom: aFormOrCursor
	"Return a ColorForm copied from the given Form or Cursor with white mapped to transparent."

	| f map |
	aFormOrCursor depth <= 8 ifFalse: [
		^ self error: 'argument depth must be 8-bits per pixel or less'].
	(aFormOrCursor isColorForm) ifTrue: [
		f := aFormOrCursor deepCopy.
		map := aFormOrCursor colors.
	] ifFalse: [
		f := ColorForm extent: aFormOrCursor extent depth: aFormOrCursor depth.
		f copyBits: aFormOrCursor boundingBox
			from: aFormOrCursor
			at: 0@0
			clippingBox: aFormOrCursor boundingBox
			rule: Form over
			fillColor: nil.
		map := Color indexedColors copyFrom: 1 to: (1 bitShift: aFormOrCursor depth)].
	map := map collect: [:c |
		c = Color white ifTrue: [Color transparent] ifFalse: [c]].
	f colors: map.
	^ f
! !

!ColorForm class methodsFor: 'as yet unclassified' stamp: ''!
twoToneFromDisplay: aRectangle using: oldForm backgroundColor: bgColor
	"Return a 1-bit deep ColorForm copied from the given rectangle of the display. All colors except the background color will be mapped to black."

	| f |
	((oldForm ~~ nil) and: [oldForm extent = aRectangle extent]) ifTrue: [
		f := oldForm fromDisplay: aRectangle.
	] ifFalse: [
		f := ColorForm extent: aRectangle extent depth: 1.
		f twoToneFromDisplay: aRectangle backgroundColor: bgColor.
		f colors: (Array
			with: bgColor
			with: Color black)].
	^ f
! !

!ColorForm methodsFor: 'copying' stamp: 'RAA 8/14/2000 10:45'!
asCursorForm

	^ (self asFormOfDepth: 32) offset: offset; as: StaticForm! !

!ColorForm methodsFor: 'postscript generation' stamp: ''!
asFormWithSingleTransparentColors
	| transparentIndexes |
	transparentIndexes := self transparentColorIndexes.
	transparentIndexes size <= 1 ifTrue:[^self]
		ifFalse:[^self mapTransparencies:transparentIndexes].! !

!ColorForm methodsFor: 'color manipulation' stamp: 'di 11/11/1998 13:20'!
asGrayScale
	"Return a grayscale ColorForm computed by mapping each color into its grayscale equivalent"
	^ self copy colors:
		(colors collect:
			[:c | c isTransparent ifTrue: [c]
						ifFalse: [Color gray: c luminance]])! !

!ColorForm methodsFor: 'copying' stamp: 'ar 10/24/2005 22:25'!
blankCopyOf: aRectangle scaledBy: scale
	^Form extent: (aRectangle extent * scale) truncated depth: 32! !

!ColorForm methodsFor: 'private' stamp: 'jm 11/16/97 09:07'!
clearColormapCache

	cachedDepth := nil.
	cachedColormap := nil.
! !

!ColorForm methodsFor: 'pixel accessing' stamp: 'jm 11/14/97 17:25'!
colorAt: aPoint
	"Return the color of the pixel at aPoint."

	^ self colors at: (self pixelValueAt: aPoint) + 1
! !

!ColorForm methodsFor: 'pixel accessing' stamp: 'jm 11/14/97 17:25'!
colorAt: aPoint put: aColor
	"Store the given color into the pixel at aPoint. The given color must match one of the colors in the receiver's colormap."

	| i |
	i := self colors indexOf: aColor
		ifAbsent: [^ self error: 'trying to use a color that is not in my colormap'].
	self pixelValueAt: aPoint put: i - 1.
! !

!ColorForm methodsFor: 'color mapping' stamp: 'ar 5/17/2001 15:44'!
colormapIfNeededFor: destForm
	| newMap color pv |
	(self hasNonStandardPalette or:[destForm hasNonStandardPalette]) ifFalse:[
		^self colormapIfNeededForDepth: destForm depth.
	].
	colors == nil ifTrue: [
		"use the standard colormap"
		^ super colormapIfNeededFor: destForm].

	(destForm depth = cachedDepth and:[cachedColormap isColormap]) 
		ifTrue: [^ cachedColormap].
	newMap := WordArray new: (1 bitShift: self depth).
	1 to: colors size do: [:i |
		color := colors at: i.
		pv := destForm pixelValueFor: color.
		(pv = 0 and:[color isTransparent not]) ifTrue:[pv := 1].
		newMap at: i put: pv].

	cachedDepth := destForm depth.
	^cachedColormap := ColorMap shifts: nil masks: nil colors: newMap.! !

!ColorForm methodsFor: 'color manipulation' stamp: 'ar 5/17/2001 15:44'!
colormapIfNeededForDepth: destDepth
	"Return a colormap for displaying the receiver at the given depth, or nil if no colormap is needed."

	| newMap |
	colors == nil ifTrue: [
		"use the standard colormap"
		^ Color colorMapIfNeededFrom: self depth to: destDepth].

	(destDepth = cachedDepth and:[cachedColormap isColormap not]) 
		ifTrue: [^ cachedColormap].
	newMap := Bitmap new: colors size.
	1 to: colors size do: [:i |
		newMap
			at: i
			put: ((colors at: i) pixelValueForDepth: destDepth)].

	cachedDepth := destDepth.
	^ cachedColormap := newMap.
! !

!ColorForm methodsFor: 'accessing' stamp: 'jm 11/14/97 17:39'!
colors
	"Return my color palette."

	self ensureColorArrayExists.
	^ colors
! !

!ColorForm methodsFor: 'accessing' stamp: 'ar 5/17/2001 15:45'!
colors: colorList
	"Set my color palette to the given collection."

	| colorArray colorCount newColors |
	colorList ifNil: [
		colors := cachedDepth := cachedColormap := nil.
		^ self].

	colorArray := colorList asArray.
	colorCount := colorArray size.
	newColors := Array new: (1 bitShift: self depth).
	1 to: newColors size do: [:i |
		i <= colorCount
			ifTrue: [newColors at: i put: (colorArray at: i)]
			ifFalse: [newColors at: i put: Color transparent]].

	colors := newColors.
	cachedDepth := nil.
	cachedColormap := nil.
! !

!ColorForm methodsFor: 'accessing' stamp: 'mir 7/21/1999 11:51'!
colorsFromArray: colorArray
	| colorList |
	colorList := colorArray collect: [:colorDef |
		Color fromArray: colorDef].
	self colors: colorList! !

!ColorForm methodsFor: 'color manipulation' stamp: 'jm 4/18/98 20:34'!
colorsUsed
	"Return a list of the colors actually used by this ColorForm."

	| myColor list |
	myColor := self colors.
	list := OrderedCollection new.
	self tallyPixelValues doWithIndex: [:count :i |
		count > 0 ifTrue: [list add: (myColor at: i)]].
	^ list asArray
! !

!ColorForm methodsFor: 'copying' stamp: 'ar 5/28/2000 12:06'!
copy: aRect
 	"Return a new ColorForm containing the portion of the receiver delineated by aRect."

	| newForm |
	newForm := self class extent: aRect extent depth: depth.
	((BitBlt current
		destForm: newForm
		sourceForm: self
		fillColor: nil
		combinationRule: Form over
		destOrigin: 0@0
		sourceOrigin: aRect origin
		extent: aRect extent
		clipRect: newForm boundingBox)
		colorMap: nil) copyBits.
	colors ifNotNil: [newForm colors: colors copy].
	^ newForm
! !

!ColorForm methodsFor: 'postscript generation' stamp: ''!
decodeArray
	^self depth = 1 ifTrue:['[1 0]'] ifFalse:['[0 255]'].! !

!ColorForm methodsFor: 'copying' stamp: 'jm 2/27/98 09:38'!
deepCopy

	^ self shallowCopy
		bits: bits copy;
		offset: offset copy;
		colors: colors
! !

!ColorForm methodsFor: 'private' stamp: 'jm 11/16/97 09:12'!
depth: bitsPerPixel

	bitsPerPixel > 8 ifTrue: [self error: 'ColorForms only support depths up to 8 bits'].
	super depth: bitsPerPixel.
! !

!ColorForm methodsFor: 'displaying' stamp: 'ar 5/14/2001 23:32'!
displayOn: aDisplayMedium at: aDisplayPoint clippingBox: clipRectangle rule: rule fillColor: aForm

	aDisplayMedium copyBits: self boundingBox
		from: self
		at: aDisplayPoint + self offset
		clippingBox: clipRectangle
		rule: rule
		fillColor: aForm
		map: (self colormapIfNeededFor: aDisplayMedium).
! !

!ColorForm methodsFor: 'displaying' stamp: 'di 7/17/97 10:04'!
displayOnPort: port at: location

	port copyForm: self to: location rule: Form paint! !

!ColorForm methodsFor: 'private' stamp: 'ar 5/17/2001 15:44'!
ensureColorArrayExists
	"Return my color palette."

	colors ifNil: [
		self depth > 8 ifTrue: [^ self error: 'ColorForms only support depths up to 8 bits'].
		self colors: (Color indexedColors copyFrom: 1 to: (1 bitShift: self depth))].
! !

!ColorForm methodsFor: 'color manipulation' stamp: 'jm 11/16/97 11:18'!
ensureTransparentColor
	"Ensure that the receiver (a) includes Color transparent in its color map and (b) that the entry for Color transparent is the first entry in its color map."

	| i |
self error: 'not yet implemented'.
	(colors includes: Color transparent)
		ifTrue: [
			(colors indexOf: Color transparent) = 1 ifTrue: [^ self].
			"shift the entry for color transparent"]
		ifFalse: [
			i := self unusedColormapEntry.
			i = 0 ifTrue: [self error: 'no color map entry is available'].
			colors at: i put: Color transparent.
			"shift the entry for color transparent"].
! !

!ColorForm methodsFor: 'scaling, rotation' stamp: 'ar 3/15/1999 14:28'!
flipBy: direction centerAt: aPoint
	| oldColors newForm |
	oldColors := colors.
	self colors: nil.
	newForm := super flipBy: direction centerAt: aPoint.
	self colors: oldColors.
	newForm colors: oldColors.
	^newForm ! !

!ColorForm methodsFor: 'postscript generation' stamp: ''!
getTransparencyUnificationLUT
	| lut transparentIndex |
	lut := Array new:colors size.
	transparentIndex := self indexOfColor:Color transparent.
	1 to: colors size do:
		[ :i | lut at:i put:(((colors at:i) = Color transparent) ifTrue:[transparentIndex] ifFalse:[i])].
 ! !

!ColorForm methodsFor: 'fileIn/Out' stamp: 'ar 3/3/2001 20:07'!
hibernate
	"Make myself take up less space. See comment in Form>hibernate."

	super hibernate.
	self clearColormapCache.
	colors ifNotNil:[colors := colors asColorArray].! !

!ColorForm methodsFor: 'color manipulation' stamp: 'di 8/28/1998 15:48'!
indexOfColor: aColor
	"Return the index of aColor in my color array"

	self ensureColorArrayExists.
	^ colors indexOf: aColor ifAbsent: [0]! !

!ColorForm methodsFor: 'testing' stamp: 'ar 5/27/2001 16:34'!
isColorForm
	^true! !

!ColorForm methodsFor: 'testing' stamp: 'ar 2/10/2004 17:18'!
isTranslucent
	"Answer whether this form may be translucent"
	^true! !

!ColorForm methodsFor: 'pixel accessing' stamp: 'tk 10/21/97 12:27'!
isTransparentAt: aPoint 
	"Return true if the receiver is transparent at the given point."

	^ (self colorAt: aPoint) isTransparent
! !

!ColorForm methodsFor: 'color manipulation' stamp: 'jm 10/19/1998 10:52'!
mapColor: oldColor to: newColor
	"Replace all occurances of the given color with the given new color in my color map."

	self ensureColorArrayExists.
	1 to: colors size do: [:i | 
		(colors at: i) = oldColor ifTrue: [colors at: i put: newColor]].
	self clearColormapCache.
! !

!ColorForm methodsFor: 'postscript generation' stamp: ''!
mapTransparencies:transparentIndexes
	^self deepCopy mapColors:transparentIndexes to:(transparentIndexes at:1).! !

!ColorForm methodsFor: 'displaying' stamp: 'ar 12/14/2001 18:14'!
maskingMap
	"Return a color map that maps all colors except transparent to words of all ones. Used to create a mask for a Form whose transparent pixel value is zero."
	| maskingMap |
	maskingMap := Bitmap new: (1 bitShift: depth) withAll: 16rFFFFFFFF.
	1 to: colors size do:[:i|
		(colors at: i) isTransparent ifTrue:[maskingMap at: i put: 0].
	].
	colors size+1 to: maskingMap size do:[:i| maskingMap at: i put: 0].
	^maskingMap! !

!ColorForm methodsFor: 'pixel accessing' stamp: 'ar 5/28/2000 12:06'!
pixelValueAt: aPoint 
	"Return the raw pixel value at the given point. Typical clients use colorAt: to get a Color."
	"Details: To get the raw pixel value, be sure the peeker's colorMap is nil."

	^ (BitBlt current bitPeekerFromForm: self) colorMap: nil; pixelAt: aPoint
! !

!ColorForm methodsFor: 'fileIn/Out' stamp: 'mu 8/17/2003 00:46'!
readAttributesFrom: aBinaryStream
	super readAttributesFrom: aBinaryStream.
	colors := ColorArray new: (2 raisedTo: depth).
	1 to: colors size do: [:idx | 
		colors basicAt: idx put: (aBinaryStream nextLittleEndianNumber: 4).
	]. 
	! !

!ColorForm methodsFor: 'color manipulation' stamp: 'jm 11/16/97 09:08'!
replaceColor: oldColor with: newColor
	"Replace all occurances of the given color with the given new color in my color map."

	self ensureColorArrayExists.
	1 to: colors size do: [:i | 
		(colors at: i) = oldColor ifTrue: [colors at: i put: newColor]].
	self clearColormapCache.
! !

!ColorForm methodsFor: 'color manipulation' stamp: 'tk 3/2/98 15:42'!
replaceColorAt: aPoint with: newColor
	"Replace a color map entry with newColor.  The entry replaced is the one used by aPoint.  If there are are two entries in the colorMap for the oldColor, just replace ONE!!!!  There are often two whites or two blacks, and this is what you want, when replacing one."

	| oldIndex |
	self ensureColorArrayExists.
	oldIndex := self pixelValueAt: aPoint.
	colors at: oldIndex+1 put: newColor.
	self clearColormapCache.
! !

!ColorForm methodsFor: 'color manipulation' stamp: 'di 8/28/1998 15:49'!
replaceColorAtIndex: index with: newColor
	"Replace a color map entry with newColor."

	self ensureColorArrayExists.
	colors at: index put: newColor.
	cachedColormap == nil ifFalse:
		[cachedColormap at: index put: (newColor pixelValueForDepth: cachedDepth)]! !

!ColorForm methodsFor: 'scaling, rotation' stamp: 'RAA 8/5/2000 18:12'!
scaledToSize: newExtent

	"super method did not seem to work so well on ColorForms"

	^(self asFormOfDepth: 16) scaledToSize: newExtent! !

!ColorForm methodsFor: 'private' stamp: 'jm 4/5/1999 10:11'!
setColors: colorArray cachedColormap: aBitmap depth: anInteger
	"Semi-private. Set the color array, cached colormap, and cached colormap depth to avoid having to recompute the colormap when switching color palettes in animations."

	colors := colorArray.
	cachedDepth := anInteger.
	cachedColormap := aBitmap.
! !

!ColorForm methodsFor: 'postscript generation' stamp: ''!
setColorspaceOn:aStream
	self depth = 1 ifTrue:[
		aStream print:'/DeviceRGB setcolorspace 0 setgray'; cr.
	]
	ifFalse:[
	aStream print:'[ /Indexed /DeviceRGB ';
	write:self colors size-1;
	print:' <'.
	(self colormapIfNeededForDepth: 32 ) storeBits:20 to:0 on:aStream.
	aStream print:'> ] setcolorspace'; cr.].
! !

!ColorForm methodsFor: 'private' stamp: 'jm 11/16/97 08:37'!
setExtent: extent depth: bitsPerPixel
	"Create a virtual bit map with the given extent and bitsPerPixel."

	bitsPerPixel > 8 ifTrue: [self error: 'ColorForms only support depths up to 8 bits'].
	super setExtent: extent depth: bitsPerPixel.
! !

!ColorForm methodsFor: 'fileIn/Out' stamp: 'bf 5/25/2000 16:31'!
storeOn: aStream
	aStream nextPut: $(.
	super storeOn: aStream.
	aStream
		cr; tab;
		nextPutAll: 'colorsFromArray: #('.
	self colors do: [:color |
		color storeArrayOn: aStream].
	aStream nextPutAll: ' ))'.! !

!ColorForm methodsFor: 'color manipulation' stamp: 'tk 3/2/98 11:26'!
transparentAllPixelsLike: aPoint
	"Make all occurances of the given pixel value transparent.  Very useful when two entries in the colorMap have the same value.  This only changes ONE."

	self replaceColorAt: aPoint with: Color transparent.
! !

!ColorForm methodsFor: 'color manipulation' stamp: 'tk 3/2/98 11:27'!
transparentColor: aColor
	"Make all occurances of the given color transparent.  Note: for colors like black and white, which have two entries in the colorMap, this changes BOTH of them.  Not always what you want."

	self replaceColor: aColor with: Color transparent.
! !

!ColorForm methodsFor: 'postscript generation' stamp: ''!
transparentColorIndexes
	^(1 to: colors size) select: [ :index | (colors at:index) isTransparent ].
! !

!ColorForm methodsFor: 'color manipulation' stamp: 'ar 5/28/2000 12:06'!
twoToneFromDisplay: aRectangle backgroundColor: bgColor
	"Copy one-bit deep ColorForm from the Display using a color map that maps all colors except the background color to black. Used for caching the contents of inactive MVC windows."

	| map |
	(width = aRectangle width and: [height = aRectangle height])
		ifFalse: [self setExtent: aRectangle extent depth: depth].

	"make a color map mapping the background color
	 to zero and all other colors to one"
	map := Bitmap new: (1 bitShift: (Display depth min: 9)).
	1 to: map size do: [:i | map at: i put: 16rFFFFFFFF].
	map at: (bgColor indexInMap: map) put: 0.

	(BitBlt current toForm: self)
		destOrigin: 0@0;
		sourceForm: Display;
		sourceRect: aRectangle;
		combinationRule: Form over;
		colorMap: map;
		copyBits.
! !

!ColorForm methodsFor: 'fileIn/Out' stamp: 'ar 3/3/2001 20:07'!
unhibernate
	colors ifNotNil:[colors := colors asArray].
	^super unhibernate.
! !

!ColorForm methodsFor: 'private' stamp: 'jm 2/24/98 18:53'!
unusedColormapEntry
	"Return the index of an unused color map entry, or zero if there isn't one."

	| tallies |
	tallies := self tallyPixelValues.
	1 to: tallies size do: [:i |
		(tallies at: i) = 0 ifTrue: [^ i]].
	^ 0
! !

!ColorForm methodsFor: 'fileIn/Out' stamp: 'mu 8/17/2003 00:42'!
writeAttributesOn: file
	| colorArray |
	super writeAttributesOn: file.
	colorArray := self colors asColorArray.
	1 to: (2 raisedTo: depth) do: [:idx |
		file nextLittleEndianNumber: 4 put: (colorArray basicAt: idx).
	] ! !

Form subclass: #Cursor
	instanceVariableNames: ''
	classVariableNames: 'BlankCursor BottomLeftCursor BottomRightCursor CornerCursor CrossHairCursor CurrentCursor DownCursor MarkerCursor MenuCursor MoveCursor NormalCursor OriginCursor ReadCursor ResizeLeftCursor ResizeTopCursor ResizeTopLeftCursor ResizeTopRightCursor RightArrowCursor SquareCursor TargetCursor TopLeftCursor TopRightCursor UpCursor WaitCursor WebLinkCursor WriteCursor XeqCursor'
	poolDictionaries: ''
	category: 'Graphics-Display Objects'!

!Cursor commentStamp: '<historical>' prior: 0!
I am a Form that is a possible appearance for a mouse cursor.  My size is always 16x16, ever since the original implementation on the Alto.

There are many examples available in the "current cursor" category of class methods.  For example, "Cursor normal" and "Cursor wait".  For example:

	Cursor wait show

!

!Cursor class methodsFor: 'constants' stamp: ''!
blank
	"Answer the instance of me that is all white."

	^BlankCursor! !

!Cursor class methodsFor: 'constants' stamp: 'JMM 10/21/2003 19:13'!
bottomLeft
	"Cursor bottomLeft showWhile: [Sensor waitButton]"
	^BottomLeftCursor
! !

!Cursor class methodsFor: 'constants' stamp: 'JMM 10/21/2003 19:13'!
bottomRight
	"Cursor bottomRight showWhile: [Sensor waitButton]"
	^BottomRightCursor
! !

!Cursor class methodsFor: 'constants' stamp: ''!
corner
	"Answer the instance of me that is the shape of the bottom right corner 
	of a rectangle."

	^CornerCursor! !

!Cursor class methodsFor: 'constants' stamp: ''!
crossHair
	"Answer the instance of me that is the shape of a cross."

	^CrossHairCursor! !

!Cursor class methodsFor: 'current cursor' stamp: ''!
currentCursor
	"Answer the instance of Cursor that is the one currently displayed."

	^CurrentCursor! !

!Cursor class methodsFor: 'current cursor' stamp: 'di 10/6/1998 13:57'!
currentCursor: aCursor 
	"Make the instance of cursor, aCursor, be the current cursor. Display it. 
	Create an error if the argument is not a Cursor."

	(aCursor isKindOf: self)
		ifTrue: [CurrentCursor := aCursor.
				aCursor beCursor]
		ifFalse: [self error: 'The new cursor must be an instance of class Cursor']! !

!Cursor class methodsFor: 'constants' stamp: ''!
down
	"Answer the instance of me that is the shape of an arrow facing 
	downward."

	^DownCursor! !

!Cursor class methodsFor: 'constants' stamp: ''!
execute
	"Answer the instance of me that is the shape of an arrow slanted left 
	with a star next to it."

	^XeqCursor! !

!Cursor class methodsFor: 'instance creation' stamp: ''!
extent: extentPoint fromArray: anArray offset: offsetPoint 
	"Answer a new instance of me with width and height specified by
	extentPoint, offset by offsetPoint, and bits from anArray.
	NOTE: This has been kluged to take an array of 16-bit constants,
	and shift them over so they are left-justified in a 32-bit bitmap"

	extentPoint = (16 @ 16)
		ifTrue: 
			[^ super
				extent: extentPoint
				fromArray: (anArray collect: [:bits | bits bitShift: 16])
				offset: offsetPoint]
		ifFalse: [self error: 'cursors must be 16@16']! !

!Cursor class methodsFor: 'class initialization' stamp: 'JMM 10/21/2003 18:57'!
initBottomLeft

	BottomLeftCursor := 
		(Cursor extent: 16@16
			fromArray: #(
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1111111111111111
		2r1111111111111111)
			offset: 0@-16).
! !

!Cursor class methodsFor: 'class initialization' stamp: 'JMM 10/21/2003 18:57'!
initBottomRight

	BottomRightCursor := 
		(Cursor extent: 16@16
			fromArray: #(
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r1111111111111111
		2r1111111111111111)
			offset: -16@-16).
! !

!Cursor class methodsFor: 'class initialization' stamp: ''!
initCorner

	CornerCursor := 
		(Cursor 
			extent: 16@16
			fromArray: #(
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r1111111111111111
		2r1111111111111111)
			offset: -16@-16).
! !

!Cursor class methodsFor: 'class initialization' stamp: 'kfr 7/12/2003 21:02'!
initCrossHair

	CrossHairCursor :=   
		(Cursor
			extent: 16@16
			fromArray: #(
		2r0000000000000000
		2r0000000100000000
		2r0000000100000000
		2r0000000100000000
		2r0000000100000000
		2r0000000100000000
		2r0000000100000000
		2r0111111111111100
		2r0000000100000000
		2r0000000100000000
		2r0000000100000000
		2r0000000100000000
		2r0000000100000000
		2r0000000100000000
		2r0000000000000000
		2r0)
			offset: -7@-7).
	
	! !

!Cursor class methodsFor: 'class initialization' stamp: ''!
initDown

	DownCursor  :=
		     (Cursor
	extent: 16@16
	fromArray: #(
		2r11000000000000
		2r11000000000000
		2r11000000000000
		2r11000000000000
		2r11000000000000
		2r11000000000000
		2r11000000000000
		2r1111110000000000
		2r111100000000000
		2r11000000000000
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0)
	offset: 0@0).
! !

!Cursor class methodsFor: 'class initialization' stamp: ''!
initMarker

	MarkerCursor := 
		Cursor
			extent: 16@16
			fromArray: #(
		2r0111000000000000
		2r1111100000000000
		2r1111100000000000
		2r0111000000000000
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0)
			offset: 0@0.
! !

!Cursor class methodsFor: 'class initialization' stamp: 'di 7/30/2001 10:32'!
initMenu 

	MenuCursor  :=
		        (Cursor
	extent: 16@16
	fromArray: #(
		2r1111111111100000
		2r1000000000100000
		2r1010011000100000
		2r1000000000100000
		2r1101001101100000
		2r1111111111100000
		2r1000000000100000
		2r1011001010100000
		2r1000000000100000
		2r1010110010100000
		2r1000000000100000
		2r1010010100100000
		2r1000000000100000
		2r1111111111100000
		0)
	offset: 0@0).
! !

!Cursor class methodsFor: 'class initialization' stamp: 'kfr 7/12/2003 21:10'!
initMove

	MoveCursor := 
		Cursor 
			extent: 16@16
			fromArray: #(
		2r1111111111111100
		2r1111111111111100
		2r1100001100001100
		2r1100001100001100
		2r1100001100001100
		2r1100001100001100
		2r1111111111111100
		2r1111111111111100
		2r1100001100001100
		2r1100001100001100
		2r1100001100001100
		2r1100001100001100
		2r1111111111111100
		2r1111111111111100
          0)
			offset: 0@0.
! !

!Cursor class methodsFor: 'class initialization' stamp: ''!
initNormal

	NormalCursor :=   
		(Cursor
			extent: 16@16
			fromArray: #(
		2r1000000000000000
		2r1100000000000000
		2r1110000000000000
		2r1111000000000000
		2r1111100000000000
		2r1111110000000000
		2r1111111000000000
		2r1111100000000000
		2r1111100000000000
		2r1001100000000000
		2r0000110000000000
		2r0000110000000000
		2r0000011000000000
		2r0000011000000000
		2r0000001100000000
		2r0000001100000000)
	offset: 0@0).

	
	! !

!Cursor class methodsFor: 'class initialization' stamp: 'di 10/8/1998 17:04'!
initNormalWithMask    "Cursor initNormalWithMask.  Cursor normal show"
	"Next two lines work simply for any cursor..."
	self initNormal.
	NormalCursor := CursorWithMask derivedFrom: NormalCursor.

	"But for a good looking cursor, you have to tweak things..."
	NormalCursor := (CursorWithMask extent: 16@16 depth: 1
			fromArray: #( 0 1073741824 1610612736 1879048192
				2013265920 2080374784 2113929216 2130706432
				2080374784 2080374784 1275068416 100663296
				100663296 50331648 50331648 0)
			offset: -1@-1)
		setMaskForm: (Form extent: 16@16 depth: 1
			fromArray: #( 3221225472 3758096384 4026531840 4160749568
				4227858432 4261412864 4278190080 4286578688
				4278190080 4261412864 4261412864 3472883712
				251658240 125829120 125829120 50331648)
			offset: 0@0).! !

!Cursor class methodsFor: 'class initialization' stamp: ''!
initOrigin

	OriginCursor :=   
		(Cursor
			extent: 16@16
			fromArray: #(
		2r1111111111111111
		2r1111111111111111
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000)
			offset: 0@0).
! !

!Cursor class methodsFor: 'class initialization' stamp: 'kfr 7/12/2003 22:55'!
initRead

	ReadCursor :=  
		(Cursor
			extent: 16@16
			fromArray: #(
		2r0000000000000000
		2r0000000000000000
		2r0001000000001000
		2r0010100000010100
		2r0100000000100000
		2r1111101111100000
		2r1000010000100000
		2r1000010000100000
		2r1011010110100000
		2r0111101111000000
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0)
	offset: 0@0).
! !

!Cursor class methodsFor: 'class initialization' stamp: 'jrp 8/6/2005 22:50'!
initResizeLeft

       ResizeLeftCursor :=
               (Cursor extent: 16@16 fromArray: #(
               2r0000000000000000
               2r0000000000000000
               2r0000000000000000
               2r0000000000000000
               2r0000100000010000
               2r0001100000011000
               2r0011100000011100
               2r0111111111111110
               2r0011100000011100
               2r0001100000011000
               2r0000100000010000
               2r0000000000000000
               2r0000000000000000
               2r0000000000000000
               2r0000000000000000
               2r0000000000000000 )
       offset: -7@-7 ) withMask! !

!Cursor class methodsFor: 'class initialization' stamp: 'jrp 8/6/2005 22:54'!
initResizeTop
    "Cursor initResizeTop"
       ResizeTopCursor :=
               (Cursor extent: 16@16 fromArray: #(
               2r000000100000000
               2r000001110000000
               2r000011111000000
               2r000111111100000
               2r000000100000000
               2r000000100000000
               2r000000100000000
               2r000000100000000
               2r000000100000000
               2r000000100000000
               2r000111111100000
               2r000011111000000
               2r000001110000000
               2r000000100000000
               2r000000000000000)
       offset: -7@-7) withMask! !

!Cursor class methodsFor: 'class initialization' stamp: 'jrp 8/6/2005 22:55'!
initResizeTopLeft

       ResizeTopLeftCursor :=
               (Cursor extent: 16@16 fromArray: #(
               2r0000000000000000
               2r0111110000000000
               2r0111100000000000
               2r0111000000000000
               2r0110100000000000
               2r0100010000000000
               2r0000001000000000
               2r0000000100000000
               2r0000000010000000
               2r0000000001000100
               2r0000000000101100
               2r0000000000011100
               2r0000000000111100
               2r0000000001111100
               2r0000000000000000
               2r0000000000000000)
       offset: -7@-7) withMask! !

!Cursor class methodsFor: 'class initialization' stamp: 'jrp 8/7/2005 07:54'!
initResizeTopRight

       ResizeTopRightCursor :=
               (Cursor extent: 16@16 fromArray: #(
               2r0000000000000000
               2r0000000001111100
               2r0000000000111100
               2r0000000000011100
               2r0000000000101100
               2r0000000001000100
               2r0000000010000000
               2r0000000100000000
               2r0000001000000000
               2r0100010000000000
               2r0110100000000000
               2r0111000000000000
               2r0111100000000000
               2r0111110000000000
               2r0000000000000000
               2r0000000000000000)
       offset: -7@-7) withMask! !

!Cursor class methodsFor: 'class initialization' stamp: ''!
initRightArrow 

	RightArrowCursor  :=
		      (Cursor
	extent: 16@16
	fromArray: #(
		2r100000000000
		2r111000000000
		2r1111111110000000
		2r111000000000
		2r100000000000
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0)
	offset: 0@0).
	
	"Cursor initRightArrow"! !

!Cursor class methodsFor: 'class initialization' stamp: ''!
initSquare

	SquareCursor := 
		(Cursor
			extent: 16@16
			fromArray: #(
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0000001111000000
		2r0000001111000000
		2r0000001111000000
		2r0000001111000000
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0)
	offset: -8@-8).

	! !

!Cursor class methodsFor: 'class initialization' stamp: 'ar 3/1/2006 22:42'!
initTarget
	^TargetCursor := Cursor
				extent: 16 @ 16
				fromArray:  #(1984 6448 8456 16644 17284 33026 35106 65278 35106 33026 17284 16644 8456 6448 1984 0)
				offset: -7 @ -7! !

!Cursor class methodsFor: 'class initialization' stamp: 'JMM 10/21/2003 19:01'!
initTopLeft
	TopLeftCursor := 
		(Cursor extent: 16@16
			fromArray: #(
		2r1111111111111111
		2r1111111111111111
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000
		2r1100000000000000)
			offset: 0@0).
! !

!Cursor class methodsFor: 'class initialization' stamp: 'JMM 10/21/2003 19:02'!
initTopRight
	TopRightCursor := 
		(Cursor extent: 16@16
			fromArray: #(
		2r1111111111111111
		2r1111111111111111
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011
		2r0000000000000011)
			offset: -16@0).
! !

!Cursor class methodsFor: 'class initialization' stamp: ''!
initUp

	UpCursor := 
		    (Cursor
	extent: 16@16
	fromArray: #(
		2r11000000000000
		2r111100000000000
		2r1111110000000000
		2r11000000000000
		2r11000000000000
		2r11000000000000
		2r11000000000000
		2r11000000000000
		2r11000000000000
		2r11000000000000
		2r0
		2r0
		2r0
		2r0
		2r0
		2r0)
	offset: 0@0).
! !

!Cursor class methodsFor: 'class initialization' stamp: 'kfr 7/12/2003 21:27'!
initWait

	WaitCursor := 
		  (Cursor
			extent: 16@16
			fromArray: #(
		2r1111111111111100
		2r1000000000000100
		2r0100000000001000
		2r0010000000010000
		2r0001110011100000
		2r0000111111000000
		2r0000011110000000
		2r0000011110000000
		2r0000100101000000
		2r0001000100100000
		2r0010000110010000
		2r0100001111001000
		2r1000111111110100
		2r1111111111111100
		0)
			offset: 0@0).
! !

!Cursor class methodsFor: 'class initialization' stamp: 'kfr 7/12/2003 22:52'!
initWrite

	WriteCursor := (Cursor
	extent: 16@16
	fromArray: #(
		2r0000000000011000
		2r0000000000111100
		2r0000000001001000
		2r0000000010010000
		2r0000000100100000
		2r0000001001000100
		2r0000010010000100
		2r0000100100001100
		2r0001001000010000
		2r0010010000010000
		2r0111100000001000
		2r0101000011111000
		2r1110000110000000
		2r0111111100000000
		2r0
		2r0)
	offset: 0@0).
! !

!Cursor class methodsFor: 'class initialization' stamp: ''!
initXeq

	XeqCursor := 
		(Cursor
			extent: 16@16
			fromArray: #(
		2r1000000000010000
		2r1100000000010000
		2r1110000000111000
		2r1111000111111111
		2r1111100011000110
		2r1111110001000100
		2r1111111001111100
		2r1111000001101100
		2r1101100011000110
		2r1001100010000010
		2r0000110000000000
		2r0000110000000000
		2r0000011000000000
		2r0000011000000000
		2r0000001100000000
		2r0000001100000000)
	offset: 0@0).
! !

!Cursor class methodsFor: 'class initialization' stamp: 'JMM 10/21/2003 19:04'!
initialize
	"Create all the standard cursors..."
		self initOrigin.
		self initRightArrow.
		self initMenu.
		self initCorner.
		self initRead.
		self initWrite.
		self initWait.
		BlankCursor := Cursor new.
		self initXeq.
		self initSquare.
		self initNormalWithMask.
		self initCrossHair.
		self initMarker.
		self initUp.
		self initDown.
		self initMove.
		self initBottomLeft.
		self initBottomRight.
		self initResizeLeft.
		self initResizeTop.
		self initResizeTopLeft.
		self initResizeTopRight.
		self initTopLeft.
		self initTopRight.
		self makeCursorsWithMask.

		"Cursor initialize"
! !

!Cursor class methodsFor: 'class initialization' stamp: 'ar 7/18/2009 20:42'!
makeCursorsWithMask
	"Cursor initialize;makeCursorsWithMask"

	self classPool associationsDo: [:var |
		((var value isKindOf: Cursor) and:[var value hasMask not])
			ifTrue: [var value: var value withMask]]! !

!Cursor class methodsFor: 'constants' stamp: ''!
marker
	"Answer the instance of me that is the shape of a small ball."

	^MarkerCursor! !

!Cursor class methodsFor: 'constants' stamp: ''!
menu 
	"Answer the instance of me that is the shape of a menu."

	^MenuCursor! !

!Cursor class methodsFor: 'constants' stamp: ''!
move
	"Answer the instance of me that is the shape of a cross inside a square."

	^MoveCursor! !

!Cursor class methodsFor: 'instance creation' stamp: 'di 10/6/1998 13:53'!
new

	^ self extent: 16 @ 16
		fromArray: (Array new: 16 withAll: 0)
		offset: 0 @ 0

	"Cursor new bitEdit show"! !

!Cursor class methodsFor: 'constants' stamp: ''!
normal
	"Answer the instance of me that is the shape of an arrow slanted left."

	^NormalCursor! !

!Cursor class methodsFor: 'constants' stamp: ''!
origin
	"Answer the instance of me that is the shape of the top left corner of a 
	rectangle."

	^OriginCursor! !

!Cursor class methodsFor: 'constants' stamp: ''!
read
	"Answer the instance of me that is the shape of eyeglasses."

	^ReadCursor! !

!Cursor class methodsFor: 'constants' stamp: 'ar 8/16/2001 14:48'!
resizeBottom
	"Cursor resizeBottom showWhile: [Sensor waitButton]"
	^self resizeTop! !

!Cursor class methodsFor: 'constants' stamp: 'ar 8/16/2001 14:46'!
resizeBottomLeft
	"Cursor resizeBottomLeft showWhile: [Sensor waitButton]"
	^self resizeTopRight! !

!Cursor class methodsFor: 'constants' stamp: 'ar 8/16/2001 14:45'!
resizeBottomRight
	"Cursor resizeBottomRight showWhile: [Sensor waitButton]"
	^self resizeTopLeft! !

!Cursor class methodsFor: 'instance creation' stamp: 'ar 8/16/2001 15:52'!
resizeForEdge: aSymbol
	"Cursor resizeForEdge: #top"
	"Cursor resizeForEdge: #bottomLeft"
	^self perform: ('resize', aSymbol first asString asUppercase, (aSymbol copyFrom: 2 to: aSymbol size)) asSymbol.! !

!Cursor class methodsFor: 'constants' stamp: 'JMM 10/21/2003 18:58'!
resizeLeft
	"Cursor resizeLeft showWhile: [Sensor waitButton]"
	^ResizeLeftCursor! !

!Cursor class methodsFor: 'constants' stamp: 'ar 8/16/2001 14:45'!
resizeRight
	"Cursor resizeRight showWhile: [Sensor waitButton]"
	^self resizeLeft! !

!Cursor class methodsFor: 'constants' stamp: 'JMM 10/21/2003 19:19'!
resizeTop
	"Cursor resizeTop showWhile: [Sensor waitButton]"
	^ResizeTopCursor! !

!Cursor class methodsFor: 'constants' stamp: 'JMM 10/21/2003 19:00'!
resizeTopLeft
	"Cursor resizeTopLeft showWhile: [Sensor waitButton]"
	^ ResizeTopLeftCursor! !

!Cursor class methodsFor: 'constants' stamp: 'JMM 10/21/2003 19:00'!
resizeTopRight
	"Cursor resizeTopRight showWhile: [Sensor waitButton]"
	^ResizeTopRightCursor! !

!Cursor class methodsFor: 'constants' stamp: ''!
rightArrow 
	"Answer the instance of me that is the shape of an arrow pointing to the right."

	^RightArrowCursor! !

!Cursor class methodsFor: 'constants' stamp: ''!
square
	"Answer the instance of me that is the shape of a square."

	^SquareCursor! !

!Cursor class methodsFor: 'class initialization' stamp: ''!
startUp
	self currentCursor: self currentCursor! !

!Cursor class methodsFor: 'constants' stamp: 'ar 3/1/2006 22:42'!
target
	"Answer the instance of me that is the shape of a gunsight."
	"Cursor target show"
	^TargetCursor ifNil:[self initTarget]! !

!Cursor class methodsFor: 'constants' stamp: 'JMM 10/21/2003 19:01'!
topLeft
	"Cursor topLeft showWhile: [Sensor waitButton]"
	^ TopLeftCursor! !

!Cursor class methodsFor: 'constants' stamp: 'JMM 10/21/2003 19:02'!
topRight
	"Cursor topRight showWhile: [Sensor waitButton]"
	^ TopRightCursor! !

!Cursor class methodsFor: 'constants' stamp: ''!
up
	"Answer the instance of me that is the shape of an arrow facing upward."

	^UpCursor! !

!Cursor class methodsFor: 'constants' stamp: 'sw 8/15/97 13:28'!
wait
	"Answer the instance of me that is the shape of an Hourglass (was in the 
	shape of three small balls)."

	^WaitCursor! !

!Cursor class methodsFor: 'constants' stamp: 'ar 9/26/2001 22:37'!
webLink
	"Return a cursor that can be used for emphasizing web links"
	"Cursor webLink showWhile: [Sensor waitButton]"
	^WebLinkCursor ifNil:[
		WebLinkCursor :=  (CursorWithMask extent: 16@16
			fromArray: #(3072 4608 4608 4608 4608 5046 4681 29257 37449 37449 32769 32769 49155 16386 24582 16380 )
			offset: -5@0) setMaskForm:
		(Form extent: 16@16 
			fromArray: (#(3072 7680 7680 7680 7680 8118 8191 32767 65535 65535 65535 65535 65535 32766 32766 16380 )  collect: [:bits | bits bitShift: 16])
			offset: 0@0)].! !

!Cursor class methodsFor: 'constants' stamp: ''!
write
	"Answer the instance of me that is the shape of a pen writing."

	^WriteCursor! !

!Cursor methodsFor: 'converting' stamp: 'RAA 8/14/2000 10:14'!
asCursorForm
	| form |
	form := StaticForm extent: self extent depth: 8.
	form fillShape: self fillColor: Color black at: offset negated.
	^ form offset: offset! !

!Cursor methodsFor: 'primitives' stamp: ''!
beCursor
	"Primitive. Tell the interpreter to use the receiver as the current cursor 
	image. Fail if the receiver does not match the size expected by the 
	hardware. Essential. See Object documentation whatIsAPrimitive."

	<primitive: 101>
	self primitiveFailed! !

!Cursor methodsFor: 'primitives' stamp: 'jm 9/22/1998 23:33'!
beCursorWithMask: maskForm
	"Primitive. Tell the interpreter to use the receiver as the current cursor image with the given mask Form. Both the receiver and the mask should have extent 16@16 and a depth of one. The mask and cursor bits are combined as follow:
			mask	cursor	effect
			 0		  0		transparent (underlying pixel shows through)
			 1		  1		opaque black
			 1		  0		opaque white
			 0		  1		invert the underlying pixel"
"Essential. See Object documentation whatIsAPrimitive."

	<primitive: 101>
	self primitiveFailed
! !

!Cursor methodsFor: 'updating' stamp: 'ls 6/17/2002 12:00'!
changed: aParameter
	"overriden to reinstall the cursor if it is the active cursor, in case the appearance has changed.  (Is this used anywhere?  Do cursors really change in place these days?)"
	self == CurrentCursor ifTrue: [self beCursor].
	super changed: aParameter! !

!Cursor methodsFor: 'testing' stamp: 'bf 2/2/1999 19:34'!
hasMask
	^false! !

!Cursor methodsFor: 'printing' stamp: ''!
printOn: aStream

	self storeOn: aStream base: 2! !

!Cursor methodsFor: 'displaying' stamp: 'ls 6/17/2002 11:56'!
show
	"Make the hardware's mouse cursor look like the receiver"

	Sensor currentCursor: self! !

!Cursor methodsFor: 'displaying' stamp: ''!
showGridded: gridPoint 
	"Make the current cursor shape be the receiver, forcing the location of the cursor to the point nearest gridPoint."
	
	Sensor cursorPoint: (Sensor cursorPoint grid: gridPoint).
	Sensor currentCursor: self! !

!Cursor methodsFor: 'displaying' stamp: 'bf 10/13/1999 13:05'!
showWhile: aBlock 
	"While evaluating the argument, aBlock, make the receiver be the cursor 
	shape."

	| oldcursor |
	oldcursor := Sensor currentCursor.
	self show.
	^aBlock ensure: [oldcursor show]
! !

!Cursor methodsFor: 'converting' stamp: 'bf 2/2/1999 19:32'!
withMask
	^CursorWithMask derivedFrom: self! !

Cursor subclass: #CursorWithMask
	instanceVariableNames: 'maskForm'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Display Objects'!

!CursorWithMask commentStamp: '<historical>' prior: 0!
A Cursor which additionally has a 16x16 transparency bitmap called a "mask".  See the comment of beCursorWithMask: for details on how the mask is treated.!

!CursorWithMask class methodsFor: 'as yet unclassified' stamp: 'di 2/18/1999 08:56'!
derivedFrom: aForm      "Cursor initNormalWithMask.  Cursor normal show"
	"aForm is presumably a cursor"
	| cursor mask ext |
	ext := aForm extent.
	cursor := self extent: ext.
	cursor copy: (1@1 extent: ext) from: 0@0 in: aForm rule: Form over.
	mask := Form extent: ext.
	(1@1) eightNeighbors do:
		[:p | mask copy: (p extent: ext) from: 0@0 in: aForm rule: Form under].
	cursor setMaskForm: mask.
	cursor offset: ((aForm offset - (1@1)) max: ext negated).
	^ cursor! !

!CursorWithMask methodsFor: 'converting' stamp: 'RAA 8/14/2000 10:14'!
asCursorForm
	| form |
	form := StaticForm extent: self extent depth: 8.
	form fillShape: maskForm fillColor: Color white.
	form fillShape: self fillColor: Color black at: offset negated.
	^ form offset: offset! !

!CursorWithMask methodsFor: 'primitives' stamp: 'di 10/6/1998 15:16'!
beCursor
	maskForm unhibernate.
	^ self beCursorWithMask: maskForm! !

!CursorWithMask methodsFor: 'mask' stamp: 'bf 2/2/1999 19:34'!
hasMask
	^true! !

!CursorWithMask methodsFor: 'mask' stamp: 'di 10/8/1998 16:46'!
maskForm
	^ maskForm! !

!CursorWithMask methodsFor: 'mask' stamp: 'di 10/8/1998 16:46'!
setMaskForm: aForm
	maskForm := aForm! !

!CursorWithMask methodsFor: 'mask' stamp: 'bf 2/2/1999 19:30'!
storeOn: aStream base: anInteger

	aStream nextPut: $(.
	super storeOn: aStream base: anInteger.
	aStream nextPutAll: ' setMaskForm: '.
	maskForm storeOn: aStream base: anInteger.
	aStream nextPut: $)! !

!CursorWithMask methodsFor: 'mask' stamp: 'bf 2/2/1999 19:31'!
withMask
	^self! !

Form subclass: #DisplayScreen
	instanceVariableNames: 'clippingBox extraRegions'
	classVariableNames: 'DeferringUpdates DisplayChangeSignature ScreenSave'
	poolDictionaries: ''
	category: 'Graphics-Display Objects'!

!DisplayScreen commentStamp: '<historical>' prior: 0!
There is only one instance of me, Display. It is a global and is used to handle general user requests to deal with the whole display screen. 
	Although I offer no protocol, my name provides a way to distinguish this special instance from all other Forms. This is useful, for example, in dealing with saving and restoring the system.
	To change the depth of your Display...
		Display newDepth: 16.
		Display newDepth: 8.
		Display newDepth: 1.
Valid display depths are 1, 2, 4, 8, 16 and 32.  It is suggested that you run with your monitors setting the same, for better speed and color fidelity.  Note that this can add up to 4Mb for the Display form.  Finally, note that newDepth: ends by executing a 'ControlManager restore' which currently terminates the active process, so nothing that follows in the doit will get executed.

Depths 1, 2, 4 and 8 bits go through a color map to put color on the screen, but 16 and 32-bit color use the pixel values directly for RGB color (5 and 8 bits per, respectivlely).  The color choice an be observed by executing Color fromUser in whatever depth you are using.
!

!DisplayScreen class methodsFor: 'snapshots' stamp: 'ar 2/5/2001 17:24'!
actualScreenDepth
	<primitive: 'primitiveScreenDepth'>
	^ Display depth! !

!DisplayScreen class methodsFor: 'snapshots' stamp: ''!
actualScreenSize
	<primitive: 106>
	^ 640@480! !

!DisplayScreen class methodsFor: 'display box access' stamp: ''!
boundingBox
	"Answer the bounding box for the form representing the current display 
	screen."

	^Display boundingBox! !

!DisplayScreen class methodsFor: 'display box access' stamp: 'sw 10/31/2001 07:18'!
checkForNewScreenSize
	"Check whether the screen size has changed and if so take appropriate actions"

	Display extent = DisplayScreen actualScreenSize ifTrue: [^ self].
	DisplayScreen startUp.
	Smalltalk isMorphic
		ifTrue:
			[World restoreMorphicDisplay.
			World repositionFlapsAfterScreenSizeChange]
		ifFalse:
			[ScheduledControllers restore; searchForActiveController]! !

!DisplayScreen class methodsFor: 'display box access' stamp: 'sma 4/28/2000 19:07'!
depth: depthInteger width: widthInteger height: heightInteger fullscreen: aBoolean
	"Force Squeak's window (if there's one) into a new size and depth."
	"DisplayScreen depth: 8 width: 1024 height: 768 fullscreen: false"

	<primitive: 92>
	self primitiveFail! !

!DisplayScreen class methodsFor: 'snapshots' stamp: 'ar 5/28/2000 11:26'!
shutDown 
	"Minimize Display memory saved in image"
	Display shutDown.! !

!DisplayScreen class methodsFor: 'snapshots' stamp: 'ar 5/17/2001 15:50'!
startUp  "DisplayScreen startUp"
	Display setExtent: self actualScreenSize depth: Display nativeDepth.
	Display beDisplay! !

!DisplayScreen methodsFor: 'displaying' stamp: 'ar 4/19/2001 05:44'!
addExtraRegion: aRectangle for: regionDrawer
	"Register the given rectangle as a region which is drawn by the specified region drawer. The region will be excluded from any updates when #forceDamageToScreen: is called. Note that the rectangle is only valid for a single update cycle; once #forceDamageToScreen: has been called, the region drawer and its region are being removed from the list"
	extraRegions ifNil:[extraRegions := #()].
	extraRegions := extraRegions copyWith: (Array with: regionDrawer with: aRectangle).
! !

!DisplayScreen methodsFor: 'private' stamp: ''!
beDisplay
	"Primitive. Tell the interpreter to use the receiver as the current display 
	image. Fail if the form is too wide to fit on the physical display. 
	Essential. See Object documentation whatIsAPrimitive."

	<primitive: 102>
	self primitiveFailed! !

!DisplayScreen methodsFor: 'other' stamp: ''!
boundingBox
	clippingBox == nil
		ifTrue: [clippingBox := super boundingBox].
	^ clippingBox! !

!DisplayScreen methodsFor: 'other' stamp: ''!
clippingTo: aRect do: aBlock
	"Display clippingTo: Rectangle fromUser do:
	[ScheduledControllers restore: Display fullBoundingBox]"
	| saveClip |
	saveClip := clippingBox.
	clippingBox := aRect.
	aBlock value.
	clippingBox := saveClip! !

!DisplayScreen methodsFor: 'displaying' stamp: 'ar 5/28/2000 12:07'!
copyBits: rect from: sf at: destOrigin clippingBox: clipRect rule: cr fillColor: hf 
	(BitBlt current
		destForm: self
		sourceForm: sf
		fillColor: hf
		combinationRule: cr
		destOrigin: destOrigin
		sourceOrigin: rect origin
		extent: rect extent
		clipRect: (clipRect intersect: clippingBox)) copyBits! !

!DisplayScreen methodsFor: 'displaying' stamp: 'ar 5/28/2000 12:07'!
copyBits: rect from: sf at: destOrigin clippingBox: clipRect rule: cr fillColor: hf map: map
	((BitBlt current
		destForm: self
		sourceForm: sf
		fillColor: hf
		combinationRule: cr
		destOrigin: destOrigin
		sourceOrigin: rect origin
		extent: rect extent
		clipRect: (clipRect intersect: clippingBox)) colorMap: map) copyBits! !

!DisplayScreen methodsFor: 'private' stamp: 'di 3/3/1999 10:00'!
copyFrom: aForm
	"Take on all state of aForm, with complete sharing"

	super copyFrom: aForm.
	clippingBox := super boundingBox! !

!DisplayScreen methodsFor: 'blitter defaults' stamp: 'ar 5/28/2000 12:01'!
defaultBitBltClass
	"Return the BitBlt version to use when I am active"
	^BitBlt! !

!DisplayScreen methodsFor: 'blitter defaults' stamp: 'ar 5/28/2000 12:02'!
defaultCanvasClass
	"Return the WarpBlt version to use when I am active"
	^FormCanvas! !

!DisplayScreen methodsFor: 'blitter defaults' stamp: 'ar 5/28/2000 12:01'!
defaultWarpBltClass
	"Return the WarpBlt version to use when I am active"
	^WarpBlt! !

!DisplayScreen methodsFor: 'other' stamp: 'hmm 6/18/2000 19:16'!
deferUpdates: aBoolean
	| wasDeferred |
	"Set the deferUpdates flag in the virtual machine. When this flag is true, BitBlt operations on the Display are not automatically propagated to the screen. If this underlying platform does not support deferred updates, this primitive will fail. Answer whether updates were deferred before if the primitive succeeds, nil if it fails."

	wasDeferred := DeferringUpdates == true.
	DeferringUpdates := aBoolean.
	^(self primitiveDeferUpdates: aBoolean) ifNotNil: [wasDeferred]! !

!DisplayScreen methodsFor: 'other' stamp: 'hmm 2/2/2001 10:14'!
deferUpdatesIn: aRectangle while: aBlock
	| result |
	(self deferUpdates: true) ifTrue: [^aBlock value].
	result := aBlock value.
	self deferUpdates: false.
	self forceToScreen: aRectangle.
	^result! !

!DisplayScreen methodsFor: 'other' stamp: 'RAA 11/27/1999 15:48'!
displayChangeSignature

	^DisplayChangeSignature! !

!DisplayScreen methodsFor: 'private' stamp: 'ar 5/25/2000 23:43'!
findAnyDisplayDepth
	"Return any display depth that is supported on this system."
	^self findAnyDisplayDepthIfNone:[
		"Ugh .... now this is a biggie - a system that does not support
		any of the Squeak display depths at all."
		Smalltalk
			logError:'Fatal error: This system has no support for any display depth at all.'
			inContext: thisContext
			to: 'SqueakDebug.log'.
		Smalltalk quitPrimitive. "There is no way to continue from here"
	].! !

!DisplayScreen methodsFor: 'private' stamp: 'ar 5/17/2001 21:03'!
findAnyDisplayDepthIfNone: aBlock
	"Return any display depth that is supported on this system.
	If there is none, evaluate aBlock."
	#(1 2 4 8 16 32 -1 -2 -4 -8 -16 -32) do:[:bpp|
		(self supportsDisplayDepth: bpp) ifTrue:[^bpp].
	].
	^aBlock value! !

!DisplayScreen methodsFor: 'displaying' stamp: 'jm 5/22/1998 01:23'!
flash: aRectangle 
	"Flash the area of the screen defined by the given rectangle."

	self reverse: aRectangle.
	self forceDisplayUpdate.
	(Delay forMilliseconds: 100) wait.
	self reverse: aRectangle.
	self forceDisplayUpdate.
! !

!DisplayScreen methodsFor: 'displaying' stamp: 'RAA 6/2/2000 12:09'!
flash: aRectangle andWait: msecs
	"Flash the area of the screen defined by the given rectangle."

	self reverse: aRectangle.
	self forceDisplayUpdate.
	(Delay forMilliseconds: msecs) wait.
	self reverse: aRectangle.
	self forceDisplayUpdate.
	(Delay forMilliseconds: msecs) wait.
! !

!DisplayScreen methodsFor: 'displaying' stamp: 'sw 1/1/2005 01:31'!
flashAll: rectangleList andWait: msecs
	"Flash the areas of the screen defined by the given rectangles."

	rectangleList do: [:aRectangle | self reverse: aRectangle].
	self forceDisplayUpdate.
	(Delay forMilliseconds: msecs) wait.
	rectangleList do: [:aRectangle | self reverse: aRectangle].
	self forceDisplayUpdate.
	(Delay forMilliseconds: msecs) wait.
! !

!DisplayScreen methodsFor: 'displaying' stamp: 'ar 5/15/2001 20:08'!
forceDamageToScreen: allDamage
	"Force all the damage rects to the screen."
	| rectList excluded remaining regions |
	rectList := allDamage.
	"Note: Reset extra regions at the beginning to prevent repeated errors"
	regions := extraRegions.
	extraRegions := nil.
	regions ifNotNil:[
		"exclude extra regions"
		regions do:[:drawerAndRect|
			excluded := drawerAndRect at: 2.
			remaining := WriteStream on: #().
			rectList do:[:r|
				remaining nextPutAll:(r areasOutside: excluded)].
			rectList := remaining contents].
	].
	rectList do:[:r| self forceToScreen: r].
	regions ifNotNil:[
		"Have the drawers paint what is needed"
		regions do:[:drawerAndRect| (drawerAndRect at: 1) forceToScreen].
	].! !

!DisplayScreen methodsFor: 'other' stamp: 'jm 5/21/1998 23:48'!
forceDisplayUpdate
	"On platforms that buffer screen updates, force the screen to be updated immediately. On other platforms, or if the primitive is not implemented, do nothing."

	<primitive: 231>
	"do nothing if primitive fails"! !

!DisplayScreen methodsFor: 'other' stamp: 'ar 2/11/1999 18:14'!
forceToScreen
	"Force the entire display area to the screen"
	^self forceToScreen: self boundingBox! !

!DisplayScreen methodsFor: 'other' stamp: 'jm 5/19/1998 17:50'!
forceToScreen: aRectangle
	"Force the given rectangular section of the Display to be copied to the screen. The primitive call does nothing if the primitive is not implemented. Typically used when the deferUpdates flag in the virtual machine is on; see deferUpdates:."

	self primShowRectLeft: aRectangle left
		right: aRectangle right
		top: aRectangle top
		bottom: aRectangle bottom.
! !

!DisplayScreen methodsFor: 'other' stamp: ''!
fullBoundingBox
	^ super boundingBox! !

!DisplayScreen methodsFor: 'other' stamp: ''!
fullScreen   "Display fullScreen"

	ScreenSave notNil ifTrue: [Display := ScreenSave].
	clippingBox := super boundingBox! !

!DisplayScreen methodsFor: 'other' stamp: 'sd 6/7/2003 19:46'!
fullScreenMode: aBoolean
	"On platforms that support it, set full-screen mode to the value of the argument. (Note: you'll need to restore the Display after calling this primitive."
	"Display fullScreenMode: true. Display newDepth: Display depth"

	<primitive: 233>
	self primitiveFailed
! !

!DisplayScreen methodsFor: 'other' stamp: ''!
height
	^ self boundingBox height! !

!DisplayScreen methodsFor: 'testing' stamp: 'ar 5/25/2000 23:34'!
isDisplayScreen
	^true! !

!DisplayScreen methodsFor: 'other' stamp: 'ar 5/5/1999 23:44'!
newDepth: pixelSize
"
	Display newDepth: 8.
	Display newDepth: 1.
"
	(self supportsDisplayDepth: pixelSize)
		ifFalse:[^self inform:'Display depth ', pixelSize printString, ' is not supported on this system'].
	self newDepthNoRestore: pixelSize.
	self restore.! !

!DisplayScreen methodsFor: 'private' stamp: 'ar 5/17/2001 15:44'!
newDepthNoRestore: pixelSize
	"Change depths.  Check if there is enough space!!  , di"
	| area need |
	pixelSize = depth ifTrue: [^ self  "no change"].
	pixelSize abs < self depth ifFalse:
		["Make sure there is enough space"
		area := Display boundingBox area. "pixels"
		Smalltalk isMorphic ifFalse:
			[ScheduledControllers scheduledWindowControllers do:
				[:aController | "This should be refined..."
				aController view cacheBitsAsTwoTone ifFalse:
					[area := area + aController view windowBox area]]].
		need := (area * (pixelSize abs - self depth) // 8)  "new bytes needed"
				+ Smalltalk lowSpaceThreshold.
		(Smalltalk garbageCollectMost <= need
			and: [Smalltalk garbageCollect <= need])
			ifTrue: [self error: 'Insufficient free space']].
	self setExtent: self extent depth: pixelSize.
	Smalltalk isMorphic ifFalse: [ScheduledControllers updateGray].
	DisplayScreen startUp! !

!DisplayScreen methodsFor: 'disk I/O' stamp: 'tk 9/28/2000 15:41'!
objectForDataStream: refStrm
	| dp |
	"I am about to be written on an object file.  Write a reference to the Display in the other system instead.  "

	"A path to me"
	dp := DiskProxy global: #Display selector: #yourself args: #().
	refStrm replace: self with: dp.
	^ dp
! !

!DisplayScreen methodsFor: 'private' stamp: 'jm 6/3/1998 13:00'!
primRetryShowRectLeft: l right: r top: t bottom: b
	"Copy the given rectangular section of the Display to to the screen. This primitive is not implemented on all platforms. Do nothing if it fails. "

	<primitive: 127>
	"do nothing if primitive fails"
! !

!DisplayScreen methodsFor: 'private' stamp: 'jm 6/3/1998 13:02'!
primShowRectLeft: l right: r top: t bottom: b
	"Copy the given rectangular section of the Display to to the screen. This primitive is not implemented on all platforms. If this fails, retry integer coordinates."

	<primitive: 127>
	"if this fails, coerce coordinates to integers and try again"
	self primRetryShowRectLeft: l truncated
		right: r rounded
		top: t truncated
		bottom: b rounded.
! !

!DisplayScreen methodsFor: 'other' stamp: 'hmm 6/18/2000 19:14'!
primitiveDeferUpdates: aBoolean
	"Set the deferUpdates flag in the virtual machine. When this flag is true, BitBlt operations on the Display are not automatically propagated to the screen. If this underlying platform does not support deferred updates, this primitive will fail. Answer the receiver if the primitive succeeds, nil if it fails."

	<primitive: 126>
	^ nil  "answer nil if primitive fails"
! !

!DisplayScreen methodsFor: 'initialize-release' stamp: 'ar 5/26/2000 00:07'!
release
	"I am no longer Display. Release any resources if necessary"! !

!DisplayScreen methodsFor: 'other' stamp: ''!
replacedBy: aForm do: aBlock
	"Permits normal display to draw on aForm instead of the display."

	ScreenSave := self.
	Display := aForm.
	aBlock value.
	Display := self.
	ScreenSave := nil.! !

!DisplayScreen methodsFor: 'other' stamp: 'ar 3/17/2001 23:53'!
restore
	Smalltalk isMorphic
		ifTrue: [World fullRepaintNeeded]
		ifFalse: [ScheduledControllers unCacheWindows; restore].! !

!DisplayScreen methodsFor: 'other' stamp: 'ar 3/17/2001 23:53'!
restoreAfter: aBlock
	"Evaluate the block, wait for a mouse click, and then restore the screen."

	aBlock value.
	Sensor waitButton.
	Smalltalk isMorphic
		ifTrue: [World fullRepaintNeeded]
		ifFalse: [(ScheduledControllers restore; activeController) view emphasize]! !

!DisplayScreen methodsFor: 'private' stamp: 'bf 5/16/2006 11:35'!
setExtent: aPoint depth: bitsPerPixel  "DisplayScreen startUp"
	"This method is critical.  If the setExtent fails, there will be no
	proper display on which to show the error condition..."
	"ar 5/1/1999: ... and that is exactly why we check for the available display depths first."

	"RAA 27 Nov 99 - if depth and extent are the same and acceptable, why go through this.
	also - record when we change so worlds can tell if it is time to repaint"

	(depth == bitsPerPixel and: [aPoint = self extent and: 
					[self supportsDisplayDepth: bitsPerPixel]]) ifFalse: [
		bits := nil.  "Free up old bitmap in case space is low"
		DisplayChangeSignature := (DisplayChangeSignature ifNil: [0]) + 1.
		(self supportsDisplayDepth: bitsPerPixel)
			ifTrue:[super setExtent: aPoint depth: bitsPerPixel]
			ifFalse:[(self supportsDisplayDepth: bitsPerPixel negated)
				ifTrue:[super setExtent: aPoint depth: bitsPerPixel negated]
				ifFalse:["Search for a suitable depth"
					super setExtent: aPoint depth: self findAnyDisplayDepth]].
	].
	clippingBox := super boundingBox! !

!DisplayScreen methodsFor: 'initialize-release' stamp: 'ar 5/28/2000 11:25'!
shutDown 
	"Minimize Display memory saved in image"
	self setExtent: 240@120 depth: depth! !

!DisplayScreen methodsFor: 'other' stamp: 'ar 5/17/2001 21:02'!
supportedDisplayDepths
	"Return all pixel depths supported on the current host platform."
	^#(1 2 4 8 16 32 -1 -2 -4 -8 -16 -32) select: [:d | self supportsDisplayDepth: d]! !

!DisplayScreen methodsFor: 'other' stamp: 'ar 5/5/1999 23:45'!
supportsDisplayDepth: pixelDepth
	"Return true if this pixel depth is supported on the current host platform.
	Primitive. Optional."
	<primitive: 91>
	^#(1 2 4 8 16 32) includes: pixelDepth! !

!DisplayScreen methodsFor: 'other' stamp: ''!
usableArea
	"Answer the usable area of the receiver.  5/22/96 sw."

	^ self boundingBox deepCopy! !

!DisplayScreen methodsFor: 'other' stamp: ''!
width
	^ self boundingBox width! !

!Form class methodsFor: 'mode constants' stamp: ''!
and
	"Answer the integer denoting the logical 'and' combination rule."

	^1! !

!Form class methodsFor: 'mode constants' stamp: ''!
blend
	"Answer the integer denoting BitBlt's alpha blend combination rule."
	^24! !

!Form class methodsFor: 'mode constants' stamp: 'di 12/31/1998 14:02'!
blendAlpha
	"Answer the integer denoting BitBlt's blend-with-constant-alpha rule."

	^ 30! !

!Form class methodsFor: 'instance creation' stamp: 'ar 5/28/2000 12:07'!
dotOfSize: diameter
	"Create a form which contains a round black dot."
	| radius form bb rect centerX centerY centerYBias centerXBias radiusSquared xOverY maxy dx |
	radius := diameter//2.
	form := self extent: diameter@diameter offset: (0@0) - (radius@radius).	
	bb := (BitBlt current toForm: form)
		sourceX: 0; sourceY: 0;
		combinationRule: Form over;
		fillColor: Color black.
	rect := form boundingBox.
	centerX := rect center x.
	centerY := rect center y.
	centerYBias := rect height odd ifTrue: [0] ifFalse: [1].
	centerXBias := rect width odd ifTrue: [0] ifFalse: [1].
	radiusSquared := (rect height asFloat / 2.0) squared - 0.01.
	xOverY := rect width asFloat / rect height asFloat.
	maxy := rect height - 1 // 2.

	"First do the inner fill, and collect x values"
	0 to: maxy do:
		[:dy |
		dx := ((radiusSquared - (dy * dy) asFloat) sqrt * xOverY) truncated.
		bb	destX: centerX - centerXBias - dx
			destY: centerY - centerYBias - dy
			width: dx + dx + centerXBias + 1
			height: 1;
			copyBits.
		bb	destY: centerY + dy;
			copyBits].
	^ form
"
Time millisecondsToRun:
	[1 to: 20 do: [:i | (Form dotOfSize: i) displayAt: (i*20)@(i*20)]]
"! !

!Form class methodsFor: 'mode constants' stamp: ''!
erase
	"Answer the integer denoting mode erase."

	^4! !

!Form class methodsFor: 'mode constants' stamp: ''!
erase1bitShape
	"Answer the integer denoting mode erase."

	^ 26! !

!Form class methodsFor: 'examples' stamp: ''!
exampleBorder    "Form exampleBorder"
	"This example demonstrates the border finding algorithm. Start
	by having the user sketch on the screen (end with option-click) and then select a rectangular
	area of the screen which includes all of the area to be filled. Finally,
	(with crosshair cursor), the user points at the interior of the region to be
	outlined, and the region begins with that place as its seed."
	| f r interiorPoint |
	Form exampleSketch.		"sketch a little area with an enclosed region"
	r := Rectangle fromUser.
	f := Form fromDisplay: r.
	Cursor crossHair showWhile:
		[interiorPoint := Sensor waitButton - r origin].
	Cursor execute showWhile:
		[f shapeBorder: Color blue width: 2 interiorPoint: interiorPoint
			sharpCorners: false internal: false].
	f displayOn: Display at: r origin	! !

!Form class methodsFor: 'examples' stamp: ''!
exampleEdits
	"In Form category editing are messages edit and bitEdit that make it possible to 
	create editors on instances of Form. 
	 
	This is the general form editor:
	| f | 
	f := Form fromUser. 
	f edit. 
	 
	This is the general bit editor:
	| f | 
	f := Form fromUser. 
	f bitEdit."! !

!Form class methodsFor: 'examples' stamp: ''!
exampleMagnify

	| f m |
	f := Form fromUser.
	m := f magnify: f boundingBox by: 5 @ 5.
	m displayOn: Display at: Sensor waitButton

	"Form exampleMagnify."! !

!Form class methodsFor: 'examples' stamp: ''!
exampleShrink

	| f s |
	f := Form fromUser.
	s := f shrink: f boundingBox by: 2 @ 5.
	s displayOn: Display at: Sensor waitButton	

	"Form exampleShrink."! !

!Form class methodsFor: 'examples' stamp: ''!
exampleSketch
	"This is a simple drawing algorithm to get a sketch on the display screen.
	Draws whenever mouse button down.  Ends with option-click."
	| aPen color |
	aPen := Pen new.
	color := 0.
	[Sensor yellowButtonPressed]
		whileFalse:
		[aPen place: Sensor cursorPoint; color: (color := color + 1).
		[Sensor redButtonPressed]
			whileTrue: [aPen goto: Sensor cursorPoint]].
	Sensor waitNoButton.

	"Form exampleSketch"! !

!Form class methodsFor: 'examples' stamp: ''!
exampleSpaceFill    "Form exampleSpaceFill"
	"This example demonstrates the area filling algorithm. Starts by having
	the user sketch on the screen (ended by option-click) and then select a rectangular
	area of the screen which includes all of the area to be filled. Finally,
	(with crosshair cursor), the user points at the interior of some region to be
	filled, and the filling begins with that place as its seed."
	| f r interiorPoint |
	Form exampleSketch.		"sketch a little area with an enclosed region"
	r := Rectangle fromUser.
	f := Form fromDisplay: r.
	Cursor crossHair showWhile:
		[interiorPoint := Sensor waitButton - r origin].
	Cursor execute showWhile:
		[f shapeFill: Color gray interiorPoint: interiorPoint].
	f displayOn: Display at: r origin	! !

!Form class methodsFor: 'instance creation' stamp: 'jm 3/27/98 16:27'!
extent: extentPoint
	"Answer an instance of me with a blank bitmap of depth 1."

	^ self extent: extentPoint depth: 1
! !

!Form class methodsFor: 'instance creation' stamp: 'jm 3/27/98 16:36'!
extent: extentPoint depth: bitsPerPixel
	"Answer an instance of me with blank bitmap of the given dimensions and depth."

	^ self basicNew setExtent: extentPoint depth: bitsPerPixel
! !

!Form class methodsFor: 'instance creation' stamp: 'ar 10/9/1998 23:44'!
extent: extentPoint depth: bitsPerPixel bits: aBitmap
	"Answer an instance of me with blank bitmap of the given dimensions and depth."

	^ self basicNew setExtent: extentPoint depth: bitsPerPixel bits: aBitmap! !

!Form class methodsFor: 'instance creation' stamp: 'jm 3/27/98 16:35'!
extent: extentPoint depth: bitsPerPixel fromArray: anArray offset: offsetPoint 
	"Answer an instance of me with a pixmap of the given depth initialized from anArray."

	^ (self extent: extentPoint depth: bitsPerPixel)
		offset: offsetPoint;
		initFromArray: anArray
! !

!Form class methodsFor: 'instance creation' stamp: 'jm 3/27/98 16:33'!
extent: extentPoint fromArray: anArray offset: offsetPoint 
	"Answer an instance of me of depth 1 with bitmap initialized from anArray."

	^ (self extent: extentPoint depth: 1)
		offset: offsetPoint;
		initFromArray: anArray
! !

!Form class methodsFor: 'instance creation' stamp: 'jm 3/27/98 16:28'!
extent: extentPoint fromStipple: fourNibbles
	"Answer an instance of me with bitmap initialized from
	a repeating 4x4 bit stipple encoded in a 16-bit constant."
	| nibble |
	^ (self extent: extentPoint depth: 1)
		initFromArray: ((1 to: 4) collect:
				[:i | nibble := (fourNibbles bitShift: -4*(4-i)) bitAnd: 16rF.
				16r11111111 * nibble])  "fill 32 bits with each 4-bit nibble"
! !

!Form class methodsFor: 'instance creation' stamp: 'jm 3/27/98 16:26'!
extent: extentPoint offset: offsetPoint 
	"Answer an instance of me with a blank bitmap of depth 1."

	^ (self extent: extentPoint depth: 1) offset: offsetPoint
! !

!Form class methodsFor: 'file list services' stamp: 'nk 6/12/2004 12:56'!
fileReaderServicesForDirectory: aFileDirectory
	^{
		self serviceImageImportDirectory.
		self serviceImageImportDirectoryWithSubdirectories.
	}! !

!Form class methodsFor: 'file list services' stamp: 'nk 7/16/2003 18:01'!
fileReaderServicesForFile: fullName suffix: suffix

	^((ImageReadWriter allTypicalFileExtensions add: '*'; add: 'form'; yourself)
		includes: suffix)
		ifTrue: [ self services ]
		ifFalse: [#()]
! !

!Form class methodsFor: 'BMP file reading' stamp: 'ar 6/16/2002 17:41'!
fromBMPFile: aBinaryStream
	"Obsolete"
	^self fromBinaryStream: aBinaryStream.! !

!Form class methodsFor: 'BMP file reading' stamp: 'ar 6/16/2002 17:41'!
fromBMPFileNamed: fileName
	"Obsolete"
	^self fromFileNamed: fileName
! !

!Form class methodsFor: 'instance creation' stamp: 'nk 7/7/2003 18:19'!
fromBinaryStream: aBinaryStream
	"Read a Form or ColorForm from given file, using the first byte of the file to guess its format. Currently handles: GIF, uncompressed BMP, and both old and new DisplayObject writeOn: formats, JPEG, and PCX. Return nil if the file could not be read or was of an unrecognized format."

	| firstByte |
	aBinaryStream binary.
	firstByte := aBinaryStream next.
	firstByte = 1 ifTrue: [
		"old Squeakform format"
		^ self new readFromOldFormat: aBinaryStream].
	firstByte = 2 ifTrue: [
		"new Squeak form format"
		^ self new readFrom: aBinaryStream].

	"Try for JPG, GIF, or PCX..."
	"Note: The following call closes the stream."
	^ ImageReadWriter formFromStream: aBinaryStream
! !

!Form class methodsFor: 'instance creation' stamp: ''!
fromDisplay: aRectangle 
	"Answer an instance of me with bitmap initialized from the area of the 
	display screen defined by aRectangle."

	^ (self extent: aRectangle extent depth: Display depth)
		fromDisplay: aRectangle! !

!Form class methodsFor: 'instance creation' stamp: ''!
fromDisplay: aRectangle using: oldForm
	"Like fromDisplay: only if oldForm is the right size, copy into it and answer it instead."

	((oldForm ~~ nil) and: [oldForm extent = aRectangle extent])
		ifTrue:
			[oldForm fromDisplay: aRectangle.
			 ^ oldForm]
		ifFalse:
			[^ self fromDisplay: aRectangle]! !

!Form class methodsFor: 'instance creation' stamp: 'mir 11/19/2001 14:13'!
fromFileNamed: fileName
	"Read a Form or ColorForm from the given file."

	| file form |
	file := (FileStream readOnlyFileNamed: fileName) binary.
	form := self fromBinaryStream: file.
	Smalltalk isMorphic ifTrue:[
		Project current resourceManager
			addResource: form
			url: (FileDirectory urlForFileNamed: file name) asString].
	file close.
	^ form
! !

!Form class methodsFor: 'instance creation' stamp: ''!
fromUser
	"Answer an instance of me with bitmap initialized from the area of the 
	display screen designated by the user. The grid for selecting an area is 
	1@1."

	^self fromUser: 1 @ 1! !

!Form class methodsFor: 'instance creation' stamp: 'ar 3/1/2006 22:50'!
fromUser: gridPoint
	"Answer an instance of me with bitmap initialized from the area of the 
	display screen designated by the user. The grid for selecting an area is 
	aPoint. Ensures that the returned form has positive extent."
	| rect |
	rect := Rectangle fromUser: gridPoint.
	^ self fromDisplay: (rect origin extent: (rect extent max: gridPoint))! !

!Form class methodsFor: 'instance creation' stamp: 'jm 12/5/97 19:32'!
fromUserWithExtent: anExtent
	"Answer an instance of me with bitmap initialized from the area of the 
	display screen whose origin is designated by the user and whose size is anExtent"

	^ self fromDisplay: (Rectangle originFromUser: anExtent)

"(Form fromUserWithExtent: 50@50) displayAt: 10@10"! !

!Form class methodsFor: 'fileIn/Out' stamp: 'nk 6/12/2004 12:47'!
importImage: fullName
	"Import the given image file and store the resulting Form in the default Imports.
	The image is named with the short filename up to the first period, possibly with additions from the directory path to make it unique."

	Imports default importImageFromFileNamed: fullName.
! !

!Form class methodsFor: 'fileIn/Out' stamp: 'nk 6/12/2004 13:08'!
importImageDirectory: dir
	"Import the given image file and store the resulting Form in the default Imports.
	The image is named with the short filename up to the first period, possibly with additions from the directory path to make it unique."

	Imports default importImageDirectory: dir
! !

!Form class methodsFor: 'fileIn/Out' stamp: 'nk 6/12/2004 12:55'!
importImageDirectoryWithSubdirectories: dir
	"Import the given image file and store the resulting Form in the default Imports.
	The image is named with the short filename up to the first period, possibly with additions from the directory path to make it unique."

	Imports default importImageDirectoryWithSubdirectories: dir
! !

!Form class methodsFor: 'initialize-release' stamp: 'hg 8/3/2000 16:25'!
initialize

	FileList registerFileReader: self! !

!Form class methodsFor: 'examples' stamp: ''!
makeStar  "See the similar example in OpaqueForm"
	| sampleForm pen |
	sampleForm := Form extent: 50@50.  "Make a form"
	pen := Pen newOnForm: sampleForm.
	pen place: 24@50; turn: 18.		"Draw a 5-pointed star on it."
	1 to: 5 do: [:i | pen go: 19; turn: 72; go: 19; turn: -144].
	^ sampleForm
"
Form makeStar follow: [Sensor cursorPoint]
				while: [Sensor noButtonPressed]
"! !

!Form class methodsFor: 'mode constants' stamp: ''!
oldErase1bitShape
	"Answer the integer denoting mode erase."

	^ 17! !

!Form class methodsFor: 'mode constants' stamp: ''!
oldPaint
	"Answer the integer denoting the 'paint' combination rule."

	^16! !

!Form class methodsFor: 'file list services' stamp: 'hg 8/3/2000 16:26'!
openAsBackground: fullName
	"Set an image as a background image.  Support Squeak's common file format 
	(GIF, JPG, PNG, 'Form stoteOn: (run coded)' and BMP)"

	(self fromFileNamed: fullName) setAsBackground! !

!Form class methodsFor: 'file list services' stamp: 'nk 1/6/2004 12:36'!
openImageInWindow: fullName
	"Handle five file formats: GIF, JPG, PNG, Form storeOn: (run coded), and BMP.
	Fail if file format is not recognized."

	| image myStream |

	myStream := (FileStream readOnlyFileNamed: fullName) binary.
	image := self fromBinaryStream: myStream.
	myStream close.

	Smalltalk isMorphic ifTrue:[
		Project current resourceManager 
			addResource: image 
			url: (FileDirectory urlForFileNamed: fullName) asString.
	].

	Smalltalk isMorphic
		ifTrue: [(World drawingClass withForm: image) openInWorld]
		ifFalse: [FormView open: image named: fullName]! !

!Form class methodsFor: 'mode constants' stamp: ''!
over
	"Answer the integer denoting mode over."

	^3! !

!Form class methodsFor: 'mode constants' stamp: ''!
paint
	"Answer the integer denoting the 'paint' combination rule."

	^25! !

!Form class methodsFor: 'mode constants' stamp: 'di 12/31/1998 14:02'!
paintAlpha
	"Answer the integer denoting BitBlt's paint-with-constant-alpha rule."

	^ 31! !

!Form class methodsFor: 'mode constants' stamp: ''!
reverse
	"Answer the integer denoting mode reverse."

	^6! !

!Form class methodsFor: 'mode constants' stamp: 'hg 1/29/2001 17:28'!
rgbMul
	"Answer the integer denoting 'Multiply each color component, 
	 their values regarded as fractions of 1' rule."

	^ 37! !

!Form class methodsFor: 'file list services' stamp: 'sw 2/17/2002 01:38'!
serviceImageAsBackground
	"Answer a service for setting the desktop background from a given graphical file's contents"

	^ SimpleServiceEntry 
		provider: self 
		label: 'use graphic as background'
		selector: #openAsBackground:
		description: 'use the graphic as the background for the desktop'
		buttonLabel: 'background'! !

!Form class methodsFor: 'file list services' stamp: 'nk 6/12/2004 13:16'!
serviceImageImportDirectory
	"Answer a service for reading a graphic into ImageImports"

	^(SimpleServiceEntry
			provider: self 
			label: 'import all images from this directory'
			selector: #importImageDirectory:
			description: 'Load all graphics found in this directory, adding them to the ImageImports repository.'
			buttonLabel: 'import dir')
			argumentGetter: [ :fileList | fileList directory ];
			yourself
! !

!Form class methodsFor: 'file list services' stamp: 'nk 6/12/2004 13:15'!
serviceImageImportDirectoryWithSubdirectories
	"Answer a service for reading all graphics from a directory and its subdirectories into ImageImports"

	^(SimpleServiceEntry
			provider: self 
			label: 'import all images from here and subdirectories'
			selector: #importImageDirectoryWithSubdirectories:
			description: 'Load all graphics found in this directory and its subdirectories, adding them to the ImageImports repository.'
			buttonLabel: 'import subdirs')
			argumentGetter: [ :fileList | fileList directory ];
			yourself
! !

!Form class methodsFor: 'file list services' stamp: 'sw 2/17/2002 01:39'!
serviceImageImports
	"Answer a service for reading a graphic into ImageImports"

	^	SimpleServiceEntry
			provider: self 
			label: 'read graphic into ImageImports'
			selector: #importImage:
			description: 'Load a graphic, placing it in the ImageImports repository.'
			buttonLabel: 'import'! !

!Form class methodsFor: 'file list services' stamp: 'sw 2/17/2002 00:31'!
serviceOpenImageInWindow
	"Answer a service for opening a graphic in a window"

	^ SimpleServiceEntry 
		provider: self 
		label: 'open graphic in a window'
		selector: #openImageInWindow:
		description: 'open a graphic file in a window'
		buttonLabel: 'open'! !

!Form class methodsFor: 'file list services' stamp: 'sd 2/1/2002 21:43'!
services

	^ Array 
		with: self serviceImageImports
		with: self serviceOpenImageInWindow
		with: self serviceImageAsBackground ! !

!Form class methodsFor: 'shut down' stamp: 'ar 5/28/2000 23:35'!
shutDown  "Form shutDown"
	"Compress all instances in the system.  Will decompress on demand..."
	Form allInstancesDo: [:f | f hibernate].
	ColorForm allInstancesDo: [:f | f hibernate].! !

!Form class methodsFor: 'examples' stamp: 'tk 7/4/2000 12:08'!
toothpaste: diam		"Display restoreAfter: [Form toothpaste: 30]"
	"Draws wormlike lines by laying down images of spheres.
	See Ken Knowlton, Computer Graphics, vol. 15 no. 4 p352.
	Draw with mouse button down; terminate by option-click."
	| facade ball filter point queue port color q colors colr colr2 |
	colors := Display depth = 1
		ifTrue: [Array with: Color black]
		ifFalse: [Color red wheel: 12].
	facade := Form extent: diam@diam offset: (diam//-2) asPoint.
	(Form dotOfSize: diam) displayOn: facade
			at: (diam//2) asPoint clippingBox: facade boundingBox
			rule: Form under fillColor: Color white.
	#(1 2 3) do:
		[:x |  "simulate facade by circles of gray"
		(Form dotOfSize: x*diam//5) displayOn: facade
			at: (diam*2//5) asPoint clippingBox: facade boundingBox
			rule: Form under
			fillColor: (Color perform: 
					(#(black gray lightGray) at: x)).
		"facade displayAt: 50*x@50"].
	ball := Form dotOfSize: diam.
	color := 8.
	[ true ] whileTrue:
		[port := BitBlt current toForm: Display.
		"Expand 1-bit forms to any pixel depth"
		port colorMap: (Bitmap with: 0 with: 16rFFFFFFFF).
		queue := OrderedCollection new: 32.
		16 timesRepeat: [queue addLast: -20@-20].
		Sensor waitButton.
		Sensor yellowButtonPressed ifTrue: [^ self].
		filter := Sensor cursorPoint.
		colr := colors atWrap: (color := color + 5).  "choose increment relatively prime to colors size"
		colr2 := colr alphaMixed: 0.3 with: Color white.
		[Sensor redButtonPressed or: [queue size > 0]] whileTrue:
			[filter := filter * 4 + Sensor cursorPoint // 5.
			point := Sensor redButtonPressed
				ifTrue: [filter] ifFalse: [-20@-20].
			port copyForm: ball to: point rule: Form paint fillColor: colr.
			(q := queue removeFirst) == nil ifTrue: [^ self].	"exit"
			Display depth = 1
				ifTrue: [port copyForm: facade to: q rule: Form erase]
				ifFalse: [port copyForm: facade to: q rule: Form paint fillColor: colr2].
			Sensor redButtonPressed ifTrue: [queue addLast: point]]].
! !

!Form class methodsFor: 'mode constants' stamp: ''!
under
	"Answer the integer denoting mode under."

	^7! !

!Form class methodsFor: 'class initialization' stamp: 'SD 11/15/2001 22:21'!
unload

	FileList unregisterFileReader: self ! !

!Form class methodsFor: 'examples' stamp: ''!
xorHack: size  "Display restoreAfter: [Form xorHack: 256]"
	"Draw a smiley face or stick figure, and end with option-click.
	Thereafter image gets 'processed' as long as you have button down.
	If you stop at just the right time, you'll see you figure upside down,
	and at the end of a full cycle, you'll see it perfectly restored.
	Dude -- this works in color too!!"
	| rect form i bb |
	rect := 5@5 extent: size@size.
	Display fillWhite: rect; border: (rect expandBy: 2) width: 2.
	Display border: (rect topRight - (0@2) extent: rect extent*2 + 4) width: 2.
	Form exampleSketch.
	form := Form fromDisplay: rect.
	bb := form boundingBox.
	i := 0.
	[Sensor yellowButtonPressed] whileFalse:
		[[Sensor redButtonPressed] whileTrue:
			[i := i + 1.
			(Array with: 0@1 with: 0@-1 with: 1@0 with: -1@0) do:
				[:d | form copyBits: bb from: form at: d
					clippingBox: bb rule: Form reverse fillColor: nil].
			form displayAt: rect topLeft.
			i+2\\size < 4 ifTrue: [(Delay forMilliseconds: 300) wait]].
		(form magnify: form boundingBox by: 2@2) displayAt: rect topRight + (2@0).
		Sensor waitButton].! !

!Form methodsFor: 'initialize-release' stamp: 'ar 5/17/2001 22:54'!
allocateForm: extentPoint
	"Allocate a new form which is similar to the receiver and can be used for accelerated blts"
	^Form extent: extentPoint depth: self nativeDepth! !

!Form methodsFor: 'filling' stamp: 'di 2/19/1999 07:07'!
anyShapeFill
	"Fill the interior of the outermost outlined region in the receiver, a 1-bit deep form.  Typically the resulting form is used with fillShape:fillColor: to paint a solid color.  See also convexShapeFill:"

	| shape |
	"Draw a seed line around the edge and fill inward from the outside."
	shape := self findShapeAroundSeedBlock: [:f | f borderWidth: 1].
	"Reverse so that this becomes solid in the middle"
	shape := shape reverse.
	"Finally erase any bits from the original so the fill is only elsewhere"
	shape copy: shape boundingBox from: self to: 0@0 rule: Form erase.
	^ shape! !

!Form methodsFor: 'converting' stamp: 'jm 11/12/97 19:28'!
as8BitColorForm
	"Simple conversion of zero pixels to transparent.  Force it to 8 bits."

	| f map |
	f := ColorForm extent: self extent depth: 8.
	self displayOn: f at: self offset negated.
	map := Color indexedColors copy.
	map at: 1 put: Color transparent.
	f colors: map.
	f offset: self offset.
	^ f
! !

!Form methodsFor: 'converting' stamp: 'RAA 8/14/2000 10:13'!
asCursorForm

	^ self as: StaticForm! !

!Form methodsFor: 'converting' stamp: 'ar 6/16/2002 17:44'!
asFormOfDepth: d
	| newForm |
	d = self depth ifTrue:[^self].
	newForm := Form extent: self extent depth: d.
	(BitBlt current toForm: newForm)
		colorMap: (self colormapIfNeededFor: newForm);
		copy: (self boundingBox)
		from: 0@0 in: self
		fillColor: nil rule: Form over.
	^newForm! !

!Form methodsFor: 'converting' stamp: 'ar 5/17/2001 15:39'!
asGrayScale
	"Assume the receiver is a grayscale image. Return a grayscale ColorForm computed by extracting the brightness levels of one color component. This technique allows a 32-bit Form to be converted to an 8-bit ColorForm to save space while retaining a full 255 levels of gray. (The usual colormapping technique quantizes to 8, 16, or 32 levels, which loses information.)"
	| f32 srcForm result map bb grays |
	self depth = 32 ifFalse: [
		f32 := Form extent: width@height depth: 32.
		self displayOn: f32.
		^ f32 asGrayScale].
	self unhibernate.
	srcForm := Form extent: (width * 4)@height depth: 8.
	srcForm bits: bits.
	result := ColorForm extent: width@height depth: 8.
	map := Bitmap new: 256.
	2 to: 256 do: [:i | map at: i put: i - 1].
	map at: 1 put: 1.  "map zero pixel values to near-black"
	bb := (BitBlt current toForm: result)
		sourceForm: srcForm;
		combinationRule: Form over;
		colorMap: map.
	0 to: width - 1 do: [:dstX |
		bb  sourceRect: (((dstX * 4) + 2)@0 extent: 1@height);
			destOrigin: dstX@0;
			copyBits].

	"final BitBlt to zero-out pixels that were truely transparent in the original"
	map := Bitmap new: 512.
	map at: 1 put: 16rFF.
	(BitBlt current toForm: result)
		sourceForm: self;
		sourceRect: self boundingBox;
		destOrigin: 0@0;
		combinationRule: Form erase;
		colorMap: map;
		copyBits.
	
	grays := (0 to: 255) collect: [:brightness | Color gray: brightness asFloat / 255.0].
	grays at: 1 put: Color transparent.
	result colors: grays.
	^ result
! !

!Form methodsFor: 'converting' stamp: 'ar 11/7/1999 20:29'!
asMorph
	^ImageMorph new image: self! !

!Form methodsFor: 'converting' stamp: 'ar 2/7/2004 18:16'!
asSourceForm
	^self! !

!Form methodsFor: 'color mapping' stamp: 'ar 5/15/2001 16:16'!
balancedPatternFor: aColor
	"Return the pixel word for representing the given color on the receiver"
	self hasNonStandardPalette
		ifTrue:[^self bitPatternFor: aColor]
		ifFalse:[^aColor balancedPatternForDepth: self depth]! !

!Form methodsFor: 'editing' stamp: 'bf 10/11/1999 15:38'!
bitEdit
	"Create and schedule a view located in an area designated by the user 
	that contains a view of the receiver magnified by 8@8 that can be 
	modified using the Bit Editor. It also contains a view of the original 
	form."

	Smalltalk isMorphic
		ifFalse: [BitEditor openOnForm: self]
		ifTrue: [self currentHand attachMorph: (FatBitsPaint new
			editForm: self;
			magnification: 8;
			brushColor: Color black;
			penSize: 1;
			yourself)].

	"Note that using direct messages to BitEditor, fixed locations and scales can be created.
	That is, also try:
		BitEditor openOnForm: self at: <some point>
		BitEditor openOnForm: self at: <some point> scale: <some point>"! !

!Form methodsFor: 'editing' stamp: ''!
bitEditAt: magnifiedFormLocation scale: scaleFactor 
	"Create and schedule a view whose top left corner is magnifiedLocation 
	and that contains a view of the receiver magnified by scaleFactor that 
	can be modified using the Bit Editor. It also contains a view of the 
	original form."

	BitEditor openOnForm: self at: magnifiedFormLocation scale: scaleFactor ! !

!Form methodsFor: 'color mapping' stamp: 'ar 5/15/2001 16:16'!
bitPatternFor: aColor
	"Return the pixel word for representing the given color on the receiver"
	aColor isColor ifFalse:[^aColor bitPatternForDepth: self depth].
	self hasNonStandardPalette
		ifTrue:[^Bitmap with: (self pixelWordFor: aColor)]
		ifFalse:[^aColor bitPatternForDepth: self depth]! !

!Form methodsFor: 'filling' stamp: ''!
bitPatternForDepth: suspectedDepth
	"Only called when a Form is being used as a fillColor.  Use a Pattern or InfiniteForm instead for this purpose.
	Interpret me as an array of (32/depth) Color pixelValues.  BitBlt aligns the first element of this array with the top scanline of the destinationForm, the second with the second, and so on, cycling through the color array as necessary. 6/18/96 tk"

	^ self! !

!Form methodsFor: 'accessing' stamp: ''!
bits
	"Answer the receiver's Bitmap containing its bits."

	^ bits! !

!Form methodsFor: 'accessing' stamp: ''!
bits: aBitmap 
	"Reset the Bitmap containing the receiver's bits."

	bits := aBitmap! !

!Form methodsFor: 'postscript generation' stamp: 'ar 5/17/2001 15:36'!
bitsPerComponent
	^self depth <= 8 ifTrue:[self depth] ifFalse:[8].
! !

!Form methodsFor: 'accessing' stamp: 'ar 5/17/2001 15:41'!
bitsSize
	| pixPerWord |
	depth == nil ifTrue: [depth := 1].
	pixPerWord := 32 // self depth.
	^ width + pixPerWord - 1 // pixPerWord * height! !

!Form methodsFor: 'copying' stamp: 'RAA 9/28/1999 11:20'!
blankCopyOf: aRectangle scaledBy: scale

        ^ self class extent: (aRectangle extent * scale) truncated depth: depth! !

!Form methodsFor: 'bordering' stamp: 'ar 5/28/2000 12:07'!
border: rect width: borderWidth rule: rule fillColor: fillColor
        "Paint a border whose rectangular area is defined by rect. The
width of the border of each side is borderWidth. Uses fillColor for drawing
the border."
        | blt |
        blt := (BitBlt current toForm: self) combinationRule: rule; fillColor: fillColor.
        blt sourceOrigin: 0@0.
        blt destOrigin: rect origin.
        blt width: rect width; height: borderWidth; copyBits.
        blt destY: rect corner y - borderWidth; copyBits.
        blt destY: rect origin y + borderWidth.
        blt height: rect height - borderWidth - borderWidth; width:
borderWidth; copyBits.
        blt destX: rect corner x - borderWidth; copyBits! !

!Form methodsFor: 'bordering' stamp: 'ar 5/17/2001 15:42'!
borderFormOfWidth: borderWidth sharpCorners: sharpen
	"Smear this form around and then subtract the original to produce
	an outline.  If sharpen is true, then cause right angles to be outlined
	by right angles (takes an additional diagonal smears ANDed with both
	horizontal and vertical smears)."
	| smearForm bigForm smearPort all cornerForm cornerPort nbrs |
	self depth > 1 ifTrue: [self halt]. "Only meaningful for B/W forms."
	bigForm := self deepCopy.
	all := bigForm boundingBox.
	smearForm := Form extent: self extent.
	smearPort := BitBlt current toForm: smearForm.
	sharpen ifTrue:
		[cornerForm := Form extent: self extent.
		cornerPort := BitBlt current toForm: cornerForm].
	nbrs := (0@0) fourNeighbors.
	1 to: borderWidth do:
		[:i |  "Iterate to get several layers of 'skin'"
		nbrs do:
			[:d |  "Smear the self in 4 directions to grow each layer of skin"
			smearPort copyForm: bigForm to: d rule: Form under].
		sharpen ifTrue:
			["Special treatment to smear sharp corners"
			nbrs with: ((2 to: 5) collect: [:i2 | nbrs atWrap: i2]) do:
				[:d1 :d2 |
				"Copy corner points diagonally"
				cornerPort copyForm: bigForm to: d1+d2 rule: Form over.
				"But only preserve if there were dots on either side"
				cornerPort copyForm: bigForm to: d1+d1+d2 rule: Form and.
				cornerPort copyForm: bigForm to: d1+d2+d2 rule: Form and.
				smearPort copyForm: cornerForm to: 0@0 rule: Form under].
			].
		bigForm copy: all from: 0@0 in: smearForm rule: Form over.
		].
	"Now erase the original shape to obtain the outline"
	bigForm copy: all from: 0@0 in: self rule: Form erase.
	^ bigForm! !

!Form methodsFor: 'bordering' stamp: ''!
borderWidth: anInteger 
	"Set the width of the border for the receiver to be anInteger and paint it 
	using black as the border color."

	self border: self boundingBox width: anInteger fillColor: Color black! !

!Form methodsFor: 'bordering' stamp: ''!
borderWidth: anInteger color: aMask
	"Set the width of the border for the receiver to be anInteger and paint it 
	using aMask as the border color."

	self border: self boundingBox width: anInteger fillColor: aMask! !

!Form methodsFor: 'bordering' stamp: ''!
borderWidth: anInteger fillColor: aMask
	"Set the width of the border for the receiver to be anInteger and paint it 
	using aMask as the border color."

	self border: self boundingBox width: anInteger fillColor: aMask! !

!Form methodsFor: 'display box access' stamp: ''!
boundingBox
	^ Rectangle origin: 0 @ 0
			corner: width @ height! !

!Form methodsFor: 'postscript generation' stamp: 'mpw 11/14/1999 22:22'!
bytesPerRow
	^ self numComponents * self paddedWidth * self bitsPerComponent / 8.! !

!Form methodsFor: 'accessing' stamp: 'tk 3/9/97'!
center
	"Note that offset is ignored here.  Are we really going to embrace offset?  "
	^ (width @ height) // 2! !

!Form methodsFor: 'analyzing' stamp: ''!
cgForPixelValue: pv orNot: not
	"Return the center of gravity for all pixels of value pv.
	Note:  If orNot is true, then produce the center of gravity for all pixels
	that are DIFFERENT from the supplied (background) value"
	| pixCount weighted xAndY |
	xAndY := (Array with: (self xTallyPixelValue: pv orNot: not)
					with: (self yTallyPixelValue: pv orNot: not)) collect:
		[:profile |	"For both x and y profiles..."
		pixCount := 0.  weighted := 0.
		profile doWithIndex:
			[:t :i | pixCount := pixCount + t.
			weighted := weighted + (t*i)].
		pixCount = 0  "Produce average of nPixels weighted by coordinate"
			ifTrue: [0.0]
			ifFalse: [weighted asFloat / pixCount asFloat - 1.0]].

	^ xAndY first @ xAndY last
"
| f cg |
[Sensor anyButtonPressed] whileFalse:
	[f := Form fromDisplay: (Sensor cursorPoint extent: 50@50).
	cg := f cgForPixelValue: (Color black pixelValueForDepth: f depth) orNot: false.
	f displayAt: 0@0.
	Display fill: (cg extent: 2@2) fillColor: Color red].
	ScheduledControllers restore
"! !

!Form methodsFor: 'pixel access' stamp: 'ar 5/17/2001 15:42'!
colorAt: aPoint
	"Return the color in the pixel at the given point.  "

	^ Color 
		colorFromPixelValue: (self pixelValueAt: aPoint)
		depth: self depth
! !

!Form methodsFor: 'pixel access' stamp: 'ar 5/14/2001 23:46'!
colorAt: aPoint put: aColor
	"Store a Color into the pixel at coordinate aPoint.  "

	self pixelValueAt: aPoint put: (self pixelValueFor: aColor).

"[Sensor anyButtonPressed] whileFalse:
	[Display colorAt: Sensor cursorPoint put: Color red]"
! !

!Form methodsFor: 'converting' stamp: 'jm 4/5/1999 19:20'!
colorReduced
	"Return a color-reduced ColorForm version of the receiver, if possible, or the receiver itself if not."

	| tally tallyDepth colorCount newForm cm oldPixelValues newFormColors nextColorIndex c |
	tally := self tallyPixelValues asArray.
	tallyDepth := (tally size log: 2) asInteger.
	colorCount := 0.
	tally do: [:n | n > 0 ifTrue: [colorCount := colorCount + 1]].
	(tally at: 1) = 0 ifTrue: [colorCount := colorCount + 1].  "include transparent"
	colorCount > 256 ifTrue: [^ self].  "cannot reduce"
	newForm := self formForColorCount: colorCount.

	"build an array of just the colors used, and a color map to translate
	 old pixel values to their indices into this color array"
	cm := Bitmap new: tally size.
	oldPixelValues := self colormapIfNeededForDepth: 32.
	newFormColors := Array new: colorCount.
	newFormColors at: 1 put: Color transparent.
	nextColorIndex := 2.
	2 to: cm size do: [:i |
		(tally at: i) > 0 ifTrue: [
			oldPixelValues = nil
				ifTrue: [c := Color colorFromPixelValue: i - 1 depth: tallyDepth]
				ifFalse: [c := Color colorFromPixelValue: (oldPixelValues at: i) depth: 32].
			newFormColors at: nextColorIndex put: c.
			cm at: i put: nextColorIndex - 1.  "pixel values are zero-based indices"
			nextColorIndex := nextColorIndex + 1]].

	"copy pixels into new ColorForm, mapping to new pixel values"
	newForm copyBits: self boundingBox
		from: self
		at: 0@0
		clippingBox: self boundingBox
		rule: Form over
		fillColor: nil
		map: cm.
	newForm colors: newFormColors.
	newForm offset: offset.
	^ newForm
! !

!Form methodsFor: 'color mapping' stamp: 'ar 5/15/2001 16:16'!
colormapFromARGB
	"Return a ColorMap mapping from canonical ARGB space into the receiver.
	Note: This version is optimized for Squeak forms."
	| map nBits |
	self hasNonStandardPalette 
		ifTrue:[^ColorMap mappingFromARGB: self rgbaBitMasks].
	self depth <= 8 ifTrue:[
		map := Color colorMapIfNeededFrom: 32 to: self depth.
		map size = 512 ifTrue:[nBits := 3].
		map size = 4096 ifTrue:[nBits := 4].
		map size = 32768 ifTrue:[nBits := 5].
		^ColorMap
			shifts: (Array 
						with: 3 * nBits - 24
						with: 2 * nBits - 16
						with: 1 * nBits - 8
						with: 0)
			masks: (Array
						with: (1 << nBits) - 1 << (24 - nBits)
						with: (1 << nBits) - 1 << (16 - nBits)
						with: (1 << nBits) - 1 << (8 - nBits)
						with: 0)
			colors: map].
	self depth = 16 ifTrue:[
		^ColorMap
			shifts: #(-9 -6 -3 0)
			masks: #(16rF80000 16rF800 16rF8 0)].
	self depth = 32 ifTrue:[
		^ColorMap
			shifts: #(0 0 0 0)
			masks: #(16rFF0000 16rFF00 16rFF 16rFF000000)].
	self error:'Bad depth'! !

!Form methodsFor: 'color mapping' stamp: 'ar 5/16/2001 22:23'!
colormapIfNeededFor: destForm
	"Return a ColorMap mapping from the receiver to destForm."
	(self hasNonStandardPalette or:[destForm hasNonStandardPalette]) 
		ifTrue:[^self colormapFromARGB mappingTo: destForm colormapFromARGB]
		ifFalse:[^self colormapIfNeededForDepth: destForm depth]! !

!Form methodsFor: 'color mapping' stamp: 'ar 5/17/2001 15:42'!
colormapIfNeededForDepth: destDepth
	"Return a colormap for displaying the receiver at the given depth, or nil if no colormap is needed."

	self depth = destDepth ifTrue: [^ nil].  "not needed if depths are the same"
	^ Color colorMapIfNeededFrom: self depth to: destDepth
! !

!Form methodsFor: 'color mapping' stamp: 'ar 5/15/2001 16:16'!
colormapToARGB
	"Return a ColorMap mapping from the receiver into canonical ARGB space."
	self hasNonStandardPalette 
		ifTrue:[^self colormapFromARGB inverseMap].
	self depth <= 8 ifTrue:[
		^ColorMap
			shifts: #(0 0 0 0)
			masks: #(16rFF0000 16rFF00 16rFF 16rFF000000)
			colors: (Color colorMapIfNeededFrom: self depth to: 32)].
	self depth = 16 ifTrue:[
		^ColorMap 
			shifts: #( 9 6 3 0) 
			masks: #(16r7C00 16r3E0 16r1F 0)].
	self depth = 32 ifTrue:[
		^ColorMap
			shifts: #(0 0 0 0) 
			masks: #(16rFF0000 16rFF00 16rFF 16rFF000000)].
	self error:'Bad depth'! !

!Form methodsFor: 'analyzing' stamp: 'jm 12/5/97 19:48'!
colorsUsed
	"Return a list of the Colors this form uses."

	| tallies tallyDepth usedColors |
	tallies := self tallyPixelValues.
	tallyDepth := (tallies size log: 2) asInteger.
	usedColors := OrderedCollection new.
	tallies doWithIndex: [:count :i |
		count > 0 ifTrue: [
			usedColors add: (Color colorFromPixelValue: i - 1 depth: tallyDepth)]].
	^ usedColors asArray
! !

!Form methodsFor: 'fileIn/Out' stamp: 'ar 2/24/2001 22:41'!
comeFullyUpOnReload: smartRefStream
	bits isForm ifFalse:[^self].
	"make sure the resource gets loaded afterwards"
	ResourceCollector current ifNil:[^self].
	ResourceCollector current noteResource: bits replacing: self.
! !

!Form methodsFor: 'display box access' stamp: ''!
computeBoundingBox
	^ Rectangle origin: 0 @ 0
			corner: width @ height! !

!Form methodsFor: 'copying' stamp: 'ar 6/9/2000 18:59'!
contentsOfArea: aRect
 	"Return a new form which derives from the portion of the original form delineated by aRect."
	^self contentsOfArea: aRect 
		into: (self class extent: aRect extent depth: depth).! !

!Form methodsFor: 'copying' stamp: 'ar 6/9/2000 19:00'!
contentsOfArea: aRect into: newForm
 	"Return a new form which derives from the portion of the original form delineated by aRect."
	^ newForm copyBits: aRect from: self at: 0@0
		clippingBox: newForm boundingBox rule: Form over fillColor: nil! !

!Form methodsFor: 'filling' stamp: 'di 9/11/1998 16:25'!
convexShapeFill: aMask 
	"Fill the interior of the outtermost outlined region in the receiver.  The outlined region must not be concave by more than 90 degrees.  Typically aMask is Color black, to produce a solid fill. then the resulting form is used with fillShape: to paint a solid color.  See also anyShapeFill"
	| destForm tempForm |
	destForm := Form extent: self extent.  destForm fillBlack.
	tempForm := Form extent: self extent.
	(0@0) fourNeighbors do:
		[:dir |  "Smear self in all 4 directions, and AND the result"
		self displayOn: tempForm at: (0@0) - self offset.
		tempForm smear: dir distance: (dir dotProduct: tempForm extent) abs.
		tempForm displayOn: destForm at: 0@0
			clippingBox: destForm boundingBox
			rule: Form and fillColor: nil].
	destForm displayOn: self at: 0@0
		clippingBox: self boundingBox
		rule: Form over fillColor: aMask! !

!Form methodsFor: 'copying' stamp: ''!
copy: aRect
 	"Return a new form which derives from the portion of the original form delineated by aRect."
	| newForm |
	newForm := self class extent: aRect extent depth: depth.
	^ newForm copyBits: aRect from: self at: 0@0
		clippingBox: newForm boundingBox rule: Form over fillColor: nil! !

!Form methodsFor: 'copying' stamp: 'ar 5/28/2000 12:08'!
copy: destRectangle from: sourcePt in: sourceForm rule: rule 
	"Make up a BitBlt table and copy the bits."
	(BitBlt current toForm: self)
		copy: destRectangle
		from: sourcePt in: sourceForm
		fillColor: nil rule: rule! !

!Form methodsFor: 'copying' stamp: ''!
copy: sourceRectangle from: sourceForm to: destPt rule: rule
	^ self copy: (destPt extent: sourceRectangle extent)
		from: sourceRectangle topLeft in: sourceForm rule: rule! !

!Form methodsFor: 'copying' stamp: 'ar 5/28/2000 12:08'!
copyBits: sourceForm at: destOrigin translucent: factor
	"Make up a BitBlt table and copy the bits with the given colorMap."
	(BitBlt current 
		destForm: self
		sourceForm: sourceForm
		halftoneForm: nil
		combinationRule: 30
		destOrigin: destOrigin
		sourceOrigin: 0@0
		extent: sourceForm extent
		clipRect: self boundingBox)
		copyBitsTranslucent: ((0 max: (factor*255.0) asInteger) min: 255)
"
 | f f2 f3 | f := Form fromUser. f2 := Form fromDisplay: (0@0 extent: f extent). f3 := f2 deepCopy.
0.0 to: 1.0 by: 1.0/32 do:
	[:t | f3 := f2 deepCopy. f3 copyBits: f at: 0@0 translucent: t.
	f3 displayAt: 0@0. (Delay forMilliseconds: 100) wait].
"! !

!Form methodsFor: 'copying' stamp: 'ar 5/28/2000 12:08'!
copyBits: sourceRect from: sourceForm at: destOrigin clippingBox: clipRect rule: rule fillColor: aForm 
	"Make up a BitBlt table and copy the bits."

	(BitBlt current 
		destForm: self
		sourceForm: sourceForm
		fillColor: aForm
		combinationRule: rule
		destOrigin: destOrigin
		sourceOrigin: sourceRect origin
		extent: sourceRect extent
		clipRect: clipRect) copyBits! !

!Form methodsFor: 'copying' stamp: 'ar 5/28/2000 12:08'!
copyBits: sourceRect from: sourceForm at: destOrigin clippingBox: clipRect rule: rule fillColor: aForm map: map
	"Make up a BitBlt table and copy the bits.  Use a colorMap."

	((BitBlt current 
		destForm: self
		sourceForm: sourceForm
		fillColor: aForm
		combinationRule: rule
		destOrigin: destOrigin
		sourceOrigin: sourceRect origin
		extent: sourceRect extent
		clipRect: clipRect) colorMap: map) copyBits! !

!Form methodsFor: 'copying' stamp: 'ar 5/28/2000 12:08'!
copyBits: sourceRect from: sourceForm at: destOrigin colorMap: map 
	"Make up a BitBlt table and copy the bits with the given colorMap."
	((BitBlt current 
		destForm: self
		sourceForm: sourceForm
		halftoneForm: nil
		combinationRule: Form over
		destOrigin: destOrigin
		sourceOrigin: sourceRect origin
		extent: sourceRect extent
		clipRect: self boundingBox) colorMap: map) copyBits! !

!Form methodsFor: 'converting' stamp: 'di 10/16/2001 19:23'!
copyWithColorsReducedTo: nColors
	"Note: this has not been engineered.
	There are better solutions in the literature."
	| palette colorMap pc closest |
	palette := self reducedPaletteOfSize: nColors.
	colorMap := (1 to: (1 bitShift: depth)) collect:
		[:i | pc := Color colorFromPixelValue: i-1 depth: depth.
		closest := palette detectMin: [:c | c diff: pc].
		closest pixelValueForDepth: depth].
	^ self deepCopy copyBits: self boundingBox from: self at: 0@0 colorMap: (colorMap as: Bitmap)
		! !

!Form methodsFor: 'postscript generation' stamp: 'ar 5/17/2001 15:39'!
decodeArray
	^self depth <= 8 ifTrue:['[1 0]'] ifFalse:['[0 1 0 1 0 1 ]'].
! !

!Form methodsFor: 'copying' stamp: 'jm 2/27/98 09:35'!
deepCopy

	^ self shallowCopy
		bits: bits copy;
		offset: offset copy
! !

!Form methodsFor: 'accessing' stamp: 'ar 5/28/2000 12:03'!
defaultCanvasClass
	"Return the default canvas used for drawing onto the receiver"
	^Display defaultCanvasClass! !

!Form methodsFor: 'accessing' stamp: 'ar 5/17/2001 15:45'!
depth
	^ depth < 0 ifTrue:[0-depth] ifFalse:[depth]! !

!Form methodsFor: 'accessing' stamp: ''!
depth: bitsPerPixel
	(bitsPerPixel > 32 or:
		[(bitsPerPixel bitAnd: bitsPerPixel-1) ~= 0])
		ifTrue: [self halt: 'bitsPerPixel must be 1, 2, 4, 8, 16 or 32'].
	depth := bitsPerPixel! !

!Form methodsFor: 'displaying' stamp: 'ar 2/13/2001 22:13'!
displayInterpolatedIn: aRectangle on: aForm
	"Display the receiver on aForm, using interpolation if necessary.
		Form fromUser displayInterpolatedOn: Display.
	Note: When scaling we attempt to use bilinear interpolation based
	on the 3D engine. If the engine is not there then we use WarpBlt.
	"
	| engine adjustedR |
	self extent = aRectangle extent ifTrue:[^self displayOn: aForm at: aRectangle origin].
	Smalltalk at: #B3DRenderEngine 
		ifPresent:[:engineClass| engine := (engineClass defaultForPlatformOn: aForm)].
	engine ifNil:[
		"We've got no bilinear interpolation. Use WarpBlt instead"
		(WarpBlt current toForm: aForm)
			sourceForm: self destRect: aRectangle;
			combinationRule: 3;
			cellSize: 2;
			warpBits.
		^self
	].

	"Otherwise use the 3D engine for our purposes"

	"there seems to be a slight bug in B3D which the following adjusts for"
	adjustedR := (aRectangle withRight: aRectangle right + 1) translateBy: 0@1.
	engine viewport: adjustedR.
	engine material: ((Smalltalk at: #B3DMaterial) new emission: Color white).
	engine texture: self.
	engine render: ((Smalltalk at: #B3DIndexedQuadMesh) new plainTextureRect).
	engine finish.! !

!Form methodsFor: 'displaying' stamp: 'ar 2/13/2001 22:12'!
displayInterpolatedOn: aForm
	"Display the receiver on aForm, using interpolation if necessary.
		Form fromUser displayInterpolatedOn: Display.
	Note: When scaling we attempt to use bilinear interpolation based
	on the 3D engine. If the engine is not there then we use WarpBlt.
	"
	| engine |
	self extent = aForm extent ifTrue:[^self displayOn: aForm].
	Smalltalk at: #B3DRenderEngine 
		ifPresent:[:engineClass| engine := (engineClass defaultForPlatformOn: aForm)].
	engine ifNil:[
		"We've got no bilinear interpolation. Use WarpBlt instead"
		(WarpBlt current toForm: aForm)
			sourceForm: self destRect: aForm boundingBox;
			combinationRule: 3;
			cellSize: 2;
			warpBits.
		^self
	].
	"Otherwise use the 3D engine for our purposes"
	engine viewport: aForm boundingBox.
	engine material: ((Smalltalk at: #B3DMaterial) new emission: Color white).
	engine texture: self.
	engine render: ((Smalltalk at: #B3DIndexedQuadMesh) new plainTextureRect).
	engine finish.! !

!Form methodsFor: 'displaying' stamp: 'ar 5/14/2001 23:33'!
displayOn: aDisplayMedium at: aDisplayPoint clippingBox: clipRectangle rule: rule fillColor: aForm

	aDisplayMedium copyBits: self boundingBox
		from: self
		at: aDisplayPoint + self offset
		clippingBox: clipRectangle
		rule: rule
		fillColor: aForm
		map: (self colormapIfNeededFor: aDisplayMedium).
! !

!Form methodsFor: 'displaying' stamp: ''!
displayOn: aDisplayMedium transformation: displayTransformation clippingBox: clipRectangle align: alignmentPoint with: relativePoint rule: ruleInteger fillColor: aForm 
	"Graphically, it means nothing to scale a Form by floating point values.  
	Because scales and other display parameters are kept in floating point to 
	minimize round off errors, we are forced in this routine to round off to the 
	nearest integer."

	| absolutePoint scale magnifiedForm |
	absolutePoint := displayTransformation applyTo: relativePoint.
	absolutePoint := absolutePoint x asInteger @ absolutePoint y asInteger.
	displayTransformation noScale
		ifTrue: [magnifiedForm := self]
		ifFalse: 
			[scale := displayTransformation scale.
			scale := scale x @ scale y.
			(1@1 = scale)
					ifTrue: [scale := nil. magnifiedForm := self]
					ifFalse: [magnifiedForm := self magnify: self boundingBox by: scale]].
	magnifiedForm
		displayOn: aDisplayMedium
		at: absolutePoint - alignmentPoint
		clippingBox: clipRectangle
		rule: ruleInteger
		fillColor: aForm! !

!Form methodsFor: 'displaying' stamp: ''!
displayOnPort: port at: location
	port copyForm: self to: location rule: Form over! !

!Form methodsFor: 'displaying' stamp: 'ar 5/17/2001 15:40'!
displayResourceFormOn: aForm
	"a special display method for blowing up resource thumbnails"
	| engine tx cmap blitter |
	self extent = aForm extent ifTrue:[^self displayOn: aForm].
	Smalltalk at: #B3DRenderEngine ifPresentAndInMemory:
		[:engineClass | engine := engineClass defaultForPlatformOn: aForm].
	engine ifNil:[
		"We've got no bilinear interpolation. Use WarpBlt instead"
		(WarpBlt current toForm: aForm)
			sourceForm: self destRect: aForm boundingBox;
			combinationRule: 3;
			cellSize: 2;
			warpBits.
		^self
	].
	tx := self asTexture.
	(blitter := BitBlt current toForm: tx)
		sourceForm: self; destRect: aForm boundingBox;
		sourceOrigin: 0@0;
		combinationRule: Form paint.
	"map transparency to current World background color"
	(World color respondsTo: #pixelWordForDepth:) ifTrue: [
		cmap := Bitmap new: (self depth <= 8 ifTrue: [1 << self depth] ifFalse: [4096]).
		cmap at: 1 put: (tx pixelWordFor: World color).
		blitter colorMap: cmap.
	].
	blitter copyBits.
	engine viewport: aForm boundingBox.
	engine material: ((Smalltalk at: #B3DMaterial) new emission: Color white).
	engine texture: tx.
	engine render: ((Smalltalk at: #B3DIndexedQuadMesh) new plainTextureRect).
	engine finish.
	"the above, using bilinear interpolation doesn't leave transparent pixel values intact"
	(WarpBlt current toForm: aForm)
		sourceForm: self destRect: aForm boundingBox;
		combinationRule: Form and;
		colorMap: (Color maskingMap: self depth);
		warpBits.! !

!Form methodsFor: 'displaying' stamp: 'ar 3/2/2001 21:32'!
displayScaledOn: aForm
	"Display the receiver on aForm, scaling if necessary.
		Form fromUser displayScaledOn: Display.
	"
	self extent = aForm extent ifTrue:[^self displayOn: aForm].
	(WarpBlt current toForm: aForm)
		sourceForm: self destRect: aForm boundingBox;
		combinationRule: Form paint;
		cellSize: 2;
		warpBits.! !

!Form methodsFor: 'accessing' stamp: 'ar 5/27/2000 16:56'!
displayScreen
	"Return the display screen the receiver is allocated on. 
	Forms in general are Squeak internal and not allocated on any particular display."
	^nil! !

!Form methodsFor: 'analyzing' stamp: 'ar 5/17/2001 15:40'!
dominantColor
	| tally max maxi |
	self depth > 16 ifTrue:
		[^(self asFormOfDepth: 16) dominantColor].
	tally := self tallyPixelValues.
	max := maxi := 0.
	tally withIndexDo: [:n :i | n > max ifTrue: [max := n. maxi := i]].
	^ Color colorFromPixelValue: maxi - 1 depth: self depth! !

!Form methodsFor: 'displaying' stamp: 'ar 5/28/2000 12:08'!
drawLine: sourceForm from: beginPoint to: endPoint clippingBox: clipRect rule: anInteger fillColor: aForm 
	"Refer to the comment in 
	DisplayMedium|drawLine:from:to:clippingBox:rule:mask:." 
	
	| dotSetter |
	"set up an instance of BitBlt for display"
	dotSetter := BitBlt current
		destForm: self
		sourceForm: sourceForm
		fillColor: aForm
		combinationRule: anInteger
		destOrigin: beginPoint
		sourceOrigin: 0 @ 0
		extent: sourceForm extent
		clipRect: clipRect.
	dotSetter drawFrom: beginPoint to: endPoint! !

!Form methodsFor: 'editing' stamp: ''!
edit
	"Start up an instance of the FormEditor on this form. Typically the form 
	is not visible on the screen. The editor menu is located at the bottom of 
	the form editing frame. The form is displayed centered in the frame. 
	YellowButtonMenu accept is used to modify the form to reflect the 
	changes made on the screen version; cancel restores the original form to 
	the screen. Note that the changes are clipped to the original size of the 
	form."
 
	FormEditor openOnForm: self! !

!Form methodsFor: 'filling' stamp: 'di 10/17/2001 10:09'!
eraseShape: bwForm
	"use bwForm as a mask to clear all pixels where bwForm has 1's"
	((BitBlt current destForm: self sourceForm: bwForm 
		fillColor: nil
		combinationRule: Form erase1bitShape	"Cut a hole in the picture with my mask"
		destOrigin: bwForm offset 
		sourceOrigin: 0@0
		extent: self extent clipRect: self boundingBox)
		colorMap: (Bitmap with: 0 with: 16rFFFFFFFF))
		copyBits.
! !

!Form methodsFor: 'accessing' stamp: ''!
extent
	^ width @ height! !

!Form methodsFor: 'transitions' stamp: 'jm 5/21/1998 23:46'!
fadeImage: otherImage at: topLeft
	indexAndMaskDo: indexAndMaskBlock
	"This fade uses halftones as a blending hack.
	Zeros in the halftone produce the original image (self), and 
	ones in the halftone produce the 'otherImage'.
	IndexAndMaskBlock gets evaluated prior to each cycle,
	and the resulting boolean determines whether to continue cycling."
	| index imageRect maskForm resultForm |
	imageRect := otherImage boundingBox.
	resultForm := self copy: (topLeft extent: imageRect extent).
	maskForm := Form extent: 32@32.
	index := 0.
	[indexAndMaskBlock value: (index := index+1) value: maskForm]
	whileTrue:
		[maskForm reverse.
		resultForm copyBits: imageRect from: resultForm at: 0@0
			clippingBox: imageRect rule: Form over fillColor: maskForm.
		maskForm reverse.
		resultForm copyBits: imageRect from: otherImage at: 0@0
			clippingBox: imageRect rule: Form under fillColor: maskForm.
		self copyBits: imageRect from: resultForm at: topLeft
				clippingBox: self boundingBox rule: Form over fillColor: nil.
		Display forceDisplayUpdate]! !

!Form methodsFor: 'transitions' stamp: 'ar 5/17/2001 15:42'!
fadeImageCoarse: otherImage at: topLeft
	"Display fadeImageCoarse: (Form fromDisplay: (40@40 extent: 300@300)) reverse at: 40@40"
	| pix j d |
	d := self depth.
	^ self fadeImage: otherImage at: topLeft indexAndMaskDo:
		[:i :mask |
		i=1 ifTrue: [pix := (1 bitShift: d) - 1.
					1 to: 8//d-1 do: [:q | pix := pix bitOr: (pix bitShift: d*4)]].
		i <= 16 ifTrue:
		[j := i-1//4+1.
		(0 to: 28 by: 4) do: [:k |
			mask bits at: j+k
				put: ((mask bits at: j+k) bitOr: (pix bitShift: i-1\\4*d))].
		"mask display." true]
		ifFalse: [false]]! !

!Form methodsFor: 'transitions' stamp: 'ar 5/17/2001 15:41'!
fadeImageFine: otherImage at: topLeft
	"Display fadeImageFine: (Form fromDisplay: (40@40 extent: 300@300)) reverse at: 40@40"
	| pix j ii d |
	d := self depth.
	^ self fadeImage: otherImage at: topLeft indexAndMaskDo:
		[:i :mask |
		i=1 ifTrue: [pix := (1 bitShift: d) - 1.
					1 to: 8//d-1 do:
						[:q | pix := pix bitOr: (pix bitShift: d*4)]].
		i <= 16 ifTrue:
		[ii := #(0 10 2 8 7 13 5 15 1 11 3 9 6 12 4 14) at: i.
		j := ii//4+1.
		(0 to: 28 by: 4) do:
			[:k | mask bits at: j+k put:
				((mask bits at: j+k) bitOr: (pix bitShift: ii\\4*d))].
		true]
		ifFalse: [false]]! !

!Form methodsFor: 'transitions' stamp: ''!
fadeImageHor: otherImage at: topLeft
	"Display fadeImageHor: (Form fromDisplay: (10@10 extent: 300@300)) reverse at: 10@10"
	^ self fadeImage: otherImage at: topLeft indexAndMaskDo:
		[:i :mask |
		mask fill: (0@(mask height//2-i) extent: mask width@(i*2)) fillColor: Color black.
		(i*2) <= mask width]! !

!Form methodsFor: 'transitions' stamp: ''!
fadeImageHorFine: otherImage at: topLeft
	"Display fadeImageHorFine: (Form fromDisplay: (10@10 extent: 300@300)) reverse at: 10@10"
	^ self fadeImage: otherImage at: topLeft indexAndMaskDo:
		[:i :mask |
		mask fill: (0@(i-1) extent: mask width@1) fillColor: Color black.
		mask fill: (0@(i-1+16) extent: mask width@1) fillColor: Color black.
		(i*2) <= mask width]! !

!Form methodsFor: 'transitions' stamp: ''!
fadeImageSquares: otherImage at: topLeft 
	"Display fadeImageSquares: (Form fromDisplay: (40@40 extent: 300@300)) reverse at: 40@40"
	^ self fadeImage: otherImage at: topLeft indexAndMaskDo:
		[:i :mask |
		mask fill: ((16-i) asPoint extent: (i*2) asPoint) fillColor: Color black.
		i <= 16]! !

!Form methodsFor: 'transitions' stamp: 'ar 5/17/2001 15:39'!
fadeImageVert: otherImage at: topLeft
	"Display fadeImageVert: (Form fromDisplay: (10@10 extent: 300@300)) reverse at: 10@10"
	| d |
	d := self depth.
	^ self fadeImage: otherImage at: topLeft indexAndMaskDo:
		[:i :mask |
		mask fill: ((mask width//2//d-i*d)@0 extent: i*2*d@mask height) fillColor: Color black.
		i <= (mask width//d)]! !

!Form methodsFor: 'filling' stamp: 'ar 5/28/2000 12:08'!
fill: aRectangle rule: anInteger fillColor: aForm 
	"Replace a rectangular area of the receiver with the pattern described by aForm 
	according to the rule anInteger."
	(BitBlt current toForm: self)
		copy: aRectangle
		from: 0@0 in: nil
		fillColor: aForm rule: anInteger! !

!Form methodsFor: 'filling' stamp: ''!
fillFromXColorBlock: colorBlock
	"Horizontal Gradient Fill.
	Supply relative x in [0.0 ... 1.0] to colorBlock,
	and paint each pixel with the color that comes back"
	| xRel |
	0 to: width-1 do:
		[:x |  xRel := x asFloat / (width-1) asFloat.
		self fill: (x@0 extent: 1@height) 
			fillColor: (colorBlock value: xRel)]
"
((Form extent: 100@100 depth: Display depth)
	fillFromXColorBlock: [:x | Color r: x g: 0.0 b: 0.5]) display
"! !

!Form methodsFor: 'filling' stamp: 'ar 5/17/2001 15:38'!
fillFromXYColorBlock: colorBlock
	"General Gradient Fill.
	Supply relative x and y in [0.0 ... 1.0] to colorBlock,
	and paint each pixel with the color that comes back"
	| poker yRel xRel |
	poker := BitBlt current bitPokerToForm: self.
	0 to: height-1 do:
		[:y | yRel := y asFloat / (height-1) asFloat.
		0 to: width-1 do:
			[:x |  xRel := x asFloat / (width-1) asFloat.
			poker pixelAt: x@y
				put: ((colorBlock value: xRel value: yRel) pixelWordForDepth: self depth)]]
"
 | d |
((Form extent: 100@20 depth: Display depth)
	fillFromXYColorBlock:
	[:x :y | d := 1.0 - (x - 0.5) abs - (y - 0.5) abs.
	Color r: d g: 0 b: 1.0-d]) display
"! !

!Form methodsFor: 'filling' stamp: ''!
fillFromYColorBlock: colorBlock
	"Vertical Gradient Fill.
	Supply relative y in [0.0 ... 1.0] to colorBlock,
	and paint each pixel with the color that comes back"
	| yRel |
	0 to: height-1 do:
		[:y |  yRel := y asFloat / (height-1) asFloat.
		self fill: (0@y extent: width@1) 
			fillColor: (colorBlock value: yRel)]
"
((Form extent: 100@100 depth: Display depth)
	fillFromYColorBlock: [:y | Color r: y g: 0.0 b: 0.5]) display
"! !

!Form methodsFor: 'filling' stamp: 'ar 5/17/2001 15:38'!
findShapeAroundSeedBlock: seedBlock
	"Build a shape that is black in any region marked by seedBlock. 
	SeedBlock will be supplied a form, in which to blacken various
	pixels as 'seeds'.  Then the seeds are smeared until 
	there is no change in the smear when it fills the region, ie,
	when smearing hits a black border and thus goes no further."
	| smearForm previousSmear all count smearPort |
	self depth > 1 ifTrue: [self halt]. "Only meaningful for B/W forms."
	all := self boundingBox.
	smearForm := Form extent: self extent.
	smearPort := BitBlt current toForm: smearForm.
	seedBlock value: smearForm.		"Blacken seeds to be smeared"
	smearPort copyForm: self to: 0@0 rule: Form erase.  "Clear any in black"
	previousSmear := smearForm deepCopy.
	count := 1.
	[count = 10 and:   "check for no change every 10 smears"
		[count := 1.
		previousSmear copy: all from: 0@0 in: smearForm rule: Form reverse.
		previousSmear isAllWhite]]
		whileFalse: 
			[smearPort copyForm: smearForm to: 1@0 rule: Form under.
			smearPort copyForm: smearForm to: -1@0 rule: Form under.
			"After horiz smear, trim around the region border"
			smearPort copyForm: self to: 0@0 rule: Form erase.
			smearPort copyForm: smearForm to: 0@1 rule: Form under.
			smearPort copyForm: smearForm to: 0@-1 rule: Form under.
			"After vert smear, trim around the region border"
			smearPort copyForm: self to: 0@0 rule: Form erase.
			count := count+1.
			count = 9 ifTrue: "Save penultimate smear for comparison"
				[previousSmear copy: all from: 0@0 in: smearForm rule: Form over]].
	"Now paint the filled region in me with aHalftone"
	^ smearForm! !

!Form methodsFor: 'initialize-release' stamp: 'ar 5/26/2000 00:46'!
finish
	"If there are any pending operations on the receiver complete them. Do not return before all modifications have taken effect."! !

!Form methodsFor: 'other' stamp: 'ar 12/12/2003 18:24'!
fixAlpha
	"Fix the alpha channel if the receiver is 32bit"
	| bb |
	self depth = 32 ifFalse:[^self].
	bb := BitBlt toForm: self.
	bb combinationRule: 40 "fixAlpha:with:".
	bb copyBits.! !

!Form methodsFor: 'scaling, rotation' stamp: 'ar 5/14/2001 23:33'!
flipBy: direction centerAt: aPoint
	"Return a copy of the receiver flipped either #vertical or #horizontal."
	| newForm quad |
	newForm := self class extent: self extent depth: depth.
	quad := self boundingBox innerCorners.
	quad := (direction = #vertical ifTrue: [#(2 1 4 3)] ifFalse: [#(4 3 2 1)])
		collect: [:i | quad at: i].
	(WarpBlt current toForm: newForm)
		sourceForm: self;
		colorMap: (self colormapIfNeededFor: newForm);
		combinationRule: 3;
		copyQuad: quad toRect: newForm boundingBox.
	newForm offset: (self offset flipBy: direction centerAt: aPoint).
	^ newForm
"
[Sensor anyButtonPressed] whileFalse:
	[((Form fromDisplay: (Sensor cursorPoint extent: 130@66))
			flipBy: #vertical centerAt: 0@0) display]
"
"Consistency test...
 | f f2 p | [Sensor anyButtonPressed] whileFalse:
	[f := Form fromDisplay: ((p := Sensor cursorPoint) extent: 31@41).
	Display fillBlack: (p extent: 31@41).
	f2 := f flipBy: #vertical centerAt: 0@0.
	(f2 flipBy: #vertical centerAt: 0@0) displayAt: p]
"
! !

!Form methodsFor: 'filling' stamp: 'ar 5/14/2001 23:46'!
floodFill2: aColor at: interiorPoint
	"Fill the shape (4-connected) at interiorPoint.  The algorithm is based on Paul Heckbert's 'A Seed Fill Algorithm', Graphic Gems I, Academic Press, 1990.
	NOTE: This is a less optimized variant for flood filling which is precisely along the lines of Heckbert's algorithm. For almost all cases #floodFill:at: will be faster (see the comment there) but this method is left in both as reference and as a fallback if such a strange case is encountered in reality."
	| peeker poker stack old new x y top x1 x2 dy left goRight |
	peeker := BitBlt current bitPeekerFromForm: self.
	poker := BitBlt current bitPokerToForm: self.
	stack := OrderedCollection new: 50.
	"read old pixel value"
	old := peeker pixelAt: interiorPoint.
	"compute new value"
	new := self pixelValueFor: aColor.
	old = new ifTrue:[^self]. "no point, is there?!!"

	x := interiorPoint x.
	y := interiorPoint y.
	(y >= 0 and:[y < height]) ifTrue:[
		stack addLast: {y. x. x. 1}. "y, left, right, dy"
		stack addLast: {y+1. x. x. -1}].
	[stack isEmpty] whileFalse:[
		top := stack removeLast.
		y := top at: 1. x1 := top at: 2. x2 := top at: 3. dy := top at: 4.
		y := y + dy.
		"Segment of scanline (y-dy) for x1 <= x <= x2 was previously filled.
		Now explore adjacent pixels in scanline y."
		x := x1.
		[x >= 0 and:[(peeker pixelAt: x@y) = old]] whileTrue:[
			poker pixelAt: x@y put: new.
			x := x - 1].
		goRight := x < x1.
		left := x+1.
		(left < x1 and:[y-dy >= 0 and:[y-dy < height]]) 
			ifTrue:[stack addLast: {y. left. x1-1. 0-dy}].
		goRight ifTrue:[x := x1 + 1].
		[
			goRight ifTrue:[
				[x < width and:[(peeker pixelAt: x@y) = old]] whileTrue:[
					poker pixelAt: x@y put: new.
					x := x + 1].
				(y+dy >= 0 and:[y+dy < height]) 
					ifTrue:[stack addLast: {y. left. x-1. dy}].
				(x > (x2+1) and:[y-dy >= 0 and:[y-dy >= 0]]) 
					ifTrue:[stack addLast: {y. x2+1. x-1. 0-dy}]].
			[(x := x + 1) <= x2 and:[(peeker pixelAt: x@y) ~= old]] whileTrue.
			left := x.
			goRight := true.
		x <= x2] whileTrue.
	].
! !

!Form methodsFor: 'filling' stamp: 'di 10/20/2001 22:03'!
floodFill: aColor at: interiorPoint
	Preferences areaFillsAreVeryTolerant ifTrue:
		[^ self floodFill: aColor at: interiorPoint tolerance: 0.2].
	Preferences areaFillsAreTolerant ifTrue:
		[^ self floodFill: aColor at: interiorPoint tolerance: 0.1].
	^ self floodFill: aColor at: interiorPoint tolerance: 0
! !

!Form methodsFor: 'filling' stamp: 'ar 8/6/2009 23:12'!
floodFill: aColor at: interiorPoint tolerance: tolerance
	"Fill the shape (4-connected) at interiorPoint.  The algorithm is based on Paul Heckbert's 'A Seed Fill Algorithm', Graphic Gems I, Academic Press, 1990.
	NOTE (ar): This variant has been heavily optimized to prevent the overhead of repeated calls to BitBlt. Usually this is a really big winner but the runtime now depends a bit on the complexity of the shape to be filled. For extremely complex shapes (say, a Hilbert curve) with very few pixels to fill it can be slower than #floodFill2:at: since it needs to repeatedly read the source bits. However, in all practical cases I found this variant to be 15-20 times faster than anything else.
	Further note (di):  I have added a feature that allows this routine to fill areas of approximately constant color (such as  photos, scans, and jpegs).  It does this by computing a color map for the peeker that maps all colors close to 'old' into colors identical to old.  This mild colorblindness achieves the desired effect with no further change or degradation of the algorithm.  tolerance should be 0 (exact match), or a value corresponding to those returned by Color>>diff:, with 0.1 being a reasonable starting choice."

	| peeker poker stack old new x y top x1 x2 dy left goRight span spanBits w box debug |
	debug := false. "set it to true to see the filling process"
	box := interiorPoint extent: 1@1.
	span := Form extent: width@1 depth: 32.
	spanBits := span bits.

	peeker := BitBlt current toForm: span.
	peeker 
		sourceForm: self; 
		combinationRule: 3; 
		width: width; 
		height: 1.

	"read old pixel value"
	peeker sourceOrigin: interiorPoint; destOrigin: interiorPoint x @ 0; width: 1; copyBits.
	old := spanBits at: interiorPoint x + 1.

	"compute new value (take care since the algorithm will fail if old = new)"
	new := self privateFloodFillValue: aColor.
	old = new ifTrue: [^ box].
	tolerance > 0 ifTrue:
		["Set up color map for approximate fills"
		peeker colorMap: (self floodFillMapFrom: self to: span mappingColorsWithin: tolerance to: old)].

	poker := BitBlt current toForm: self.
	poker 
		fillColor: aColor;
		combinationRule: 3;
		width: width;
		height: 1.

	stack := OrderedCollection new: 50.
	x := interiorPoint x.
	y := interiorPoint y.
	(y >= 0 and:[y < height]) ifTrue:[
		stack addLast: {y. x. x. 1}. "y, left, right, dy"
		stack addLast: {y+1. x. x. -1}].

	[stack isEmpty] whileFalse:[
		debug ifTrue:[self displayOn: Display].
		top := stack removeLast.
		y := top at: 1. x1 := top at: 2. x2 := top at: 3. dy := top at: 4.
		y := y + dy.
		debug ifTrue:[
			Display 
				drawLine: (Form extent: 1@1 depth: 8) fillWhite
				from: (x1-1)@y to: (x2+1)@y 
				clippingBox: Display boundingBox
				rule: Form over fillColor: nil].
		"Segment of scanline (y-dy) for x1 <= x <= x2 was previously filled.
		Now explore adjacent pixels in scanline y."
		peeker sourceOrigin: 0@y; destOrigin: 0@0; width: width; copyBits.
			"Note: above is necessary since we don't know where we'll end up filling"
		x := x1.
		w := 0.
		[x >= 0 and:[(spanBits at: x+1) = old]] whileTrue:[
			w := w + 1.
			x := x - 1].
		w > 0 ifTrue:[
			"overwrite pixels"
			poker destOrigin: x+1@y; width: w; copyBits.
			box := box quickMerge: ((x+1@y) extent: w@1)].
		goRight := x < x1.
		left := x+1.
		(left < x1 and:[y-dy >= 0 and:[y-dy < height]]) 
			ifTrue:[stack addLast: {y. left. x1-1. 0-dy}].
		goRight ifTrue:[x := x1 + 1].
		[
			goRight ifTrue:[
				w := 0.
				[x < width and:[(spanBits at: x+1) = old]] whileTrue:[
					w := w + 1.
					x := x + 1].
				w > 0 ifTrue:[
					"overwrite pixels"
					poker destOrigin: (x-w)@y; width: w; copyBits.
					box := box quickMerge: ((x-w@y) extent: w@1)].
				(y+dy >= 0 and:[y+dy < height]) 
					ifTrue:[stack addLast: {y. left. x-1. dy}].
				(x > (x2+1) and:[y-dy >= 0 and:[y-dy >= 0]]) 
					ifTrue:[stack addLast: {y. x2+1. x-1. 0-dy}]].
			[(x := x + 1) <= x2 and:[(spanBits at: x+1) ~= old]] whileTrue.
			left := x.
			goRight := true.
		x <= x2] whileTrue.
	].
	^box! !

!Form methodsFor: 'filling' stamp: 'di 10/20/2001 10:09'!
floodFillMapFrom: sourceForm to: scanlineForm mappingColorsWithin: dist to: centerPixVal
	"This is a helper routine for floodFill.  It's written for clarity (scanning the entire
	map using colors) rather than speed (which would require hacking rgb components
	in the nieghborhood of centerPixVal.  Note that some day a better proximity metric
	would be (h s v) where tolerance could be reduced in hue."

	| colorMap centerColor |
	scanlineForm depth = 32 ifFalse: [self error: 'depth 32 assumed'].
	"First get a modifiable identity map"
	colorMap := 	(Color cachedColormapFrom: sourceForm depth to: scanlineForm depth) copy.
	centerColor := Color colorFromPixelValue: (centerPixVal bitOr: 16rFFe6) depth: scanlineForm depth.
	"Now replace all entries that are close to the centerColor"
	1 to: colorMap size do:
		[:i | ((Color colorFromPixelValue: ((colorMap at: i) bitOr: 16rFFe6) depth: scanlineForm depth)
				diff: centerColor) <= dist ifTrue: [colorMap at: i put: centerPixVal]].
	^ colorMap! !

!Form methodsFor: 'initialize-release' stamp: 'ar 5/26/2000 00:45'!
flush
	"If there are any pending operations on the receiver start doing them. In time, they will show up on the receiver but not necessarily immediately after this method returns."! !

!Form methodsFor: 'accessing' stamp: ''!
form
	"Answer the receiver's form.  For vanilla Forms, this degenerates to self.  Makes several methods that operate on both Forms and MaskedForms much more straightforward.   6/1/96 sw"

	^ self! !

!Form methodsFor: 'other' stamp: 'jm 9/27/97 21:02'!
formForColorCount: colorCount
	"Return a ColorForm of sufficient depth to represent the given number of colors. The maximum number of colors is 256."

	colorCount > 256 ifTrue: [^ self error: 'too many colors'].

	colorCount > 16 ifTrue: [^ ColorForm extent: self extent depth: 8].
	colorCount > 4 ifTrue: [^ ColorForm extent: self extent depth: 4].
	colorCount > 2 ifTrue: [^ ColorForm extent: self extent depth: 2].
	^ ColorForm extent: self extent depth: 1
! !

!Form methodsFor: 'initialize-release' stamp: ''!
fromDisplay: aRectangle 
	"Create a virtual bit map from a user specified rectangular area on the 
	display screen. Reallocates bitmap only if aRectangle ~= the receiver's 
	extent."

	(width = aRectangle width and: [height = aRectangle height])
		ifFalse: [self setExtent: aRectangle extent depth: depth].
	self
		copyBits: (aRectangle origin extent: self extent)
		from: Display
		at: 0 @ 0
		clippingBox: self boundingBox
		rule: Form over
		fillColor: nil! !

!Form methodsFor: 'accessing' stamp: 'ar 5/28/2000 12:03'!
getCanvas
	"Return a Canvas that can be used to draw onto the receiver"
	^self defaultCanvasClass on: self! !

!Form methodsFor: 'private' stamp: 'tk 3/13/2000 15:21'!
hackBits: bitThing
	"This method provides an initialization so that BitBlt may be used, eg, to 
	copy ByteArrays and other non-pointer objects efficiently.
	The resulting form looks 4 wide, 8 deep, and bitThing-size-in-words high."
	width := 4.
	depth := 8.
	bitThing class isBits ifFalse: [self error: 'bitThing must be a non-pointer object'].
	bitThing class isBytes
		ifTrue: [height := bitThing basicSize // 4]
		ifFalse: [height := bitThing basicSize].
	bits := bitThing! !

!Form methodsFor: 'accessing' stamp: 'ar 5/28/2000 00:48'!
hasBeenModified
	"Return true if something *might* have been drawn into the receiver"
	^(bits == nil or:[bits class == ByteArray]) not
	"Read the above as: If the receiver has forgotten its contents (bits == nil) 
	or is still hibernated it can't be modified."! !

!Form methodsFor: 'accessing' stamp: 'ar 5/28/2000 00:48'!
hasBeenModified: aBool
	"Change the receiver to reflect the modification state"
	aBool ifTrue:[^self unhibernate].
	self shouldPreserveContents
		ifTrue:[self hibernate]
		ifFalse:[bits := nil]! !

!Form methodsFor: 'testing' stamp: 'ar 5/15/2001 16:14'!
hasNonStandardPalette
	"Return true if the receiver has a non-standard palette.
	Non-standard means that RGBA components may be located
	at positions differing from the standard Squeak RGBA layout
	at the receiver's depth."
	^false! !

!Form methodsFor: 'accessing' stamp: ''!
height
	^ height! !

!Form methodsFor: 'fileIn/Out' stamp: 'di 8/5/1998 11:37'!
hibernate
	"Replace my bitmap with a compactly encoded representation (a ByteArray).  It is vital that BitBlt and any other access to the bitmap (such as writing to a file) not be used when in this state.  Since BitBlt will fail if the bitmap size is wrong (not = bitsSize), we do not allow replacement by a byteArray of the same (or larger) size."

	"NOTE: This method copies code from Bitmap compressToByteArray so that it can
	nil out the old bits during the copy, thus avoiding 2x need for extra storage."
	| compactBits lastByte |
	(bits isMemberOf: Bitmap) ifFalse: [^ self  "already hibernated or weird state"].
	compactBits := ByteArray new: (bits size*4) + 7 + (bits size//1984*3).
	lastByte := bits compress: bits toByteArray: compactBits.
	lastByte < (bits size*4) ifTrue:
		[bits := nil.  "Let GC reclaim the old bits before the copy if necessary"
		bits := compactBits copyFrom: 1 to: lastByte]! !

!Form methodsFor: 'private' stamp: ''!
initFromArray: array
	"Fill the bitmap from array.  If the array is shorter,
	then cycle around in its contents until the bitmap is filled."
	| ax aSize array32 i j word16 |
	ax := 0.
	aSize := array size.
	aSize > bits size ifTrue:
		["backward compatibility with old 16-bit bitmaps and their forms"
		array32 := Array new: height * (width + 31 // 32).
		i := j := 0.
		1 to: height do:
			[:y | 1 to: width+15//16 do:
				[:x16 | word16 := array at: (i := i + 1).
				x16 odd ifTrue: [array32 at: (j := j+1) put: (word16 bitShift: 16)]
						ifFalse: [array32 at: j put: ((array32 at: j) bitOr: word16)]]].
		^ self initFromArray: array32].
	1 to: bits size do:
		[:index |
		(ax := ax + 1) > aSize ifTrue: [ax := 1].
		bits at: index put: (array at: ax)]! !

!Form methodsFor: 'analyzing' stamp: ''!
innerPixelRectFor: pv orNot: not
	"Return a rectangle describing the smallest part of me that includes 
	all pixels of value pv.
	Note:  If orNot is true, then produce a copy that includes all pixels
	that are DIFFERENT from the supplied (background) value"

	| xTally yTally |
	xTally := self xTallyPixelValue: pv orNot: not.
	yTally := self yTallyPixelValue: pv orNot: not.
	^ ((xTally findFirst: [:t | t>0]) - 1) @ ((yTally findFirst: [:t | t>0]) - 1)
		corner:
			(xTally findLast: [:t | t>0])@(yTally findLast: [:t | t>0])! !

!Form methodsFor: 'testing' stamp: 'ar 7/21/2007 21:37'!
isAllWhite
	"Answer whether all bits in the receiver are white"
	| word |
	self unhibernate.
	word := Color white pixelWordForDepth: self depth.
	1 to: bits size do: [:i | (bits at: i) = word ifFalse: [^ false]].
	^ true! !

!Form methodsFor: 'testing' stamp: 'ar 5/17/2001 15:46'!
isBigEndian
	"Return true if the receiver contains big endian pixels, meaning the left-most pixel is stored in the most significant bits of a word."
	^depth > 0! !

!Form methodsFor: 'testing' stamp: 'ar 5/28/2000 14:58'!
isBltAccelerated: ruleInteger for: sourceForm
	"Return true if the receiver can perform accelerated blts operations by itself"
	^false! !

!Form methodsFor: 'testing' stamp: 'ar 5/28/2000 15:04'!
isDisplayScreen
	^false! !

!Form methodsFor: 'testing' stamp: 'ar 5/27/2000 16:54'!
isExternalForm
	^false! !

!Form methodsFor: 'testing' stamp: 'ar 5/28/2000 14:58'!
isFillAccelerated: ruleInteger for: aColor
	"Return true if the receiver can perform accelerated fill operations by itself"
	^false! !

!Form methodsFor: 'testing' stamp: 'ar 10/30/2000 23:23'!
isForm
	^true! !

!Form methodsFor: 'testing' stamp: 'ar 5/17/2001 15:47'!
isLittleEndian
	"Return true if the receiver contains little endian pixels, meaning the left-most pixel is stored in the least significant bits of a word."
	^depth < 0! !

!Form methodsFor: 'testing' stamp: 'RAA 8/14/2000 10:00'!
isStatic

	^false! !

!Form methodsFor: 'testing' stamp: 'ar 2/10/2004 17:18'!
isTranslucent
	"Answer whether this form may be translucent"
	^self depth = 32! !

!Form methodsFor: 'pixel access' stamp: 'ar 5/17/2001 15:39'!
isTransparentAt: aPoint 
	"Return true if the receiver is transparent at the given point."

	self depth = 1 ifTrue: [^ false].  "no transparency at depth 1"
	^ (self pixelValueAt: aPoint) = (self pixelValueFor: Color transparent)
! !

!Form methodsFor: 'scaling, rotation' stamp: ''!
magnify: aRectangle by: scale 
	"Answer a Form created as a scaling of the receiver.
	Scale may be a Float, and may be greater or less than 1.0."
	^ self magnify: aRectangle by: scale smoothing: 1

"Dynamic test...
[Sensor anyButtonPressed] whileFalse:
	[(Display magnify: (Sensor cursorPoint extent: 31@41) by: 5@3) display]
"
"Scaling test...
| f cp | f := Form fromDisplay: (Rectangle originFromUser: 100@100).
Display restoreAfter: [Sensor waitNoButton.
[Sensor anyButtonPressed] whileFalse:
	[cp := Sensor cursorPoint.
	(f magnify: f boundingBox by: (cp x asFloat@cp y asFloat)/f extent) display]]
"
"Consistency test...
 | f f2 p | [Sensor anyButtonPressed] whileFalse:
	[f := Form fromDisplay: ((p := Sensor cursorPoint) extent: 31@41).
	Display fillBlack: (p extent: 31@41).
	f2 := f magnify: f boundingBox by: 5@3.
	(f2 shrink: f2 boundingBox by: 5@3) displayAt: p]
"
! !

!Form methodsFor: 'scaling, rotation' stamp: 'tpr 9/28/2004 17:00'!
magnify: aRectangle by: scale smoothing: cellSize
        "Answer a Form created as a scaling of the receiver.
        Scale may be a Float or even a Point, and may be greater or less than 1.0."
        | newForm |
        newForm := self blankCopyOf: aRectangle scaledBy: scale.
        (WarpBlt current toForm: newForm)
                sourceForm: self;
                colorMap: (self colormapIfNeededFor: newForm);
                cellSize: cellSize;  "installs a new colormap if cellSize > 1"
                combinationRule: 3;
                copyQuad: aRectangle innerCorners toRect: newForm boundingBox.
        ^ newForm

"Dynamic test...
[Sensor anyButtonPressed] whileFalse:
        [(Display magnify: (Sensor cursorPoint extent: 131@81) by: 0.5 smoothing: 2) display]
"
"Scaling test...
| f cp | f := Form fromDisplay: (Rectangle originFromUser: 100@100).
Display restoreAfter: [Sensor waitNoButton.
[Sensor anyButtonPressed] whileFalse:
        [cp := Sensor cursorPoint.
        (f magnify: f boundingBox by: (cp x asFloat@cp y asFloat)/f extent smoothing: 2) display]]
"! !

!Form methodsFor: 'scaling, rotation' stamp: 'tpr 9/28/2004 17:00'!
magnifyBy: scale 
	"Answer a Form created as a scaling of the receiver.
	Scale may be a Float or even a Point, and may be greater or less than 1.0."

	^ self magnify: self boundingBox by: scale
			smoothing: (scale < 1 ifTrue: [2] ifFalse: [1])! !

!Form methodsFor: 'color mapping' stamp: ''!
makeBWForm: foregroundColor
	"Map this form into a B/W form with 1's in the foreground regions."
	| bwForm map |
	bwForm := Form extent: self extent.
	map := self newColorMap.  "All non-foreground go to 0's"
	map at: (foregroundColor indexInMap: map) put: 1.
	bwForm copyBits: self boundingBox from: self at: 0@0 colorMap: map.
	^ bwForm! !

!Form methodsFor: 'color mapping' stamp: 'ar 5/17/2001 15:38'!
mapColor: oldColor to: newColor
	"Make all pixels of the given color in this Form to the given new color."
	"Warnings: This method modifies the receiver. It may lose some color accuracy on 32-bit Forms, since the transformation uses a color map with only 15-bit resolution."

	| map |
	map := (Color cachedColormapFrom: self depth to: self depth) copy.
	map at: (oldColor indexInMap: map) put: (newColor pixelWordForDepth: self depth).
	(BitBlt current toForm: self)
		sourceForm: self;
		sourceOrigin: 0@0;
		combinationRule: Form over;
		destX: 0 destY: 0 width: width height: height;
		colorMap: map;
		copyBits.
! !

!Form methodsFor: 'color mapping' stamp: 'ar 5/17/2001 15:40'!
mapColors: oldColorBitsCollection to: newColorBits
	"Make all pixels of the given color in this Form to the given new color."
	"Warnings: This method modifies the receiver. It may lose some color accuracy on 32-bit Forms, since the transformation uses a color map with only 15-bit resolution."

	| map |
	self depth < 16
		ifTrue: [map := (Color cachedColormapFrom: self depth to: self depth) copy]
		ifFalse: [
			"use maximum resolution color map"
			"source is 16-bit or 32-bit RGB; use colormap with 5 bits per color component"
			map := Color computeRGBColormapFor: self depth bitsPerColor: 5].
	oldColorBitsCollection do:[ :oldColor | map at: oldColor put: newColorBits].

	(BitBlt current toForm: self)
		sourceForm: self;
		sourceOrigin: 0@0;
		combinationRule: Form over;
		destX: 0 destY: 0 width: width height: height;
		colorMap: map;
		copyBits.
! !

!Form methodsFor: 'color mapping' stamp: 'ar 12/14/2001 18:11'!
maskingMap
	"Return a color map that maps all colors except transparent to words of all ones. Used to create a mask for a Form whose transparent pixel value is zero."
	^Color maskingMap: self depth! !

!Form methodsFor: 'editing' stamp: 'RAA 9/28/1999 09:11'!
morphEdit

        ^ FatBitsPaint new openWith: self! !

!Form methodsFor: 'accessing' stamp: 'ar 5/17/2001 15:50'!
nativeDepth
	"Return the 'native' depth of the receiver, e.g., including the endianess"
	^depth! !

!Form methodsFor: 'color mapping' stamp: 'ar 5/17/2001 15:41'!
newColorMap 
	"Return an uninitialized color map array appropriate to this Form's depth."

	^ Bitmap new: (1 bitShift: (self depth min: 15))
! !

!Form methodsFor: 'postscript generation' stamp: 'ar 5/17/2001 15:43'!
numComponents
	^self depth <= 8 ifTrue:[1] ifFalse:[3].
! !

!Form methodsFor: 'fileIn/Out' stamp: 'ar 3/3/2001 16:16'!
objectForDataStream: refStream
	| prj repl |
	prj := refStream project.
	prj ifNil:[^super objectForDataStream: refStream].
	ResourceCollector current ifNil:[^super objectForDataStream: refStream].
	repl := ResourceCollector current objectForDataStream: refStream fromForm: self.
	^repl! !

!Form methodsFor: 'accessing' stamp: 'ar 2/16/2000 22:00'!
offset
	^offset ifNil:[0@0]! !

!Form methodsFor: 'accessing' stamp: ''!
offset: aPoint

	offset := aPoint! !

!Form methodsFor: 'converting' stamp: 'ar 7/23/1999 17:04'!
orderedDither32To16
	"Do an ordered dithering for converting from 32 to 16 bit depth."
	| ditherMatrix ii out inBits outBits index pv dmv r di dmi dmo g b pvOut outIndex |
	self depth = 32 ifFalse:[^self error:'Must be 32bit for this'].
	ditherMatrix := #(	0	8	2	10
						12	4	14	6
						3	11	1	9
						15	7	13	5).
	ii := (0 to: 31) collect:[:i| i].
	out := Form extent: self extent depth: 16.
	inBits := self bits.
	outBits := out bits.
	index := outIndex := 0.
	pvOut := 0.
	0 to: self height-1 do:[:y|
		0 to: self width-1 do:[:x|
			pv := inBits at: (index := index + 1).
			dmv := ditherMatrix at: (y bitAnd: 3) * 4 + (x bitAnd: 3) + 1.
			r := pv bitAnd: 255.	di := r * 496 bitShift: -8.
			dmi := di bitAnd: 15.	dmo := di bitShift: -4.
			r := dmv < dmi ifTrue:[ii at: 2+dmo] ifFalse:[ii at: 1+dmo].
			g := (pv bitShift: -8) bitAnd: 255.	di := g * 496 bitShift: -8.
			dmi := di bitAnd: 15.	dmo := di bitShift: -4.
			g := dmv < dmi ifTrue:[ii at: 2+dmo] ifFalse:[ii at: 1+dmo].
			b := (pv bitShift: -16) bitAnd: 255.	di := b * 496 bitShift: -8.
			dmi := di bitAnd: 15.	dmo := di bitShift: -4.
			b := dmv < dmi ifTrue:[ii at: 2+dmo] ifFalse:[ii at: 1+dmo].
			pvOut := (pvOut bitShift: 16) + 
						(b bitShift: 10) + (g bitShift: 5) + r.
			(x bitAnd: 1) = 1 ifTrue:[
				outBits at: (outIndex := outIndex+1) put: pvOut.
				pvOut := 0].
		].
		(self width bitAnd: 1) = 1 ifTrue:[
			outBits at: (outIndex := outIndex+1) put: (pvOut bitShift: -16).
			pvOut := 0].
	].
	^out! !

!Form methodsFor: 'postscript generation' stamp: ''!
paddedWidth
	^ (self width + (self rowPadding-1)// self rowPadding) * self rowPadding.! !

!Form methodsFor: 'transitions' stamp: 'jm 6/1/1998 10:55'!
pageImage: otherImage at: topLeft corner: corner
	"Produce a page-turning illusion that gradually reveals otherImage
	located at topLeft in this form.  Corner specifies which corner, as
		1=topLeft, 2=topRight, 3=bottomRight, 4=bottomLeft."
	| bb maskForm resultForm delta maskLoc maskRect stepSize cornerSel smallRect |
	stepSize := 10.
	bb := otherImage boundingBox.
	resultForm := self copy: (topLeft extent: bb extent).
	maskForm := Form extent: ((otherImage width min: otherImage height) + stepSize) asPoint.

	"maskLoc := starting loc rel to topLeft"
	otherImage width > otherImage height
		ifTrue: ["wide image; motion is horizontal."
				(corner between: 2 and: 3) not ifTrue:
					["motion is to the right"
					delta := 1@0.
					maskLoc := bb topLeft - (corner = 1
						ifTrue: [maskForm width@0]
						ifFalse: [maskForm width@stepSize])]
					ifFalse:
					["motion is to the left"
					delta := -1@0.
					maskLoc := bb topRight - (corner = 2
						ifTrue: [0@0]
						ifFalse: [0@stepSize])]]
		ifFalse: ["tall image; motion is vertical."
				corner <= 2 ifTrue:
					["motion is downward"
					delta := 0@1.
					maskLoc := bb topLeft - (corner = 1
						ifTrue: [0@maskForm height]
						ifFalse: [stepSize@maskForm height])]
					ifFalse:
					["motion is upward"
					delta := 0@-1.
					maskLoc := bb bottomLeft - (corner = 3
						ifTrue: [stepSize@0]
						ifFalse: [0@0])]].

	"Build a solid triangle in the mask form"
	(Pen newOnForm: maskForm) in: [:p |
		corner even  "Draw 45-degree line"
			ifTrue: [p place: 0@0; turn: 135; go: maskForm width*3//2]
			ifFalse: [p place: 0@(maskForm height-1); turn: 45; go: maskForm width*3//2]].
	maskForm smear: delta negated distance: maskForm width.
	"Copy the mask to full resolution for speed.  Make it be the reversed
	so that it can be used for ORing in the page-corner color"
	maskForm := (Form extent: maskForm extent depth: otherImage depth)
		copyBits: maskForm boundingBox from: maskForm at: 0@0
		colorMap: (Bitmap with: 16rFFFFFFFF with: 0).

	"Now move the triangle maskForm across the resultForm selecting the
	triangular part of otherImage to display, and across the resultForm,
	selecting the part of the original image to erase."
	cornerSel := #(topLeft topRight bottomRight bottomLeft) at: corner.
	1 to: (otherImage width + otherImage height // stepSize)+1 do:
		[:i |		"Determine the affected square"
		maskRect := (maskLoc extent: maskForm extent) intersect: bb.
		((maskLoc x*delta x) + (maskLoc y*delta y)) < 0 ifTrue:
			[smallRect := 0@0 extent: (maskRect width min: maskRect height) asPoint.
			maskRect := smallRect align: (smallRect perform: cornerSel)
								with: (maskRect perform: cornerSel)].

		"AND otherForm with triangle mask, and OR into result"
		resultForm copyBits: bb from: otherImage at: 0@0
				clippingBox: maskRect rule: Form over fillColor: nil.
		resultForm copyBits: maskForm boundingBox from: maskForm at: maskLoc
				clippingBox: maskRect rule: Form erase fillColor: nil.
		resultForm copyBits: maskForm boundingBox from: maskForm at: maskLoc
				clippingBox: maskRect rule: Form under fillColor: Color lightBrown.

		"Now update Display in a single BLT."
		self copyBits: maskRect from: resultForm at: topLeft + maskRect topLeft
				clippingBox: self boundingBox rule: Form over fillColor: nil.
		Display forceDisplayUpdate.
		maskLoc := maskLoc + (delta*stepSize)]
"
1 to: 4 do: [:corner | Display pageImage:
				(Form fromDisplay: (10@10 extent: 200@300)) reverse
			at: 10@10 corner: corner]
"
! !

!Form methodsFor: 'transitions' stamp: 'ar 5/28/2000 12:12'!
pageWarp: otherImage at: topLeft forward: forward
	"Produce a page-turning illusion that gradually reveals otherImage
	located at topLeft in this form.
	forward == true means turn pages toward you, else away. [ignored for now]"
	| pageRect oldPage nSteps buffer p leafRect sourceQuad warp oldBottom d |
	pageRect := otherImage boundingBox.
	oldPage := self copy: (pageRect translateBy: topLeft).
	(forward ifTrue: [oldPage] ifFalse: [otherImage])
		border: pageRect
		widthRectangle: (Rectangle
				left: 0
				right: 2
				top: 1
				bottom: 1)
		rule: Form over
		fillColor: Color black.
	oldBottom := self copy: ((pageRect bottomLeft + topLeft) extent: (pageRect width@(pageRect height//4))).
	nSteps := 8.
	buffer := Form extent: otherImage extent + (0@(pageRect height//4)) depth: self depth.
	d := pageRect topLeft + (0@(pageRect height//4)) - pageRect topRight.
	1 to: nSteps-1 do:
		[:i | forward
			ifTrue: [buffer copy: pageRect from: otherImage to: 0@0 rule: Form over.
					p := pageRect topRight + (d * i // nSteps)]
			ifFalse: [buffer copy: pageRect from: oldPage to: 0@0 rule: Form over.
					p := pageRect topRight + (d * (nSteps-i) // nSteps)].
		buffer copy: oldBottom boundingBox from: oldBottom to: pageRect bottomLeft rule: Form over.
		leafRect := pageRect topLeft corner: p x @ (pageRect bottom + p y).
		sourceQuad := Array with: pageRect topLeft
			with: pageRect bottomLeft + (0@p y)
			with: pageRect bottomRight
			with: pageRect topRight - (0@p y).
		warp := (WarpBlt current toForm: buffer)
				clipRect: leafRect;
				sourceForm: (forward ifTrue: [oldPage] ifFalse: [otherImage]);
				combinationRule: Form paint.
		warp copyQuad: sourceQuad toRect: leafRect.
		self copy: buffer boundingBox from: buffer to: topLeft rule: Form over.
		Display forceDisplayUpdate].

	buffer copy: pageRect from: otherImage to: 0@0 rule: Form over.
	buffer copy: oldBottom boundingBox from: oldBottom to: pageRect bottomLeft rule: Form over.
	self copy: buffer boundingBox from: buffer to: topLeft rule: Form over.
	Display forceDisplayUpdate.
"
1 to: 4 do: [:corner | Display pageWarp:
				(Form fromDisplay: (10@10 extent: 200@300)) reverse
			at: 10@10 forward: false]
"
! !

!Form methodsFor: 'displaying' stamp: 'ar 5/28/2000 12:08'!
paintBits: sourceForm at: destOrigin translucent: factor
	"Make up a BitBlt table and copy the bits with the given colorMap."
	(BitBlt current destForm: self
		sourceForm: sourceForm
		halftoneForm: nil
		combinationRule: 31
		destOrigin: destOrigin
		sourceOrigin: 0@0
		extent: sourceForm extent
		clipRect: self boundingBox)
		copyBitsTranslucent: ((0 max: (factor*255.0) asInteger) min: 255)
"
 | f f2 f3 | f := Form fromUser. f replaceColor: f peripheralColor withColor: Color transparent.
f2 := Form fromDisplay: (0@0 extent: f extent). f3 := f2 deepCopy.
0.0 to: 1.0 by: 1.0/32 do:
	[:t | f3 := f2 deepCopy. f3 paintBits: f at: 0@0 translucent: t.
	f3 displayAt: 0@0. (Delay forMilliseconds: 100) wait].
"! !

!Form methodsFor: 'analyzing' stamp: 'ar 5/17/2001 15:40'!
pixelCompare: aRect with: otherForm at: otherLoc
	"Compare the selected bits of this form (those within aRect) against
	those in a similar rectangle of otherFrom.  Return the sum of the
	absolute value of the differences of the color values of every pixel.
	Obviously, this is most useful for rgb (16- or 32-bit) pixels but,
	in the case of 8-bits or less, this will return the sum of the differing
	bits of the corresponding pixel values (somewhat less useful)"
	| pixPerWord temp |
	pixPerWord := 32//self depth.
	(aRect left\\pixPerWord = 0 and: [aRect right\\pixPerWord = 0]) ifTrue:
		["If word-aligned, use on-the-fly difference"
		^ (BitBlt current toForm: self) copy: aRect from: otherLoc in: otherForm
				fillColor: nil rule: 32].
	"Otherwise, combine in a word-sized form and then compute difference"
	temp := self copy: aRect.
	temp copy: aRect from: otherLoc in: otherForm rule: 21.
	^ (BitBlt current toForm: temp) copy: aRect from: otherLoc in: nil
				fillColor: (Bitmap with: 0) rule: 32
"  Dumb example prints zero only when you move over the original rectangle...
 | f diff | f := Form fromUser.
[Sensor anyButtonPressed] whileFalse:
	[diff := f pixelCompare: f boundingBox
		with: Display at: Sensor cursorPoint.
	diff printString , '        ' displayAt: 0@0]
"! !

!Form methodsFor: 'pixel access' stamp: 'ar 5/28/2000 12:08'!
pixelValueAt: aPoint 
	"Return the raw pixel value at the given point. This pixel value depends on the receiver's depth. Typical clients use colorAt: to get a Color.  "

	^ (BitBlt current bitPeekerFromForm: self) pixelAt: aPoint
! !

!Form methodsFor: 'pixel access' stamp: 'ar 5/28/2000 12:08'!
pixelValueAt: aPoint put: pixelValue
	"Store the given raw pixel value at the given point. Typical clients use colorAt:put: to store a color. "

	(BitBlt current bitPokerToForm: self) pixelAt: aPoint put: pixelValue.
! !

!Form methodsFor: 'color mapping' stamp: 'ar 5/15/2001 16:16'!
pixelValueFor: aColor
	"Return the pixel word for representing the given color on the receiver"
	self hasNonStandardPalette
		ifTrue:[^self colormapFromARGB mapPixel: (aColor pixelValueForDepth: 32)]
		ifFalse:[^aColor pixelValueForDepth: self depth]! !

!Form methodsFor: 'color mapping' stamp: 'ar 5/15/2001 16:16'!
pixelWordFor: aColor
	"Return the pixel word for representing the given color on the receiver"
	| basicPattern |
	self hasNonStandardPalette 
		ifFalse:[^aColor pixelWordForDepth: self depth].
	basicPattern := self pixelValueFor: aColor.
	self depth = 32 
		ifTrue:[^basicPattern]
		ifFalse:[^aColor pixelWordFor: self depth filledWith: basicPattern]! !

!Form methodsFor: 'analyzing' stamp: 'ar 5/17/2001 15:42'!
primCountBits
	"Count the non-zero pixels of this form."
	self depth > 8 ifTrue:
		[^(self asFormOfDepth: 8) primCountBits].
	^ (BitBlt current toForm: self)
		fillColor: (Bitmap with: 0);
		destRect: (0@0 extent: width@height);
		combinationRule: 32;
		copyBits! !

!Form methodsFor: 'other' stamp: 'jm 1/6/98 10:37'!
primPrintHScale: hScale vScale: vScale landscape: aBoolean
	"On platforms that support it, this primitive prints the receiver, assumed to be a Form, to the default printer."
	"(Form extent: 10@10) primPrintHScale: 1.0 vScale: 1.0 landscape: true"

	<primitive: 232>
	self primitiveFailed
! !

!Form methodsFor: 'fileIn/Out' stamp: 'di 3/15/1999 14:50'!
printOn: aStream
    aStream
        nextPutAll: self class name;
        nextPut: $(; print: width;
        nextPut: $x; print: height;
        nextPut: $x; print: depth;
        nextPut: $).
! !

!Form methodsFor: 'postscript generation' stamp: 'nk 12/31/2003 15:46'!
printPostscript: aStream operator: operator 
	aStream preserveStateDuring: 
			[:inner | 
			inner rectclip: (0 @ 0 extent: width @ height).
			self setColorspaceOn: inner.
			inner
				print: '[ ';
				cr;
				print: '/ImageType 1';
				cr;
				print: '/ImageMatrix [1 0 0 1 0 0]';
				cr;
				print: '/MultipleDataSources false';
				cr;
				print: '/DataSource level1 { { currentfile ';
				write: self bytesPerRow;
				print: ' string readhexstring pop }} bind { currentfile /ASCIIHexDecode filter } ifelse';
				cr;
				print: '/Width ';
				write: self paddedWidth;
				cr;
				print: '/Height ';
				write: self height;
				cr;
				print: '/Decode ';
				print: self decodeArray;
				cr;
				print: '/BitsPerComponent ';
				write: self bitsPerComponent;
				cr;
				print: 'makeDict ';
				print: operator;
				cr.
			self storePostscriptHexOn: inner.
			inner
				print: $>;
				cr.
			inner cr].
	aStream cr! !

!Form methodsFor: 'private' stamp: 'ar 12/19/2000 16:23'!
privateFloodFillValue: aColor
	"Private. Compute the pixel value in the receiver's depth but take into account implicit color conversions by BitBlt."
	| f1 f2 bb |
	f1 := Form extent: 1@1 depth: depth.
	f2 := Form extent: 1@1 depth: 32.
	bb := BitBlt toForm: f1.
	bb fillColor: aColor; 
		destRect: (0@0 corner: 1@1); 
		combinationRule: 3; 
		copyBits.
	bb := BitBlt toForm: f2.
	bb sourceForm: f1; 
		sourceOrigin: 0@0;
		destRect: (0@0 corner: 1@1);
		combinationRule: 3;
		copyBits.
	^f2 pixelValueAt: 0@0.! !

!Form methodsFor: 'fileIn/Out' stamp: 'mu 8/17/2003 00:44'!
readAttributesFrom: aBinaryStream
	| offsetX offsetY |
	depth := aBinaryStream next.
	(self depth isPowerOfTwo and: [self depth between: 1 and: 32])
		ifFalse: [self error: 'invalid depth; bad Form file?'].
	width := aBinaryStream nextWord.
	height := aBinaryStream nextWord.
	offsetX  := aBinaryStream nextWord.
	offsetY := aBinaryStream nextWord.
	offsetX > 32767 ifTrue: [offsetX := offsetX - 65536].
	offsetY > 32767 ifTrue: [offsetY := offsetY - 65536].
	offset := Point x: offsetX y: offsetY.
	
! !

!Form methodsFor: 'fileIn/Out' stamp: 'mu 8/17/2003 00:43'!
readBitsFrom: aBinaryStream
	
	bits := Bitmap newFromStream: aBinaryStream.
	bits size = self bitsSize ifFalse: [self error: 'wrong bitmap size; bad Form file?'].
	^ self
! !

!Form methodsFor: 'fileIn/Out' stamp: 'mu 8/17/2003 00:44'!
readFrom: aBinaryStream
	"Reads the receiver from the given binary stream with the format:
		depth, extent, offset, bits."
	self readAttributesFrom: aBinaryStream.
	self readBitsFrom: aBinaryStream! !

!Form methodsFor: 'fileIn/Out' stamp: 'jm 3/27/98 16:54'!
readFromOldFormat: aBinaryStream
	"Read a Form in the original ST-80 format."

	| w h offsetX offsetY newForm theBits pos |
	self error: 'this method must be updated to read into 32-bit word bitmaps'.
	w := aBinaryStream nextWord.
	h := aBinaryStream nextWord.
	offsetX  := aBinaryStream nextWord.
	offsetY := aBinaryStream nextWord.
	offsetX > 32767 ifTrue: [offsetX := offsetX - 65536].
	offsetY > 32767 ifTrue: [offsetY := offsetY - 65536].
	newForm := Form extent: w @ h offset: offsetX @ offsetY.
	theBits := newForm bits.
	pos := 0.
	1 to: w + 15 // 16 do: [:j |
		1 to: h do: [:i |
			theBits at: (pos := pos+1) put: aBinaryStream nextWord]].
	newForm bits: theBits.
	^ newForm
! !

!Form methodsFor: 'resources' stamp: 'ar 12/9/2002 16:04'!
readNativeResourceFrom: byteStream
	| img aStream |
	(byteStream isKindOf: FileStream) ifTrue:[
		"Ugly, but ImageReadWriter will send #reset which is implemented as #reopen and we may not be able to do so."
		aStream := RWBinaryOrTextStream with: byteStream contents.
	] ifFalse:[
		aStream := byteStream.
	].
	img := [ImageReadWriter formFromStream: aStream] on: Error do:[:ex| nil].
	img ifNil:[^nil].
	(img isColorForm and:[self isColorForm]) ifTrue:[
		| cc |
		cc := img colors.
		img colors: nil.
		img displayOn: self.
		img colors: cc.
	] ifFalse:[
		img displayOn: self.
	].
	img := nil.! !

!Form methodsFor: 'resources' stamp: 'nk 7/30/2004 17:53'!
readResourceFrom: aStream 
	"Store a resource representation of the receiver on aStream.
	Must be specific to the receiver so that no code is filed out."

	| bitsSize msb |
	(aStream next: 4) asString = self resourceTag 
		ifFalse: 
			[aStream position: aStream position - 4.
			^self readNativeResourceFrom: aStream].
	width := aStream nextNumber: 4.
	height := aStream nextNumber: 4.
	depth := aStream nextNumber: 4.
	bitsSize := aStream nextNumber: 4.
	bitsSize = 0 
		ifFalse: 
			[bits := aStream next: bitsSize.
			^self].
	msb := (aStream nextNumber: 4) = 1.
	bitsSize := aStream nextNumber: 4.
	bits := Bitmap new: self bitsSize.
	(Form 
		extent: width @ height
		depth: depth
		bits: (aStream next: bitsSize * 4)) displayOn: self.
	msb = SmalltalkImage current  isBigEndian 
		ifFalse: 
			[Bitmap 
				swapBytesIn: bits
				from: 1
				to: bits size]! !

!Form methodsFor: 'analyzing' stamp: 'ar 5/17/2001 15:37'!
rectangleEnclosingPixelsNotOfColor: aColor
	"Answer the smallest rectangle enclosing all the pixels of me that are different from the given color. Useful for extracting a foreground graphic from its background."

	| cm slice copyBlt countBlt top bottom newH left right |
	"map the specified color to 1 and all others to 0"
	cm := Bitmap new: (1 bitShift: (self depth min: 15)).
	cm primFill: 1.
	cm at: (aColor indexInMap: cm) put: 0.

	"build a 1-pixel high horizontal slice and BitBlts for counting pixels of interest"
	slice := Form extent: width@1 depth: 1.
	copyBlt := (BitBlt current toForm: slice)
		sourceForm: self;
		combinationRule: Form over;
		destX: 0 destY: 0 width: width height: 1;
		colorMap: cm.
	countBlt := (BitBlt current toForm: slice)
		fillColor: (Bitmap with: 0);
		destRect: (0@0 extent: slice extent);
		combinationRule: 32.

	"scan in from top and bottom"
	top := (0 to: height)
		detect: [:y |
			copyBlt sourceOrigin: 0@y; copyBits.
			countBlt copyBits > 0]
		ifNone: [^ 0@0 extent: 0@0].
	bottom := (height - 1 to: top by: -1)
		detect: [:y |
			copyBlt sourceOrigin: 0@y; copyBits.
			countBlt copyBits > 0].

	"build a 1-pixel wide vertical slice and BitBlts for counting pixels of interest"
	newH := bottom - top + 1.
	slice := Form extent: 1@newH depth: 1.
	copyBlt := (BitBlt current toForm: slice)
		sourceForm: self;
		combinationRule: Form over;
		destX: 0 destY: 0 width: 1 height: newH;
		colorMap: cm.
	countBlt := (BitBlt current toForm: slice)
		fillColor: (Bitmap with: 0);
		destRect: (0@0 extent: slice extent);
		combinationRule: 32.

	"scan in from left and right"
	left := (0 to: width)
		detect: [:x |
			copyBlt sourceOrigin: x@top; copyBits.
			countBlt copyBits > 0].
	right := (width - 1 to: left by: -1)
		detect: [:x |
			copyBlt sourceOrigin: x@top; copyBits.
			countBlt copyBits > 0].

	^ left@top corner: (right + 1)@(bottom + 1)
! !

!Form methodsFor: 'color mapping' stamp: 'di 10/16/2001 15:23'!
reducedPaletteOfSize: nColors
	"Return an array of colors of size nColors, such that those colors
	represent well the pixel values actually found in this form."
	| threshold tallies colorTallies dist delta palette cts top cluster |
	tallies := self tallyPixelValues.  "An array of tallies for each pixel value"
	threshold := width * height // 500.

	"Make an array of (color -> tally) for all tallies over threshold"
	colorTallies := Array streamContents:
		[:s | tallies withIndexDo:
			[:v :i | v >= threshold ifTrue:
				[s nextPut: (Color colorFromPixelValue: i-1 depth: depth) -> v]]].

	"Extract a set of clusters by picking the top tally, and then removing all others
	whose color is within dist of it.  Iterate the process, adjusting dist until we get nColors."
	dist := 0.2.  delta := dist / 2.
		[cts := colorTallies copy.
		palette := Array streamContents: [:s |
			[cts isEmpty] whileFalse:
				[top := cts detectMax: [:a | a value].
				cluster := cts select: [:a | (a key diff: top key) < dist].
				s nextPut: top key -> (cluster detectSum: [:a | a value]).
				cts := cts copyWithoutAll: cluster]].
		palette size = nColors or: [delta < 0.001]]
		whileFalse:
			[palette size > nColors
				ifTrue: [dist := dist + delta]
				ifFalse: [dist := dist - delta].
			delta := delta / 2].
	^ palette collect: [:a | a key]
! !

!Form methodsFor: 'other' stamp: 'RAA 1/30/2002 16:42'!
relativeTextAnchorPosition

	^nil		"so forms can be in TextAnchors"! !

!Form methodsFor: 'fileIn/Out' stamp: 'ar 2/24/2001 22:39'!
replaceByResource: aForm
	"Replace the receiver by some resource that just got loaded"
	(self extent = aForm extent and:[self depth = aForm depth]) ifTrue:[
		bits := aForm bits.
	].! !

!Form methodsFor: 'image manipulation' stamp: 'ar 5/17/2001 15:40'!
replaceColor: oldColor withColor: newColor
	"Replace one color with another everywhere is this form"

	| cm newInd target ff |
	self depth = 32
		ifTrue: [cm := (Color  cachedColormapFrom: 16 to: 32) copy]
		ifFalse: [cm := Bitmap new: (1 bitShift: (self depth min: 15)).
				1 to: cm size do: [:i | cm at: i put: i - 1]].
	newInd := newColor pixelValueForDepth: self depth.
	cm at: (oldColor pixelValueForDepth: (self depth min: 16))+1 put: newInd.
	target := newColor isTransparent 
		ifTrue: [ff := Form extent: self extent depth: depth.
			ff fillWithColor: newColor.  ff]
		ifFalse: [self].
	(BitBlt current toForm: target)
		sourceForm: self;
		sourceOrigin: 0@0;
		combinationRule: Form paint;
		destX: 0 destY: 0 width: width height: height;
		colorMap: cm;
		copyBits.
	newColor = Color transparent 
		ifTrue: [target displayOn: self].! !

!Form methodsFor: 'resources' stamp: 'ar 2/27/2001 14:56'!
resourceTag
	^'FORM'! !

!Form methodsFor: 'color mapping' stamp: 'ar 5/27/2000 20:14'!
rgbaBitMasks
	"Return the masks for specifying the R,G,B, and A components in the receiver"
	self depth <= 8
		ifTrue:[^#(16rFF0000 16rFF00 16rFF 16rFF000000)].
	self depth = 16
		ifTrue:[^#(16r7C00 16r3E0 16r1F 16r0)].
	self depth = 32
		ifTrue:[^#(16rFF0000 16rFF00 16rFF 16rFF000000)].
	self error:'Bad depth for form'! !

!Form methodsFor: 'scaling, rotation' stamp: ''!
rotateBy: deg
	"Rotate the receiver by the indicated number of degrees."
	"rot is the destination form, bit enough for any angle."

	^ self rotateBy: deg smoothing: 1
"
 | a f |  f := Form fromDisplay: (0@0 extent: 200@200).  a := 0.
[Sensor anyButtonPressed] whileFalse:
	[((Form fromDisplay: (Sensor cursorPoint extent: 130@66))
		rotateBy: (a := a+5)) display].
f display
"! !

!Form methodsFor: 'scaling, rotation' stamp: 'wiz 1/22/2006 01:15'!
rotateBy: direction centerAt: aPoint
	"Return a rotated copy of the receiver. 
	direction = #none, #right, #left, or #pi"
	| newForm quad rot scale |
	direction == #none ifTrue: [^ self].
	scale :=  (direction = #pi ifTrue: [width@height] ifFalse: [height@width]) / self extent .
	newForm := self blankCopyOf: self boundingBox scaledBy: scale.
	quad := self boundingBox innerCorners.
	rot := #(right pi left) indexOf: direction.
	(WarpBlt current toForm: newForm)
		sourceForm: self;
		colorMap: (self colormapIfNeededFor: newForm);
		combinationRule: 3;
		copyQuad: ((1+rot to: 4+rot) collect: [:i | quad atWrap: i])
			 toRect: newForm boundingBox.
	newForm offset: (self offset rotateBy: direction centerAt: aPoint).
	^ newForm
"
[Sensor anyButtonPressed] whileFalse:
	[((Form fromDisplay: (Sensor cursorPoint extent: 130@66))
		rotateBy: #left centerAt: 0@0) display]
"
"Consistency test...
 | f f2 p | [Sensor anyButtonPressed] whileFalse:
	[f := Form fromDisplay: ((p := Sensor cursorPoint) extent: 31@41).
	Display fillBlack: (p extent: 31@41).
	f2 := f rotateBy: #left centerAt: 0@0.
	(f2 rotateBy: #right centerAt: 0@0) displayAt: p]
"
! !

!Form methodsFor: 'scaling, rotation' stamp: 'ar 3/1/2006 23:04'!
rotateBy: deg magnify: scale smoothing: cellSize
	"Rotate the receiver by the indicated number of degrees and magnify. scale can be a Point to make for interesting 3D effects "
	"rot is the destination form, big enough for any angle."

	| side rot warp r1 pts p bigSide |
	side := 1 + self extent r asInteger.
	bigSide := (side asPoint * scale) rounded.
	rot := self blankCopyOf: self boundingBox scaledBy: ( bigSide / self extent ).
	warp := (WarpBlt current toForm: rot)
		sourceForm: self;
		colorMap: (self colormapIfNeededFor: rot);
		cellSize: cellSize;  "installs a new colormap if cellSize > 1"
		combinationRule: Form paint.
	r1 := (0@0 extent: side@side) align: (side@side)//2 with: self boundingBox center.

	"Rotate the corners of the source rectangle." 
	pts := r1 innerCorners collect:
		[:pt | p := pt - r1 center.
		(r1 center x asFloat + (p x asFloat*deg degreeCos) + (p y asFloat*deg degreeSin)) @
		(r1 center y asFloat - (p x asFloat*deg degreeSin) + (p y asFloat*deg degreeCos))].
	warp copyQuad: pts toRect: rot boundingBox.
	^ rot
"
 | a f |  f := Form fromDisplay: (0@0 extent: 200@200).  a := 0.
[Sensor anyButtonPressed] whileFalse:
	[((Form fromDisplay: (Sensor cursorPoint extent: 130@66))
		rotateBy: (a := a+5) magnify: 0.75@2 smoothing: 2) display].
f display
"! !

!Form methodsFor: 'scaling, rotation' stamp: 'tpr 9/28/2004 16:55'!
rotateBy: deg smoothing: cellSize
	"Rotate the receiver by the indicated number of degrees."
	^self rotateBy: deg magnify: 1 smoothing: cellSize
"
 | a f |  f := Form fromDisplay: (0@0 extent: 200@200).  a := 0.
[Sensor anyButtonPressed] whileFalse:
	[((Form fromDisplay: (Sensor cursorPoint extent: 130@66))
		rotateBy: (a := a+5) smoothing: 2) display].
f display
"! !

!Form methodsFor: 'postscript generation' stamp: 'mpw 11/15/1999 08:34'!
rowPadding
	^ 32 // self depth! !

!Form methodsFor: 'scaling, rotation' stamp: 'RAA 7/13/2000 12:09'!
scaledToSize: newExtent

	| scale |

	newExtent = self extent ifTrue: [^self].
	scale := newExtent x / self width min: newExtent y / self height.
	^self magnify: self boundingBox by: scale smoothing: 2.
! !

!Form methodsFor: 'other' stamp: 'dgd 8/26/2003 21:44'!
setAsBackground
	"Set this form as a background image."

	| world newColor |
	Smalltalk isMorphic 
		ifTrue:
			[world := self currentWorld.
			newColor := InfiniteForm with: self.
			self rememberCommand:
				(Command new cmdWording: 'set background to a picture' translated;
					undoTarget: world selector: #color: argument: world color;
					redoTarget: world selector: #color: argument: newColor).
			world color: newColor]
		ifFalse:
			[ScheduledControllers screenController model form: self.
			Display restoreAfter: []]! !

!Form methodsFor: 'postscript generation' stamp: ''!
setColorspaceOn:aStream
	self numComponents = 1 ifTrue:[aStream print:'/DeviceGray setcolorspace 0 setgray'; cr.]
		ifFalse:[aStream print:'/DeviceRGB setcolorspace'; cr.].! !

!Form methodsFor: 'private' stamp: '6/9/97 16:10 di'!
setExtent: extent depth: bitsPerPixel
	"Create a virtual bit map with the given extent and bitsPerPixel."

	width := extent x asInteger.
	width < 0 ifTrue: [width := 0].
	height := extent y asInteger.
	height < 0 ifTrue: [height := 0].
	depth := bitsPerPixel.
	bits := Bitmap new: self bitsSize! !

!Form methodsFor: 'private' stamp: 'ar 5/28/2000 15:49'!
setExtent: extent depth: bitsPerPixel bits: bitmap
	"Create a virtual bit map with the given extent and bitsPerPixel."

	width := extent x asInteger.
	width < 0 ifTrue: [width := 0].
	height := extent y asInteger.
	height < 0 ifTrue: [height := 0].
	depth := bitsPerPixel.
	(bits isNil or:[self bitsSize = bitmap size]) ifFalse:[^self error:'Bad dimensions'].
	bits := bitmap! !

!Form methodsFor: 'private' stamp: 'ar 10/30/2000 23:22'!
setResourceBits: aForm
	"Private. Really. Used for setting the 'resource bits' when externalizing some form"
	bits := aForm.! !

!Form methodsFor: 'bordering' stamp: 'di 10/21/2001 09:39'!
shapeBorder: aColor width: borderWidth
	"A simplified version for shapes surrounded by transparency (as SketchMorphs).
	Note also this returns a new form that may be larger, and does not affect the original."
	| shapeForm borderForm newForm |
	newForm := Form extent: self extent + (borderWidth*2) depth: self depth.
	newForm fillColor: Color transparent.
	self displayOn: newForm at: (0@0) + borderWidth.
	"First identify the shape in question as a B/W form"
	shapeForm := (newForm makeBWForm: Color transparent) reverse.
	"Now find the border of that shape"
	borderForm := shapeForm borderFormOfWidth: borderWidth sharpCorners: false.
	"Finally use that shape as a mask to paint the border with color"
	^ newForm fillShape: borderForm fillColor: aColor! !

!Form methodsFor: 'bordering' stamp: ''!
shapeBorder: aColor width: borderWidth interiorPoint: interiorPoint
	sharpCorners: sharpen internal: internal
	"Identify the shape (region of identical color) at interiorPoint,
	and then add an outline of width=borderWidth and color=aColor.
	If sharpen is true, then cause right angles to be outlined by
	right angles.  If internal is true, then produce a border that lies
	within the identified shape.  Thus one can put an internal border
	around the whole background, thus effecting a normal border
	around every other foreground image."
	| shapeForm borderForm interiorColor |
	"First identify the shape in question as a B/W form"
	interiorColor := self colorAt: interiorPoint.
	shapeForm := (self makeBWForm: interiorColor) reverse
				findShapeAroundSeedBlock:
					[:form | form pixelValueAt: interiorPoint put: 1].
	"Reverse the image to grow the outline inward"
	internal ifTrue: [shapeForm reverse].
	"Now find the border fo that shape"
	borderForm := shapeForm borderFormOfWidth: borderWidth sharpCorners: sharpen.
	"Finally use that shape as a mask to paint the border with color"
	self fillShape: borderForm fillColor: aColor! !

!Form methodsFor: 'filling' stamp: 'di 10/17/2001 10:10'!
shapeFill: aColor interiorPoint: interiorPoint
	"Identify the shape (region of identical color) at interiorPoint,
	and then fill that shape with the new color, aColor
	: modified di's original method such that it returns the bwForm, for potential use by the caller"

	| bwForm interiorPixVal map ppd color ind |
	self depth = 1 ifTrue:
		[^ self shapeFill: aColor
			seedBlock: [:form | form pixelValueAt: interiorPoint put: 1]].

	"First map this form into a B/W form with 0's in the interior region."
		"bwForm := self makeBWForm: interiorColor."	"won't work for two whites"
	interiorPixVal := self pixelValueAt: interiorPoint.
	bwForm := Form extent: self extent.
	map := Bitmap new: (1 bitShift: (self depth min: 12)).  "Not calling newColorMap.  All 
			non-foreground go to 0.  Length is 2 to 4096."
	ppd := self depth.	"256 long color map in depth 8 is not one of the following cases"
	3 to: 5 do: [:bitsPerColor | 
		(2 raisedTo: bitsPerColor*3) = map size 
			ifTrue: [ppd := bitsPerColor*3]].	"ready for longer maps than 512"

	ppd <= 8
		ifTrue: [map at: interiorPixVal+1 put: 1]
		ifFalse: [interiorPixVal = 0 
			ifFalse: [color := Color colorFromPixelValue: interiorPixVal depth: self depth.
				ind := color pixelValueForDepth: ppd.
				map at: ind+1 put: 1]
			ifTrue: [map at: 1 put: 1]].
	bwForm copyBits: self boundingBox from: self at: 0@0 colorMap: map.
	bwForm reverse.  "Make interior region be 0's"

	"Now fill the interior region and return that shape"
	bwForm := bwForm findShapeAroundSeedBlock:
					[:form | form pixelValueAt: interiorPoint put: 1].

	"Finally use that shape as a mask to flood the region with color"
	self eraseShape: bwForm.
	self fillShape: bwForm fillColor: aColor.
	^ bwForm! !

!Form methodsFor: 'filling' stamp: 'ar 5/17/2001 15:38'!
shapeFill: aColor seedBlock: seedBlock
	self depth > 1 ifTrue: [self error: 'This call only meaningful for B/W forms'].
	(self findShapeAroundSeedBlock: seedBlock)
		displayOn: self at: 0@0 clippingBox: self boundingBox
		rule: Form under fillColor: aColor ! !

!Form methodsFor: 'testing' stamp: 'ar 5/28/2000 14:58'!
shouldPreserveContents
	"Return true if the receiver should preserve it's contents when flagged to be clean. Most forms can not be trivially restored by some drawing operation but some may."
	^true! !

!Form methodsFor: 'scaling, rotation' stamp: ''!
shrink: aRectangle by: scale 
	| scalePt |
	scalePt := scale asPoint.
	^ self magnify: aRectangle by: (1.0 / scalePt x asFloat) @ (1.0 / scalePt y asFloat)! !

!Form methodsFor: 'initialize-release' stamp: 'ar 5/28/2000 18:45'!
shutDown
	"The system is going down. Try to preserve some space"
	self hibernate! !

!Form methodsFor: 'accessing' stamp: ''!
size
	"Should no longer be used -- use bitsSize instead.  length of variable part of instance."
	^ super size! !

!Form methodsFor: 'transitions' stamp: 'jm 5/21/1998 23:46'!
slideImage: otherImage at: topLeft delta: delta
	"Display slideImage: (Form fromDisplay: (40@40 extent: 300@300)) reverse
		at: 40@40 delta: 3@-4"
	| bb nSteps clipRect |
	bb := otherImage boundingBox.
	clipRect := topLeft extent: otherImage extent.
	nSteps := 1.
	delta x = 0 ifFalse: [nSteps := nSteps max: (bb width//delta x abs) + 1].
	delta y = 0 ifFalse: [nSteps := nSteps max: (bb height//delta y abs) + 1].
	1 to: nSteps do:
			[:i | self copyBits: bb from: otherImage
				at: delta*(i-nSteps) + topLeft
				clippingBox: clipRect rule: Form paint fillColor: nil.
			Display forceDisplayUpdate]! !

!Form methodsFor: 'image manipulation' stamp: 'ar 5/28/2000 12:09'!
smear: dir distance: dist
	"Smear any black pixels in this form in the direction dir in Log N steps"
	| skew bb |
	bb := BitBlt current destForm: self sourceForm: self fillColor: nil
		combinationRule: Form under destOrigin: 0@0 sourceOrigin: 0@0
		extent: self extent clipRect: self boundingBox.
	skew := 1.
	[skew < dist] whileTrue:
		[bb destOrigin: dir*skew; copyBits.
		skew := skew+skew]! !

!Form methodsFor: 'image manipulation' stamp: 'LB 8/26/2002 18:08'!
stencil
	"return a 1-bit deep, black-and-white stencil of myself"

	| canvas |
	canvas := FormCanvas extent: self extent depth: 1.
	canvas fillColor: (Color white).

	canvas stencil: self at: 0@0  
				sourceRect: (Rectangle origin: 0@0 corner: self extent) color: Color black.

	^ canvas form
! !

!Form methodsFor: 'fileIn/Out' stamp: 'nk 12/31/2003 16:06'!
store15To24HexBitsOn:aStream

	| buf i lineWidth |

	"write data for 16-bit form, optimized for encoders writing directly to files to do one single file write rather than 12. I'm not sure I understand the significance of the shifting pattern, but I think I faithfully translated it from the original"

	lineWidth := 0.
	buf := String new: 12.
	bits do: [:word | 
		i := 0.
		"upper pixel"
		buf at: (i := i + 1) put: ((word bitShift: -27) bitAnd: 15) asHexDigit.
		buf at: (i := i + 1) put: ((word bitShift: -32) bitAnd: 8) asHexDigit.

		buf at: (i := i + 1) put: ((word bitShift: -22) bitAnd: 15) asHexDigit.
		buf at: (i := i + 1) put: ((word bitShift: -27) bitAnd: 8) asHexDigit.

		buf at: (i := i + 1) put: ((word bitShift: -17) bitAnd: 15) asHexDigit.
		buf at: (i := i + 1) put: ((word bitShift: -22) bitAnd: 8) asHexDigit.

		"lower pixel"

		buf at: (i := i + 1) put: ((word bitShift: -11) bitAnd: 15) asHexDigit.
		buf at: (i := i + 1) put: ((word bitShift: -16) bitAnd: 8) asHexDigit.

		buf at: (i := i + 1) put: ((word bitShift: -6) bitAnd: 15) asHexDigit.
		buf at: (i := i + 1) put: ((word bitShift: -11) bitAnd: 8) asHexDigit.

		buf at: (i := i + 1) put: ((word bitShift: -1) bitAnd: 15) asHexDigit.
		buf at: (i := i + 1) put: ((word bitShift: -6) bitAnd: 8) asHexDigit.
		aStream nextPutAll: buf.
		lineWidth := lineWidth + 12.
		lineWidth > 100 ifTrue: [ aStream cr. lineWidth := 0 ].
		"#( 31 26 21 15 10 5 )  do:[:startBit | ]"
	].! !

!Form methodsFor: 'fileIn/Out' stamp: ''!
store32To24HexBitsOn:aStream
	^self storeBits:20 to:0 on:aStream.! !

!Form methodsFor: 'fileIn/Out' stamp: ''!
storeBits:startBit to:stopBit on:aStream
	bits storeBits:startBit to:stopBit on:aStream.! !

!Form methodsFor: 'fileIn/Out' stamp: 'laza 3/29/2004 12:21'!
storeBitsOn:aStream base:anInteger
	bits do: [:word | 
		anInteger = 10
			ifTrue: [aStream space]
			ifFalse: [aStream crtab: 2].
		word storeOn: aStream base: anInteger].
! !

!Form methodsFor: 'fileIn/Out' stamp: ''!
storeHexBitsOn:aStream
	^self storeBits:28 to:0 on:aStream.! !

!Form methodsFor: 'fileIn/Out' stamp: ''!
storeOn: aStream

	self storeOn: aStream base: 10! !

!Form methodsFor: 'fileIn/Out' stamp: ''!
storeOn: aStream base: anInteger 
	"Store the receiver out as an expression that can be evaluated to recreate a Form with the same contents as the original."

	self unhibernate.
	aStream nextPut: $(.
	aStream nextPutAll: self species name.
	aStream crtab: 1.
	aStream nextPutAll: 'extent: '.
	self extent printOn: aStream.
	aStream crtab: 1.
	aStream nextPutAll: 'depth: '.
	self depth printOn: aStream.
	aStream crtab: 1.
	aStream nextPutAll: 'fromArray: #('.
	self storeBitsOn:aStream base:anInteger.
	aStream nextPut: $).
	aStream crtab: 1.
	aStream nextPutAll: 'offset: '.
	self offset printOn: aStream.
	aStream nextPut: $).
! !

!Form methodsFor: 'postscript generation' stamp: 'nk 12/31/2003 15:46'!
storePostscriptHexOn: inner 
	self depth <= 8 ifTrue: [self storeHexBitsOn: inner].
	self depth = 16 ifTrue: [self store15To24HexBitsOn: inner].
	self depth = 32 ifTrue: [self store32To24HexBitsOn: inner]! !

!Form methodsFor: 'resources' stamp: 'sd 9/30/2003 13:41'!
storeResourceOn: aStream
	"Store a resource representation of the receiver on aStream.
	Must be specific to the receiver so that no code is filed out."
	self hibernate.
	aStream nextPutAll: self resourceTag asByteArray. "tag"
	aStream nextNumber: 4 put: width.
	aStream nextNumber: 4 put: height.
	aStream nextNumber: 4 put: depth.
	(bits isMemberOf: ByteArray) ifFalse:[
		"must store bitmap"
		aStream nextNumber: 4 put: 0. "tag"
		aStream nextNumber: 4 put: (SmalltalkImage current  isBigEndian ifTrue:[1] ifFalse:[0]).
	].
	aStream nextNumber: 4 put: bits size.
	aStream nextPutAll: bits.
! !

!Form methodsFor: 'initialize-release' stamp: 'ar 6/16/2002 18:39'!
swapEndianness
	"Swap from big to little endian pixels and vice versa"
	depth := 0 - depth.! !

!Form methodsFor: 'analyzing' stamp: 'jm 6/18/1999 18:41'!
tallyPixelValues
	"Answer a Bitmap whose elements contain the number of pixels in this Form with the pixel value corresponding to their index. Note that the pixels of multiple Forms can be tallied together using tallyPixelValuesInRect:into:."

	^ self tallyPixelValuesInRect: self boundingBox
		into: (Bitmap new: (1 bitShift: (self depth min: 15)))
"
Move a little rectangle around the screen and print its tallies...
 | r tallies nonZero |
Cursor blank showWhile: [
[Sensor anyButtonPressed] whileFalse:
	[r := Sensor cursorPoint extent: 10@10.
	Display border: (r expandBy: 2) width: 2 rule: Form reverse fillColor: nil.
	tallies := (Display copy: r) tallyPixelValues.
	nonZero := (1 to: tallies size) select: [:i | (tallies at: i) > 0]
			thenCollect: [:i | (tallies at: i) -> (i-1)].
	nonZero printString , '          ' displayAt: 0@0.
	Display border: (r expandBy: 2) width: 2 rule: Form reverse fillColor: nil]]
"
! !

!Form methodsFor: 'analyzing' stamp: 'ar 5/28/2000 12:09'!
tallyPixelValuesInRect: destRect into: valueTable
	"Tally the selected pixels of this Form into valueTable, a Bitmap of depth 2^depth similar to a color map. Answer valueTable."

	(BitBlt current toForm: self)
		sourceForm: self;  "src must be given for color map ops"
		sourceOrigin: 0@0;
		tallyMap: valueTable;
		combinationRule: 33;
		destRect: destRect;
		copyBits.
	^ valueTable

"
Move a little rectangle around the screen and print its tallies...
 | r tallies nonZero |
Cursor blank showWhile: [
[Sensor anyButtonPressed] whileFalse:
	[r := Sensor cursorPoint extent: 10@10.
	Display border: (r expandBy: 2) width: 2 rule: Form reverse fillColor: nil.
	tallies := (Display copy: r) tallyPixelValues.
	nonZero := (1 to: tallies size) select: [:i | (tallies at: i) > 0]
			thenCollect: [:i | (tallies at: i) -> (i-1)].
	nonZero printString , '          ' displayAt: 0@0.
	Display border: (r expandBy: 2) width: 2 rule: Form reverse fillColor: nil]]
"! !

!Form methodsFor: 'image manipulation' stamp: 'jm 6/30/1999 15:36'!
trimBordersOfColor: aColor
	"Answer a copy of this Form with each edge trimmed in to the first pixel that is not of the given color. (That is, border strips of the given color are removed)."

	| r |
	r := self rectangleEnclosingPixelsNotOfColor: aColor.
	^ self copy: r
! !

!Form methodsFor: 'fileIn/Out' stamp: 'ar 3/3/2001 15:50'!
unhibernate
	"If my bitmap has been compressed into a ByteArray,
	then expand it now, and return true."
	| resBits |
	bits isForm ifTrue:[
		resBits := bits.
		bits := Bitmap new: self bitsSize.
		resBits displayResourceFormOn: self.
		^true].
	bits == nil ifTrue:[bits := Bitmap new: self bitsSize. ^true].
	(bits isMemberOf: ByteArray)
		ifTrue: [bits := Bitmap decompressFromByteArray: bits. ^ true].
	^ false! !

!Form methodsFor: 'copying' stamp: 'tk 8/19/1998 16:11'!
veryDeepCopyWith: deepCopier
	"Return self.  I am immutable in the Morphic world.  Do not record me."
	^ self! !

!Form methodsFor: 'accessing' stamp: ''!
width
	^ width! !

!Form methodsFor: 'transitions' stamp: 'jm 6/18/1998 12:57'!
wipeImage: otherImage at: topLeft clippingBox: clipBox rectForIndex: rectForIndexBlock

	| i clipRect t rectOrList waitTime |
	i := 0.
	clipRect := topLeft extent: otherImage extent.
	clipBox ifNotNil: [clipRect := clipRect intersect: clipBox].
	[rectOrList := rectForIndexBlock value: (i := i + 1).
	 rectOrList == nil]
		whileFalse: [
			t := Time millisecondClockValue.
			rectOrList asOrderedCollection do: [:r |
				self copyBits: r from: otherImage at: topLeft + r topLeft
					clippingBox: clipRect rule: Form over fillColor: nil].
			Display forceDisplayUpdate.
			waitTime := 3 - (Time millisecondClockValue - t).
			waitTime > 0 ifTrue:
				["(Delay forMilliseconds: waitTime) wait"]].
! !

!Form methodsFor: 'transitions' stamp: 'jm 10/16/97 15:21'!
wipeImage: otherImage at: topLeft delta: delta
	"Display wipeImage: (Form fromDisplay: (40@40 extent: 300@300)) reverse
		at: 40@40 delta: 0@-2"

	self wipeImage: otherImage at: topLeft delta: delta clippingBox: nil.
! !

!Form methodsFor: 'transitions' stamp: 'jm 10/16/97 15:17'!
wipeImage: otherImage at: topLeft delta: delta clippingBox: clipBox

	| wipeRect bb nSteps |
	bb := otherImage boundingBox.
	wipeRect := delta x = 0
		ifTrue:
		[delta y = 0 ifTrue: [nSteps := 1. bb "allow 0@0"] ifFalse: [
		nSteps := bb height//delta y abs + 1.  "Vertical movement"
		delta y > 0
			ifTrue: [bb topLeft extent: bb width@delta y]
			ifFalse: [bb bottomLeft+delta extent: bb width@delta y negated]]]
		ifFalse:
		[nSteps := bb width//delta x abs + 1.  "Horizontal movement"
		delta x > 0
			ifTrue: [bb topLeft extent: delta x@bb height]
			ifFalse: [bb topRight+delta extent: delta x negated@bb height]].
	^ self wipeImage: otherImage at: topLeft clippingBox: clipBox rectForIndex:
		[:i | i <= nSteps
			ifTrue: [wipeRect translateBy: (delta* (i-1))]
			ifFalse: [nil]]! !

!Form methodsFor: 'fileIn/Out' stamp: 'mu 8/17/2003 00:35'!
writeAttributesOn: file
	self unhibernate.
	file nextPut: depth.
	file nextWordPut: width.
	file nextWordPut: height.
	file nextWordPut: ((self offset x) >=0
					ifTrue: [self offset x]
					ifFalse: [self offset x + 65536]).
	file nextWordPut: ((self offset y) >=0
					ifTrue: [self offset y]
					ifFalse: [self offset y + 65536]).
	! !

!Form methodsFor: 'fileIn/Out' stamp: 'ar 6/16/2002 17:53'!
writeBMPfileNamed: fName  "Display writeBMPfileNamed: 'display.bmp'"
	BMPReadWriter putForm: self onFileNamed: fName! !

!Form methodsFor: 'fileIn/Out' stamp: 'mu 8/17/2003 00:35'!
writeBitsOn: file
	bits writeOn: file! !

!Form methodsFor: 'fileIn/Out' stamp: 'sw 2/20/2002 15:37'!
writeJPEGfileNamed: fileName 
	"Write a JPEG file to the given filename using default settings"

	self writeJPEGfileNamed: fileName progressive: false

"
Display writeJPEGfileNamed: 'display.jpeg'
Form fromUser writeJPEGfileNamed: 'yourPatch.jpeg'
"! !

!Form methodsFor: 'fileIn/Out' stamp: 'sw 2/20/2002 15:29'!
writeJPEGfileNamed: fileName  progressive: aBoolean
	"Write a JPEG file to the given filename using default settings.  Make it progressive or not, depending on the boolean argument"

	JPEGReadWriter2 putForm: self quality: -1 "default" progressiveJPEG: aBoolean onFileNamed: fileName

"
Display writeJPEGfileNamed: 'display.jpeg' progressive: false.
Form fromUser writeJPEGfileNamed: 'yourPatch.jpeg' progressive: true
"! !

!Form methodsFor: 'fileIn/Out' stamp: 'mu 8/17/2003 00:36'!
writeOn: file
	"Write the receiver on the file in the format
		depth, extent, offset, bits."
	self writeAttributesOn: file.
	self writeBitsOn: file! !

!Form methodsFor: 'fileIn/Out' stamp: 'di 7/6/1998 23:00'!
writeOnMovie: file
	"Write just my bits on the file."
	self unhibernate.
	bits writeUncompressedOn: file! !

!Form methodsFor: 'fileIn/Out' stamp: 'tk 2/19/1999 07:30'!
writeUncompressedOn: file
	"Write the receiver on the file in the format depth, extent, offset, bits.  Warning:  Caller must put header info on file!!  Use writeUncompressedOnFileNamed: instead."
	self unhibernate.
	file binary.
	file nextPut: depth.
	file nextWordPut: width.
	file nextWordPut: height.
	file nextWordPut: ((self offset x) >=0
					ifTrue: [self offset x]
					ifFalse: [self offset x + 65536]).
	file nextWordPut: ((self offset y) >=0
					ifTrue: [self offset y]
					ifFalse: [self offset y + 65536]).
	bits writeUncompressedOn: file! !

!Form methodsFor: 'analyzing' stamp: 'ar 5/28/2000 12:09'!
xTallyPixelValue: pv orNot: not
	"Return an array of the number of pixels with value pv by x-value.
	Note that if not is true, then this will tally those different from pv."
	| cm slice countBlt copyBlt |
	cm := self newColorMap.		"Map all colors but pv to zero"
	not ifTrue: [cm atAllPut: 1].		"... or all but pv to one"
	cm at: pv+1 put: 1 - (cm at: pv+1).
	slice := Form extent: 1@height.
	copyBlt := (BitBlt current destForm: slice sourceForm: self
				halftoneForm: nil combinationRule: Form over
				destOrigin: 0@0 sourceOrigin: 0@0 extent: 1 @ slice height
				clipRect: slice boundingBox) colorMap: cm.
	countBlt := (BitBlt current toForm: slice)
				fillColor: (Bitmap with: 0);
				destRect: (0@0 extent: slice extent);
				combinationRule: 32.
	^ (0 to: width-1) collect:
		[:x |
		copyBlt sourceOrigin: x@0; copyBits.
		countBlt copyBits]! !

!Form methodsFor: 'analyzing' stamp: 'ar 5/28/2000 12:09'!
yTallyPixelValue: pv orNot: not
	"Return an array of the number of pixels with value pv by y-value.
	Note that if not is true, then this will tally those different from pv."
	| cm slice copyBlt countBlt |
	cm := self newColorMap.		"Map all colors but pv to zero"
	not ifTrue: [cm atAllPut: 1].		"... or all but pv to one"
	cm at: pv+1 put: 1 - (cm at: pv+1).
	slice := Form extent: width@1.
	copyBlt := (BitBlt current destForm: slice sourceForm: self
				halftoneForm: nil combinationRule: Form over
				destOrigin: 0@0 sourceOrigin: 0@0 extent: slice width @ 1
				clipRect: slice boundingBox) colorMap: cm.
	countBlt := (BitBlt current toForm: slice)
				fillColor: (Bitmap with: 0);
				destRect: (0@0 extent: slice extent);
				combinationRule: 32.
	^ (0 to: height-1) collect:
		[:y |
		copyBlt sourceOrigin: 0@y; copyBits.
		countBlt copyBits]! !

!Form methodsFor: 'transitions' stamp: 'di 1/28/1999 09:20'!
zoomIn: goingIn orOutTo: otherImage at: topLeft vanishingPoint: vp 
	"Display zoomInTo: (Form fromDisplay: (40@40 extent: 300@300)) reverse at: 40@40.
	Display zoomOutTo: (Form fromDisplay: (40@40 extent: 300@300)) reverse at: 40@40."
	| nSteps j bigR lilR minTime startTime lead |
	nSteps := 16.
	minTime := 500.  "milliseconds"
	startTime := Time millisecondClockValue.
	^ self wipeImage: otherImage at: topLeft clippingBox: nil rectForIndex:
		[:i | "i runs from 1 to nsteps"
		i > nSteps
			ifTrue: [nil "indicates all done"]
			ifFalse:
			["If we are going too fast, delay for a bit"
			lead := startTime + (i-1*minTime//nSteps) - Time millisecondClockValue.
			lead > 10 ifTrue: [(Delay forMilliseconds: lead) wait].

			"Return an array with the difference rectangles for this step."
			j := goingIn ifTrue: [i] ifFalse: [nSteps+1-i].
			bigR := vp - (vp*(j)//nSteps) corner:
				vp + (otherImage extent-vp*(j)//nSteps).
			lilR := vp - (vp*(j-1)//nSteps) corner:
				vp + (otherImage extent-vp*(j-1)//nSteps).
			bigR areasOutside: lilR]]! !

!Form methodsFor: 'transitions' stamp: 'di 3/2/98 09:14'!
zoomInTo: otherImage at: topLeft
	"Display zoomInTo: (Form fromDisplay: (40@40 extent: 300@300)) reverse at: 40@40"
	^ self zoomIn: true orOutTo: otherImage at: topLeft
		vanishingPoint: otherImage extent//2+topLeft! !

!Form methodsFor: 'transitions' stamp: 'di 3/2/98 09:15'!
zoomOutTo: otherImage at: topLeft
	"Display zoomOutTo: (Form fromDisplay: (40@40 extent: 300@300)) reverse at: 40@40"
	^ self zoomIn: false orOutTo: otherImage at: topLeft
		vanishingPoint: otherImage extent//2+topLeft! !

Form subclass: #FormStub
	instanceVariableNames: 'locator'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Display Objects'!

!FormStub methodsFor: 'accessing' stamp: 'ar 2/24/2001 22:02'!
locator
	^locator! !

!FormStub methodsFor: 'accessing' stamp: 'ar 2/24/2001 22:02'!
locator: aString
	locator := aString! !

!FormStub methodsFor: 'fileIn/Out' stamp: 'ar 2/27/2001 21:36'!
objectForDataStream: refStream
	"Force me into outPointers so that I get notified about startup"
	refStream replace: self with: self.
	^self! !

Form subclass: #StaticForm
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Display Objects'!

!StaticForm commentStamp: '<historical>' prior: 0!
An optimization for Nebraska - a StaticForm does not change once created so it may be cached on the remote end.!

!StaticForm methodsFor: 'as yet unclassified' stamp: 'RAA 8/14/2000 09:59'!
isStatic

	^true! !

!DisplayObject class methodsFor: 'fileIn/Out' stamp: 'mdr 8/31/2000 19:11'!
collectionFromFileNamed: fileName 
	"Answer a collection of Forms read from the external file 
	named fileName. The file format is: fileCode, {depth, extent, offset, bits}."

	| formList f fileCode |
	formList := OrderedCollection new.
	f := (FileStream readOnlyFileNamed: fileName) binary.
	fileCode := f next.
	fileCode = 1
		ifTrue: [
			[f atEnd] whileFalse: [formList add: (self new readFromOldFormat: f)]]
		ifFalse: [
			fileCode = 2 ifFalse: [self error: 'unknown Form file format'. ^ formList].
			[f atEnd] whileFalse: [formList add: (self new readFrom: f)]].
	f close.
	^ formList
! !

!DisplayObject class methodsFor: 'fileIn/Out' stamp: ''!
writeCollection: coll onFileNamed: fileName 
	"Saves a collection of Forms on the file fileName in the format:
		fileCode, {depth, extent, offset, bits}."
	| file |
	file := FileStream newFileNamed: fileName.
	file binary.
	file nextPut: 2.  "file code = 2"
	coll do: [:f | f writeOn: file].
	file close
"
 | f c | c := OrderedCollection new.
[(f := Form fromUser) boundingBox area>25] whileTrue: [c add: f].
Form writeCollection: c onFileNamed: 'test.forms'.
c := Form collectionFromFileNamed: 'test.forms'.
1 to: c size do: [:i | (c at: i) displayAt: 0@(i*100)].
"! !

!DisplayObject methodsFor: 'transforming' stamp: ''!
align: alignmentPoint with: relativePoint 
	"Translate the receiver's offset such that alignmentPoint aligns with 
	relativePoint."

	self offset: (self offset translateBy: relativePoint - alignmentPoint)! !

!DisplayObject methodsFor: 'display box access' stamp: ''!
boundingBox
	"Answer the rectangular area that represents the boundaries of the 
	receiver's space of information."

	^self computeBoundingBox! !

!DisplayObject methodsFor: 'display box access' stamp: ''!
center

	^ self boundingBox center! !

!DisplayObject methodsFor: 'display box access' stamp: ''!
computeBoundingBox
	"Answer the rectangular area that represents the boundaries of the 
	receiver's area for displaying information. This is the primitive for 
	computing the area if it is not already known."

	self subclassResponsibility! !

!DisplayObject methodsFor: 'displaying-Display' stamp: ''!
display 
	"Display the receiver on the Display at location 0,0."

	self displayOn: Display! !

!DisplayObject methodsFor: 'displaying-generic' stamp: ''!
displayAt: aDisplayPoint 
	"Display the receiver located at aDisplayPoint with default settings for 
	the displayMedium, rule and halftone."

	self displayOn: Display
		at: aDisplayPoint
		clippingBox: Display boundingBox
		rule: Form over
		fillColor: nil! !

!DisplayObject methodsFor: 'displaying-generic' stamp: ''!
displayOn: aDisplayMedium
	"Simple default display in order to see the receiver in the upper left 
	corner of screen."

	self displayOn: aDisplayMedium at: 0 @ 0! !

!DisplayObject methodsFor: 'displaying-generic' stamp: ''!
displayOn: aDisplayMedium at: aDisplayPoint 
	"Display the receiver located at aDisplayPoint with default settings for 
	rule and halftone."

	self displayOn: aDisplayMedium
		at: aDisplayPoint
		clippingBox: aDisplayMedium boundingBox
		rule: Form over
		fillColor: nil! !

!DisplayObject methodsFor: 'displaying-generic' stamp: ''!
displayOn: aDisplayMedium at: aDisplayPoint clippingBox: clipRectangle 
	"Display the receiver located at aDisplayPoint with default settings for 
	rule and halftone. Information to be displayed must be confined to the 
	area that intersects with clipRectangle."

	self displayOn: aDisplayMedium
		at: aDisplayPoint
		clippingBox: clipRectangle
		rule: Form over
		fillColor: nil! !

!DisplayObject methodsFor: 'displaying-generic' stamp: ''!
displayOn: aDisplayMedium at: aDisplayPoint clippingBox: clipRectangle rule: ruleInteger fillColor: aForm
	"This is the basic display primitive for graphic display objects. Display 
	the receiver located at aDisplayPoint with rule, ruleInteger, and mask, 
	aForm. Information to be displayed must be confined to the area that 
	intersects with clipRectangle."

	self subclassResponsibility! !

!DisplayObject methodsFor: 'displaying-generic' stamp: ''!
displayOn: aDisplayMedium at: aDisplayPoint rule: ruleInteger
	"Display the receiver located at aPoint with default setting for the 
	halftone and clippingBox."

	self displayOn: aDisplayMedium
		at: aDisplayPoint
		clippingBox: aDisplayMedium boundingBox
		rule: ruleInteger
		fillColor: nil! !

!DisplayObject methodsFor: 'displaying-generic' stamp: ''!
displayOn: aDisplayMedium transformation: displayTransformation clippingBox: clipRectangle 
	"Display primitive for the receiver where a DisplayTransformation is 
	provided as an argument. Alignment is defaulted to the receiver's 
	rectangle. Information to be displayed must be confined to the area that 
	intersects with clipRectangle."

	self displayOn: aDisplayMedium
		transformation: displayTransformation
		clippingBox: clipRectangle
		align: self relativeRectangle center
		with: self relativeRectangle center
		rule: Form over
		fillColor: nil! !

!DisplayObject methodsFor: 'displaying-generic' stamp: ''!
displayOn: aDisplayMedium transformation: displayTransformation clippingBox: clipRectangle align: alignmentPoint with: relativePoint 
	"Display primitive where a DisplayTransformation is provided as an 
	argument, rule is over and mask is Form black. Information to be 
	displayed must be confined to the area that intersects with clipRectangle."

	self displayOn: aDisplayMedium
		transformation: displayTransformation
		clippingBox: clipRectangle
		align: alignmentPoint
		with: relativePoint
		rule: Form over
		fillColor: nil! !

!DisplayObject methodsFor: 'displaying-generic' stamp: ''!
displayOn: aDisplayMedium transformation: displayTransformation clippingBox: clipRectangle align: alignmentPoint with: relativePoint rule: ruleInteger fillColor: aForm 
	"Display the receiver where a DisplayTransformation is provided as an 
	argument, rule is ruleInteger and mask is aForm. Translate by 
	relativePoint-alignmentPoint. Information to be displayed must be 
	confined to the area that intersects with clipRectangle."

	| absolutePoint |
	absolutePoint := displayTransformation applyTo: relativePoint.
	self displayOn: aDisplayMedium
		at: (absolutePoint - alignmentPoint) 
		clippingBox: clipRectangle 
		rule: ruleInteger 
		fillColor: aForm ! !

!DisplayObject methodsFor: 'displaying-generic' stamp: ''!
displayOn: aDisplayMedium transformation: displayTransformation clippingBox: clipRectangle fixedPoint: aPoint 
	"Display the receiver where a DisplayTransformation is provided as an 
	argument, rule is over and mask is Form black. No translation. 
	Information to be displayed must be confined to the area that intersects 
	with clipRectangle."

	self displayOn: aDisplayMedium
		transformation: displayTransformation
		clippingBox: clipRectangle
		align: aPoint
		with: aPoint
		rule: Form over
		fillColor: nil! !

!DisplayObject methodsFor: 'displaying-generic' stamp: ''!
displayOn: aDisplayMedium transformation: displayTransformation clippingBox: clipRectangle rule: ruleInteger fillColor: aForm 
	"Display the receiver where a DisplayTransformation is provided as an 
	argument, rule is ruleInteger and mask is aForm. No translation. 
	Information to be displayed must be confined to the area that intersects 
	with clipRectangle."

	self displayOn: aDisplayMedium
		transformation: displayTransformation
		clippingBox: clipRectangle
		align: self relativeRectangle origin
		with: self relativeRectangle origin
		rule: ruleInteger
		fillColor: aForm! !

!DisplayObject methodsFor: 'displaying-generic' stamp: ''!
displayOnPort: aPort 
	self displayOnPort: aPort at: 0@0! !

!DisplayObject methodsFor: 'displaying-generic' stamp: 'jm 10/21/97 16:56'!
displayOnPort: port at: location rule: rule

	port copyForm: self to: location rule: rule.
! !

!DisplayObject methodsFor: 'accessing' stamp: ''!
extent
	"Answer the point that represents the width and height of the receiver's 
	bounding box."

	^self boundingBox extent! !

!DisplayObject methodsFor: 'displaying-Display' stamp: ''!
follow: locationBlock while: durationBlock
   "Move an image around on the Display. Restore the background
   continuously without causing flashing. The argument, locationBlock,
   supplies each new location, and the argument, durationBlock, supplies
   true to continue, and then false to stop.
   8/20/96 sw: call follow:while:bitsBehind: to do the real work.  Note that th
method
   now returns the final bits behind as method value."
 
   | bitsBehind loc |
   bitsBehind := Form fromDisplay: ((loc := locationBlock value) extent: self extent).
   ^ self follow: locationBlock while: durationBlock bitsBehind: bitsBehind startingLoc: loc! !

!DisplayObject methodsFor: 'displaying-Display' stamp: 'ar 5/28/2000 12:06'!
follow: locationBlock while: durationBlock bitsBehind: initialBitsBehind startingLoc: loc
   "Move an image around on the Display. Restore the background continuously without causing flashing. The argument, locationBlock, supplies each new location, and the argument, durationBlock, supplies true to continue or false to stop. This variant takes the bitsBehind as an input argument, and returns the final saved saved bits as method value."

   | location rect1 save1 save1Blt buffer bufferBlt newLoc rect2 bothRects |
   location := loc.
   rect1 := location extent: self extent.
   save1 := initialBitsBehind.
   save1Blt := BitBlt current toForm: save1.
   buffer := Form extent: self extent*2 depth: Display depth.  "Holds overlapping region"
   bufferBlt := BitBlt current toForm: buffer.
   Display deferUpdates: true.
   self displayOn: Display at: location rule: Form paint.
   Display deferUpdates: false; forceToScreen: (location extent: self extent).
   [durationBlock value] whileTrue: [
		newLoc := locationBlock value.
		newLoc ~= location ifTrue: [
			rect2 := newLoc extent: self extent.
			bothRects := rect1 merge: rect2.
			(rect1 intersects: rect2)
				ifTrue: [  "when overlap, buffer background for both rectangles"
					bufferBlt copyFrom: bothRects in: Display to: 0@0.
					bufferBlt copyFrom: save1 boundingBox in: save1 to: rect1 origin - bothRects origin.
					"now buffer is clean background; get new bits for save1"
					save1Blt copy: (0@0 extent: self extent) from: rect2 origin - bothRects origin in: buffer.
					self displayOnPort: bufferBlt at: rect2 origin - bothRects origin rule: Form paint.
					Display deferUpdates: true.
					Display copy: bothRects from: 0@0 in: buffer rule: Form over.
					Display deferUpdates: false; forceToScreen: bothRects]
				ifFalse: [  "when no overlap, do the simple thing (both rects might be too big)"
					Display deferUpdates: true.
					Display copy: (location extent: save1 extent) from: 0@0 in: save1 rule: Form over.
					save1Blt copyFrom: rect2 in: Display to: 0@0.
					self displayOn: Display at: newLoc rule: Form paint.
					Display deferUpdates: false; 
						forceToScreen: (location extent: save1 extent); 
						forceToScreen: (newLoc extent: self extent)].
			location := newLoc.
			rect1 := rect2]].

	^ save1 displayOn: Display at: location
! !

!DisplayObject methodsFor: 'displaying-generic' stamp: ''!
followCursor
	"Just show the Form following the mouse. 6/21/96 tk"
	Cursor blank showWhile:
		[self follow: [Sensor cursorPoint] while: [Sensor noButtonPressed]]
! !

!DisplayObject methodsFor: 'accessing' stamp: ''!
height
	"Answer the number that represents the height of the receiver's 
	bounding box."

	^self boundingBox height! !

!DisplayObject methodsFor: 'display box access' stamp: ''!
initialExtent
	"Included here for when a FormView is being opened
	as a window.  (4@4) covers border widths."

	^ self extent + (4@4) ! !

!DisplayObject methodsFor: 'displaying-Display' stamp: 'di 9/12/97 11:09'!
isTransparent
	^ false! !

!DisplayObject methodsFor: 'accessing' stamp: ''!
offset
	"Answer the amount by which the receiver should be offset when it is 
	displayed or its position is tested."

	self subclassResponsibility! !

!DisplayObject methodsFor: 'accessing' stamp: ''!
offset: aPoint 
	"Set the amount by which the receiver's position is offset."

	^self! !

!DisplayObject methodsFor: 'accessing' stamp: ''!
relativeRectangle
	"Answer a Rectangle whose top left corner is the receiver's offset position 
	and whose width and height are the same as the receiver."

	^Rectangle origin: self offset extent: self extent! !

!DisplayObject methodsFor: 'truncation and round off' stamp: ''!
rounded
	"Convert the offset of the receiver to integer coordinates."

	self offset: self offset rounded! !

!DisplayObject methodsFor: 'transforming' stamp: ''!
scaleBy: aPoint 
	"Scale the receiver's offset by aPoint."

	self offset: (self offset scaleBy: aPoint)! !

!DisplayObject methodsFor: 'displaying-Display' stamp: ''!
slideFrom: startPoint to: stopPoint nSteps: nSteps 
	"does not display at the first point, but does at the last"
	| i p delta |
	i:=0.  p:= startPoint.
	delta := (stopPoint-startPoint) // nSteps.
	^ self follow: [p:= p+delta]
		while: [(i:=i+1) < nSteps]! !

!DisplayObject methodsFor: 'displaying-Display' stamp: 'jm 10/22/97 07:43'!
slideFrom: startPoint to: stopPoint nSteps: nSteps delay: milliSecs
	"Slide this object across the display over the given number of steps, pausing for the given number of milliseconds after each step."
	"Note: Does not display at the first point, but does at the last."

	| i p delta |
	i := 0.
	p := startPoint.
	delta := (stopPoint - startPoint) / nSteps asFloat.
	^ self
		follow: [(p := p + delta) truncated]
		while: [
			(Delay forMilliseconds: milliSecs) wait.
			(i := i + 1) < nSteps]
! !

!DisplayObject methodsFor: 'displaying-Display' stamp: 'di 10/19/97 12:05'!
slideFrom: startPoint to: stopPoint nSteps: nSteps delay: milliSecs andStay: stayAtEnd
	"Does not display at the first point, but does at the last.
	Moreover, if stayAtEnd is true, it leaves the dragged image at the stopPoint"
	| i done |
	i := 0.
	^ self follow: [startPoint + ((stopPoint-startPoint) * i // nSteps)]
		while: [milliSecs ifNotNil: [(Delay forMilliseconds: milliSecs) wait].
				((done := (i := i+1) > nSteps) and: [stayAtEnd])
					ifTrue: [^ self "Return without clearing the image"].
				done not]! !

!DisplayObject methodsFor: 'displaying-Display' stamp: 'sr 6/6/2000 05:37'!
slideWithFirstFrom: startPoint to: stopPoint nSteps: nSteps delay: milliSecs 
	"Slide this object across the display over the given number of steps, 
	pausing for the given number of milliseconds after each step."
	"Note: Does display at the first point and at the last."
	| i p delta |
	i := 0.
	delta := stopPoint - startPoint / nSteps asFloat.
	p := startPoint - delta.
	^ self follow: [(p := p + delta) truncated]
		while: 
			[(Delay forMilliseconds: milliSecs) wait.
			(i := i + 1) <= nSteps]! !

!DisplayObject methodsFor: 'transforming' stamp: ''!
translateBy: aPoint 
	"Translate the receiver's offset."

	self offset: (self offset translateBy: aPoint)! !

!DisplayObject methodsFor: 'accessing' stamp: ''!
width
	"Answer the number that represents the width of the receiver's bounding 
	box."

	^self boundingBox width! !

!DisplayObject methodsFor: 'fileIn/Out' stamp: ''!
writeOnFileNamed: fileName 
	"Saves the receiver on the file fileName in the format:
		fileCode, depth, extent, offset, bits."
	| file |
	file := FileStream newFileNamed: fileName.
	file binary.
	file nextPut: 2.  "file code = 2"
	self writeOn: file.
	file close
"
 | f |
[(f := Form fromUser) boundingBox area>25] whileTrue:
	[f writeOnFileNamed: 'test.form'.
	(Form newFromFileNamed: 'test.form') display].
"! !

!DisplayObject methodsFor: 'fileIn/Out' stamp: 'tk 2/19/1999 07:20'!
writeUncompressedOnFileNamed: fileName 
	"Saves the receiver on the file fileName in the format:
		fileCode, depth, extent, offset, bits."
	| file |
	file := FileStream newFileNamed: fileName.
	file binary.
	file nextPut: 2.  "file code = 2"
	self writeUncompressedOn: file.
	file close
"
 | f |
[(f := Form fromUser) boundingBox area>25] whileTrue:
	[f writeUncompressedOnFileNamed: 'test.form'.
	(Form fromBinaryStream: (FileStream oldFileNamed: 'test.form')) display].
"! !

DisplayObject subclass: #InfiniteForm
	instanceVariableNames: 'patternForm'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Display Objects'!

!InfiniteForm commentStamp: '<historical>' prior: 0!
I represent a Form obtained by replicating a pattern form indefinitely in all directions.!

!InfiniteForm class methodsFor: 'instance creation' stamp: ''!
with: aForm 
	"Answer an instance of me whose pattern form is the argument, aForm."

	^self new form: aForm! !

!InfiniteForm methodsFor: 'as yet unclassified' stamp: 'RAA 6/1/2000 10:50'!
addFillStyleMenuItems: aMenu hand: aHand from: aMorph
	"Add the items for changing the current fill style of the receiver"

	"prevents a walkback when control menu is built for morph with me as color"! !

!InfiniteForm methodsFor: 'accessing' stamp: 'mjg 7/9/2001 14:12'!
asColor
	^ patternForm dominantColor! !

!InfiniteForm methodsFor: 'accessing' stamp: ''!
asForm
	^ patternForm! !

!InfiniteForm methodsFor: 'fillstyle protocol' stamp: 'bolot 9/15/1999 10:13'!
bitPatternForDepth: suspectedDepth
	^ patternForm! !

!InfiniteForm methodsFor: 'displaying' stamp: 'sw 2/16/98 03:42'!
colorForInsets
	^ Color transparent! !

!InfiniteForm methodsFor: 'display box access' stamp: ''!
computeBoundingBox 
	"Refer to the comment in DisplayObject|computeBoundingBox."

	^0 @ 0 corner: SmallInteger maxVal @ SmallInteger maxVal! !

!InfiniteForm methodsFor: 'fillstyle protocol' stamp: 'ar 7/2/1999 14:56'!
direction
	^patternForm width @ 0! !

!InfiniteForm methodsFor: 'displaying' stamp: 'nk 4/17/2004 19:48'!
displayOn: aDisplayMedium at: aDisplayPoint clippingBox: clipRectangle rule: ruleInteger fillColor: aForm
	"This is the real display message, but it doesn't get used until the new
	display protocol is installed."
	| targetBox patternBox bb |
	(patternForm isForm) ifFalse:
		[^ aDisplayMedium fill: clipRectangle rule: ruleInteger fillColor: patternForm].

	"Do it iteratively"
	targetBox := aDisplayMedium boundingBox intersect: clipRectangle.
	patternBox := patternForm boundingBox.
	bb := BitBlt current destForm: aDisplayMedium sourceForm: patternForm fillColor: aForm
		combinationRule: ruleInteger destOrigin: 0@0 sourceOrigin: 0@0
		extent: patternBox extent clipRect: clipRectangle.
	bb colorMap:
		(patternForm colormapIfNeededFor: aDisplayMedium).
	(targetBox left truncateTo: patternBox width)
		to: targetBox right - 1 by: patternBox width do:
		[:x |
		(targetBox top truncateTo: patternBox height)
			to: targetBox bottom - 1 by: patternBox height do:
			[:y |
			bb destOrigin: x@y; copyBits]]! !

!InfiniteForm methodsFor: 'displaying' stamp: 'nk 4/17/2004 19:48'!
displayOnPort: aPort at: offset

	| targetBox patternBox savedMap top left |

	self flag: #bob.

	"this *may* not get called at the moment. I have been trying to figure out the right way for this to work and am using #displayOnPort:offsetBy: as my current offering - Bob"

	(patternForm isForm) ifFalse: [
		"patternForm is a Pattern or Color; just use it as a mask for BitBlt"
		^ aPort fill: aPort clipRect fillColor: patternForm rule: Form over].

	"do it iteratively"
	targetBox := aPort clipRect.
	patternBox := patternForm boundingBox.
	savedMap := aPort colorMap.
	aPort sourceForm: patternForm;
		fillColor: nil;
		combinationRule: Form paint;
		sourceRect: (0@0 extent: patternBox extent);
		colorMap: (patternForm colormapIfNeededFor: aPort destForm).
	top := (targetBox top truncateTo: patternBox height) "- (offset y \\ patternBox height)".
	left :=  (targetBox left truncateTo: patternBox width) "- (offset x \\ patternBox width)".
	left to: (targetBox right - 1) by: patternBox width do:
		[:x | top to: (targetBox bottom - 1) by: patternBox height do:
			[:y | aPort destOrigin: x@y; copyBits]].
	aPort colorMap: savedMap.
! !

!InfiniteForm methodsFor: 'displaying' stamp: 'nk 4/17/2004 19:48'!
displayOnPort: aPort offsetBy: offset

	| targetBox patternBox savedMap top left |

	"this version tries to get the form aligned where the user wants it and not just aligned with the cliprect"

	(patternForm isForm) ifFalse: [
		"patternForm is a Pattern or Color; just use it as a mask for BitBlt"
		^ aPort fill: aPort clipRect fillColor: patternForm rule: Form over].

	"do it iteratively"
	targetBox := aPort clipRect.
	patternBox := patternForm boundingBox.
	savedMap := aPort colorMap.
	aPort sourceForm: patternForm;
		fillColor: nil;
		combinationRule: Form paint;
		sourceRect: (0@0 extent: patternBox extent);
		colorMap: (patternForm colormapIfNeededFor: aPort destForm).
	top := (targetBox top truncateTo: patternBox height) + offset y.
	left :=  (targetBox left truncateTo: patternBox width) + offset x.

	left to: (targetBox right - 1) by: patternBox width do:
		[:x | top to: (targetBox bottom - 1) by: patternBox height do:
			[:y | aPort destOrigin: x@y; copyBits]].
	aPort colorMap: savedMap.
! !

!InfiniteForm methodsFor: 'accessing' stamp: 'di 9/2/97 20:21'!
dominantColor
	^ patternForm dominantColor! !

!InfiniteForm methodsFor: 'fillstyle protocol' stamp: 'ar 7/2/1999 14:54'!
form
	"Bitmap fills respond to #form"
	^patternForm! !

!InfiniteForm methodsFor: 'private' stamp: ''!
form: aForm

	patternForm := aForm! !

!InfiniteForm methodsFor: 'fillstyle protocol' stamp: 'ar 7/2/1999 14:54'!
isBitmapFill
	^true! !

!InfiniteForm methodsFor: 'fillstyle protocol' stamp: 'ar 7/2/1999 14:54'!
isGradientFill
	^false! !

!InfiniteForm methodsFor: 'fillstyle protocol' stamp: 'ar 7/2/1999 14:59'!
isOrientedFill
	^true! !

!InfiniteForm methodsFor: 'fillstyle protocol' stamp: 'ar 7/2/1999 14:54'!
isSolidFill
	^false! !

!InfiniteForm methodsFor: 'fillstyle protocol' stamp: 'ar 7/2/1999 14:55'!
isTiled
	"Return true if the receiver should be drawn as a tiled pattern"
	^true! !

!InfiniteForm methodsFor: 'fillstyle protocol' stamp: 'ar 9/2/1999 14:32'!
isTranslucent
	"Return true since the bitmap may be translucent and we don't really want to check"
	^true! !

!InfiniteForm methodsFor: 'fillstyle protocol' stamp: 'ar 7/2/1999 14:57'!
normal
	^0 @ patternForm height! !

!InfiniteForm methodsFor: 'accessing' stamp: ''!
offset 
	"Refer to the comment in DisplayObject|offset."

	^0 @ 0! !

!InfiniteForm methodsFor: 'fillstyle protocol' stamp: 'ar 7/2/1999 14:56'!
origin
	^0@0! !

!InfiniteForm methodsFor: 'fillstyle protocol' stamp: 'ar 7/2/1999 14:56'!
origin: aPoint
	"Ignored"
! !

!InfiniteForm methodsFor: 'displaying' stamp: 'ar 8/16/2001 12:47'!
raisedColor
	^ Color transparent! !

Object subclass: #DisplayTransform
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Transformations'!

!DisplayTransform commentStamp: '<historical>' prior: 0!
This class represents a base for generic transformations of 2D points between different coordinate systems (including scaling and rotation). The transformations map objects between one coordinate system and another where it is assumed that a nested hierarchy of transformations can be defined.

It is assumed that transformations deal with Integer points. All transformations should return Integer coordinates (even though float points may be passed in as argument).

Compositions of transformations MUST work in the following order. A 'global' transformation (the argument in #composedWithGlobal:) is defined as a transformation that takes place between the receiver (the 'local') transformation and any 'global' point computations, whereas a 'local' transformation (e.g., the argument in #composedWithLocal:) takes place between the receiver ('global') and any 'local' points. For the transformation methods this means that combining a global and a local transformation will result in the following order:

		globalPointToLocal: globalPoint
			"globalPoint -> globalTransform -> localTransform -> locaPoint"
			^localTransform globalPointToLocal:
				(globalTransform globalPointToLocal: globalPoint)

		localPointToGlobal: localPoint
			"localPoint -> localTransform -> globalTransform -> globalPoint"
			^globalTransform localPointToGlobal:
				(localTransform localPointToGlobal: localPoint)

!

DisplayTransform subclass: #CompositeTransform
	instanceVariableNames: 'globalTransform localTransform'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Transformations'!

!CompositeTransform commentStamp: '<historical>' prior: 0!
A composite transform provides the effect of several levels of coordinate transformations.!

!CompositeTransform class methodsFor: 'instance creation' stamp: 'ls 3/19/2000 16:44'!
globalTransform: gt localTransform: lt
	^self new globalTransform: gt localTransform: lt! !

!CompositeTransform methodsFor: 'accessing' stamp: 'di 10/26/1999 17:06'!
angle
	^ localTransform angle + globalTransform angle! !

!CompositeTransform methodsFor: 'converting' stamp: 'ar 11/2/1998 20:00'!
asCompositeTransform
	^self! !

!CompositeTransform methodsFor: 'converting' stamp: 'ar 11/2/1998 19:56'!
asMatrixTransform2x3
	^globalTransform asMatrixTransform2x3
		composedWithLocal: localTransform asMatrixTransform2x3! !

!CompositeTransform methodsFor: 'converting' stamp: 'di 10/26/1999 17:03'!
asMorphicTransform
	"Squash a composite transform down to a simple one"
	^ MorphicTransform offset: self offset angle: self angle scale: self scale! !

!CompositeTransform methodsFor: 'initialization' stamp: 'di 10/26/1999 17:08'!
composedWith: aTransform
	"Return a new transform that has the effect of transforming points first by the receiver and then by the argument."

	self isIdentity ifTrue: [^ aTransform].
	aTransform isIdentity ifTrue: [^ self].
	^ CompositeTransform new globalTransform: self
							localTransform: aTransform! !

!CompositeTransform methodsFor: 'transforming points' stamp: 'ar 11/2/1998 16:39'!
globalPointToLocal: aPoint
	"Transform aPoint from global coordinates into local coordinates"
	^localTransform globalPointToLocal:
		(globalTransform globalPointToLocal: aPoint)! !

!CompositeTransform methodsFor: 'initialization' stamp: 'di 3/4/98 19:17'!
globalTransform: gt localTransform: lt
	globalTransform := gt.
	localTransform := lt! !

!CompositeTransform methodsFor: 'accessing' stamp: 'ar 11/2/1998 19:45'!
inverseTransformation
	"Return the inverse transformation of the receiver"
	^self species new
		globalTransform: localTransform inverseTransformation
		localTransform: globalTransform inverseTransformation! !

!CompositeTransform methodsFor: 'transformations' stamp: 'di 10/1/1998 13:51'!
invert: aPoint
	^ globalTransform invert: (localTransform invert: aPoint)! !

!CompositeTransform methodsFor: 'testing' stamp: 'ar 11/2/1998 20:00'!
isCompositeTransform
	^true! !

!CompositeTransform methodsFor: 'testing' stamp: 'di 3/4/98 19:18'!
isIdentity
	^ globalTransform isIdentity and: [localTransform isIdentity]! !

!CompositeTransform methodsFor: 'testing' stamp: 'di 3/4/98 19:18'!
isPureTranslation
	^ globalTransform isPureTranslation and: [localTransform isPureTranslation]! !

!CompositeTransform methodsFor: 'transforming points' stamp: 'ar 11/2/1998 16:39'!
localPointToGlobal: aPoint
	"Transform aPoint from global coordinates into local coordinates"
	^globalTransform localPointToGlobal:
		(localTransform localPointToGlobal: aPoint)! !

!CompositeTransform methodsFor: 'accessing' stamp: 'di 10/26/1999 15:40'!
offset
	^ (self localPointToGlobal: 0@0) negated! !

!CompositeTransform methodsFor: 'accessing' stamp: 'di 10/26/1999 15:39'!
scale
	^ localTransform scale * globalTransform scale! !

!CompositeTransform methodsFor: 'transformations' stamp: 'di 3/4/98 19:20'!
transform: aPoint
	^ localTransform transform: (globalTransform transform: aPoint)! !

!DisplayTransform class methodsFor: 'instance creation' stamp: 'ar 11/2/1998 20:55'!
identity
	^self new setIdentity! !

!DisplayTransform methodsFor: 'converting' stamp: 'ar 11/2/1998 19:59'!
asCompositeTransform
	"Represent the receiver as a composite transformation"
	^CompositeTransform new
		globalTransform: self
		localTransform: self species identity! !

!DisplayTransform methodsFor: 'converting' stamp: 'ar 11/2/1998 20:01'!
asMatrixTransform2x3
	"Represent the receiver as a 2x3 matrix transformation"
	^self subclassResponsibility! !

!DisplayTransform methodsFor: 'composing' stamp: 'ar 11/2/1998 16:15'!
composedWithGlobal: aTransformation
	"Return the composition of the receiver and the global transformation passed in.
	A 'global' transformation is defined as a transformation that takes place
	between the receiver (the 'local') transformation and any 'global' point
	computations, e.g., for the methods

		globalPointToLocal: globalPoint
			globalPoint -> globalTransform -> localTransform -> locaPoint

		localPointToGlobal: localPoint
			localPoint -> localTransform -> globalTransform -> globalPoint

		"
	^aTransformation composedWithLocal: self! !

!DisplayTransform methodsFor: 'composing' stamp: 'ar 11/2/1998 16:41'!
composedWithLocal: aTransformation
	"Return the composition of the receiver and the local transformation passed in.
	A 'local' transformation is defined as a transformation that takes place
	between the receiver (the 'global') transformation and any 'local' point
	computations, e.g., for the methods

		globalPointToLocal: globalPoint
			globalPoint -> globalTransform -> localTransform -> locaPoint

		localPointToGlobal: localPoint
			localPoint -> localTransform -> globalTransform -> globalPoint

		"
	self isIdentity ifTrue:[^ aTransformation].
	aTransformation isIdentity ifTrue:[^ self].
	^ CompositeTransform new globalTransform: self
							localTransform: aTransformation! !

!DisplayTransform methodsFor: 'transforming rects' stamp: 'ar 11/2/1998 16:19'!
globalBoundsToLocal: aRectangle
	"Transform aRectangle from global coordinates into local coordinates"
	^Rectangle encompassing: (self globalPointsToLocal: aRectangle corners)! !

!DisplayTransform methodsFor: 'transforming points' stamp: 'ar 11/2/1998 16:17'!
globalPointToLocal: aPoint
	"Transform aPoint from global coordinates into local coordinates"
	^self subclassResponsibility! !

!DisplayTransform methodsFor: 'transforming points' stamp: 'ar 11/9/1998 14:35'!
globalPointsToLocal: inArray
	"Transform all the points of inArray from global into local coordinates"
	^inArray collect:[:pt| self globalPointToLocal: pt]! !

!DisplayTransform methodsFor: 'accessing' stamp: 'ar 11/2/1998 19:43'!
inverseTransformation
	"Return the inverse transformation of the receiver"
	^self subclassResponsibility! !

!DisplayTransform methodsFor: 'transforming points' stamp: 'gh 10/22/2001 13:24'!
invertBoundsRect: aRectangle
	"Return a rectangle whose coordinates have been transformed
	from local back to global coordinates."

	^self subclassResponsibility! !

!DisplayTransform methodsFor: 'testing' stamp: 'ar 11/2/1998 22:47'!
isCompositeTransform
	"Return true if the receiver is a composite transformation.
	Composite transformations may have impact on the accuracy."
	^false! !

!DisplayTransform methodsFor: 'testing' stamp: 'ar 11/2/1998 16:17'!
isIdentity
	"Return true if the receiver is the identity transform; that is, if applying to a point returns the point itself."
	^self subclassResponsibility! !

!DisplayTransform methodsFor: 'testing' stamp: 'ar 11/2/1998 22:48'!
isMatrixTransform2x3
	"Return true if the receiver is 2x3 matrix transformation"
	^false! !

!DisplayTransform methodsFor: 'testing' stamp: 'ar 11/2/1998 22:48'!
isMorphicTransform
	"Return true if the receiver is a MorphicTransform, that is specifies the transformation values explicitly."
	^false! !

!DisplayTransform methodsFor: 'testing' stamp: 'ar 11/2/1998 16:16'!
isPureTranslation
	"Return true if the receiver specifies no rotation or scaling."
	^self subclassResponsibility! !

!DisplayTransform methodsFor: 'transforming rects' stamp: 'ar 11/2/1998 16:19'!
localBoundsToGlobal: aRectangle
	"Transform aRectangle from local coordinates into global coordinates"
	^Rectangle encompassing: (self localPointsToGlobal: aRectangle corners)! !

!DisplayTransform methodsFor: 'transforming points' stamp: 'ar 11/2/1998 16:18'!
localPointToGlobal: aPoint
	"Transform aPoint from local coordinates into global coordinates"
	^self subclassResponsibility! !

!DisplayTransform methodsFor: 'transforming points' stamp: 'ar 11/9/1998 14:35'!
localPointsToGlobal: inArray
	"Transform all the points of inArray from local into global coordinates"
	^inArray collect:[:pt| self localPointToGlobal: pt]! !

!DisplayTransform methodsFor: 'initialize' stamp: 'ar 11/2/1998 23:18'!
setIdentity
	"Initialize the receiver to the identity transformation (e.g., not affecting points)"
	^self subclassResponsibility! !

!DisplayTransform methodsFor: 'transforming rects' stamp: 'di 10/25/1999 12:49'!
sourceQuadFor: aRectangle
	^ aRectangle innerCorners collect: 
		[:p | self globalPointToLocal: p]! !

DisplayTransform subclass: #IdentityTransform
	instanceVariableNames: ''
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'Graphics-Transformations'!

!IdentityTransform class methodsFor: 'class initialization' stamp: 'ar 9/11/2000 21:24'!
initialize
	"IdentityTransform initialize"
	Default := self basicNew.! !

!IdentityTransform class methodsFor: 'instance creation' stamp: 'ar 9/11/2000 21:24'!
new
	"There can be only one"
	^Default! !

!IdentityTransform methodsFor: 'accessing' stamp: 'di 9/29/2000 09:04'!
angle
	^ 0.0! !

!IdentityTransform methodsFor: 'converting' stamp: 'ar 9/11/2000 21:21'!
asMatrixTransform2x3
	"Represent the receiver as a 2x3 matrix transformation"
	^MatrixTransform2x3 identity! !

!IdentityTransform methodsFor: 'composing' stamp: 'ar 9/11/2000 21:27'!
composedWith: aTransform
	^aTransform! !

!IdentityTransform methodsFor: 'composing' stamp: 'ar 9/11/2000 21:19'!
composedWithGlobal: aTransformation
	^aTransformation! !

!IdentityTransform methodsFor: 'composing' stamp: 'ar 9/11/2000 21:19'!
composedWithLocal: aTransformation
	^aTransformation! !

!IdentityTransform methodsFor: 'transforming rects' stamp: 'ar 9/11/2000 21:20'!
globalBoundsToLocal: aRectangle
	"Transform aRectangle from global coordinates into local coordinates"
	^aRectangle! !

!IdentityTransform methodsFor: 'transforming points' stamp: 'ar 9/11/2000 21:19'!
globalPointToLocal: aPoint
	"Transform aPoint from global coordinates into local coordinates"
	^aPoint! !

!IdentityTransform methodsFor: 'transforming points' stamp: 'ar 9/11/2000 21:20'!
globalPointsToLocal: inArray
	"Transform all the points of inArray from global into local coordinates"
	^inArray! !

!IdentityTransform methodsFor: 'accessing' stamp: 'ar 9/11/2000 21:18'!
inverseTransformation
	"Return the inverse transformation of the receiver"
	^self! !

!IdentityTransform methodsFor: 'transforming points' stamp: 'gh 10/22/2001 13:24'!
invertBoundsRect: aRectangle
	"Return a rectangle whose coordinates have been transformed
	from local back to global coordinates. Since I am the identity matrix
	no transformation is made."

	^aRectangle
! !

!IdentityTransform methodsFor: 'testing' stamp: 'ar 9/11/2000 21:18'!
isIdentity
	"Return true if the receiver is the identity transform; that is, if applying to a point returns the point itself."
	^true! !

!IdentityTransform methodsFor: 'testing' stamp: 'ar 9/11/2000 21:19'!
isPureTranslation
	"Return true if the receiver specifies no rotation or scaling."
	^true! !

!IdentityTransform methodsFor: 'transforming rects' stamp: 'ar 9/11/2000 21:20'!
localBoundsToGlobal: aRectangle
	"Transform aRectangle from local coordinates into global coordinates"
	^aRectangle! !

!IdentityTransform methodsFor: 'transforming points' stamp: 'ar 9/11/2000 21:20'!
localPointToGlobal: aPoint
	"Transform aPoint from local coordinates into global coordinates"
	^aPoint! !

!IdentityTransform methodsFor: 'transforming points' stamp: 'ar 9/11/2000 21:20'!
localPointsToGlobal: inArray
	"Transform all the points of inArray from local into global coordinates"
	^inArray! !

!IdentityTransform methodsFor: 'accessing' stamp: 'ar 4/19/2001 06:01'!
offset
	^0@0! !

!IdentityTransform methodsFor: 'initialize' stamp: 'ar 9/11/2000 21:18'!
setIdentity
	"I *am* the identity transform"
	^self! !

!IdentityTransform methodsFor: 'transforming rects' stamp: 'ar 9/11/2000 21:21'!
sourceQuadFor: aRectangle
	^ aRectangle innerCorners! !

DisplayTransform variableWordSubclass: #MatrixTransform2x3
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Transformations'!

!MatrixTransform2x3 commentStamp: '<historical>' prior: 0!
This class represents a transformation for points, that is a combination of scale, offset, and rotation. It is implemented as a 2x3 matrix containing the transformation from the local coordinate system in the global coordinate system. Thus, transforming points from local to global coordinates is fast and cheap whereas transformations from global to local coordinate systems are relatively expensive.

Implementation Note: It is assumed that the transformation deals with Integer points. All transformations will return Integer coordinates (even though float points may be passed in here).!

!MatrixTransform2x3 class methodsFor: 'instance creation' stamp: 'ar 11/2/1998 22:50'!
identity
	^self new setScale: 1.0! !

!MatrixTransform2x3 class methodsFor: 'instance creation' stamp: 'ar 7/9/1998 20:09'!
new
	^self new: 6! !

!MatrixTransform2x3 class methodsFor: 'instance creation' stamp: 'mir 6/12/2001 15:34'!
newFromStream: s
	"Only meant for my subclasses that are raw bits and word-like.  For quick unpack form the disk."
	self isPointers | self isWords not ifTrue: [^ super newFromStream: s].
		"super may cause an error, but will not be called."
	^ s nextWordsInto: (self new: 6)! !

!MatrixTransform2x3 class methodsFor: 'instance creation' stamp: 'ar 11/12/1998 01:25'!
transformFromLocal: localBounds toGlobal: globalBounds
	^((self withOffset: (globalBounds center)) composedWithLocal:
		(self withScale: (globalBounds extent / localBounds extent) asFloatPoint))
			composedWithLocal: (self withOffset: localBounds center negated)
"
	^(self identity)
		setScale: (globalBounds extent / localBounds extent) asFloatPoint;
		setOffset: localBounds center negated asFloatPoint;
		composedWithGlobal:(self withOffset: globalBounds center asFloatPoint)
"! !

!MatrixTransform2x3 class methodsFor: 'instance creation' stamp: 'ar 11/2/1998 02:49'!
withAngle: angle
	^self new setAngle: angle! !

!MatrixTransform2x3 class methodsFor: 'instance creation' stamp: 'ar 11/3/1998 02:52'!
withOffset: aPoint
	^self identity setOffset: aPoint! !

!MatrixTransform2x3 class methodsFor: 'instance creation' stamp: 'ar 11/2/1998 23:17'!
withRotation: angle
	^self new setAngle: angle! !

!MatrixTransform2x3 class methodsFor: 'instance creation' stamp: 'ar 11/2/1998 02:49'!
withScale: aPoint
	^self new setScale: aPoint! !

!MatrixTransform2x3 methodsFor: 'comparing' stamp: 'ar 2/2/2001 15:47'!
= aMatrixTransform2x3 
	| length |
	<primitive: 'primitiveEqual' module: 'FloatArrayPlugin'>
	self class = aMatrixTransform2x3 class ifFalse: [^ false].
	length := self size.
	length = aMatrixTransform2x3 size ifFalse: [^ false].
	1 to: self size do: [:i | (self at: i)
			= (aMatrixTransform2x3 at: i) ifFalse: [^ false]].
	^ true! !

!MatrixTransform2x3 methodsFor: 'element access' stamp: 'ar 11/2/1998 22:56'!
a11
	^self at: 1! !

!MatrixTransform2x3 methodsFor: 'element access' stamp: 'ar 11/2/1998 22:57'!
a11: value
	 self at: 1 put: value! !

!MatrixTransform2x3 methodsFor: 'element access' stamp: 'ar 11/2/1998 22:56'!
a12
	^self at: 2! !

!MatrixTransform2x3 methodsFor: 'element access' stamp: 'ar 11/2/1998 22:57'!
a12: value
	 self at: 2 put: value! !

!MatrixTransform2x3 methodsFor: 'element access' stamp: 'ar 11/2/1998 22:56'!
a13
	^self at: 3! !

!MatrixTransform2x3 methodsFor: 'element access' stamp: 'ar 11/2/1998 22:57'!
a13: value
	 self at: 3 put: value! !

!MatrixTransform2x3 methodsFor: 'element access' stamp: 'ar 11/2/1998 22:56'!
a21
	 ^self at: 4! !

!MatrixTransform2x3 methodsFor: 'element access' stamp: 'ar 11/2/1998 22:57'!
a21: value
	 self at: 4 put: value! !

!MatrixTransform2x3 methodsFor: 'element access' stamp: 'ar 11/2/1998 22:56'!
a22
	 ^self at: 5! !

!MatrixTransform2x3 methodsFor: 'element access' stamp: 'ar 11/2/1998 22:57'!
a22: value
	 self at: 5 put: value! !

!MatrixTransform2x3 methodsFor: 'element access' stamp: 'ar 11/2/1998 22:56'!
a23
	 ^self at: 6! !

!MatrixTransform2x3 methodsFor: 'element access' stamp: 'ar 11/2/1998 22:57'!
a23: value
	 self at: 6 put: value! !

!MatrixTransform2x3 methodsFor: 'converting' stamp: 'ar 11/2/1998 15:34'!
asMatrixTransform2x3
	^self! !

!MatrixTransform2x3 methodsFor: 'accessing' stamp: 'ar 2/2/2001 15:47'!
at: index
	<primitive: 'primitiveAt' module: 'FloatArrayPlugin'>
	^Float fromIEEE32Bit: (self basicAt: index)! !

!MatrixTransform2x3 methodsFor: 'accessing' stamp: 'ar 2/2/2001 15:47'!
at: index put: value
	<primitive: 'primitiveAtPut' module: 'FloatArrayPlugin'>
	value isFloat 
		ifTrue:[self basicAt: index put: value asIEEE32BitWord]
		ifFalse:[self at: index put: value asFloat].
	^value! !

!MatrixTransform2x3 methodsFor: 'objects from disk' stamp: 'nk 3/17/2004 16:06'!
byteSize
	^self basicSize * self bytesPerBasicElement! !

!MatrixTransform2x3 methodsFor: 'objects from disk' stamp: 'nk 3/17/2004 15:04'!
bytesPerBasicElement
	"Answer the number of bytes that each of my basic elements requires.
	In other words:
		self basicSize * self bytesPerBasicElement
	should equal the space required on disk by my variable sized representation."
	^4! !

!MatrixTransform2x3 methodsFor: 'objects from disk' stamp: 'yo 3/6/2004 12:57'!
bytesPerElement

	^ 4.
! !

!MatrixTransform2x3 methodsFor: 'composing' stamp: 'RAA 9/20/2000 13:10'!
composedWithLocal: aTransformation
	"Return the composition of the receiver and the local transformation passed in"
	aTransformation isMatrixTransform2x3 ifFalse:[^super composedWithLocal: aTransformation].
	^self composedWithLocal: aTransformation asMatrixTransform2x3 into: self class new! !

!MatrixTransform2x3 methodsFor: 'composing' stamp: 'ar 2/2/2001 15:47'!
composedWithLocal: aTransformation into: result
	"Return the composition of the receiver and the local transformation passed in.
	Store the composed matrix into result."
	| a11 a12 a13 a21 a22 a23 b11 b12 b13 b21 b22 b23 matrix |
	<primitive: 'primitiveComposeMatrix' module: 'Matrix2x3Plugin'>
	matrix := aTransformation asMatrixTransform2x3.
	a11 := self a11.		b11 := matrix a11.
	a12 := self a12.		b12 := matrix a12.
	a13 := self a13.		b13 := matrix a13.
	a21 := self a21.		b21 := matrix a21.
	a22 := self a22.		b22 := matrix a22.
	a23 := self a23.		b23 := matrix a23.
	result a11: (a11 * b11) + (a12 * b21).
	result a12: (a11 * b12) + (a12 * b22).
	result a13: a13 + (a11 * b13) + (a12 * b23).
	result a21: (a21 * b11) + (a22 * b21).
	result a22: (a21 * b12) + (a22 * b22).
	result a23: a23 + (a21 * b13) + (a22 * b23).
	^result! !

!MatrixTransform2x3 methodsFor: 'transforming rects' stamp: 'ar 2/2/2001 15:47'!
globalBounds: srcRect toLocal: dstRect
	"Transform aRectangle from global coordinates into local coordinates"
	<primitive: 'primitiveInvertRectInto' module: 'Matrix2x3Plugin'>
	^super globalBoundsToLocal: srcRect! !

!MatrixTransform2x3 methodsFor: 'transforming rects' stamp: 'ar 11/9/1998 14:40'!
globalBoundsToLocal: aRectangle
	"Transform aRectangle from global coordinates into local coordinates"
	^self globalBounds: aRectangle toLocal: Rectangle new! !

!MatrixTransform2x3 methodsFor: 'transforming points' stamp: 'ar 2/2/2001 15:47'!
globalPointToLocal: aPoint
	"Transform aPoint from global coordinates into local coordinates"
	<primitive: 'primitiveInvertPoint' module: 'Matrix2x3Plugin'>
	^(self invertPoint: aPoint) rounded! !

!MatrixTransform2x3 methodsFor: 'comparing' stamp: 'ar 5/3/2001 13:02'!
hash
	| result |
	<primitive: 'primitiveHashArray' module: 'FloatArrayPlugin'>
	result := 0.
	1 to: self size do:[:i| result := result + (self basicAt: i) ].
	^result bitAnd: 16r1FFFFFFF! !

!MatrixTransform2x3 methodsFor: 'accessing' stamp: 'ar 11/18/1998 14:03'!
inverseTransformation
	"Return the inverse transformation of the receiver.
	The inverse transformation is computed by first calculating
	the inverse offset and then computing transformations
	for the two identity vectors (1@0) and (0@1)"
	| r1 r2 r3 m |
	r3 := self invertPoint: 0@0.
	r1 := (self invertPoint: 1@0) - r3.
	r2 := (self invertPoint: 0@1) - r3.
	m := self species new.
	m
		a11: r1 x; a12: r2 x; a13: r3 x;
		a21: r1 y; a22: r2 y; a23: r3 y.
	^m! !

!MatrixTransform2x3 methodsFor: 'transforming points' stamp: 'ar 11/16/1998 23:46'!
invertPoint: aPoint
	"Transform aPoint from global coordinates into local coordinates"
	| x y det a11 a12 a21 a22 detX detY |
	x := aPoint x asFloat - (self a13).
	y := aPoint y asFloat - (self a23).
	a11 := self a11.	a12 := self a12.
	a21 := self a21.	a22 := self a22.
	det := (a11 * a22) - (a12 * a21).
	det = 0.0 ifTrue:[^0@0]. "So we have at least a valid result"
	det := 1.0 / det.
	detX := (x * a22) - (a12 * y).
	detY := (a11 * y) - (x * a21).
	^(detX * det) @ (detY * det)! !

!MatrixTransform2x3 methodsFor: 'testing' stamp: 'ar 2/2/2001 15:47'!
isIdentity
	"Return true if the receiver is the identity transform; that is, if applying to a point returns the point itself."
	<primitive: 'primitiveIsIdentity' module: 'Matrix2x3Plugin'>
	^self isPureTranslation and:[self a13 = 0.0 and:[self a23 = 0.0]]! !

!MatrixTransform2x3 methodsFor: 'testing' stamp: 'ar 11/2/1998 23:15'!
isMatrixTransform2x3
	"Return true if the receiver is 2x3 matrix transformation"
	^true! !

!MatrixTransform2x3 methodsFor: 'testing' stamp: 'ar 2/2/2001 15:47'!
isPureTranslation
	"Return true if the receiver specifies no rotation or scaling."
	<primitive: 'primitiveIsPureTranslation' module: 'Matrix2x3Plugin'>
	^self a11 = 1.0 and:[self a12 = 0.0 and:[self a22 = 0.0 and:[self a21 = 1.0]]]! !

!MatrixTransform2x3 methodsFor: 'transforming rects' stamp: 'ar 2/2/2001 15:47'!
localBounds: srcRect toGlobal: dstRect
	"Transform aRectangle from local coordinates into global coordinates"
	<primitive: 'primitiveTransformRectInto' module: 'Matrix2x3Plugin'>
	^super localBoundsToGlobal: srcRect! !

!MatrixTransform2x3 methodsFor: 'transforming rects' stamp: 'ar 11/9/1998 14:40'!
localBoundsToGlobal: aRectangle
	"Transform aRectangle from local coordinates into global coordinates"
	^self localBounds: aRectangle toGlobal: Rectangle new! !

!MatrixTransform2x3 methodsFor: 'transforming points' stamp: 'ar 2/2/2001 15:47'!
localPointToGlobal: aPoint
	"Transform aPoint from local coordinates into global coordinates"
	<primitive: 'primitiveTransformPoint' module: 'Matrix2x3Plugin'>
	^(self transformPoint: aPoint) rounded! !

!MatrixTransform2x3 methodsFor: 'accessing' stamp: 'ar 11/2/1998 23:19'!
offset
	^self a13 @ self a23! !

!MatrixTransform2x3 methodsFor: 'accessing' stamp: 'ar 11/2/1998 23:05'!
offset: aPoint
	self a13: aPoint x asFloat.
	self a23: aPoint y asFloat.! !

!MatrixTransform2x3 methodsFor: 'printing' stamp: 'ar 11/2/1998 23:11'!
printOn: aStream
	aStream 
		nextPutAll: self class name;
		nextPut: $(;
		cr; print: self a11; tab; print: self a12; tab; print: self a13;
		cr; print: self a21; tab; print: self a22; tab; print: self a23;
		cr; nextPut:$).! !

!MatrixTransform2x3 methodsFor: 'objects from disk' stamp: 'yo 3/6/2004 15:33'!
restoreEndianness
	"This word object was just read in from a stream.  It was stored in Big Endian (Mac) format.  Swap each pair of bytes (16-bit word), if the current machine is Little Endian.
	Why is this the right thing to do?  We are using memory as a byteStream.  High and low bytes are reversed in each 16-bit word, but the stream of words ascends through memory.  Different from a Bitmap."

	| w b1 b2 b3 b4 |
	SmalltalkImage current  isLittleEndian ifTrue: [
		1 to: self basicSize do: [:i |
			w := self basicAt: i.
			b1 := w digitAt: 1.
			b2 := w digitAt: 2.
			b3 := w digitAt: 3.
			b4 := w digitAt: 4.
			w := (b1 << 24) + (b2 << 16) + (b3 << 8) + b4.
			self basicAt: i put: w.
		]
	].

! !

!MatrixTransform2x3 methodsFor: 'private' stamp: 'ar 11/2/1998 23:17'!
setAngle: angle
	"Set the raw rotation angle in the receiver"
	| rad s c |
	rad := angle degreesToRadians.
	s := rad sin.
	c := rad cos.
	self a11: c.
	self a12: s negated.
	self a21: s.
	self a22: c.! !

!MatrixTransform2x3 methodsFor: 'initialize' stamp: 'ar 11/2/1998 23:17'!
setIdentiy
	"Initialize the receiver to the identity transformation (e.g., not affecting points)"
	self
		a11: 1.0; a12: 0.0; a13: 0.0;
		a21: 0.0; a22: 1.0; a23: 0.0.! !

!MatrixTransform2x3 methodsFor: 'private' stamp: 'ar 11/2/1998 23:17'!
setOffset: aPoint
	"Set the raw offset in the receiver"
	| pt |
	pt := aPoint asPoint.
	self a13: pt x asFloat.
	self a23: pt y asFloat.! !

!MatrixTransform2x3 methodsFor: 'private' stamp: 'ar 11/2/1998 23:16'!
setScale: aPoint
	"Set the raw scale in the receiver"
	| pt |
	pt := aPoint asPoint.
	self a11: pt x asFloat.
	self a22: pt y asFloat.! !

!MatrixTransform2x3 methodsFor: 'transforming points' stamp: 'ar 8/26/2001 20:54'!
transformDirection: aPoint
	"Transform aPoint from local coordinates into global coordinates"
	| x y |
	x := (aPoint x * self a11) + (aPoint y * self a12).
	y := (aPoint x * self a21) + (aPoint y * self a22).
	^x @ y! !

!MatrixTransform2x3 methodsFor: 'transforming points' stamp: 'ar 11/2/1998 23:09'!
transformPoint: aPoint
	"Transform aPoint from local coordinates into global coordinates"
	| x y |
	x := (aPoint x * self a11) + (aPoint y * self a12) + self a13.
	y := (aPoint x * self a21) + (aPoint y * self a22) + self a23.
	^x @ y! !

!MatrixTransform2x3 methodsFor: 'objects from disk' stamp: 'ar 8/6/2001 17:52'!
writeOn: aStream
	aStream nextWordsPutAll: self.! !

DisplayTransform subclass: #MorphicTransform
	instanceVariableNames: 'offset angle scale'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Transformations'!

!MorphicTransform commentStamp: '<historical>' prior: 0!
This class implements simple translation, scaling and rotation for points, as well as inverse transformations.  These transformations are used in TransformMorphs (clipping scrollers) and TransformationMorphs (general flex-morph wrappers) to map, eg, global mouse coords into local coords, and to invert, eg, local damage rectangles into global damage rectangles.!

!MorphicTransform class methodsFor: 'instance creation' stamp: ''!
identity

	^ self offset: 0@0 angle: 0.0 scale: 1.0! !

!MorphicTransform class methodsFor: 'instance creation' stamp: ''!
new

	^ self offset: 0@0
! !

!MorphicTransform class methodsFor: 'instance creation' stamp: ''!
offset: aPoint

	^ self offset: aPoint angle: 0.0 scale: 1.0! !

!MorphicTransform class methodsFor: 'instance creation' stamp: ''!
offset: aPoint angle: a scale: s

	^ self basicNew setOffset: aPoint angle: a scale: s! !

!MorphicTransform methodsFor: 'accessing' stamp: ''!
angle
	^ angle! !

!MorphicTransform methodsFor: 'converting' stamp: 'ar 11/2/1998 20:14'!
asMatrixTransform2x3
	^((MatrixTransform2x3 withRotation: angle radiansToDegrees negated) composedWithLocal:
		(MatrixTransform2x3 withScale: scale))
			offset: offset negated! !

!MorphicTransform methodsFor: 'converting' stamp: 'di 10/26/1999 17:03'!
asMorphicTransform

	^ self! !

!MorphicTransform methodsFor: 'transformations' stamp: 'di 3/4/98 19:10'!
composedWith: aTransform
	"Return a new transform that has the effect of transforming points first by the receiver and then by the argument."

	self isIdentity ifTrue: [^ aTransform].
	aTransform isIdentity ifTrue: [^ self].
	^ CompositeTransform new globalTransform: self
							localTransform: aTransform! !

!MorphicTransform methodsFor: 'composing' stamp: 'nk 3/9/2001 13:55'!
composedWithLocal: aTransform
	aTransform isIdentity ifTrue:[^self].
	self isIdentity ifTrue:[^aTransform].
	aTransform isMorphicTransform ifFalse:[^super composedWithLocal: aTransform].
	self isPureTranslation ifTrue:[
		^aTransform withOffset: aTransform offset + self offset].
	aTransform isPureTranslation ifTrue:[
		^self withOffset: (self localPointToGlobal: aTransform offset negated) negated].
	^super composedWithLocal: aTransform.! !

!MorphicTransform methodsFor: 'transforming points' stamp: 'ar 11/2/1998 16:13'!
globalPointToLocal: aPoint
	"Transform aPoint from global coordinates into local coordinates"
	^self transform: aPoint! !

!MorphicTransform methodsFor: 'accessing' stamp: 'ar 11/9/1998 14:33'!
inverseTransformation
	"Return the inverse transformation of the receiver"
	^MorphicTransform
		offset: (self transform: 0@0) - (self transform: offset)
		angle: angle negated
		scale: scale reciprocal! !

!MorphicTransform methodsFor: 'transformations' stamp: 'di 10/28/1999 09:10'!
invert: aPoint
	"Transform the given point from local to global coordinates."
	| p3 p2 |
	self isPureTranslation ifTrue: [^ aPoint - offset].
	p3 :=  aPoint * scale.
	p2 := ((p3 x * angle cos) + (p3 y * angle sin))
		@ ((p3 y * angle cos) - (p3 x * angle sin)).
	^ (p2 - offset)
! !

!MorphicTransform methodsFor: 'transformations' stamp: 'di 10/3/1998 00:18'!
invertBoundsRect: aRectangle
	"Return a rectangle whose coordinates have been transformed
	from local back to global coordinates.  NOTE: if the transformation
	is not just a translation, then it will compute the bounding box
	in global coordinates."
	| outerRect |
	self isPureTranslation
	ifTrue:
		[^ (self invert: aRectangle topLeft)
			corner: (self invert: aRectangle bottomRight)]
	ifFalse:
		[outerRect := Rectangle encompassing:
			(aRectangle innerCorners collect: [:p | self invert: p]).
		"Following asymmetry due to likely subsequent truncation"
		^ outerRect topLeft - (1@1) corner: outerRect bottomRight + (2@2)]! !

!MorphicTransform methodsFor: 'transformations' stamp: 'di 10/2/1998 08:54'!
invertRect: aRectangle

	self error: 'method name changed to emphasize enclosing bounds'.
	^ self invertBoundsRect: aRectangle! !

!MorphicTransform methodsFor: 'testing' stamp: 'ar 11/2/1998 20:57'!
isIdentity
	"Return true if the receiver is the identity transform; that is, if applying to a point returns the point itself."

	^ self isPureTranslation and: [offset = (0@0)]
! !

!MorphicTransform methodsFor: 'testing' stamp: 'ar 11/2/1998 19:51'!
isMorphicTransform
	^true! !

!MorphicTransform methodsFor: 'testing' stamp: 'ar 11/2/1998 20:57'!
isPureTranslation
	"Return true if the receiver specifies no rotation or scaling."

	^ angle = 0.0 and: [scale = 1.0]
! !

!MorphicTransform methodsFor: 'transforming points' stamp: 'ar 11/2/1998 16:32'!
localPointToGlobal: aPoint
	"Transform aPoint from global coordinates into local coordinates"
	^self invert: aPoint! !

!MorphicTransform methodsFor: 'accessing' stamp: ''!
offset
	^ offset
! !

!MorphicTransform methodsFor: 'printing' stamp: 'ar 5/19/1999 18:21'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut:$(;
		nextPutAll:'angle = '; print: angle;
		nextPutAll:'; scale = '; print: scale;
		nextPutAll:'; offset = '; print: offset;
		nextPut:$).! !

!MorphicTransform methodsFor: 'accessing' stamp: ''!
scale
	^ scale! !

!MorphicTransform methodsFor: 'private' stamp: ''!
setAngle: aFloat

	angle := aFloat.
! !

!MorphicTransform methodsFor: 'initialize' stamp: 'ar 11/2/1998 20:58'!
setIdentiy
	scale := 1.0.
	offset := 0@0.
	angle := 0.0.! !

!MorphicTransform methodsFor: 'private' stamp: ''!
setOffset: aPoint

	offset := aPoint.
! !

!MorphicTransform methodsFor: 'private' stamp: ''!
setOffset: aPoint angle: a scale: s

	offset := aPoint.
	angle := a.
	scale := s! !

!MorphicTransform methodsFor: 'private' stamp: ''!
setScale: aFloat

	scale := aFloat.
! !

!MorphicTransform methodsFor: 'transformations' stamp: 'di 10/28/1999 09:05'!
transform: aPoint
	"Transform the given point from global to local coordinates."
	| p2 p3 |
	self isPureTranslation ifTrue: [^ aPoint + offset].
	p2 := aPoint + offset.
	p3 := (((p2 x * angle cos) - (p2 y * angle sin))
		@ ((p2 y * angle cos) + (p2 x * angle sin)))
			/ scale.
	^ p3! !

!MorphicTransform methodsFor: 'transformations' stamp: 'di 10/3/1998 00:18'!
transformBoundsRect: aRectangle
	"Return a rectangle whose coordinates have been transformed
	from global to local coordinates.  NOTE: if the transformation
	is not just a translation, then it will compute the bounding box
	in global coordinates."
	| outerRect |
	self isPureTranslation
	ifTrue:
		[^ (self transform: aRectangle topLeft)
			corner: (self transform: aRectangle bottomRight)]
	ifFalse:
		[outerRect := Rectangle encompassing:
			(aRectangle innerCorners collect: [:p | self transform: p]).
		"Following asymmetry due to likely subsequent truncation"
		^ outerRect topLeft - (1@1) corner: outerRect bottomRight + (2@2)]! !

!MorphicTransform methodsFor: 'accessing' stamp: ''!
withAngle: a
	"Return a copy of me with a different Angle"
	^ self copy setAngle: a! !

!MorphicTransform methodsFor: 'accessing' stamp: ''!
withOffset: a
	"Return a copy of me with a different Offset"
	^ self copy setOffset: a! !

!MorphicTransform methodsFor: 'accessing' stamp: ''!
withScale: a
	"Return a copy of me with a different Scale"
	^ self copy setScale: a! !

Object subclass: #FontSet
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Fonts'!

!FontSet commentStamp: '<historical>' prior: 0!
FontSet provides a mechanism for storing a set of fonts as a class that can be conveniently filedOut, filedIn, and installed as a TextStyle.

The most common use is...
	Find a font you like.
	Use BitFont to convert a bunch of sizes to data files named, eg, LovelyNN.BF
	Use FontSet convertFontsNamed: 'Lovely' to produce a FontSet named Lovely.
	FileOut that FontSet for later use.
	Use Lovely installAsTextStyle to make all sizes available in a TextStyle
		named #Lovely in the TextConstants dictionary.
	Use ctrl-k in any text pane to select the new Lovely style for that paragraph.
	Then use cmd-1 through 5 or cmd-k to set the point-size for any selection.
!

!FontSet class methodsFor: 'compiling' stamp: 'sma 12/29/1999 11:48'!
acceptsLoggingOfCompilation
	"Dont log sources for my subclasses, so as not to waste time
	and space storing printString versions of the string literals."

	^ self == FontSet! !

!FontSet class methodsFor: 'compiling' stamp: 'di 1/24/2005 12:40'!
compileFont: strikeFont 
	| tempName literalString header sizeStr familyName |
	tempName := 'FontTemp.sf2'.
	strikeFont writeAsStrike2named: tempName.
	literalString := (Base64MimeConverter mimeEncode: (FileStream readOnlyFileNamed: tempName) binary) contents fullPrintString.
	sizeStr := strikeFont pointSize asString.
	familyName := strikeFont name first: (strikeFont name findLast: [:x | x isDigit not]).

	header := 'size' , sizeStr , '
	^ self fontNamed: ''' , familyName , sizeStr , ''' fromMimeLiteral:
' .
	self class
		compile: header , literalString
		classified: 'fonts'
		notifying: nil.
	FileDirectory default deleteFileNamed: tempName
! !

!FontSet class methodsFor: 'as yet unclassified' stamp: 'di 9/15/97 12:01'!
convertFontsNamed: familyName  "FontSet convertFontsNamed: 'Palatino' "
	^ self convertFontsNamed: familyName inDirectoryNamed: ''! !

!FontSet class methodsFor: 'converting' stamp: 'sma 12/29/1999 21:18'!
convertFontsNamed: familyName inDirectoryNamed: dirName
		"FontSet convertFontsNamed: 'Tekton' inDirectoryNamed: 'Tekton Fonts' "
	"This utility is for use after you have used BitFont to produce data files 
	for the fonts you wish to use.  It will read the BitFont files and build
	a fontset class from them.  If one already exists, the sizes that can be
	found will be overwritten."
	"For this utility to work as is, the BitFont data files must be named 'familyNN.BF',
	and must reside in the directory named by dirName (use '' for the current directory)."

	| allFontNames fontSet dir |
	"Check first for matching file names and usable FontSet class name."
	dir := dirName isEmpty
		ifTrue: [FileDirectory default]
		ifFalse: [FileDirectory default directoryNamed: dirName].
	allFontNames := dir fileNamesMatching: familyName , '##.BF'.
	allFontNames isEmpty ifTrue: [^ self error: 'No files found like ' , familyName , 'NN.BF'].
	fontSet := self fontSetClass: familyName.
	allFontNames do:
		[:each |
		Transcript cr; show: each.
		fontSet compileFont: (StrikeFont new readFromBitFont: (dir fullNameFor: each))]! !

!FontSet class methodsFor: 'converting' stamp: 'sma 12/29/1999 12:27'!
convertTextStyleNamed: aString
	| style fontSet |
	(style := TextStyle named: aString) ifNil: [^ self error: 'unknown text style ' , aString].
	fontSet := self fontSetClass: aString.
	style fontArray do: [:each | fontSet compileFont: each]! !

!FontSet class methodsFor: 'filein/out' stamp: 'sma 12/29/1999 11:49'!
fileOut
	"FileOut and then change the properties of the file so that it won't be
	treated as text by, eg, email attachment facilities"

	super fileOut.
	(FileStream oldFileNamed: self name , '.st') setFileTypeToObject; close! !

!FontSet class methodsFor: 'private' stamp: 'sma 12/29/1999 12:58'!
fontCategory
	^ 'Graphics-Fonts' asSymbol! !

!FontSet class methodsFor: 'private' stamp: 'RAA 6/20/2000 13:29'!
fontName

	self flag: #bob.		"temporary hack until I figure out what's happening here"
	(self name beginsWith: superclass name) ifFalse: [^self name].
	^ (self name copyFrom: superclass name size + 1 to: self name size) asSymbol! !

!FontSet class methodsFor: 'installing' stamp: 'di 1/24/2005 12:48'!
fontNamed: fontName fromLiteral: aString
	"NOTE -- THIS IS AN OBSOLETE METHOD THAT MAY CAUSE ERRORS.

The old form of fileOut for FontSets produced binary literal strings which may not be accurately read in systems with support for international character sets.  If possible, file the FontSet out again from a system that produces the newer MIME encoding (current def of compileFont:), and uses the corresponding altered version of this method.  If this is not easy, then
	file the fontSet into an older system (3.7 or earlier),
	assume it is called FontSetZork...
	execute FontSetZork installAsTextStyle.
	copy the compileFont: method from this system into that older one.
	remove the class FontSetZork.
	Execute:  FontSet convertTextStyleNamed: 'Zork', and see that it creates a new FontSetZork.
	FileOut the new class FontSetZork.
	The resulting file should be able to be read into this system.
"

	^ StrikeFont new 
		name: fontName;
		readFromStrike2Stream: (ReadStream on: aString asByteArray)! !

!FontSet class methodsFor: 'installing' stamp: 'di 1/24/2005 11:13'!
fontNamed: fontName fromMimeLiteral: aString
	"This method allows a font set to be captured as sourcecode in a subclass.
	The string literals will presumably be created by printing, eg,
		(FileStream readOnlyFileNamed: 'Palatino24.sf2') contentsOfEntireFile,
		and following the logic in compileFont: to encode and add a heading.

	See the method installAsTextStyle to see how this can be used."

	^ StrikeFont new 
		name: fontName;
		readFromStrike2Stream: (Base64MimeConverter mimeDecodeToBytes: aString readStream)! !

!FontSet class methodsFor: 'private' stamp: 'sma 12/29/1999 12:58'!
fontSetClass: aString
	| className fontSet |
	className := (self name , (aString select: [:c | c isAlphaNumeric]) capitalized) asSymbol.
	fontSet := Smalltalk
		at: className
		ifAbsentPut: [self
			subclass: className
			instanceVariableNames: ''
			classVariableNames: ''
			poolDictionaries: ''
			category: self fontCategory].
	(fontSet inheritsFrom: self) ifFalse: [^ self error: 'The name ' , className , ' is already in use'].
	^ fontSet! !

!FontSet class methodsFor: 'installing' stamp: 'ar 8/6/2009 18:22'!
installAsDefault  "FontSetNewYork installAsDefault"
	(UIManager default confirm: 'Do you want to install
''' , self fontName , ''' as default font?')
		ifFalse: [^ self].
	self installAsTextStyle.
	"TextConstants at: #OldDefaultTextStyle put: TextStyle default."
	TextConstants at: #DefaultTextStyle put: (TextStyle named: self fontName).
	#(ListParagraph PopUpMenu StandardSystemView) do:[:className|
		Smalltalk at: className ifPresent:[:aClass| aClass initialize].
	].! !

!FontSet class methodsFor: 'installing' stamp: 'sma 12/30/1999 15:05'!
installAsTextStyle  "FontSetNewYork installAsTextStyle"
	| selectors |
	(TextConstants includesKey: self fontName) ifTrue:
		[(self confirm: 
self fontName , ' is already defined in TextConstants.
Do you want to replace that definition?')
			ifFalse: [^ self]].
	selectors := (self class selectors select: [:s | s beginsWith: 'size']) asSortedCollection.
	TextConstants
		at: self fontName
		put: (TextStyle fontArray: (selectors collect: [:each | self perform: each]))! !

!FontSet class methodsFor: 'installing' stamp: 'nk 8/31/2004 09:23'!
size: pointSize fromLiteral: aString 
	"This method allows a font set to be captured as sourcecode in a subclass.
	The string literals will presumably be created by printing, eg,
		(FileStream readOnlyFileNamed: 'Palatino24.sf2') contentsOfEntireFile,
		and then pasting into a browser after a heading like, eg,
size24
	^ self size: 24 fromLiteral:
	'--unreadable binary data--'

	See the method installAsTextStyle to see how this can be used."

	"This method is old and for backward compatibility only.
	please use fontNamed:fromLiteral: instead."

	self flag: #bob.	"used in Alan's projects"
	^(StrikeFont new)
		name: self fontName , (pointSize < 10 
							ifTrue: ['0' , pointSize printString]
							ifFalse: [pointSize printString]);
		readFromStrike2Stream: ((RWBinaryOrTextStream with: aString)
					reset;
					binary);
		yourself! !

Object subclass: #IdentityGlyphMap
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Fonts'!

!IdentityGlyphMap methodsFor: 'as yet unclassified' stamp: 'yo 2/13/2004 04:07'!
at: index

	^ index - 1.
! !

Object subclass: #ImageReadWriter
	instanceVariableNames: 'stream'
	classVariableNames: 'ImageNotStoredSignal MagicNumberErrorSignal'
	poolDictionaries: ''
	category: 'Graphics-Files'!

!ImageReadWriter commentStamp: '<historical>' prior: 0!
Copyright (c) Kazuki Yasumatsu, 1995. All rights reserved.

I am an abstract class to provide for encoding and/or decoding an image on a stream.

Instance Variables:
	stream		<ReadStream | WriteStream>	stream for image storages

Class Variables:
	ImageNotStoredSignal		<Signal>	image not stored error signal
	MagicNumberErrorSignal		<Signal>	magic number error signal

Subclasses must implement the following messages:
	accessing
		nextImage
		nextPutImage:
	testing
		canUnderstand         (added tao 10/26/97)!

ImageReadWriter subclass: #BMPReadWriter
	instanceVariableNames: 'bfType bfSize bfOffBits biSize biWidth biHeight biPlanes biBitCount biCompression biSizeImage biXPelsPerMeter biYPelsPerMeter biClrUsed biClrImportant'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Files'!

!BMPReadWriter class methodsFor: 'testing' stamp: 'ar 6/16/2002 18:55'!
displayAllFrom: fd
	"BMPReadWriter displayAllFrom: FileDirectory default"
	fd fileNames do:[:fName|
		(fName endsWith: '.bmp') ifTrue:[
			[(Form fromBinaryStream: (fd readOnlyFileNamed: fName)) display.
			Display forceDisplayUpdate] on: Error do:[:nix|].
		].
	].
	fd directoryNames do:[:fdName|
		self displayAllFrom: (fd directoryNamed: fdName)
	].! !

!BMPReadWriter class methodsFor: 'testing' stamp: 'ar 6/16/2002 18:56'!
readAllFrom: fd
	"MessageTally spyOn:[BMPReadWriter readAllFrom: FileDirectory default]"
	fd fileNames do:[:fName|
		(fName endsWith: '.bmp') ifTrue:[
			[Form fromBinaryStream: (fd readOnlyFileNamed: fName)] on: Error do:[:nix].
		].
	].
	fd directoryNames do:[:fdName|
		self readAllFrom: (fd directoryNamed: fdName)
	].! !

!BMPReadWriter class methodsFor: 'image reading/writing' stamp: 'nk 7/16/2003 17:56'!
typicalFileExtensions
	"Answer a collection of file extensions (lowercase) which files that I can read might commonly have"
	^#('bmp')! !

!BMPReadWriter methodsFor: 'reading' stamp: 'ar 6/16/2002 15:36'!
nextImage
	| colors |
	stream binary.
	self readHeader.
	biBitCount = 24 ifTrue:[^self read24BmpFile].
	"read the color map"
	colors := self readColorMap.
	^self readIndexedBmpFile: colors! !

!BMPReadWriter methodsFor: 'writing' stamp: 'yo 2/18/2004 17:57'!
nextPutImage: aForm
	| bhSize rowBytes rgb data colorValues depth image ppw scanLineLen |
	depth := aForm depth.
	[#(1 4 8 32) includes: depth] whileFalse:[depth := depth + 1 asLargerPowerOfTwo].
	image := aForm asFormOfDepth: depth.
	image unhibernate.
	bhSize := 14.  "# bytes in file header"
	biSize := 40.  "info header size in bytes"
	biWidth := image width.
	biHeight := image height.
	biClrUsed := depth = 32 ifTrue: [0] ifFalse:[1 << depth].  "No. color table entries"
	bfOffBits := biSize + bhSize + (4*biClrUsed).
	rowBytes := ((depth min: 24) * biWidth + 31 // 32) * 4.
	biSizeImage := biHeight * rowBytes.

	"Write the file header"
	stream position: 0.
	stream nextLittleEndianNumber: 2 put: 19778.  "bfType = BM"
	stream nextLittleEndianNumber: 4 put: bfOffBits + biSizeImage.  "Entire file size in bytes"
	stream nextLittleEndianNumber: 4 put: 0.  "bfReserved"
	stream nextLittleEndianNumber: 4 put: bfOffBits.  "Offset of bitmap data from start of hdr (and file)"

	"Write the bitmap info header"
	stream position: bhSize.
	stream nextLittleEndianNumber: 4 put: biSize.  "info header size in bytes"
	stream nextLittleEndianNumber: 4 put: image width.  "biWidth"
	stream nextLittleEndianNumber: 4 put: image height.  "biHeight"
	stream nextLittleEndianNumber: 2 put: 1.  "biPlanes"
	stream nextLittleEndianNumber: 2 put: (depth min: 24).  "biBitCount"
	stream nextLittleEndianNumber: 4 put: 0.  "biCompression"
	stream nextLittleEndianNumber: 4 put: biSizeImage.  "size of image section in bytes"
	stream nextLittleEndianNumber: 4 put: 2800.  "biXPelsPerMeter"
	stream nextLittleEndianNumber: 4 put: 2800.  "biYPelsPerMeter"
	stream nextLittleEndianNumber: 4 put: biClrUsed.
	stream nextLittleEndianNumber: 4 put: 0.  "biClrImportant"
	biClrUsed > 0 ifTrue: [
		"write color map; this works for ColorForms, too"
		colorValues := image colormapIfNeededForDepth: 32.
		1 to: biClrUsed do: [:i |
			rgb := colorValues at: i.
			0 to: 24 by: 8 do: [:j | stream nextPut: (rgb >> j bitAnd: 16rFF)]]].

	depth < 32 ifTrue: [
		"depth = 1, 4 or 8."
		data := image bits asByteArray.
		ppw := 32 // depth.
		scanLineLen := biWidth + ppw - 1 // ppw * 4.  "# of bytes in line"
		1 to: biHeight do: [:i |
			stream next: scanLineLen putAll: data startingAt: (biHeight-i)*scanLineLen+1.
		].
	] ifFalse: [
		1 to: biHeight do:[:i |
			data := (image copy: (0@(biHeight-i) extent: biWidth@1)) bits.
			1 to: data size do: [:j | stream nextLittleEndianNumber: 3 put: (data at: j)].
			1 to: (data size*3)+3//4*4-(data size*3) do: [:j | stream nextPut: 0 "pad to 32-bits"]
		].
	].
	stream position = (bfOffBits + biSizeImage) ifFalse: [self error:'Write failure'].
	stream close.! !

!BMPReadWriter methodsFor: 'reading' stamp: 'ar 10/25/2005 13:48'!
read24BmpFile
	"Read 24-bit pixel data from the given a BMP stream."
	| form formBits pixelLine bitsIndex bitBlt |
	form := Form extent: biWidth@biHeight depth: 32.
	pixelLine := ByteArray new: (((24 * biWidth) + 31) // 32) * 4.
	bitsIndex := form height - 1 * biWidth + 1.
	formBits := form bits.
	1 to: biHeight do: [:i |
		pixelLine := stream nextInto: pixelLine.
		self read24BmpLine: pixelLine into: formBits startingAt: bitsIndex width: biWidth.
		bitsIndex := bitsIndex - biWidth.
	].
	bitBlt := BitBlt toForm: form.
	bitBlt combinationRule: 7 "bitOr:with:".
	bitBlt halftoneForm: (Bitmap with: 16rFF000000).
	bitBlt copyBits.
	^ form
! !

!BMPReadWriter methodsFor: 'reading' stamp: 'ar 10/25/2005 13:47'!
read24BmpLine: pixelLine into: formBits startingAt: formBitsIndex width: width
	| pixIndex rgb bitsIndex |
	pixIndex := 0. "pre-increment"
	bitsIndex := formBitsIndex-1. "pre-increment"
	1 to: width do: [:j |
		rgb := 
			(pixelLine at: (pixIndex := pixIndex+1)) +
			((pixelLine at: (pixIndex := pixIndex+1)) bitShift: 8) +
			((pixelLine at: (pixIndex := pixIndex+1)) bitShift: 16).
		formBits at: (bitsIndex := bitsIndex+1) put: rgb.
	].
! !

!BMPReadWriter methodsFor: 'reading' stamp: 'ar 10/24/2005 20:57'!
readColorMap
	"Read colorCount BMP color map entries from the given binary stream. Answer an array of Colors."
	| colorCount colors maxLevel b g r ccStream |
	colorCount := (bfOffBits - 54) // 4.
	"Note: some programs (e.g. Photoshop 4.0) apparently do not set colorCount; assume that any data between the end of the header and the start of the pixel data is the color map"
	biBitCount >= 16 ifTrue:[^nil].
	colorCount = 0 ifTrue: [ "this BMP file does not have a color map"
		"default monochrome color map"
		biBitCount = 1 ifTrue: [^ Array with: Color white with: Color black].
		"default gray-scale color map"
		maxLevel := (2 raisedTo: biBitCount) - 1.
		^ (0 to: maxLevel) collect: [:level | Color gray: (level asFloat / maxLevel)]].
	ccStream := ReadStream on: (stream next: colorCount*4).
	colors := Array new: colorCount.
	1 to: colorCount do: [:i |
		b := ccStream next.
		g := ccStream next.
		r := ccStream next.
		ccStream next. "skip reserved"
		colors at: i put: (Color r: r g: g b: b range: 255)].
	^ colors
! !

!BMPReadWriter methodsFor: 'reading' stamp: 'ar 6/16/2002 15:20'!
readHeader
	| reserved |
	bfType := stream nextLittleEndianNumber: 2.
	bfSize := stream nextLittleEndianNumber: 4.
	reserved := stream nextLittleEndianNumber: 4.
	bfOffBits := stream nextLittleEndianNumber: 4.
	biSize := stream nextLittleEndianNumber: 4.
	biWidth := stream nextLittleEndianNumber: 4.
	biHeight := stream nextLittleEndianNumber: 4.
	biPlanes := stream nextLittleEndianNumber: 2.
	biBitCount := stream nextLittleEndianNumber: 2.
	biCompression := stream nextLittleEndianNumber: 4.
	biSizeImage := stream nextLittleEndianNumber: 4.
	biXPelsPerMeter := stream nextLittleEndianNumber: 4.
	biYPelsPerMeter := stream nextLittleEndianNumber: 4.
	biClrUsed := stream nextLittleEndianNumber: 4.
	biClrImportant := stream nextLittleEndianNumber: 4.
! !

!BMPReadWriter methodsFor: 'reading' stamp: 'ar 10/25/2005 13:45'!
readIndexedBmpFile: colors
	"Read uncompressed pixel data of depth d from the given BMP stream, where d is 1, 4, 8, or 16"
	| form bytesPerRow pixelData pixelLine startIndex map bitBlt mask |
	colors 
		ifNil:[form := Form extent: biWidth@biHeight depth: biBitCount]
		ifNotNil:[form := ColorForm extent: biWidth@biHeight depth: biBitCount.
				form colors: colors].
	bytesPerRow := (((biBitCount* biWidth) + 31) // 32) * 4.
	pixelData := ByteArray new: bytesPerRow * biHeight.
	biHeight to: 1 by: -1 do: [:y |
		pixelLine := stream next: bytesPerRow.
		startIndex := ((y - 1) * bytesPerRow) + 1.
		pixelData 
			replaceFrom: startIndex 
			to: startIndex + bytesPerRow - 1 
			with: pixelLine 
			startingAt: 1].
	form bits copyFromByteArray: pixelData.
	biBitCount = 16 ifTrue:[
		map := ColorMap shifts: #(8 -8 0 0) masks: #(16rFF 16rFF00 0 0).
		mask := 16r80008000.
	].
	biBitCount = 32 ifTrue:[
		map := ColorMap shifts: #(24 8 -8 -24) masks: #(16rFF 16rFF00 16rFF0000 16rFF000000).
		mask := 16rFF000000.
	].
	map ifNotNil:[
		bitBlt := BitBlt toForm: form.
		bitBlt sourceForm: form.
		bitBlt colorMap: map.
		bitBlt combinationRule: Form over.
		bitBlt copyBits.
	].
	mask ifNotNil:[
		bitBlt := BitBlt toForm: form.
		bitBlt combinationRule: 7 "bitOr:with:".
		bitBlt halftoneForm: (Bitmap with: mask).
		bitBlt copyBits.
	].
	^ form
! !

!BMPReadWriter methodsFor: 'testing' stamp: 'ar 6/16/2002 15:27'!
understandsImageFormat
	stream size < 54 ifTrue:[^false]. "min size = BITMAPFILEHEADER+BITMAPINFOHEADER"
	self readHeader.
	bfType = 19778 "BM" ifFalse:[^false].
	biSize = 40 ifFalse:[^false].
	biPlanes = 1 ifFalse:[^false].
	bfSize <= stream size ifFalse:[^false].
	biCompression = 0 ifFalse:[^false].
	^true! !

ImageReadWriter subclass: #GIFReadWriter
	instanceVariableNames: 'width height bitsPerPixel colorPalette rowByteSize xpos ypos pass interlace codeSize clearCode eoiCode freeCode maxCode prefixTable suffixTable remainBitCount bufByte bufStream transparentIndex mapOf32 localColorTable delay loopCount offset'
	classVariableNames: 'Extension ImageSeparator Terminator'
	poolDictionaries: ''
	category: 'Graphics-Files'!

!GIFReadWriter commentStamp: '<historical>' prior: 0!
Copyright (c) Kazuki Yasumatsu, 1995. All rights reserved.

Used with permission.  Modified for use in Squeak.!

GIFReadWriter subclass: #AnimatedGIFReadWriter
	instanceVariableNames: 'forms delays comments'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Files'!

!AnimatedGIFReadWriter class methodsFor: 'image reading/writing' stamp: 'mir 11/18/2003 17:00'!
formsFromFileNamed: fileName 
	| stream |
	stream := FileStream readOnlyFileNamed: fileName.
	^ self formsFromStream: stream! !

!AnimatedGIFReadWriter class methodsFor: 'image reading/writing' stamp: 'mir 11/18/2003 17:00'!
formsFromStream: stream 
	| reader |
	reader := self new on: stream reset.
	Cursor read
		showWhile: [reader allImages.
			reader close].
	^reader! !

!AnimatedGIFReadWriter class methodsFor: 'image reading/writing' stamp: 'nk 6/12/2004 13:12'!
typicalFileExtensions
	"Answer a collection of file extensions (lowercase) which files that I can read might commonly have"
	^#('gif')! !

!AnimatedGIFReadWriter class methodsFor: 'image reading/writing' stamp: 'asm 12/11/2003 21:29'!
wantsToHandleGIFs
	^true! !

!AnimatedGIFReadWriter methodsFor: 'accessing' stamp: 'bf 2/25/2005 11:11'!
allImages
	| body colorTable |
	stream class == ReadWriteStream ifFalse: [
		stream binary.
		self on: (ReadWriteStream with: (stream contentsOfEntireFile))].
	localColorTable := nil.
	forms := OrderedCollection new.
	delays := OrderedCollection new.
	comments := OrderedCollection new.
	self readHeader.
	[(body := self readBody) == nil]
		whileFalse: [colorTable := localColorTable
						ifNil: [colorPalette].
			transparentIndex
				ifNotNil: [transparentIndex + 1 > colorTable size
						ifTrue: [colorTable := colorTable forceTo: transparentIndex + 1 paddingWith: Color white].
					colorTable at: transparentIndex + 1 put: Color transparent].
			body colors: colorTable.
			forms add: body.
			delays add: delay].
	^ forms! !

!AnimatedGIFReadWriter methodsFor: 'private' stamp: 'mir 11/19/2003 12:25'!
comment: aString
	comments add: aString! !

!AnimatedGIFReadWriter methodsFor: 'accessing' stamp: 'mir 11/19/2003 14:16'!
delays
	^ delays! !

!AnimatedGIFReadWriter methodsFor: 'accessing' stamp: 'mir 11/19/2003 14:16'!
forms
	^ forms! !

!AnimatedGIFReadWriter methodsFor: 'private-decoding' stamp: 'mir 11/19/2003 12:21'!
readBitData
	| form |
	form := super readBitData.
	form offset: offset.
	^form! !

!GIFReadWriter class methodsFor: 'examples' stamp: 'bf 5/29/2003 01:56'!
exampleAnim
	"GIFReadWriter exampleAnim"

	| writer extent center |
	writer := GIFReadWriter on: (FileStream newFileNamed: 'anim.gif').
	writer loopCount: 20.		"Repeat 20 times"
	writer delay: 10.		"Wait 10/100 seconds"
	extent := 42@42.
	center := extent / 2.
	Cursor write showWhile: [
		[2 to: center x - 1 by: 2 do: [:r |
			"Make a fancy anim without using Canvas - inefficient as hell"
			| image |
			image := ColorForm extent: extent depth: 8.
			0.0 to: 359.0 do: [:theta | image colorAt: (center + (Point r: r degrees: theta)) rounded put: Color red].
			writer nextPutImage: image]
		]	ensure: [writer close]].! !

!GIFReadWriter class methodsFor: 'examples' stamp: 'nk 7/30/2004 21:40'!
grabScreenAndSaveOnDisk
	"GIFReaderWriter grabScreenAndSaveOnDisk"

	| form fileName |
	form := Form fromUser.
	form bits size = 0 ifTrue: [^Beeper beep].
	fileName := FileDirectory default nextNameFor: 'Squeak' extension: 'gif'.
	Utilities informUser: 'Writing ' , fileName
		during: [GIFReadWriter putForm: form onFileNamed: fileName]! !

!GIFReadWriter class methodsFor: 'class initialization' stamp: ''!
initialize
	"GIFReadWriter initialize"

	ImageSeparator := $, asInteger.
	Extension := $!! asInteger.
	Terminator := $; asInteger.
! !

!GIFReadWriter class methodsFor: 'image reading/writing' stamp: 'asm 12/11/2003 21:29'!
typicalFileExtensions
	"Answer a collection of file extensions (lowercase) which files that I can 
	read might commonly have"

	self
		allSubclasses detect: [:cls | cls wantsToHandleGIFs ]
					 ifNone: ["if none of my subclasses wants , then i''ll have to do"
							^ #('gif' )].
	^ #( )! !

!GIFReadWriter class methodsFor: 'image reading/writing' stamp: 'asm 12/11/2003 21:29'!
wantsToHandleGIFs
	^ false! !

!GIFReadWriter methodsFor: 'private' stamp: ''!
checkCodeSize
	(freeCode > maxCode and: [codeSize < 12])
		ifTrue:
			[codeSize := codeSize + 1.
			maxCode := (1 bitShift: codeSize) - 1]! !

!GIFReadWriter methodsFor: 'stream access' stamp: 'bf 5/29/2003 01:23'!
close
	"Write terminator"
	self nextPut: Terminator.
	^super close! !

!GIFReadWriter methodsFor: 'accessing' stamp: 'bf 5/29/2003 01:43'!
delay: aNumberOrNil
	"Set delay for next image in hundredth (1/100) of seconds"
	delay := aNumberOrNil! !

!GIFReadWriter methodsFor: 'private-packing' stamp: ''!
fillBuffer
	| packSize |
	packSize := self next.
	bufStream := ReadStream on: (self next: packSize)! !

!GIFReadWriter methodsFor: 'private-bits access' stamp: ''!
flushBits
	remainBitCount = 0 ifFalse:
		[self nextBytePut: bufByte.
		remainBitCount := 0].
	self flushBuffer! !

!GIFReadWriter methodsFor: 'private-packing' stamp: ''!
flushBuffer
	bufStream isEmpty ifTrue: [^self].
	self nextPut: bufStream size.
	self nextPutAll: bufStream contents.
	bufStream := WriteStream on: (ByteArray new: 256)! !

!GIFReadWriter methodsFor: 'private-encoding' stamp: ''!
flushCode
	self flushBits! !

!GIFReadWriter methodsFor: 'accessing' stamp: 'bf 5/29/2003 01:39'!
loopCount: aNumber
	"Set looping. This must be done before any image is written!!"
	loopCount := aNumber! !

!GIFReadWriter methodsFor: 'private-bits access' stamp: ''!
nextBits
	| integer readBitCount shiftCount byte |
	integer := 0.
	remainBitCount = 0
		ifTrue:
			[readBitCount := 8.
			shiftCount := 0]
		ifFalse:
			[readBitCount := remainBitCount.
			shiftCount := remainBitCount - 8].
	[readBitCount < codeSize]
		whileTrue:
			[byte := self nextByte.
			byte == nil ifTrue: [^eoiCode].
			integer := integer + (byte bitShift: shiftCount).
			shiftCount := shiftCount + 8.
			readBitCount := readBitCount + 8].
	(remainBitCount := readBitCount - codeSize) = 0
		ifTrue:	[byte := self nextByte]
		ifFalse:	[byte := self peekByte].
	byte == nil ifTrue: [^eoiCode].
	^(integer + (byte bitShift: shiftCount)) bitAnd: maxCode! !

!GIFReadWriter methodsFor: 'private-bits access' stamp: ''!
nextBitsPut: anInteger
	| integer writeBitCount shiftCount |
	shiftCount := 0.
	remainBitCount = 0
		ifTrue:
			[writeBitCount := 8.
			integer := anInteger]
		ifFalse:
			[writeBitCount := remainBitCount.
			integer := bufByte + (anInteger bitShift: 8 - remainBitCount)].
	[writeBitCount < codeSize]
		whileTrue:
			[self nextBytePut: ((integer bitShift: shiftCount) bitAnd: 255).
			shiftCount := shiftCount - 8.
			writeBitCount := writeBitCount + 8].
	(remainBitCount := writeBitCount - codeSize) = 0
		ifTrue: [self nextBytePut: (integer bitShift: shiftCount)]
		ifFalse: [bufByte := integer bitShift: shiftCount].
	^anInteger! !

!GIFReadWriter methodsFor: 'private-packing' stamp: ''!
nextByte
	bufStream atEnd
		ifTrue:
			[self atEnd ifTrue: [^nil].
			self fillBuffer].
	^bufStream next! !

!GIFReadWriter methodsFor: 'private-packing' stamp: ''!
nextBytePut: aByte
	bufStream nextPut: aByte.
	bufStream size >= 254 ifTrue: [self flushBuffer]! !

!GIFReadWriter methodsFor: 'accessing' stamp: 'sd 1/30/2004 15:18'!
nextImage
	"Read in the next GIF image from the stream. Read it all into
memory first for speed."

	| f thisImageColorTable |
	stream class == ReadWriteStream ifFalse: [
		stream binary.
		self on: (ReadWriteStream with: (stream contentsOfEntireFile))].

	localColorTable := nil.
	self readHeader.
	f := self readBody.
	self close.
	f == nil ifTrue: [^ self error: 'corrupt GIF file'].

	thisImageColorTable := localColorTable ifNil: [colorPalette].
	transparentIndex ifNotNil: [
		transparentIndex + 1 > thisImageColorTable size ifTrue: [
			thisImageColorTable := thisImageColorTable 
				forceTo: transparentIndex + 1 
				paddingWith: Color white
		].
		thisImageColorTable at: transparentIndex + 1 put: Color transparent
	].
	f colors: thisImageColorTable.
	^ f
! !

!GIFReadWriter methodsFor: 'accessing' stamp: 'ar 10/24/2005 22:52'!
nextPutImage: aForm

	| reduced tempForm |
	aForm unhibernate.
	aForm depth > 8 ifTrue:[
		reduced := aForm colorReduced.  "minimize depth"
		reduced depth > 8 ifTrue: [
			"Not enough color space; do it the hard way."
			reduced := reduced asFormOfDepth: 8].
	] ifFalse:[reduced := aForm].
	reduced depth < 8 ifTrue: [
		"writeBitData: expects depth of 8"
		tempForm := reduced class extent: reduced extent depth: 8.
		(reduced isColorForm) ifTrue:[
			tempForm
				copyBits: reduced boundingBox
				from: reduced at: 0@0
				clippingBox: reduced boundingBox
				rule: Form over
				fillColor: nil
				map: nil.
			tempForm colors: reduced colors.
		] ifFalse: [reduced displayOn: tempForm].
		reduced := tempForm.
	].
	(reduced isColorForm) ifTrue:[
		(reduced colorsUsed includes: Color transparent) ifTrue: [
			transparentIndex := (reduced colors indexOf: Color transparent) - 1.
		]
	] ifFalse: [transparentIndex := nil].
	width := reduced width.
	height := reduced height.
	bitsPerPixel := reduced depth.
	colorPalette := reduced colormapIfNeededForDepth: 32.
	interlace := false.
	self writeHeader.
	self writeBitData: reduced bits.
! !

!GIFReadWriter methodsFor: 'private-packing' stamp: ''!
peekByte
	bufStream atEnd
		ifTrue:
			[self atEnd ifTrue: [^nil].
			self fillBuffer].
	^bufStream peek! !

!GIFReadWriter methodsFor: 'private-decoding' stamp: 'mir 11/19/2003 12:19'!
readBitData
	"using modified Lempel-Ziv Welch algorithm."

	| outCodes outCount bitMask initCodeSize code curCode oldCode inCode finChar i bytes f c packedBits hasLocalColor localColorSize maxOutCodes |

	maxOutCodes := 4096.
	offset := self readWord@self readWord. "Image Left@Image Top"
	width := self readWord.
	height := self readWord.

	"---
	Local Color Table Flag        1 Bit
	Interlace Flag                1 Bit
	Sort Flag                     1 Bit
	Reserved                      2 Bits
	Size of Local Color Table     3 Bits
	----"
	packedBits := self next.
	interlace := (packedBits bitAnd: 16r40) ~= 0.
	hasLocalColor := (packedBits bitAnd: 16r80) ~= 0.
	localColorSize := 1 bitShift: ((packedBits bitAnd: 16r7) + 1).
	hasLocalColor ifTrue: [localColorTable := self readColorTable: localColorSize].

	pass := 0.
	xpos := 0.
	ypos := 0.
	rowByteSize := ((width + 3) // 4) * 4.
	remainBitCount := 0.
	bufByte := 0.
	bufStream := ReadStream on: ByteArray new.

	outCodes := ByteArray new: maxOutCodes + 1.
	outCount := 0.
	bitMask := (1 bitShift: bitsPerPixel) - 1.
	prefixTable := Array new: 4096.
	suffixTable := Array new: 4096.

	initCodeSize := self next.

	self setParameters: initCodeSize.
	bitsPerPixel > 8 ifTrue: [^self error: 'never heard of a GIF that deep'].
	bytes := ByteArray new: rowByteSize * height.
	[(code := self readCode) = eoiCode] whileFalse:
		[code = clearCode
			ifTrue:
				[self setParameters: initCodeSize.
				curCode := oldCode := code := self readCode.
				finChar := curCode bitAnd: bitMask.
				"Horrible hack to avoid running off the end of the bitmap.  Seems to cure problem reading some gifs!!? tk 6/24/97 20:16"
				xpos = 0 ifTrue: [
						ypos < height ifTrue: [
							bytes at: (ypos * rowByteSize) + xpos + 1 put: finChar]]
					ifFalse: [bytes at: (ypos * rowByteSize) + xpos + 1 put: finChar].
				self updatePixelPosition]
			ifFalse:
				[curCode := inCode := code.
				curCode >= freeCode ifTrue:
					[curCode := oldCode.
					outCodes at: (outCount := outCount + 1) put: finChar].
				[curCode > bitMask] whileTrue:
					[outCount > maxOutCodes
						ifTrue: [^self error: 'corrupt GIF file (OutCount)'].
					outCodes at: (outCount := outCount + 1)
						put: (suffixTable at: curCode + 1).
					curCode := prefixTable at: curCode + 1].
				finChar := curCode bitAnd: bitMask.
				outCodes at: (outCount := outCount + 1) put: finChar.
				i := outCount.
				[i > 0] whileTrue:
					["self writePixel: (outCodes at: i) to: bits"
					bytes at: (ypos * rowByteSize) + xpos + 1 put: (outCodes at: i).
					self updatePixelPosition.
					i := i - 1].
				outCount := 0.
				prefixTable at: freeCode + 1 put: oldCode.
				suffixTable at: freeCode + 1 put: finChar.
				oldCode := inCode.
				freeCode := freeCode + 1.
				self checkCodeSize]].
	prefixTable := suffixTable := nil.

	f := ColorForm extent: width@height depth: 8.
	f bits copyFromByteArray: bytes.
	"Squeak can handle depths 1, 2, 4, and 8"
	bitsPerPixel > 4 ifTrue: [^ f].
	"reduce depth to save space"
	c := ColorForm extent: width@height
		depth: (bitsPerPixel = 3 ifTrue: [4] ifFalse: [bitsPerPixel]).
	f displayOn: c.
	^ c
! !

!GIFReadWriter methodsFor: 'private-decoding' stamp: 'KLC 1/25/2004 14:04'!
readBody
	"Read the GIF blocks. Modified to return a form.  "

	| form extype block blocksize packedFields delay1 |
	form := nil.
	[stream atEnd] whileFalse: [
		block := self next.
		block = Terminator ifTrue: [^ form].
		block = ImageSeparator ifTrue: [
			form isNil
				ifTrue: [form := self readBitData]
				ifFalse: [self skipBitData].
		] ifFalse: [
			block = Extension
				ifFalse: [^ form "^ self error: 'Unknown block type'"].
			"Extension block"
			extype := self next.	"extension type"
			extype = 16rF9 ifTrue: [  "graphics control"
				self next = 4 ifFalse: [^ form "^ self error: 'corrupt GIF file'"].
				"====
				Reserved                      3 Bits
				Disposal Method               3 Bits
				User Input Flag               1 Bit
				Transparent Color Flag        1 Bit
				==="
 
				packedFields := self next.
				delay1 := self next.	"delay time 1"
				delay := (self next*256 + delay1) *10.	 "delay time 2"
				transparentIndex := self next.
				(packedFields bitAnd: 1) = 0 ifTrue: [transparentIndex := nil].
				self next = 0 ifFalse: [^ form "^ self error: 'corrupt GIF file'"].
			] ifFalse: [
				"Skip blocks"
				[(blocksize := self next) > 0]
					whileTrue: [
						"Read the block and ignore it and eat the block terminator"
						self next: blocksize]]]]! !

!GIFReadWriter methodsFor: 'private-decoding' stamp: ''!
readCode
	^self nextBits! !

!GIFReadWriter methodsFor: 'private-decoding' stamp: 'RAA 4/25/2001 08:48'!
readColorTable: numberOfEntries

	| array r g b |

	array := Array new: numberOfEntries.
	1 to: array size do: [ :i |
		r := self next.  
		g := self next.  
		b := self next.
		array at: i put: (Color r: r g: g b: b range: 255)
	].
	^array! !

!GIFReadWriter methodsFor: 'private-decoding' stamp: 'RAA 4/25/2001 08:49'!
readHeader
	| is89 byte hasColorMap |
	(self hasMagicNumber: 'GIF87a' asByteArray)
		ifTrue: [is89 := false]
		ifFalse: [(self hasMagicNumber: 'GIF89a' asByteArray)
			ifTrue: [is89 := true]
			ifFalse: [^ self error: 'This does not appear to be a GIF file']].
	self readWord.	"skip Screen Width"
	self readWord.	"skip Screen Height"
	byte := self next.
	hasColorMap := (byte bitAnd: 16r80) ~= 0.
	bitsPerPixel := (byte bitAnd: 7) + 1.
	byte := self next.	"skip background color."
	self next ~= 0
		ifTrue: [is89
			ifFalse: [^self error: 'corrupt GIF file (screen descriptor)']].
	hasColorMap
		ifTrue:
			[colorPalette := self readColorTable: (1 bitShift: bitsPerPixel)]
		ifFalse:
			["Transcript cr; show: 'GIF file does not have a color map.'."
			colorPalette := nil "Palette monochromeDefault"].! !

!GIFReadWriter methodsFor: 'private-encoding' stamp: 'tk 9/14/97 16:25'!
readPixelFrom: bits
	"Since bits is a Bitmap with 32 bit values, watch out for the
padding at the end of each row.  But, GIF format already wants padding to
32 bit boundary!!  OK as is.  tk 9/14/97"

	| pixel |
	ypos >= height ifTrue: [^nil].
	pixel := bits byteAt: (ypos * rowByteSize + xpos + 1).
	self updatePixelPosition.
	^pixel! !

!GIFReadWriter methodsFor: 'private-decoding' stamp: ''!
readWord
	^self next + (self next bitShift: 8)! !

!GIFReadWriter methodsFor: 'private' stamp: ''!
setParameters: initCodeSize
	clearCode := 1 bitShift: initCodeSize.
	eoiCode := clearCode + 1.
	freeCode := clearCode + 2.
	codeSize := initCodeSize + 1.
	maxCode := (1 bitShift: codeSize) - 1! !

!GIFReadWriter methodsFor: 'accessing' stamp: '6/18/97 13:18 '!
setStream: aStream
	"Feed it in from an existing source"
	stream := aStream! !

!GIFReadWriter methodsFor: 'private-decoding' stamp: ''!
skipBitData
	| misc blocksize |
	self readWord.  "skip Image Left"
	self readWord.  "skip Image Top"
	self readWord.  "width"
	self readWord.  "height"
	misc := self next.
	(misc bitAnd: 16r80) = 0 ifFalse: [ "skip colormap"
		1 to: (1 bitShift: (misc bitAnd: 7) + 1) do: [:i |
			self next; next; next]].
	self next.  "minimum code size"
	[(blocksize := self next) > 0]
		whileTrue: [self next: blocksize]! !

!GIFReadWriter methodsFor: 'accessing' stamp: 'di 9/15/1998 09:53'!
understandsImageFormat
	^('abc' collect: [:x | stream next asCharacter]) = 'GIF'! !

!GIFReadWriter methodsFor: 'private' stamp: ''!
updatePixelPosition
	(xpos := xpos + 1) >= width ifFalse: [^self].
	xpos := 0.
	interlace
		ifFalse: [ypos := ypos + 1. ^self].
	pass = 0 ifTrue:
		[(ypos := ypos + 8) >= height
			ifTrue:
				[pass := pass + 1.
				ypos := 4].
		^self].
	pass = 1 ifTrue:
		[(ypos := ypos + 8) >= height
			ifTrue:
				[pass := pass + 1.
				ypos := 2].
		^self].
	pass = 2 ifTrue:
		[(ypos := ypos + 4) >= height
			ifTrue:
				[pass := pass + 1.
				ypos := 1].
		^self].
	pass = 3 ifTrue:
		[ypos := ypos + 2.
		^self].

	^self error: 'can''t happen'! !

!GIFReadWriter methodsFor: 'private-encoding' stamp: 'bf 5/29/2003 01:21'!
writeBitData: bits
	"using modified Lempel-Ziv Welch algorithm."

	| maxBits maxMaxCode tSize initCodeSize ent tShift fCode pixel index disp nomatch |
	pass := 0.
	xpos := 0.
	ypos := 0.
	rowByteSize := width * 8 + 31 // 32 * 4.
	remainBitCount := 0.
	bufByte := 0.
	bufStream := WriteStream on: (ByteArray new: 256).

	maxBits := 12.
	maxMaxCode := 1 bitShift: maxBits.
	tSize := 5003.
	prefixTable := Array new: tSize.
	suffixTable := Array new: tSize.

	initCodeSize := bitsPerPixel <= 1 ifTrue: [2] ifFalse: [bitsPerPixel].
	self nextPut: initCodeSize.
	self setParameters: initCodeSize.

	tShift := 0.
	fCode := tSize.
	[fCode < 65536] whileTrue:
		[tShift := tShift + 1.
		fCode := fCode * 2].
	tShift := 8 - tShift.
	1 to: tSize do: [:i | suffixTable at: i put: -1].

	self writeCodeAndCheckCodeSize: clearCode.
	ent := self readPixelFrom: bits.
	[(pixel := self readPixelFrom: bits) == nil] whileFalse:
		[
		fCode := (pixel bitShift: maxBits) + ent.
		index := ((pixel bitShift: tShift) bitXor: ent) + 1.
		(suffixTable at: index) = fCode
			ifTrue: [ent := prefixTable at: index]
			ifFalse:
				[nomatch := true.
				(suffixTable at: index) >= 0
					ifTrue:
						[disp := tSize - index + 1.
						index = 1 ifTrue: [disp := 1].
						"probe"
						[(index := index - disp) < 1 ifTrue: [index := index + tSize].
						(suffixTable at: index) = fCode
							ifTrue:
								[ent := prefixTable at: index.
								nomatch := false.
								"continue whileFalse:"].
						nomatch and: [(suffixTable at: index) > 0]]
							whileTrue: ["probe"]].
				"nomatch"
				nomatch ifTrue:
					[self writeCodeAndCheckCodeSize: ent.
					ent := pixel.
					freeCode < maxMaxCode
						ifTrue:
							[prefixTable at: index put: freeCode.
							suffixTable at: index put: fCode.
							freeCode := freeCode + 1]
						ifFalse:
							[self writeCodeAndCheckCodeSize: clearCode.
							1 to: tSize do: [:i | suffixTable at: i put: -1].
							self setParameters: initCodeSize]]]].
	prefixTable := suffixTable := nil.
	self writeCodeAndCheckCodeSize: ent.
	self writeCodeAndCheckCodeSize: eoiCode.
	self flushCode.

	self nextPut: 0.	"zero-length packet"
! !

!GIFReadWriter methodsFor: 'private-encoding' stamp: ''!
writeCode: aCode
	self nextBitsPut: aCode! !

!GIFReadWriter methodsFor: 'private-encoding' stamp: ''!
writeCodeAndCheckCodeSize: aCode
	self writeCode: aCode.
	self checkCodeSize! !

!GIFReadWriter methodsFor: 'private-encoding' stamp: 'bf 5/29/2003 01:38'!
writeHeader

	| byte |
	stream position = 0 ifTrue: [
		"For first image only"
		self nextPutAll: 'GIF89a' asByteArray.
		self writeWord: width.	"Screen Width"
		self writeWord: height.	"Screen Height"
		byte := 16r80.  "has color map"
		byte := byte bitOr: ((bitsPerPixel - 1) bitShift: 5).  "color resolution"
		byte := byte bitOr: bitsPerPixel - 1.  "bits per pixel"
		self nextPut: byte.
		self nextPut: 0.		"background color."
		self nextPut: 0.		"reserved"
		colorPalette do: [:pixelValue |
			self	nextPut: ((pixelValue bitShift: -16) bitAnd: 255);
				nextPut: ((pixelValue bitShift: -8) bitAnd: 255);
				nextPut: (pixelValue bitAnd: 255)].
		loopCount notNil ifTrue: [
			"Write a Netscape loop chunk"
			self nextPut: Extension.
			self nextPutAll: #(255 11 78 69 84 83 67 65 80 69 50 46 48 3 1) asByteArray.
			self writeWord: loopCount.
			self nextPut: 0]].

	delay notNil | transparentIndex notNil ifTrue: [
		self nextPut: Extension;
			nextPutAll: #(16rF9 4) asByteArray;
			nextPut: (transparentIndex isNil ifTrue: [0] ifFalse: [9]);
			writeWord: (delay isNil ifTrue: [0] ifFalse: [delay]);
			nextPut: (transparentIndex isNil ifTrue: [0] ifFalse: [transparentIndex]);
			nextPut: 0].

	self nextPut: ImageSeparator.
	self writeWord: 0.		"Image Left"
	self writeWord: 0.		"Image Top"
	self writeWord: width.	"Image Width"
	self writeWord: height.	"Image Height"
	byte := interlace ifTrue: [16r40] ifFalse: [0].
	self nextPut: byte.
! !

!GIFReadWriter methodsFor: 'private-encoding' stamp: ''!
writeWord: aWord
	self nextPut: (aWord bitAnd: 255).
	self nextPut: ((aWord bitShift: -8) bitAnd: 255).
	^aWord! !

!ImageReadWriter class methodsFor: 'image reading/writing' stamp: 'nk 7/16/2003 17:59'!
allTypicalFileExtensions
	"Answer a collection of file extensions (lowercase) which files that my subclasses can read might commonly have"
	"ImageReadWriter allTypicalFileExtensions"
	| extensions |
	extensions := Set new.
	self allSubclassesDo: [ :cls | extensions addAll: cls typicalFileExtensions ].
	^extensions! !

!ImageReadWriter class methodsFor: 'image reading/writing' stamp: 'ls 9/15/1998 19:08'!
formFromFileNamed: fileName
	"Answer a ColorForm stored on the file with the given name."
	| stream |
	stream := FileStream readOnlyFileNamed: fileName.
	^self formFromStream: stream! !

!ImageReadWriter class methodsFor: 'image reading/writing' stamp: 'tk 5/7/1998 17:35'!
formFromServerFile: fileName
	"Answer a ColorForm stored on the file with the given name.  Meant to be called from during the getting of updates from the server.  That assures that (Utilities serverUrls) returns the right group of servers."

	| form urls doc |
	urls := Utilities serverUrls collect:
		[:url | url, fileName].  " fileName starts with: 'updates/'  "
	urls do: [:aURL |
		(fileName findTokens: '.') last asLowercase = 'gif' ifTrue: [
			form := HTTPSocket httpGif: aURL.
			form = (ColorForm extent: 20@20 depth: 8) 
				ifTrue: [self inform: 'The file ',aURL,' is ill formed.'].
			^ form].
		(fileName findTokens: '.') last asLowercase = 'bmp' ifTrue: [
			doc := HTTPSocket httpGet: aURL accept: 'image/bmp'.
			form := Form fromBMPFile: doc.
			doc close.
			form ifNil: [self inform: 'The file ',aURL,' is ill formed.'. ^ Form new]
				ifNotNil: [^ form]].
		self inform: 'File ', fileName, 'does not end with .gif or .bmp'].
	self inform: 'That file not found on any server we know'.! !

!ImageReadWriter class methodsFor: 'image reading/writing' stamp: 'sd 1/30/2004 15:18'!
formFromStream: aBinaryStream
	"Answer a ColorForm stored on the given stream.  closes the stream"
	| reader readerClass form  |

	readerClass := self withAllSubclasses
		detect: [:subclass | subclass understandsImageFormat: aBinaryStream]
		ifNone: [
			aBinaryStream close.
			^self error: 'image format not recognized'].
	reader := readerClass new on: aBinaryStream reset.
	Cursor read showWhile: [
		form := reader nextImage.
		reader close].
	^ form
! !

!ImageReadWriter class methodsFor: 'instance creation' stamp: ''!
on: aStream
	"Answer an instance of the receiver for encoding and/or decoding images on the given."

	^ self new on: aStream
! !

!ImageReadWriter class methodsFor: 'image reading/writing' stamp: ''!
putForm: aForm onFileNamed: fileName
	"Store the given form on a file of the given name."

	| writer |
	writer := self on: (FileStream newFileNamed: fileName) binary.
	Cursor write showWhile: [writer nextPutImage: aForm].
	writer close.
! !

!ImageReadWriter class methodsFor: 'image reading/writing' stamp: 'tk
9/13/97 16:13'!
putForm: aForm onStream: aWriteStream
	"Store the given form on a file of the given name."

	| writer |
	writer := self on: aWriteStream.
	Cursor write showWhile: [writer nextPutImage: aForm].
	writer close.
! !

!ImageReadWriter class methodsFor: 'image reading/writing' stamp: 'nk 7/16/2003 17:55'!
typicalFileExtensions
	"Answer a collection of file extensions (lowercase) which files that I can read might commonly have"
	^#()! !

!ImageReadWriter class methodsFor: 'image reading/writing' stamp: 'ar 3/1/2006 22:59'!
understandsImageFormat: aStream 
	^[(self new on: aStream) understandsImageFormat] on: Error do:[:ex| ex return: false]! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
atEnd

	^stream atEnd! !

!ImageReadWriter methodsFor: 'private' stamp: ''!
changePadOfBits: bits width: width height: height depth: depth from: oldPad
to: newPad
	"Change padding size of bits."

	| srcRowByteSize dstRowByteSize newBits srcRowBase rowEndOffset |
	(#(8 16 32) includes: oldPad)
		ifFalse: [^self error: 'Invalid pad: ', oldPad printString].
	(#(8 16 32) includes: newPad)
		ifFalse: [^self error: 'Invalid pad: ', newPad printString].
	srcRowByteSize := width * depth + oldPad - 1 // oldPad * (oldPad / 8).
	srcRowByteSize * height = bits size
		ifFalse: [^self error: 'Incorrect bitmap array size.'].
	dstRowByteSize := width * depth + newPad - 1 // newPad * (newPad / 8).
	newBits := ByteArray new: dstRowByteSize * height.
	srcRowBase := 1.
	rowEndOffset := dstRowByteSize - 1.
	1 to: newBits size by: dstRowByteSize do:
		[:dstRowBase |
		newBits replaceFrom: dstRowBase
			to: dstRowBase + rowEndOffset
			with: bits
			startingAt: srcRowBase.
		srcRowBase := srcRowBase + srcRowByteSize].
	^newBits! !

!ImageReadWriter methodsFor: 'stream access' stamp: 'sd 1/30/2004 15:18'!
close
	
	stream close! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
contents

	^stream contents! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
cr

	^stream nextPut: Character cr asInteger! !

!ImageReadWriter methodsFor: 'private' stamp: ''!
hasMagicNumber: aByteArray
	| position |
	position := stream position.
	((stream size - position) >= aByteArray size and:
	[(stream next: aByteArray size)  = aByteArray])
		ifTrue: [^true].
	stream position: position.
	^false! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
lf
	"PPM and PBM are used LF as CR."

	^stream nextPut: Character lf asInteger! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
next

	^stream next! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
next: size

	^stream next: size! !

!ImageReadWriter methodsFor: 'accessing' stamp: ''!
nextImage
	"Dencoding an image on stream and answer the image."

	^self subclassResponsibility! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
nextLong
	"Read a 32-bit quantity from the input stream."

	^(stream next bitShift: 24) + (stream next bitShift: 16) +
		(stream next bitShift: 8) + stream next! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
nextLongPut: a32BitW
	"Write out a 32-bit integer as 32 bits."

	stream nextPut: ((a32BitW bitShift: -24) bitAnd: 16rFF).
	stream nextPut: ((a32BitW bitShift: -16) bitAnd: 16rFF).
	stream nextPut: ((a32BitW bitShift: -8) bitAnd: 16rFF).
	stream nextPut: (a32BitW bitAnd: 16rFF).
	^a32BitW! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
nextPut: aByte

	^stream nextPut: aByte! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
nextPutAll: aByteArray

	^stream nextPutAll: aByteArray! !

!ImageReadWriter methodsFor: 'accessing' stamp: ''!
nextPutImage: anImage
	"Encoding anImage on stream."

	^self subclassResponsibility! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
nextWord
	"Read a 16-bit quantity from the input stream."

	^(stream next bitShift: 8) + stream next! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
nextWordPut: a16BitW
	"Write out a 16-bit integer as 16 bits."

	stream nextPut: ((a16BitW bitShift: -8) bitAnd: 16rFF).
	stream nextPut: (a16BitW bitAnd: 16rFF).
	^a16BitW! !

!ImageReadWriter methodsFor: 'private' stamp: 'sd 1/30/2004 15:18'!
on: aStream
	(stream := aStream) reset.
	stream binary.
	"Note that 'reset' makes a file be text.  Must do this after."! !

!ImageReadWriter methodsFor: 'stream access' stamp: 'tao 10/23/97 18:00'!
peekFor: aValue

	^stream peekFor: aValue! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
position

	^stream position! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
position: anInteger

	^stream position: anInteger! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
size

	^stream size! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
skip: anInteger

	^stream skip: anInteger! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
space

	^stream nextPut: Character space asInteger! !

!ImageReadWriter methodsFor: 'stream access' stamp: ''!
tab

	^stream nextPut: Character tab asInteger! !

!ImageReadWriter methodsFor: 'testing' stamp: 'tao 10/27/97 09:26'!
understandsImageFormat
	"Test to see if the image stream format is understood by this decoder.
	This should be implemented in each subclass of ImageReadWriter so that
	a proper decoder can be selected without ImageReadWriter having to know
	about all possible image file types."

	^ false! !

!ImageReadWriter methodsFor: 'private' stamp: ''!
unpackBits: bits depthTo8From: depth with: width height: height pad: pad
	"Unpack bits of depth 1, 2, or 4 image to it of depth 8 image."

	| bitMask pixelInByte bitsWidth upBitsWidth stopWidth
	 trailingSize upBits bitIndex upBitIndex val |
	(#(1 2 4) includes: depth)
		ifFalse: [^self error: 'depth must be 1, 2, or 4'].
	(#(8 16 32) includes: pad)
		ifFalse: [^self error: 'pad must be 8, 16, or 32'].
	bitMask := (1 bitShift: depth) - 1.
	pixelInByte := 8 / depth.
	bitsWidth := width * depth + pad - 1 // pad * (pad / 8).
	upBitsWidth := width * 8 + pad - 1 // pad * (pad / 8).
	stopWidth := width * depth + 7 // 8.
	trailingSize := width - (stopWidth - 1 * pixelInByte).
	upBits := ByteArray new: upBitsWidth * height.
	1 to: height do: [:i |
		bitIndex := i - 1 * bitsWidth.
		upBitIndex := i - 1 * upBitsWidth.
		1 to: stopWidth - 1 do: [:j |
			val := bits at: (bitIndex := bitIndex + 1).
			upBitIndex := upBitIndex + pixelInByte.
			1 to: pixelInByte do: [:k |
				upBits at: (upBitIndex - k + 1) put: (val bitAnd: bitMask).
				val := val bitShift: depth negated]].
		val := (bits at: (bitIndex := bitIndex + 1))
				bitShift: depth negated * (pixelInByte - trailingSize).
		upBitIndex := upBitIndex + trailingSize.
		1 to: trailingSize do: [:k |
			upBits at: (upBitIndex - k + 1) put: (val bitAnd: bitMask).
			val := val bitShift: depth negated]].
	^ upBits
! !

ImageReadWriter subclass: #JPEGReadWriter
	instanceVariableNames: 'width height components currentComponents qTable hACTable hDCTable restartInterval restartsToGo mcuWidth mcuHeight mcusPerRow mcuRowsInScan mcuMembership mcuSampleBuffer mcuImageBuffer majorVersion minorVersion dataPrecision densityUnit xDensity yDensity ss se ah al sosSeen residuals ditherMask'
	classVariableNames: 'ConstBits DCTK1 DCTK2 DCTK3 DCTK4 DCTSize DCTSize2 DitherMasks FIXn0n298631336 FIXn0n34414 FIXn0n390180644 FIXn0n541196100 FIXn0n71414 FIXn0n765366865 FIXn0n899976223 FIXn1n175875602 FIXn1n40200 FIXn1n501321110 FIXn1n77200 FIXn1n847759065 FIXn1n961570560 FIXn2n053119869 FIXn2n562915447 FIXn3n072711026 FloatSampleOffset HuffmanTableSize JFIFMarkerParser JPEGNaturalOrder MaxSample Pass1Bits Pass1Div Pass2Div QTableScaleFactor QuantizationTableSize SampleOffset'
	poolDictionaries: ''
	category: 'Graphics-Files'!

!JPEGReadWriter commentStamp: '<historical>' prior: 0!
I am a subclass of ImageReadWriter that understands JFIF file streams, and can decode JPEG images.
This code is based upon the Independent Joint Photographic Experts Group (IJPEG) software, originally written in C by Tom Lane, Philip Gladstone, Luis Ortiz, Jim Boucher, Lee Crocker, Julian Minguillon, George Phillips, Davide Rossi, Ge' Weijers, and other members of the Independent JPEG Group.

!

!JPEGReadWriter class methodsFor: 'initialization' stamp: 'ar 3/3/2001 23:07'!
initialize
	"JPEGReadWriter initialize"
	"general constants"
	DCTSize := 8.
	MaxSample := (2 raisedToInteger: DCTSize) - 1.
	SampleOffset := MaxSample // 2.
	FloatSampleOffset := SampleOffset asFloat.
	DCTSize2 := DCTSize squared.
	QuantizationTableSize := 4.
	HuffmanTableSize := 4.

	"floating-point Inverse Discrete Cosine Transform (IDCT) constants"
	ConstBits := 13.
	Pass1Bits := 2.
	DCTK1 := 2 sqrt.
	DCTK2 := 1.847759065.
	DCTK3 := 1.082392200.
	DCTK4 := -2.613125930.
	Pass1Div := 1 bitShift: ConstBits - Pass1Bits.
	Pass2Div := 1 bitShift: ConstBits + Pass1Bits + 3.

	"fixed-point Inverse Discrete Cosine Transform (IDCT) constants"
	FIXn0n298631336 := 2446.
	FIXn0n390180644 := 3196.
	FIXn0n541196100 := 4433.
	FIXn0n765366865 := 6270.
	FIXn0n899976223 := 7373.
	FIXn1n175875602 := 9633.
	FIXn1n501321110 := 12299.
	FIXn1n847759065 := 15137.
	FIXn1n961570560 := 16069.
	FIXn2n053119869 := 16819.
	FIXn2n562915447 := 20995.
	FIXn3n072711026 := 25172.

	"fixed-point color conversion constants"
	FIXn0n34414 := 22554.
	FIXn0n71414 := 46802.
	FIXn1n40200 := 91881.
	FIXn1n77200 :=  116130.

	"reordering table from JPEG zig-zag order"
	JPEGNaturalOrder := #(
		1 2 9 17 10 3 4 11
		18 25 33 26 19 12 5 6
		13 20 27 34 41 49 42 35
		28 21 14 7 8 15 22 29
		36 43 50 57 58 51 44 37
		30 23 16 24 31 38 45 52
		59 60 53 46 39 32 40 47
		54 61 62 55 48 56 63 64).

	"scale factors for the values in the Quantization Tables"
	QTableScaleFactor := (0 to: DCTSize-1) collect:
		[:k | k = 0
			ifTrue: [1.0]
			ifFalse: [(k * Float pi / 16) cos * 2 sqrt]].

	"dithering masks"
	(DitherMasks := Dictionary new)
		add: 0 -> 0;
		add: 1 -> 127;
		add: 2 -> 63;
		add: 4 -> 63;
		add: 8 -> 31;
		add: 16 -> 7;
		add: 32 -> 0.

	"dictionary of marker parsers"
	(JFIFMarkerParser := Dictionary new)
		add: (16r01 -> #parseNOP);
		add: (16rC0 -> #parseStartOfFile);
		add: (16rC4 -> #parseHuffmanTable);
		addAll: ((16rD0 to: 16rD7) collect: [:m | Association key: m value: #parseNOP]);
		add: (16rD8 -> #parseStartOfInput);
		add: (16rD9 -> #parseEndOfInput);
		add: (16rDA -> #parseStartOfScan);
		add: (16rDB -> #parseQuantizationTable);
		add: (16rDD -> #parseDecoderRestartInterval);
		add: (16rE0 -> #parseAPPn);
		add: (16rE1 -> #parseAPPn)! !

!JPEGReadWriter class methodsFor: 'image reading/writing' stamp: 'nk 7/16/2003 17:56'!
typicalFileExtensions
	"Answer a collection of file extensions (lowercase) which files that I can read might commonly have"
	^#('jpg' 'jpeg')! !

!JPEGReadWriter class methodsFor: 'image reading/writing' stamp: 'ar 6/16/2002 18:54'!
understandsImageFormat: aStream
	(JPEGReadWriter2 understandsImageFormat: aStream) ifTrue:[^false].
	aStream reset.
	aStream next = 16rFF ifFalse: [^ false].
	aStream next = 16rD8 ifFalse: [^ false].
	^true! !

!JPEGReadWriter methodsFor: 'colorspace conversion' stamp: 'ar 3/4/2001 22:18'!
colorConvertFloatYCbCrMCU

	| ySampleStream crSampleStream cbSampleStream y cb cr red green blue bits |
	ySampleStream := currentComponents at: 1.
	cbSampleStream := currentComponents at: 2.
	crSampleStream := currentComponents at: 3.
	ySampleStream resetSampleStream.
	cbSampleStream resetSampleStream.
	crSampleStream resetSampleStream.
	bits := mcuImageBuffer bits.
	1 to: bits size do:
		[:i |
		y := ySampleStream nextSample.
		cb := cbSampleStream nextSample - FloatSampleOffset.
		cr := crSampleStream nextSample - FloatSampleOffset.
		red := self sampleFloatRangeLimit: (y + (1.40200 * cr)).
		green := self sampleFloatRangeLimit: (y - (0.34414 * cb) - (0.71414 * cr)).
		blue := self sampleFloatRangeLimit: (y + (1.77200 * cb)).
		bits at: i put: 16rFF000000 + (red << 16) + (green << 8) + blue].
	! !

!JPEGReadWriter methodsFor: 'colorspace conversion' stamp: 'ar 3/4/2001 22:17'!
colorConvertGrayscaleMCU

	| ySampleStream y bits |
	ySampleStream := currentComponents at: 1.
	ySampleStream resetSampleStream.
	bits := mcuImageBuffer bits.
	1 to: bits size do:
		[:i |
		y := (ySampleStream nextSample) + (residuals at: 2).
		y > MaxSample ifTrue: [y := MaxSample].
		residuals at: 2 put: (y bitAnd: ditherMask).
		y := y bitAnd: MaxSample - ditherMask.
		y < 1 ifTrue: [y := 1].
		bits at: i put: 16rFF000000 + (y<<16) + (y<<8) + y].
	! !

!JPEGReadWriter methodsFor: 'colorspace conversion' stamp: 'ar 3/4/2001 22:18'!
colorConvertIntYCbCrMCU

	| ySampleStream crSampleStream cbSampleStream y cb cr red green blue bits |
	ySampleStream := currentComponents at: 1.
	cbSampleStream := currentComponents at: 2.
	crSampleStream := currentComponents at: 3.
	ySampleStream resetSampleStream.
	cbSampleStream resetSampleStream.
	crSampleStream resetSampleStream.
	bits := mcuImageBuffer bits.
	1 to: bits size do:
		[:i |
		y := ySampleStream nextSample.
		cb := cbSampleStream nextSample - SampleOffset.
		cr := crSampleStream nextSample - SampleOffset.
		red := y + ((FIXn1n40200 * cr) // 65536) + (residuals at: 1).
		red > MaxSample
			ifTrue: [red := MaxSample]
			ifFalse: [red < 0 ifTrue: [red := 0]].
		residuals at: 1 put: (red bitAnd: ditherMask).
		red := red bitAnd: MaxSample - ditherMask.
		red < 1 ifTrue: [red := 1].
		green := y - ((FIXn0n34414 * cb) // 65536) -
			((FIXn0n71414 * cr) // 65536) + (residuals at: 2).
		green > MaxSample
			ifTrue: [green := MaxSample]
			ifFalse: [green < 0 ifTrue: [green := 0]].
		residuals at: 2 put: (green bitAnd: ditherMask).
		green := green bitAnd: MaxSample - ditherMask.
		green < 1 ifTrue: [green := 1].
		blue := y + ((FIXn1n77200 * cb) // 65536) + (residuals at: 3).
		blue > MaxSample
			ifTrue: [blue := MaxSample]
			ifFalse: [blue < 0 ifTrue: [blue := 0]].
		residuals at: 3 put: (blue bitAnd: ditherMask).
		blue := blue bitAnd: MaxSample - ditherMask.
		blue < 1 ifTrue: [blue := 1].
		bits at: i put: 16rFF000000 + (red bitShift: 16) + (green bitShift: 8) + blue].
	! !

!JPEGReadWriter methodsFor: 'colorspace conversion' stamp: 'ar 3/7/2001 01:02'!
colorConvertMCU

	^ currentComponents size = 3
		ifTrue:
			[self useFloatingPoint
				ifTrue: [self colorConvertFloatYCbCrMCU]
				ifFalse: [self primColorConvertYCbCrMCU: currentComponents
								bits: mcuImageBuffer bits
								residuals: residuals
								ditherMask: ditherMask.]]
		ifFalse: [self primColorConvertGrayscaleMCU]! !

!JPEGReadWriter methodsFor: 'dct' stamp: 'tao 10/26/97 15:16'!
dctFloatRangeLimit: value

	^ (value / 8.0) + FloatSampleOffset.! !

!JPEGReadWriter methodsFor: 'huffman encoding' stamp: 'ar 3/4/2001 20:55'!
decodeBlockInto: anArray component: aColorComponent dcTable: huffmanDC acTable: huffmanAC

	| byte i zeroCount |
	byte := stream decodeValueFrom: huffmanDC.
	byte ~= 0 ifTrue: [byte := self scaleAndSignExtend: ( self getBits: byte) inFieldWidth: byte].
	byte := aColorComponent updateDCValue: byte.
	anArray atAllPut: 0.
	anArray at: 1 put: byte.
	i := 2.
	[i <= DCTSize2] whileTrue:
		[byte := stream decodeValueFrom: huffmanAC.
		zeroCount := byte >> 4.
		byte := byte bitAnd: 16r0F.
		byte ~= 0
			ifTrue:
				[i := i + zeroCount.
				byte := self scaleAndSignExtend: ( self getBits: byte) inFieldWidth: byte.
				anArray at:	 (JPEGNaturalOrder at: i) put: byte]
			ifFalse:
				[zeroCount = 15 ifTrue: [i := i + zeroCount] ifFalse: [^ self]].
		i := i + 1]
		! !

!JPEGReadWriter methodsFor: 'huffman encoding' stamp: 'ar 3/7/2001 01:00'!
decodeMCU

	| comp ci |
	(restartInterval ~= 0 and: [restartsToGo = 0]) ifTrue: [self processRestart].
	1 to: mcuMembership size do:[:i|
		ci := mcuMembership at: i.
		comp := currentComponents at: ci.
		self
			primDecodeBlockInto: (mcuSampleBuffer at: i)
			component: comp
			dcTable: (hDCTable at: comp dcTableIndex)
			acTable: (hACTable at: comp acTableIndex)
			stream: stream.
	].
	restartsToGo := restartsToGo - 1.! !

!JPEGReadWriter methodsFor: 'public access' stamp: 'ar 3/7/2001 00:18'!
decompressionTest
	"Test decompression; don't generate actual image"
	| xStep yStep x y |
MessageTally spyOn:[
	ditherMask := DitherMasks at: 32.
	residuals := WordArray new: 3.
	sosSeen := false.
	self parseFirstMarker.
	[sosSeen] whileFalse: [self parseNextMarker].
	xStep := mcuWidth * DCTSize.
	yStep := mcuHeight * DCTSize.
	y := 0.
	1 to: mcuRowsInScan do:
		[:row |
		x := 0.
		1 to: mcusPerRow do:
			[:col |
			self decodeMCU.
			self idctMCU.
			self colorConvertMCU.
			x := x + xStep].
		y := y + yStep].
].! !

!JPEGReadWriter methodsFor: 'huffman encoding' stamp: 'ar 3/4/2001 17:27'!
getBits: requestedBits
	^stream getBits: requestedBits! !

!JPEGReadWriter methodsFor: 'accessing' stamp: 'tao 10/19/97 13:46'!
hACTable

	hACTable ifNil: [hACTable := Array new: HuffmanTableSize].
	^ hACTable! !

!JPEGReadWriter methodsFor: 'accessing' stamp: 'tao 10/19/97 13:46'!
hDCTable

	hDCTable ifNil: [hDCTable := Array new: HuffmanTableSize].
	^ hDCTable! !

!JPEGReadWriter methodsFor: 'dct' stamp: 'tao 10/26/97 16:34'!
idctBlockFloat: anArray component: aColorComponent

	| t0 t1 t2 t3 t4 t5 t6 t7 t10 t11 t12 t13 z5 z10 z11 z12 z13 qt ws |
	qt := self qTable at: (aColorComponent qTableIndex).
	ws := Array new: DCTSize2.

	"Pass 1: process columns from input, store into work array"
	1 to: DCTSize do: [:i |
		t0 := (anArray at: i) * (qt at: i).
		t1 := (anArray at: (DCTSize*2 + i)) * (qt at: (DCTSize*2 + i)).
		t2 := (anArray at: (DCTSize*4 + i)) * (qt at: (DCTSize*4 + i)).
		t3 := (anArray at: (DCTSize*6 + i)) * (qt at: (DCTSize*6 + i)).
		t10 := t0 + t2.
		t11 := t0 - t2.
		t13 := t1 + t3.
		t12 := (t1 - t3) * DCTK1 - t13.
		t0 := t10 + t13.
		t3 := t10 - t13.
		t1 := t11 + t12.
		t2 := t11 - t12.
		t4 := (anArray at: (DCTSize + i)) * (qt at: (DCTSize + i)).
		t5 := (anArray at: (DCTSize*3 + i)) * (qt at: (DCTSize*3 + i)).
		t6 := (anArray at: (DCTSize*5 + i)) * (qt at: (DCTSize*5 + i)).
		t7 := (anArray at: (DCTSize*7 + i)) * (qt at: (DCTSize*7 + i)).
		z13 := t6 + t5.
		z10 := t6 - t5.
		z11 := t4 + t7.
		z12 := t4 - t7.
		t7 := z11 + z13.
		t11 := (z11 - z13) * DCTK1.
		z5 := (z10 + z12) * DCTK2.
		t10 := DCTK3 * z12 - z5.
		t12 := DCTK4 * z10 + z5.
		t6 := t12 - t7.
		t5 := t11 - t6.
		t4 := t10 + t5.
		ws at: i put: t0 + t7.
		ws at: (DCTSize*7 + i) put: t0 - t7.
		ws at: (DCTSize + i) put: t1 + t6.
		ws at: (DCTSize*6 + i) put: t1 - t6.
		ws at: (DCTSize*2 + i) put: t2 + t5.
		ws at: (DCTSize*5 + i) put: t2 - t5.
		ws at: (DCTSize*4 + i) put: t3 + t4.
		ws at: (DCTSize*3 + i) put: t3 - t4].

		"Pass 2: process rows from the workspace"
	(0 to: DCTSize2-DCTSize by: DCTSize) do: [:i |
		t10 := (ws at: (i+1)) + (ws at: (i+5)).
		t11 := (ws at: (i+1)) - (ws at: (i+5)).
		t13 := (ws at: (i+3)) + (ws at: (i+7)).
		t12 := ((ws at: (i+3)) - (ws at: (i+7))) * DCTK1 - t13.
		t0 := t10 + t13.
		t3 := t10 - t13.
		t1 := t11 + t12.
		t2 := t11 - t12.
		z13 := (ws at: (i+6)) + (ws at: (i+4)).
		z10 := (ws at: (i+6)) - (ws at: (i+4)).
		z11 := (ws at: (i+2)) + (ws at: (i+8)).
		z12 := (ws at: (i+2)) - (ws at: (i+8)).
		t7 := z11 + z13.
		t11 := (z11 - z13) * DCTK1.
		z5 := (z10 + z12) * DCTK2.
		t10 := DCTK3 * z12 - z5.
		t12 := DCTK4 * z10 + z5.
		t6 := t12 - t7.
		t5 := t11 - t6.
		t4 := t10 + t5.

		"final output stage: scale down by a factor of 8 and range-limit"
		anArray at: (i+1) put: (self dctFloatRangeLimit: (t0 + t7)).
		anArray at: (i+8) put: (self dctFloatRangeLimit: (t0 - t7)).
		anArray at: (i+2) put: (self dctFloatRangeLimit: (t1 + t6)).
		anArray at: (i+7) put: (self dctFloatRangeLimit: (t1 - t6)).
		anArray at: (i+3) put: (self dctFloatRangeLimit: (t2 + t5)).
		anArray at: (i+6) put: (self dctFloatRangeLimit: (t2 - t5)).
		anArray at: (i+5) put: (self dctFloatRangeLimit: (t3 + t4)).
		anArray at: (i+4) put: (self dctFloatRangeLimit: (t3 - t4))]


! !

!JPEGReadWriter methodsFor: 'dct' stamp: 'ar 3/4/2001 21:35'!
idctBlockInt: anArray component: aColorComponent
	^self idctBlockInt: anArray qt: (self qTable at: aColorComponent qTableIndex)! !

!JPEGReadWriter methodsFor: 'dct' stamp: 'ar 3/4/2001 21:34'!
idctBlockInt: anArray qt: qt

	|  ws anACTerm dcval z1 z2 z3 z4 z5 t0 t1 t2 t3 t10 t11 t12 t13 shift |
	ws := Array new: DCTSize2.

	"Pass 1: process columns from anArray, store into work array"
	shift := 1 bitShift: ConstBits - Pass1Bits.
	1 to: DCTSize do:
		[:i |
		anACTerm := nil.
		1 to: DCTSize-1 do:[:row|
			anACTerm ifNil:[
				(anArray at: row * DCTSize + i) = 0 ifFalse:[anACTerm := row]]].
		anACTerm == nil
			ifTrue:
				[dcval := (anArray at: i) * (qt at: 1) bitShift: Pass1Bits.
				0 to: DCTSize-1 do: [:j | ws at: (j * DCTSize + i) put: dcval]]
			ifFalse:
				[z2 := (anArray at: (DCTSize * 2 + i)) * (qt at: (DCTSize * 2 + i)).
				z3 := (anArray at: (DCTSize * 6 + i)) * (qt at: (DCTSize * 6 + i)).
				z1 := (z2 + z3) * FIXn0n541196100.
				t2 := z1 + (z3 * FIXn1n847759065 negated).
				t3 := z1 + (z2 * FIXn0n765366865).
				z2 := (anArray at: i) * (qt at: i).
				z3 := (anArray at: (DCTSize * 4 + i)) * (qt at: (DCTSize * 4 + i)).
				t0 := (z2 + z3) bitShift: ConstBits.
				t1 := (z2 - z3) bitShift: ConstBits.
				t10 := t0 + t3.
				t13 := t0 - t3.
				t11 := t1 + t2.
				t12 := t1 - t2.
				t0 := (anArray at: (DCTSize * 7 + i)) * (qt at: (DCTSize * 7 + i)).
				t1 := (anArray at: (DCTSize * 5 + i)) * (qt at: (DCTSize * 5 + i)).
				t2 := (anArray at: (DCTSize * 3 + i)) * (qt at: (DCTSize * 3 + i)).
				t3 := (anArray at: (DCTSize + i)) * (qt at: (DCTSize + i)).
				z1 := t0 + t3.
				z2 := t1 + t2.
				z3 := t0 + t2.
				z4 := t1 + t3.
				z5 := (z3 + z4) * FIXn1n175875602.
				t0 := t0 * FIXn0n298631336.
				t1 := t1 * FIXn2n053119869.
				t2 := t2 * FIXn3n072711026.
				t3 := t3 * FIXn1n501321110.
				z1 := z1 * FIXn0n899976223 negated.
				z2 := z2 * FIXn2n562915447 negated.
				z3 := z3 * FIXn1n961570560 negated.
				z4 := z4 * FIXn0n390180644 negated.
				z3 := z3 + z5.
				z4 := z4 + z5.
				t0 := t0 + z1 + z3.
				t1 := t1 +z2 +z4.
				t2 := t2 + z2 + z3.
				t3 := t3 + z1 + z4.
				ws at: i put: (t10 + t3) >> (ConstBits - Pass1Bits).
				ws at: (DCTSize * 7 + i) put: (t10 - t3) // shift.
				ws at: (DCTSize * 1 + i) put: (t11 + t2) // shift.
				ws at: (DCTSize * 6 + i) put: (t11 - t2) // shift.
				ws at: (DCTSize * 2 + i) put: (t12 + t1) // shift.
				ws at: (DCTSize * 5 + i) put: (t12 - t1) // shift.
				ws at: (DCTSize * 3 + i) put: (t13 + t0) // shift.
				ws at: (DCTSize * 4 + i) put: (t13 - t0) // shift]].

	"Pass 2: process rows from work array, store back into anArray"
	shift := 1 bitShift: ConstBits + Pass1Bits + 3.
	0 to: DCTSize2-DCTSize by: DCTSize do:
		[:i |
		z2 := ws at: i + 3.
		z3 := ws at: i + 7.
		z1 := (z2 + z3) * FIXn0n541196100.
		t2 := z1 + (z3 * FIXn1n847759065 negated).
		t3 := z1 + (z2 * FIXn0n765366865).
		t0 := (ws at: (i + 1)) + (ws at: (i + 5)) bitShift: ConstBits.
		t1 := (ws at: (i + 1)) - (ws at: (i + 5)) bitShift: ConstBits.
		t10 := t0 + t3.
		t13 := t0 - t3.
		t11 := t1 + t2.
		t12 := t1 -t2.
		t0 := ws at: (i + 8).
		t1 := ws at: (i + 6).
		t2 := ws at: (i + 4).
		t3 := ws at: (i + 2).
		z1 := t0 + t3.
		z2 := t1 + t2.
		z3 := t0 + t2.
		z4 := t1 + t3.
		z5 := (z3 + z4) * FIXn1n175875602.
		t0 := t0 * FIXn0n298631336.
		t1 := t1 * FIXn2n053119869.
		t2 := t2 * FIXn3n072711026.
		t3 := t3 * FIXn1n501321110.
		z1 := z1 * FIXn0n899976223 negated.
		z2 := z2 * FIXn2n562915447 negated.
		z3 := z3 * FIXn1n961570560 negated.
		z4 := z4 * FIXn0n390180644 negated.
		z3 := z3 + z5.
		z4 := z4 + z5.
		t0 := t0 + z1 + z3.
		t1 := t1 + z2 + z4.
		t2 := t2 + z2 + z3.
		t3 := t3 + z1 + z4.
		anArray at: (i + 1) put: (self sampleRangeLimit: (t10 + t3) // shift + SampleOffset).
		anArray at: (i + 8) put: (self sampleRangeLimit: (t10 - t3) // shift + SampleOffset).
		anArray at: (i + 2) put: (self sampleRangeLimit: (t11 + t2) // shift + SampleOffset).
		anArray at: (i + 7) put: (self sampleRangeLimit: (t11 - t2) // shift + SampleOffset).
		anArray at: (i + 3) put: (self sampleRangeLimit: (t12 + t1) // shift + SampleOffset).
		anArray at: (i + 6) put: (self sampleRangeLimit: (t12 - t1) // shift + SampleOffset).
		anArray at: (i + 4) put: (self sampleRangeLimit: (t13 + t0) // shift + SampleOffset).
		anArray at: (i + 5) put: (self sampleRangeLimit: (t13 - t0) // shift + SampleOffset)].


! !

!JPEGReadWriter methodsFor: 'dct' stamp: 'ar 3/7/2001 00:58'!
idctMCU

	| comp fp ci |
	fp := self useFloatingPoint.
	1 to: mcuMembership size do:[:i|
		ci := mcuMembership at: i.
		comp := currentComponents at: ci.
		fp ifTrue:[
			self idctBlockFloat: (mcuSampleBuffer at: i) component: comp.
		] ifFalse:[
			self primIdctInt: (mcuSampleBuffer at: i)
				qt: (qTable at: comp qTableIndex)]].! !

!JPEGReadWriter methodsFor: 'marker parsing' stamp: 'ar 3/4/2001 18:38'!
initialSOSSetup

	mcuWidth := (components detectMax: [:c | c widthInBlocks]) widthInBlocks.
	mcuHeight := (components detectMax: [:c | c heightInBlocks]) heightInBlocks.
	components do:[:c |
		c mcuWidth: mcuWidth mcuHeight: mcuHeight dctSize: DCTSize].
	stream resetBitBuffer.! !

!JPEGReadWriter methodsFor: 'public access' stamp: 'tao 9/18/1998 08:53'!
nextImage

	^ self nextImageDitheredToDepth: Display depth
! !

!JPEGReadWriter methodsFor: 'public access' stamp: 'ar 10/28/2001 16:25'!
nextImageDitheredToDepth: depth

	| form xStep yStep x y bb |
	ditherMask := DitherMasks
		at: depth
		ifAbsent: [self error: 'can only dither to display depths'].
	residuals := WordArray new: 3.
	sosSeen := false.
	self parseFirstMarker.
	[sosSeen] whileFalse: [self parseNextMarker].
	form := Form extent: (width @ height) depth: depth.
	bb := BitBlt current toForm: form.
	bb sourceForm: mcuImageBuffer.
	bb colorMap: (mcuImageBuffer colormapIfNeededFor: form).
	bb sourceRect: mcuImageBuffer boundingBox.
	bb combinationRule: Form over.
	xStep := mcuWidth * DCTSize.
	yStep := mcuHeight * DCTSize.
	y := 0.
	1 to: mcuRowsInScan do:
		[:row |
		x := 0.
		1 to: mcusPerRow do:
			[:col |
			self decodeMCU.
			self idctMCU.
			self colorConvertMCU.
			bb destX: x; destY: y; copyBits.
			x := x + xStep].
		y := y + yStep].
	^ form! !

!JPEGReadWriter methodsFor: 'error handling' stamp: 'tao 10/19/97 12:25'!
notSupported: aString

	self error: aString , ' is not currently supported'! !

!JPEGReadWriter methodsFor: 'marker parsing' stamp: 'mir 6/13/2001 13:06'!
okToIgnoreMarker: aMarker

	^ (((16rE0 to: 16rEF) includes: aMarker) "unhandled APPn markers"
		or: [aMarker = 16rDC or: [aMarker = 16rFE]]) "DNL or COM markers"
		or: [aMarker = 16r99] "Whatever that is"! !

!JPEGReadWriter methodsFor: 'private' stamp: 'ar 3/4/2001 17:34'!
on: aStream
	super on: aStream.
	stream := JPEGReadStream on: stream upToEnd.! !

!JPEGReadWriter methodsFor: 'marker parsing' stamp: 'ar 3/6/2001 22:28'!
parseAPPn

	| length buffer thumbnailLength markerStart |
	markerStart := self position.
	length := self nextWord.
	buffer := self next: 4.
	(buffer asString = 'JFIF') ifFalse: [
		"Skip APPs that we're not interested in"
		stream next: length-6.
		^self].
	self next.
	majorVersion := self next.
	minorVersion := self next.
	densityUnit := self next.
	xDensity := self nextWord.
	yDensity := self nextWord.
	thumbnailLength := self next * self next * 3.
	length := length - (self position - markerStart).
	length = thumbnailLength ifFalse: [self error: 'APP0 thumbnail length is incorrect.'].
	self next: length! !

!JPEGReadWriter methodsFor: 'marker parsing' stamp: 'tao 10/20/97 13:36'!
parseDecoderRestartInterval

	| length |
	length := self nextWord.
	length = 4 ifFalse: [self error: 'DRI length incorrect'].
	restartInterval := self nextWord.! !

!JPEGReadWriter methodsFor: 'marker parsing' stamp: 'mir 6/12/2001 11:28'!
parseFirstMarker

	| marker |
	self next = 16rFF ifFalse: [self error: 'JFIF marker expected'].
	marker := self next.
	marker = 16rD9
		ifTrue: [^self "halt: 'EOI encountered.'"].
	marker = 16rD8 ifFalse: [self error: 'SOI marker expected'].
	self parseStartOfInput.
! !

!JPEGReadWriter methodsFor: 'marker parsing' stamp: 'ar 3/4/2001 18:36'!
parseHuffmanTable

	| length markerStart index bits count huffVal isACTable hTable |
	markerStart := self position.
	length := self nextWord.
	[self position - markerStart >= length] whileFalse:
		[index := self next.
		isACTable := (index bitAnd: 16r10) ~= 0.
		index := (index bitAnd: 16r0F) + 1.
		index > HuffmanTableSize
			ifTrue: [self error: 'image has more than ', HuffmanTableSize printString,
				' quantization tables'].
		bits := self next: 16.
		count := bits sum.
		(count > 256 or: [(count > (length - (self position - markerStart)))])
			ifTrue: [self error: 'Huffman Table count is incorrect'].
		huffVal := self next: count.
		hTable := stream buildLookupTable: huffVal counts: bits.
		isACTable
			ifTrue:
				[self hACTable at: index put: hTable]
			ifFalse:
				[self hDCTable at: index put: hTable]].! !

!JPEGReadWriter methodsFor: 'marker parsing' stamp: 'tao 10/24/97 17:32'!
parseNOP

	"don't need to do anything, here"! !

!JPEGReadWriter methodsFor: 'marker parsing' stamp: 'BG 3/16/2005 08:24'!
parseNextMarker
	"Parse the next marker of the stream"

	| byte discardedBytes |
	discardedBytes := 0.
	[(byte := self next) = 16rFF] whileFalse: [discardedBytes := discardedBytes + 1].	
	[[(byte := self next) = 16rFF] whileTrue. byte = 16r00] whileTrue:
		[discardedBytes := discardedBytes + 2].
	discardedBytes > 0 ifTrue: [self "notifyWithLabel: 'warning: extraneous data discarded'"].
	self perform:
		(JFIFMarkerParser
			at: byte
			ifAbsent:
				[(self okToIgnoreMarker: byte)
					ifTrue: [#skipMarker]
					ifFalse: [self error: 'marker ', byte printStringHex , ' cannot be handled']])! !

!JPEGReadWriter methodsFor: 'marker parsing' stamp: 'ar 3/3/2001 22:19'!
parseQuantizationTable

	| length markerStart n prec value table |
	markerStart := self position.
	length := self nextWord.
	[self position - markerStart >= length] whileFalse:
		[value := self next.
		n := (value bitAnd: 16r0F) + 1.
		prec := (value >> 4) > 0.
		n > QuantizationTableSize
			 ifTrue: [self error: 'image has more than ',
				QuantizationTableSize printString,
				' quantization tables'].
		table := IntegerArray new: DCTSize2.
		1 to: DCTSize2 do:
			[:i |
			value := (prec
				ifTrue: [self nextWord]
				ifFalse: [self next]).
			table at: (JPEGNaturalOrder at: i) put: value].
		self useFloatingPoint ifTrue: [self scaleQuantizationTable: table].
		self qTable at: n put: table]! !

!JPEGReadWriter methodsFor: 'marker parsing' stamp: 'tao 10/24/97 17:07'!
parseStartOfFile

	| length markerStart value n |
	markerStart := self position.
	length := self nextWord.
	dataPrecision := self next.
	dataPrecision = 8
		ifFalse: [self error: 'cannot handle ', dataPrecision printString, '-bit components'].
	height := self nextWord.
	width := self nextWord.
	n := self next.
	(height = 0) | (width = 0) | (n = 0) ifTrue: [self error: 'empty image'].
	(length - (self position - markerStart)) ~= (n * 3)
		ifTrue: [self error: 'component length is incorrect'].
	components := Array new: n.
	1 to: components size do:
		[:i |
		components
			at: i
			put:
				(JPEGColorComponent new
					id: self next;
					"heightInBlocks: (((value := self next) >> 4) bitAnd: 16r0F);
					widthInBlocks: (value bitAnd: 16r0F);"
					widthInBlocks: (((value := self next) >> 4) bitAnd: 16r0F);
					heightInBlocks: (value bitAnd: 16r0F);

					qTableIndex: self next + 1)]! !

!JPEGReadWriter methodsFor: 'marker parsing' stamp: 'tao 10/22/97 17:38'!
parseStartOfInput

	restartInterval := 0.
	densityUnit := 0.
	xDensity := 1.
	yDensity := 1
! !

!JPEGReadWriter methodsFor: 'marker parsing' stamp: 'tao 10/26/97 21:53'!
parseStartOfScan

	| length n id value dcNum acNum comp |
	length := self nextWord.
	n := self next.
	(length ~= (n*2 + 6)) | (n < 1) ifTrue: [self error: 'SOS length is incorrect'].
	currentComponents := Array new: n.
	1 to: n do: [:i |
		id := self next.
		value := self next.
		dcNum := (value >> 4) bitAnd: 16r0F.
		acNum := value bitAnd: 16r0F.
		comp := components detect: [:c | c id = id].
		comp
			dcTableIndex: dcNum+1;
			acTableIndex: acNum+1.
		currentComponents at: i put: comp].
	ss := self next.
	se := self next.
	value := self next.
	ah := (value >> 4) bitAnd: 16r0F.
	al := value bitAnd: 16r0F.
	self initialSOSSetup.
	self perScanSetup.
	sosSeen := true! !

!JPEGReadWriter methodsFor: 'marker parsing' stamp: 'ar 3/7/2001 01:01'!
perScanSetup

	mcusPerRow := (width / (mcuWidth * DCTSize)) ceiling.
	mcuRowsInScan := (height / (mcuHeight * DCTSize)) ceiling.
	(currentComponents size = 3 or: [currentComponents size = 1])
		ifFalse: [self error: 'JPEG color space not recognized'].
	mcuMembership := OrderedCollection new.
	currentComponents withIndexDo:
		[:c :i |
		c priorDCValue: 0.
		mcuMembership addAll: ((1 to: c totalMcuBlocks) collect: [:b | i])].
	mcuMembership := mcuMembership asArray.
	mcuSampleBuffer := (1 to: mcuMembership size) collect: [:i | IntegerArray new: DCTSize2].
	currentComponents withIndexDo:
		[:c :i |
			c initializeSampleStreamBlocks:
				((1 to: mcuMembership size)
					select: [:j | i = (mcuMembership at: j)]
					thenCollect: [:j | mcuSampleBuffer at: j])].
	mcuImageBuffer := Form
		extent: (mcuWidth @ mcuHeight) * DCTSize
		depth: 32.
	restartsToGo := restartInterval.! !

!JPEGReadWriter methodsFor: 'colorspace conversion' stamp: 'ar 3/4/2001 22:19'!
primColorConvertGrayscaleMCU
	self primColorConvertGrayscaleMCU: (currentComponents at: 1)
			bits: mcuImageBuffer bits
			residuals: residuals
			ditherMask: ditherMask.! !

!JPEGReadWriter methodsFor: 'colorspace conversion' stamp: 'ar 3/15/2001 18:11'!
primColorConvertGrayscaleMCU: componentArray bits: bits residuals: residualArray ditherMask: mask
	<primitive: 'primitiveColorConvertGrayscaleMCU' module: 'JPEGReaderPlugin'>
	"JPEGReaderPlugin doPrimitive: #primitiveColorConvertGrayscaleMCU."
	^self colorConvertGrayscaleMCU! !

!JPEGReadWriter methodsFor: 'colorspace conversion' stamp: 'ar 3/4/2001 21:36'!
primColorConvertIntYCbCrMCU
	self primColorConvertYCbCrMCU: currentComponents
			bits: mcuImageBuffer bits
			residuals: residuals
			ditherMask: ditherMask.! !

!JPEGReadWriter methodsFor: 'colorspace conversion' stamp: 'ar 3/4/2001 21:36'!
primColorConvertYCbCrMCU: componentArray bits: bits residuals: residualArray ditherMask: mask
	<primitive: 'primitiveColorConvertMCU' module: 'JPEGReaderPlugin'>
	^self colorConvertIntYCbCrMCU! !

!JPEGReadWriter methodsFor: 'huffman encoding' stamp: 'ar 3/4/2001 21:32'!
primDecodeBlockInto: sampleBuffer component: comp dcTable: dcTable acTable: acTable stream: jpegStream
	<primitive: 'primitiveDecodeMCU' module: 'JPEGReaderPlugin'>
	^self decodeBlockInto: sampleBuffer component: comp dcTable: dcTable acTable: acTable! !

!JPEGReadWriter methodsFor: 'dct' stamp: 'ar 3/4/2001 21:37'!
primIdctBlockInt: anArray component: aColorComponent
	^self primIdctInt: anArray qt: (self qTable at: aColorComponent qTableIndex)! !

!JPEGReadWriter methodsFor: 'dct' stamp: 'ar 3/4/2001 21:35'!
primIdctInt: anArray qt: qt
	<primitive: 'primitiveIdctInt' module: 'JPEGReaderPlugin'>
	^self idctBlockInt: anArray qt: qt! !

!JPEGReadWriter methodsFor: 'huffman encoding' stamp: 'ar 3/4/2001 17:40'!
processRestart
	stream resetBitBuffer.
	self parseNextMarker.
	currentComponents do: [:c | c priorDCValue: 0].
	restartsToGo := restartInterval.! !

!JPEGReadWriter methodsFor: 'accessing' stamp: 'tao 10/19/97 08:46'!
qTable

	qTable ifNil: [qTable := Array new: QuantizationTableSize].
	^ qTable! !

!JPEGReadWriter methodsFor: 'colorspace conversion' stamp: 'tao 10/26/97 15:43'!
sampleFloatRangeLimit: aNumber

	^ (aNumber rounded max: 0) min: MaxSample! !

!JPEGReadWriter methodsFor: 'colorspace conversion' stamp: 'di 9/15/1998 14:30'!
sampleRangeLimit: aNumber

	aNumber < 0 ifTrue: [^ 0].
	aNumber > MaxSample ifTrue: [^ MaxSample].
	^ aNumber! !

!JPEGReadWriter methodsFor: 'huffman encoding' stamp: 'ar 3/4/2001 01:17'!
scaleAndSignExtend: aNumber inFieldWidth: w

	aNumber < (1 bitShift: (w - 1))
		ifTrue: [^aNumber - (1 bitShift: w) + 1]
		ifFalse: [^aNumber]! !

!JPEGReadWriter methodsFor: 'dct' stamp: 'tao 10/26/97 16:16'!
scaleQuantizationTable: table

	| index |

	index := 1.
	1 to: DCTSize do:
		[:row |
		1 to: DCTSize do:
			[:col |
			table at: index
				put: ((table at: index) * (QTableScaleFactor at: row) *
					(QTableScaleFactor at: col)) rounded.
			index := index + 1]].
	^ table
! !

!JPEGReadWriter methodsFor: 'public access' stamp: 'ar 3/4/2001 17:26'!
setStream: aStream
	"Feed it in from an existing source"
	stream := JPEGReadStream on: aStream upToEnd.! !

!JPEGReadWriter methodsFor: 'marker parsing' stamp: 'tao 10/22/97 16:27'!
skipMarker

	| length markerStart |
	markerStart := self position.
	length := self nextWord.
	self next: length - (self position - markerStart)
! !

!JPEGReadWriter methodsFor: 'testing' stamp: 'ar 3/4/2001 00:50'!
understandsImageFormat
	"Answer true if the image stream format is understood by this decoder."
	self next = 16rFF ifFalse: [^ false].
	self next = 16rD8 ifFalse: [^ false].
	^ true
! !

!JPEGReadWriter methodsFor: 'preferences' stamp: 'tao 10/26/97 22:09'!
useFloatingPoint

	^ false! !

ImageReadWriter subclass: #JPEGReadWriter2
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Files'!

!JPEGReadWriter2 commentStamp: '<historical>' prior: 0!
I provide fast JPEG compression and decompression. I require the VM pluginJPEGReadWriter2Plugin, which is typically stored in same directory as the Squeak virtual machine.

JPEGReadWriter2Plugin is based on LIBJPEG library. This sentence applies to the plugin:
   "This software is based in part on the work of the Independent JPEG Group".

The LIBJPEG license allows it to be used free for any purpose so long as its origin and copyright are acknowledged. You can read more about LIBJPEG and get the complete source code at www.ijg.org.
!

!JPEGReadWriter2 class methodsFor: 'image reading/writing' stamp: 'ar 6/16/2002 18:54'!
primJPEGPluginIsPresent
	<primitive: 'primJPEGPluginIsPresent' module: 'JPEGReadWriter2Plugin'>
	^false! !

!JPEGReadWriter2 class methodsFor: 'image reading/writing' stamp: 'jm 12/22/2001 11:55'!
putForm: aForm quality: quality progressiveJPEG: progressiveFlag onFileNamed: fileName
	"Store the given Form as a JPEG file of the given name, overwriting any existing file of that name. Quality goes from 0 (low) to 100 (high), where -1 means default. If progressiveFlag is true, encode as a progressive JPEG."

	| writer |
	FileDirectory deleteFilePath: fileName.
	writer := self on: (FileStream newFileNamed: fileName) binary.
	Cursor write showWhile: [
		writer nextPutImage: aForm quality: quality progressiveJPEG: progressiveFlag].
	writer close.
! !

!JPEGReadWriter2 class methodsFor: 'image reading/writing' stamp: 'nk 7/16/2003 17:56'!
typicalFileExtensions
	"Answer a collection of file extensions (lowercase) which files that I can read might commonly have"
	^#('jpg' 'jpeg')! !

!JPEGReadWriter2 methodsFor: 'public access' stamp: 'JMM 1/11/2005 14:56'!
compress: aForm quality: quality
	"Encode the given Form and answer the compressed ByteArray. Quality goes from 0 (low) to 100 (high), where -1 means default."

	| sourceForm jpegCompressStruct jpegErrorMgr2Struct buffer byteCount |
	aForm unhibernate.
	"odd width images of depth 16 give problems; avoid them."
	sourceForm := (aForm depth = 32) | (aForm width even & (aForm depth = 16))
		ifTrue: [aForm]
		ifFalse: [aForm asFormOfDepth: 32].
	jpegCompressStruct := ByteArray new: self primJPEGCompressStructSize.
	jpegErrorMgr2Struct := ByteArray new: self primJPEGErrorMgr2StructSize.
	buffer := ByteArray new: sourceForm width * sourceForm height + 1024.
	byteCount := self primJPEGWriteImage: jpegCompressStruct 
		onByteArray: buffer
		form: sourceForm
		quality: quality
		progressiveJPEG: false
		errorMgr: jpegErrorMgr2Struct.
	byteCount = 0 ifTrue: [self error: 'buffer too small for compressed data'].
	^ buffer copyFrom: 1 to: byteCount
! !

!JPEGReadWriter2 methodsFor: 'public access' stamp: 'jm 11/20/2001 10:01'!
imageExtent: aByteArray 
	"Answer the extent of the compressed image encoded in the given ByteArray."

	| jpegDecompressStruct jpegErrorMgr2Struct w h |
	jpegDecompressStruct := ByteArray new: self primJPEGDecompressStructSize.
	jpegErrorMgr2Struct := ByteArray new: self primJPEGErrorMgr2StructSize.
	self
		primJPEGReadHeader: jpegDecompressStruct 
		fromByteArray: aByteArray
		errorMgr: jpegErrorMgr2Struct.
	w := self primImageWidth: jpegDecompressStruct.
	h := self primImageHeight: jpegDecompressStruct.
	^ w @ h
! !

!JPEGReadWriter2 methodsFor: 'testing' stamp: 'ar 11/27/2001 00:40'!
isPluginPresent
	^self primJPEGPluginIsPresent! !

!JPEGReadWriter2 methodsFor: 'public access' stamp: 'jm 11/20/2001 10:23'!
nextImage
	"Decode and answer a Form from my stream."

	^ self nextImageSuggestedDepth: Display depth
! !

!JPEGReadWriter2 methodsFor: 'public access' stamp: 'sd 1/30/2004 15:19'!
nextImageSuggestedDepth: depth
	"Decode and answer a Form of the given depth from my stream. Close the stream if it is a file stream. Possible depths are 16-bit and 32-bit."

	| bytes width height form jpegDecompressStruct jpegErrorMgr2Struct depthToUse |
	bytes := stream upToEnd.
	stream close.
	jpegDecompressStruct := ByteArray new: self primJPEGDecompressStructSize.
	jpegErrorMgr2Struct := ByteArray new: self primJPEGErrorMgr2StructSize.
	self 
		primJPEGReadHeader: jpegDecompressStruct 
		fromByteArray: bytes
		errorMgr: jpegErrorMgr2Struct.
	width := self primImageWidth: jpegDecompressStruct.
	height := self primImageHeight: jpegDecompressStruct.
	"Odd width images of depth 16 gave problems. Avoid them (or check carefully!!)"
	depthToUse := ((depth = 32) | width odd) ifTrue: [32] ifFalse: [16].
	form := Form extent: width@height depth: depthToUse.
	(width = 0 or: [height = 0]) ifTrue: [^ form].
	self
		primJPEGReadImage: jpegDecompressStruct
		fromByteArray: bytes
		onForm: form
		doDithering: true
		errorMgr: jpegErrorMgr2Struct.
	^ form
! !

!JPEGReadWriter2 methodsFor: 'public access' stamp: 'jm 11/20/2001 10:21'!
nextPutImage: aForm
	"Encode the given Form on my stream with default quality."

	^ self nextPutImage: aForm quality: -1 progressiveJPEG: false
! !

!JPEGReadWriter2 methodsFor: 'public access' stamp: 'JMM 1/11/2005 14:45'!
nextPutImage: aForm quality: quality progressiveJPEG: progressiveFlag
	"Encode the given Form on my stream with the given settings. Quality goes from 0 (low) to 100 (high), where -1 means default. If progressiveFlag is true, encode as a progressive JPEG."

	| sourceForm jpegCompressStruct jpegErrorMgr2Struct buffer byteCount |
	aForm unhibernate.
	"odd width images of depth 16 give problems; avoid them."
	sourceForm := (aForm depth = 32) | (aForm width even & (aForm depth = 16))
		ifTrue: [aForm]
		ifFalse: [aForm asFormOfDepth: 32].
	jpegCompressStruct := ByteArray new: self primJPEGCompressStructSize.
	jpegErrorMgr2Struct := ByteArray new: self primJPEGErrorMgr2StructSize.
	buffer := ByteArray new: sourceForm width * sourceForm height + 1024.
	byteCount := self primJPEGWriteImage: jpegCompressStruct 
		onByteArray: buffer
		form: sourceForm
		quality: quality
		progressiveJPEG: progressiveFlag
		errorMgr: jpegErrorMgr2Struct.
	byteCount = 0 ifTrue: [self error: 'buffer too small for compressed data'].
	stream next: byteCount putAll: buffer startingAt: 1.
	self close.
! !

!JPEGReadWriter2 methodsFor: 'primitives' stamp: 'jm 11/20/2001 10:34'!
primImageHeight: aJPEGCompressStruct

	<primitive: 'primImageHeight' module: 'JPEGReadWriter2Plugin'>
	self primitiveFailed
! !

!JPEGReadWriter2 methodsFor: 'primitives' stamp: 'jm 11/20/2001 10:35'!
primImageWidth: aJPEGCompressStruct

	<primitive: 'primImageWidth' module: 'JPEGReadWriter2Plugin'>
	self primitiveFailed
! !

!JPEGReadWriter2 methodsFor: 'primitives' stamp: 'jm 11/20/2001 10:35'!
primJPEGCompressStructSize

	<primitive: 'primJPEGCompressStructSize' module: 'JPEGReadWriter2Plugin'>
	self primitiveFailed
! !

!JPEGReadWriter2 methodsFor: 'primitives' stamp: 'jm 11/20/2001 10:35'!
primJPEGDecompressStructSize

	<primitive: 'primJPEGDecompressStructSize' module: 'JPEGReadWriter2Plugin'>
	self primitiveFailed
! !

!JPEGReadWriter2 methodsFor: 'primitives' stamp: 'jm 11/20/2001 10:35'!
primJPEGErrorMgr2StructSize

	<primitive: 'primJPEGErrorMgr2StructSize' module: 'JPEGReadWriter2Plugin'>
	self primitiveFailed
! !

!JPEGReadWriter2 methodsFor: 'primitives' stamp: 'ar 11/27/2001 00:39'!
primJPEGPluginIsPresent
	<primitive: 'primJPEGPluginIsPresent' module: 'JPEGReadWriter2Plugin'>
	^false! !

!JPEGReadWriter2 methodsFor: 'primitives' stamp: 'jm 11/20/2001 10:35'!
primJPEGReadHeader: aJPEGDecompressStruct fromByteArray: source errorMgr: aJPEGErrorMgr2Struct

	<primitive: 'primJPEGReadHeaderfromByteArrayerrorMgr' module: 'JPEGReadWriter2Plugin'>
	self primitiveFailed
! !

!JPEGReadWriter2 methodsFor: 'primitives' stamp: 'jmv 12/7/2001 13:45'!
primJPEGReadImage: aJPEGDecompressStruct fromByteArray: source onForm: form doDithering: ditherFlag errorMgr: aJPEGErrorMgr2Struct

	<primitive: 'primJPEGReadImagefromByteArrayonFormdoDitheringerrorMgr' module: 'JPEGReadWriter2Plugin'>
	self primitiveFailed
! !

!JPEGReadWriter2 methodsFor: 'primitives' stamp: 'jm 11/20/2001 10:35'!
primJPEGWriteImage: aJPEGCompressStruct onByteArray: destination form: form quality: quality progressiveJPEG: progressiveFlag errorMgr: aJPEGErrorMgr2Struct

	<primitive: 'primJPEGWriteImageonByteArrayformqualityprogressiveJPEGerrorMgr' module: 'JPEGReadWriter2Plugin'>
	self primitiveFailed
! !

!JPEGReadWriter2 methodsFor: 'public access' stamp: 'jmv 12/7/2001 13:49'!
uncompress: aByteArray into: aForm
	"Uncompress an image from the given ByteArray into the given Form.
	Fails if the given Form has the wrong dimensions or depth.
	If aForm has depth 16, do ordered dithering."

	| jpegDecompressStruct jpegErrorMgr2Struct w h |
	aForm unhibernate.
	jpegDecompressStruct := ByteArray new: self primJPEGDecompressStructSize.
	jpegErrorMgr2Struct := ByteArray new: self primJPEGErrorMgr2StructSize.
	self 
		primJPEGReadHeader: jpegDecompressStruct 
		fromByteArray: aByteArray
		errorMgr: jpegErrorMgr2Struct.
	w := self primImageWidth: jpegDecompressStruct.
	h := self primImageHeight: jpegDecompressStruct.
	((aForm width = w) & (aForm height = h)) ifFalse: [
		^ self error: 'form dimensions do not match'].

	"odd width images of depth 16 give problems; avoid them"
	w odd
		ifTrue: [
			aForm depth = 32 ifFalse: [^ self error: 'must use depth 32 with odd width']]
		ifFalse: [
			((aForm depth = 16) | (aForm depth = 32)) ifFalse: [^ self error: 'must use depth 16 or 32']].

	self primJPEGReadImage: jpegDecompressStruct
		fromByteArray: aByteArray
		onForm: aForm
		doDithering: true
		errorMgr: jpegErrorMgr2Struct.
! !

!JPEGReadWriter2 methodsFor: 'public access' stamp: 'jmv 12/7/2001 13:48'!
uncompress: aByteArray into: aForm doDithering: ditherFlag
	"Uncompress an image from the given ByteArray into the given Form. 
	Fails if the given Form has the wrong dimensions or depth.
	If aForm has depth 16 and ditherFlag = true, do ordered dithering."

	| jpegDecompressStruct jpegErrorMgr2Struct w h |
	aForm unhibernate.
	jpegDecompressStruct := ByteArray new: self primJPEGDecompressStructSize.
	jpegErrorMgr2Struct := ByteArray new: self primJPEGErrorMgr2StructSize.
	self 
		primJPEGReadHeader: jpegDecompressStruct 
		fromByteArray: aByteArray
		errorMgr: jpegErrorMgr2Struct.
	w := self primImageWidth: jpegDecompressStruct.
	h := self primImageHeight: jpegDecompressStruct.
	((aForm width = w) & (aForm height = h)) ifFalse: [
		^ self error: 'form dimensions do not match'].

	"odd width images of depth 16 give problems; avoid them"
	w odd
		ifTrue: [
			aForm depth = 32 ifFalse: [^ self error: 'must use depth 32 with odd width']]
		ifFalse: [
			((aForm depth = 16) | (aForm depth = 32)) ifFalse: [^ self error: 'must use depth 16 or 32']].

	self primJPEGReadImage: jpegDecompressStruct
		fromByteArray: aByteArray
		onForm: aForm
		doDithering: ditherFlag
		errorMgr: jpegErrorMgr2Struct.
! !

!JPEGReadWriter2 methodsFor: 'testing' stamp: 'ar 11/27/2001 00:39'!
understandsImageFormat
	"Answer true if the image stream format is understood by this decoder."
	self isPluginPresent ifFalse:[^false]. "cannot read it otherwise"
	self next = 16rFF ifFalse: [^ false].
	self next = 16rD8 ifFalse: [^ false].
	^ true
! !

ImageReadWriter subclass: #PCXReadWriter
	instanceVariableNames: 'version encoding colorPlanes isGrayScale width height bitsPerPixel colorPalette rowByteSize'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Files'!

!PCXReadWriter class methodsFor: 'image reading/writing' stamp: 'nk 7/16/2003 17:57'!
typicalFileExtensions
	"Answer a collection of file extensions (lowercase) which files that I can read might commonly have"
	^#('pcx')! !

!PCXReadWriter methodsFor: 'accessing' stamp: 'tao 10/6/97 10:11'!
nextImage
	"Read in the next PCX image from the stream."

	| bytes form |
	self readHeader.
	bytes := self readBody.
	colorPalette := self readPalette.
	self close.
	form := ColorForm extent: width@height depth: bitsPerPixel.
	(Form new hackBits: bytes) displayOn: (Form new hackBits: form bits).
	form colors: colorPalette.
	^ form
! !

!PCXReadWriter methodsFor: 'private-decoding' stamp: 'tao 10/6/97 08:38'!
nextWord
	^self next + (self next bitShift: 8)! !

!PCXReadWriter methodsFor: 'private-decoding' stamp: 'tao 10/6/97 10:07'!
readBody

	| array scanLine rowBytes position byte count pad |
	pad := #(0 3 2 1) at: (width \\ 4 + 1).
	array := ByteArray new: ((width + pad) * height * bitsPerPixel) // 8.
	scanLine := ByteArray new: rowByteSize.
	position := 1.
	1 to: height do:
		[:line |
		rowBytes := 0.
		[rowBytes < rowByteSize] whileTrue:
			[byte := self next.
			byte < 16rC0
				ifTrue:
					[rowBytes := rowBytes + 1.
					scanLine at: rowBytes put: byte]
				ifFalse:
					[count := byte - 16rC0.
					byte := self next.
					1 to: count do: [:i | scanLine at: rowBytes + i put: byte].
					rowBytes := rowBytes + count]].
		array
			replaceFrom: position
			to: position + width - 1
			with: scanLine
			startingAt: 1.
		position := position + width + pad].
	^ array
! !

!PCXReadWriter methodsFor: 'private-decoding' stamp: 'md 11/14/2003 16:51'!
readHeader

	| xMin xMax yMin yMax |
	self next.	"skip over manufacturer field"
	version := self next.
	encoding := self next.
	bitsPerPixel := self next.
	xMin := self nextWord.
	yMin := self nextWord.
	xMax := self nextWord.
	yMax := self nextWord.
	width := xMax - xMin + 1.
	height := yMax - yMin + 1.
	self next: 4. "skip over device resolution"
	self next: 49. "skip over EGA color palette"
	colorPlanes := self next.
	rowByteSize := self nextWord.
	isGrayScale := (self next: 2) = 2.
	self next: 58. "skip over filler"



! !

!PCXReadWriter methodsFor: 'private-decoding' stamp: 'tao 10/6/97 08:29'!
readPalette

	| r g b array |
	self next = 12 ifFalse: [self error: 'no Color Palette!!'].
	array := Array new: (1 bitShift: bitsPerPixel).
	1 to: array size do:
		[:i |
		r := self next.  g := self next.  b := self next.
		array at: i put: (Color r: r g: g b: b range: 255)].
	^ array.
! !

ImageReadWriter subclass: #PNGReadWriter
	instanceVariableNames: 'chunk form width height depth backColor bitsPerChannel colorType interlaceMethod bitsPerPixel bytesPerScanline thisScanline prevScanline rowSize globalDataChunk unknownChunks palette transparentPixelValue filtersSeen swizzleMap cachedDecoderMap bigEndian'
	classVariableNames: 'BPP BlockHeight BlockWidth Debugging StandardColors StandardSwizzleMaps'
	poolDictionaries: ''
	category: 'Graphics-Files'!

!PNGReadWriter commentStamp: '<historical>' prior: 0!
I am a subclass of ImageReadWriter that decodes Portable Network Graphics
(PNG) images.

Submitted by Duane Maxwell!

!PNGReadWriter class methodsFor: 'as yet unclassified' stamp: 'ar 2/11/2004 00:54'!
computeSwizzleMapForDepth: depth
	"Answer a map that maps pixels in a word to their opposite location. Used for 'middle-endian' forms where the byte-order is different from the bit order (good joke, eh?)."
	| map swizzled |
	map := Bitmap new: 256.
	depth = 4 ifTrue:[
		0 to: 255 do:[:pix|
			swizzled := 0.
			swizzled := swizzled bitOr: (((pix bitShift: 0) bitAnd: 15) bitShift: 4).
			swizzled := swizzled bitOr: (((pix bitShift: -4) bitAnd: 15) bitShift: 0).
			map at: pix+1 put: swizzled.
		].
		^ColorMap colors: map
	].

	depth = 2 ifTrue:[
		0 to: 255 do:[:pix|
			swizzled := 0.
			swizzled := swizzled bitOr: (((pix bitShift: 0) bitAnd: 3) bitShift: 6).
			swizzled := swizzled bitOr: (((pix bitShift: -2) bitAnd: 3) bitShift: 4).
			swizzled := swizzled bitOr: (((pix bitShift: -4) bitAnd: 3) bitShift: 2).
			swizzled := swizzled bitOr: (((pix bitShift: -6) bitAnd: 3) bitShift: 0).
			map at: pix+1 put: swizzled.
		].
		^ColorMap colors: map
	].

	depth = 1 ifTrue:[
		0 to: 255 do:[:pix|
			swizzled := 0.
			swizzled := swizzled bitOr: (((pix bitShift: 0) bitAnd: 1) bitShift: 7).
			swizzled := swizzled bitOr: (((pix bitShift: -1) bitAnd: 1) bitShift: 6).
			swizzled := swizzled bitOr: (((pix bitShift: -2) bitAnd: 1) bitShift: 5).
			swizzled := swizzled bitOr: (((pix bitShift: -3) bitAnd: 1) bitShift: 4).
			swizzled := swizzled bitOr: (((pix bitShift: -4) bitAnd: 1) bitShift: 3).
			swizzled := swizzled bitOr: (((pix bitShift: -5) bitAnd: 1) bitShift: 2).
			swizzled := swizzled bitOr: (((pix bitShift: -6) bitAnd: 1) bitShift: 1).
			swizzled := swizzled bitOr: (((pix bitShift: -7) bitAnd: 1) bitShift: 0).
			map at: pix+1 put: swizzled.
		].
		^ColorMap colors: map
	].
	self error: 'Unrecognized depth'! !

!PNGReadWriter class methodsFor: 'as yet unclassified' stamp: 'RAA 11/7/2000 09:22'!
createAFormFrom: data

	| error f |

	error := ''.
	f := [
		self formFromStream: (RWBinaryOrTextStream with: data)
	] ifError: [ :a :b |
		error := a printString,'  ',b printString.
		(StringMorph contents: error) color: Color red; imageForm
	].
	^{f. error}! !

!PNGReadWriter class methodsFor: 'as yet unclassified' stamp: 'RAA 11/7/2000 09:20'!
debugging: aBoolean

	Debugging := aBoolean! !

!PNGReadWriter class methodsFor: 'as yet unclassified' stamp: 'ar 2/11/2004 00:55'!
initialize
	"
	PNGReadWriter initialize
	"

	BPP := {	#(1 2 4 8 16).
			#(0 0 0 0 0).
			#(0 0 0 24 48).
			#(1 2 4 8 0).
			#(0 0 0 16 32).
			#(0 0 0 0 0).
			#(0 0 0 32 64).
			#(0 0 0 0 0) }.

	BlockHeight := #(8 8 4 4 2 2 1).
	BlockWidth := #(8 4 4 2 2 1 1).

	StandardColors := Color indexedColors collect:[:aColor|
		Color 
			r: (aColor red * 255) truncated / 255
			g: (aColor green * 255) truncated / 255
			b: (aColor blue * 255) truncated / 255.
	].

	StandardSwizzleMaps := Array new: 4.
	#(1 2 4) do:[:i| StandardSwizzleMaps at: i put: (self computeSwizzleMapForDepth: i)].! !

!PNGReadWriter class methodsFor: 'as yet unclassified' stamp: 'RAA 11/7/2000 09:15'!
insertMorph: aMorph named: aString into: aBook

	| newPage |

	aBook ifNil: [^self].
	newPage := aBook insertPageLabel: aString morphs: {aMorph}.
	newPage color: Color lightYellow.
	newPage extent: (
		newPage submorphs inject: 10@10 into: [ :ex :m |
			m left: 10.
			ex max: m width @ m bottom
		]
	) + (20@20).
! !

!PNGReadWriter class methodsFor: 'as yet unclassified' stamp: 'RAA 12/10/2000 10:11'!
test1
"PNGReadWriter test1"
	| data t error d0 d1 f fileInfo book result d2 |

	Debugging _ true.
	1 = 1 ifTrue: [
		book _ BookMorph new.
		book setProperty: #transitionSpec toValue: {'silence'. #none. #none}.
	].
	d0 _ FileDirectory default.
	d1 _ d0 directoryNamed: 'PngSuite Folder'.
	d2 _ d0 directoryNamed: 'BIG PNG'.
	{d0. d1. d2}.		"keep compiler quiet"
"==
citrus_none_sub.png
citrus_adm7_adap.png
citrus_adm7_aver.png
citrus_adm7_non.png
citrus_adm7_paeth.png
pngs-img-ie5mac.png
=="
	fileInfo _ {
		d2. {'citrus_adm7_adap.png'}.
		"d1. d1 fileNames."
	}.
	fileInfo pairsDo: [ :dir :fileNames |
		fileNames do: [ :each |
			Transcript cr; show: each.
			data _ (dir fileNamed: each) contentsOfEntireFile.
			error _ ''.
			MessageTally spyOn: [
				t _ [
					result _ self createAFormFrom: data.
					f_ result first.
					error _ result second.
				] timeToRun.].
			self insertMorph: f asMorph named: each into: book.
			Transcript show: each,'  ',data size printString,' = ',t printString,' ms',error; cr.
		].
	].
	book ifNotNil: [book openInWorld].
	Debugging _ false.! !

!PNGReadWriter class methodsFor: 'image reading/writing' stamp: 'nk 7/16/2003 17:57'!
typicalFileExtensions
	"Answer a collection of file extensions (lowercase) which files that I can read might commonly have"
	^#('png')! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'DSM 3/26/2000 21:32'!
copyPixels: y
	"Handle non-interlaced pixels of supported colorTypes"

	| s |
	s := #(copyPixelsGray: nil copyPixelsRGB: copyPixelsIndexed:
		  copyPixelsGrayAlpha: nil copyPixelsRGBA:) at: colorType+1.
	self perform: s asSymbol with: y
! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'RAA 11/4/2000 16:08'!
copyPixels: y at: startX by: incX
	"Handle interlaced pixels of supported colorTypes"

	| s |
	s := #(copyPixelsGray:at:by: nil copyPixelsRGB:at:by: copyPixelsIndexed:at:by:
		  copyPixelsGrayAlpha:at:by: nil copyPixelsRGBA:at:by:) at: colorType+1.
	self perform: s asSymbol with: y with: startX with: incX
! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'RAA 11/4/2000 16:12'!
copyPixelsGray: y 
	"Handle non-interlaced grayscale color mode (colorType = 0)"
	| blitter pixPerByte mask shifts pixelNumber rawByte pixel transparentIndex |
	blitter := BitBlt current bitPokerToForm: form.
	transparentIndex := form colors size.
	bitsPerChannel = 16
		ifTrue: [0
				to: width - 1
				do: [:x | blitter pixelAt: x @ y put: 255
							- (thisScanline at: x << 1 + 1)].
			^ self]
		ifFalse: [bitsPerChannel = 8
				ifTrue: [1
						to: width
						do: [:x | blitter
								pixelAt: x - 1 @ y
								put: (thisScanline at: x)].
					^ self].
			bitsPerChannel = 1
				ifTrue: [pixPerByte := 8.
					mask := 1.
					shifts := #(7 6 5 4 3 2 1 0 )].
			bitsPerChannel = 2
				ifTrue: [pixPerByte := 4.
					mask := 3.
					shifts := #(6 4 2 0 )].
			bitsPerChannel = 4
				ifTrue: [pixPerByte := 2.
					mask := 15.
					shifts := #(4 0 )].
			pixelNumber := 0.
			0 to: width - 1 do: [:x | 
				rawByte := thisScanline at: pixelNumber // pixPerByte + 1.
				pixel := rawByte
							>> (shifts at: pixelNumber \\ pixPerByte + 1) bitAnd: mask.
				pixel = transparentPixelValue ifTrue: [pixel := transparentIndex].
				blitter pixelAt: x @ y put: pixel.
				pixelNumber := pixelNumber + 1
			]
		]! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'RAA 11/4/2000 16:09'!
copyPixelsGray: y at: startX by: incX
	"Handle interlaced grayscale color mode (colorType = 0)"

	| b offset bits w pixel mask blitter pixelNumber pixPerByte rawByte
shifts |
	bitsPerChannel = 16
		ifTrue: [
			b := BitBlt current bitPokerToForm: form.
			startX to: width-1 by: incX do: [ :x |
				b pixelAt: x@y put: 255 - (thisScanline at: (x//incX<<1)+1).
				].
			^ self
			].
	offset := y*rowSize+1.
	bits := form bits.
	bitsPerChannel = 8 ifTrue: [
		startX to: width-1 by: incX do: [ :x |
			w := offset + (x>>2).
			b := 3- (x \\ 4) * 8.
			pixel := (thisScanline at: x // incX + 1)<<b.
			mask := (255<<b) bitInvert32.
			bits at: w put: (((bits at: w) bitAnd: mask) bitOr: pixel)
		].
		^ self
	].
	bitsPerChannel = 1 ifTrue: [
		pixPerByte := 8.
		mask := 1.
		shifts := #(7 6 5 4 3 2 1 0).
	].
	bitsPerChannel = 2 ifTrue: [
		pixPerByte := 4.
		mask := 3.
		shifts := #(6 4 2 0).
	].
	bitsPerChannel = 4 ifTrue: [
		pixPerByte := 2.
		mask := 15.
		shifts := #(4 0).
	].

	blitter := BitBlt current bitPokerToForm: form.
	pixelNumber := 0.
	startX to: width-1 by: incX do: [ :x |
		rawByte := thisScanline at: (pixelNumber // pixPerByte) + 1.
		pixel := (rawByte >> (shifts at: (pixelNumber \\ pixPerByte) + 1)) bitAnd: mask.
		blitter pixelAt: (x@y) put: pixel.
		pixelNumber := pixelNumber + 1.
	].
! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'RAA 11/4/2000 16:09'!
copyPixelsGrayAlpha: y
	"Handle non-interlaced grayscale with alpha color mode (colorType = 4)"

	| i pixel gray b |
	b := BitBlt current bitPokerToForm: form.
	bitsPerChannel = 8
		ifTrue: [
			0 to: width-1 do: [ :x |
				i := (x << 1) + 1.
				gray := thisScanline at: i.
				pixel := ((thisScanline at: i+1)<<24) + (gray<<16) + (gray<<8) + gray.
				b pixelAt: x@y put: pixel.
				]
			]
		ifFalse: [
			0 to: width-1 do: [ :x |
				i := (x << 2) + 1.
				gray := thisScanline at: i.
				pixel := ((thisScanline at: i+2)<<24) + (gray<<16) + (gray<<8) + gray.
				b pixelAt: x@y put: pixel.
				]
			]
! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'RAA 11/4/2000 16:09'!
copyPixelsGrayAlpha: y at: startX by: incX
	"Handle interlaced grayscale with alpha color mode (colorType = 4)"

	| i pixel gray b |
	b := BitBlt current bitPokerToForm: form.
	bitsPerChannel = 8
		ifTrue: [
			startX to: width-1 by: incX do: [ :x |
				i := (x // incX << 1) + 1.
				gray := thisScanline at: i.
				pixel := ((thisScanline at: i+1)<<24) + (gray<<16) + (gray<<8) + gray.
				b pixelAt: x@y put: pixel.
				]
			]
		ifFalse: [
			startX to: width-1 by: incX do: [ :x |
				i := (x // incX << 2) + 1.
				gray := thisScanline at: i.
				pixel := ((thisScanline at: i+2)<<24) + (gray<<16) + (gray<<8) + gray.
				b pixelAt: x@y put: pixel.
				]
			]
! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'ar 1/1/1970 21:00'!
copyPixelsIndexed: y
	"Handle non-interlaced indexed color mode (colorType = 3)"
	| hack hackBlt swizzleHack swizzleBlt scanline hackDepth |
	scanline := ByteArray new: bytesPerScanline + 3 // 4 * 4.
	scanline replaceFrom: 1 to: thisScanline size with: thisScanline startingAt: 1.
	hackDepth := bigEndian ifTrue:[form depth] ifFalse:[form depth negated].
	hack := Form extent: width@1 depth: hackDepth bits: scanline.
	hackBlt := BitBlt toForm: form.
	hackBlt sourceForm: hack.
	hackBlt combinationRule: Form over.
	hackBlt destOrigin: 0@y.
	hackBlt width: width; height: 1.

	(form depth < 8 and:[bigEndian not]) ifTrue:[
		swizzleHack := Form new hackBits: scanline.
		swizzleBlt := BitBlt toForm: swizzleHack.
		swizzleBlt sourceForm: swizzleHack.
		swizzleBlt combinationRule: Form over.
		swizzleBlt colorMap: (StandardSwizzleMaps at: form depth).
		swizzleBlt copyBits.
	].

	hackBlt copyBits.! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'RAA 11/4/2000 16:23'!
copyPixelsIndexed: y at: startX by: incX
	"Handle interlaced indexed color mode (colorType = 3)"

	| offset b bits w pixel mask pixPerByte shifts blitter pixelNumber rawByte |
	offset := y*rowSize+1.
	bits := form bits.
	bitsPerChannel = 8
		ifTrue: [
			startX to: width-1 by: incX do: [ :x |
				w := offset + (x>>2).
				b := 3 - (x \\ 4) * 8.
				pixel := (thisScanline at: x // incX + 1)<<b.
				mask := (255<<b) bitInvert32.
				bits at: w put: (((bits at: w) bitAnd: mask) bitOr: pixel)].
			^ self ].
	bitsPerChannel = 1 ifTrue: [
		pixPerByte := 8.
		mask := 1.
		shifts := #(7 6 5 4 3 2 1 0).
	].
	bitsPerChannel = 2 ifTrue: [
		pixPerByte := 4.
		mask := 3.
		shifts := #(6 4 2 0).
	].
	bitsPerChannel = 4 ifTrue: [
		pixPerByte := 2.
		mask := 15.
		shifts := #(4 0).
	].

	blitter := BitBlt current bitPokerToForm: form.
	pixelNumber := 0.
	startX to: width-1 by: incX do: [ :x |
		rawByte := thisScanline at: (pixelNumber // pixPerByte) + 1.
		pixel := (rawByte >> (shifts at: (pixelNumber \\ pixPerByte) + 1)) bitAnd: mask.
		blitter pixelAt: (x@y) put: pixel.
		pixelNumber := pixelNumber + 1.
	].
! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'RAA 11/7/2000 09:30'!
copyPixelsRGB: y
	"Handle non-interlaced RGB color mode (colorType = 2)"

	| i pixel tempForm tempBits |

	tempForm := Form extent: width@1 depth: 32.
	tempBits := tempForm bits.
	pixel := LargePositiveInteger new: 4.
	pixel at: 4 put: 16rFF.
	bitsPerChannel = 8 ifTrue: [
		i := 1.
		1 to: width do: [ :x |
			pixel
				at: 3 put: (thisScanline at: i);
				at: 2 put: (thisScanline at: i+1);
				at: 1 put: (thisScanline at: i+2).
			tempBits at: x put: pixel.
			i := i + 3.
		]
	] ifFalse: [
		i := 1.
		1 to: width do: [ :x |
			pixel
				at: 3 put: (thisScanline at: i);
				at: 2 put: (thisScanline at: i+2);
				at: 1 put: (thisScanline at: i+4).
			tempBits at: x put: pixel.
			i := i + 6.
		]
	].
	transparentPixelValue ifNotNil: [
		1 to: width do: [ :x |
			(tempBits at: x) = transparentPixelValue ifTrue: [
				tempBits at: x put: 0.
			].
		].
	].
	tempForm displayOn: form at: 0@y rule: Form paint.
! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'nk 7/27/2004 17:18'!
copyPixelsRGB: y at: startX by: incX
	"Handle interlaced RGB color mode (colorType = 2)"

	| i pixel tempForm tempBits xx loopsToDo |

	tempForm := Form extent: width@1 depth: 32.
	tempBits := tempForm bits.
	pixel := LargePositiveInteger new: 4.
	pixel at: 4 put: 16rFF.
	loopsToDo := width - startX + incX - 1 // incX.
	bitsPerChannel = 8 ifTrue: [
		i := (startX // incX * 3) + 1.
		xx := startX+1.
		1 to: loopsToDo do: [ :j |
			pixel
				at: 3 put: (thisScanline at: i);
				at: 2 put: (thisScanline at: i+1);
				at: 1 put: (thisScanline at: i+2).
			tempBits at: xx put: pixel.
			i := i + 3.
			xx := xx + incX.
		]
	] ifFalse: [
		i := (startX // incX * 6) + 1.
		xx := startX+1.
		1 to: loopsToDo do: [ :j |
			pixel
				at: 3 put: (thisScanline at: i);
				at: 2 put: (thisScanline at: i+2);
				at: 1 put: (thisScanline at: i+4).
			tempBits at: xx put: pixel.
			i := i + 6.
			xx := xx + incX.
		].
	].
	transparentPixelValue ifNotNil: [
		startX to: width-1 by: incX do: [ :x |
			(tempBits at: x+1) = transparentPixelValue ifTrue: [
				tempBits at: x+1 put: 0.
			].
		].
	].
	tempForm displayOn: form at: 0@y rule: Form paint.

! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'ar 2/18/2004 23:58'!
copyPixelsRGBA: y
	"Handle non-interlaced RGBA color modes (colorType = 6)"

	| i pixel tempForm tempBits ff |
	bitsPerChannel = 8 ifTrue: [
		ff := Form extent: width@1 depth: 32 bits: thisScanline.
		cachedDecoderMap 
			ifNil:[cachedDecoderMap := self rgbaDecoderMapForDepth: depth].
		(BitBlt toForm: form)
			sourceForm: ff;
			destOrigin: 0@y;
			combinationRule: Form over;
			colorMap: cachedDecoderMap;
			copyBits.
		^self.
	].
	tempForm := Form extent: width@1 depth: 32.
	tempBits := tempForm bits.
	pixel := LargePositiveInteger new: 4.
	i := -7.
	0 to: width-1 do: [ :x |
			i := i + 8.
			pixel at: 4 put: (thisScanline at: i+6);
				at: 3 put: (thisScanline at: i);
				at: 2 put: (thisScanline at: i+2);
				at: 1 put: (thisScanline at: i+4).
			tempBits at: x+1 put: pixel.
	].
	tempForm displayOn: form at: 0@y rule: Form over.
! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'nk 7/27/2004 17:57'!
copyPixelsRGBA: y at: startX by: incX
	"Handle interlaced RGBA color modes (colorType = 6)"

	| i pixel tempForm tempBits |

	tempForm := Form extent: width@1 depth: 32.
	tempBits := tempForm bits.
	pixel := LargePositiveInteger new: 4.
	bitsPerChannel = 8 ifTrue: [
		i := (startX // incX << 2) + 1.
		startX to: width-1 by: incX do: [ :x |
			pixel at: 4 put: (thisScanline at: i+3);
				at: 3 put: (thisScanline at: i);
				at: 2 put: (thisScanline at: i+1);
				at: 1 put: (thisScanline at: i+2).
			tempBits at: x+1 put: pixel.
			i := i + 4.
		]
	] ifFalse: [
		i := (startX // incX << 3) +1.
		startX to: width-1 by: incX do: [ :x |
			pixel at: 4 put: (thisScanline at: i+6);
				at: 3 put: (thisScanline at: i);
				at: 2 put: (thisScanline at: i+2);
				at: 1 put: (thisScanline at: i+4).
			tempBits at: x+1 put: pixel.
			i := i + 8.
		].
	].
	tempForm displayOn: form at: 0@y rule: Form paintAlpha.

! !

!PNGReadWriter methodsFor: 'accessing' stamp: 'RAA 11/7/2000 09:20'!
debugging

	^Debugging == true! !

!PNGReadWriter methodsFor: 'miscellaneous' stamp: 'DSM 4/27/2000 13:09'!
doPass: pass
	"Certain interlace passes are skipped with certain small image
dimensions"

	pass = 1 ifTrue: [ ^ true ].
	((width = 1) and: [height = 1]) ifTrue: [ ^ false ].
	pass = 2 ifTrue: [ ^ width >= 5 ].
	pass = 3 ifTrue: [ ^ height >= 5 ].
	pass = 4 ifTrue: [ ^ (width >=3 ) or: [height >= 5] ].
	pass = 5 ifTrue: [ ^ height >=3 ].
	pass = 6 ifTrue: [ ^ width >=2 ].
	pass = 7 ifTrue: [ ^ height >=2 ].

! !

!PNGReadWriter methodsFor: 'filtering' stamp: 'RAA 11/7/2000 09:43'!
filterAverage: count
	"Use the average of the pixel to the left and the pixel above as a predictor"

	| delta |
	delta := bitsPerPixel // 8 max: 1.
	1 to: delta do: [:i |
		thisScanline at: i put: ((thisScanline at: i) + ((prevScanline at: i) // 2) bitAnd: 255)].
	delta + 1 to: count do: [:i |
		thisScanline at: i put:
			((thisScanline at: i)
			+ ((prevScanline at: i)
			+ (thisScanline at: i - delta) // 2) bitAnd: 255)]! !

!PNGReadWriter methodsFor: 'filtering' stamp: 'DSM 3/25/2000 17:54'!
filterHorizontal: count
	"Use the pixel to the left as a predictor"

	| delta |
	delta := bitsPerPixel // 8 max: 1.
	delta+1 to: count do: [ :i |
		thisScanline at: i put: (((thisScanline at: i) +
(thisScanline at: i-delta)) bitAnd: 255) ]


! !

!PNGReadWriter methodsFor: 'filtering' stamp: 'DSM 3/25/2000 17:55'!
filterNone: count
! !

!PNGReadWriter methodsFor: 'filtering' stamp: 'RAA 11/7/2000 09:45'!
filterPaeth: count
	"Select one of (the pixel to the left, the pixel above and the pixel to above left) to
	predict the value of this pixel"

	| delta |
	delta := bitsPerPixel // 8 max: 1.
	1 to: delta do: [ :i |
		thisScanline at: i put:
			(((thisScanline at: i) + (prevScanline at: i)) bitAnd: 255)].
	delta+1 to: count do: [ :i |
		thisScanline
			at: i
			put: (((thisScanline at: i) + (self
				paethPredictLeft: (thisScanline at: i-delta)
				above: (prevScanline at: i)
				aboveLeft: (prevScanline at: i-delta)))
					bitAnd: 255)]

! !

!PNGReadWriter methodsFor: 'filtering' stamp: 'eat 9/11/2000 20:08'!
filterScanline: filterType count: count

	self
		perform: (
			#(filterNone: filterHorizontal: filterVertical: filterAverage: filterPaeth:)
				at: filterType+1)
		with: count.

! !

!PNGReadWriter methodsFor: 'filtering' stamp: 'DSM 3/25/2000 17:54'!
filterVertical: count
	"Use the pixel above as a predictor"

	1 to: count do: [ :i |
		thisScanline at: i put: (((thisScanline at: i) +
(prevScanline at: i)) bitAnd: 255) ]

! !

!PNGReadWriter methodsFor: 'miscellaneous' stamp: 'ar 2/11/2004 01:27'!
grayColorsFor: d
	"return a color table for a gray image"

	palette := Array new: 1<<d.
	d = 1 ifTrue: [
		palette at: 1 put: Color black.
		palette at: 2 put: Color white.
		^ palette,{Color transparent}
		].
	d = 2 ifTrue: [
		palette at: 1 put: Color black.
		palette at: 2 put: (Color gray: 85.0 / 255.0).
		palette at: 3 put: (Color gray: 170.0 / 255.0).
		palette at: 4 put: Color white.
		^ palette,{Color transparent}.
		].
	d = 4 ifTrue: [
		0 to: 15 do: [ :g |
			palette at: g+1 put: (Color gray: (g/15) asFloat) ].
		^ palette,{Color transparent}
		].
	d = 8 ifTrue: [
		0 to: 255 do: [ :g |
			palette at: g+1 put: (Color gray: (g/255) asFloat) ].
		^ palette		"??transparent??"
		].
! !

!PNGReadWriter methodsFor: 'accessing' stamp: 'nk 7/30/2004 17:51'!
nextImage
	bigEndian := SmalltalkImage current isBigEndian.
	filtersSeen := Bag new.
	globalDataChunk := nil.
	transparentPixelValue := nil.
	unknownChunks := Set new.
	stream reset.
	stream binary.
	stream skip: 8.
	[stream atEnd] whileFalse: [self processNextChunk].
	"Set up our form"
	palette ifNotNil: 
			["Dump the palette if it's the same as our standard palette"

			palette = (StandardColors copyFrom: 1 to: palette size) 
				ifTrue: [palette := nil]].
	(depth <= 8 and: [palette notNil]) 
		ifTrue: 
			[form := ColorForm extent: width @ height depth: depth.
			form colors: palette]
		ifFalse: [form := Form extent: width @ height depth: depth].
	backColor ifNotNil: [form fillColor: backColor].
	chunk := globalDataChunk ifNil: [self error: 'image data is missing'].
	chunk ifNotNil: [self processIDATChunk].
	unknownChunks isEmpty 
		ifFalse: 
			["Transcript show: ' ',unknownChunks asSortedCollection asArray printString."

			].
	self debugging 
		ifTrue: 
			[Transcript
				cr;
				show: 'form = ' , form printString.
			Transcript
				cr;
				show: 'colorType = ' , colorType printString.
			Transcript
				cr;
				show: 'interlaceMethod = ' , interlaceMethod printString.
			Transcript
				cr;
				show: 'filters = ' , filtersSeen sortedCounts asArray printString].
	^form! !

!PNGReadWriter methodsFor: 'writing' stamp: 'ar 12/12/2003 16:37'!
nextPutImage: aForm
	"Write out the given form. We're keeping it simple here, no interlacing, no filters."
	^self nextPutImage: aForm interlace: 0 filter: 0. "no filtering"! !

!PNGReadWriter methodsFor: 'writing' stamp: 'nk 7/30/2004 17:51'!
nextPutImage: aForm interlace: aMethod filter: aFilterType 
	"Note: For now we keep it simple - interlace and filtering are simply ignored"

	| crcStream |
	bigEndian := SmalltalkImage current isBigEndian.
	form := aForm.
	width := aForm width.
	height := aForm height.
	aForm depth <= 8 
		ifTrue: 
			[bitsPerChannel := aForm depth.
			colorType := 3.
			bytesPerScanline := (width * aForm depth + 7) // 8]
		ifFalse: 
			[bitsPerChannel := 8.
			colorType := 6.
			bytesPerScanline := width * 4].
	self writeFileSignature.
	crcStream := WriteStream on: (ByteArray new: 1000).
	crcStream resetToStart.
	self writeIHDRChunkOn: crcStream.
	self writeChunk: crcStream.
	form depth <= 8 
		ifTrue: 
			[crcStream resetToStart.
			self writePLTEChunkOn: crcStream.
			self writeChunk: crcStream.
			form isColorForm 
				ifTrue: 
					[crcStream resetToStart.
					self writeTRNSChunkOn: crcStream.
					self writeChunk: crcStream]].
	form depth = 16 
		ifTrue: 
			[crcStream resetToStart.
			self writeSBITChunkOn: crcStream.
			self writeChunk: crcStream].
	crcStream resetToStart.
	self writeIDATChunkOn: crcStream.
	self writeChunk: crcStream.
	crcStream resetToStart.
	self writeIENDChunkOn: crcStream.
	self writeChunk: crcStream! !

!PNGReadWriter methodsFor: 'filtering' stamp: 'eat 9/11/2000 20:05'!
paethPredictLeft: a above: b aboveLeft: c
	"Predicts the value of a pixel based on nearby pixels, based on
Paeth (GG II, 1991)"

	| pa pb pc |
	pa := b > c ifTrue: [b - c] ifFalse: [c - b].
	pb := a > c ifTrue: [a - c] ifFalse: [c - a].
	pc := a + b - c - c.
	pc < 0 ifTrue: [
		pc := pc * -1].
	((pa <= pb) and: [pa <= pc]) ifTrue: [^ a].
	(pb <= pc) ifTrue: [^ b].
	^ c
! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'ar 2/10/2004 23:55'!
processBackgroundChunk

	| val red green blue max |

	"Transcript show: '  BACKGROUND: ',chunk printString."
	colorType = 3 ifTrue: [
		backColor := palette at: chunk first + 1.
		^self
	].
	max := (2 raisedTo: bitsPerChannel) - 1.
	(colorType = 0 or: [colorType = 4]) ifTrue: [
		val := chunk unsignedShortAt: 1 bigEndian: true.
		backColor := Color gray: val / max.
		^self
	].
	(colorType = 2 or: [colorType = 6]) ifTrue: [
		red := chunk unsignedShortAt: 1 bigEndian: true.
		green := chunk unsignedShortAt: 3 bigEndian: true.
		blue := chunk unsignedShortAt: 5 bigEndian: true.
		backColor := Color r: red/max g: green/max b: blue/max.
		^self
	].
"self halt."

"====
The bKGD chunk specifies a default background color to present the image against. Note that viewers are not bound to honor this chunk; a viewer can choose to use a different background. 

For color type 3 (indexed color), the bKGD chunk contains: 


   Palette index:  1 byte

The value is the palette index of the color to be used as background. 

For color types 0 and 4 (grayscale, with or without alpha), bKGD contains: 


   Gray:  2 bytes, range 0 .. (2^bitdepth)-1

(For consistency, 2 bytes are used regardless of the image bit depth.) The value is the gray level to be used as background. 

For color types 2 and 6 (truecolor, with or without alpha), bKGD contains: 


   Red:   2 bytes, range 0 .. (2^bitdepth)-1
   Green: 2 bytes, range 0 .. (2^bitdepth)-1
   Blue:  2 bytes, range 0 .. (2^bitdepth)-1

(For consistency, 2 bytes per sample are used regardless of the image bit depth.) This is the RGB color to be used as background. 

When present, the bKGD chunk must precede the first IDAT chunk, and must follow the PLTE chunk, if any. 
==="
! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'RAA 11/4/2000 17:00'!
processIDATChunk

	interlaceMethod = 0
		ifTrue: [ self processNonInterlaced ]
		ifFalse: [ self processInterlaced ]
! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'ar 2/10/2004 23:55'!
processIHDRChunk
	width := chunk longAt: 1 bigEndian: true.
	height := chunk longAt: 5 bigEndian: true.
	bitsPerChannel := chunk at: 9.
	colorType := chunk at: 10.
	"compression := chunk at: 11." "TODO - validate compression"
	"filterMethod := chunk at: 12." "TODO - validate filterMethod"
	interlaceMethod := chunk at: 13. "TODO - validate interlace method"
	(#(2 4 6) includes: colorType)
		ifTrue: [depth := 32].
	(#(0 3) includes: colorType) ifTrue: [
		depth := bitsPerChannel min: 8.
		colorType = 0 ifTrue: [ "grayscale"
			palette := self grayColorsFor: depth.
		].
	].
	bitsPerPixel := (BPP at: colorType+1) at: bitsPerChannel highBit.
	bytesPerScanline := width * bitsPerPixel + 7 // 8.
	rowSize := width * depth + 31 >> 5.
! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'ar 2/29/2004 04:19'!
processInterlaced
	| z filter bytesPerPass startingCol colIncrement rowIncrement startingRow cx sc temp |
	startingCol := #(0 4 0 2 0 1 0 ).
	colIncrement := #(8 8 4 4 2 2 1 ).
	rowIncrement := #(8 8 8 4 4 2 2 ).
	startingRow := #(0 0 4 0 2 0 1 ).
	z := ZLibReadStream on: chunk from: 1 to: chunk size.
	1 to: 7 do: [:pass |
		(self doPass: pass)
			ifTrue:
				[cx := colIncrement at: pass.
				sc := startingCol at: pass.
				bytesPerPass := width - sc + cx - 1 // cx * bitsPerPixel + 7 // 8.
				prevScanline := ByteArray new: bytesPerPass.
				thisScanline := ByteArray new: bytesPerScanline.
				(startingRow at: pass)
					to: height - 1
					by: (rowIncrement at: pass)
					do: [:y |
						filter := z next.
						filtersSeen add: filter.
						(filter isNil or: [(filter between: 0 and: 4) not])
							ifTrue: [^ self].
						thisScanline := z next: bytesPerPass into: thisScanline startingAt: 1.
						self filterScanline: filter count: bytesPerPass.
						self copyPixels: y at: sc by: cx.
						temp := prevScanline.
						prevScanline := thisScanline.
						thisScanline := temp.
					]
				]
	].
	z atEnd ifFalse:[self error:'Unexpected data'].! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'ar 2/11/2004 12:14'!
processNextChunk

	| length chunkType crc chunkCrc |

	length := self nextLong.

	chunkType := (self next: 4) asString.
	chunk := self next: length.
	chunkCrc := self nextLong bitXor: 16rFFFFFFFF.
	crc := self updateCrc: 16rFFFFFFFF from: 1 to: 4 in: chunkType.
	crc := self updateCrc: crc from: 1 to: length in: chunk.
	crc = chunkCrc ifFalse:[
		self error: 'PNGReadWriter crc error in chunk ', chunkType.
	].

	chunkType = 'IEND' ifTrue: [^self	"*should* be the last chunk"].
	chunkType = 'sBIT' ifTrue: [^self processSBITChunk "could indicate unusual sample depth in original"].
	chunkType = 'gAMA' ifTrue: [^self 	"indicates gamma correction value"].
	chunkType = 'bKGD' ifTrue: [^self processBackgroundChunk].
	chunkType = 'pHYs' ifTrue: [^self processPhysicalPixelChunk].
	chunkType = 'tRNS' ifTrue: [^self processTransparencyChunk].

	chunkType = 'IHDR' ifTrue: [^self processIHDRChunk].
	chunkType = 'PLTE' ifTrue: [^self processPLTEChunk].
	chunkType = 'IDAT' ifTrue: [
		"---since the compressed data can span multiple
		chunks, stitch them all together first. later,
		if memory is an issue, we need to figure out how
		to do this on the fly---"
		globalDataChunk := globalDataChunk ifNil: [chunk] ifNotNil:
			[globalDataChunk,chunk].
		^self
	].
	unknownChunks add: chunkType.
! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'ar 2/29/2004 04:19'!
processNonInterlaced
	| z filter temp copyMethod debug |
	debug := self debugging.
	copyMethod := #(copyPixelsGray: nil copyPixelsRGB: copyPixelsIndexed:
		  copyPixelsGrayAlpha: nil copyPixelsRGBA:) at: colorType+1.
	debug ifTrue: [ Transcript cr; nextPutAll: 'NI chunk size='; print: chunk size ].
	z := ZLibReadStream on: chunk from: 1 to: chunk size.
	prevScanline := ByteArray new: bytesPerScanline.
	thisScanline := ByteArray new: bytesPerScanline.
	0 to: height-1 do: [ :y |
		filter := (z next: 1) first.
		debug ifTrue:[filtersSeen add: filter].
		thisScanline := z next: bytesPerScanline into: thisScanline startingAt: 1.
		(debug and: [ thisScanline size < bytesPerScanline ]) ifTrue: [ Transcript nextPutAll: ('wanted {1} but only got {2}' format: { bytesPerScanline. thisScanline size }); cr ].
		filter = 0 ifFalse:[self filterScanline: filter count: bytesPerScanline].
		self perform: copyMethod with: y.
		temp := prevScanline.
		prevScanline := thisScanline.
		thisScanline := temp.
		].
	z atEnd ifFalse:[self error:'Unexpected data'].
	debug ifTrue: [Transcript  nextPutAll: ' compressed size='; print: z position  ].
! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'ar 2/11/2004 01:02'!
processPLTEChunk

	| colorCount i |

	colorCount := chunk size // 3. "TODO - validate colorCount against depth"
	palette := Array new: colorCount.
	0 to: colorCount-1 do: [ :index |
		i := index * 3 + 1.
		palette at: index+1 put:
			(Color r: (chunk at: i)/255.0 g: (chunk at: i+1)/255.0 b: (chunk at: i+2)/255.0)
		].! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'RAA 11/5/2000 11:24'!
processPhysicalPixelChunk

	"Transcript show: '  PHYSICAL: ',chunk printString."
! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'ar 12/12/2003 18:33'!
processSBITChunk
	| rBits gBits bBits aBits |
	colorType = 6 ifFalse:[^self].
	rBits := chunk at: 1.
	gBits := chunk at: 2.
	bBits := chunk at: 3.
	aBits := chunk at: 4.
	(rBits = 5 and:[gBits = 5 and:[bBits = 5 and:[aBits = 1]]]) ifTrue:[
		depth := 16.
	].! !

!PNGReadWriter methodsFor: 'chunks' stamp: 'RAA 11/4/2000 16:22'!
processTransparencyChunk

	| red green blue |

	"Transcript show: '  TRANSPARENCY ',chunk printString."
	colorType = 0 ifTrue: [
		transparentPixelValue := chunk unsignedShortAt: 1 bigEndian: true.
		^self
	].
	colorType = 2 ifTrue: [
		red := chunk at: 2.
		green := chunk at: 2.
		blue := chunk at: 2.
		transparentPixelValue := 16rFF00 + red << 8 + green << 8 + blue.
		^self
	].
	colorType = 3 ifTrue: [
		chunk withIndexDo: [ :alpha :index |
			palette at: index put: ((palette at: index) alpha: alpha/255)
		].
		^self
	].
! !

!PNGReadWriter methodsFor: 'pixel copies' stamp: 'ar 2/19/2004 00:10'!
rgbaDecoderMapForDepth: decoderDepth
	bigEndian ifTrue:[
		depth = 16 ifTrue:[
			"Big endian, 32 -> 16 color mapping."
			^ColorMap
				shifts: #(-17 -14 -11 0)
				masks: #(16rF8000000 16rF80000 16rF800 16r00)
		] ifFalse:[
			"Big endian, 32 -> 32 color mapping"
			^ColorMap 
				shifts: #(-8 -8 -8 24) 
				masks: #(16rFF000000 16rFF0000 16rFF00 16rFF).
		].
	].
	depth = 16 ifTrue:[
		"Little endian, 32 -> 16 color mapping."
		^ColorMap
			shifts: #(7 -6 -19 0)
			masks: #(16rF8 16rF800 16rF80000 0)
	] ifFalse:[
		"Little endian, 32 -> 32 color mapping"
		^ColorMap 
			shifts: #(-16 0 16 0) 
			masks: #(16rFF0000 16rFF00 16rFF 16rFF000000).
	].! !

!PNGReadWriter methodsFor: 'accessing' stamp: 'DSM 3/24/2000 01:12'!
understandsImageFormat
	#(137 80 78 71 13 10 26 10) do: [ :byte |
		stream next = byte ifFalse: [^ false]].
	^ true
! !

!PNGReadWriter methodsFor: 'writing' stamp: 'nk 2/17/2004 16:51'!
updateCrc: oldCrc from: start to: stop in: aCollection
	^ZipWriteStream updateCrc: oldCrc from: start to: stop in: aCollection! !

!PNGReadWriter methodsFor: 'writing' stamp: 'BG 3/16/2005 08:26'!
writeChunk: crcStream
	| bytes length crc debug |
	debug := self debugging.
	bytes := crcStream originalContents.
	length := crcStream position.
	crc := self updateCrc: 16rFFFFFFFF from: 1 to: length in: bytes.
	crc := crc bitXor: 16rFFFFFFFF.
	debug ifTrue: [ Transcript cr;
		print: stream position; space;
		nextPutAll: (bytes copyFrom: 1 to: 4) asString;
		nextPutAll: ' len='; print: length;
		nextPutAll: ' crc=0x'; nextPutAll: crc printStringHex  ].
	stream nextNumber: 4 put: length-4. "exclude chunk name"
	stream next: length putAll: bytes startingAt: 1.
	stream nextNumber: 4 put: crc.
	debug ifTrue: [ Transcript nextPutAll: ' afterPos='; print: stream position ].
	crcStream resetToStart.! !

!PNGReadWriter methodsFor: 'writing' stamp: 'ar 12/12/2003 16:40'!
writeFileSignature
	stream nextPutAll: #(16r89 16r50 16r4E  16r47 16r0D 16r0A 16r1A 16r0A) asByteArray! !

!PNGReadWriter methodsFor: 'writing' stamp: 'nk 2/17/2004 14:57'!
writeIDATChunkOn: aStream
	"Write the IDAT chunk"
	| z |
	aStream nextPutAll: 'IDAT' asByteArray.
	z := ZLibWriteStream on: aStream.
	form depth <= 8 
		ifTrue:[self writeType3DataOn: z]
		ifFalse:[ self writeType6DataOn: z].
	self debugging ifTrue: [
		Transcript cr;
			nextPutAll: 'compressed size=';
			print: aStream position;
			nextPutAll: ' uncompressed size=';
			print: z position  ]
! !

!PNGReadWriter methodsFor: 'writing' stamp: 'ar 12/12/2003 17:08'!
writeIENDChunkOn: aStream
	"Write the IEND chunk"
	aStream nextPutAll: 'IEND' asByteArray.! !

!PNGReadWriter methodsFor: 'writing' stamp: 'ar 12/12/2003 17:21'!
writeIHDRChunkOn: aStream
	"Write the IHDR chunk"
	aStream nextPutAll: 'IHDR' asByteArray.
	aStream nextInt32Put: width.
	aStream nextInt32Put: height.
	aStream nextNumber: 1 put: bitsPerChannel.
	aStream nextNumber: 1 put: colorType.
	aStream nextNumber: 1 put: 0. "compression"
	aStream nextNumber: 1 put: 0. "filter method"
	aStream nextNumber: 1 put: 0. "interlace method"
! !

!PNGReadWriter methodsFor: 'writing' stamp: 'nk 4/17/2004 19:44'!
writePLTEChunkOn: aStream
	"Write the PLTE chunk"
	| r g b colors |
	aStream nextPutAll: 'PLTE' asByteArray.
	(form isColorForm) 
		ifTrue:[colors := form colors]
		ifFalse:[colors := Color indexedColors copyFrom: 1 to: (1 bitShift: form depth)].
	colors do:[:aColor|
		r := (aColor red * 255) truncated.
		g := (aColor green * 255) truncated.
		b := (aColor blue * 255) truncated.
		aStream nextPut: r; nextPut: g; nextPut: b.
	].! !

!PNGReadWriter methodsFor: 'writing' stamp: 'ar 12/12/2003 18:29'!
writeSBITChunkOn: aStream
	"Write the IDAT chunk"
	aStream nextPutAll: 'sBIT' asByteArray.
	form depth = 16 ifFalse:[self error: 'Unimplemented feature'].
	aStream nextPut: 5.
	aStream nextPut: 5.
	aStream nextPut: 5.
	aStream nextPut: 1.! !

!PNGReadWriter methodsFor: 'writing' stamp: 'ar 12/12/2003 17:34'!
writeTRNSChunkOn: aStream
	"Write out tRNS chunk"
	aStream nextPutAll: 'tRNS' asByteArray.
	form colors do:[:aColor|
		aStream nextPut: (aColor alpha * 255) truncated.
	].! !

!PNGReadWriter methodsFor: 'writing' stamp: 'ar 1/1/1970 20:58'!
writeType3DataOn: zStream
	"Write color indexed data."
	| scanline hack hackBlt swizzleBlt swizzleHack hackDepth |
	scanline := ByteArray new: bytesPerScanline + 3 // 4 * 4.
	hackDepth := bigEndian ifTrue:[form depth] ifFalse:[form depth negated].
	hack := Form extent: width@1 depth: hackDepth bits: scanline.
	hackBlt := BitBlt toForm: hack.
	hackBlt sourceForm: form.
	hackBlt combinationRule: Form over.
	hackBlt destOrigin: 0@0.
	hackBlt width: width; height: 1.
	(form depth < 8 and:[bigEndian not]) ifTrue:[
		swizzleHack := Form new hackBits: scanline.
		swizzleBlt := BitBlt toForm: swizzleHack.
		swizzleBlt sourceForm: swizzleHack.
		swizzleBlt combinationRule: Form over.
		swizzleBlt colorMap: (StandardSwizzleMaps at: form depth).
	].
	0 to: height-1 do:[:i|
		hackBlt sourceOrigin: 0@i; copyBits.
		swizzleBlt ifNotNil:[swizzleBlt copyBits].
		zStream nextPut: 0. "filterType"
		zStream next: bytesPerScanline putAll: scanline startingAt: 1.
	].
	zStream close.! !

!PNGReadWriter methodsFor: 'writing' stamp: 'ar 2/19/2004 00:10'!
writeType6DataOn: zStream
	"Write RGBA data."
	| scanline hack hackBlt cm miscBlt |
	scanline := ByteArray new: bytesPerScanline.
	hack := Form extent: width@1 depth: 32 bits: scanline.
	form depth = 16 ifTrue:[
		"Expand 16 -> 32"
		miscBlt := BitBlt toForm: hack.
		miscBlt sourceForm: form.
		miscBlt combinationRule: Form over.
		miscBlt destOrigin: 0@0.
		miscBlt width: width; height: 1.
	].
	hackBlt := BitBlt toForm: hack.
	hackBlt sourceForm: (miscBlt ifNil:[form] ifNotNil:[hack]).
	hackBlt combinationRule: Form over.
	hackBlt destOrigin: 0@0.
	hackBlt width: width; height: 1.
	bigEndian ifTrue:[
		cm := ColorMap 
			shifts: #(8 8 8 -24) 
			masks: #(16rFF0000 16rFF00 16rFF 16rFF000000).
	] ifFalse:[
		cm := ColorMap 
			shifts: #(-16 0 16 0) 
			masks: #(16rFF0000 16rFF00 16rFF 16rFF000000).
	].
	hackBlt colorMap: cm.
	0 to: height-1 do:[:i|
		miscBlt ifNil:[
			hackBlt sourceOrigin: 0@i; copyBits.
		] ifNotNil:[
			miscBlt sourceOrigin: 0@i; copyBits.
			hack fixAlpha.
			hackBlt copyBits.
		].
		zStream nextPut: 0. "filterType"
		zStream nextPutAll: scanline.
	].
	zStream close.! !

ImageReadWriter subclass: #PNMReadWriter
	instanceVariableNames: 'first type origin cols rows depth maxValue tupleType pragma'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Files'!

!PNMReadWriter commentStamp: 'jdr 10/20/2003 17:08' prior: 0!
I am a subclass of ImageReadWriter that decodes portable anymap file formats
(pbm, pgm, ppm and  pam) images.

I accept the #origin pragma for SE files as described in:
Algoritms For Image Processing And Computer Vision. J. R. Parker

Don't work with 2 bytes samples (16 bit grays, > 32 bits color, etc...), 
pam files preliminary support.

f _ ImageReadWriter formFromFileNamed: 'Tools:Squeak3.4:Carmen.ppm'.
f morphEdit

Submitted by Javier Diaz Reinoso, Oct/2003!

!PNMReadWriter class methodsFor: 'testing' stamp: 'jdr 10/11/2003 14:49'!
testFromSEFile: filename
	"read SE file, check origin
		PNMReadWriter testFromSEFile: 'Tools:Squeak3.4:eliseSE.pbm'.
	"
	| prw f |
	prw := self new.
	prw stream: (FileStream readOnlyFileNamed: filename).
	f := prw nextImage.
	f morphEdit.
	prw inspect! !

!PNMReadWriter class methodsFor: 'testing' stamp: 'jdr 10/16/2003 17:22'!
testFromString
	"read SE file from string
		PNMReadWriter testFromString
	"
	| prw f s |
	prw := self new.
	s := 
'P1
#origin 1 0
3 1
1	01'.
	prw stream: (ReadStream on: s from: 1 to: (s size)).
	f := prw nextImage.
	f morphEdit.
	Transcript cr;show:'Origin=', prw origin asString; cr.! !

!PNMReadWriter class methodsFor: 'testing' stamp: 'jdr 7/18/2005 17:03'!
testMultiFile: filename
	"write two files from user, then read
		PNMReadWriter testMultiFile: 'Tools:Squeak3.6:outMulti.pbm'.
	"
	| prw f |
	prw := self new.
	prw stream: ((FileStream newFileNamed: filename) binary).
	prw pragma: '#Squeak test', String lf.
	f := Form fromUser. prw nextPutImage: f. 
	f := Form fromUser.prw nextPutImage: f.	
	prw close.
	prw stream: (StandardFileStream readOnlyFileNamed: filename).
	f := prw nextImage. (SketchMorph withForm: f) openInWorld.
	f := prw nextImage. (SketchMorph withForm: f) openInWorld.
! !

!PNMReadWriter class methodsFor: 'testing' stamp: 'jdr 10/15/2003 15:43'!
testToSEFile: filename
	"write SE file with origin
		PNMReadWriter testToSEFile: 'Tools:Squeak3.4:outSE.pbm'.
	"
	| prw f |
	prw := self new.
	prw stream: ((FileStream newFileNamed: filename) binary).
	prw pragma: '#origin 10 10', String lf.
	f := Form fromUser.
	prw nextPutImage: f! !

!PNMReadWriter class methodsFor: 'image reading/writing' stamp: 'jdr 7/18/2005 16:25'!
typicalFileExtensions
	"Answer a collection of file extensions (lowercase) which files that I can read might commonly have"
	^#('pbm' 'pnm' 'ppm' 'pam')! !

!PNMReadWriter methodsFor: 'reading' stamp: 'jdr 10/16/2003 17:19'!
cleanLine
	"upTo LF or CR, tab as space"

	| line loop b |
	line := WriteStream with: ''.
	loop := true.
	[loop] whileTrue: [
		b := stream next.
		b ifNil:[
			loop := false		"EOS"
		]
		ifNotNil: [
			(b = (Character cr) or:[b = Character lf]) ifTrue:[
				loop := false.
			]
			ifFalse:[
				b = (Character tab) ifTrue:[b := Character space].
				line nextPut: b.
			]
		]
	].
	^line contents! !

!PNMReadWriter methodsFor: 'reading' stamp: 'jdr 10/15/2003 12:20'!
getTokenPbm: aCollection
	"get a number, return rest of collection"
	| line tokens token |
	tokens := aCollection.
	tokens size = 0 ifTrue:[
		[
			line := self pbmGetLine.
			line ifNil:[^{nil . nil}].
			tokens := line findTokens: ' '.
			tokens size = 0
		] whileTrue:[].
	].
	"Transcript cr; show: tokens asString."
	token := tokens removeFirst.
	^{token asInteger . tokens}
! !

!PNMReadWriter methodsFor: 'reading' stamp: 'jdr 10/16/2003 16:39'!
nextImage
	"read one image"
	| data p |
	first ifNil:[
		first _ false.
		data _ stream contentsOfEntireFile.
		stream _ (RWBinaryOrTextStream with: data) reset.
	]
	ifNotNil:[
		type < 4 ifTrue:[
			self error:'Plain PBM, PGM or PPM have only one image'
		].
	].
	stream ascii.
	p _ stream next.
	type _ (stream next) asInteger - 48.
	(p = $P and:[type > 0 and:[type < 8]]) ifFalse:[
		self error:'Not a PNM file'
	].
	type = 7 ifTrue:[
		self readHeaderPAM
	]
	ifFalse: [
		self readHeader
	].
	type caseOf: {
		[1] 	->	[^self readPlainBW].
		[2] 	->	[^self readPlainGray].
		[3] 	->	[^self readPlainRGB].
		[4] 	->	[^self readBWreverse: false].
		[5] 	->	[^self readGray].
		[6] 	->	[^self readRGB].
		[7] 	->	[	"PAM"
					(tupleType asUppercase) caseOf: {
						['BLACKANDWHITE'] 		-> [^self readBWreverse: true].
						['GRAYSCALE'] 			-> [^self readGray].
						['RGB'] 					-> [^self readRGB].
						['RGB_ALPHA'] 			-> [^self error:'Not implemented'].
						['GRAYSCALE_ALPHA'] 	-> [^self error:'Not implemented'].
					} otherwise: [^self readData].
				]
	}! !

!PNMReadWriter methodsFor: 'writing' stamp: 'jdr 10/16/2003 16:08'!
nextPutBW: aForm reverse: flagXor
	| myType val nBytes bytesRow |
	cols := aForm width.
	rows := aForm height.
	depth := aForm depth.
	"stream position: 0."
	aForm depth = 1 ifTrue:[myType := $4] ifFalse:[myType := $5].
	self writeHeader: myType.
	stream binary.
	nBytes := (cols/8) ceiling.
	bytesRow := (cols/32) ceiling * 4.
	0 to: rows-1 do: [:y | | i |
		i := 1 + (bytesRow*y).
		0 to: nBytes-1 do: [:x |
			val := aForm bits byteAt: i.
			flagXor ifTrue:[val := val bitXor: 16rFF].
			stream nextPut: val.
			i := i+1.
		]
	].
! !

!PNMReadWriter methodsFor: 'writing' stamp: 'jdr 10/16/2003 16:08'!
nextPutGray: aForm
	| myType peeker val |
	cols := aForm width.
	rows := aForm height.
	depth := aForm depth.
	"stream position: 0."
	aForm depth = 1 ifTrue:[myType := $4] ifFalse:[myType := $5].
	self writeHeader: myType.
	peeker := BitBlt current bitPeekerFromForm: aForm.
	0 to: rows-1 do: [:y |
		0 to: cols-1 do: [:x |
			val := peeker pixelAt: x@y.
			stream nextPut: val.
		]
	].
! !

!PNMReadWriter methodsFor: 'writing' stamp: 'jdr 10/16/2003 14:22'!
nextPutImage: aForm
	aForm unhibernate.
	aForm depth	 caseOf: {
		[1] 		-> [self nextPutBW: aForm reverse: false].
		[16] 	-> [self nextPutRGB: aForm].
		[32] 	-> [self nextPutRGB: aForm].
	} otherwise: [
		(aForm respondsTo: #colors) ifTrue:[
			aForm colors ifNil: [
				self nextPutGray: aForm
			]
			ifNotNil: [
				self nextPutRGB: aForm
			]
		]
		ifFalse:[
			self nextPutGray: aForm
		]
	]! !

!PNMReadWriter methodsFor: 'writing' stamp: 'jdr 10/16/2003 16:08'!
nextPutRGB: aForm
	| myType peeker f shift mask |
	cols := aForm width.
	rows := aForm height.
	depth := aForm depth.
	f := aForm.
	depth < 16 ifTrue:[
		f := aForm asFormOfDepth: 32.
		depth := 32.
	].
	myType := $6.
	"stream position: 0."
	self writeHeader: myType.
	depth = 32 ifTrue:[shift := 8. mask := 16rFF] ifFalse:[shift := 5. mask := 16r1F].
	peeker := BitBlt current bitPeekerFromForm: f.
	0 to: rows-1 do: [:y |
		0 to: cols-1 do: [:x | | p r g b |
			p := peeker pixelAt: x@y.
			b := p bitAnd: mask. p := p >> shift.
			g := p bitAnd: mask. p := p >> shift.
			r := p bitAnd: mask.
			stream nextPut: r.
			stream nextPut: g.
			stream nextPut: b.
		]
	].
! !

!PNMReadWriter methodsFor: 'accessing' stamp: 'jdr 10/16/2003 14:52'!
origin
	^origin! !

!PNMReadWriter methodsFor: 'reading' stamp: 'jdr 10/10/2003 15:09'!
pbmGetLine
	"Get the next non-comment line from the PBM stream
	Look for 'pragmas' - commands hidden in the comments"
	
 	| line |
	[
		line := self cleanLine.
		line ifNil: [^nil].
		(line size > 0 and:[(line at: 1) = $#]) ifTrue:[
			self pbmParam: line.
		].
		(line size = 0) or:[(line at: 1) = $#]
	]
 	whileTrue: [].
	^line! !

!PNMReadWriter methodsFor: 'reading' stamp: 'jdr 10/10/2003 15:11'!
pbmParam: line
	"Look for a parameter hidden in a comment"
	| key tokens |
	tokens := line findTokens: ' '.
	key := (tokens at: 1) asLowercase.
	(key = '#origin' and:[tokens size = 3]) ifTrue:[	"ORIGIN key word"
		"This is for SE files as described in:
		Algoritms For Image Processing And Computer Vision. J. R. Parker"
		origin := ((tokens at: 2) asInteger) @ ((tokens at: 3) asInteger)
	].
! !

!PNMReadWriter methodsFor: 'accessing' stamp: 'jdr 10/15/2003 15:35'!
pragma: s
	pragma := s! !

!PNMReadWriter methodsFor: 'reading' stamp: 'jdr 7/18/2005 17:02'!
r: r g: g b: b for: aDepth
	"integer value according depth"
	| val |
	aDepth = 16 ifTrue: [
		val := (1 << 15) + (r << 10) + (g << 5) + b.
	]
	ifFalse:[
		val := (16rFF << 24) + (r << 16) + (g << 8) + b.
	].
	^val
! !

!PNMReadWriter methodsFor: 'reading' stamp: 'jdr 10/16/2003 15:49'!
readBWreverse: flagXor
	"B&W for PAM"
	| val form bytesRow nBytes |
	stream binary.
	form := Form extent: cols@rows depth: 1.
	nBytes := (cols/8) ceiling.
	bytesRow := (cols/32) ceiling * 4.
	0 to: rows-1 do: [:y | | i |
		i := 1 + (bytesRow*y).
		0 to: nBytes-1 do: [:x |
			val := stream next.
			flagXor ifTrue:[val := val bitXor: 16rFF].
			form bits byteAt: i put: val.
			i := i+1.
		]
	].
	^form
! !

!PNMReadWriter methodsFor: 'reading' stamp: 'md 10/20/2004 15:45'!
readData
	"generic data"
	| data nBits nBytes val sample |
	stream binary.
	data := OrderedCollection new.
	nBits := maxValue floorLog:2.
	nBytes := (nBits+1) >> 3.
	(nBits+1 rem: 8) > 0 ifTrue:[nBytes := nBytes+1].

	0 to: rows-1 do: [:y |
		0 to: cols-1 do: [:x |
			val := 0.
			1 to: nBytes do: [:n |
				sample := stream next.
				val := val << 8 + sample.
			].
			data add: val.
		]
	].
	^data

! !

!PNMReadWriter methodsFor: 'reading' stamp: 'jdr 10/16/2003 15:44'!
readGray
	"gray form"
	| val form poker |
	maxValue > 255 ifTrue:[self error:'Gray value > 8 bits not supported in Squeak'].
	stream binary.
	form := Form extent: cols@rows depth: depth.
	poker := BitBlt current bitPokerToForm: form.
	0 to: rows-1 do: [:y |
		0 to: cols-1 do: [:x |
			val := stream next.
			poker pixelAt: x@y put: val.
		]
	].
	^form
! !

!PNMReadWriter methodsFor: 'reading' stamp: 'jdr 10/15/2003 15:44'!
readHeader
	"read header for pbm, pgm or ppm"
	| tokens aux d c  |
	tokens := OrderedCollection new.
	aux := self getTokenPbm: tokens.
	cols := aux at: 1. tokens := aux at: 2.
	aux := self getTokenPbm: tokens.
	rows := aux at: 1. tokens := aux at: 2.

	(type = 1 or:[type = 4]) ifTrue:[
		maxValue := 1
	]
	ifFalse: [
		aux := self getTokenPbm: tokens.
		maxValue := aux at: 1. tokens := aux at: 2.
	].
	d := {1 . 2 . 4 . 	8 . 		16 . 32}.
	c := {2 . 4 . 16 . 256 . 32768 . 16777216}. 
	(type = 3 or:[type = 6]) ifTrue: [
		maxValue >= 65536 ifTrue:[
			self error:'Pixmap > 48 bits not supported in PPM'
		].
		maxValue >= 256 ifTrue:[
			self error:'Pixmap > 32 bits are not supported in Squeak'
		].
		maxValue < 32 ifTrue:[depth := 16] ifFalse:[depth := 32].
	]
	ifFalse: [
		depth := nil.
		1 to: c size do:[:i| ((c at: i) > maxValue and:[depth = nil]) ifTrue:[depth:=d at: i]].
	].
	Transcript cr; show: 'PBM file class ', type asString, ' size ', cols asString, ' x ', 
		rows asString, ' maxValue =', maxValue asString, ' depth=', depth asString.
! !

!PNMReadWriter methodsFor: 'reading' stamp: 'jdr 10/15/2003 12:35'!
readHeaderPAM
	"read pam header, not tested"
	| loop line tokens key val |
	tupleType := ''.
	loop := true.
	loop whileTrue:[
		line := self pbmGetLine.
		tokens := line findTokens: ' '.
		tokens size = 2 ifTrue:[
			key := tokens at: 1 asUppercase.
			val := tokens at: 2.
			key caseOf: {
				['WIDTH'] 		-> [cols := val asInteger].
				['HEIGHT'] 		-> [rows := val asInteger].
				['DEPTH'] 		-> [depth := val asInteger].
				['MAXVAL']		-> [maxValue := val asInteger].
				['TUPLETYPE']	-> [tupleType := tupleType, ' ', val].
				['ENDHDR']		-> [loop := false].
			}
		]
	].
	Transcript cr; show: 'PAM file class ', type asString, ' size ', cols asString, ' x ', 
		rows asString, ' maxValue =', maxValue asString, ' depth=', depth asString.
! !

!PNMReadWriter methodsFor: 'reading' stamp: 'jdr 10/16/2003 16:03'!
readPlainBW
	"plain BW"
	| val form poker |
	form := Form extent: cols@rows depth: depth.
	poker := BitBlt current bitPokerToForm: form.
	0 to: rows-1 do: [:y |
		0 to: cols-1 do: [:x |
			[val := stream next. (val = $0 or:[val = $1])] whileFalse:[
				val ifNil:[self error:'End of file reading PBM'].
			].
			poker pixelAt: x@y put: (val asInteger).
		]
	].
	^form
! !

!PNMReadWriter methodsFor: 'reading' stamp: 'jdr 10/16/2003 15:44'!
readPlainGray
	"plain gray"
	| val form poker aux tokens |
	form := Form extent: cols@rows depth: depth.
	poker := BitBlt current bitPokerToForm: form.
	tokens := OrderedCollection new.
	0 to: rows-1 do: [:y |
		0 to: cols-1 do: [:x |
			aux := self getTokenPbm: tokens.
			val := aux at: 1. tokens := aux at: 2.
			poker pixelAt: x@y put: val.
		]
	].
	^form
! !

!PNMReadWriter methodsFor: 'reading' stamp: 'jdr 10/15/2003 12:49'!
readPlainRGB
	"RGB form, use 32 bits"
	| val form poker tokens aux |
	maxValue > 255 ifTrue:[self error:'RGB value > 32 bits not supported in Squeak'].
	form := Form extent: cols@rows depth: 32.
	poker := BitBlt current bitPokerToForm: form.
	tokens := OrderedCollection new.
	0 to: rows-1 do: [:y |
		0 to: cols-1 do: [:x | | r g b|
			aux := self getTokenPbm: tokens. r := aux at: 1. tokens := aux at: 2.
			aux := self getTokenPbm: tokens. g := aux at: 1. tokens := aux at: 2.
			aux := self getTokenPbm: tokens. b := aux at: 1. tokens := aux at: 2.
			val := self r: r g: g b: b for: depth.
			poker pixelAt: x@y put: val.
		]
	].
	^form
! !

!PNMReadWriter methodsFor: 'reading' stamp: 'jdr 7/18/2005 16:58'!
readRGB
	"RGB form, use 16/32 bits"
	| val form poker sample shift |
	maxValue > 255 ifTrue:[self error:'RGB value > 32 bits not supported in Squeak'].
	stream binary.
	form := Form extent: cols@rows depth: depth.
	poker := BitBlt current bitPokerToForm: form.
	depth = 32 ifTrue:[shift := 8] ifFalse:[shift := 5].
	0 to: rows-1 do: [:y |
		0 to: cols-1 do: [:x |
			val := 16rFF.	"no transparency"
			1 to: 3 do: [:i |
				sample := stream next.
				val := val << shift + sample.
			].
			poker pixelAt: x@y put: val.
		]
	].
	^form
! !

!PNMReadWriter methodsFor: 'accessing' stamp: 'jdr 10/10/2003 18:04'!
stream: s
	stream := s! !

!PNMReadWriter methodsFor: 'accessing' stamp: 'jdr 10/16/2003 14:53'!
tupleType
	^tupleType! !

!PNMReadWriter methodsFor: 'testing' stamp: 'jdr 10/11/2003 14:52'!
understandsImageFormat
	"P1 to P7"
	| p  |
	p := stream next asCharacter.
	type := stream next - 48.
	^(p = $P and:[type > 0 and:[type < 8]])
	! !

!PNMReadWriter methodsFor: 'writing' stamp: 'jdr 10/15/2003 15:48'!
writeHeader: myType
	"this is ascii"
	stream nextPut: ($P asciiValue).
	stream nextPut: (myType asciiValue).
	stream nextPut: 10.		"nl"
	pragma ifNotNil:[
		stream nextPutAll: (pragma asByteArray).
	].
	stream nextPutAll: (cols printString) asByteArray.
	stream nextPut: 32.		" "
	stream nextPutAll: (rows printString) asByteArray.
	stream nextPut: 10.		"nl"
	depth > 1 ifTrue: [| d c maxV |
		d := {1 . 2 . 4  . 8   . 16 . 32}.
		c := {1 . 3 . 15 . 255 . 31 . 255}. 
		maxV := nil.
		1 to: d size do:[:i| ((d at: i) = depth and:[maxV = nil]) ifTrue:[maxV := c at: i]].
		stream nextPutAll: (maxV printString) asByteArray.
		stream nextPut: 10.		"nl"
	]
	! !

ImageReadWriter subclass: #XBMReadWriter
	instanceVariableNames: 'width height'
	classVariableNames: 'Flipbits'
	poolDictionaries: ''
	category: 'Graphics-Files'!

!XBMReadWriter class methodsFor: 'class initialization' stamp: 'wdc 2/2/1999 15:28'!
initialize
	"XBMReadWriter initialize"
	| flippedByte |
	Flipbits := (0 to: 255) collect:
     [:n |  "Compute the bit-reversal of the 8-bit value, n"
     flippedByte := 0.
     0 to: 7 do: 
         [:i | 
         flippedByte := flippedByte bitOr: ((n >> i bitAnd: 1) << (7-i))].
         flippedByte]! !

!XBMReadWriter class methodsFor: 'class initialization' stamp: 'nk 7/16/2003 17:57'!
typicalFileExtensions
	"Answer a collection of file extensions (lowercase) which files that I can read might commonly have"
	^#('xbm')! !

!XBMReadWriter methodsFor: 'accessing' stamp: 'wdc 2/1/1999 07:15'!
nextImage
	"Read in the next xbm image from the stream."
	| form long incount chunks byteWidth pad fourway outcount total |
	stream reset.
	stream ascii.
	self readHeader.
	form := ColorForm extent: width@height depth: 1.
	incount := 0.	outcount :=1.
	chunks := Array new: 4.	byteWidth := width + 7 // 8.
	total := byteWidth * height.
	byteWidth > 4
		ifTrue: [ pad := byteWidth \\ 4]
		ifFalse: [ pad := 4 - byteWidth ].
	fourway := 0.
	[(incount = total)] whileFalse: [
		incount := incount + 1.
		fourway := fourway + 1.
		chunks at: fourway put: (Flipbits at: ((self parseByteValue) +1)).
		(pad > 0 and: [(incount \\ byteWidth) = 0]) ifTrue: [
			1 to: pad do:
				[:q |	
  			fourway := fourway + 1.	
			chunks at: fourway put: 0]
		].
		fourway = 4 ifTrue: [
			long := Integer
				byte1: (chunks at: 4)
				byte2: (chunks at: 3)
				byte3: (chunks at: 2)
				byte4: (chunks at: 1).
			(form bits) at: outcount put: long.
			fourway := 0.
			outcount := outcount + 1].
		].
	 ^ form ! !

!XBMReadWriter methodsFor: 'private' stamp: 'wdc 2/2/1999 15:20'!
parseByteValue
	"skip over separators and return next bytevalue parsed as a C language number:
		0ddd is an octal digit.
		0xddd is a hex digit.
		ddd is decimal."
	| source mybase |
	stream skipSeparators.
	source := ReadWriteStream on: String new.
	[stream atEnd or: [ stream peek isSeparator ]]
		whileFalse: [source nextPut: self next asUppercase].
	mybase := 10. "Base 10 default"
	source reset.
	(source peek = $0) ifTrue: [
		mybase := 8. "Octal or Hex, say its Octal unless overridden."
		source next.
		(source peek = $X) ifTrue: [
			mybase := 16. "Ah.  It's Hex."
			source next.
			]
		].
	^ Integer readFrom: source base: mybase! !

!XBMReadWriter methodsFor: 'private' stamp: 'nk 7/16/2003 18:17'!
readHeader
	"Set width and height, and position stream at start of bytes"
	| number setwidth setheight fieldName |
	setwidth _ setheight _ false.
		[((stream atEnd) or: [setwidth and: [setheight]])]
		whileFalse: [
	  	self skipCComments.
		(stream nextMatchAll: '#define ') ifFalse: [^ false].
		(stream skipTo: $_) ifFalse: [^ false].
		fieldName _ String streamContents:
			[:source |
			[(stream atEnd) or: [ stream peek isSeparator ]]
				whileFalse: [ source nextPut: stream next]].
	  	(fieldName = 'width') ifTrue: [
			stream skipSeparators.
			number _ Integer readFrom: stream.
			(number > 0) ifTrue: [setwidth _true].
	  		width _ number.].
		(fieldName = 'height') ifTrue: [
			stream skipSeparators.
			number _ Integer readFrom: stream.
			(number > 0) ifTrue: [setheight _ true].
			height _ number.
			].
		].
	(setwidth & setheight) ifFalse: [^ false].
	^ stream skipTo: ${
! !

!XBMReadWriter methodsFor: 'private' stamp: 'nk 7/16/2003 18:15'!
skipCComments
	[ stream skipSeparators.
	stream peekFor: $/ ] whileTrue: [
		stream next.		"skip next *"
		[ (stream skipTo: $*) ifFalse: [ ^false ].
			stream peekFor: $/ ] whileFalse ]! !

!XBMReadWriter methodsFor: 'accessing' stamp: 'nk 7/16/2003 18:18'!
understandsImageFormat
	"Test to see if the image stream format is understood by this decoder.	This should be implemented in each subclass of ImageReadWriter so that	a proper decoder can be selected without ImageReadWriter having to know about all possible image file types."
	| first |
	stream ascii.
	self skipCComments.
	first := (stream next: 7) asString.
	stream binary.
	^ (first = '#define')! !

Object subclass: #JPEGColorComponent
	instanceVariableNames: 'currentX currentY hSampleFactor vSampleFactor mcuBlocks widthInBlocks heightInBlocks dctSize mcuWidth mcuHeight priorDCValue id qTableIndex dcTableIndex acTableIndex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Files'!

!JPEGColorComponent commentStamp: '<historical>' prior: 0!
I represent a single component of color in JPEG YCbCr color space.  I can accept a list of blocks in my component from the current MCU, then stream the samples from this block for use in color conversion.  I also store the running DC sample value for my component, used by the Huffman decoder.

The following layout is fixed for the JPEG primitives to work:
	currentX 		<SmallInteger>
	currentY 		<SmallInteger>
	hSampleFactor 	<SmallInteger>
	vSampleFactor 	<SmallInteger>
	mcuBlocks 		<Array of: <IntegerArray of: DCTSize2 * Integer>>
	widthInBlocks 	<SmallInteger>
	heightInBlocks 	<SmallInteger>
	dctSize 			<SmallInteger>
	mcuWidth 		<SmallInteger>
	mcuHeight 		<SmallInteger>
	priorDCValue 	<SmallInteger>
!

!JPEGColorComponent methodsFor: 'accessing' stamp: 'tao 10/23/97 12:21'!
acTableIndex

	^acTableIndex! !

!JPEGColorComponent methodsFor: 'accessing' stamp: 'tao 10/23/97 12:20'!
acTableIndex: anInteger

	acTableIndex := anInteger! !

!JPEGColorComponent methodsFor: 'accessing' stamp: 'tao 10/23/97 12:21'!
dcTableIndex

	^dcTableIndex! !

!JPEGColorComponent methodsFor: 'accessing' stamp: 'tao 10/23/97 12:20'!
dcTableIndex: anInteger

	dcTableIndex := anInteger! !

!JPEGColorComponent methodsFor: 'accessing' stamp: 'tao 10/23/97 12:11'!
heightInBlocks

	^heightInBlocks! !

!JPEGColorComponent methodsFor: 'accessing' stamp: 'tao 10/23/97 12:10'!
heightInBlocks: anInteger

	heightInBlocks := anInteger! !

!JPEGColorComponent methodsFor: 'accessing' stamp: 'tao 10/23/97 12:09'!
id

	^id! !

!JPEGColorComponent methodsFor: 'accessing' stamp: 'tao 10/23/97 12:09'!
id: anObject

	id := anObject! !

!JPEGColorComponent methodsFor: 'sample streaming' stamp: 'tao 10/23/97 12:24'!
initializeSampleStreamBlocks: aCollection

	mcuBlocks := aCollection.
	self resetSampleStream! !

!JPEGColorComponent methodsFor: 'accessing' stamp: 'ar 3/4/2001 01:19'!
mcuWidth: mw mcuHeight: mh dctSize: ds

	mcuWidth := mw.
	mcuHeight := mh.
	dctSize := ds.
	hSampleFactor := mcuWidth // widthInBlocks.
	vSampleFactor := mcuHeight // heightInBlocks! !

!JPEGColorComponent methodsFor: 'sample streaming' stamp: 'ar 3/4/2001 22:16'!
nextSample

	| dx dy blockIndex sampleIndex sample |
	dx := currentX // hSampleFactor.
	dy := currentY // vSampleFactor.
	blockIndex := dy // dctSize * widthInBlocks + (dx // dctSize) + 1.
	sampleIndex := dy \\ dctSize * dctSize + (dx \\ dctSize) + 1.
	sample := (mcuBlocks at: blockIndex) at: sampleIndex.
	currentX := currentX + 1.
	currentX < (mcuWidth * dctSize)
		ifFalse:
			[currentX := 0.
			currentY := currentY + 1].
	^ sample! !

!JPEGColorComponent methodsFor: 'accessing' stamp: 'tao 10/23/97 12:15'!
priorDCValue: aNumber

	priorDCValue := aNumber! !

!JPEGColorComponent methodsFor: 'accessing' stamp: 'tao 10/23/97 12:43'!
qTableIndex
	^qTableIndex! !

!JPEGColorComponent methodsFor: 'accessing' stamp: 'tao 10/23/97 12:11'!
qTableIndex: anInteger

	qTableIndex := anInteger! !

!JPEGColorComponent methodsFor: 'sample streaming' stamp: 'tao 10/23/97 12:24'!
resetSampleStream

	currentX := 0.
	currentY := 0! !

!JPEGColorComponent methodsFor: 'accessing' stamp: 'tao 10/23/97 12:18'!
totalMcuBlocks

	^ heightInBlocks * widthInBlocks! !

!JPEGColorComponent methodsFor: 'accessing' stamp: 'tao 10/23/97 12:16'!
updateDCValue: aNumber

	priorDCValue := priorDCValue + aNumber.
	^priorDCValue! !

!JPEGColorComponent methodsFor: 'accessing' stamp: 'tao 10/23/97 12:11'!
widthInBlocks

	^widthInBlocks! !

!JPEGColorComponent methodsFor: 'accessing' stamp: 'tao 10/23/97 12:10'!
widthInBlocks: anInteger

	widthInBlocks := anInteger! !

Object subclass: #JPEGHuffmanTable
	instanceVariableNames: 'bits values mincode maxcode valptr lookaheadBits lookaheadSymbol'
	classVariableNames: 'BitBufferSize Lookahead'
	poolDictionaries: ''
	category: 'Graphics-Files'!

!JPEGHuffmanTable commentStamp: '<historical>' prior: 0!
I represent the table of values used to decode Huffman entropy-encoded bitstreams.  From the JFIF file header entropy values, I build a derived table of codes and values for faster decoding.!

!JPEGHuffmanTable class methodsFor: 'initialization' stamp: 'tao 10/21/97 11:06'!
initialize

	Lookahead := 8.
	BitBufferSize := 16! !

!JPEGHuffmanTable class methodsFor: 'constants' stamp: 'tao 10/21/97 22:15'!
lookahead

	^ Lookahead! !

!JPEGHuffmanTable methodsFor: 'accessing' stamp: 'tao 10/20/97 14:27'!
bits: anObject

	bits := anObject! !

!JPEGHuffmanTable methodsFor: 'accessing' stamp: 'tao 10/21/97 23:31'!
lookaheadBits
	^lookaheadBits! !

!JPEGHuffmanTable methodsFor: 'accessing' stamp: 'tao 10/21/97 23:38'!
lookaheadSymbol
	^lookaheadSymbol! !

!JPEGHuffmanTable methodsFor: 'computation' stamp: 'tao 10/24/97 12:28'!
makeDerivedTables

	| huffSize huffCode code si index lookbits |
	mincode := Array new: 16.
	maxcode := Array new: 17.
	valptr := Array new: 17.
	huffSize := OrderedCollection new.
	1 to: 16 do: [:l | 1 to: (bits at: l) do: [:i | huffSize add: l]].
	huffSize add: 0.
	code := 0.
	huffCode := Array new: huffSize size.
	si := huffSize at: 1.
	index := 1.
	[(huffSize at: index) ~= 0] whileTrue:
		[[(huffSize at: index) = si] whileTrue:
			[huffCode at: index put: code.
			index := index + 1.
			code := code + 1].
		code := code << 1.
		si := si + 1].

	index := 1.
	1 to: 16 do:
		[:l |
		(bits at: l) ~= 0
			ifTrue:
				[valptr at: l put: index.
				mincode at: l put: (huffCode at: index).
				index := index + (bits at: l).
				maxcode at: l put: (huffCode at: index-1)]
			ifFalse:
				[maxcode at: l put: -1]].
	maxcode at: 17 put: 16rFFFFF.

	lookaheadBits := (Array new: 1 << Lookahead) atAllPut: 0.
	lookaheadSymbol := Array new: 1 << Lookahead.
	index := 1.
	1 to: Lookahead do:
		[:l |
		1 to: (bits at: l) do:
			[:i |
			lookbits := (huffCode at: index) << (Lookahead - l) + 1.
			(1 << (Lookahead - l) to: 1 by: -1) do:
				[:ctr |
				lookaheadBits at: lookbits put: l.
				lookaheadSymbol at: lookbits put: (values at: index).
				lookbits := lookbits + 1].
			index := index + 1]]! !

!JPEGHuffmanTable methodsFor: 'accessing' stamp: 'tao 10/21/97 23:59'!
maxcode
	^maxcode! !

!JPEGHuffmanTable methodsFor: 'computation' stamp: 'tao 10/21/97 22:44'!
valueForCode: code length: length

	^ values at: ((valptr at: length) + code - (mincode at: length))! !

!JPEGHuffmanTable methodsFor: 'accessing' stamp: 'tao 10/20/97 14:27'!
values: anObject

	values := anObject! !

Object subclass: #Point
	instanceVariableNames: 'x y'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Primitives'!

!Point commentStamp: '<historical>' prior: 0!
I represent an x-y pair of numbers usually designating a location on the screen.!

!Point class methodsFor: 'instance creation' stamp: 'sw 9/20/97 15:34'!
fromUser
	Sensor waitNoButton.
	Cursor crossHair show.
	Sensor waitButton.
	Cursor normal show.
	^ Sensor cursorPoint

"Point fromUser"! !

!Point class methodsFor: 'instance creation' stamp: 'wiz 5/4/2006 00:04'!
fromUserWithCursor: aCursor
	Sensor waitNoButton.
	aCursor showWhile:[Sensor waitButton].
	^ Sensor cursorPoint

"Point fromUserWithCursor: Cursor target"! !

!Point class methodsFor: 'instance creation' stamp: 'md 12/2/2004 23:44'!
r: rho degrees: degrees
	"Answer an instance of me with polar coordinates rho and theta."

	^self basicNew setR: rho degrees: degrees! !

!Point class methodsFor: 'instance creation' stamp: 'md 12/2/2004 23:44'!
x: xInteger y: yInteger 
	"Answer an instance of me with coordinates xInteger and yInteger."

	^self basicNew setX: xInteger setY: yInteger! !

!Point methodsFor: 'arithmetic' stamp: 'di 11/6/1998 14:01'!
* arg 
	"Answer a Point that is the product of the receiver and arg."

	arg isPoint ifTrue: [^ (x * arg x) @ (y * arg y)].
	^ arg adaptToPoint: self andSend: #*! !

!Point methodsFor: 'arithmetic' stamp: 'di 11/6/1998 14:01'!
+ arg 
	"Answer a Point that is the sum of the receiver and arg."

	arg isPoint ifTrue: [^ (x + arg x) @ (y + arg y)].
	^ arg adaptToPoint: self andSend: #+! !

!Point methodsFor: 'arithmetic' stamp: 'di 11/6/1998 14:02'!
- arg 
	"Answer a Point that is the difference of the receiver and arg."

	arg isPoint ifTrue: [^ (x - arg x) @ (y - arg y)].
	^ arg adaptToPoint: self andSend: #-! !

!Point methodsFor: 'arithmetic' stamp: 'di 11/6/1998 14:02'!
/ arg 
	"Answer a Point that is the quotient of the receiver and arg."

	arg isPoint ifTrue: [^ (x / arg x) @ (y / arg y)].
	^ arg adaptToPoint: self andSend: #/! !

!Point methodsFor: 'arithmetic' stamp: 'di 11/6/1998 14:02'!
// arg 
	"Answer a Point that is the quotient of the receiver and arg."

	arg isPoint ifTrue: [^ (x // arg x) @ (y // arg y)].
	^ arg adaptToPoint: self andSend: #//! !

!Point methodsFor: 'comparing' stamp: ''!
< aPoint 
	"Answer whether the receiver is above and to the left of aPoint."

	^x < aPoint x and: [y < aPoint y]! !

!Point methodsFor: 'comparing' stamp: ''!
<= aPoint 
	"Answer whether the receiver is neither below nor to the right of aPoint."

	^x <= aPoint x and: [y <= aPoint y]! !

!Point methodsFor: 'comparing' stamp: ''!
= aPoint

	self species = aPoint species
		ifTrue: [^x = aPoint 
	"Refer to the comment in Object|=." x and: [y = aPoint y]]
		ifFalse: [^false]! !

!Point methodsFor: 'comparing' stamp: ''!
> aPoint 
	"Answer whether the receiver is below and to the right of aPoint."

	^x > aPoint x and: [y > aPoint y]! !

!Point methodsFor: 'comparing' stamp: ''!
>= aPoint 
	"Answer whether the receiver is neither above nor to the left of aPoint."

	^x >= aPoint x and: [y >= aPoint y]! !

!Point methodsFor: 'arithmetic' stamp: 'di 11/6/1998 14:02'!
\\ arg 
	"Answer a Point that is the mod of the receiver and arg."

	arg isPoint ifTrue: [^ (x \\ arg x) @ (y \\ arg y)].
	^ arg adaptToPoint: self andSend: #\\! !

!Point methodsFor: 'arithmetic' stamp: ''!
abs
	"Answer a Point whose x and y are the absolute values of the receiver's x 
	and y."

	^ x abs @ y abs! !

!Point methodsFor: 'converting' stamp: 'di 11/6/1998 13:45'!
adaptToCollection: rcvr andSend: selector
	"If I am involved in arithmetic with a Collection, return a Collection of
	the results of each element combined with me in that expression."

	^ rcvr collect: [:element | element perform: selector with: self]! !

!Point methodsFor: 'converting' stamp: 'di 11/9/1998 12:44'!
adaptToNumber: rcvr andSend: selector
	"If I am involved in arithmetic with an Integer, convert it to a Point."
	^ rcvr@rcvr perform: selector with: self! !

!Point methodsFor: 'converting' stamp: 'di 11/6/1998 13:47'!
adaptToString: rcvr andSend: selector
	"If I am involved in arithmetic with a String, convert it to a Number."
	^ rcvr asNumber perform: selector with: self! !

!Point methodsFor: 'transforming' stamp: 'di 4/30/1998 11:16'!
adhereTo: aRectangle
	"If the receiver lies outside aRectangle, return the nearest point on the boundary of the rectangle, otherwise return self."

	(aRectangle containsPoint: self) ifTrue: [^ self].
	^ ((x max: aRectangle left) min: aRectangle right)
		@ ((y max: aRectangle top) min: aRectangle bottom)! !

!Point methodsFor: 'converting' stamp: ''!
asFloatPoint
	^ x asFloat @ y asFloat! !

!Point methodsFor: 'converting' stamp: ''!
asIntegerPoint
	^ x asInteger @ y asInteger! !

!Point methodsFor: 'converting' stamp: 'wiz 11/25/2004 12:48'!
asNonFractionalPoint
(x isFraction or: [y isFraction])
	ifTrue:[^ x asFloat @ y asFloat]! !

!Point methodsFor: 'converting' stamp: ''!
asPoint
	"Answer the receiver itself."

	^self! !

!Point methodsFor: 'printing' stamp: 'sw 9/27/2001 17:26'!
basicType
	"Answer a symbol representing the inherent type of the receiver"

	^ #Point! !

!Point methodsFor: 'point functions' stamp: 'FBS 1/5/2004 13:08'!
bearingToPoint: anotherPoint
    "Return the bearing, in degrees, from the receiver to anotherPoint.
     Adapted from Playground, where the ultimate provenance of the algorithm was a wild earlier method of Jay Fenton's which I never checked carefully, but the thing has always seemed to work"

    | deltaX deltaY  |
    deltaX := anotherPoint x -  x.
    deltaY := anotherPoint y - y.

    deltaX abs < 0.001
        ifTrue:
            [^ deltaY > 0 ifTrue: [180] ifFalse: [0]].

    ^ ((deltaX >= 0 ifTrue: [90] ifFalse: [270])
            - ((deltaY / deltaX) arcTan negated radiansToDegrees)) rounded
! !

!Point methodsFor: 'private' stamp: 'ar 4/4/1999 00:40'!
bitShiftPoint: bits
	x := x bitShift: bits.
	y := y bitShift: bits.! !

!Point methodsFor: 'converting' stamp: ''!
corner: aPoint 
	"Answer a Rectangle whose origin is the receiver and whose corner is 
	aPoint. This is one of the infix ways of expressing the creation of a 
	rectangle."

	^Rectangle origin: self corner: aPoint! !

!Point methodsFor: 'point functions' stamp: 'ar 10/30/1998 03:05'!
crossProduct: aPoint 
	"Answer a number that is the cross product of the receiver and the 
	argument, aPoint."

	^ (x * aPoint y) - (y * aPoint x)! !

!Point methodsFor: 'copying' stamp: ''!
deepCopy
	"Implemented here for better performance."

	^x deepCopy @ y deepCopy! !

!Point methodsFor: 'polar coordinates' stamp: 'di 6/12/97 12:18'!
degrees
	"Answer the angle the receiver makes with origin in degrees. right is 0; down is 90."
	| tan theta |
	x = 0
		ifTrue: [y >= 0
				ifTrue: [^ 90.0]
				ifFalse: [^ 270.0]]
		ifFalse: 
			[tan := y asFloat / x asFloat.
			theta := tan arcTan.
			x >= 0
				ifTrue: [y >= 0
						ifTrue: [^ theta radiansToDegrees]
						ifFalse: [^ 360.0 + theta radiansToDegrees]]
				ifFalse: [^ 180.0 + theta radiansToDegrees]]! !

!Point methodsFor: 'point functions' stamp: 'spfa 11/6/2004 16:01'!
dist: aPoint 
	"Answer the distance between aPoint and the receiver."

	| dx dy |

	dx := aPoint x - x.
	dy := aPoint y - y.

	^ ((dx * dx) + (dy * dy)) sqrt! !

!Point methodsFor: 'point functions' stamp: 'di 9/11/1998 16:22'!
dotProduct: aPoint 
	"Answer a number that is the dot product of the receiver and the 
	argument, aPoint. That is, the two points are multipled and the 
	coordinates of the result summed."

	^ (x * aPoint x) + (y * aPoint y)! !

!Point methodsFor: 'point functions' stamp: ''!
eightNeighbors
	^ (Array with: self + (1@0)
		with: self + (1@1)
		with: self + (0@1)
		with: self + (-1@1)) ,
	(Array with: self + (-1@0)
		with: self + (-1@-1)
		with: self + (0@-1)
		with: self + (1@-1))
! !

!Point methodsFor: 'converting' stamp: ''!
extent: aPoint 
	"Answer a Rectangle whose origin is the receiver and whose extent is 
	aPoint. This is one of the infix ways of expressing the creation of a 
	rectangle."

	^Rectangle origin: self extent: aPoint! !

!Point methodsFor: 'point functions' stamp: 'di 6/11/97 16:08'!
flipBy: direction centerAt: c
	"Answer a Point which is flipped according to the direction about the point c.
	Direction must be #vertical or #horizontal."
	direction == #vertical ifTrue: [^ x @ (c y * 2 - y)].
	direction == #horizontal ifTrue: [^ (c x * 2 - x) @ y].
	self error: 'unrecognizable direction'! !

!Point methodsFor: 'point functions' stamp: ''!
fourNeighbors
	^ Array with: self + (1@0)
		with: self + (0@1)
		with: self + (-1@0)
		with: self + (0@-1)
! !

!Point methodsFor: 'point functions' stamp: ''!
grid: aPoint 
	"Answer a Point to the nearest rounded grid modules specified by aPoint."

	| newX newY |
	newX := x + (aPoint x // 2) truncateTo: aPoint x.
	newY := y + (aPoint y // 2) truncateTo: aPoint y.
	^newX @ newY! !

!Point methodsFor: 'comparing' stamp: 'SqR 11/3/2000 17:08'!
hash
	"Hash is reimplemented because = is implemented."

	^(x hash hashMultiply + y hash) hashMultiply! !

!Point methodsFor: 'comparing' stamp: ''!
hashMappedBy: map
	"My hash is independent of my oop."

	^self hash! !

!Point methodsFor: 'point functions' stamp: 'ar 5/22/2001 23:46'!
insideTriangle: p1 with: p2 with: p3
	"Return true if the receiver is within the triangle defined by the three coordinates.
	Note: This method computes the barycentric coordinates for the receiver and tests those coordinates."
	| p0 b0 b1 b2 b3 |
	p0 := self.
	b0 := ((p2 x - p1 x) * (p3 y - p1 y)) - ((p3 x - p1 x) * (p2 y - p1 y)).
	b0 isZero ifTrue:[^false]. "degenerate"
	b0 := 1.0 / b0.
	b1 := (((p2 x - p0 x) * (p3 y - p0 y)) - ((p3 x - p0 x) * (p2 y - p0 y))) * b0.
	b2 := (((p3 x - p0 x) * (p1 y - p0 y)) - ((p1 x - p0 x) * (p3 y - p0 y))) * b0.
	b3 := (((p1 x - p0 x) * (p2 y - p0 y)) - ((p2 x - p0 x) * (p1 y - p0 y))) * b0.
	b1 < 0.0 ifTrue:[^false].
	b2 < 0.0 ifTrue:[^false].
	b3 < 0.0 ifTrue:[^false].
	^true

! !

!Point methodsFor: 'interpolating' stamp: 'jsp 3/22/1999 16:31'!
interpolateTo: end at: amountDone
	"Interpolate between the instance and end after the specified amount has been done (0 - 1)."

	^ self + ((end - self) * amountDone).! !

!Point methodsFor: 'geometry' stamp: 'laza 1/24/2000 03:44'!
isInsideCircle: a with: b with: c 
	"Returns TRUE if self is inside the circle defined by the     
	points a, b, c. See Guibas and Stolfi (1985) p.107"
	^ (a dotProduct: a)
		* (b triangleArea: c with: self) - ((b dotProduct: b)
			* (a triangleArea: c with: self)) + ((c dotProduct: c)
			* (a triangleArea: b with: self)) - ((self dotProduct: self)
			* (a triangleArea: b with: c)) > 0.0! !

!Point methodsFor: 'converting' stamp: 'di 11/6/1998 07:45'!
isPoint
	^ true! !

!Point methodsFor: 'self evaluating' stamp: 'sd 7/31/2005 21:48'!
isSelfEvaluating
	^ self class == Point! !

!Point methodsFor: 'testing' stamp: 'ar 10/29/2000 19:02'!
isZero
	^x isZero and:[y isZero]! !

!Point methodsFor: 'comparing' stamp: ''!
max: aPoint 
	"Answer the lower right corner of the rectangle uniquely defined by the 
	receiver and the argument, aPoint."

	^ (x max: aPoint x) @ (y max: aPoint y)! !

!Point methodsFor: 'comparing' stamp: ''!
min: aPoint 
	"Answer the upper left corner of the rectangle uniquely defined by the 
	receiver and the argument, aPoint."

	^ (x min: aPoint x) @ (y min: aPoint y)! !

!Point methodsFor: 'comparing' stamp: ''!
min: aMin max: aMax 

	^ (self min: aMin) max: aMax! !

!Point methodsFor: 'point functions' stamp: 'di 12/1/97 12:37'!
nearestPointAlongLineFrom: p1 to: p2
	"Note this will give points beyond the endpoints.
	Streamlined by Gerardo Richarte 11/3/97"
	| x21 y21 t x1 y1 |
	p1 x = p2 x ifTrue: [^ p1 x @ y].
	p1 y = p2 y ifTrue: [^ x @ p1 y].
	x1 := p1 x asFloat.
	y1 := p1 y asFloat.
	x21 := p2 x asFloat - x1.
	y21 := p2 y asFloat - y1.
	t := ((y asFloat - y1 / x21) + (x asFloat - x1 / y21))
			/ ((x21 / y21) + (y21 / x21)).
	^ (x1 + (t * x21)) @ (y1 + (t * y21))
"
	| old new |
	Pen new place: 200@100; goto: (old := 500@300).
	Display reverse: (old extent: 10@10).
	[Sensor anyButtonPressed] whileFalse:
		[(new := (Sensor cursorPoint nearestPointAlongLineFrom: 200@100 to: 500@300) )
			= old ifFalse:
				[Display reverse: (old extent: 10@10).
				Display reverse: ((old := new) extent: 10@10)]]
"
! !

!Point methodsFor: 'point functions' stamp: 'di 12/1/97 12:40'!
nearestPointOnLineFrom: p1 to: p2
	"This will not give points beyond the endpoints"
	^ (self nearestPointAlongLineFrom: p1 to: p2)
		adhereTo: (p1 rect: p2)! !

!Point methodsFor: 'transforming' stamp: 'ar 8/26/2001 22:14'!
negated
	"Answer a point whose x and y coordinates are the negatives of those of the receiver.  6/6/96 sw"
	"Optimized for speed -- ar 8/26/2001"
	^ (0 - x) @ (0 - y)! !

!Point methodsFor: 'point functions' stamp: ''!
normal
	"Answer a Point representing the unit vector rotated 90 deg clockwise."

	| n |
	n := y negated @ x.
	^n / (n x * n x + (n y * n y)) sqrt! !

!Point methodsFor: 'point functions' stamp: 'ar 8/26/2001 22:15'!
normalized
	"Optimized for speed -- ar 8/26/2001"
	| r |
	r := ((x*x) + (y * y)) sqrt.
	^(x / r) @ (y / r)! !

!Point methodsFor: 'point functions' stamp: 'laza 12/13/1999 11:43'!
octantOf: otherPoint 
	"Return 1..8 indicating relative direction to otherPoint.  
	1=ESE, 2=SSE, ... etc. clockwise to 8=ENE"
	"[Sensor anyButtonPressed] whileFalse: [(Display boundingBox center 
	octantOf: Sensor cursorPoint) printString displayAt: 0@0]"
	| quad moreHoriz |
	(x = otherPoint x and: [y > otherPoint y])
		ifTrue: [^ 6].
	"special case"
	(y = otherPoint y and: [x < otherPoint x])
		ifTrue: [^ 8].
	quad := self quadrantOf: otherPoint.
	moreHoriz := (x - otherPoint x) abs >= (y - otherPoint y) abs.
	(quad even eqv: moreHoriz)
		ifTrue: [^ quad * 2]
		ifFalse: [^ quad * 2 - 1]! !

!Point methodsFor: 'point functions' stamp: 'di 12/1/97 12:12'!
onLineFrom: p1 to: p2
	^ self onLineFrom: p1 to: p2 within: 2! !

!Point methodsFor: 'point functions' stamp: 'jm 2/24/98 08:34'!
onLineFrom: p1 to: p2 within: epsilon
	"Answer true if the receiver lies on the given line segment between p1 and p2 within a small epsilon."

	"is this point within the box spanning p1 and p2 expanded by epsilon? (optimized)"
	p1 x < p2 x
		ifTrue: [
			((x < (p1 x - epsilon)) or: [x > (p2 x + epsilon)]) ifTrue: [^ false]]
		ifFalse: [
			((x < (p2 x - epsilon)) or: [x > (p1 x + epsilon)]) ifTrue: [^ false]].
	p1 y < p2 y
		ifTrue: [
			((y < (p1 y - epsilon)) or: [y > (p2 y + epsilon)]) ifTrue: [^ false]]
		ifFalse: [
			((y < (p2 y - epsilon)) or: [y > (p1 y + epsilon)]) ifTrue: [^ false]].

	"it's in the box; is it on the line?"
	^ (self dist: (self nearestPointAlongLineFrom: p1 to: p2)) <= epsilon! !

!Point methodsFor: 'printing' stamp: ''!
printOn: aStream 
	"The receiver prints on aStream in terms of infix notation."

	x printOn: aStream.
	aStream nextPut: $@.
	y printOn: aStream! !

!Point methodsFor: 'point functions' stamp: '6/9/97 14:51 di'!
quadrantOf: otherPoint
	"Return 1..4 indicating relative direction to otherPoint.
	1 is downRight, 2=downLeft, 3=upLeft, 4=upRight"
	^ x <= otherPoint x
		ifTrue: [y < otherPoint y ifTrue: [1] ifFalse: [4]]
		ifFalse: [y <= otherPoint y ifTrue: [2] ifFalse: [3]]
"
[Sensor anyButtonPressed] whileFalse:
	[(Display boundingBox center quadrantOf: Sensor cursorPoint) printString displayAt: 0@0]
"! !

!Point methodsFor: 'polar coordinates' stamp: ''!
r
	"Answer the receiver's radius in polar coordinate system."

	^(self dotProduct: self) sqrt! !

!Point methodsFor: 'arithmetic' stamp: 'TRee 6/3/2004 11:09'!
reciprocal
    " Answer a Point with coordinates that are the reciprocals of mine. "
    " Method was missing from release. "
    " 20040301 20:50:35 TRee(Squeak3.6-5429-tree07.38) "

    ^ x reciprocal @ y reciprocal.
! !

!Point methodsFor: 'converting' stamp: 'di 12/3/97 19:00'!
rect: aPoint 
	"Answer a Rectangle that encompasses the receiver and aPoint.
	This is the most general infix way to create a rectangle."

	^ Rectangle 
		origin: (self min: aPoint)
		corner: (self max: aPoint)! !

!Point methodsFor: 'transforming' stamp: ''!
rotateBy: angle about: center
	"Even though Point.theta is measured CW, this rotates with the more conventional CCW interpretateion of angle."

	| p r theta |
	p := self - center.
	r := p r.
	theta := angle asFloat - p theta.
	^ (center x asFloat + (r * theta cos)) @
	  (center y asFloat - (r * theta sin))! !

!Point methodsFor: 'point functions' stamp: 'di 6/11/97 15:12'!
rotateBy: direction centerAt: c
	"Answer a Point which is rotated according to direction, about the point c.
	Direction must be one of #right (CW), #left (CCW) or #pi (180 degrees)."
	| offset |
	offset := self - c.
	direction == #right ifTrue: [^ (offset y negated @ offset x) + c].
	direction == #left ifTrue: [^ (offset y @ offset x negated) + c].
	direction == #pi ifTrue: [^ c - offset].
	self error: 'unrecognizable direction'! !

!Point methodsFor: 'truncation and round off' stamp: 'nice 2/5/2006 16:35'!
roundTo: grid
	"Answer a Point that is the receiver's x and y rounded to grid x and 
	grid y."
	
	| gridPoint |
	gridPoint := grid asPoint.
	^(x roundTo: gridPoint x) @ (y roundTo: gridPoint y)! !

!Point methodsFor: 'truncation and round off' stamp: 'jm 6/3/1998 12:21'!
rounded
	"Answer a Point that is the receiver's x and y rounded. Answer the receiver if its coordinates are already integral."

	(x isInteger and: [y isInteger]) ifTrue: [^ self].
	^ x rounded @ y rounded
! !

!Point methodsFor: 'transforming' stamp: ''!
scaleBy: factor 
	"Answer a Point scaled by factor (an instance of Point)."

	^(factor x * x) @ (factor y * y)! !

!Point methodsFor: 'transforming' stamp: 'di 12/4/97 14:34'!
scaleFrom: rect1 to: rect2
	"Produce a point stretched according to the stretch from rect1 to rect2"
	^ rect2 topLeft + (((x-rect1 left) * rect2 width // rect1 width)
					@ ((y-rect1 top) * rect2 height // rect1 height))! !

!Point methodsFor: 'private' stamp: 'tk 10/4/2001 16:16'!
setR: rho degrees: degrees 

	| radians |
	radians := degrees asFloat degreesToRadians.
	x := rho asFloat * radians cos.
	y := rho asFloat * radians sin.! !

!Point methodsFor: 'private' stamp: 'sw 3/21/2000 13:24'!
setX: xValue setY: yValue

	x := xValue.
	y := yValue! !

!Point methodsFor: 'geometry' stamp: 'laza 1/6/2000 10:30'!
sideOf: otherPoint 
	"Returns #left, #right or #center if the otherPoint lies to the left, right 
	or on the line given by the vector from 0@0 to self"
	| side |
	side := (self crossProduct: otherPoint) sign.
	^ {#right. #center. #left} at: side + 2
! !

!Point methodsFor: 'point functions' stamp: 'ar 4/18/1999 05:17'!
sortsBefore: otherPoint
	"Return true if the receiver sorts before the other point"
	^y = otherPoint y
		ifTrue:[x <= otherPoint x]
		ifFalse:[y <= otherPoint y]! !

!Point methodsFor: 'point functions' stamp: 'ar 5/23/2001 21:29'!
squaredDistanceTo: aPoint
	"Answer the distance between aPoint and the receiver."
	| delta |
	delta := aPoint - self.
	^delta dotProduct: delta! !

!Point methodsFor: 'printing' stamp: 'ar 7/8/2006 19:15'!
storeOn: aStream 
	"x@y printed form is good for storing too"
	aStream nextPut: $(.
	self printOn: aStream.
	aStream nextPut: $).
! !

!Point methodsFor: 'polar coordinates' stamp: 'hk 11/10/2005 10:07'!
theta
	"Answer the angle the receiver makes with origin in radians. right is 0; 
	down is 90. 
	Corrected the constants from single precision to 64 Bit precision 
	and changed the sends in case of overflow to constants HK 2005-07-23"

	| tan theta |
	x = 0
		ifTrue: [y >= 0
				ifTrue: [^ 1.570796326794897 "90.0 degreesToRadians"]
				ifFalse: [^ 4.71238898038469 "270.0 degreesToRadians"]]
		ifFalse: 
			[tan := y asFloat / x asFloat.
			theta := tan arcTan.
			x >= 0
				ifTrue: [y >= 0
						ifTrue: [^theta]
						ifFalse: [^"360.0 degreesToRadians" 6.283185307179586 + theta]]
				ifFalse: [^"180.0 degreesToRadians" 3.141592653589793 + theta]]! !

!Point methodsFor: 'geometry' stamp: 'ar 4/6/2000 18:37'!
to: end1 intersects: start2 to: end2 
	"Returns true if the linesegment from start1 (=self) to end1 intersects      
	    with the segment from start2 to end2, otherwise false."
	| start1 sideStart sideEnd |
	start1 := self.
	(((start1 = start2 or: [end1 = end2])
		or: [start1 = end2])
		or: [start2 = end1])
		ifTrue: [^ true].
	sideStart := start1 to: end1 sideOf: start2.
	sideEnd := start1 to: end1 sideOf: end2.
	sideStart = sideEnd ifTrue: [^ false].
	sideStart := start2 to: end2 sideOf: start1.
	sideEnd := start2 to: end2 sideOf: end1.
	sideStart = sideEnd ifTrue: [^ false].
	^ true! !

!Point methodsFor: 'geometry' stamp: 'laza 1/5/2000 11:50'!
to: end sideOf: otherPoint 
	"Returns #left, #right, #center if the otherPoint lies to the left, right or on the line given by the vector from self to end"
	^ end - self sideOf: otherPoint - self! !

!Point methodsFor: 'transforming' stamp: ''!
translateBy: delta 
	"Answer a Point translated by delta (an instance of Point)."

	^(delta x + x) @ (delta y + y)! !

!Point methodsFor: 'point functions' stamp: 'ar 11/12/1998 01:44'!
transposed
	^y@x! !

!Point methodsFor: 'geometry' stamp: 'laza 1/17/2000 15:47'!
triangleArea: b with: c
	"Returns twice the area of the oriented triangle (a, b, c), i.e., the   
	area is positive if the triangle is oriented counterclockwise"
	^ b x - self x * (c y - self y) - (b y - self y * (c x - self x))! !

!Point methodsFor: 'truncation and round off' stamp: ''!
truncateTo: grid
	"Answer a Point that is the receiver's x and y truncated to grid x and 
	grid y."
	| gridPoint |
	gridPoint := grid asPoint.
	^(x truncateTo: gridPoint x) @ (y truncateTo: gridPoint y)! !

!Point methodsFor: 'truncation and round off' stamp: 'jm 5/29/1998 15:53'!
truncated
	"Answer a Point whose x and y coordinates are integers. Answer the receiver if its coordinates are already integral."

	(x isInteger and: [y isInteger]) ifTrue: [^ self].
	^ x truncated @ y truncated
! !

!Point methodsFor: 'copying' stamp: 'tk 8/19/1998 16:05'!
veryDeepCopyWith: deepCopier
	"Return self.  I am immutable in the Morphic world.  Do not record me."! !

!Point methodsFor: 'accessing' stamp: ''!
x
	"Answer the x coordinate."

	^x! !

!Point methodsFor: 'accessing' stamp: ''!
y
	"Answer the y coordinate."

	^y! !

Object subclass: #Rectangle
	instanceVariableNames: 'origin corner'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Primitives'!

!Rectangle commentStamp: '<historical>' prior: 0!
I represent a rectangular area of the screen. Arithmetic functions take points as arguments and carry out scaling and translating operations to create new instances of me. Rectangle functions create new instances by determining intersections of rectangles with rectangles.!

Rectangle subclass: #Quadrangle
	instanceVariableNames: 'borderWidth borderColor insideColor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Primitives'!

!Quadrangle commentStamp: '<historical>' prior: 0!
I represent a particular kind of Rectangle that has a border and inside color.!

!Quadrangle class methodsFor: 'instance creation' stamp: 'sw 5/4/2001 17:12'!
exampleInViewer
	"Create a sample Quadrangle and open a Viewer on it"

	(self region: (100@100 extent: 100@50) borderWidth: (1 + (6 atRandom)) borderColor: Color black insideColor: (Color perform: #(green red blue yellow) atRandom)) beViewed

"Quadrangle exampleInViewer"! !

!Quadrangle class methodsFor: 'instance creation' stamp: 'fc 1/16/2005 21:20'!
origin: originPoint corner: cornerPoint 
	"Override Rectangles origin:corner: in order to get initialized.

	Answer an instance of me whose corners (top left and bottom right) are 
	determined by the arguments."

	^self new setOrigin: originPoint corner: cornerPoint! !

!Quadrangle class methodsFor: 'instance creation' stamp: ''!
region: aRectangle borderWidth: anInteger borderColor: aMask1 insideColor: aMask2
	"Answer an instance of me with rectangle, border width and color, and 
	inside color determined by the arguments."

	^super new
		setRegion: aRectangle
		borderWidth: anInteger
		borderColor: aMask1
		insideColor: aMask2! !

!Quadrangle methodsFor: 'transforming' stamp: ''!
align: aPoint1 with: aPoint2 
	"Answer a new Quadrangle translated by aPoint2 - aPoint1.
	 5/24/96 sw: removed hard-coded class name so subclasses can gain same functionality."

	^ self class
		region: (super translateBy: aPoint2 - aPoint1)
		borderWidth: borderWidth
		borderColor: borderColor
		insideColor: insideColor! !

!Quadrangle methodsFor: 'transforming' stamp: ''!
alignedTo: alignPointSelector
	"Return a copy with offset according to alignPointSelector which is one of...
	#(topLeft, topCenter, topRight, leftCenter, center, etc)
	 5/24/96 sw: removed hard-coded class name so subclasses can gain same functionality."

	^ self class
		region: (super translateBy: (0@0) - (self perform: alignPointSelector))
		borderWidth: borderWidth
		borderColor: borderColor
		insideColor: insideColor! !

!Quadrangle methodsFor: 'bordering' stamp: ''!
borderColor
	"Answer the form that is the borderColor of the receiver."

	^borderColor! !

!Quadrangle methodsFor: 'bordering' stamp: ''!
borderColor: aColor 
	"Set the borderColor of the receiver to aColor, a Form."

	borderColor := aColor! !

!Quadrangle methodsFor: 'bordering' stamp: ''!
borderWidth
	"Answer the borderWidth of the receiver."

	^borderWidth! !

!Quadrangle methodsFor: 'bordering' stamp: ''!
borderWidth: anInteger 
	"Set the borderWidth of the receiver to anInteger."

	borderWidth := anInteger! !

!Quadrangle methodsFor: 'bordering' stamp: ''!
borderWidthLeft: anInteger1 right: anInteger2 top: anInteger3 bottom: anInteger4
	"Set the border width of the receiver to a Rectangle that represents the 
	left, right, top, and bottom border widths."

	borderWidth := anInteger1 @ anInteger3 corner: anInteger2 @ anInteger4! !

!Quadrangle methodsFor: 'displaying-Display' stamp: ''!
display 
	"Display the border and insideRegion of the receiver on the Display."

	self displayOn: Display! !

!Quadrangle methodsFor: 'displaying-Display' stamp: ''!
displayAlign: aPoint1 with: aPoint2 clippingBox: aRectangle 
	"Display the border and region of the receiver on the Display so that its 
	position at aPoint1 is aligned with position aPoint2. The displayed 
	information should be clipped so that only information with the area 
	determined by aRectangle is displayed." 

	self displayOn: Display align: aPoint1 with: aPoint2 clippingBox: aRectangle! !

!Quadrangle methodsFor: 'displaying-generic' stamp: ''!
displayOn: aDisplayMedium
	"Display the border and insideRegion of the receiver."

	borderWidth ~~ 0
		ifTrue:	[aDisplayMedium
				border: self region
				widthRectangle: borderWidth
				rule: Form over
				fillColor: borderColor].
	insideColor ~~ nil
		ifTrue:	[aDisplayMedium fill: self inside fillColor: insideColor]! !

!Quadrangle methodsFor: 'displaying-generic' stamp: ''!
displayOn: aDisplayMedium align: aPoint1 with: aPoint2 clippingBox: aRectangle
	"Display the border and region of the receiver so that its position at 
	aPoint1 is aligned with position aPoint2. The displayed information 
	should be clipped so that only information with the area determined by 
	aRectangle is displayed."

	| savedRegion |
	savedRegion := self region.
	self region: ((savedRegion align: aPoint1 with: aPoint2) intersect: aRectangle).
	self displayOn: aDisplayMedium.
	self region: savedRegion! !

!Quadrangle methodsFor: 'displaying-generic' stamp: ''!
displayOn: aDisplayMedium transformation: aWindowingTransformation clippingBox: aRectangle
	"Display the border and region of the receiver so that it is scaled and 
	translated with respect to aWindowingTransformation. The displayed 
	information should be clipped so that only information with the area 
	determined by aRectangle is displayed."

	| screenRectangle |
	screenRectangle := 
		(aWindowingTransformation applyTo: self) intersect: aRectangle.
	borderWidth ~~ 0 & (insideColor ~~ nil)
		ifTrue: 
			[aDisplayMedium fill: screenRectangle fillColor: Color black "borderColor".
			aDisplayMedium
				fill: (screenRectangle insetBy: borderWidth)
				fillColor: insideColor]! !

!Quadrangle methodsFor: 'displaying-generic' stamp: ''!
displayOnPort: aPort at: p
	"Display the border and insideRegion of the receiver."

	(insideColor == nil or: [borderWidth <= 0])
		ifFalse: [aPort fill: (self region translateBy: p) 
			fillColor: borderColor rule: Form over].
	insideColor == nil
		ifFalse: [aPort fill: (self inside translateBy: p) 
			fillColor: insideColor rule: Form over]! !

!Quadrangle methodsFor: 'displaying-Display' stamp: ''!
displayTransformation: aWindowingTransformation clippingBox: aRectangle 
	"Display the border and region of the receiver on the Display so that it 
	is scaled and translated with respect to aWindowingTransformation. The 
	displayed information should be clipped so that only information with 
	the area determined by aRectangle is displayed." 

	self displayOn: Display transformation: aWindowingTransformation clippingBox: aRectangle! !

!Quadrangle methodsFor: 'initialize-release' stamp: ''!
initialize
	"Initialize the region to a null Rectangle, the borderWidth to 1, the 
	borderColor to black, and the insideColor to white."

	origin := 0 @ 0.
	corner := 0 @ 0.
	borderWidth := 1.
	borderColor := Color black.
	insideColor := Color white.
! !

!Quadrangle methodsFor: 'bordering' stamp: ''!
inside
	"Answer a Rectangle that is the receiver inset by the borderWidth."

	^self insetBy: borderWidth! !

!Quadrangle methodsFor: 'bordering' stamp: ''!
insideColor
	"Answer the form that is the insideColor of the receiver."

	^insideColor! !

!Quadrangle methodsFor: 'bordering' stamp: ''!
insideColor: aColor 
	"Set the insideColor of the receiver to aColor, a Form."

	insideColor := aColor! !

!Quadrangle methodsFor: 'rectangle functions' stamp: ''!
intersect: aRectangle 
	"Answer a new Quadrangle whose region is the intersection of the 
	receiver's area and aRectangle.
	 5/24/96 sw: removed hard-coded class name so subclasses can gain same functionality."

	^ self class
	 	region: (super intersect: aRectangle)
		borderWidth: borderWidth
		borderColor: borderColor
		insideColor: insideColor! !

!Quadrangle methodsFor: 'bordering' stamp: ''!
region
	"Answer a Rectangle that defines the area of the receiver."

	^origin corner: corner! !

!Quadrangle methodsFor: 'bordering' stamp: ''!
region: aRectangle 
	"Set the rectangular area of the receiver to aRectangle."

	origin := aRectangle origin.
	corner := aRectangle corner! !

!Quadrangle methodsFor: 'transforming' stamp: ''!
scaleBy: aPoint 
	"Answer a new Quadrangle scaled by aPoint.
	 5/24/96 sw: removed hard-coded class name so subclasses can gain same functionality."

	^ self class
		region: (super scaleBy: aPoint)
		borderWidth: borderWidth
		borderColor: borderColor
		insideColor: insideColor! !

!Quadrangle methodsFor: 'bordering' stamp: 'sw 5/4/2001 18:30'!
setHeight: aNumber 
	"Set the receiver's height"

	self region: (origin extent: (self width @ aNumber))! !

!Quadrangle methodsFor: 'bordering' stamp: 'sw 5/4/2001 17:54'!
setLeft: aNumber 
	"Move the receiver so that its left edge is given by aNumber.  An example of a setter to go with #left"

	self region: ((aNumber @ origin y) extent: self extent)! !

!Quadrangle methodsFor: 'private' stamp: ''!
setRegion: aRectangle borderWidth: anInteger borderColor: aMask1 insideColor: aMask2

	origin := aRectangle origin.
	corner := aRectangle corner.
	borderWidth := anInteger.
	borderColor := aMask1.
	insideColor := aMask2! !

!Quadrangle methodsFor: 'bordering' stamp: 'sw 5/4/2001 18:19'!
setRight: aNumber 
	"Move the receiver so that its right edge is given by aNumber.  An example of a setter to go with #right"

	self region: ((origin x + (aNumber - self right) @ origin y) extent: self extent)! !

!Quadrangle methodsFor: 'bordering' stamp: 'sw 5/4/2001 18:26'!
setWidth: aNumber 
	"Set the receiver's width"

	self region: (origin extent: (aNumber @ self height))! !

!Quadrangle methodsFor: 'transforming' stamp: ''!
translateBy: aPoint 
	"Answer a new Quadrangle translated by aPoint.
	 5/24/96 sw: removed hard-coded class name so subclasses can gain same functionality."

	^ self class
		region: (super translateBy: aPoint)
		borderWidth: borderWidth
		borderColor: borderColor
		insideColor: insideColor! !

!Rectangle class methodsFor: 'instance creation' stamp: 'tk 3/9/97'!
center: centerPoint extent: extentPoint 
	"Answer an instance of me whose center is centerPoint and width 
	by height is extentPoint.  "

	^self origin: centerPoint - (extentPoint//2) extent: extentPoint! !

!Rectangle class methodsFor: 'instance creation' stamp: 'di 12/1/97 10:42'!
encompassing: listOfPoints
	"A number of callers of encompass: should use this method."
	| topLeft bottomRight |
	topLeft := bottomRight := nil.
	listOfPoints do:
		[:p | topLeft == nil
			ifTrue: [topLeft := bottomRight := p]
			ifFalse: [topLeft := topLeft min: p.
					bottomRight := bottomRight max: p]].
	^ topLeft corner: bottomRight! !

!Rectangle class methodsFor: 'instance creation' stamp: ''!
fromUser
	"Answer an instance of me that is determined by having the user 
	designate the top left and bottom right corners. The gridding for user 
	selection is 1@1."

	^self fromUser: 1 @ 1! !

!Rectangle class methodsFor: 'instance creation' stamp: ''!
fromUser: gridPoint
	"Answer a Rectangle that is determined by having the user 
	designate the top left and bottom right corners. 
	The cursor reamins linked with the sensor, but
	the outline is kept gridded."
	| originRect |
	originRect := Cursor origin showWhile: 
		[((Sensor cursorPoint grid: gridPoint) extent: 0@0) newRectFrom:
			[:f | (Sensor cursorPoint grid: gridPoint) extent: 0@0]].
	^ Cursor corner showWhile:
		[originRect newRectFrom:
			[:f | f origin corner: (Sensor cursorPoint grid: gridPoint)]]! !

!Rectangle class methodsFor: 'instance creation' stamp: 'md 12/3/2004 00:02'!
left: leftNumber right: rightNumber top: topNumber bottom: bottomNumber 
	"Answer an instance of me whose left, right, top, and bottom coordinates 
	are determined by the arguments."

	^ self basicNew setOrigin: leftNumber @ topNumber corner: rightNumber @ bottomNumber! !

!Rectangle class methodsFor: 'instance creation' stamp: 'btr 2/14/2003 16:29'!
merging: listOfRects 
	"A number of callers of merge: should use this method."
	| minX minY maxX maxY |
	listOfRects
		do: [:r | minX
				ifNil: [minX := r topLeft x. minY := r topLeft y.
					maxX := r bottomRight x. maxY := r bottomRight y]
				ifNotNil: [minX := minX min: r topLeft x. minY := minY min: r topLeft y.
					maxX := maxX max: r bottomRight x. maxY := maxY max: r bottomRight y]].
	^ minX@minY corner: maxX@maxY! !

!Rectangle class methodsFor: 'instance creation' stamp: 'md 12/3/2004 00:04'!
origin: originPoint corner: cornerPoint 
	"Answer an instance of me whose corners (top left and bottom right) are 
	determined by the arguments."

	^self basicNew setOrigin: originPoint corner: cornerPoint! !

!Rectangle class methodsFor: 'instance creation' stamp: 'md 12/3/2004 00:03'!
origin: originPoint extent: extentPoint 
	"Answer an instance of me whose top left corner is originPoint and width 
	by height is extentPoint."

	^self basicNew setOrigin: originPoint corner: originPoint + extentPoint! !

!Rectangle class methodsFor: 'instance creation' stamp: ''!
originFromUser: extentPoint 
	"Answer an instance of me that is determined by having the user 
	designate the top left corner. The width and height are determined by 
	extentPoint. The gridding for user selection is 1@1."

	^self originFromUser: extentPoint grid: 1 @ 1! !

!Rectangle class methodsFor: 'instance creation' stamp: ''!
originFromUser: extentPoint grid: gridPoint 
	"Answer an instance of me that is determined by having the user 
	designate the top left corner. The width and height are determined by 
	extentPoint. The gridding for user selection is scaleFactor. Assumes that 
	the sender has determined an extent that is a proper multiple of 
	scaleFactor."

	^ Cursor origin showWhile: 
		[((Sensor cursorPoint grid: gridPoint) extent: extentPoint) newRectFrom:
			[:f | (Sensor cursorPoint grid: gridPoint) extent: extentPoint]].
! !

!Rectangle methodsFor: 'comparing' stamp: ''!
= aRectangle 
	"Answer true if the receiver's species, origin and corner match aRectangle's."

	self species = aRectangle species
		ifTrue: [^origin = aRectangle origin and: [corner = aRectangle corner]]
		ifFalse: [^false]! !

!Rectangle methodsFor: 'accessing' stamp: 'acg 2/23/2000 00:52'!
aboveCenter
	"Answer the point slightly above the center of the receiver."

	^self topLeft + self bottomRight // (2@3)! !

!Rectangle methodsFor: 'rectangle functions' stamp: 'di 10/22/1998 16:11'!
adjustTo: newRect along: side 
	"Return a copy adjusted to fit a neighbor that has changed size."
	side = #left ifTrue: [^ self withRight: newRect left].
	side = #right ifTrue: [^ self withLeft: newRect right].
	side = #top ifTrue: [^ self withBottom: newRect top].
	side = #bottom ifTrue: [^ self withTop: newRect bottom].! !

!Rectangle methodsFor: 'transforming' stamp: ''!
align: aPoint1 with: aPoint2 
	"Answer a Rectangle that is a translated by aPoint2 - aPoint1."

	^self translateBy: aPoint2 - aPoint1! !

!Rectangle methodsFor: 'rectangle functions' stamp: 'ar 1/5/2002 18:04'!
allAreasOutsideList: aCollection do: aBlock
	"Enumerate aBlock with all areas of the receiver not overlapping 
	any rectangle in the given collection"
	^self allAreasOutsideList: aCollection startingAt: 1 do: aBlock! !

!Rectangle methodsFor: 'rectangle functions' stamp: 'ar 1/5/2002 18:03'!
allAreasOutsideList: aCollection startingAt: startIndex do: aBlock
	"Enumerate aBlock with all areas of the receiver not overlapping 
	any rectangle in the given collection"
	| yOrigin yCorner aRectangle index rr |
	index := startIndex.

	"Find the next intersecting rectangle from aCollection"
	[index <= aCollection size ifFalse:[^aBlock value: self].
	aRectangle := aCollection at: index.
	origin <= aRectangle corner and: [aRectangle origin <= corner]] 
		whileFalse:[index := index + 1].

	"aRectangle is intersecting; process it"
	aRectangle origin y > origin y 
		ifTrue: [rr := origin corner: corner x @ (yOrigin := aRectangle origin y).
				rr allAreasOutsideList: aCollection startingAt: index+1 do: aBlock]
		ifFalse: [yOrigin := origin y].
	aRectangle corner y < corner y
		ifTrue: [rr := origin x @ (yCorner := aRectangle corner y) corner: corner.
				rr allAreasOutsideList: aCollection startingAt: index+1 do: aBlock]
		ifFalse: [yCorner := corner y].
	aRectangle origin x > origin x 
		ifTrue: [rr := origin x @ yOrigin corner: aRectangle origin x @ yCorner.
				rr allAreasOutsideList: aCollection startingAt: index+1 do: aBlock].
	aRectangle corner x < corner x 
		ifTrue: [rr := aRectangle corner x @ yOrigin corner: corner x @ yCorner.
				rr allAreasOutsideList: aCollection startingAt: index+1 do: aBlock].! !

!Rectangle methodsFor: 'rectangle functions' stamp: ''!
amountToTranslateWithin: aRectangle
	"Answer a Point, delta, such that self + delta is forced within aRectangle."
	"Altered so as to prefer to keep self topLeft inside when all of self
	cannot be made to fit 7/27/96 di"
	| dx dy |
	dx := 0.  dy := 0.
	self right > aRectangle right ifTrue: [dx := aRectangle right - self right].
	self bottom > aRectangle bottom ifTrue: [dy := aRectangle bottom - self bottom].
	(self left + dx) < aRectangle left ifTrue: [dx := aRectangle left - self left].
	(self top + dy) < aRectangle top ifTrue: [dy := aRectangle top - self top].
	^ dx@dy! !

!Rectangle methodsFor: 'accessing' stamp: 'di 11/17/2001 14:01'!
area
	"Answer the receiver's area, the product of width and height."
	| w |
	(w := self width) <= 0 ifTrue: [^ 0].
	^ w * self height max: 0! !

!Rectangle methodsFor: 'rectangle functions' stamp: 'ar 10/25/2005 17:22'!
areasOutside: aRectangle
	"Answer an Array of Rectangles comprising the parts of the receiver not 
	intersecting aRectangle."

	| areas yOrigin yCorner |
	"Make sure the intersection is non-empty"
     (self intersects: aRectangle)
		ifFalse: [^Array with: self].
	areas := OrderedCollection new.
	aRectangle origin y > origin y
		ifTrue: [areas addLast: (origin corner: corner x @ (yOrigin := aRectangle origin y))]
		ifFalse: [yOrigin := origin y].
	aRectangle corner y < corner y
		ifTrue: [areas addLast: (origin x @ (yCorner := aRectangle corner y) corner: corner)]
		ifFalse: [yCorner := corner y].
	aRectangle origin x > origin x 
		ifTrue: [areas addLast: (origin x @ yOrigin corner: aRectangle origin x @ yCorner)].
	aRectangle corner x < corner x 
		ifTrue: [areas addLast: (aRectangle corner x @ yOrigin corner: corner x @ yCorner)].
	^areas! !

!Rectangle methodsFor: 'rectangle functions' stamp: 'di 10/21/1998 16:00'!
bordersOn: her along: herSide 
	(herSide = #right and: [self left = her right])
	| (herSide = #left and: [self right = her left])
		ifTrue:
		[^ (self top max: her top) < (self bottom min: her bottom)].
	(herSide = #bottom and: [self top = her bottom])
	| (herSide = #top and: [self bottom = her top])
		ifTrue:
		[^ (self left max: her left) < (self right min: her right)].
	^ false! !

!Rectangle methodsFor: 'accessing' stamp: ''!
bottom
	"Answer the position of the receiver's bottom horizontal line."

	^corner y! !

!Rectangle methodsFor: 'accessing' stamp: 'ar 10/26/2000 22:17'!
bottom: aNumber
	^origin corner: corner x @ aNumber! !

!Rectangle methodsFor: 'accessing' stamp: ''!
bottomCenter
	"Answer the point at the center of the bottom horizontal line of the 
	receiver."

	^self center x @ self bottom! !

!Rectangle methodsFor: 'accessing' stamp: ''!
bottomLeft
	"Answer the point at the left edge of the bottom horizontal line of the 
	receiver."

	^origin x @ corner y! !

!Rectangle methodsFor: 'accessing' stamp: ''!
bottomRight
	"Answer the point at the right edge of the bottom horizontal line of the 
	receiver."

	^corner! !

!Rectangle methodsFor: 'accessing' stamp: ''!
boundingBox
	^ self! !

!Rectangle methodsFor: 'accessing' stamp: ''!
center
	"Answer the point at the center of the receiver."

	^self topLeft + self bottomRight // 2! !

!Rectangle methodsFor: 'transforming' stamp: ''!
centeredBeneath: aRectangle
	 "Move the reciever so that its top center point coincides with the bottom center point of aRectangle.  5/20/96 sw:"

	^ self align: self topCenter with: aRectangle bottomCenter! !

!Rectangle methodsFor: 'testing' stamp: ''!
containsPoint: aPoint 
	"Answer whether aPoint is within the receiver."

	^origin <= aPoint and: [aPoint < corner]! !

!Rectangle methodsFor: 'testing' stamp: ''!
containsRect: aRect
	"Answer whether aRect is within the receiver (OK to coincide)."

	^ aRect origin >= origin and: [aRect corner <= corner]
! !

!Rectangle methodsFor: 'accessing' stamp: ''!
corner
	"Answer the point at the bottom right corner of the receiver."

	^corner! !

!Rectangle methodsFor: 'accessing' stamp: ''!
corners
	"Return an array of corner points in the order of a quadrilateral spec for WarpBlt."

	^ Array
		with: self topLeft
		with: self bottomLeft
		with: self bottomRight
		with: self topRight
! !

!Rectangle methodsFor: 'FMP' stamp: 'RAA 6/1/1999 17:41'!
deltaToEnsureInOrCentered: r extra: aNumber

	| dX dY halfXDiff halfYDiff |
	dX := dY := 0.
	halfXDiff := (r width - self width * aNumber) truncated.
	halfYDiff := (r height - self height  * aNumber) truncated.
	self left < r left
		ifTrue: [dX := self left - r left - halfXDiff]
		ifFalse: [self right > r right ifTrue: [dX := self right - r right + halfXDiff]].
	self top < r top
		ifTrue: [dY := self top - r top - halfYDiff]
		ifFalse: [self bottom > r bottom ifTrue: [dY := self bottom - r bottom + halfYDiff]].
	^dX @ dY
! !

!Rectangle methodsFor: 'rectangle functions' stamp: ''!
encompass: aPoint 
	"Answer a Rectangle that contains both the receiver and aPoint.  5/30/96 sw"

	^ Rectangle 
		origin: (origin min: aPoint)
		corner: (corner max:  aPoint)! !

!Rectangle methodsFor: 'rectangle functions' stamp: 'md 10/4/2005 14:42'!
expandBy: delta 
	"Answer a Rectangle that is outset from the receiver by delta. delta is a 
	Rectangle, Point, or scalar."

	(delta isRectangle)
		ifTrue: [^Rectangle 
					origin: origin - delta origin 
					corner: corner + delta corner]
		ifFalse: [^Rectangle 
					origin: origin - delta 
					corner: corner + delta]! !

!Rectangle methodsFor: 'rectangle functions' stamp: 'md 10/4/2005 14:42'!
extendBy: delta 
	"Answer a Rectangle with the same origin as the receiver, but whose corner is offset by delta. delta is a 
	Rectangle, Point, or scalar."

	(delta isRectangle)
		ifTrue: [^Rectangle 
					origin: origin
					corner: corner + delta corner]
		ifFalse: [^Rectangle 
					origin: origin
					corner: corner + delta]! !

!Rectangle methodsFor: 'accessing' stamp: ''!
extent
	"Answer with a rectangle with origin 0@0 and corner the receiver's 
	width @ the receiver's height."

	^corner - origin! !

!Rectangle methodsFor: 'transforming' stamp: 'di 6/11/97 16:24'!
flipBy: direction centerAt: aPoint 
	"Return a copy flipped #vertical or #horizontal, about aPoint."
	^ (origin flipBy: direction centerAt: aPoint)
		rect: (corner flipBy: direction centerAt: aPoint)! !

!Rectangle methodsFor: 'rectangle functions' stamp: 'di 10/21/1998 16:11'!
forPoint: aPoint closestSideDistLen: sideDistLenBlock
	"Evaluate the block with my side (symbol) closest to aPoint,
		the approx distance of aPoint from that side, and
		the length of the side (or 0 if aPoint is beyond the side)"
	| side |
	side := self sideNearestTo: aPoint.
	side == #right ifTrue:
		[^ sideDistLenBlock value: side value: (self right - aPoint x) abs
			value: ((aPoint y between: self top and: self bottom)
						ifTrue: [self height] ifFalse: [0])].
	side == #left ifTrue:
		[^ sideDistLenBlock value: side value: (self left - aPoint x) abs
			value: ((aPoint y between: self top and: self bottom)
						ifTrue: [self height] ifFalse: [0])].
	side == #bottom ifTrue:
		[^ sideDistLenBlock value: side value: (self bottom - aPoint y) abs
			value: ((aPoint x between: self left and: self right)
						ifTrue: [self width] ifFalse: [0])].
	side == #top ifTrue:
		[^ sideDistLenBlock value: side value: (self top - aPoint y) abs
			value: ((aPoint x between: self left and: self right)
						ifTrue: [self width] ifFalse: [0])].! !

!Rectangle methodsFor: 'testing' stamp: ''!
hasPositiveExtent
	^ (corner x > origin x) and: [corner y > origin y]! !

!Rectangle methodsFor: 'comparing' stamp: ''!
hash
	"Hash is reimplemented because = is implemented."

	^origin hash bitXor: corner hash! !

!Rectangle methodsFor: 'comparing' stamp: ''!
hashMappedBy: map
	"My hash is independent of my oop."

	^self hash! !

!Rectangle methodsFor: 'accessing' stamp: ''!
height
	"Answer the height of the receiver."

	^corner y - origin y! !

!Rectangle methodsFor: 'accessing' stamp: ''!
innerCorners
	"Return an array of inner corner points,
	ie, the most extreme pixels included,
	in the order of a quadrilateral spec for WarpBlt"
	| r1 |
	r1 := self topLeft corner: self bottomRight - (1@1).
	^ Array with: r1 topLeft with: r1 bottomLeft with: r1 bottomRight with: r1 topRight! !

!Rectangle methodsFor: 'rectangle functions' stamp: 'md 10/2/2005 21:51'!
insetBy: delta 
	"Answer a Rectangle that is inset from the receiver by delta. delta is a 
	Rectangle, Point, or scalar."

	
	(delta isRectangle)
		ifTrue: [^Rectangle 
					origin: origin + delta origin 
					corner: corner - delta corner]
		ifFalse: [^Rectangle 
					origin: origin + delta 
					corner: corner - delta]! !

!Rectangle methodsFor: 'rectangle functions' stamp: ''!
insetOriginBy: originDeltaPoint cornerBy: cornerDeltaPoint 
	"Answer a Rectangle that is inset from the receiver by a given amount in 
	the origin and corner."

	^Rectangle
		origin: origin + originDeltaPoint
		corner: corner - cornerDeltaPoint! !

!Rectangle methodsFor: 'rectangle functions' stamp: 'ar 11/12/2000 19:10'!
intersect: aRectangle 
	"Answer a Rectangle that is the area in which the receiver overlaps with 
	aRectangle. Optimized for speed; old code read:
		^Rectangle 
			origin: (origin max: aRectangle origin)
			corner: (corner min: aRectangle corner)
	"
	| aPoint left right top bottom |
	aPoint := aRectangle origin.
	aPoint x > origin x ifTrue:[left := aPoint x] ifFalse:[left := origin x].
	aPoint y > origin y ifTrue:[top := aPoint y] ifFalse:[top := origin y].
	aPoint := aRectangle corner.
	aPoint x < corner x ifTrue:[right := aPoint x] ifFalse:[right := corner x].
	aPoint y < corner y ifTrue:[bottom := aPoint y] ifFalse:[bottom := corner y].
	^Rectangle
		origin: (left@top)
		corner: (right@bottom)
! !

!Rectangle methodsFor: 'testing' stamp: 'ar 9/10/2000 15:29'!
intersects: aRectangle 
	"Answer whether aRectangle intersects the receiver anywhere."
	"Optimized; old code answered:
		(origin max: aRectangle origin) < (corner min: aRectangle corner)"

	| rOrigin rCorner |
	rOrigin := aRectangle origin.
	rCorner := aRectangle corner.
	rCorner x <= origin x	ifTrue: [^ false].
	rCorner y <= origin y	ifTrue: [^ false].
	rOrigin x >= corner x	ifTrue: [^ false].
	rOrigin y >= corner y	ifTrue: [^ false].
	^ true
! !

!Rectangle methodsFor: 'testing' stamp: 'md 10/2/2005 21:51'!
isRectangle
	^true! !

!Rectangle methodsFor: 'self evaluating' stamp: 'sd 7/31/2005 21:49'!
isSelfEvaluating
	^ self class == Rectangle! !

!Rectangle methodsFor: 'testing' stamp: ''!
isTall
	^ self height > self width! !

!Rectangle methodsFor: 'testing' stamp: ''!
isWide
	^ self width > self height! !

!Rectangle methodsFor: 'testing' stamp: 'ar 10/29/2000 19:03'!
isZero
	^origin isZero and:[corner isZero]! !

!Rectangle methodsFor: 'accessing' stamp: ''!
left
	"Answer the position of the receiver's left vertical line."

	^origin x! !

!Rectangle methodsFor: 'accessing' stamp: 'ar 10/26/2000 22:16'!
left: aNumber
	^aNumber @ origin y corner: corner! !

!Rectangle methodsFor: 'accessing' stamp: ''!
leftCenter
	"Answer the point at the center of the receiver's left vertical line."

	^self left @ self center y! !

!Rectangle methodsFor: 'rectangle functions' stamp: ''!
merge: aRectangle 
	"Answer a Rectangle that contains both the receiver and aRectangle."

	^Rectangle 
		origin: (origin min: aRectangle origin)
		corner: (corner max: aRectangle corner)! !

!Rectangle methodsFor: 'transforming' stamp: 'JMM 10/21/2003 17:26'!
newRectButtonPressedDo: newRectBlock 
	"Track the outline of a new rectangle until mouse button 
	changes. newFrameBlock produces each new rectangle from the 
	previous. Only tracks while mouse is down."
	| rect newRect buttonNow aHand delay |
	delay := Delay forMilliseconds: 10.
	buttonNow := Sensor anyButtonPressed.
	rect := self.
	Display
		border: rect
		width: 2
		rule: Form reverse
		fillColor: Color gray.
	[buttonNow]
		whileTrue: [delay wait.
			buttonNow := Sensor anyButtonPressed.
			newRect := newRectBlock value: rect.
			newRect = rect
				ifFalse: [Display
						border: rect
						width: 2
						rule: Form reverse
						fillColor: Color gray.
					Display
						border: newRect
						width: 2
						rule: Form reverse
						fillColor: Color gray.
					rect := newRect]].
	Display
		border: rect
		width: 2
		rule: Form reverse
		fillColor: Color gray.
	" pay the price for reading the sensor directly ; get this party started "
	Smalltalk isMorphic
		ifTrue: [aHand := World activeHand.
			aHand newMouseFocus: nil;
				 showTemporaryCursor: nil;
				 flushEvents].
	Sensor processEvent: Sensor createMouseEvent.
	^ rect! !

!Rectangle methodsFor: 'transforming' stamp: 'JMM 10/21/2003 17:28'!
newRectFrom: newRectBlock
	"Track the outline of a new rectangle until mouse button changes.
	newFrameBlock produces each new rectangle from the previous"
	| rect newRect buttonStart buttonNow aHand delay |
	delay := Delay forMilliseconds: 10.
	buttonStart := buttonNow := Sensor anyButtonPressed.
	rect := self.
	Display border: rect width: 2 rule: Form reverse fillColor: Color gray.
	[buttonNow == buttonStart] whileTrue: 
		[delay wait.
		buttonNow := Sensor anyButtonPressed.
		newRect := newRectBlock value: rect.
		newRect = rect ifFalse:
			[Display border: rect width: 2 rule: Form reverse fillColor: Color gray.
			Display border: newRect width: 2 rule: Form reverse fillColor: Color gray.
			rect := newRect]].
	Display border: rect width: 2 rule: Form reverse fillColor: Color gray.
	" pay the price for reading the sensor directly ; get this party started "
	Smalltalk isMorphic
		ifTrue: [aHand := World activeHand.
			aHand newMouseFocus: nil;
				 showTemporaryCursor: nil;
				 flushEvents].
	Sensor processEvent: Sensor createMouseEvent.
	^ rect! !

!Rectangle methodsFor: 'accessing' stamp: ''!
origin
	"Answer the point at the top left corner of the receiver."

	^origin! !

!Rectangle methodsFor: 'rectangle functions' stamp: 'md 10/4/2005 14:41'!
outsetBy: delta 
	"Answer a Rectangle that is outset from the receiver by delta. delta is a 
	Rectangle, Point, or scalar."

	(delta isRectangle)
		ifTrue: [^Rectangle 
					origin: origin - delta origin 
					corner: corner + delta corner]
		ifFalse: [^Rectangle 
					origin: origin - delta 
					corner: corner + delta]! !

!Rectangle methodsFor: 'rectangle functions' stamp: 'di 4/30/1998 11:22'!
pointNearestTo: aPoint
	"Return the point on my border closest to aPoint"
	| side |
	(self containsPoint: aPoint)
		ifTrue:
			[side := self sideNearestTo: aPoint.
			side == #right ifTrue: [^ self right @ aPoint y].
			side == #left ifTrue: [^ self left @ aPoint y].
			side == #bottom ifTrue: [^ aPoint x @ self bottom].
			side == #top ifTrue: [^ aPoint x @ self top]]
		ifFalse:
			[^ aPoint adhereTo: self]! !

!Rectangle methodsFor: 'printing' stamp: ''!
printOn: aStream 
	"Refer to the comment in Object|printOn:."

	origin printOn: aStream.
	aStream nextPutAll: ' corner: '.
	corner printOn: aStream! !

!Rectangle methodsFor: 'printing' stamp: 'MPW 1/4/1901 08:18'!
propertyListOn: aStream 
	" {x=a; y=b; width=c; height=d} "
	aStream print:'{ x='; write:origin x;
			print:' y='; write:origin y;
			print:' width='; write:self extent x;
			print:' height='; write:self extent y;
			print:'};'.
! !

!Rectangle methodsFor: 'rectangle functions' stamp: ''!
quickMerge: aRectangle 
	"Answer the receiver if it encloses the given rectangle or the merge of the two rectangles if it doesn't. THis method is an optimization to reduce extra rectangle creations."

	| useRcvr rOrigin rCorner minX maxX minY maxY |
	useRcvr := true.
	rOrigin := aRectangle topLeft.
	rCorner := aRectangle bottomRight.
	minX := rOrigin x < origin x ifTrue: [useRcvr := false. rOrigin x] ifFalse: [origin x].
	maxX := rCorner x > corner x ifTrue: [useRcvr := false. rCorner x] ifFalse: [corner x].
	minY := rOrigin y < origin y ifTrue: [useRcvr := false. rOrigin y] ifFalse: [origin y].
	maxY := rCorner y > corner y ifTrue: [useRcvr := false. rCorner y] ifFalse: [corner y].

	useRcvr
		ifTrue: [^ self]
		ifFalse: [^ Rectangle origin: minX@minY corner: maxX@maxY].
! !

!Rectangle methodsFor: 'rectangle functions' stamp: 'di 10/20/97 23:01'!
rectanglesAt: y height: ht
	(y+ht) > self bottom ifTrue: [^ Array new].
	^ Array with: (origin x @ y corner: corner x @ (y+ht))! !

!Rectangle methodsFor: 'accessing' stamp: ''!
right
	"Answer the position of the receiver's right vertical line."

	^corner x! !

!Rectangle methodsFor: 'accessing' stamp: 'ar 10/26/2000 22:17'!
right: aNumber
	^origin corner: aNumber @ corner y! !

!Rectangle methodsFor: 'accessing' stamp: ''!
rightCenter
	"Answer the point at the center of the receiver's right vertical line."

	^self right @ self center y! !

!Rectangle methodsFor: 'transforming' stamp: 'di 6/11/97 15:11'!
rotateBy: direction centerAt: aPoint
	"Return a copy rotated #right, #left, or #pi about aPoint"
	^ (origin rotateBy: direction centerAt: aPoint)
		rect: (corner rotateBy: direction centerAt: aPoint)! !

!Rectangle methodsFor: 'truncation and round off' stamp: ''!
rounded
	"Answer a Rectangle whose origin and corner are rounded."

	^Rectangle origin: origin rounded corner: corner rounded! !

!Rectangle methodsFor: 'transforming' stamp: ''!
scaleBy: scale 
	"Answer a Rectangle scaled by scale, a Point or a scalar."

	^Rectangle origin: origin * scale corner: corner * scale! !

!Rectangle methodsFor: 'transforming' stamp: ''!
scaleFrom: rect1 to: rect2
	"Produce a rectangle stretched according to the stretch from rect1 to rect2"
	^ (origin scaleFrom: rect1 to: rect2)
		corner: (corner scaleFrom: rect1 to: rect2)! !

!Rectangle methodsFor: 'private' stamp: ''!
setOrigin: topLeft corner: bottomRight
	origin := topLeft.
	corner := bottomRight! !

!Rectangle methodsFor: 'rectangle functions' stamp: 'di 10/21/1998 15:09'!
sideNearestTo: aPoint
	| distToLeft distToRight distToTop distToBottom closest side |
	distToLeft := aPoint x - self left.
	distToRight := self right - aPoint x.
	distToTop := aPoint y - self top.
	distToBottom := self bottom - aPoint y.
	closest := distToLeft. side := #left.
	distToRight < closest ifTrue: [closest := distToRight. side := #right].
	distToTop < closest ifTrue: [closest := distToTop. side := #top].
	distToBottom < closest ifTrue: [closest := distToBottom. side := #bottom].
	^ side
"
 | r | r := Rectangle fromUser.
Display border: r width: 1.
[Sensor anyButtonPressed] whileFalse:
	[(r sideNearestTo: Sensor cursorPoint) , '      ' displayAt: 0@0]
"! !

!Rectangle methodsFor: 'transforming' stamp: 'sw 5/21/96'!
squishedWithin: aRectangle
	"Return an adjustment of the receiver that fits within aRectangle by reducing its size, not by changing its origin.  "

	^ origin corner: (corner min: aRectangle bottomRight)

"(50 @ 50 corner: 160 @ 100) squishedWithin:  (20 @ 10 corner: 90 @ 85)"
! !

!Rectangle methodsFor: 'printing' stamp: ''!
storeOn: aStream 
	"printed form is good for storing too"
	self printOn: aStream! !

!Rectangle methodsFor: 'accessing' stamp: ''!
top
	"Answer the position of the receiver's top horizontal line."

	^origin y! !

!Rectangle methodsFor: 'accessing' stamp: 'ar 10/26/2000 22:17'!
top: aNumber
	^origin x @ aNumber corner: corner! !

!Rectangle methodsFor: 'accessing' stamp: ''!
topCenter
	"Answer the point at the center of the receiver's top horizontal line."

	^self center x @ self top! !

!Rectangle methodsFor: 'accessing' stamp: ''!
topLeft
	"Answer the point at the top left corner of the receiver's top horizontal line."

	^origin
! !

!Rectangle methodsFor: 'accessing' stamp: ''!
topRight
	"Answer the point at the top right corner of the receiver's top horizontal 
	line."

	^corner x @ origin y! !

!Rectangle methodsFor: 'transforming' stamp: ''!
translateBy: factor 
	"Answer a Rectangle translated by factor, a Point or a scalar."

	^Rectangle origin: origin + factor corner: corner + factor! !

!Rectangle methodsFor: 'transforming' stamp: 'nk 7/5/2003 08:31'!
translatedAndSquishedToBeWithin: aRectangle
	"Return an adjustment of the receiver that fits within aRectangle by
		- translating it to be within aRectangle if necessary, then
		- reducing its size, if necessary"

	^ (self translatedToBeWithin: aRectangle) squishedWithin: aRectangle! !

!Rectangle methodsFor: 'rectangle functions' stamp: ''!
translatedToBeWithin: aRectangle
	"Answer a copy of the receiver that does not extend beyond aRectangle.  7/8/96 sw"

	^ self translateBy: (self amountToTranslateWithin: aRectangle)! !

!Rectangle methodsFor: 'truncation and round off' stamp: ''!
truncateTo: grid
	"Answer a Rectangle whose origin and corner are truncated to grid x and grid y."

	^Rectangle origin: (origin truncateTo: grid)
				corner: (corner truncateTo: grid)! !

!Rectangle methodsFor: 'truncation and round off' stamp: 'jm 5/29/1998 15:53'!
truncated
	"Answer a Rectangle whose origin and corner have any fractional parts removed. Answer the receiver if its coordinates are already integral."

	(origin x isInteger and:
	[origin y isInteger and:
	[corner x isInteger and:
	[corner y isInteger]]])
		ifTrue: [^ self].

	^ Rectangle origin: origin truncated corner: corner truncated
! !

!Rectangle methodsFor: 'accessing' stamp: ''!
width
	"Answer the width of the receiver."

	^corner x - origin x! !

!Rectangle methodsFor: 'rectangle functions' stamp: ''!
withBottom: y 
	"Return a copy of me with a different bottom y"
	^ origin x @ origin y corner: corner x @ y! !

!Rectangle methodsFor: 'rectangle functions' stamp: ''!
withHeight: height 
	"Return a copy of me with a different height"
	^ origin corner: corner x @ (origin y + height)! !

!Rectangle methodsFor: 'rectangle functions' stamp: ''!
withLeft: x 
	"Return a copy of me with a different left x"
	^ x @ origin y corner: corner x @ corner y! !

!Rectangle methodsFor: 'rectangle functions' stamp: ''!
withRight: x 
	"Return a copy of me with a different right x"
	^ origin x @ origin y corner: x @ corner y! !

!Rectangle methodsFor: 'rectangle functions' stamp: ''!
withSide: side setTo: value  "return a copy with side set to value"
	^ self perform: (#(withLeft: withRight: withTop: withBottom: )
							at: (#(left right top bottom) indexOf: side))
		with: value! !

!Rectangle methodsFor: 'rectangle functions' stamp: 'di 9/8/1999 21:25'!
withSideOrCorner: side setToPoint: newPoint
	"Return a copy with side set to newPoint"

	^ self withSideOrCorner: side setToPoint: newPoint minExtent: 0@0! !

!Rectangle methodsFor: 'rectangle functions' stamp: 'bf 9/10/1999 16:16'!
withSideOrCorner: side setToPoint: newPoint minExtent: minExtent
	"Return a copy with side set to newPoint"
	^self withSideOrCorner: side setToPoint: newPoint minExtent: minExtent
		limit: ((#(left top) includes: side) ifTrue: [SmallInteger minVal] ifFalse: [SmallInteger maxVal])! !

!Rectangle methodsFor: 'rectangle functions' stamp: 'bf 9/10/1999 16:07'!
withSideOrCorner: side setToPoint: newPoint minExtent: minExtent limit: limit
	"Return a copy with side set to newPoint"
	side = #top ifTrue: [^ self withTop: (newPoint y min: corner y - minExtent y max: limit + minExtent y)].
	side = #bottom ifTrue: [^ self withBottom: (newPoint y min: limit - minExtent y max: origin y + minExtent y)].
	side = #left ifTrue: [^ self withLeft: (newPoint x min: corner x - minExtent x max: limit + minExtent x)].
	side = #right ifTrue: [^ self withRight: (newPoint x min: limit - minExtent x max: origin x + minExtent x)].
	side = #topLeft ifTrue: [^ (newPoint min: corner - minExtent) corner: self bottomRight].
	side = #bottomRight ifTrue: [^ self topLeft corner: (newPoint max: origin + minExtent)].
	side = #bottomLeft ifTrue: [^ self topRight rect: ((newPoint x min: corner x - minExtent x) @ (newPoint y max: origin y + minExtent y))].
	side = #topRight ifTrue: [^ self bottomLeft rect: ((newPoint x max: origin x + minExtent x) @ (newPoint y min: corner y - minExtent y))].! !

!Rectangle methodsFor: 'rectangle functions' stamp: ''!
withTop: y 
	"Return a copy of me with a different top y"
	^ origin x @ y corner: corner x @ corner y! !

!Rectangle methodsFor: 'rectangle functions' stamp: ''!
withWidth: width 
	"Return a copy of me with a different width"
	^ origin corner: (origin x + width) @ corner y! !

Object subclass: #TextPrinter
	instanceVariableNames: 'form para paperSize landscape resolution depth offset columns docTitle noHeader noFooter'
	classVariableNames: 'DefaultPaperSize DefaultTextPrinter'
	poolDictionaries: ''
	category: 'Graphics-Text'!

!TextPrinter class methodsFor: 'accessing' stamp: 'ar 4/30/98 18:31'!
defaultPaperSize
	^DefaultPaperSize! !

!TextPrinter class methodsFor: 'accessing' stamp: 'ar 4/30/98 18:31'!
defaultPaperSize: aPoint
	DefaultPaperSize := aPoint! !

!TextPrinter class methodsFor: 'accessing' stamp: 'dew 3/7/2000 20:39'!
defaultTextPrinter
	"This is the global default TextPrinter instance."
	DefaultTextPrinter isNil ifTrue: [DefaultTextPrinter := self new].
	^DefaultTextPrinter! !

!TextPrinter class methodsFor: 'class initialization' stamp: 'ar 4/30/98 18:30'!
initialize
	"TextPrinter initialize"
	self defaultPaperSize: self paperSizeA4.! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:30'!
mm2in: aPoint
	"Convert aPoint from millimeters to inches"
	^aPoint / 25.4! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:27'!
paperSize10x14
	^10.0@14.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:27'!
paperSize11x17
	^11.0@17.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:27'!
paperSizeA3
	^self mm2in: 297@420! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:27'!
paperSizeA4
	^self mm2in: 210@297! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:27'!
paperSizeA5
	^self mm2in: 148@210! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:27'!
paperSizeB4
	^self mm2in: 250@354! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:27'!
paperSizeB5
	^self mm2in: 182@257! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeCSheet
	^17.0@22.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeDSheet
	^22.0@34.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeESheet
	^34.0@44.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelope10
	^4.125@9.5
! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelope11
	^4.5@10.375! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelope12
	^4.75@11! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelope14
	^5.0@11.5! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelope9
	^3.875@8.875! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelopeB4
	^self mm2in: 250@353! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelopeB5
	^self mm2in: 176@250! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelopeB6
	^self mm2in: 176@125! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:28'!
paperSizeEnvelopeC3
	^self mm2in: 324@458! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeEnvelopeC4
	^self mm2in: 229@324! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeEnvelopeC5
	^self mm2in: 162@229! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeEnvelopeC6
	^self mm2in: 114@162! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeEnvelopeC65
	^self mm2in: 114@229! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeFanfoldGerman
	"German standard fanfold"
	^8.5@12.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeFanfoldLegalGerman
	"German legal fanfold"
	^8.5@13.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeFanfoldUS
	"US standard fanfold"
	^14.875@11.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeFolio
	^8.5@13.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeLegal
	^8.5@14.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeLetter
	^8.5@11.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:29'!
paperSizeNote
	^8.5@11.0! !

!TextPrinter class methodsFor: 'paper sizes' stamp: 'ar 4/30/98 18:30'!
paperSizeTabloid
	^11.0@17.0! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:42'!
bestColor
	"Set the reproduction quality to true color"
	depth := 32.! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:42'!
blackAndWhite
	"Set the reproduction quality to black and white"
	depth := 1.! !

!TextPrinter methodsFor: 'formatting' stamp: 'ar 4/30/98 19:25'!
columnRect: n
	"Return a rectangle describing the n-th column"
	| area left right |
	area := self textArea.
	left := area left + ((n-1) * self columnWidth).
	left := left + ((n-1) * self columnSkip).
	right := left + self columnWidth.
	^(self in2pix: left @ area top) corner: 
		(self in2pix: right @ area bottom)! !

!TextPrinter methodsFor: 'formatting' stamp: 'ar 4/30/98 19:20'!
columnSkip
	"Return the separating space between two columns in inches"
	^0.2! !

!TextPrinter methodsFor: 'formatting' stamp: 'ar 4/30/98 19:21'!
columnWidth
	^(self textWidth - ((self columns-1) * self columnSkip)) / self columns! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:50'!
columns
	^columns! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:50'!
columns: aNumber
	columns := aNumber asInteger max: 1.! !

!TextPrinter methodsFor: 'initialize' stamp: 'ar 4/30/98 19:26'!
defaultPaperSize
	"Return the default paper size (inches) for printing"
	^self class defaultPaperSize! !

!TextPrinter methodsFor: 'initialize' stamp: 'nk 4/2/2004 11:32'!
defaultResolution
	"Return the default resolution (DPI) for printing"
	^TextStyle pixelsPerInch asPoint! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 20:14'!
documentTitle
	^docTitle! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 20:14'!
documentTitle: aString
	docTitle := aString! !

!TextPrinter methodsFor: 'printing' stamp: 'ar 4/30/98 20:41'!
flushPage
	"The current page has been set up. Send it to the printer."
	form primPrintHScale: self resolution x vScale: self resolution y landscape: self landscape.
	"Uncomment the following for testing"
	"form displayOn: Display. (Delay forSeconds: 5) wait."
! !

!TextPrinter methodsFor: 'footer' stamp: 'ar 4/30/98 19:23'!
footerHeight
	"Return the (additional) height of the footer in inches."
	self noFooter ifTrue:[^0.0].
	^(self pix2in: 0@TextStyle default lineGrid) y * 2! !

!TextPrinter methodsFor: 'footer' stamp: 'ar 4/30/98 20:11'!
footerParagraph
	"Return a paragraph for the footer"
	| fPara rect |
	fPara := Paragraph new.
	fPara destinationForm: form.
	rect := (self in2pix: self textArea bottomLeft) corner: 
				(self in2pix: self textArea bottomRight + (0.0@self footerHeight)).
	fPara clippingRectangle: rect.
	fPara compositionRectangle: rect.
	^fPara! !

!TextPrinter methodsFor: 'formatting' stamp: 'ar 4/30/98 19:29'!
formatColumn: columnNum startingWith: anIndex
	"Format a new column starting at the given string index. Return the string index indicating the start of the next column or nil if no more columns need printing."
	| colRect blk |
	colRect := self columnRect: columnNum.
	anIndex > 1 ifTrue:[para text: (para text copyFrom: anIndex to: para text size)].
	para compositionRectangle: colRect.
	para clippingRectangle: colRect.
	para composeAll.
	para displayOn: form.
	para visibleRectangle corner y <= colRect extent y ifTrue:[^nil].
	"More columns -- find the character block of the last line and adjust clip rect"
	blk := para characterBlockAtPoint: para visibleRectangle bottomLeft.
	para clearVisibleRectangle. "Make sure that the background is clean"
	para clippingRectangle: (colRect topLeft corner: colRect right@blk top).
	para displayOn: form.
	^blk stringIndex.! !

!TextPrinter methodsFor: 'formatting' stamp: 'ar 4/30/98 19:29'!
formatPage: pageNum startingWith: anIndex
	"Format a new page starting at the given string index. Return the string index indicating the start of the next page or nil if no more pages need printing."
	| nextIndex |
	nextIndex := anIndex.
	1 to: self columns do:[:i|
		nextIndex := self formatColumn: i startingWith: nextIndex.
		nextIndex isNil ifTrue:[^nil].
	].
	^nextIndex! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:42'!
goodColor
	"Set the reproduction quality to 8 bit color depth"
	depth := 8.! !

!TextPrinter methodsFor: 'header' stamp: 'ar 4/30/98 19:23'!
headerHeight
	"Return the (additional) height of the header in inches."
	self noHeader ifTrue:[^0.0].
	^(self pix2in: 0@TextStyle default lineGrid) y * 2! !

!TextPrinter methodsFor: 'header' stamp: 'ar 4/30/98 20:11'!
headerParagraph
	"Return a paragraph for the footer"
	| hPara rect |
	hPara := Paragraph new.
	hPara destinationForm: form.
	rect := (self in2pix: self textArea topLeft - (0.0@self headerHeight)) corner: 
				(self in2pix: self textArea topRight).
	hPara clippingRectangle: rect.
	hPara compositionRectangle: rect.
	^hPara! !

!TextPrinter methodsFor: 'other' stamp: 'ar 4/30/98 18:39'!
in2mm: aPoint
	"Convert aPoint from millimeters to inches"
	^aPoint * 25.4! !

!TextPrinter methodsFor: 'other' stamp: 'ar 4/30/98 18:38'!
in2pix: aPoint
	"Convert aPoint from inches to actual pixels"
	^(aPoint * self resolution) rounded! !

!TextPrinter methodsFor: 'initialize' stamp: 'ar 4/30/98 19:39'!
initialize
	self paperSize: self defaultPaperSize.
	self resolution: self defaultResolution.
	self blackAndWhite.
	self landscape: false.
	self offsetRect: (1.0@1.0 corner: 1.0@1.0).
	self columns: 1.
	self noHeader: false.
	self noFooter: false.
	self documentTitle: 'Squeak Document (from ', Date today printString,')'.! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:42'!
landscape
	^landscape! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:42'!
landscape: aBoolean
	landscape := aBoolean! !

!TextPrinter methodsFor: 'other' stamp: 'ar 4/30/98 18:38'!
mm2in: aPoint
	"Convert aPoint from millimeters to inches"
	^aPoint / 25.4! !

!TextPrinter methodsFor: 'other' stamp: 'ar 4/30/98 18:40'!
mm2pix: aPoint
	"Convert aPoint from millimeters to actual pixels"
	^self in2pix: (self mm2in: aPoint)! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 19:23'!
noFooter
	^noFooter! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 19:22'!
noFooter: aBoolean
	"Turn off footer printing"
	noFooter := aBoolean.! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 19:22'!
noHeader
	^noHeader! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 19:22'!
noHeader: aBoolean
	"Turn off header printing"
	noHeader := aBoolean.! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 19:27'!
offsetRect
	^offset! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 19:27'!
offsetRect: aRectangle
	"Set the offset rectangle"
	offset := aRectangle! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:42'!
paperSize
	^paperSize! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:42'!
paperSize: aPoint
	paperSize := aPoint! !

!TextPrinter methodsFor: 'other' stamp: 'ar 4/30/98 18:39'!
pix2in: aPoint
	"Convert aPoint from a pixel value to inches"
	^aPoint / self resolution! !

!TextPrinter methodsFor: 'other' stamp: 'ar 4/30/98 18:40'!
pix2mm: aPoint
	"Convert aPoint from a pixel value to millimeters"
	^self in2mm: (self pix2in: aPoint)! !

!TextPrinter methodsFor: 'private' stamp: 'ar 4/30/98 19:40'!
pixelSize
	"Return the size of the page in pixels"
	^self in2pix: (self realPaperSize)! !

!TextPrinter methodsFor: 'footer' stamp: 'ar 4/30/98 19:24'!
printFooter: pageNumber
	"Print the footer for the given page number"
	| fPara |
	self noFooter ifTrue:[^self].
	fPara := self footerParagraph.
	fPara centered.
	fPara text: ('Page ', pageNumber printString) asText.
	fPara displayOn: form.! !

!TextPrinter methodsFor: 'header' stamp: 'ar 4/30/98 19:23'!
printHeader: pageNumber
	"Print the header for the given page number"
	| fPara |
	self noHeader ifTrue:[^self].
	fPara := self headerParagraph.
	fPara centered.
	fPara text: self documentTitle asText.
	fPara displayOn: form.! !

!TextPrinter methodsFor: 'printing' stamp: 'ar 4/30/98 19:19'!
printParagraph
	| pageNum nextIndex |
	para destinationForm: form.
	pageNum := 1.
	nextIndex := 1.
	[form fillColor: Color white.
	self printHeader: pageNum.
	self printFooter: pageNum.
	nextIndex := self formatPage: pageNum startingWith: nextIndex.
	self flushPage.
	nextIndex isNil] whileFalse:[pageNum := pageNum + 1].! !

!TextPrinter methodsFor: 'printing' stamp: 'ar 4/30/98 18:55'!
printText: aText
	"Print aText"
	form isNil ifTrue:[
		form := Form extent: self pixelSize depth: depth.
	].
	para := Paragraph withText: aText asText.
	Cursor wait showWhile:[
		self printParagraph.
	].! !

!TextPrinter methodsFor: 'private' stamp: 'ar 4/30/98 19:39'!
realPaperSize
	^self landscape
		ifTrue:[self paperSize y @ self paperSize x]
		ifFalse:[self paperSize]! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:43'!
resolution
	^resolution! !

!TextPrinter methodsFor: 'accessing' stamp: 'ar 4/30/98 18:43'!
resolution: aPoint
	resolution := aPoint! !

!TextPrinter methodsFor: 'formatting' stamp: 'ar 4/30/98 19:58'!
textArea
	^(self offsetRect origin + (0.0@self headerHeight)) corner:
		(self realPaperSize - self offsetRect corner - (0.0@self footerHeight))! !

!TextPrinter methodsFor: 'formatting' stamp: 'ar 4/30/98 19:23'!
textWidth
	^self textArea extent x! !

ReadStream subclass: #JPEGReadStream
	instanceVariableNames: 'bitBuffer bitsInBuffer'
	classVariableNames: 'MaxBits'
	poolDictionaries: ''
	category: 'Graphics-Files'!

!JPEGReadStream commentStamp: '<historical>' prior: 0!
Encapsulates huffman encoded access to JPEG data.

The following layout is fixed for the JPEG primitives to work:

	collection	<ByteArray | String>
	position		<SmallInteger>
	readLimit	<SmallInteger>
	bitBuffer	<SmallInteger>
	bitsInBuffer	<SmallInteger>!

!JPEGReadStream class methodsFor: 'class initialization' stamp: 'ar 3/4/2001 18:32'!
initialize
	"JPEGReadStream initialize"
	MaxBits := 16.! !

!JPEGReadStream methodsFor: 'huffman trees' stamp: 'ar 3/4/2001 21:06'!
buildLookupTable: values counts: counts
	| min max |
	min := max := nil.
	1 to: counts size do:[:i|
		(counts at: i) = 0 ifFalse:[
			min ifNil:[min := i-1].
			max := i]].
	^self
		createHuffmanTables: values 
		counts: {0},counts 
		from: min+1 
		to: max.! !

!JPEGReadStream methodsFor: 'huffman trees' stamp: 'ar 3/4/2001 18:46'!
createHuffmanTables: values counts: counts from: minBits to: maxBits
	"Create the actual tables"
	| table tableStart tableSize tableEnd 
	valueIndex tableStack numValues deltaBits maxEntries
	lastTable lastTableStart tableIndex lastTableIndex |

	table := WordArray new: ((4 bitShift: minBits) max: 16).

	"Create the first entry - this is a dummy.
	It gives us information about how many bits to fetch initially."
	table at: 1 put: (minBits bitShift: 24) + 2. "First actual table starts at index 2"

	"Create the first table from scratch."
	tableStart := 2. "See above"
	tableSize := 1 bitShift: minBits.
	tableEnd := tableStart + tableSize.
	"Store the terminal symbols"
	valueIndex := (counts at: minBits+1).
	tableIndex := 0.
	1 to: valueIndex do:[:i|
		table at: tableStart + tableIndex put: (values at: i).
		tableIndex := tableIndex + 1].
	"Fill up remaining entries with invalid entries"
	tableStack := OrderedCollection new: 10. "Should be more than enough"
	tableStack addLast: 
		(Array 
			with: minBits	"Number of bits (e.g., depth) for this table"
			with: tableStart	"Start of table"
			with: tableIndex "Next index in table"
			with: minBits	"Number of delta bits encoded in table"
			with: tableSize - valueIndex "Entries remaining in table").
	"Go to next value index"
	valueIndex := valueIndex + 1.
	"Walk over remaining bit lengths and create new subtables"
	minBits+1 to: maxBits do:[:bits|
		numValues := counts at: bits+1.
		[numValues > 0] whileTrue:["Create a new subtable"
			lastTable := tableStack last.
			lastTableStart := lastTable at: 2.
			lastTableIndex := lastTable at: 3.
			deltaBits := bits - (lastTable at: 1).
			"Make up a table of deltaBits size"
			tableSize := 1 bitShift: deltaBits.
			tableStart := tableEnd.
			tableEnd := tableEnd + tableSize.
			[tableEnd > table size ]
				whileTrue:[table := self growHuffmanTable: table].
			"Connect to last table"
			self assert:[(table at: lastTableStart + lastTableIndex) = 0]."Entry must be unused"
			table at: lastTableStart + lastTableIndex put: (deltaBits bitShift: 24) + tableStart.
			lastTable at: 3 put: lastTableIndex+1.
			lastTable at: 5 put: (lastTable at: 5) - 1.
			self assert:[(lastTable at: 5) >= 0]. "Don't exceed tableSize"
			"Store terminal values"
			maxEntries := numValues min: tableSize.
			tableIndex := 0.
			1 to: maxEntries do:[:i|
				table at: tableStart + tableIndex put: (values at: valueIndex).
				valueIndex := valueIndex + 1.
				numValues := numValues - 1.
				tableIndex := tableIndex+1].
			"Check if we have filled up the current table completely"
			maxEntries = tableSize ifTrue:[
				"Table has been filled. Back up to the last table with space left."
				[tableStack isEmpty not and:[(tableStack last at: 5) = 0]]
						whileTrue:[tableStack removeLast].
			] ifFalse:[
				"Table not yet filled. Put it back on the stack."
				tableStack addLast:
					(Array
						with: bits		"Nr. of bits in this table"
						with: tableStart	"Start of table"
						with: tableIndex "Index in table"
						with: deltaBits	"delta bits of table"
						with: tableSize - maxEntries "Unused entries in table").
			].
		].
	].
	 ^table copyFrom: 1 to: tableEnd-1! !

!JPEGReadStream methodsFor: 'huffman trees' stamp: 'ar 3/4/2001 18:44'!
decodeValueFrom: table
	"Decode the next value in the receiver using the given huffman table."
	| bits bitsNeeded tableIndex value |
	bitsNeeded := (table at: 1) bitShift: -24.	"Initial bits needed"
	tableIndex := 2.							"First real table"
	[bits := self getBits: bitsNeeded.			"Get bits"
	value := table at: (tableIndex + bits).		"Lookup entry in table"
	(value bitAnd: 16r3F000000) = 0] 			"Check if it is a non-leaf node"
		whileFalse:["Fetch sub table"
			tableIndex := value bitAnd: 16rFFFF.	"Table offset in low 16 bit"
			bitsNeeded := (value bitShift: -24) bitAnd: 255. "Additional bits in high 8 bit"
			bitsNeeded > MaxBits ifTrue:[^self error:'Invalid huffman table entry']].
	^value! !

!JPEGReadStream methodsFor: 'accessing' stamp: 'ar 3/4/2001 20:42'!
fillBuffer

	| byte |
	[bitsInBuffer <= 16]
		whileTrue:[
			byte := self next.
			(byte = 16rFF and: [(self peekFor: 16r00) not])
					ifTrue:
						[self position: self position - 1.
						^0].
			bitBuffer := (bitBuffer bitShift: 8) bitOr: byte.
			bitsInBuffer := bitsInBuffer + 8].
	^ bitsInBuffer! !

!JPEGReadStream methodsFor: 'accessing' stamp: 'ar 3/4/2001 18:43'!
getBits: requestedBits
	| value |
	requestedBits > bitsInBuffer ifTrue:[
		self fillBuffer.
		requestedBits > bitsInBuffer ifTrue:[
			self error: 'not enough bits available to decode']].
	value := bitBuffer bitShift: (requestedBits - bitsInBuffer).
	bitBuffer := bitBuffer bitAnd: (1 bitShift: (bitsInBuffer - requestedBits)) -1.
	bitsInBuffer := bitsInBuffer - requestedBits.
	^ value! !

!JPEGReadStream methodsFor: 'huffman trees' stamp: 'ar 3/4/2001 18:21'!
growHuffmanTable: table
	| newTable |
	newTable := table species new: table size * 2.
	newTable replaceFrom: 1 to: table size with: table startingAt: 1.
	^newTable! !

!JPEGReadStream methodsFor: 'accessing' stamp: 'ar 3/6/2001 12:34'!
nextByte
	^self next asInteger! !

!JPEGReadStream methodsFor: 'accessing' stamp: 'ar 3/6/2001 12:35'!
nextBytes: n
	^(self next: n) asByteArray! !

!JPEGReadStream methodsFor: 'accessing' stamp: 'ar 3/4/2001 17:40'!
reset
	super reset.
	self resetBitBuffer! !

!JPEGReadStream methodsFor: 'accessing' stamp: 'ar 3/4/2001 18:44'!
resetBitBuffer
	bitBuffer := 0.
	bitsInBuffer := 0.
! !

ArrayedCollection variableWordSubclass: #Bitmap
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Primitives'!

!Bitmap commentStamp: '<historical>' prior: 0!
My instances provide contiguous storage of bits, primarily to hold the graphical data of Forms. Forms and their subclasses provide the additional structural information as to how the bits should be interpreted in two dimensions.!

!Bitmap class methodsFor: 'instance creation' stamp: 'di 2/9/98 16:02'!
decodeIntFrom: s
	"Decode an integer in stream s as follows...
		0-223	0-223
		224-254	(0-30)*256 + next byte (0-7935)
		255		next 4 bytes	"		
	| int |
	int := s next.
	int <= 223 ifTrue: [^ int].
	int <= 254 ifTrue: [^ (int-224)*256 + s next].
	int := s next.
	1 to: 3 do: [:j | int := (int bitShift: 8) + s next].
	^ int! !

!Bitmap class methodsFor: 'instance creation' stamp: 'di 2/12/98 14:34'!
decompressFromByteArray: byteArray
	| s bitmap size |
	s := ReadStream on: byteArray.
	size := self decodeIntFrom: s.
	bitmap := self new: size.
	bitmap decompress: bitmap fromByteArray: byteArray at: s position+1.
	^ bitmap! !

!Bitmap class methodsFor: 'instance creation' stamp: 'ar 12/23/1999 14:35'!
newFromStream: s
	| len |
	s next = 16r80 ifTrue:
		["New compressed format"
		len := self decodeIntFrom: s.
		^ Bitmap decompressFromByteArray: (s nextInto: (ByteArray new: len))].
	s skip: -1.
	len := s nextInt32.
	len <= 0
		ifTrue: ["Old compressed format"
				^ (self new: len negated) readCompressedFrom: s]
		ifFalse: ["Old raw data format"
				^ s nextWordsInto: (self new: len)]! !

!Bitmap class methodsFor: 'utilities' stamp: 'sd 6/28/2003 09:33'!
swapBytesIn: aNonPointerThing from: start to: stop
	"Perform a bigEndian/littleEndian byte reversal of my words.
	We only intend this for non-pointer arrays.  Do nothing if I contain pointers."
	| hack blt |
	"The implementation is a hack, but fast for large ranges"
	hack := Form new hackBits: aNonPointerThing.
	blt := (BitBlt toForm: hack) sourceForm: hack.
	blt combinationRule: Form reverse.  "XOR"
	blt sourceY: start-1; destY: start-1; height: stop-start+1; width: 1.
	blt sourceX: 0; destX: 3; copyBits.  "Exchange bytes 0 and 3"
	blt sourceX: 3; destX: 0; copyBits.
	blt sourceX: 0; destX: 3; copyBits.
	blt sourceX: 1; destX: 2; copyBits.  "Exchange bytes 1 and 2"
	blt sourceX: 2; destX: 1; copyBits.
	blt sourceX: 1; destX: 2; copyBits.
! !

!Bitmap methodsFor: 'as yet unclassified' stamp: 'yo 2/18/2004 18:28'!
asByteArray
	"Faster way to make a byte array from me.
	copyFromByteArray: makes equal Bitmap."
	| f bytes hack |
	f := Form extent: 4@self size depth: 8 bits: self.
	bytes := ByteArray new: self size * 4.
	hack := Form new hackBits: bytes.
	SmalltalkImage current isLittleEndian ifTrue:[hack swapEndianness].
	hack copyBits: f boundingBox
		from: f
		at: (0@0)
		clippingBox: hack boundingBox
		rule: Form over
		fillColor: nil
		map: nil.

	"f displayOn: hack."
	^ bytes.
! !

!Bitmap methodsFor: 'accessing' stamp: 'ar 3/3/2001 16:11'!
atAllPut: value
	"Fill the receiver, an indexable bytes or words object, with the given positive integer. The range of possible fill values is [0..255] for byte arrays and [0..(2^32 - 1)] for word arrays."
	<primitive: 145>
	super atAllPut: value.! !

!Bitmap methodsFor: 'accessing' stamp: ''!
bitPatternForDepth: depth
	"The raw call on BitBlt needs a Bitmap to represent this color.  I already am Bitmap like.  I am already adjusted for a specific depth.  Interpret me as an array of (32/depth) Color pixelValues.  BitBlt aligns the first element of this array with the top scanline of the destinationForm, the second with the second, and so on, cycling through the color array as necessary. 6/18/96 tk"

	^ self! !

!Bitmap methodsFor: 'accessing' stamp: ''!
byteAt: byteAddress
	"Extract a byte from a Bitmap.  Note that this is a byte address and it is one-order.  For repeated use, create an instance of BitBlt and use pixelAt:.  See Form pixelAt:  7/1/96 tk"
	| lowBits |
	lowBits := byteAddress - 1 bitAnd: 3.
	^((self at: byteAddress - 1 - lowBits // 4 + 1)
		bitShift: (lowBits - 3) * 8)
		bitAnd: 16rFF! !

!Bitmap methodsFor: 'accessing' stamp: 'ar 9/21/2001 23:06'!
byteAt: byteAddress put: byte
	"Insert a byte into a Bitmap.  Note that this is a byte address and it is one-order.  For repeated use, create an instance of BitBlt and use pixelAt:put:.  See Form pixelAt:put:  7/1/96 tk"
	| longWord shift lowBits longAddr |
	(byte < 0 or:[byte > 255]) ifTrue:[^self errorImproperStore].
	lowBits := byteAddress - 1 bitAnd: 3.
	longWord := self at: (longAddr := (byteAddress - 1 - lowBits) // 4 + 1).
	shift := (3 - lowBits) * 8.
	longWord := longWord - (longWord bitAnd: (16rFF bitShift: shift)) 
		+ (byte bitShift: shift).
	self at: longAddr put: longWord.
	^ byte! !

!Bitmap methodsFor: 'accessing' stamp: 'ar 3/3/2001 16:18'!
byteSize
	^self size * 4! !

!Bitmap methodsFor: 'filing' stamp: 'ar 2/3/2001 16:11'!
compress: bm toByteArray: ba
	"Store a run-coded compression of the receiver into the byteArray ba,
	and return the last index stored into. ba is assumed to be large enough.
	The encoding is as follows...
		S {N D}*.
		S is the size of the original bitmap, followed by run-coded pairs.
		N is a run-length * 4 + data code.
		D, the data, depends on the data code...
			0	skip N words, D is absent
			1	N words with all 4 bytes = D (1 byte)
			2	N words all = D (4 bytes)
			3	N words follow in D (4N bytes)
		S and N are encoded as follows...
			0-223	0-223
			224-254	(0-30)*256 + next byte (0-7935)
			255		next 4 bytes"		
	| size k word j lowByte eqBytes i |
	<primitive: 'primitiveCompressToByteArray' module: 'MiscPrimitivePlugin'>
	self var: #bm declareC: 'int *bm'.
	self var: #ba declareC: 'unsigned char *ba'.
	size := bm size.
	i := self encodeInt: size in: ba at: 1.
	k := 1.
	[k <= size] whileTrue:
		[word := bm at: k.
		lowByte := word bitAnd: 16rFF.
		eqBytes := ((word >> 8) bitAnd: 16rFF) = lowByte
				and: [((word >> 16) bitAnd: 16rFF) = lowByte
				and: [((word >> 24) bitAnd: 16rFF) = lowByte]].
		j := k.
		[j < size and: [word = (bm at: j+1)]]  "scan for = words..."
			whileTrue: [j := j+1].
		j > k ifTrue:
			["We have two or more = words, ending at j"
			eqBytes
				ifTrue: ["Actually words of = bytes"
						i := self encodeInt: j-k+1*4+1 in: ba at: i.
						ba at: i put: lowByte.  i := i+1]
				ifFalse: [i := self encodeInt: j-k+1*4+2 in: ba at: i.
						i := self encodeBytesOf: word in: ba at: i].
			k := j+1]
			ifFalse:
			["Check for word of 4 = bytes"
			eqBytes ifTrue:
				["Note 1 word of 4 = bytes"
				i := self encodeInt: 1*4+1 in: ba at: i.
				ba at: i put: lowByte.  i := i+1.
				k := k + 1]
				ifFalse:
				["Finally, check for junk"
				[j < size and: [(bm at: j) ~= (bm at: j+1)]]  "scan for ~= words..."
					whileTrue: [j := j+1].
				j = size ifTrue: [j := j + 1].
				"We have one or more unmatching words, ending at j-1"
				i := self encodeInt: j-k*4+3 in: ba at: i.
				k to: j-1 do:
					[:m | i := self encodeBytesOf: (bm at: m) in: ba at: i].
				k := j]]].
	^ i - 1  "number of bytes actually stored"
"
Space check:
 | n rawBytes myBytes b |
n := rawBytes := myBytes := 0.
Form allInstancesDo:
	[:f | f unhibernate.
	b := f bits.
	n := n + 1.
	rawBytes := rawBytes + (b size*4).
	myBytes := myBytes + (b compressToByteArray size).
	f hibernate].
Array with: n with: rawBytes with: myBytes
ColorForms: (116 230324 160318 )
Forms: (113 1887808 1325055 )

Integerity check:
Form allInstances do:
	[:f | f unhibernate.
	f bits = (Bitmap decompressFromByteArray: f bits compressToByteArray)
		ifFalse: [self halt].
	f hibernate]

Speed test:
MessageTally spyOn: [Form allInstances do:
	[:f | Bitmap decompressFromByteArray: f bits compressToByteArray]]
"! !

!Bitmap methodsFor: 'filing' stamp: 'RAA 7/28/2000 08:40'!
compressGZip
	| ba hackwa hackba blt rowsAtATime sourceOrigin rowsRemaining bufferStream gZipStream |

"just hacking around to see if further compression would help Nebraska"

	bufferStream := RWBinaryOrTextStream on: (ByteArray new: 5000).
	gZipStream := GZipWriteStream on: bufferStream.

	ba := nil.
	rowsAtATime := 20000.		"or 80000 bytes"
	hackwa := Form new hackBits: self.
	sourceOrigin := 0@0.
	[(rowsRemaining := hackwa height - sourceOrigin y) > 0] whileTrue: [
		rowsAtATime := rowsAtATime min: rowsRemaining.
		(ba isNil or: [ba size ~= (rowsAtATime * 4)]) ifTrue: [
			ba := ByteArray new: rowsAtATime * 4.
			hackba := Form new hackBits: ba.
			blt := (BitBlt toForm: hackba) sourceForm: hackwa.
		].
		blt 
			combinationRule: Form over;
			sourceOrigin: sourceOrigin;
			destX: 0 destY: 0 width: 4 height: rowsAtATime;
			copyBits.
		"bufferStream nextPutAll: ba."
		sourceOrigin := sourceOrigin x @ (sourceOrigin y + rowsAtATime).
	].
	gZipStream close.
	^bufferStream contents
! !

!Bitmap methodsFor: 'filing' stamp: 'di 8/5/1998 11:31'!
compressToByteArray
	"Return a run-coded compression of this bitmap into a byteArray"		
	| byteArray lastByte |
	"Without skip codes, it is unlikely that the compressed bitmap will be any larger than was the original.  The run-code cases are...
	N >= 1 words of equal bytes:  4N bytes -> 2 bytes (at worst 4 -> 2)
	N > 1 equal words:  4N bytes -> 5 bytes (at worst 8 -> 5)
	N > 1 unequal words:  4N bytes -> 4N + M, where M is the number of bytes required to encode the run length.

The worst that can happen is that the method begins with unequal words, and than has interspersed occurrences of a word with equal bytes.  Thus we require a run-length at the beginning, and after every interspersed word of equal bytes.  However, each of these saves 2 bytes, so it must be followed by a run of 1984 (7936//4) or more (for which M jumps from 2 to 5) to add any extra overhead.  Therefore the worst case is a series of runs of 1984 or more, with single interspersed words of equal bytes.  At each break we save 2 bytes, but add 5.  Thus the overhead would be no more than 5 (encoded size) + 2 (first run len) + (S//1984*3)."
	
"NOTE: This code is copied in Form hibernate for reasons given there."
	byteArray := ByteArray new: (self size*4) + 7 + (self size//1984*3).
	lastByte := self compress: self toByteArray: byteArray.
	^ byteArray copyFrom: 1 to: lastByte! !

!Bitmap methodsFor: 'as yet unclassified' stamp: 'RAA 7/28/2000 21:51'!
copy

	^self clone! !

!Bitmap methodsFor: 'accessing' stamp: 'nk 7/30/2004 17:53'!
copyFromByteArray: byteArray 
	"This method should work with either byte orderings"

	| myHack byteHack |
	myHack := Form new hackBits: self.
	byteHack := Form new hackBits: byteArray.
	SmalltalkImage current  isLittleEndian ifTrue: [byteHack swapEndianness].
	byteHack displayOn: myHack! !

!Bitmap methodsFor: 'filing' stamp: 'ar 2/3/2001 16:11'!
decompress: bm fromByteArray: ba at: index
	"Decompress the body of a byteArray encoded by compressToByteArray (qv)...
	The format is simply a sequence of run-coded pairs, {N D}*.
		N is a run-length * 4 + data code.
		D, the data, depends on the data code...
			0	skip N words, D is absent
				(could be used to skip from one raster line to the next)
			1	N words with all 4 bytes = D (1 byte)
			2	N words all = D (4 bytes)
			3	N words follow in D (4N bytes)
		S and N are encoded as follows (see decodeIntFrom:)...
			0-223	0-223
			224-254	(0-30)*256 + next byte (0-7935)
			255		next 4 bytes"	
	"NOTE:  If fed with garbage, this routine could read past the end of ba, but it should fail before writing past the ned of bm."
	| i code n anInt data end k pastEnd |
	<primitive: 'primitiveDecompressFromByteArray' module: 'MiscPrimitivePlugin'>
	self var: #bm declareC: 'int *bm'.
	self var: #ba declareC: 'unsigned char *ba'.
	i := index.  "byteArray read index"
	end := ba size.
	k := 1.  "bitmap write index"
	pastEnd := bm size + 1.
	[i <= end] whileTrue:
		["Decode next run start N"
		anInt := ba at: i.  i := i+1.
		anInt <= 223 ifFalse:
			[anInt <= 254
				ifTrue: [anInt := (anInt-224)*256 + (ba at: i).  i := i+1]
				ifFalse: [anInt := 0.
						1 to: 4 do: [:j | anInt := (anInt bitShift: 8) + (ba at: i).  i := i+1]]].
		n := anInt >> 2.
		(k + n) > pastEnd ifTrue: [^ self primitiveFail].
		code := anInt bitAnd: 3.
		code = 0 ifTrue: ["skip"].
		code = 1 ifTrue: ["n consecutive words of 4 bytes = the following byte"
						data := ba at: i.  i := i+1.
						data := data bitOr: (data bitShift: 8).
						data := data bitOr: (data bitShift: 16).
						1 to: n do: [:j | bm at: k put: data.  k := k+1]].
		code = 2 ifTrue: ["n consecutive words = 4 following bytes"
						data := 0.
						1 to: 4 do: [:j | data := (data bitShift: 8) bitOr: (ba at: i).  i := i+1].
						1 to: n do: [:j | bm at: k put: data.  k := k+1]].
		code = 3 ifTrue: ["n consecutive words from the data..."
						1 to: n do:
							[:m | data := 0.
							1 to: 4 do: [:j | data := (data bitShift: 8) bitOr: (ba at: i).  i := i+1].
							bm at: k put: data.  k := k+1]]]! !

!Bitmap methodsFor: 'accessing' stamp: 'ar 11/2/1998 12:19'!
defaultElement
	"Return the default element of the receiver"
	^0! !

!Bitmap methodsFor: 'filing' stamp: 'jm 2/15/98 17:27'!
encodeBytesOf: anInt in: ba at: i
	"Copy the integer anInt into byteArray ba at index i, and return the next index"

	self inline: true.
	self var: #ba declareC: 'unsigned char *ba'.
	0 to: 3 do:
		[:j | ba at: i+j put: (anInt >> (3-j*8) bitAnd: 16rFF)].
	^ i+4! !

!Bitmap methodsFor: 'filing' stamp: 'jm 2/12/98 17:32'!
encodeInt: int
	"Encode the integer int as per encodeInt:in:at:, and return it as a ByteArray"
	| byteArray next |
	byteArray := ByteArray new: 5.
	next := self encodeInt: int in: byteArray at: 1.
	^ byteArray copyFrom: 1 to: next - 1
! !

!Bitmap methodsFor: 'filing' stamp: 'jm 2/15/98 17:26'!
encodeInt: anInt in: ba at: i
	"Encode the integer anInt in byteArray ba at index i, and return the next index.
	The encoding is as follows...
		0-223	0-223
		224-254	(0-30)*256 + next byte (0-7935)
		255		next 4 bytes"		

	self inline: true.
	self var: #ba declareC: 'unsigned char *ba'.
	anInt <= 223 ifTrue: [ba at: i put: anInt. ^ i+1].
	anInt <= 7935 ifTrue: [ba at: i put: anInt//256+224. ba at: i+1 put: anInt\\256.  ^ i+2].
	ba at: i put: 255.
	^ self encodeBytesOf: anInt in: ba at: i+1! !

!Bitmap methodsFor: 'initialize-release' stamp: 'ar 12/23/1999 14:35'!
fromByteStream: aStream 
	"Initialize the array of bits by reading integers from the argument, 
	aStream."
	aStream nextWordsInto: self! !

!Bitmap methodsFor: 'accessing' stamp: 'ar 3/3/2001 22:41'!
integerAt: index
	"Return the integer at the given index"
	| word |
	<primitive: 165>
	word := self basicAt: index.
	word < 16r3FFFFFFF ifTrue:[^word]. "Avoid LargeInteger computations"
	^word >= 16r80000000	"Negative?!!"
		ifTrue:["word - 16r100000000"
				(word bitInvert32 + 1) negated]
		ifFalse:[word]! !

!Bitmap methodsFor: 'accessing' stamp: 'ar 3/3/2001 22:42'!
integerAt: index put: anInteger
	"Store the integer at the given index"
	| word |
	<primitive: 166>
	anInteger < 0
		ifTrue:["word := 16r100000000 + anInteger"
				word := (anInteger + 1) negated bitInvert32]
		ifFalse:[word := anInteger].
	self  basicAt: index put: word.
	^anInteger! !

!Bitmap methodsFor: 'testing' stamp: 'ar 5/25/2000 19:42'!
isColormap
	"Bitmaps were used as color maps for BitBlt.
	This method allows to recognize real color maps."
	^false! !

!Bitmap methodsFor: 'accessing' stamp: 'tk 3/15/97'!
pixelValueForDepth: depth
	"Self is being used to represent a single color.  Answer bits that appear in ONE pixel of this color in a Bitmap of the given depth. The depth must be one of 1, 2, 4, 8, 16, or 32.  Returns an integer.  First pixel only.  "

	^ (self at: 1) bitAnd: (1 bitShift: depth) - 1! !

!Bitmap methodsFor: 'accessing' stamp: ''!
primFill: aPositiveInteger
	"Fill the receiver, an indexable bytes or words object, with the given positive integer. The range of possible fill values is [0..255] for byte arrays and [0..(2^32 - 1)] for word arrays."

	<primitive: 145>
	self errorImproperStore.! !

!Bitmap methodsFor: 'printing' stamp: 'sma 6/1/2000 09:42'!
printOn: aStream
	self printNameOn: aStream.
	aStream nextPutAll: ' of length '; print: self size! !

!Bitmap methodsFor: 'printing' stamp: 'MPW 1/1/1901 22:00'!
printOnStream: aStream

	aStream print: 'a Bitmap of length '; write:self size.
! !

!Bitmap methodsFor: 'filing' stamp: 'di 2/11/98 21:34'!
readCompressedFrom: strm
	"Decompress an old-style run-coded stream into this bitmap:
		[0 means end of runs]
		[n = 1..127] [(n+3) copies of next byte]
		[n = 128..191] [(n-127) next bytes as is]
		[n = 192..255] [(n-190) copies of next 4 bytes]"
	| n byte out outBuff bytes |
	out := WriteStream on: (outBuff := ByteArray new: self size*4).
	[(n := strm next) > 0] whileTrue:
		[(n between: 1 and: 127) ifTrue:
			[byte := strm next.
			1 to: n+3 do: [:i | out nextPut: byte]].
		(n between: 128 and: 191) ifTrue:
			[1 to: n-127 do: [:i | out nextPut: strm next]].
		(n between: 192 and: 255) ifTrue:
			[bytes := (1 to: 4) collect: [:i | strm next].
			1 to: n-190 do: [:i | bytes do: [:b | out nextPut: b]]]].
	out position = outBuff size ifFalse: [self error: 'Decompression size error'].
	"Copy the final byteArray into self"
	self copyFromByteArray: outBuff.! !

!Bitmap methodsFor: 'accessing' stamp: ''!
replaceFrom: start to: stop with: replacement startingAt: repStart 
	"Primitive. This destructively replaces elements from start to stop in the receiver starting at index, repStart, in the collection, replacement. Answer the receiver. Range checks are performed in the primitive only. Optional. See Object documentation whatIsAPrimitive."
	<primitive: 105>
	super replaceFrom: start to: stop with: replacement startingAt: repStart! !

!Bitmap methodsFor: 'filing' stamp: 'tk 1/24/2000 22:37'!
restoreEndianness
	"This word object was just read in from a stream.  Bitmaps are always compressed and serialized in a machine-independent way.  Do not correct the Endianness."

	"^ self"
! !

!Bitmap methodsFor: 'filing' stamp: 'nk 12/31/2003 16:02'!
storeBits: startBit to: stopBit on: aStream 
	"Store my bits as a hex string, breaking the lines every 100 bytes or 
	so to comply with the maximum line length limits of Postscript (255 
	bytes). "
	| lineWidth |
	lineWidth := 0.
	self
		do: [:word | 
			startBit
				to: stopBit
				by: -4
				do: [:shift | 
					aStream nextPut: (word >> shift bitAnd: 15) asHexDigit.
					lineWidth := lineWidth + 1].
			(lineWidth > 100)
				ifTrue: [aStream cr.
					lineWidth := 0]].
	lineWidth > 0 ifTrue: [ aStream cr ].! !

!Bitmap methodsFor: 'filing' stamp: 'jm 2/18/98 14:19'!
writeOn: aStream 
	"Store the array of bits onto the argument, aStream. A leading byte of 16r80 identifies this as compressed by compressToByteArray (qv)."

	| b |
	aStream nextPut: 16r80.
	b := self compressToByteArray.
	aStream
		nextPutAll: (self encodeInt: b size);
		nextPutAll: b.
! !

!Bitmap methodsFor: 'filing' stamp: 'tk 2/19/1999 07:36'!
writeUncompressedOn: aStream 
	"Store the array of bits onto the argument, aStream.
	(leading byte ~= 16r80) identifies this as raw bits (uncompressed)."

	aStream nextInt32Put: self size.
	aStream nextPutAll: self
! !

Rectangle subclass: #CharacterBlock
	instanceVariableNames: 'stringIndex text textLine'
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

!CharacterBlock commentStamp: '<historical>' prior: 0!
My instances contain information about displayed characters. They are used to return the results of methods:
	Paragraph characterBlockAtPoint: aPoint and
	Paragraph characterBlockForIndex: stringIndex.
Any recomposition or movement of a Paragraph can make the instance obsolete.!

!CharacterBlock methodsFor: 'comparing' stamp: ''!
< aCharacterBlock 
	"Answer whether the string index of the receiver precedes that of 
	aCharacterBlock."

	^stringIndex < aCharacterBlock stringIndex! !

!CharacterBlock methodsFor: 'comparing' stamp: ''!
<= aCharacterBlock 
	"Answer whether the string index of the receiver does not come after that 
	of aCharacterBlock."

	^(self > aCharacterBlock) not! !

!CharacterBlock methodsFor: 'comparing' stamp: ''!
= aCharacterBlock

	self species = aCharacterBlock species
		ifTrue: [^stringIndex = aCharacterBlock stringIndex]
		ifFalse: [^false]! !

!CharacterBlock methodsFor: 'comparing' stamp: ''!
> aCharacterBlock 
	"Answer whether the string index of the receiver comes after that of 
	aCharacterBlock."

	^aCharacterBlock < self! !

!CharacterBlock methodsFor: 'comparing' stamp: ''!
>= aCharacterBlock 
	"Answer whether the string index of the receiver does not precede that of 
	aCharacterBlock."

	^(self < aCharacterBlock) not! !

!CharacterBlock methodsFor: 'accessing' stamp: 'di 6/7/2000 17:33'!
copy
	"Overridden because Rectangle does a deepCopy, which goes nuts with the text"

	^ self clone! !

!CharacterBlock methodsFor: 'comparing' stamp: 'th 9/17/2002 11:54'!
max: aCharacterBlock
	aCharacterBlock ifNil:[^self].
	^aCharacterBlock > self
		ifTrue:[ aCharacterBlock]
		ifFalse:[self].! !

!CharacterBlock methodsFor: 'comparing' stamp: 'th 9/17/2002 11:54'!
min: aCharacterBlock
	aCharacterBlock ifNil:[^self].
	^aCharacterBlock < self
		ifTrue:[ aCharacterBlock]
		ifFalse:[self].! !

!CharacterBlock methodsFor: 'private' stamp: ''!
moveBy: aPoint 
	"Change the corner positions of the receiver so that its area translates by 
	the amount defined by the argument, aPoint."

	origin := origin + aPoint.
	corner := corner + aPoint! !

!CharacterBlock methodsFor: 'printing' stamp: 'di 12/2/97 19:15'!
printOn: aStream

	aStream nextPutAll: 'a CharacterBlock with index '.
	stringIndex printOn: aStream.
	(text ~~ nil and: [text size> 0 and: [stringIndex between: 1 and: text size]])
		ifTrue: [aStream nextPutAll: ' and character '.
				(text at: stringIndex) printOn: aStream].
	aStream nextPutAll: ' and rectangle '.
	super printOn: aStream.
	textLine ifNotNil: [aStream cr; nextPutAll: ' in '.
				textLine printOn: aStream].
! !

!CharacterBlock methodsFor: 'accessing' stamp: ''!
stringIndex
	"Answer the position of the receiver in the string it indexes."

	^stringIndex! !

!CharacterBlock methodsFor: 'private' stamp: 'di 10/23/97 22:33'!
stringIndex: anInteger text: aText topLeft: topLeft extent: extent

	stringIndex := anInteger.
	text := aText.
	super setOrigin: topLeft corner: topLeft + extent ! !

!CharacterBlock methodsFor: 'accessing' stamp: 'di 12/2/97 14:33'!
textLine
	^ textLine! !

!CharacterBlock methodsFor: 'accessing' stamp: 'di 12/2/97 14:33'!
textLine: aLine
	textLine := aLine! !

Object subclass: #CharacterScanner
	instanceVariableNames: 'destX lastIndex xTable destY stopConditions text textStyle alignment leftMargin rightMargin font line runStopIndex spaceCount spaceWidth emphasisCode kern indentationLevel wantsColumnBreaks'
	classVariableNames: 'DefaultStopConditions NilCondition PaddedSpaceCondition SpaceCondition'
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

!CharacterScanner commentStamp: '<historical>' prior: 0!
My instances hold the state associated with scanning text. My subclasses scan characters for specified purposes, such as computing a CharacterBlock or placing characters into Forms.!

CharacterScanner subclass: #CharacterBlockScanner
	instanceVariableNames: 'characterPoint characterIndex lastCharacter lastCharacterExtent lastSpaceOrTabExtent nextLeftMargin specialWidth'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Text'!

!CharacterBlockScanner commentStamp: '<historical>' prior: 0!
My instances are used to scan text to compute the CharacterBlock for a character specified by its index in the text or its proximity to the cursor location.!

!CharacterBlockScanner methodsFor: 'private' stamp: 'tlk 5/7/2006 16:42'!
buildCharacterBlockIn: para
	| lineIndex runLength lineStop done stopCondition |
	"handle nullText"
	(para numberOfLines = 0 or: [text size = 0])
		ifTrue:	[^ CharacterBlock new stringIndex: 1  "like being off end of string"
					text: para text
					topLeft: (para leftMarginForDisplayForLine: 1 alignment: (alignment ifNil:[textStyle alignment]))
								@ para compositionRectangle top
					extent: 0 @ textStyle lineGrid].
	"find the line"
	lineIndex := para lineIndexOfTop: characterPoint y.
	destY := para topAtLineIndex: lineIndex.
	line := para lines at: lineIndex.
	lastIndex := line first.

	rightMargin := para rightMarginForDisplay.
     self setStopConditions.  " also loads the font and loads all emphasis attributes "


	(lineIndex = para numberOfLines and:
		[(destY + line lineHeight) < characterPoint y])
			ifTrue:	["if beyond lastLine, force search to last character"
					self characterPointSetX: rightMargin]
			ifFalse:	[characterPoint y < (para compositionRectangle) top
						ifTrue: ["force search to first line"
								characterPoint := (para compositionRectangle) topLeft].
					characterPoint x > rightMargin
						ifTrue:	[self characterPointSetX: rightMargin]].
	destX := (leftMargin := para leftMarginForDisplayForLine: lineIndex alignment: (alignment ifNil:[textStyle alignment])).
	nextLeftMargin:= para leftMarginForDisplayForLine: lineIndex+1 alignment: (alignment ifNil:[textStyle alignment]).
	lastIndex := line first.

	self setStopConditions.		"also sets font"
	runLength := (text runLengthFor: line first).
	characterIndex == nil
		ifTrue:	[lineStop := line last  "characterBlockAtPoint"]
		ifFalse:	[lineStop := characterIndex  "characterBlockForIndex"].
	(runStopIndex := lastIndex + (runLength - 1)) > lineStop
		ifTrue:	[runStopIndex := lineStop].
	lastCharacterExtent := 0 @ line lineHeight.
	spaceCount := 0. done  := false.
	self handleIndentation.

	[done]
	whileFalse:
	[stopCondition := self scanCharactersFrom: lastIndex to: runStopIndex
			in: text string rightX: characterPoint x
			stopConditions: stopConditions kern: kern.

	"see setStopConditions for stopping conditions for character block 	operations."
	self lastCharacterExtentSetX: (font widthOf: (text at: lastIndex)).
	(self perform: stopCondition) ifTrue:
		[characterIndex == nil
			ifTrue: ["characterBlockAtPoint"
					^ CharacterBlock new stringIndex: lastIndex text: text
						topLeft: characterPoint + (font descentKern @ 0)
						extent: lastCharacterExtent]
			ifFalse: ["characterBlockForIndex"
					^ CharacterBlock new stringIndex: lastIndex text: text
						topLeft: characterPoint + ((font descentKern) - kern @ 0)
						extent: lastCharacterExtent]]]! !

!CharacterBlockScanner methodsFor: 'scanning' stamp: 'ar 5/17/2000 19:14'!
characterBlockAtPoint: aPoint in: aParagraph
	"Answer a CharacterBlock for character in aParagraph at point aPoint. It 
	is assumed that aPoint has been transformed into coordinates appropriate 
	to the text's destination form rectangle and the composition rectangle."

	self initializeFromParagraph: aParagraph clippedBy: aParagraph clippingRectangle.
	characterPoint := aPoint.
	^self buildCharacterBlockIn: aParagraph! !

!CharacterBlockScanner methodsFor: 'scanning' stamp: 'nk 11/22/2004 14:32'!
characterBlockAtPoint: aPoint index: index in: textLine
	"This method is the Morphic characterBlock finder.  It combines
	MVC's characterBlockAtPoint:, -ForIndex:, and buildCharcterBlock:in:"
	| runLength lineStop done stopCondition |
	line := textLine.
	rightMargin := line rightMargin.
	lastIndex := line first.
	self setStopConditions.		"also sets font"
	characterIndex := index.  " == nil means scanning for point"
	characterPoint := aPoint.
	(characterPoint isNil or: [characterPoint y > line bottom])
		ifTrue: [characterPoint := line bottomRight].
	(text isEmpty or: [(characterPoint y < line top or: [characterPoint x < line left])
				or: [characterIndex notNil and: [characterIndex < line first]]])
		ifTrue:	[^ (CharacterBlock new stringIndex: line first text: text
					topLeft: line leftMargin@line top extent: 0 @ textStyle lineGrid)
					textLine: line].
	destX := leftMargin := line leftMarginForAlignment: alignment.
	destY := line top.
	runLength := text runLengthFor: line first.
	characterIndex
		ifNotNil:	[lineStop := characterIndex  "scanning for index"]
		ifNil:	[lineStop := line last  "scanning for point"].
	runStopIndex := lastIndex + (runLength - 1) min: lineStop.
	lastCharacterExtent := 0 @ line lineHeight.
	spaceCount := 0.

	done  := false.
	[done] whileFalse:
		[stopCondition := self scanCharactersFrom: lastIndex to: runStopIndex
			in: text string rightX: characterPoint x
			stopConditions: stopConditions kern: kern.
		"see setStopConditions for stopping conditions for character block 	operations."
		self lastCharacterExtentSetX: (specialWidth
			ifNil: [font widthOf: (text at: lastIndex)]
			ifNotNil: [specialWidth]).
		(self perform: stopCondition) ifTrue:
			[characterIndex
				ifNil: [
					"Result for characterBlockAtPoint: "
					(stopCondition ~~ #cr and: [ lastIndex == line last
						and: [ aPoint x > ((characterPoint x) + (lastCharacterExtent x / 2)) ]])
							ifTrue: [ "Correct for right half of last character in line"
								^ (CharacterBlock new stringIndex: lastIndex + 1
										text: text
										topLeft: characterPoint + (lastCharacterExtent x @ 0) + (font descentKern @ 0)
										extent:  0 @ lastCharacterExtent y)
									textLine: line ].
						^ (CharacterBlock new stringIndex: lastIndex
							text: text topLeft: characterPoint + (font descentKern @ 0)
							extent: lastCharacterExtent - (font baseKern @ 0))
									textLine: line]
				ifNotNil: ["Result for characterBlockForIndex: "
						^ (CharacterBlock new stringIndex: characterIndex
							text: text topLeft: characterPoint + ((font descentKern) - kern @ 0)
							extent: lastCharacterExtent)
									textLine: line]]]! !

!CharacterBlockScanner methodsFor: 'scanning' stamp: 'ar 5/17/2000 19:14'!
characterBlockForIndex: targetIndex in: aParagraph 
	"Answer a CharacterBlock for character in aParagraph at targetIndex. The 
	coordinates in the CharacterBlock will be appropriate to the intersection 
	of the destination form rectangle and the composition rectangle."

	self 
		initializeFromParagraph: aParagraph 
		clippedBy: aParagraph clippingRectangle.
	characterIndex := targetIndex.
	characterPoint := 
		aParagraph rightMarginForDisplay @ 
			(aParagraph topAtLineIndex: 
				(aParagraph lineIndexOfCharacterIndex: characterIndex)).
	^self buildCharacterBlockIn: aParagraph! !

!CharacterBlockScanner methodsFor: 'private' stamp: 'ar 1/8/2000 14:34'!
characterPointSetX: xVal
	characterPoint := xVal @ characterPoint y! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'RAA 2/25/2001 14:55'!
cr 
	"Answer a CharacterBlock that specifies the current location of the mouse 
	relative to a carriage return stop condition that has just been 
	encountered. The ParagraphEditor convention is to denote selections by 
	CharacterBlocks, sometimes including the carriage return (cursor is at 
	the end) and sometimes not (cursor is in the middle of the text)."

	((characterIndex ~= nil
		and: [characterIndex > text size])
			or: [(line last = text size)
				and: [(destY + line lineHeight) < characterPoint y]])
		ifTrue:	["When off end of string, give data for next character"
				destY := destY +  line lineHeight.
				lastCharacter := nil.
				characterPoint := (nextLeftMargin ifNil: [leftMargin]) @ destY.
				lastIndex := lastIndex + 1.
				self lastCharacterExtentSetX: 0.
				^ true].
		lastCharacter := CR.
		characterPoint := destX @ destY.
		self lastCharacterExtentSetX: rightMargin - destX.
		^true! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'ar 12/15/2001 23:28'!
crossedX
	"Text display has wrapping. The scanner just found a character past the x 
	location of the cursor. We know that the cursor is pointing at a character 
	or before one."

	| leadingTab currentX |
	characterIndex == nil ifFalse: [
		"If the last character of the last line is a space,
		and it crosses the right margin, then locating
		the character block after it is impossible without this hack."
		characterIndex > text size ifTrue: [
			lastIndex := characterIndex.
			characterPoint := (nextLeftMargin ifNil: [leftMargin]) @ (destY + line lineHeight).
			^true]].
	characterPoint x <= (destX + (lastCharacterExtent x // 2))
		ifTrue:	[lastCharacter := (text at: lastIndex).
				characterPoint := destX @ destY.
				^true].
	lastIndex >= line last 
		ifTrue:	[lastCharacter := (text at: line last).
				characterPoint := destX @ destY.
				^true].

	"Pointing past middle of a character, return the next character."
	lastIndex := lastIndex + 1.
	lastCharacter := text at: lastIndex.
	currentX := destX + lastCharacterExtent x + kern.
	self lastCharacterExtentSetX: (font widthOf: lastCharacter).
	characterPoint := currentX @ destY.
	lastCharacter = Space ifFalse: [^ true].

	"Yukky if next character is space or tab."
	alignment = Justified ifTrue:
		[self lastCharacterExtentSetX:
			(lastCharacterExtent x + 	(line justifiedPadFor: (spaceCount + 1))).
		^ true].

	true ifTrue: [^ true].
	"NOTE:  I find no value to the following code, and so have defeated it - DI"

	"See tabForDisplay for illumination on the following awfulness."
	leadingTab := true.
	line first to: lastIndex - 1 do:
		[:index | (text at: index) ~= Tab ifTrue: [leadingTab := false]].
	(alignment ~= Justified or: [leadingTab])
		ifTrue:	[self lastCharacterExtentSetX: (textStyle nextTabXFrom: currentX
					leftMargin: leftMargin rightMargin: rightMargin) -
						currentX]
		ifFalse:	[self lastCharacterExtentSetX:  (((currentX + (textStyle tabWidth -
						(line justifiedTabDeltaFor: spaceCount))) -
							currentX) max: 0)].
	^ true! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'yo 11/18/2002 13:16'!
endOfRun
	"Before arriving at the cursor location, the selection has encountered an 
	end of run. Answer false if the selection continues, true otherwise. Set 
	up indexes for building the appropriate CharacterBlock."

	| runLength lineStop |
	(((characterIndex ~~ nil and:
		[runStopIndex < characterIndex and: [runStopIndex < text size]])
			or:	[characterIndex == nil and: [lastIndex < line last]]) or: [
				((lastIndex < line last)
				and: [((text at: lastIndex) leadingChar ~= (text at: lastIndex+1) leadingChar)
					and: [lastIndex ~= characterIndex]])])
		ifTrue:	["We're really at the end of a real run."
				runLength := (text runLengthFor: (lastIndex := lastIndex + 1)).
				characterIndex ~~ nil
					ifTrue:	[lineStop := characterIndex	"scanning for index"]
					ifFalse:	[lineStop := line last			"scanning for point"].
				(runStopIndex := lastIndex + (runLength - 1)) > lineStop
					ifTrue: 	[runStopIndex := lineStop].
				self setStopConditions.
				^false].

	lastCharacter := text at: lastIndex.
	characterPoint := destX @ destY.
	((lastCharacter = Space and: [alignment = Justified])
		or: [lastCharacter = Tab and: [lastSpaceOrTabExtent notNil]])
		ifTrue: [lastCharacterExtent := lastSpaceOrTabExtent].
	characterIndex ~~ nil
		ifTrue:	["If scanning for an index and we've stopped on that index,
				then we back destX off by the width of the character stopped on
				(it will be pointing at the right side of the character) and return"
				runStopIndex = characterIndex
					ifTrue:	[self characterPointSetX: destX - lastCharacterExtent x.
							^true].
				"Otherwise the requested index was greater than the length of the
				string.  Return string size + 1 as index, indicate further that off the
				string by setting character to nil and the extent to 0."
				lastIndex :=  lastIndex + 1.
				lastCharacter := nil.
				self lastCharacterExtentSetX: 0.
				^true].

	"Scanning for a point and either off the end of the line or off the end of the string."
	runStopIndex = text size
		ifTrue:	["off end of string"
				lastIndex :=  lastIndex + 1.
				lastCharacter := nil.
				self lastCharacterExtentSetX: 0.
				^true].
	"just off end of line without crossing x"
	lastIndex := lastIndex + 1.
	^true! !

!CharacterBlockScanner methodsFor: 'scanning' stamp: 'hmm 2/2/2001 15:07'!
indentationLevel: anInteger
	super indentationLevel: anInteger.
	nextLeftMargin := leftMargin.
	indentationLevel timesRepeat: [
		nextLeftMargin := textStyle nextTabXFrom: nextLeftMargin
					leftMargin: leftMargin
					rightMargin: rightMargin]! !

!CharacterBlockScanner methodsFor: 'private' stamp: 'ar 1/8/2000 14:34'!
lastCharacterExtentSetX: xVal
	lastCharacterExtent := xVal @ lastCharacterExtent y! !

!CharacterBlockScanner methodsFor: 'private' stamp: 'ar 1/8/2000 14:34'!
lastSpaceOrTabExtentSetX: xVal
	lastSpaceOrTabExtent := xVal @ lastSpaceOrTabExtent y! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'ar 1/9/2000 13:51'!
paddedSpace
	"When the line is justified, the spaces will not be the same as the font's 
	space character. A padding of extra space must be considered in trying 
	to find which character the cursor is pointing at. Answer whether the 
	scanning has crossed the cursor."

	| pad |
	pad := 0.
	spaceCount := spaceCount + 1.
	pad := line justifiedPadFor: spaceCount.
	lastSpaceOrTabExtent := lastCharacterExtent copy.
	self lastSpaceOrTabExtentSetX:  spaceWidth + pad.
	(destX + lastSpaceOrTabExtent x)  >= characterPoint x
		ifTrue: [lastCharacterExtent := lastSpaceOrTabExtent copy.
				^self crossedX].
	lastIndex := lastIndex + 1.
	destX := destX + lastSpaceOrTabExtent x.
	^ false
! !

!CharacterBlockScanner methodsFor: 'scanning' stamp: 'ar 12/16/2001 19:27'!
placeEmbeddedObject: anchoredMorph
	"Workaround: The following should really use #textAnchorType"
	anchoredMorph relativeTextAnchorPosition ifNotNil:[^true].
	(super placeEmbeddedObject: anchoredMorph) ifFalse: [^ false].
	specialWidth := anchoredMorph width.
	^ true! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'ar 1/8/2000 14:32'!
setFont
	specialWidth := nil.
	super setFont! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'ar 10/18/2004 14:30'!
setStopConditions
	"Set the font and the stop conditions for the current run."
	
	self setFont.
	self setConditionArray: (alignment = Justified ifTrue: [#paddedSpace]).
! !

!CharacterBlockScanner methodsFor: 'stop conditions' stamp: 'ar 12/15/2001 23:28'!
tab
	| currentX |
	currentX := (alignment == Justified and: [self leadingTab not])
		ifTrue:		"imbedded tabs in justified text are weird"
			[destX + (textStyle tabWidth - (line justifiedTabDeltaFor: spaceCount)) max: destX]
		ifFalse:
			[textStyle
				nextTabXFrom: destX
				leftMargin: leftMargin
				rightMargin: rightMargin].
	lastSpaceOrTabExtent := lastCharacterExtent copy.
	self lastSpaceOrTabExtentSetX: (currentX - destX max: 0).
	currentX >= characterPoint x
		ifTrue: 
			[lastCharacterExtent := lastSpaceOrTabExtent copy.
			^ self crossedX].
	destX := currentX.
	lastIndex := lastIndex + 1.
	^false! !

!CharacterScanner class methodsFor: 'class initialization' stamp: 'yo 12/18/2002 14:09'!
initialize
"
	CharacterScanner initialize
"
	| a |
	a := Array new: 258.
	a at: 1 + 1 put: #embeddedObject.
	a at: Tab asciiValue + 1 put: #tab.
	a at: CR asciiValue + 1 put: #cr.
	a at: EndOfRun put: #endOfRun.
	a at: CrossedX put: #crossedX.
	NilCondition := a copy.
	DefaultStopConditions := a copy.

	PaddedSpaceCondition := a copy.
	PaddedSpaceCondition at: Space asciiValue + 1 put: #paddedSpace.
	
	SpaceCondition := a copy.
	SpaceCondition at: Space asciiValue + 1 put: #space.
! !

!CharacterScanner methodsFor: 'private' stamp: 'ar 1/8/2000 14:27'!
addEmphasis: code
	"Set the bold-ital-under-strike emphasis."
	emphasisCode := emphasisCode bitOr: code! !

!CharacterScanner methodsFor: 'private' stamp: 'ar 1/8/2000 14:27'!
addKern: kernDelta
	"Set the current kern amount."
	kern := kern + kernDelta! !

!CharacterScanner methodsFor: 'scanning' stamp: 'yo 9/23/2002 16:13'!
basicScanCharactersFrom: startIndex to: stopIndex in: sourceString rightX: rightX stopConditions: stops kern: kernDelta
	"Primitive. This is the inner loop of text display--but see 
	scanCharactersFrom: to:rightX: which would get the string, 
	stopConditions and displaying from the instance. March through source 
	String from startIndex to stopIndex. If any character is flagged with a 
	non-nil entry in stops, then return the corresponding value. Determine 
	width of each character from xTable, indexed by map. 
	If dextX would exceed rightX, then return stops at: 258. 
	Advance destX by the width of the character. If stopIndex has been
	reached, then return stops at: 257. Optional. 
	See Object documentation whatIsAPrimitive."
	| ascii nextDestX char |
	<primitive: 103>
	lastIndex := startIndex.
	[lastIndex <= stopIndex]
		whileTrue: 
			[char := (sourceString at: lastIndex).
			ascii := char asciiValue + 1.
			(stops at: ascii) == nil ifFalse: [^stops at: ascii].
			"Note: The following is querying the font about the width
			since the primitive may have failed due to a non-trivial
			mapping of characters to glyphs or a non-existing xTable."
			nextDestX := destX + (font widthOf: char).
			nextDestX > rightX ifTrue: [^stops at: CrossedX].
			destX := nextDestX + kernDelta.
			lastIndex := lastIndex + 1].
	lastIndex := stopIndex.
	^stops at: EndOfRun! !

!CharacterScanner methodsFor: 'scanning' stamp: 'RAA 5/4/2001 13:53'!
columnBreak

	^true! !

!CharacterScanner methodsFor: 'scanning' stamp: 'ar 12/17/2001 01:50'!
embeddedObject
	| savedIndex |
	savedIndex := lastIndex.
	text attributesAt: lastIndex do:[:attr| 
		attr anchoredMorph ifNotNil:[
			"Following may look strange but logic gets reversed.
			If the morph fits on this line we're not done (return false for true) 
			and if the morph won't fit we're done (return true for false)"
			(self placeEmbeddedObject: attr anchoredMorph) ifFalse:[^true]]].
	lastIndex := savedIndex + 1. "for multiple(!!) embedded morphs"
	^false! !

!CharacterScanner methodsFor: 'scanning' stamp: 'hmm 7/15/2000 22:40'!
handleIndentation
	self indentationLevel timesRepeat: [
		self plainTab]! !

!CharacterScanner methodsFor: 'scanning' stamp: 'ar 5/17/2000 18:20'!
indentationLevel
	"return the number of tabs that are currently being placed at the beginning of each line"
	^indentationLevel ifNil:[0]! !

!CharacterScanner methodsFor: 'scanning' stamp: 'ar 1/8/2000 14:23'!
indentationLevel: anInteger
	"set the number of tabs to put at the beginning of each line"
	indentationLevel := anInteger! !

!CharacterScanner methodsFor: 'initialize' stamp: 'ls 1/14/2002 21:26'!
initialize
	destX := destY := leftMargin := 0.! !

!CharacterScanner methodsFor: 'private' stamp: 'ar 5/17/2000 17:13'!
initializeFromParagraph: aParagraph clippedBy: clippingRectangle

	text := aParagraph text.
	textStyle := aParagraph textStyle. 
! !

!CharacterScanner methodsFor: 'initialize' stamp: 'ar 12/31/2001 00:52'!
initializeStringMeasurer
	stopConditions := Array new: 258.
	stopConditions at: CrossedX put: #crossedX.
	stopConditions at: EndOfRun put: #endOfRun.
! !

!CharacterScanner methodsFor: 'scanner methods' stamp: 'yo 12/18/2002 12:32'!
isBreakableAtIndex: index

	^ (EncodedCharSet at: ((text at: index) leadingChar + 1)) isBreakableAt: index in: text.
! !

!CharacterScanner methodsFor: 'scanning' stamp: 'ar 1/8/2000 14:23'!
leadingTab
	"return true if only tabs lie to the left"
	line first to: lastIndex do:
		[:i | (text at: i) == Tab ifFalse: [^ false]].
	^ true! !

!CharacterScanner methodsFor: 'scanning' stamp: 'tak 3/12/2005 00:43'!
measureString: aString inFont: aFont from: startIndex to: stopIndex
	"WARNING: In order to use this method the receiver has to be set up using #initializeStringMeasurer"
	destX := destY := lastIndex := 0.
	xTable := aFont xTable.
	self scanCharactersFrom: startIndex to: stopIndex in: aString rightX: 999999 stopConditions: stopConditions kern: 0.
	^destX! !

!CharacterScanner methodsFor: 'scanning' stamp: 'ar 12/16/2001 19:27'!
placeEmbeddedObject: anchoredMorph
	"Place the anchoredMorph or return false if it cannot be placed.
	In any event, advance destX by its width."
	| w |
	"Workaround: The following should really use #textAnchorType"
	anchoredMorph relativeTextAnchorPosition ifNotNil:[^true].
	destX := destX + (w := anchoredMorph width).
	(destX > rightMargin and: [(leftMargin + w) <= rightMargin])
		ifTrue: ["Won't fit, but would on next line"
				^ false].
	lastIndex := lastIndex + 1.
	self setFont.  "Force recalculation of emphasis for next run"
	^ true! !

!CharacterScanner methodsFor: 'scanning' stamp: 'ar 12/15/2001 23:28'!
plainTab
	"This is the basic method of adjusting destX for a tab."
	destX := (alignment == Justified and: [self leadingTab not])
		ifTrue:		"embedded tabs in justified text are weird"
			[destX + (textStyle tabWidth - (line justifiedTabDeltaFor: spaceCount)) max: destX]
		ifFalse: 
			[textStyle nextTabXFrom: destX
				leftMargin: leftMargin
				rightMargin: rightMargin]! !

!CharacterScanner methodsFor: 'scanning' stamp: 'sn 7/31/2009 15:03'!
scanCharactersFrom: startIndex to: stopIndex in: sourceString rightX: rightX stopConditions: stops kern: kernDelta

	| startEncoding selector |
	(sourceString isByteString) ifTrue: [^ self basicScanCharactersFrom: startIndex to: stopIndex in: sourceString rightX: rightX stopConditions: stops kern: kernDelta.].

	(sourceString isWideString) ifTrue: [
		startIndex > stopIndex ifTrue: [lastIndex := stopIndex. ^ stops at: EndOfRun].
		startEncoding :=  (sourceString at: startIndex) leadingChar.
		selector := EncodedCharSet scanSelectorAt: startEncoding.
		^ self perform: selector withArguments: (Array with: startIndex with: stopIndex with: sourceString with: rightX with: stopConditions with: kernDelta).
	].
	
	^ stops at: EndOfRun
! !

!CharacterScanner methodsFor: 'scanner methods' stamp: 'ar 8/15/2009 11:20'!
scanJapaneseCharactersFrom: startIndex to: stopIndex in: sourceString rightX: rightX stopConditions: stops kern: kernDelta

	| ascii encoding f nextDestX startEncoding |
	lastIndex := startIndex.
	lastIndex > stopIndex ifTrue: [lastIndex := stopIndex. ^ stops at: EndOfRun].
	startEncoding := (sourceString at: startIndex) leadingChar.
	font ifNil: [font := (TextConstants at: #DefaultMultiStyle) fontArray at: 1].
	((font isMemberOf: StrikeFontSet) or: [font isKindOf: TTCFontSet]) ifTrue: [
		f := font fontArray at: startEncoding + 1.
		spaceWidth := f widthOf: Space.
	].

	[lastIndex <= stopIndex] whileTrue: [
		encoding := (sourceString at: lastIndex) leadingChar.
		encoding ~= startEncoding ifTrue: [lastIndex := lastIndex - 1. ^ stops at: EndOfRun].
		ascii := (sourceString at: lastIndex) charCode.
		(encoding = 0 and: [ascii < 256 and:[(stops at: ascii + 1) notNil]]) 
			ifTrue: [^ stops at: ascii + 1].
		nextDestX := destX + (font widthOf: (sourceString at: lastIndex)).
		nextDestX > rightX ifTrue: [^ stops at: CrossedX].
		destX := nextDestX + kernDelta.
		"destX printString displayAt: 0@(lastIndex*20)."
		lastIndex := lastIndex + 1.
	].
	lastIndex := stopIndex.
	^ stops at: EndOfRun! !

!CharacterScanner methodsFor: 'scanner methods' stamp: 'ar 8/15/2009 11:19'!
scanMultiCharactersFrom: startIndex to: stopIndex in: sourceString rightX: rightX stopConditions: stops kern: kernDelta

	| ascii encoding f nextDestX startEncoding |
	lastIndex := startIndex.
	lastIndex > stopIndex ifTrue: [lastIndex := stopIndex. ^ stops at: EndOfRun].
	startEncoding := (sourceString at: startIndex) leadingChar.
	font ifNil: [font := (TextConstants at: #DefaultMultiStyle) fontArray at: 1].
	((font isMemberOf: StrikeFontSet) or: [font isKindOf: TTCFontSet]) ifTrue: [
		f := font fontArray at: startEncoding + 1.
		spaceWidth := f widthOf: Space.
	].

	[lastIndex <= stopIndex] whileTrue: [
		encoding := (sourceString at: lastIndex) leadingChar.
		encoding ~= startEncoding ifTrue: [lastIndex := lastIndex - 1. ^ stops at: EndOfRun].
		ascii := (sourceString at: lastIndex) charCode.
		(encoding = 0 and: [ascii < 256 and:[(stops at: ascii + 1) notNil]]) 
			ifTrue: [^ stops at: ascii + 1].
		nextDestX := destX + (font widthOf: (sourceString at: lastIndex)).
		nextDestX > rightX ifTrue: [^ stops at: CrossedX].
		destX := nextDestX + kernDelta.
		"destX printString displayAt: 0@(lastIndex*20)."
		lastIndex := lastIndex + 1.
	].
	lastIndex := stopIndex.
	^ stops at: EndOfRun! !

!CharacterScanner methodsFor: 'private' stamp: 'ar 1/8/2000 14:27'!
setActualFont: aFont
	"Set the basal font to an isolated font reference."

	font := aFont! !

!CharacterScanner methodsFor: 'private' stamp: 'ar 12/15/2001 23:31'!
setAlignment: style
	alignment := style.
	! !

!CharacterScanner methodsFor: 'private' stamp: 'yo 10/7/2002 14:33'!
setConditionArray: aSymbol

	aSymbol == #paddedSpace ifTrue: [^stopConditions := PaddedSpaceCondition copy].
	aSymbol == #space ifTrue: [^stopConditions := SpaceCondition copy].
	aSymbol == nil ifTrue: [^stopConditions := NilCondition copy].
	self error: 'undefined stopcondition for space character'.
! !

!CharacterScanner methodsFor: 'private' stamp: 'tak 3/12/2005 00:43'!
setFont
	| priorFont |
	"Set the font and other emphasis."
	priorFont := font.
	text == nil ifFalse:[
		emphasisCode := 0.
		kern := 0.
		indentationLevel := 0.
		alignment := textStyle alignment.
		font := nil.
		(text attributesAt: lastIndex forStyle: textStyle)
			do: [:att | att emphasizeScanner: self]].
	font == nil ifTrue:
		[self setFont: textStyle defaultFontIndex].
	font := font emphasized: emphasisCode.
	priorFont ifNotNil: [destX := destX + priorFont descentKern].
	destX := destX - font descentKern.
	"NOTE: next statement should be removed when clipping works"
	leftMargin ifNotNil: [destX := destX max: leftMargin].
	kern := kern - font baseKern.

	"Install various parameters from the font."
	spaceWidth := font widthOf: Space.
	xTable := font xTable.
	stopConditions := DefaultStopConditions.! !

!CharacterScanner methodsFor: 'private' stamp: 'ar 1/8/2000 14:28'!
setFont: fontNumber
	"Set the font by number from the textStyle."

	self setActualFont: (textStyle fontAt: fontNumber)! !

!CharacterScanner methodsFor: 'private' stamp: 'ar 1/8/2000 14:28'!
text: t textStyle: ts
	text := t.
	textStyle := ts! !

!CharacterScanner methodsFor: 'private' stamp: 'ar 1/8/2000 14:28'!
textColor: ignored
	"Overridden in DisplayScanner"! !

!CharacterScanner methodsFor: 'initialize' stamp: 'RAA 5/7/2001 10:11'!
wantsColumnBreaks: aBoolean

	wantsColumnBreaks := aBoolean! !

CharacterScanner subclass: #CompositionScanner
	instanceVariableNames: 'spaceX spaceIndex lineHeight baseline lineHeightAtSpace baselineAtSpace'
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

!CompositionScanner commentStamp: '<historical>' prior: 0!
CompositionScanners are used to measure text and determine where line breaks and space padding should occur.!

!CompositionScanner methodsFor: 'stop conditions' stamp: 'RAA 5/4/2001 13:52'!
columnBreak

	"Answer true. Set up values for the text line interval currently being 
	composed."

	line stop: lastIndex.
	spaceX := destX.
	line paddingWidth: rightMargin - spaceX.
	^true! !

!CompositionScanner methodsFor: 'scanning' stamp: 'ar 12/17/2001 02:06'!
composeFrom: startIndex inRectangle: lineRectangle
	firstLine: firstLine leftSide: leftSide rightSide: rightSide
	"Answer an instance of TextLineInterval that represents the next line in the paragraph."
	| runLength done stopCondition |
	"Set up margins"
	leftMargin := lineRectangle left.
	leftSide ifTrue: [leftMargin := leftMargin +
						(firstLine ifTrue: [textStyle firstIndent]
								ifFalse: [textStyle restIndent])].
	destX := spaceX := leftMargin.
	rightMargin := lineRectangle right.
	rightSide ifTrue: [rightMargin := rightMargin - textStyle rightIndent].
	lastIndex := startIndex.	"scanning sets last index"
	destY := lineRectangle top.
	lineHeight := baseline := 0.  "Will be increased by setFont"
	self setStopConditions.	"also sets font"
	runLength := text runLengthFor: startIndex.
	runStopIndex := (lastIndex := startIndex) + (runLength - 1).
	line := (TextLine start: lastIndex stop: 0 internalSpaces: 0 paddingWidth: 0)
				rectangle: lineRectangle.
	spaceCount := 0.
	self handleIndentation.
	leftMargin := destX.
	line leftMargin: leftMargin.

	done := false.
	[done]
		whileFalse: 
			[stopCondition := self scanCharactersFrom: lastIndex to: runStopIndex
				in: text string rightX: rightMargin stopConditions: stopConditions
				kern: kern.
			"See setStopConditions for stopping conditions for composing."
			(self perform: stopCondition)
				ifTrue: [^ line lineHeight: lineHeight + textStyle leading
							baseline: baseline + textStyle leading]]! !

!CompositionScanner methodsFor: 'scanning' stamp: 'hmm 7/20/2000 18:24'!
composeLine: lineIndex fromCharacterIndex: startIndex inParagraph: aParagraph 
	"Answer an instance of TextLineInterval that represents the next line in the paragraph."
	| runLength done stopCondition |
	destX := spaceX := leftMargin := aParagraph leftMarginForCompositionForLine: lineIndex.
	destY := 0.
	rightMargin := aParagraph rightMarginForComposition.
	leftMargin >= rightMargin ifTrue: [self error: 'No room between margins to compose'].
	lastIndex := startIndex.	"scanning sets last index"
	lineHeight := textStyle lineGrid.  "may be increased by setFont:..."
	baseline := textStyle baseline.
	self setStopConditions.	"also sets font"
	self handleIndentation.
	runLength := text runLengthFor: startIndex.
	runStopIndex := (lastIndex := startIndex) + (runLength - 1).
	line := TextLineInterval
		start: lastIndex
		stop: 0
		internalSpaces: 0
		paddingWidth: 0.
	spaceCount := 0.
	done := false.
	[done]
		whileFalse: 
			[stopCondition := self scanCharactersFrom: lastIndex to: runStopIndex
				in: text string rightX: rightMargin stopConditions: stopConditions
				kern: kern.
			"See setStopConditions for stopping conditions for composing."
			(self perform: stopCondition)
				ifTrue: [^line lineHeight: lineHeight + textStyle leading
							baseline: baseline + textStyle leading]]! !

!CompositionScanner methodsFor: 'scanning' stamp: 'ar 8/4/2009 23:53'!
computeDefaultLineHeight
	"Compute the default line height for a potentially empty text"
	lastIndex := 1.
	self setFont.
	^lineHeight + textStyle leading! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'ar 1/9/2000 13:54'!
cr
	"Answer true. Set up values for the text line interval currently being 
	composed."

	line stop: lastIndex.
	spaceX := destX.
	line paddingWidth: rightMargin - spaceX.
	^true! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'ar 1/9/2000 13:54'!
crossedX
	"There is a word that has fallen across the right edge of the composition 
	rectangle. This signals the need for wrapping which is done to the last 
	space that was encountered, as recorded by the space stop condition."

	spaceCount >= 1 ifTrue:
		["The common case. First back off to the space at which we wrap."
		line stop: spaceIndex.
		lineHeight := lineHeightAtSpace.
		baseline := baselineAtSpace.
		spaceCount := spaceCount - 1.
		spaceIndex := spaceIndex - 1.

		"Check to see if any spaces preceding the one at which we wrap.
			Double space after punctuation, most likely."
		[(spaceCount > 1 and: [(text at: spaceIndex) = Space])]
			whileTrue:
				[spaceCount := spaceCount - 1.
				"Account for backing over a run which might
					change width of space."
				font := text fontAt: spaceIndex withStyle: textStyle.
				spaceIndex := spaceIndex - 1.
				spaceX := spaceX - (font widthOf: Space)].
		line paddingWidth: rightMargin - spaceX.
		line internalSpaces: spaceCount]
	ifFalse:
		["Neither internal nor trailing spaces -- almost never happens."
		lastIndex := lastIndex - 1.
		[destX <= rightMargin]
			whileFalse:
				[destX := destX - (font widthOf: (text at: lastIndex)).
				lastIndex := lastIndex - 1].
		spaceX := destX.
		line paddingWidth: rightMargin - destX.
		line stop: (lastIndex max: line first)].
	^true! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'ar 1/9/2000 13:54'!
endOfRun
	"Answer true if scanning has reached the end of the paragraph. 
	Otherwise step conditions (mostly install potential new font) and answer 
	false."

	| runLength |
	lastIndex = text size
	ifTrue:	[line stop: lastIndex.
			spaceX := destX.
			line paddingWidth: rightMargin - destX.
			^true]
	ifFalse:	[runLength := (text runLengthFor: (lastIndex := lastIndex + 1)).
			runStopIndex := lastIndex + (runLength - 1).
			self setStopConditions.
			^false]
! !

!CompositionScanner methodsFor: 'intialize-release' stamp: 'ar 5/17/2000 19:14'!
forParagraph: aParagraph
	"Initialize the receiver for scanning the given paragraph."

	self
		initializeFromParagraph: aParagraph
		clippedBy: aParagraph clippingRectangle.
! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'ar 12/17/2001 02:13'!
placeEmbeddedObject: anchoredMorph
	| descent |
	"Workaround: The following should really use #textAnchorType"
	anchoredMorph relativeTextAnchorPosition ifNotNil:[^true].
	(super placeEmbeddedObject: anchoredMorph) ifFalse: ["It doesn't fit"
		"But if it's the first character then leave it here"
		lastIndex < line first ifFalse:[
			line stop: lastIndex-1.
			^ false]].
	descent := lineHeight - baseline.
	lineHeight := lineHeight max: anchoredMorph height.
	baseline := lineHeight - descent.
	line stop: lastIndex.
	^ true! !

!CompositionScanner methodsFor: 'accessing' stamp: 'ar 1/8/2000 14:35'!
rightX
	"Meaningful only when a line has just been composed -- refers to the 
	line most recently composed. This is a subtrefuge to allow for easy 
	resizing of a composition rectangle to the width of the maximum line. 
	Useful only when there is only one line in the form or when each line 
	is terminated by a carriage return. Handy for sizing menus and lists."

	^spaceX! !

!CompositionScanner methodsFor: 'scanning' stamp: 'ar 1/8/2000 14:36'!
setActualFont: aFont
	"Keep track of max height and ascent for auto lineheight"
	| descent |
	super setActualFont: aFont.
	lineHeight == nil
		ifTrue: [descent := font descent.
				baseline := font ascent.
				lineHeight := baseline + descent]
		ifFalse: [descent := lineHeight - baseline max: font descent.
				baseline := baseline max: font ascent.
				lineHeight := lineHeight max: baseline + descent]! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'RAA 5/7/2001 10:12'!
setFont
	super setFont.
	stopConditions == DefaultStopConditions 
		ifTrue:[stopConditions := stopConditions copy].
	stopConditions at: Space asciiValue + 1 put: #space.
	wantsColumnBreaks == true ifTrue: [
		stopConditions at: TextComposer characterForColumnBreak asciiValue + 1 put: #columnBreak.
	].
! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'ar 1/8/2000 14:37'!
setStopConditions
	"Set the font and the stop conditions for the current run."
	
	self setFont! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'ar 1/9/2000 13:55'!
space
	"Record left x and character index of the space character just encounted. 
	Used for wrap-around. Answer whether the character has crossed the 
	right edge of the composition rectangle of the paragraph."

	spaceX := destX.
	destX := spaceX + spaceWidth.
	spaceIndex := lastIndex.
	lineHeightAtSpace := lineHeight.
	baselineAtSpace := baseline.
	lastIndex := lastIndex + 1.
	spaceCount := spaceCount + 1.
	destX > rightMargin ifTrue: 	[^self crossedX].
	^false
! !

!CompositionScanner methodsFor: 'stop conditions' stamp: 'ar 1/9/2000 13:59'!
tab
	"Advance destination x according to tab settings in the paragraph's 
	textStyle. Answer whether the character has crossed the right edge of 
	the composition rectangle of the paragraph."

	destX := textStyle
				nextTabXFrom: destX leftMargin: leftMargin rightMargin: rightMargin.
	destX > rightMargin ifTrue:	[^self crossedX].
	lastIndex := lastIndex + 1.
	^false
! !

CompositionScanner subclass: #SegmentScanner
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

!SegmentScanner methodsFor: 'as yet unclassified' stamp: 'ar 5/18/2000 16:48'!
setFont
	super setFont.
	"Make a local copy of stop conditions so we don't modify the default"
	stopConditions == DefaultStopConditions 
		ifTrue:[stopConditions := stopConditions copy].
	stopConditions at: Space asciiValue + 1 put: nil.! !

CharacterScanner subclass: #DisplayScanner
	instanceVariableNames: 'bitBlt lineY runX foregroundColor backgroundColor fillBlt lineHeight paragraph paragraphColor morphicOffset ignoreColorChanges'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Text'!

!DisplayScanner commentStamp: '<historical>' prior: 0!
My instances are used to scan text and display it on the screen or in a hidden form.!

!DisplayScanner class methodsFor: 'queries' stamp: 'ar 5/17/2000 17:39'!
defaultFont
	^ TextStyle defaultFont! !

!DisplayScanner methodsFor: 'stop conditions' stamp: 'ar 1/8/2000 14:42'!
cr
	"When a carriage return is encountered, simply increment the pointer 
	into the paragraph."

	lastIndex:= lastIndex + 1.
	^false! !

!DisplayScanner methodsFor: 'stop conditions' stamp: 'di 9/3/2000 16:24'!
crossedX
	"This condition will sometimes be reached 'legally' during display, when, 
	for instance the space that caused the line to wrap actually extends over 
	the right boundary. This character is allowed to display, even though it 
	is technically outside or straddling the clipping ectangle since it is in 
	the normal case not visible and is in any case appropriately clipped by 
	the scanner."

	^ true ! !

!DisplayScanner methodsFor: 'scanning' stamp: 'ar 7/30/2009 22:13'!
displayLine: textLine offset: offset leftInRun: leftInRun
	"The call on the primitive (scanCharactersFrom:to:in:rightX:) will be interrupted according to an array of stop conditions passed to the scanner at which time the code to handle the stop condition is run and the call on the primitive continued until a stop condition returns true (which means the line has terminated).  leftInRun is the # of characters left to scan in the current run; when 0, it is time to call setStopConditions."
	| done stopCondition nowLeftInRun startIndex string lastPos |
	line := textLine.
	morphicOffset := offset.
	lineY := line top + offset y.
	lineHeight := line lineHeight.
	rightMargin := line rightMargin + offset x.
	lastIndex := line first.
	leftInRun <= 0 ifTrue: [self setStopConditions].
	leftMargin := (line leftMarginForAlignment: alignment) + offset x.
	destX := runX := leftMargin.
	fillBlt == nil ifFalse:
		["Not right"
		fillBlt destX: line left destY: lineY
			width: line width left height: lineHeight; copyBits].
	lastIndex := line first.
	leftInRun <= 0
		ifTrue: [nowLeftInRun := text runLengthFor: lastIndex]
		ifFalse: [nowLeftInRun := leftInRun].
	destY := lineY + line baseline - font ascent.
	runStopIndex := lastIndex + (nowLeftInRun - 1) min: line last.
	spaceCount := 0.
	done := false.
	string := text string.
	[done] whileFalse:[
		startIndex := lastIndex.
		lastPos := destX@destY.
		stopCondition := self scanCharactersFrom: lastIndex to: runStopIndex
						in: string rightX: rightMargin stopConditions: stopConditions
						kern: kern.
		lastIndex >= startIndex ifTrue:[
			font displayString: string on: bitBlt 
				from: startIndex 
	"XXXX: The following is an interesting bug. All stopConditions exept #endOfRun
		have lastIndex past the last character displayed. #endOfRun sets it *on* the character.
		If we display up until lastIndex then we will also display invisible characters like
		CR and tab. This problem should be fixed in the scanner (i.e., position lastIndex
		consistently) but I don't want to deal with the fallout right now so we keep the
		fix minimally invasive."
				to: (stopCondition == #endOfRun ifTrue:[lastIndex] ifFalse:[lastIndex-1]) 
				at: lastPos kern: kern].
		"see setStopConditions for stopping conditions for displaying."
		done := self perform: stopCondition.
		lastIndex > runStopIndex ifTrue: [done := true].
	].
	^ runStopIndex - lastIndex   "Number of characters remaining in the current run"! !

!DisplayScanner methodsFor: 'MVC-compatibility' stamp: 'BG 5/31/2003 16:08'!
displayLines: linesInterval in: aParagraph clippedBy: visibleRectangle
	"The central display routine. The call on the primitive 
	(scanCharactersFrom:to:in:rightX:) will be interrupted according to an 
	array of stop conditions passed to the scanner at which time the code to 
	handle the stop condition is run and the call on the primitive continued 
	until a stop condition returns true (which means the line has 
	terminated)."
	| runLength done stopCondition leftInRun startIndex string lastPos |
	"leftInRun is the # of characters left to scan in the current run;
		when 0, it is time to call 'self setStopConditions'"
	morphicOffset := 0@0.
	leftInRun := 0.
	self initializeFromParagraph: aParagraph clippedBy: visibleRectangle.
	ignoreColorChanges := false.
	paragraph := aParagraph.
	foregroundColor := paragraphColor := aParagraph foregroundColor.
	backgroundColor := aParagraph backgroundColor.
	aParagraph backgroundColor isTransparent
		ifTrue: [fillBlt := nil]
		ifFalse: [fillBlt := bitBlt copy.  "Blt to fill spaces, tabs, margins"
				fillBlt sourceForm: nil; sourceOrigin: 0@0.
				fillBlt fillColor: aParagraph backgroundColor].
	rightMargin := aParagraph rightMarginForDisplay.
	lineY := aParagraph topAtLineIndex: linesInterval first.
	bitBlt destForm deferUpdatesIn: visibleRectangle while: [
		linesInterval do: 
			[:lineIndex | 
			line := aParagraph lines at: lineIndex.
			lastIndex := line first.
               self setStopConditions. " causes an assignment to inst var.  alignment "

			leftMargin := aParagraph leftMarginForDisplayForLine: lineIndex alignment: (alignment ifNil:[textStyle alignment]).
			destX := (runX := leftMargin).
			line := aParagraph lines at: lineIndex.
			lineHeight := line lineHeight.
			fillBlt == nil ifFalse:
				[fillBlt destX: visibleRectangle left destY: lineY
					width: visibleRectangle width height: lineHeight; copyBits].
			lastIndex := line first.
			leftInRun <= 0
				ifTrue: [self setStopConditions.  "also sets the font"
						leftInRun := text runLengthFor: line first].
			destY := lineY + line baseline - font ascent.  "Should have happened in setFont"
			runLength := leftInRun.
			runStopIndex := lastIndex + (runLength - 1) min: line last.
			leftInRun := leftInRun - (runStopIndex - lastIndex + 1).
			spaceCount := 0.
			done := false.
			string := text string.
			self handleIndentation.
			[done] whileFalse:[
				startIndex := lastIndex.
				lastPos := destX@destY.
				stopCondition := self scanCharactersFrom: lastIndex to: runStopIndex
							in: string rightX: rightMargin stopConditions: stopConditions
							kern: kern.
				lastIndex >= startIndex ifTrue:[
					font displayString: string on: bitBlt 
						from: startIndex to: lastIndex at: lastPos kern: kern].
				"see setStopConditions for stopping conditions for displaying."
				done := self perform: stopCondition].
			fillBlt == nil ifFalse:
				[fillBlt destX: destX destY: lineY width: visibleRectangle right-destX height: lineHeight; copyBits].
			lineY := lineY + lineHeight]]! !

!DisplayScanner methodsFor: 'stop conditions' stamp: 'di 9/3/2000 16:24'!
endOfRun
	"The end of a run in the display case either means that there is actually 
	a change in the style (run code) to be associated with the string or the 
	end of this line has been reached."
	| runLength |
	lastIndex = line last ifTrue: [^true].
	runX := destX.
	runLength := text runLengthFor: (lastIndex := lastIndex + 1).
	runStopIndex := lastIndex + (runLength - 1) min: line last.
	self setStopConditions.
	^ false! !

!DisplayScanner methodsFor: 'MVC-compatibility' stamp: 'yo 3/14/2005 06:48'!
initializeFromParagraph: aParagraph clippedBy: clippingRectangle

	super initializeFromParagraph: aParagraph clippedBy: clippingRectangle.
	bitBlt := BitBlt asGrafPort toForm: aParagraph destinationForm.
	bitBlt sourceX: 0; width: 0.	"Init BitBlt so that the first call to a primitive will not fail"
	bitBlt combinationRule:
		((Display depth = 1)
			ifTrue:
				[aParagraph rule]
			ifFalse:
				[Form paint]).
	bitBlt colorMap:
		(Bitmap with: 0      "Assumes 1-bit deep fonts"
				with: (bitBlt destForm pixelValueFor: aParagraph foregroundColor)).
	bitBlt clipRect: clippingRectangle! !

!DisplayScanner methodsFor: 'stop conditions' stamp: 'di 9/3/2000 16:20'!
paddedSpace
	"Each space is a stop condition when the alignment is right justified. 
	Padding must be added to the base width of the space according to 
	which space in the line this space is and according to the amount of 
	space that remained at the end of the line when it was composed."

	spaceCount := spaceCount + 1.
	destX := destX + spaceWidth + (line justifiedPadFor: spaceCount).
	lastIndex := lastIndex + 1.
	^ false! !

!DisplayScanner methodsFor: 'scanning' stamp: 'ar 12/17/2001 13:28'!
placeEmbeddedObject: anchoredMorph
	anchoredMorph relativeTextAnchorPosition ifNotNil:[
		anchoredMorph position: 
			anchoredMorph relativeTextAnchorPosition +
			(anchoredMorph owner textBounds origin x @ 0)
			- (0@morphicOffset y) + (0@lineY).
		^true
	].
	(super placeEmbeddedObject: anchoredMorph) ifFalse: [^ false].
	anchoredMorph isMorph ifTrue: [
		anchoredMorph position: ((destX - anchoredMorph width)@lineY) - morphicOffset
	] ifFalse: [
		destY := lineY.
		runX := destX.
		anchoredMorph 
			displayOn: bitBlt destForm 
			at: destX - anchoredMorph width @ destY
			clippingBox: bitBlt clipRect
	].
	^ true! !

!DisplayScanner methodsFor: 'stop conditions' stamp: 'hmm 7/16/2000 08:23'!
plainTab
	| oldX |
	oldX := destX.
	super plainTab.
	fillBlt == nil ifFalse:
		[fillBlt destX: oldX destY: destY width: destX - oldX height: font height; copyBits]! !

!DisplayScanner methodsFor: 'private' stamp: 'ar 5/17/2000 19:26'!
setDestForm: df
	bitBlt setDestForm: df.! !

!DisplayScanner methodsFor: 'private' stamp: 'di 9/3/2000 16:13'!
setFont 
	foregroundColor := paragraphColor.
	super setFont.  "Sets font and emphasis bits, and maybe foregroundColor"
	font installOn: bitBlt foregroundColor: foregroundColor backgroundColor: Color transparent.
	text ifNotNil:[destY := lineY + line baseline - font ascent]! !

!DisplayScanner methodsFor: 'private' stamp: 'hmm 9/16/2000 21:29'!
setPort: aBitBlt
	"Install the BitBlt to use"
	bitBlt := aBitBlt.
	bitBlt sourceX: 0; width: 0.	"Init BitBlt so that the first call to a primitive will not fail"
	bitBlt sourceForm: nil. "Make sure font installation won't be confused"
! !

!DisplayScanner methodsFor: 'stop conditions' stamp: 'yo 10/4/2002 20:43'!
setStopConditions
	"Set the font and the stop conditions for the current run."
	
	self setFont.
	self setConditionArray: (alignment = Justified ifTrue: [#paddedSpace]).

"
	alignment = Justified ifTrue: [
		stopConditions == DefaultStopConditions 
			ifTrue:[stopConditions := stopConditions copy].
		stopConditions at: Space asciiValue + 1 put: #paddedSpace]
"! !

!DisplayScanner methodsFor: 'stop conditions' stamp: 'hmm 7/16/2000 08:23'!
tab
	self plainTab.
	lastIndex := lastIndex + 1.
	^ false! !

!DisplayScanner methodsFor: 'private' stamp: 'ar 1/8/2000 14:51'!
text: t textStyle: ts foreground: foreColor background: backColor fillBlt: blt ignoreColorChanges: shadowMode
	text := t.
	textStyle := ts. 
	foregroundColor := paragraphColor := foreColor.
	(backgroundColor := backColor) isTransparent ifFalse:
		[fillBlt := blt.
		fillBlt fillColor: backgroundColor].
	ignoreColorChanges := shadowMode! !

!DisplayScanner methodsFor: 'private' stamp: 'ar 1/8/2000 14:51'!
textColor: textColor
	ignoreColorChanges ifTrue: [^ self].
	foregroundColor := textColor! !

CharacterScanner subclass: #CompositionScanner
	instanceVariableNames: 'spaceX spaceIndex lineHeight baseline lineHeightAtSpace baselineAtSpace'
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

!CompositionScanner commentStamp: '<historical>' prior: 0!
CompositionScanners are used to measure text and determine where line breaks and space padding should occur.!

DisplayObject subclass: #DisplayText
	instanceVariableNames: 'text textStyle offset form foreColor backColor'
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Display Objects'!

!DisplayText commentStamp: '<historical>' prior: 0!
I represent Text whose emphasis changes are mapped to a set of fonts. My instances have an offset used in determining screen placement for displaying. They get used two different ways in the system. In the user interface, they mainly hold onto some text which is viewed by some form of ParagraphEditor. However, as a DisplayObject, they may need to display efficiently, so my instances have a cache for the bits.!

!DisplayText class methodsFor: 'examples' stamp: 'tk 11/28/2001 16:03'!
example
	"Continually prints two lines of text wherever you point with the cursor.  Terminate by pressing any button on the
	mouse."
	| tx |
	tx := 'this is a line of characters and
this is the second line.' asDisplayText.
	tx foregroundColor: Color black backgroundColor: Color transparent.
	tx := tx alignedTo: #center.
	[Sensor anyButtonPressed]
		whileFalse:
			[tx displayOn: Display at: Sensor cursorPoint]

	"DisplayText example."! !

!DisplayText class methodsFor: 'instance creation' stamp: ''!
text: aText 
	"Answer an instance of me such that the text displayed is aText 
	according to the system's default text style."

	^self new
		setText: aText
		textStyle: DefaultTextStyle copy
		offset: 0 @ 0! !

!DisplayText class methodsFor: 'instance creation' stamp: ''!
text: aText textStyle: aTextStyle 
	"Answer an instance of me such that the text displayed is aText 
	according to the style specified by aTextStyle."

	^self new
		setText: aText
		textStyle: aTextStyle
		offset: 0 @ 0! !

!DisplayText class methodsFor: 'instance creation' stamp: ''!
text: aText textStyle: aTextStyle offset: aPoint 
	"Answer an instance of me such that the text displayed is aText 
	according to the style specified by aTextStyle. The display of the 
	information should be offset by the amount given as the argument, 
	aPoint."

	^self new
		setText: aText
		textStyle: aTextStyle
		offset: aPoint! !

!DisplayText methodsFor: 'accessing' stamp: ''!
alignedTo: alignPointSelector
	"Return a copy with offset according to alignPointSelector which is one of...
	#(topLeft, topCenter, topRight, leftCenter, center, etc)"
	| boundingBox |
	boundingBox := 0@0 corner: self form extent.
	^ self shallowCopy offset: (0@0) - (boundingBox perform: alignPointSelector)! !

!DisplayText methodsFor: 'converting' stamp: 'tk 10/21/97 12:28'!
asParagraph
	"Answer a Paragraph whose text and style are identical to that of the 
	receiver."
	| para |
	para := Paragraph withText: text style: textStyle.
	para foregroundColor: foreColor backgroundColor: backColor.
	backColor isTransparent ifTrue: [para rule: Form paint].
	^ para! !

!DisplayText methodsFor: 'color' stamp: ''!
backgroundColor
	backColor == nil ifTrue: [^ Color transparent].
	^ backColor! !

!DisplayText methodsFor: 'display box access' stamp: ''!
boundingBox 
	"Refer to the comment in DisplayObject|boundingBox."

	^self form boundingBox! !

!DisplayText methodsFor: 'private' stamp: 'nk 6/25/2003 12:51'!
composeForm
	"For the TT strings in MVC widgets in a Morphic world such as a progress bar, the form is created by Morphic machinery."
	| canvas tmpText |
	Smalltalk isMorphic
		ifTrue:
			[tmpText := TextMorph new contentsAsIs: text deepCopy.
			foreColor ifNotNil: [tmpText text addAttribute: (TextColor color: foreColor)].
			backColor ifNotNil: [tmpText backgroundColor: backColor].
			tmpText setTextStyle: textStyle.
			canvas := FormCanvas on: (Form extent: tmpText extent depth: 32).
			tmpText drawOn: canvas.
			form := canvas form.
		]
		ifFalse: [form := self asParagraph asForm]! !

!DisplayText methodsFor: 'display box access' stamp: ''!
computeBoundingBox 
	"Compute minimum enclosing rectangle around characters."

	| character font width carriageReturn lineWidth lineHeight |
	carriageReturn := Character cr.
	width := lineWidth := 0.
	font := textStyle defaultFont.
	lineHeight := textStyle lineGrid.
	1 to: text size do: 
		[:i | 
		character := text at: i.
		character = carriageReturn
		  ifTrue: 
			[lineWidth := lineWidth max: width.
			lineHeight := lineHeight + textStyle lineGrid.
			width := 0]
		  ifFalse: [width := width + (font widthOf: character)]].
	lineWidth := lineWidth max: width.
	^offset extent: lineWidth @ lineHeight! !

!DisplayText methodsFor: 'displaying' stamp: 'yo 6/23/2003 20:05'!
displayOn: aDisplayMedium at: aDisplayPoint clippingBox: clipRectangle rule: ruleInteger fillColor: aForm
	"For TT font, rule 34 is used if possible."
	"Refer to the comment in 
	DisplayObject|displayOn:at:clippingBox:rule:mask:."

	| form1 rule |
	form1 := self form.
	rule := (ruleInteger = Form over and: [backColor isTransparent])
				ifTrue: [form1 depth = 32 ifTrue: [rule := 34] ifFalse: [Form paint]]
				ifFalse: [ruleInteger].
	form1 depth = 32 ifTrue: [rule := 34].
	form1
		displayOn: aDisplayMedium
		at: aDisplayPoint + offset
		clippingBox: clipRectangle
		rule: rule
		fillColor: aForm! !

!DisplayText methodsFor: 'displaying' stamp: ''!
displayOn: aDisplayMedium transformation: displayTransformation clippingBox: clipRectangle align: alignmentPoint with: relativePoint rule: ruleInteger fillColor: aForm 
	"Refer to the comment in 
	DisplayObject|displayOn:transformation:clippingBox:align:with:rule:mask:."

	| absolutePoint |
	absolutePoint := displayTransformation applyTo: relativePoint.
	absolutePoint := absolutePoint x asInteger @ absolutePoint y asInteger.
	self displayOn: aDisplayMedium
		at: absolutePoint - alignmentPoint
		clippingBox: clipRectangle
		rule: ruleInteger
		fillColor: aForm! !

!DisplayText methodsFor: 'displaying' stamp: ''!
displayOnPort: aPort at: location
	self form displayOnPort: aPort at: location + offset! !

!DisplayText methodsFor: 'accessing' stamp: ''!
fontsUsed
	"Return a list of all fonts used currently in this text.  8/19/96 tk"

	^ text runs values asSet collect: [:each | textStyle fontAt: each]! !

!DisplayText methodsFor: 'color' stamp: ''!
foregroundColor
	foreColor == nil ifTrue: [^ Color black].
	^ foreColor! !

!DisplayText methodsFor: 'color' stamp: ''!
foregroundColor: cf backgroundColor: cb
	foreColor := cf.
	backColor := cb! !

!DisplayText methodsFor: 'accessing' stamp: ''!
form 
	"Answer the form into which the receiver's display bits are cached."

	form == nil ifTrue: [self composeForm].
	^form! !

!DisplayText methodsFor: 'accessing' stamp: ''!
lineGrid
	"Answer the relative space between lines of the receiver's text."

	^textStyle lineGrid! !

!DisplayText methodsFor: 'accessing' stamp: ''!
numberOfLines 
	"Answer the number of lines of text in the receiver."

	^self height // text lineGrid! !

!DisplayText methodsFor: 'accessing' stamp: ''!
offset 
	"Refer to the comment in DisplayObject|offset."

	^offset! !

!DisplayText methodsFor: 'accessing' stamp: ''!
offset: aPoint 
	"Refer to the comment in DisplayObject|offset:."

	offset := aPoint! !

!DisplayText methodsFor: 'private' stamp: ''!
setText: aText textStyle: aTextStyle offset: aPoint

	text := aText.
	textStyle := aTextStyle.
	offset := aPoint.
	form := nil! !

!DisplayText methodsFor: 'accessing' stamp: ''!
string
	"Answer the string of the characters displayed by the receiver."

	^text string! !

!DisplayText methodsFor: 'accessing' stamp: ''!
text 
	"Answer the text displayed by the receiver."

	^text! !

!DisplayText methodsFor: 'accessing' stamp: ''!
text: aText 
	"Set the receiver to display the argument, aText."
	
	text := aText.
	form := nil.
	self changed.
	! !

!DisplayText methodsFor: 'accessing' stamp: ''!
textStyle 
	"Answer the style by which the receiver displays its text."

	^textStyle! !

!DisplayText methodsFor: 'accessing' stamp: ''!
textStyle: aTextStyle 
	"Set the style by which the receiver should display its text."

	textStyle := aTextStyle.
	form := nil.
	self changed.
	! !

CompositionScanner subclass: #SegmentScanner
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

AbstractFont subclass: #StrikeFont
	instanceVariableNames: 'characterToGlyphMap xTable glyphs name type minAscii maxAscii maxWidth strikeLength ascent descent xOffset raster subscript superscript emphasis derivativeFonts pointSize fallbackFont charIndex'
	classVariableNames: 'DefaultStringScanner'
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Fonts'!

!StrikeFont commentStamp: '<historical>' prior: 0!
I represent a compact encoding of a set of Forms corresponding to characters in the ASCII character set. All the forms are placed side by side in a large form whose height is the font height, and whose width is the sum of all the character widths. The xTable variable gives the left-x coordinates of the subforms corresponding to the glyphs. Characters are mapped to glyphs by using the characterToGyphMap.

Subclasses can have non-trivial mapping rules as well as different representations for glyphs sizes (e.g., not using an xTable). If so, these classes should return nil when queried for xTable and/or the characterToGlyphMap. This will cause the CharacterScanner primitive to fail and query the font for the width of a character (so that a more programatical approach can be implemented).

For display, fonts need to implement two messages:
	#installOn: aDisplayContext foregroundColor: foregroundColor backgroundColor: backgroundColor
This method installs the receiver (a font) on the given DisplayContext (which may be an instance of BitBlt or Canvas (or any of it's subclasses). The font should take the appropriate action to initialize the display context so that further display operations can be optimized.
	#displayString: aString on: aDisplayContext from: startIndex to: stopIndex at: aPoint kern: kernDelta
This method is called for each subsequent run of characters in aString which is to be displayed with the (previously installed) settings.
!

StrikeFont subclass: #FormSetFont
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Graphics-Fonts'!

!FormSetFont commentStamp: '<historical>' prior: 0!
FormSetFonts are designed to capture individual images as character forms for imbedding in normal text.  While most often used to insert an isolated glyph in some text, the code is actually desinged to support an entire user-defined font.  The TextAttribute subclass TextFontReference is specifically designed for such in-line insertion of exceptional fonts in normal text.!

!FormSetFont class methodsFor: 'examples' stamp: 'ar 1/15/2001 18:38'!
copy: charForm toClipBoardAs: char ascent: ascent
	Clipboard clipboardText:
		(Text string: char asString
			attribute: (TextFontReference toFont: 
				(FormSetFont new
					fromFormArray: (Array with: charForm)
					asciiStart: char asciiValue
					ascent: ascent)))
"
	The S in the Squeak welcome window was installed by doing the following
	in a workspace (where the value of, eg, charForm will persist through BitEdit...
	f := TextStyle default fontAt: 4.
	oldS := f characterFormAt: $S.
	charForm := Form extent: oldS extent depth: 8.
	oldS displayOn: charForm.
	charForm bitEdit.
	...Play around with the BitEditor, then accept and close...
	FormSetFont copy: charForm toClipBoardAs: $S ascent: f ascent.
	...Then do a paste into the Welcome window
"! !

!FormSetFont class methodsFor: 'examples' stamp: ''!
example    "FormSetFont example"
	"Lets the user select a (small) area of the screen to represent the
	character A, then copies 'A' to the clipboard with that as the letter form.
	Thereafter, a paste operation will imbed that character in any text."
	| charForm |
	charForm := Form fromUser.
	self copy: charForm toClipBoardAs: $A ascent: charForm height! !

!FormSetFont methodsFor: 'as yet unclassified' stamp: 'ar 5/23/2000 12:49'!
fromFormArray: formArray asciiStart: asciiStart ascent: ascentVal
	| height width x badChar |
	type := 2.
	name := 'aFormFont'.
	minAscii := asciiStart.
	maxAscii := minAscii + formArray size - 1.
	ascent := ascentVal.
	subscript := superscript := emphasis := 0.
	height := width := 0.
	maxWidth := 0.
	formArray do:
		[:f | width := width + f width.
		maxWidth := maxWidth max: f width.
		height := height max: f height + f offset y].
	badChar := (Form extent: 7@height) borderWidth: 1.
	width := width + badChar width.
	descent := height - ascent.
	pointSize := height.
	glyphs := Form extent: width @ height depth: formArray first depth.
	xTable := Array new: maxAscii + 3 withAll: 0.
	x := 0.
	formArray doWithIndex:
		[:f :i | f displayOn: glyphs at: x@0.
		xTable at: minAscii + i+1 put: (x := x + f width)].
	badChar displayOn: glyphs at: x@0.
	xTable at: maxAscii + 3 put: x + badChar width.
	characterToGlyphMap := nil.! !

!FormSetFont methodsFor: 'as yet unclassified' stamp: ''!
reset  "Ignored by FormSetFonts"! !

StrikeFont subclass: #HostFont
	instanceVariableNames: 'fullWidth kernPairs ranges'
	classVariableNames: 'IsoToSqueakMap'
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Fonts'!

!HostFont class methodsFor: 'accessing' stamp: 'yo 2/14/2004 01:50'!
defaultRanges

	^ Array with: (Array with: 0 with: 16r2AFF).
! !

!HostFont class methodsFor: 'instance creation' stamp: 'ar 6/4/2000 23:13'!
fontName: fontName size: ptSize emphasis: emphasisCode
	"
		^HostFont fontName: (HostFont fontNameFromUser) size: 12 emphasis: 0.
	"
	^self new fontName: fontName size: ptSize emphasis: emphasisCode! !

!HostFont class methodsFor: 'instance creation' stamp: 'yo 2/14/2004 01:17'!
fontName: fontName size: ptSize emphasis: emphasisCode rangesArray: ranges
	"
		^HostFont fontName: (HostFont fontNameFromUser) size: 12 emphasis: 0.
	"
	^self new fontName: fontName size: ptSize emphasis: emphasisCode rangesArray: ranges! !

!HostFont class methodsFor: 'accessing' stamp: 'rbb 2/18/2005 13:23'!
fontNameFromUser
	"HostFont fontNameFromUser"
	| fontNames index labels |
	fontNames := self listFontNames asSortedCollection.
	labels := WriteStream on: (String new: 100).
	fontNames do:[:fn| labels nextPutAll: fn] separatedBy:[labels cr].
	index := (UIManager default chooseFrom: (labels contents substrings) title: 'Choose your font').
	index = 0 ifTrue:[^nil].
	^fontNames at: index! !

!HostFont class methodsFor: 'system defaults' stamp: 'yo 3/17/2004 00:39'!
initForSubtitles
"
	HostFont initForSubtitles
"

	HostFont textStyleFrom: 'Verdana' sizes: #(18 20 22 24 26 28) ranges: HostFont defaultRanges.

	StrikeFontSet installExternalFontFileName: 'greekFont.out' encoding: GreekEnvironment leadingChar encodingName: #Greek textStyleName: #DefaultMultiStyle.


	TTCFontReader encodingTag: SimplifiedChineseEnvironment leadingChar.
	TTCFontSet newTextStyleFromTTFile: 'C:\WINDOWS\Fonts\simhei.TTF'.

	TTCFontReader encodingTag: JapaneseEnvironment leadingChar.
	TTCFontSet newTextStyleFromTTFile: 'C:\WINDOWS\Fonts\msgothic.TTC'.

	TTCFontReader encodingTag: KoreanEnvironment leadingChar.
	TTCFontSet newTextStyleFromTTFile: 'C:\WINDOWS\Fonts\gulim.TTC'.
! !

!HostFont class methodsFor: 'system defaults' stamp: 'yo 2/13/2004 23:25'!
initWin32
	"HostFont initWin32"
	#(
			"Basic fonts"
			('Arial'				"menu/text serifless"
				(10 11 12 13 14 16 18 20 22 24 26 28 30 36 48 60 72 90))
			('Times New Roman'	"menu/text serifs"
				(10 11 12 13 14 16 18 20 22 24 26 28 30 36 48 60 72 90))
			('Courier New'			"menu/text fixed"
				(10 11 12 13 14 16 18 20 22 24 26 28 30 36 48 60 72 90))
			('Wingdings'			"deco"
				(10 11 12 13 14 16 18 20 22 24 26 28 30 36 48 60 72 90))
			('Symbol'				"deco"
				(10 11 12 13 14 16 18 20 22 24 26 28 30 36 48 60 72 90))

			"Nice fonts"
			('Verdana'			"menu/text serifless"
				(10 11 12 13 14 16 18 20 22 24 26 28 30 36 48 60 72 90))

			('Tahoma'			"menu/text serifless"
				(10 11 12 13 14 16 18 20 22 24 26 28 30 36 48 60 72 90))

			('Garamond'			"menu/text serifs"
				(10 11 12 13 14 16 18 20 22 24 26 28 30 36 48 60 72 90))
			('Georgia'			"menu/text serifs"
				(10 11 12 13 14 16 18 20 22 24 26 28 30 36 48 60 72 90))

			('Comic Sans MS'	"eToy"
				(10 11 12 13 14 16 18 20 22 24 26 28 30 36 48 60 72 90))

			"Optional fonts"
			('Impact'			"flaps"
				(10 11 12 13 14 16 18 20 22 24 26 28 30 36 48 60 72 90))

			('Webdings'			"deco"
				(10 11 12 13 14 16 18 20 22 24 26 28 30 36 48 60 72 90))

			('System'		"12pt only"
				(12))
			('Fixedsys'		"12pt only"
				(12))
		) do:[:spec| HostFont textStyleFrom: spec first sizes: spec last].

	TextConstants removeKey: #Atlanta ifAbsent: [].
	TextConstants removeKey: #ComicPlain ifAbsent: [].
	TextConstants removeKey: #ComicBold ifAbsent: [].
	TextConstants removeKey: #Courier ifAbsent: [].
	TextConstants removeKey: #Palatino ifAbsent: [].

	TextConstants at: #DefaultFixedTextStyle put: (TextConstants at: #'Courier New').
	TextConstants at: #Helvetica put:  (TextConstants at: #'Arial').

! !

!HostFont class methodsFor: 'accessing' stamp: 'ar 6/4/2000 23:18'!
listFontName: index
	<primitive:'primitiveListFont' module:'FontPlugin'>
	^nil! !

!HostFont class methodsFor: 'accessing' stamp: 'ar 6/4/2000 23:18'!
listFontNames
	"HostFont listFontNames"
	"List all the OS font names"
	| font fontNames index |
	fontNames := WriteStream on: Array new.
	index := 0.
	[font := self listFontName: index.
	font == nil] whileFalse:[
		fontNames nextPut: font.
		index := index + 1].
	^fontNames contents! !

!HostFont class methodsFor: 'accessing' stamp: 'yo 2/14/2004 01:57'!
rangesForJapanese

	| basics etc |
	basics := {
		Array with: 0 with: 255
	}.
	etc := {
		Array with: 16r370 with: 16r3FF. "greek"
		Array with: 16r400 with: 16r52F. "cyrillic"
		Array with: 16r1D00 with: 16r1D7F. "phonetic"
		Array with: 16r1E00 with: 16r1EFF. "latin extended additional"
		Array with: 16r2000 with: 16r206F. "general punctuation"
		Array with: 16r20A0 with: 16r20CF. "currency symbols"
		Array with: 16r2100 with: 16r214F. "letterlike"
		Array with: 16r2150 with: 16r218F. "number form"
		Array with: 16r2190 with: 16r21FF. "arrows"
		Array with: 16r2200 with: 16r22FF. "math operators"
		Array with: 16r2300 with: 16r23FF. "misc tech"
		Array with: 16r2460 with: 16r24FF. "enclosed alnum"
		Array with: 16r2500 with: 16r257F. "box drawing"
		Array with: 16r2580 with: 16r259F. "box elem"
		Array with: 16r25A0 with: 16r25FF. "geometric shapes"
		Array with: 16r2600 with: 16r26FF. "misc symbols"
		Array with: 16r2700 with: 16r27BF. "dingbats"
		Array with: 16r27C0 with: 16r27EF. "misc math A"
		Array with: 16r27F0 with: 16r27FF. "supplimental arrow A"
		Array with: 16r2900 with: 16r297F. "supplimental arrow B"
		Array with: 16r2980 with: 16r29FF. "misc math B"
		Array with: 16r2A00 with: 16r2AFF. "supplimental math op"
		Array with: 16r2900 with: 16r297F. "supplimental arrow B"
		Array with: 16r2E80 with: 16r2EFF. "cjk radicals suppliment"
		Array with: 16r2F00 with: 16r2FDF. "kangxi radicals"
		Array with: 16r3000 with: 16r303F. "cjk symbols"
		Array with: 16r3040 with: 16r309F. "hiragana"
		Array with: 16r30A0 with: 16r30FF. "katakana"
		Array with: 16r3190 with: 16r319F. "kanbun"
		Array with: 16r31F0 with: 16r31FF. "katakana extension"
		Array with: 16r3200 with: 16r32FF. "enclosed CJK"
		Array with: 16r3300 with: 16r33FF. "CJK compatibility"
		Array with: 16r3400 with: 16r4DBF. "CJK unified extension A"
		Array with: 16r4E00 with: 16r9FAF. "CJK ideograph"
		Array with: 16rF900 with: 16rFAFF. "CJK compatiblity ideograph"
		Array with: 16rFE30 with: 16rFE4F. "CJK compatiblity forms"
		Array with: 16rFF00 with: 16rFFEF. "half and full"
	}.

	^ basics, etc.
! !

!HostFont class methodsFor: 'accessing' stamp: 'ar 1/27/2002 19:37'!
textStyleFrom: fontName
	"HostFont textStyleFromUser"
	| styleName fonts |
	styleName := fontName asSymbol.
	"(TextConstants includesKey: styleName)
		ifTrue:[(self confirm: 
styleName , ' is already defined in TextConstants.
Do you want to replace that definition?')
			ifFalse: [^ self]]."
	fonts := #(10 11 12 13 14 16 18 20 22 24 26 28 30 36 48 60 72 90).
	('Rendering ', styleName) displayProgressAt: Sensor cursorPoint
		from: 1 to: fonts size during:[:bar|
			fonts := fonts
				collect:[:ptSize| bar value: (fonts indexOf: ptSize).
							   self fontName: styleName 
									size: ptSize
									emphasis: 0]
				thenSelect:[:font| font notNil]]. "reject those that failed"
	fonts size = 0 ifTrue:[^self error:'Could not create font style', styleName].
	TextConstants
		at: styleName
		put: (TextStyle fontArray: fonts).! !

!HostFont class methodsFor: 'accessing' stamp: 'ar 1/27/2002 20:12'!
textStyleFrom: fontName sizes: ptSizes
	| styleName fonts |
	styleName := fontName asSymbol.
	(TextConstants includesKey: styleName)
		ifTrue:[(self confirm: 
styleName , ' is already defined in TextConstants.
Do you want to replace that definition?')
			ifFalse: [^ self]].
	('Rendering ', styleName) displayProgressAt: Sensor cursorPoint
		from: 1 to: ptSizes size during:[:bar|
			fonts := ptSizes
				collect:[:ptSize| bar value: (ptSizes indexOf: ptSize).
							   self fontName: styleName 
									size: ptSize
									emphasis: 0]
				thenSelect:[:font| font notNil]]. "reject those that failed"
	fonts size = 0 ifTrue:[^self error:'Could not create font style', styleName].
	TextConstants
		at: styleName
		put: (TextStyle fontArray: fonts).! !

!HostFont class methodsFor: 'accessing' stamp: 'yo 2/14/2004 01:26'!
textStyleFrom: fontName sizes: ptSizes ranges: ranges
	| styleName fonts |
	styleName := fontName asSymbol.
	(TextConstants includesKey: styleName)
		ifTrue:[(self confirm: 
styleName , ' is already defined in TextConstants.
Do you want to replace that definition?')
			ifFalse: [^ self]].
	('Rendering ', styleName) displayProgressAt: Sensor cursorPoint
		from: 1 to: ptSizes size during:[:bar|
			fonts := ptSizes
				collect:[:ptSize| bar value: (ptSizes indexOf: ptSize).
							   self fontName: styleName 
									size: ptSize
									emphasis: 0 rangesArray: ranges
				]
				thenSelect:[:font| font notNil]]. "reject those that failed"
	fonts size = 0 ifTrue:[^self error:'Could not create font style', styleName].
	TextConstants
		at: styleName
		put: (TextStyle fontArray: fonts).! !

!HostFont class methodsFor: 'accessing' stamp: 'ar 8/28/2000 17:27'!
textStyleFromUser
	"HostFont textStyleFromUser"
	| styleName fonts |
	styleName := self fontNameFromUser ifNil:[^self].
	styleName := styleName asSymbol.
	(TextConstants includesKey: styleName)
		ifTrue:[(self confirm: 
styleName , ' is already defined in TextConstants.
Do you want to replace that definition?')
			ifFalse: [^ self]].
	fonts := #(10 12 14 16 18 20 22 24 26 28 30 36 48 60 72 90).
	('Rendering ', styleName) displayProgressAt: Sensor cursorPoint
		from: 1 to: fonts size during:[:bar|
			fonts := fonts
				collect:[:ptSize| bar value: (fonts indexOf: ptSize).
							   self fontName: styleName 
									size: ptSize
									emphasis: 0]
				thenSelect:[:font| font notNil]]. "reject those that failed"
	fonts size = 0 ifTrue:[^self error:'Could not create font style', styleName].
	TextConstants
		at: styleName
		put: (TextStyle fontArray: fonts).! !

!HostFont class methodsFor: 'system defaults' stamp: 'yo 12/2/2004 12:50'!
unloadAsianTT
"
	self unloadAsianTT
"

	TTCFontSet removeStyleName: 'MultiSimHei'.
	TTCFontSet removeStyleName: 'MultiMSGothic'.
	TTCFontSet removeStyleName: 'MultiGulim'.
! !

!HostFont methodsFor: 'accessing' stamp: 'ar 2/2/2002 18:49'!
baseKern
	^0! !

!HostFont methodsFor: 'accessing' stamp: 'yo 2/13/2004 04:06'!
createCharacterToGlyphMap

	^ IdentityGlyphMap new.
! !

!HostFont methodsFor: 'accessing' stamp: 'ar 2/2/2002 18:49'!
descentKern
	^0! !

!HostFont methodsFor: 'accessing' stamp: 'yo 1/7/2005 12:03'!
displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta 

	^ self displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: aPoint y + self ascent.
! !

!HostFont methodsFor: 'accessing' stamp: 'yo 1/7/2005 15:14'!
displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: baselineY

 	^ super displayMultiString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: baselineY.
! !

!HostFont methodsFor: 'emphasis' stamp: 'ar 8/29/2000 21:18'!
emphasized: code
	| derivative addedEmphasis base safeCode |
	code = 0 ifTrue: [^ self].
	derivativeFonts == nil ifTrue:[derivativeFonts := Array new: 32].
	derivative := derivativeFonts at: (safeCode := code min: derivativeFonts size).
	derivative == nil ifFalse: [^ derivative].  "Already have this style"

	"Dont have it -- derive from another with one with less emphasis"
	addedEmphasis := 1 bitShift: safeCode highBit - 1.
	base := self emphasized: safeCode - addedEmphasis.  "Order is Bold, Ital, Under, Narrow"
	addedEmphasis = 1 ifTrue:   "Compute synthetic bold version of the font"
		[derivative := (base copy name: base name) makeBoldGlyphs].
	addedEmphasis = 2 ifTrue:   "Compute synthetic italic version of the font"
		[ derivative := (base copy name: base name) makeItalicGlyphs].
	addedEmphasis = 4 ifTrue:   "Compute underlined version of the font"
		[derivative := (base copy name: base name) makeUnderlinedGlyphs].
	addedEmphasis = 8 ifTrue:   "Compute narrow version of the font"
		[derivative := (base copy name: base name) makeCondensedGlyphs].
	addedEmphasis = 16 ifTrue:   "Compute struck-out version of the font"
		[derivative := (base copy name: base name) makeStruckOutGlyphs].
	derivative emphasis: safeCode.
	derivativeFonts at: safeCode put: derivative.
	^ derivative! !

!HostFont methodsFor: 'private-creation' stamp: 'yo 2/14/2004 01:38'!
fontName: fontName size: ptSize emphasis: emphasisCode

	^ self fontName: fontName size: ptSize emphasis: emphasisCode rangesArray: (Array with: (Array with: 0 with: 255)).
! !

!HostFont methodsFor: 'private-creation' stamp: 'yo 2/14/2004 01:41'!
fontName: fontName size: ptSize emphasis: emphasisCode rangesArray: rangesArray
	"
		^HostFont fontName: ('MS UI Gothic') size: 12 emphasis: 0 rangesArray: EFontBDFFontReaderForRanges basicNew rangesForJapanese.
	"
	| fontHandle xStart w glyphForm fontHeight fw enc rangesStream currentRange |
	fontHandle := self primitiveCreateFont: fontName size: ptSize emphasis: emphasisCode.
	fontHandle ifNil:[^nil].
	ranges := rangesArray.
	ranges ifNil: [ranges := Array with: (Array with: 0 with: 255)].
	pointSize := ptSize.
	name := fontName.
	emphasis := emphasisCode.
	minAscii := 0.
	maxAscii := ranges last last.
	ascent := self primitiveFontAscent: fontHandle.
	descent := self primitiveFontDescent: fontHandle.
	kernPairs := Array new: (self primitiveFontNumKernPairs: fontHandle).
	1 to: kernPairs size do:[:i|
		kernPairs at: i put: (self primitiveFont: fontHandle getKernPair: i)].
	fontHeight := ascent + descent.
	xTable := Array new: maxAscii + 3.
	fullWidth := Array new: maxAscii + 1.
	xStart := maxWidth := 0.
	rangesStream := ReadStream on: (ranges collect: [:e | (e first to: e second)]).
	currentRange := rangesStream next.
	0 to: maxAscii do:[:i|
		xTable at: i+1 put: xStart.
		i > currentRange last ifTrue: [
			[rangesStream atEnd not and: [currentRange := rangesStream next. currentRange last < i]] whileTrue.
			rangesStream atEnd ifTrue: [].
		].
		(currentRange includes: i) ifTrue: [
			xTable at: i+1 put: xStart.
			fw := self primitiveFont: fontHandle fullWidthOfChar: i.
			(#(	1 "anchored morph"
				9 "tab"
				10 "LF"
				13 "CR"
			) includes: i) ifTrue:[fw := {0. 0. 0}].
			fullWidth at: i+1 put: fw.
			w := fw at: 2.
			(fw at: 1) > 0 ifTrue:[w := w + (fw at: 1)].
			(fw at: 3) > 0 ifTrue:[w := w + (fw at: 3)].
			w > maxWidth ifTrue:[maxWidth := w].
			xStart := xStart + w].
		].
	xStart = 0 ifTrue:[^nil].
	strikeLength := xStart.
	xTable at: maxAscii+1 put: xStart.
	xTable at: maxAscii+2 put: xStart.
	xTable at: maxAscii+3 put: xStart.
	glyphs := Form extent: xTable last @ fontHeight depth: 1.
	glyphForm := Form extent: maxWidth @ fontHeight depth: 1.
	0 to: maxAscii do:[:i|
		glyphForm fillWhite.
		self primitiveFont: fontHandle glyphOfChar: i into: glyphForm.
		xStart := xTable at: i+1.
		glyphForm displayOn: glyphs at: xStart@0.
		"glyphForm displayOn: Display at: xStart@0."
	].
	enc := self primitiveFontEncoding: fontHandle.
	enc = 1 ifTrue:[characterToGlyphMap := self isoToSqueakMap].
	self primitiveDestroyFont: fontHandle.
	^self! !

!HostFont methodsFor: 'accessing' stamp: 'ar 2/18/2001 20:01'!
getFontData
	| fontHandle bufSize buffer |
	fontHandle := self primitiveCreateFont: name size: pointSize emphasis: emphasis.
	fontHandle ifNil:[^nil].
	bufSize := self primitiveFontDataSize: fontHandle.
	buffer := ByteArray new: bufSize.
	self primitiveFont: fontHandle getData: buffer.
	^buffer! !

!HostFont methodsFor: 'private-creation' stamp: 'yo 2/13/2004 02:53'!
isoToSqueakMap
	^nil
! !

!HostFont methodsFor: 'emphasis' stamp: 'yo 2/14/2004 01:38'!
makeBoldGlyphs
	"First check if we can use some OS support for this"
	(self class listFontNames includes: name) ifFalse:[^super makeBoldGlyphs].
	"Now attempt a direct creation through the appropriate primitives"
	(self fontName: name size: pointSize emphasis: (emphasis bitOr: 1) rangesArray: ranges) 
		ifNil:[^super makeBoldGlyphs]. "nil means we failed"! !

!HostFont methodsFor: 'emphasis' stamp: 'yo 2/14/2004 01:39'!
makeItalicGlyphs
	"First check if we can use some OS support for this"
	(self class listFontNames includes: name) ifFalse:[^super makeItalicGlyphs].
	"Now attempt a direct creation through the appropriate primitives"
	(self fontName: name size: pointSize emphasis: (emphasis bitOr: 2) rangesArray: ranges)
		ifNil:[^super makeItalicGlyphs]. "nil means we failed"! !

!HostFont methodsFor: 'emphasis' stamp: 'yo 2/14/2004 01:39'!
makeStruckOutGlyphs
	"First check if we can use some OS support for this"
	(self class listFontNames includes: name) ifFalse:[^super makeStruckOutGlyphs].
	"Now attempt a direct creation through the appropriate primitives"
	(self fontName: name size: pointSize emphasis: (emphasis bitOr: 8) rangesArray: ranges)
		ifNil:[^super makeStruckOutGlyphs]. "nil means we failed"! !

!HostFont methodsFor: 'emphasis' stamp: 'yo 2/14/2004 01:40'!
makeUnderlinedGlyphs
	"First check if we can use some OS support for this"
	(self class listFontNames includes: name) ifFalse:[^super makeUnderlinedGlyphs].
	"Now attempt a direct creation through the appropriate primitives"
	(self fontName: name size: pointSize emphasis: (emphasis bitOr: 4) rangesArray: ranges)
		ifNil:[^super makeUnderlinedGlyphs]. "nil means we failed"! !

!HostFont methodsFor: 'primitives' stamp: 'ar 6/4/2000 23:11'!
primitiveCreateFont: fontName size: fontSize emphasis: fontFlags
	<primitive:'primitiveCreateFont' module:'FontPlugin'>
	^nil! !

!HostFont methodsFor: 'primitives' stamp: 'ar 6/4/2000 23:01'!
primitiveDestroyFont: fontHandle
	<primitive:'primitiveDestroyFont' module:'FontPlugin'>
	^self primitiveFailed! !

!HostFont methodsFor: 'primitives' stamp: 'nk 8/31/2004 09:19'!
primitiveFont: fontHandle fullWidthOfChar: aCharIndex 
	<primitive:'primitiveFontFullWidthOfChar' module:'FontPlugin'>
	^Array 
		with: 0
		with: (self primitiveFont: fontHandle widthOfChar: aCharIndex)
		with: 0! !

!HostFont methodsFor: 'primitives' stamp: 'ar 2/18/2001 19:46'!
primitiveFont: fontHandle getData: buffer
	<primitive:'primitiveGetFontData' module:'FontPlugin'>
	^self primitiveFailed! !

!HostFont methodsFor: 'primitives' stamp: 'ar 8/28/2000 16:05'!
primitiveFont: fontHandle getKernPair: kernIndex
	<primitive:'primitiveFontGetKernPair' module:'FontPlugin'>
	^0! !

!HostFont methodsFor: 'primitives' stamp: 'nk 8/31/2004 09:19'!
primitiveFont: fontHandle glyphOfChar: aCharIndex into: glyphForm 
	<primitive:'primitiveFontGlyphOfChar' module:'FontPlugin'>
	^self primitiveFailed! !

!HostFont methodsFor: 'primitives' stamp: 'nk 8/31/2004 09:19'!
primitiveFont: fontHandle widthOfChar: aCharIndex 
	<primitive:'primitiveFontWidthOfChar' module:'FontPlugin'>
	^self primitiveFailed! !

!HostFont methodsFor: 'primitives' stamp: 'ar 6/4/2000 23:01'!
primitiveFontAscent: fontHandle
	<primitive:'primitiveFontAscent' module:'FontPlugin'>
	^self primitiveFailed! !

!HostFont methodsFor: 'primitives' stamp: 'ar 2/18/2001 19:45'!
primitiveFontDataSize: fontHandle
	<primitive:'primitiveFontDataSize' module:'FontPlugin'>
	^self primitiveFailed! !

!HostFont methodsFor: 'primitives' stamp: 'ar 6/4/2000 23:01'!
primitiveFontDescent: fontHandle
	<primitive:'primitiveFontDescent' module:'FontPlugin'>
	^self primitiveFailed! !

!HostFont methodsFor: 'primitives' stamp: 'ar 2/18/2001 20:00'!
primitiveFontEmbeddingFlags: fontHandle
	<primitive:'primitiveFontEmbeddingFlags' module:'FontPlugin'>
	^self primitiveFailed! !

!HostFont methodsFor: 'primitives' stamp: 'ar 6/4/2000 23:02'!
primitiveFontEncoding: fontHandle
	<primitive:'primitiveFontEncoding' module:'FontPlugin'>
	^self primitiveFailed! !

!HostFont methodsFor: 'primitives' stamp: 'ar 8/28/2000 16:04'!
primitiveFontNumKernPairs: fontHandle
	<primitive:'primitiveFontNumKernPairs' module:'FontPlugin'>
	^0! !

!HostFont methodsFor: 'accessing' stamp: 'ar 2/18/2001 20:04'!
testEmbeddingFlags
	"HostFont basicNew testEmbeddingFlags"
	| list fontHandle |
	list := self class listFontNames.
	list do:[:fName|
		fontHandle := self primitiveCreateFont: fName size: 12 emphasis: 0.
		fontHandle ifNotNil:[
			type := self primitiveFontEmbeddingFlags: fontHandle.
			Transcript cr; show: fName,': ', type printString.
			self primitiveDestroyFont: fontHandle.
		].
	].! !

!HostFont methodsFor: 'accessing' stamp: 'yo 2/17/2004 16:23'!
widthOfString: aString from: firstIndex to: lastIndex
	
	^ (aString copyFrom: firstIndex to: lastIndex) inject: 0 into: [:s :t | s := s + (self widthOf: t)].! !

!StrikeFont class methodsFor: 'accessing' stamp: 'nk 9/1/2004 11:00'!
actualFamilyNames
	"Answer a sorted list of actual family names, without the Default aliases"

	^(self familyNames copyWithoutAll: TextStyle defaultFamilyNames) asOrderedCollection! !

!StrikeFont class methodsFor: 'examples' stamp: ''!
convertFontsNamed: familyName  " StrikeFont convertFontsNamed: 'NewYork' "
	"This utility is for use after you have used BitFont to produce data files 
	for the fonts you wish to use.  It will read the BitFont files and then 
	write them out in strike2 (*.sf2) format which is much more compact,
	and which can be read in again very quickly."
	"For this utility to work as is, the BitFont data files must be named
	'familyNN.BF', and must reside in the same directory as the image."
	| f |
	(FileDirectory default fileNamesMatching: familyName , '*.BF') do:
		[:fname | Transcript cr; show: fname.
		f := StrikeFont new readFromBitFont: fname.
		f writeAsStrike2named: f name , '.sf2']! !

!StrikeFont class methodsFor: 'font creation' stamp: 'jmv 8/4/2009 09:55'!
createDejaVu: pointSize
	"Warning: Uses the methods in 'dejaVu font data' category, that will be removed soon (or are already removed) to save space."
	
	| base bold oblique boldOblique point |
	point := pointSize asString.
	base := (StrikeFont new
		buildFromForm: (self perform: ('dejaVuSansBook', point, 'Form') asSymbol)
		data: (self perform: ('dejaVuSansBook', point, 'Data') asSymbol)
		name: 'Bitmap DejaVu Sans ', point)
			pointSize: pointSize.
	bold := (StrikeFont new
		buildFromForm:  (self perform: ('dejaVuSansBold', point, 'Form') asSymbol)
		data: (self perform: ('dejaVuSansBold', point, 'Data') asSymbol)
		name: 'Bitmap DejaVu Sans ', point, 'B')
			emphasis: 1;
			pointSize: pointSize.
	oblique := (StrikeFont new
		buildFromForm: (self perform: ('dejaVuSansOblique', point, 'Form') asSymbol)
		data: (self perform: ('dejaVuSansOblique', point, 'Data') asSymbol)
		name: 'Bitmap DejaVu Sans ', point, 'I')
			emphasis: 2;
			pointSize: pointSize.
	boldOblique := (StrikeFont new
		buildFromForm: (self perform: ('dejaVuSansBoldOblique', point, 'Form') asSymbol)
		data: (self perform: ('dejaVuSansBoldOblique', point, 'Data') asSymbol)
		name: 'Bitmap DejaVu Sans ', point, 'BI')
			emphasis: 3;
			pointSize: pointSize.
		
	base derivativeFont: bold at: 1.
	base derivativeFont: oblique at: 2.
	base derivativeFont: boldOblique at: 3.
	
	^base! !

!StrikeFont class methodsFor: 'accessing' stamp: 'rej 9/29/2007 08:50'!
defaultSized: aNumber
	| fonts f |
	"This used to be the default textstyle, but it needs to be a StrikeFont and not a TTCFont and sometimes the default textstyle is a TTCFont.  So, we use a typical StrikeFont as the default fallback font."
	fonts := (TextConstants at: #Accuny) fontArray.
	f := fonts first.
	1 to: fonts size do: [:i |
		aNumber > (fonts at: i) height ifTrue: [f := fonts at: i].
	].
	^f
! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:30'!
dejaVuSansBold12Data
	"Created using:
	Clipboard default clipboardText:
		((FileStream oldFileNamed: 'AAFonts\DejaVu Sans Bold 12.txt') contentsOfEntireFile substrings 
			collect: [ :each | each asNumber]) asString
	"
	^#(12 15 4 0 5 10 16 28 38 54 68 71 77 83 91 102 107 113 117 124 135 146 157 168 179 190 201 212 223 234 239 245 257 268 279 287 302 316 327 339 351 361 371 384 396 401 408 420 430 445 458 471 482 495 507 517 529 541 555 573 585 598 609 616 623 630 638 651 656 667 678 687 698 709 717 728 739 744 751 762 767 784 795 806 817 828 836 846 854 865 877 892 902 913 922 931 935 943 954 964 975 984 992 1005 1015 1025 1035 1045 1055 1065 1075 1085 1095 1105 1115 1125 1135 1145 1155 1165 1175 1185 1195 1205 1215 1225 1235 1245 1255 1265 1275 1285 1290 1295 1304 1315 1326 1338 1349 1358 1367 1379 1387 1396 1408 1414 1426 1432 1438 1449 1455 1462 1473 1484 1493 1497 1506 1512 1521 1530 1548 1565 1578 1587 1601 1615 1629 1642 1655 1668 1686 1698 1708 1718 1728 1738 1744 1749 1756 1762 1775 1788 1801 1814 1827 1840 1853 1864 1878 1890 1902 1914 1926 1939 1950 1961 1972 1983 1994 2005 2016 2027 2044 2053 2064 2075 2086 2097 2103 2109 2116 2122 2133 2144 2155 2166 2177 2188 2199 2210 2221 2232 2243 2254 2265 2276 2287 2298)! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:31'!
dejaVuSansBold12Form
	"Created using:
	Clipboard default clipboardText:
	 	((ByteArray streamContents:[:s|
			PNGReadWriter
				putForm: (Form fromFileNamed: 'AAFonts\DejaVu Sans Bold 12.bmp')
				onStream: s]) asString base64Encoded)
	"
	^Form fromBinaryStream: (
'iVBORw0KGgoAAAANSUhEUgAACPsAAAATCAYAAADsppoMAABsrElEQVR4XuVdB0xTbRc2ado0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'
	) base64Decoded asByteArray readStream! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:31'!
dejaVuSansBold7Data
	"Created using:
	Clipboard default clipboardText:
		((FileStream oldFileNamed: 'AAFonts\DejaVu Sans Bold 7.txt') contentsOfEntireFile substrings 
			collect: [ :each | each asNumber]) asString
	"
	^#(7 8 3 0 3 7 12 20 27 38 47 51 56 61 66 74 78 83 87 92 99 106 113 120 128 135 142 149 156 163 167 171 179 187 195 201 211 220 227 234 242 248 254 262 270 274 279 287 294 303 311 320 327 336 344 351 359 367 376 387 395 403 411 416 421 426 431 440 444 451 458 464 471 478 484 491 498 502 507 514 518 528 535 543 550 557 563 570 576 583 590 600 607 615 621 627 630 636 644 651 658 664 669 678 685 692 699 706 713 720 727 734 741 748 755 762 769 776 783 790 797 804 811 818 825 832 839 846 853 860 867 874 877 881 887 894 901 909 917 924 930 940 946 952 961 966 976 981 987 995 1000 1006 1014 1022 1029 1033 1039 1044 1050 1056 1068 1080 1089 1095 1104 1113 1122 1131 1140 1150 1160 1167 1173 1179 1185 1192 1196 1200 1205 1210 1219 1227 1236 1245 1254 1263 1272 1279 1288 1296 1304 1312 1320 1328 1335 1342 1349 1356 1363 1370 1377 1384 1395 1401 1408 1415 1422 1429 1434 1438 1443 1448 1456 1463 1471 1479 1487 1493 1501 1509 1517 1524 1531 1538 1545 1553 1560 1568)! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:31'!
dejaVuSansBold7Form
	"Created using:
	Clipboard default clipboardText:
	 	((ByteArray streamContents:[:s|
			PNGReadWriter
				putForm: (Form fromFileNamed: 'AAFonts\DejaVu Sans Bold 7.bmp')
				onStream: s]) asString base64Encoded)
	"
	^Form fromBinaryStream: (
'iVBORw0KGgoAAAANSUhEUgAABiEAAAALCAYAAAAZZi58AAA7+0lEQVR4XuVdB1BT2fdmhoFh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'
	) base64Decoded asByteArray readStream! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:32'!
dejaVuSansBold9Data
	"Created using:
	Clipboard default clipboardText:
		((FileStream oldFileNamed: 'AAFonts\DejaVu Sans Bold 9.txt') contentsOfEntireFile substrings 
			collect: [ :each | each asNumber]) asString
	"
	^#(9 11 3 0 3 7 13 23 32 45 56 59 64 69 76 85 89 94 98 104 113 122 131 140 149 158 167 176 185 194 198 202 212 222 231 238 250 261 270 279 289 297 305 315 325 329 335 345 353 365 375 386 395 406 415 424 433 442 453 467 477 487 496 502 508 514 520 530 534 543 552 560 569 578 585 594 603 607 612 620 624 636 645 654 663 672 678 686 693 702 711 722 731 740 748 756 759 766 775 783 791 798 804 814 822 830 838 846 854 862 870 878 886 894 902 910 918 926 934 942 950 958 966 974 982 990 998 1006 1014 1022 1030 1038 1041 1045 1052 1061 1070 1080 1089 1097 1104 1115 1122 1130 1139 1144 1155 1160 1166 1176 1182 1188 1198 1207 1215 1219 1227 1232 1240 1247 1261 1274 1284 1291 1302 1313 1324 1334 1344 1354 1368 1377 1385 1393 1401 1409 1413 1417 1422 1427 1438 1448 1459 1470 1481 1492 1503 1511 1522 1531 1540 1549 1558 1568 1577 1586 1595 1604 1613 1622 1631 1639 1652 1660 1669 1678 1687 1696 1701 1706 1712 1717 1726 1734 1743 1752 1761 1770 1779 1788 1797 1806 1815 1824 1833 1842 1851 1859)! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:32'!
dejaVuSansBold9Form
	"Created using:
	Clipboard default clipboardText:
	 	((ByteArray streamContents:[:s|
			PNGReadWriter
				putForm: (Form fromFileNamed: 'AAFonts\DejaVu Sans Bold 9.bmp')
				onStream: s]) asString base64Encoded)
	"
	^Form fromBinaryStream: (
'iVBORw0KGgoAAAANSUhEUgAAB0QAAAAOCAYAAAC/6687AABTd0lEQVR4XuVdB0xUWRcmmQwh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'
	) base64Decoded asByteArray readStream! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:32'!
dejaVuSansBoldOblique12Data
	"Created using:
	Clipboard default clipboardText:
		((FileStream oldFileNamed: 'AAFonts\DejaVu Sans Bold Oblique 12.txt') contentsOfEntireFile substrings 
			collect: [ :each | each asNumber]) asString
	"
	^#(12 15 4 0 5 12 18 31 41 57 71 74 82 89 97 108 114 120 125 134 145 156 168 180 192 204 215 226 238 249 255 262 273 284 295 303 318 331 344 356 370 382 394 407 421 428 438 452 462 479 493 507 520 534 546 558 570 583 597 615 630 643 656 665 670 679 687 700 705 716 727 737 749 760 769 780 791 797 806 818 824 840 851 862 874 885 894 904 913 924 935 950 962 974 985 995 999 1009 1020 1030 1041 1051 1059 1072 1082 1092 1102 1112 1122 1132 1142 1152 1162 1172 1182 1192 1202 1212 1222 1232 1242 1252 1262 1272 1282 1292 1302 1312 1322 1332 1342 1352 1357 1364 1374 1386 1399 1412 1424 1433 1443 1456 1465 1475 1487 1493 1506 1512 1519 1530 1537 1544 1557 1569 1579 1584 1595 1602 1611 1621 1640 1657 1670 1678 1691 1704 1717 1730 1743 1756 1776 1788 1800 1812 1824 1836 1843 1851 1859 1867 1881 1895 1909 1923 1937 1951 1965 1976 1993 2006 2019 2032 2045 2058 2071 2082 2093 2104 2115 2126 2137 2148 2166 2176 2187 2198 2209 2220 2225 2233 2240 2247 2258 2269 2280 2291 2302 2313 2324 2335 2347 2358 2369 2380 2391 2403 2415 2427)! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:33'!
dejaVuSansBoldOblique12Form
	"Created using:
	Clipboard default clipboardText:
	 	((ByteArray streamContents:[:s|
			PNGReadWriter
				putForm: (Form fromFileNamed: 'AAFonts\DejaVu Sans Bold Oblique 12.bmp')
				onStream: s]) asString base64Encoded)
	"
	^Form fromBinaryStream: (
'iVBORw0KGgoAAAANSUhEUgAACXwAAAATCAYAAAAp8Ve0AAB9+klEQVR4XuVdB0xTaxsmado0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'
	) base64Decoded asByteArray readStream! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:33'!
dejaVuSansBoldOblique7Data
	"Created using:
	Clipboard default clipboardText:
		((FileStream oldFileNamed: 'AAFonts\DejaVu Sans Bold Oblique 7.txt') contentsOfEntireFile substrings 
			collect: [ :each | each asNumber]) asString
	"
	^#(7 8 3 0 3 7 12 20 27 37 46 49 54 59 65 72 76 81 84 91 99 107 115 123 131 139 147 155 163 171 176 181 188 195 202 208 218 226 234 242 251 259 267 276 285 290 296 305 312 323 332 341 349 358 366 374 382 390 398 409 418 426 434 440 443 449 454 463 467 474 482 489 497 505 511 519 527 532 538 546 551 562 570 578 586 594 601 608 614 621 628 637 645 653 660 667 670 677 684 691 698 705 710 719 726 733 740 747 754 761 768 775 782 789 796 803 810 817 824 831 838 845 852 859 866 873 880 887 894 901 908 915 918 922 928 936 945 953 961 967 974 983 990 996 1003 1008 1017 1022 1027 1034 1040 1046 1055 1063 1069 1072 1079 1083 1090 1096 1108 1120 1128 1134 1142 1150 1158 1166 1174 1182 1193 1201 1209 1217 1225 1233 1238 1244 1250 1256 1265 1274 1283 1292 1301 1310 1319 1326 1336 1344 1352 1360 1368 1376 1384 1392 1399 1406 1413 1420 1427 1434 1445 1452 1460 1468 1476 1484 1489 1495 1500 1506 1514 1522 1530 1538 1546 1554 1562 1569 1577 1584 1591 1598 1605 1613 1621 1629)! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:33'!
dejaVuSansBoldOblique7Form
	"Created using:
	Clipboard default clipboardText:
	 	((ByteArray streamContents:[:s|
			PNGReadWriter
				putForm: (Form fromFileNamed: 'AAFonts\DejaVu Sans Bold Oblique 7.bmp')
				onStream: s]) asString base64Encoded)
	"
	^Form fromBinaryStream: (
'iVBORw0KGgoAAAANSUhEUgAABl4AAAALCAYAAAAeJPBpAABCM0lEQVR4XuVdB0xUW9c1IZAJ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'
	) base64Decoded asByteArray readStream! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:33'!
dejaVuSansBoldOblique9Data
	"Created using:
	Clipboard default clipboardText:
		((FileStream oldFileNamed: 'AAFonts\DejaVu Sans Bold Oblique 9.txt') contentsOfEntireFile substrings 
			collect: [ :each | each asNumber]) asString
	"
	^#(9 11 3 0 3 8 14 24 33 45 56 59 65 71 78 87 91 97 100 107 117 127 137 147 157 167 177 187 197 207 211 216 226 236 245 252 264 275 285 295 306 315 324 335 346 352 360 371 379 392 403 414 424 435 445 454 463 473 483 497 508 518 528 535 539 546 552 562 566 575 585 593 603 612 619 629 638 644 651 661 667 680 689 698 708 718 726 734 741 750 758 769 778 788 796 804 807 815 824 831 839 847 853 863 870 877 884 891 898 905 912 919 926 933 940 947 954 961 968 975 982 989 996 1003 1010 1017 1024 1031 1038 1045 1052 1059 1062 1067 1075 1085 1095 1105 1114 1122 1130 1140 1147 1154 1164 1170 1180 1185 1191 1201 1207 1213 1223 1233 1241 1245 1253 1258 1265 1272 1286 1300 1310 1316 1327 1338 1349 1360 1371 1382 1397 1407 1416 1425 1434 1443 1449 1455 1462 1469 1480 1491 1502 1513 1524 1535 1546 1554 1566 1576 1586 1596 1606 1616 1626 1636 1645 1654 1663 1672 1681 1690 1704 1712 1721 1730 1739 1748 1753 1760 1766 1772 1782 1791 1800 1809 1818 1827 1836 1845 1855 1864 1873 1882 1891 1901 1911 1921)! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:33'!
dejaVuSansBoldOblique9Form
	"Created using:
	Clipboard default clipboardText:
	 	((ByteArray streamContents:[:s|
			PNGReadWriter
				putForm: (Form fromFileNamed: 'AAFonts\DejaVu Sans Bold Oblique 9.bmp')
				onStream: s]) asString base64Encoded)
	"
	^Form fromBinaryStream: (
'iVBORw0KGgoAAAANSUhEUgAAB4IAAAAOCAYAAADKStHdAABeM0lEQVR4XuVdZ0xVSxclIRBC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'
	) base64Decoded asByteArray readStream! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:34'!
dejaVuSansBook12Data
	"Created using:
	Clipboard default clipboardText:
		((FileStream oldFileNamed: 'AAFonts\DejaVu Sans Book 12.txt') contentsOfEntireFile substrings 
			collect: [ :each | each asNumber]) asString
	"
	^#(12 15 4 0 4 7 13 25 34 49 61 64 69 74 82 93 97 102 105 112 122 132 142 152 162 172 182 192 202 212 215 219 230 241 252 260 275 287 296 306 317 326 334 346 357 360 365 375 384 396 407 419 428 440 450 459 470 481 493 509 520 531 541 546 553 558 566 579 583 592 602 610 620 630 637 647 656 659 664 673 676 690 699 709 719 729 736 744 750 759 768 781 790 800 808 815 818 825 836 846 857 866 874 887 897 907 917 927 937 947 957 967 977 987 997 1007 1017 1027 1037 1047 1057 1067 1077 1087 1097 1107 1117 1127 1137 1147 1157 1167 1171 1174 1183 1192 1203 1213 1222 1230 1238 1251 1258 1266 1278 1283 1296 1302 1308 1319 1325 1332 1343 1353 1363 1366 1374 1379 1386 1394 1411 1427 1438 1446 1458 1470 1482 1495 1507 1519 1535 1545 1554 1563 1572 1581 1585 1589 1594 1600 1613 1624 1636 1648 1660 1672 1684 1695 1707 1718 1729 1740 1751 1762 1771 1780 1789 1798 1807 1816 1825 1834 1849 1857 1867 1877 1887 1897 1902 1907 1914 1920 1929 1938 1948 1958 1968 1978 1987 1998 2008 2017 2026 2035 2044 2054 2064 2074)! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:34'!
dejaVuSansBook12Form
	"Created using:
	Clipboard default clipboardText:
	 	((ByteArray streamContents:[:s|
			PNGReadWriter
				putForm: (Form fromFileNamed: 'AAFonts\DejaVu Sans Book 12.bmp')
				onStream: s]) asString base64Encoded)
	"
	^Form fromBinaryStream: (
'iVBORw0KGgoAAAANSUhEUgAACBsAAAATCAYAAADbRJd9AABtH0lEQVR4XuVdB0xTbRc2IRJC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'
	) base64Decoded asByteArray readStream! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:34'!
dejaVuSansBook7Data
	"Created using:
	Clipboard default clipboardText:
		((FileStream oldFileNamed: 'AAFonts\DejaVu Sans Book 7.txt') contentsOfEntireFile substrings 
			collect: [ :each | each asNumber]) asString
	"
	^#(7 8 3 0 3 6 11 19 25 35 43 46 50 54 59 67 70 74 77 82 89 96 103 110 117 124 131 138 145 152 155 158 166 174 182 187 197 205 212 219 227 233 239 247 254 257 261 268 274 282 289 297 303 311 318 325 332 339 347 357 364 371 378 382 387 391 396 405 409 415 422 428 435 442 447 454 460 463 467 473 476 485 491 498 505 512 517 523 528 534 541 550 556 562 568 573 576 581 589 596 603 610 615 624 631 638 645 652 659 666 673 680 687 694 701 708 715 722 729 736 743 750 757 764 771 778 785 792 799 806 813 820 823 826 832 838 845 851 858 864 870 879 884 890 898 902 911 916 921 929 934 939 946 953 959 962 968 972 978 984 994 1004 1011 1016 1024 1032 1040 1047 1054 1061 1071 1078 1084 1090 1096 1102 1105 1108 1113 1118 1126 1133 1141 1149 1157 1165 1173 1180 1188 1195 1202 1209 1216 1223 1229 1236 1242 1248 1254 1260 1266 1273 1283 1289 1296 1303 1310 1317 1321 1325 1330 1335 1342 1348 1355 1362 1369 1376 1382 1390 1397 1403 1409 1415 1421 1427 1434 1441)! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:34'!
dejaVuSansBook7Form
	"Created using:
	Clipboard default clipboardText:
	 	((ByteArray streamContents:[:s|
			PNGReadWriter
				putForm: (Form fromFileNamed: 'AAFonts\DejaVu Sans Book 7.bmp')
				onStream: s]) asString base64Encoded)
	"
	^Form fromBinaryStream: (
'iVBORw0KGgoAAAANSUhEUgAABaIAAAALCAYAAACNtvp/AAAy3klEQVR4XuV9B1BUWd7vVL36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'
	) base64Decoded asByteArray readStream! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:35'!
dejaVuSansBook9Data
	"Created using:
	Clipboard default clipboardText:
		((FileStream oldFileNamed: 'AAFonts\DejaVu Sans Book 9.txt') contentsOfEntireFile substrings 
			collect: [ :each | each asNumber]) asString
	"
	^#(9 11 3 0 3 6 11 20 28 40 50 53 57 61 68 77 80 84 86 91 99 107 115 123 131 139 147 155 163 171 173 177 187 196 205 211 223 232 240 248 257 264 270 279 288 291 296 304 311 321 329 339 346 356 363 371 379 388 397 409 417 425 433 437 442 447 453 463 467 474 482 488 496 504 509 517 524 527 531 538 541 552 559 567 575 583 588 594 599 606 613 623 630 637 644 650 653 659 668 676 684 691 697 707 715 723 731 739 747 755 763 771 779 787 795 803 811 819 827 835 843 851 859 867 875 883 891 899 907 915 923 931 934 937 944 952 960 968 976 982 988 999 1005 1012 1022 1026 1037 1042 1047 1056 1062 1068 1076 1084 1092 1094 1101 1105 1112 1118 1131 1143 1151 1157 1166 1175 1184 1193 1202 1211 1223 1231 1238 1245 1252 1259 1262 1265 1269 1274 1285 1293 1303 1313 1323 1333 1343 1351 1361 1370 1379 1388 1397 1405 1412 1419 1426 1433 1440 1447 1454 1462 1474 1480 1488 1496 1504 1512 1516 1520 1525 1530 1538 1545 1553 1561 1569 1577 1584 1593 1601 1608 1615 1622 1629 1636 1644 1652)! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:35'!
dejaVuSansBook9Form
	"Created using:
	Clipboard default clipboardText:
	 	((ByteArray streamContents:[:s|
			PNGReadWriter
				putForm: (Form fromFileNamed: 'AAFonts\DejaVu Sans Book 9.bmp')
				onStream: s]) asString base64Encoded)
	"
	^Form fromBinaryStream: (
'iVBORw0KGgoAAAANSUhEUgAABnUAAAAOCAYAAAD5VJn9AABJ1klEQVR4XuVdB0xTaxu+yY03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'
	) base64Decoded asByteArray readStream! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:35'!
dejaVuSansOblique12Data
	"Created using:
	Clipboard default clipboardText:
		((FileStream oldFileNamed: 'AAFonts\DejaVu Sans Oblique 12.txt') contentsOfEntireFile substrings 
			collect: [ :each | each asNumber]) asString
	"
	^#(12 15 4 0 4 10 15 28 37 51 63 65 72 79 87 98 103 109 112 121 131 141 152 163 173 184 194 204 214 224 229 235 246 257 268 276 292 304 314 326 338 348 358 370 382 387 396 408 416 430 442 454 464 476 486 497 508 520 532 548 560 571 584 591 596 604 612 625 629 638 648 657 668 678 686 696 706 711 719 729 734 749 759 769 780 790 798 807 814 824 834 847 858 869 879 888 891 900 911 921 932 942 950 963 973 983 993 1003 1013 1023 1033 1043 1053 1063 1073 1083 1093 1103 1113 1123 1133 1143 1153 1163 1173 1183 1193 1203 1213 1223 1233 1243 1247 1253 1262 1274 1285 1296 1307 1316 1325 1338 1346 1355 1366 1372 1385 1391 1397 1408 1415 1422 1435 1446 1456 1459 1469 1475 1483 1492 1510 1526 1537 1544 1556 1568 1580 1592 1604 1616 1633 1645 1655 1665 1675 1685 1690 1696 1702 1709 1722 1734 1746 1758 1770 1782 1794 1805 1819 1831 1843 1855 1867 1878 1888 1898 1907 1916 1925 1934 1943 1952 1968 1977 1987 1997 2007 2017 2021 2027 2033 2039 2049 2059 2069 2079 2089 2099 2109 2120 2131 2141 2151 2161 2171 2182 2193 2204)! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:35'!
dejaVuSansOblique12Form
	"Created using:
	Clipboard default clipboardText:
	 	((ByteArray streamContents:[:s|
			PNGReadWriter
				putForm: (Form fromFileNamed: 'AAFonts\DejaVu Sans Oblique 12.bmp')
				onStream: s]) asString base64Encoded)
	"
	^Form fromBinaryStream: (
'iVBORw0KGgoAAAANSUhEUgAACJ0AAAATCAYAAAAwX+47AAB9+klEQVR4XuVdB0xT3Rc3MRJC
2tCUpk0hjDDCCsMow6iIGgQXzrhw770n7j1xfi4+t7jFLS5wADJVRBRcuFDZZe+2v/99r6Xt
62Co36f/rzdpouX0vvvuu+/ec37nd85pAeh/0/+m/03/m/43/W/63/S/6X/T/6b/Tf+b/jf9
b/rf9L/pf9P/pv9N/1vTmlRSjMS13dBtRTTyy7Nwa2Ugukw7g3cVEv2fnP94Exe/Rmz0S4jq
pGQdVCIrKQoxb4ohlkr1f3J+5ztZnYXbu1diw8k0lFZn48H+1Vh39ClEkh9/LlKpGAWJoVi+
Kgwvy5r+bkvKXuHoKHe0Gf03UorEDcrWFSTj2PLZmDFzGU6mlULSwDqSFCVj5yBnCK27YtGl
DLy5MgWdRoU3MH4JxGIxxGQOqH9L5P/WPYdfcXttP7gIODD3GoO9iYV/zLqW1uXg1qzWcJl4
Hdl10kaeWzU+X56Bto79sCNB+z1IKt/j9PjWcPbpCHuX8bj6vU7HXk/mTCyhn4vqv+vPgaRN
XWDfcx8yKpnXqP58FXPaOaPvlkdy2UpkXpgObwcX2Nj2wr6MSvn3smck61/+b8mvmXNpzRdc
HO8AoxZGcBxzEs9TDmGQ7zzEFDWwBqS1+HZjJrw8ZyAip67pa744BfsGucBj4gm8LJU0+E6J
kndjkLszXKxbY+qNbNQ1sMZqssIxwdEILYwcMfrEc6QcHgzfudEo0jFHoofT4cByxKxHRShN
CkYblhXGROQ1+F79060u5zZmt3HBuCvfGrxXxjyJRUja0RfO7Wbg7OuyBsdf8/Um5ndwglc7
OziPDcfX2gb2kPzbGGfDgvviBJQWx2KuMwv2U6Ma3CelkhKkHhgCV49JOPGy9LfOpf43/W/6
3/S//emthf5Pgf43/W/63/S/6X/T/6b/Tf+b/jf9b/rf9L/pf9P/pv9N/5v+N/1v+t+a08R5
d4kjyQZtPR3hNe08Miv1n3CiD01a/Rah/dtiaOhjRG0fDO9+IUgUiX/beCQ1Vaiqka09ilhQ
V12J6jr9W4uSnCsYas6G97oUlImiMNGWBdcF8Sj9SQeptOYrbs7rBP/1CTod3YxxlL3E4dEd
ELg2Et+qG3LuS1CWfhpzAodizfUMZEauRA//5XhUINYhX423B/ujx4Z4fEzYiT4WBmglDMC2
JyU6r1EavwhuLBtMuJOPsufr0c7YHMOuZOuUL4mdB0/vWTjxIA53/hqJNn7bkFb55ziY63Lv
Ym67Ltj4TLcTXiqtQ07Ucvi1m4AT6drlqGd6Y34XdFt8HZ8rChG7ojN8l8dpPF+pOAdXgyzA
9lqNZ2WFiJpsB5bLPMSWKOUk5a9woL8XRp//oviu+tNlzOncFXMvvkM5RfiRVODt6ano2m8d
Ir+WIPPUGHQYdpyMtQoZu7qAKwjEyU8V+HCsF/gm3bD/TfUvma+K5xvQred2PPnwCBsDTNHK
wBpDjrxFVSPvhFScjwfBndBp+eOmrfnip/hrSHsM2BqNnNqG1rwYhXEhGNh5HEKfFSI/fgO6
d1qAqHxda74CqRv90SMkGR+iNyHAtBUMrAbj8JtKne9UdUEmXr3KRF61BDWiD0h/+Q45v/ls
pu47//5idPRdhcSSxsdCEU4SQ/qj/bBdiC9omPhTmxOJ5QFdMOvCO7LvJWB9Nx8sflio8/2Q
VOXh/atXyCyshqS6EJnk3+9zq3S/T5ISPN8/DO37b8GjnBr9VwAafM4SlKYdxzQfK3D5jui1
4gY+V+s/QUf/m/43fWv6TzrR/6b/Tf+b/jf9b/rf9L/pf/t/NOjEeYic5w5T321ILdd/Q07/
m/43/W/63/S/6X/T/6b/Tf/bv6yPEgdDTSlxnBUUoVL84/poxav9GD/xb2RUSFGXF41tE+cg
7HWF/k/wH/zcy9IOYebQgRg6fR8SC8W/cSxi5EZMhmfPvcioqkN+9HoEBgQjshkZCvS//b+s
uzqUfn6HrDKx3LmehSxRdcMZEFSc7nUln5Dx8g2+l+t2epcmLEHnvvuQIipH9qPV6NZrF9Kr
9NfWprKT5Lx7ifQPItRIqpGX+QqvMvNR/asykZD3t/RLBl6+/ooysRhlX1/jZcZnlIp/Tf+1
ok94/7WMzvRCkQ2+fCtGrUT6f7TmayH6+A5f5Wu+Kj8L34tr/8hMG9K6MuQXV+v/RvWLmqQs
DQeCfNF/azKKm7BmxYWxWO9nj46LI7Ssoyq8PzsdXb37YlNyyT+2fqQ1H3C8f1sM3B6B+JgL
WNrVE9OiCv/9tVididOTu6D7knvIETc1s04Rnp9chSXbb+FTlUTLHNbgW0Qwenj5Y0lUXrMz
YFGZqm7PcoVpl+1Iq1Aj91V/Qvi0NnDot1Nx1uXcWYC2xlz4bn+Jyt/0vkurP+PiDD/4z7uJ
73VNG4OkPANHhzvDbewJZJRJtO5pOZErEejdFfMishvNfERlv5JIJH/mnlf5Ficm+KHfutgm
Z5SjsrXt6O0Ar1nh+KBtnUnKkR46EJYsW4y7nvNLM63V5URhqY89/FZHNUiO/BWthe5DrRzP
VnuBbTEcN3IlyDzUAxbea5FSXoVXO3xh0XmnVqMjL2IU7FovQWKZlCzMN9jfzRY9j2SipqHU
eDlXMHn8GWSpLF7agHkTjuUD2sLc2BhmnqPwV5KIMdFSSSlSdgbCI+go3qixjqmNIDfhKBYN
9IYNjwWWiTU8A2fjaJqS4Vt/TxoPID8aq7u6YfCh12oL6RV2dLVHH3I/1arjINf6cGwYhhxi
fv9bF31VBvZ0s4Dn8nimsp2bgCML+8PT2gRsExt0GLUV97K0KwTVn29h7/LpCOruCRuuAVrw
e+HohxqV+67G1+hjWDdjMLq4W8GEZQiOeWv0nnMIifm1GhtK9v0QTO3fGa1tBGAbGoAtdETn
0RtxI7OiwY2DStWXvM0fgpYtwOt5hPmsvp7HAGELtGih9mG3w4bnFWr9lOL1zZ2Y3b8dHMw4
MGIL4dghCFvjZIef+Hs4Bptq6Uv+YXmuYmwAmXd3YUagJ6xNWDDiWKB1r1k4GJvDWOtScTFe
XlyDkb4OELBZ4Nl1wqjNEfj4fxr9Qz3znMTTWD85EF62fLAMjWHhORTrIj43yhL/uevW4PPZ
IFi1agE+WQOZNVKGgfLt4iCYMp5XK7AEDvAZvhbXMis0Nu+PUQexOKgr3Cy5MDIwgomlO7qN
WorDDz+hQn5IaOu3FcsE5k7t0WfSOoQlZDOMLO3jUPmt3ZTG5QwcMTO6WG28VciKPYFV43rA
g353jMAxdUC73hOw8lAk3paImf2Stb/+uSY4V5d1Bv0ExvDZltYk+fr39nNYHwg4vtj+qkrj
d6rvGKUQFSTsxmA7NoS+S3D9Y6Xi+6yz/SE07ogtLyq0GpiK+eB2wa6MKk0A6fZ0OBlS8+OE
2bHFGr9rZT8Z9wqY71Ttp5MI5PPQg+zLzPmswOeHh7BsdDd6LzJmcWHh2g1j1p5HqqiuwefZ
0oALy7Z9sSgsFUVqSitliH+JPoIVY+r7NYGFmz/Grg9HmkoqWa39GvFgS/aiNZffokw9qoSM
91PUASwc4gtnsm8Z0uvVBT59J2HjDfLeZexGF67ufav+Yzb0svz6JYid4wSW/TQ81JJOtDRu
PlxYdph8X0SuXYi746xh7LUGz3QQGqrfhqK3kI02S2Jka4WcE8c+MJn/FS+2wse4BbhddhEQ
UMowetJ3dYGJoBcOZ1Yz9urEFZ7g2IzGFbX0rtVfLmGqizFMe+wgxsxHnOxvBjbROR4XSzT2
qo/nxsORY4fRp5RrQJJ/C2OsWjY8X9yu2JOhOp5c3BhhodhbHKY/Iv2XIXGJOyz6MXUXxW9K
47HA1VD+G3MMu5rDPIu/R2LrzGCEvSr7acVZnHsHc1rbY+jpz836XfXrvfATaO45vxKILoyc
CBu+Pw68/THgoTY3Fnsn+8HZlI1WLVjwWPGk2X1Upe9EZ74z5saWNChXnrIW3nwPrEgua2C+
nDDroajZfdfLzWlETv+b2toWPcf5faG4+aFS/yejQSO8Au/Cg9G3jQU4hiy4LYyHtPghptha
IOhmHnPv+XAMvQQGMCb6emDwJWT+AIBeTkWIsnXvoSYBoXinFs0krfmOe8t9iS7fEhyP2bj0
oeqn975ftY829X70X07/5fRfTv/lmivXkE39Zl83mPB6EB245qds8+zYg5gd2BZWXCMYca3h
0X8l7nyv1QqkvwzfgPH+bjDnGMmwmFGrce55oRwAjcYcZxO035iKCrU9tfr9MQyytsKAI++Y
OF5VFmKPLsOobq1hRTA3A4INCGxao3P/Kdh6J0sNtP6Gh39NQ4CrKbFn2TB17oxxfz3VwLB2
+7fBqEPnsbZXO4w7namBMUhrC5AavgmTe7WFNZ8NNp/cc6/J2HTxOQpqNTHBr3e2YFw3AvBT
WIxjDyy5/OEfxS0afF5NHLu07ClWebIYa4rX45Astf+x6fC15sLEvjuWXMpE5Q84Tuvy47Bv
WgBciQ3HNvXAiF3xKBBLFfp76bOdGNHBHnyWEfjOgVgZkcWwefO/ZCIzs6FPFgqqJD/orMpB
4skVGOXnCguuIVrWz4GwH85kNUwgkZY9wwYfF/RfOhc9PIfj6GslziitzMChScROtyI4oP1U
PBAp563620PsJvaGI8FKTOy6YlbYK5T+Aoe0uKoYud++4OOHhubqA02Y+Hkn31fc2ToPcwnW
mFUj1bDJKjOvYMPcedh846NWu3WhKws2EyJ/CI/7T8rXZuHiSGtifxKsyTwQO5OLGnT6SMja
29yJ2ygmY+J/4Jdhk83GGv8NORVsURfe2NzvVdexKGoSbFuq4mhcmNp5oseYZTj0QDlf/wZe
+F+Rb77dqY57kv7dAzBu3XmkfMvE9UWd0HZc87HHP02+uVijYr0Zd8Cm1MbJspLSDFxaOxKd
7IlfxYgDS4/+WHL6BYrFmvu36PFKtLfpoShFpbvPVOwd6A6/lZrOZAqXz40KRmfvUVg+qxNc
h19AlrpM5TucmtoOFubemBmeibyn29CjZwjzvfkchj4CDjqFvERZzlUMt2DDa/UzNX3vNfb2
7IDZ5LwpKXyBw4M8MfpG7m/wXxG/8otdCLBujzVJjZNspOICxG3tA48+G/FQSzYbqr+i5BD0
9hyI4CU94dLrIN5q0fXp0k3WBJNZFI+SohjMdmLBYdp9BSGhLicCM9wcMOK8im5Vm4PIYF+0
HnYQqSUS2dhfHsAgjz5YsXMqPD3mEr1FItcP2TAfdhnZZenY2ZkLQd8wtfsQIW49wV3Me2Nn
bDqe3AnDjhWrceq1um+sDOlHxsHD3BKdFt3A55w4rPbvh9B3mjpJRcYBBNp4IjhW1Pg80uQZ
L3hNOYf3lRKt81j6fA/6e/bF4uA+cPXfhXQtmcRqs6Pw165IfHwXjimubLLnG8Eu6Aheq8lK
RPcx1Z4Fx1mPUFSagMXuLFiPu438f4lYSN1PSfJmdLbugm2pZU0j5IxoQ+b8JrK0rB+abHRv
MXzaTUTIuj5w6al9nYmzL9NZ7TxXP0VZ4T1MorPaxaGkgecjFiViWy939Noci3wtfgwZbu3U
ZNy6+s1+dDPhoXvoe5R/OY1+Qg58t79SEKR0k07qsnCmnxAc3+14RQDJppFOyvFsjRdsiXMr
m2yUtHPHvjWWJDY86eXPNmLEqkSUqUyMOD8K8wh46jP/OCJj7yF0oguEHbfiRaVU+RBuz0eH
jgtxV429LtuUV6Ej3wxd5v2NiNhExJKXbNOU0dio4kzQRjqRVr7GkSH2cJ4QrkVZp8glvWHl
vQZPy6QoywjH4StpENWpkE7qivDq2mFcfFX2G4kBYuRETIGTzRCc/qwEHGq/3cCctjxY+C3A
oVuxiI3Yh/FuHAh67GM4AhX3GjYGXQOGYVrwckzyIM5Ot8VIKFNR4qpeYWePNvCftAah4fcQ
Gx+DmwdmoD3fEGZ9DzGAFspBl7p/CZbuOI4rUY+RnByHe2HL0dvKEGyvlUgqleo0Dr5enY62
Ng4wNTIkm3aCmlKwAK48Tyy8lIiUlBTl58U7FFRLVDa9LFyb6wWhVTfM3n0BkY+TkRR7G2d2
bcLRl7JnJanIxuvUFGY/Kcm4uaYzeIa2GB2eJR9TJV4fHgJrtjm6zSfOkJhExEeGYVkvSxgR
59quV/U1KUvwJKQ7UTrsMWB1GO49jsPdw7PQnseB57LYn6qp+tvWVkUqtvT2w+jVh3HtYRy9
hia1NkYryyCEf9OuwNYWvkLEweWYEOgNe1PKac4CnxxWgZM34WJKfoOENMWmXHAf89zYYHFZ
MHSYgehi5tqinL9sm+E4Fit/bk8e4/bhuWQPaAV+9wMqa/YDLs1pBx7LAr4TN+PkjQeIS4jF
3Qt7ML+3A4wtgnAtR8Ls12oIDtP9PsOTxGjcOrsHCwLtwTayxsB9qQpAQiFvHaQch8on9XV2
43KprxkpD+l6vCu6wcxIAO/Rq3HochRiExMQffssdi/oCycOBx02pzL6ZVmORkS+RLsCbWSD
CfcKmyQvkylG7GwnGBFj8r6IeZ8sq9G4Jf8dpaxknJkODxMOXMcdQWqxKsmiFPELXGFkPRZ3
CrQpHPL5YPNgbEjGFylSA4+eY4uPCf3sWwn6IEy+pynGQRR0IyNHzHjAVH6KY2bBke2KBXGl
KvP5GdcWdoSAY4vuM3fg7K1oxMc/wtUDc+FnbgRep1WILhCrzU/9809B8qOLWN/fBoaGLphz
v5DR7/VFPhDq6JffeS1iC9X7HYzDMbJ1lRR1Csu6m8OA7YnlCcUqSqAISdt6EGDYDoGLD+Ly
/XgkJ8Xizvl9WDYiAEGH30Fc+hmvnivXUOyJUbAzNEWfA49U1lYq0rPK5GN9i4P+JuD3OooP
GmdcNd6FBoBHAPBDBACvJ4VwzYbiSo6WZ1eXg5uTHci7NxrhX2uRdyMIZhRxKL2KcQ+P5rnC
kCLuqZFXJMVxCG5jDNtxNzSY59WZRxAo5KHr7nQFMCwpfoKQACGMXWfgalaNjNTwYC5c2NYY
dUlZF5aujxu3Hp0FAviuewyRSt+Sqly8TdV8P1NSEnBlRVcIWrHhPvMKQ/mTKfY8uEw9i3gy
l6+/V8jm6mAALNpvRGqFFkWxtgiZL8jzOD4StrxOCHlZqaI830VwRx4NohrZj8KxjJ9zmNL6
0K2ZaOu5oFmKcf6t0bAS9sXpL3X/kC5SgdRNHcBznofHJdIfAOazED7Kluh+CxH2IBnPyNy/
y6tutgGQe2MELEwH4PzXhu+zRpSJFy8yIdLB9q4r+YA0olvkVUs0+j7XAAiukGsCWK7/Te3c
SlgEN74PtqXpP+mkQWPv9T4ECE3RddkZPEh6jjc5VTpJJ5Lyr8h4GoPzSzsRXag1gpPKfoik
0b3vLkQ909xLkyPWo/cAprOVAh6uzvMC39QXC4+EYRXRl7nO43HyJ/a+X7mPNvV+9F9O/+X0
X07/5Zorp9tpU4C7E+zAabsUSWW6cJBKpG0fjWVxJTr1Cwrz6mDCh/fUvbhOYRJR57F9+XbE
iyQaOvu94A7gc5wweO0Z3H+ahtTHV7BjvCfs+55U6Oz3ZziC57OVUUZCWvUWRwZYwWrgUcY9
0VF6fSzANidO+r1XEZuagdfpKYi5fhTrJ/bHlPAsBrD9ZIMP+OZ+WHTkNmITYhBxMgTrj6dr
6LRfLw6HpQEfXTfFM3R4mS3wHKEjnGHeZgiC91/EPYLzJRKsMPzAMgzzsIDz8FCkqDiH6OCl
A5uw92Ik4pIe49rWAbBzmoi7Bf9+AFBzxi6tK8GHtOcq6+o50j4Uo+zJGnR0HYStVwmWd3Ur
BrUOwPaXzdeRKt5cxP7DV/AgPh63946Eg7kyuIyyo4rSIxERFUfszgc4ObstzMiaUOC03y9h
iFkjAQ9qwWpNnqOKNzgx2hEsRV9C9N77AAnhBPf1WUVjsI0SC86Ph5d7b6yPKWAGLNaKkPn2
M95eHgHXHn/jvXwtiwseYam3GdzH7MSVR9G4sqU/rIVdsVstEKapOGzpO4JdLx0Nf3dzsFo2
HhjSooUFhl/L+QV2VzUyDwdCaOyF1WoE+rqcSCzx5sG8104801LSojrzMHoJBegR+u6H8Lj/
onxdzk1MdjSGiRmX6JnaHdTKua/C64O9YWrqj803rmKNDxecDitxI1n9fJC9x03CJgN1YJM/
gzX+C3Kq2KIuvLG53zPOxK0+4Ah6YFfUM/r6Twk2di88FMuGuoNrKES3DbEoFEv/FbzwvyDf
7H2m+gtuLO5E9hkr+E0LwamIR4hPiMaNQ4vRw8oIBmSu7QbtQbK8zFpzsMc/Tb65WKNivVmP
xrXPxSgtLUNljQ6cveodjg6xg7X/Avx9IwZJybG4sWcsXEzcsDC2SIvOWIzEdZ3hNkqTKKLY
4yrfIWx8R/TZpOlMpp38qXsx1G86zmdWQlz4EAva+2Frajnj/Mq+Oh5dZ0bg3ZO96GdpgJbG
bTD75nem07osC+mpafhQXEfO7O/ISH2B94U1aterRfbtYHQStKKJAjYDduFJseQ3+Ucr8Gpf
bzj3PoDXDQTbUKSP+2v7wG/Gaa2ZOWiHfOYZTPYbjb9flJJ5IH6Kbh2xmDwvdQxCUpmDN2Re
3uVTpZvy8e5FKo3PKEiwRNfNvjEVrb2XKEgiCVv7oMO440iXX7su5w4W+/fFpseFENd+xpkg
Tww780Wmy5D+Mr6VQ0J0xY9pqXj1pUwF787D24QInNw8Ae3MjNCS7Y6pRx4hI6dSY5y1Wecw
wn8Jot7HYVsPU7RqyUW7JVHIE0u1nnNvDg2AS8AuvKxoeB4jl3ZD5zmX8UnHfFd/DseMbkHY
96wY4oqX2NWjPeZEFWiMr0b0CZnycdeVfkXmxzytRGsZxp+Kt1R5rJpCvCfz8zq74l/NikLZ
Oc+3B8Bl0HFkNmT/UYScGZ3QfUUksmukWm278vRQBHWZjNPvyL2XJmOtb0csjdckTdW/f5mi
GjIH1Jojc5CnO+ubpOwV/h7eHoN2JmnYV/XXzrk+nMatTzcRt6b9Uakv8alUTO8NGWRvyCxS
/lY36aTkMeY5s2A35T5E0iaSTuqycLqfLc14o1gt5c/Wwtt2CC59F2uZ6GzEHQ3GcL82sBOy
YcQxg617R/SatFvGEAwfAlvPlUiWGxZliUvgbjMStwul9ERUZBzGUI++2JOqCTJKxd9xaZg5
cTYrjQltTZ10Qi2SlBA/mDlOwOWvtTqA3b/gZ+aBlU/KUJ2TjNPLh6LbwAXYMNsPXWesx8JB
3TBkaRiScqq1zGkMZjhY0gDwP7n4peSl3dHFHN6rk5UO8ZrPODvcGlzPpYramDQ55epIWHI6
YHMDTExpZRq2+XAgHHgBX8WqWTzKUVDEXNDUQZu80hMskwCt7Dh1pfFlSCdwiOxBLbI0sy91
D/q4dsfaYyvgwxFiwPmvakpBd/BtJuBeoaSB65Tjxa7uMLcZyoh2aNILVBCF+cQ4MQ0MVTDL
pEXRmOXEgv2kCIajlI6C5iqZ0pWvdqIrT4gef6XR9StlYylBfHAbsInDPiL31x+6UnEZsr80
L21ZXXEWsoprm9Z/XRkKSmsZB0/GbuKUJo5mdcNcWifC02Mz0EFgCK5LIGZsOoorkbFIfpqM
2NtnsHNOT9jzHRG0/1mD9TGp55e6zZc85yDsCiGGtAkzap5iIB/vzQev52FmBhRJPiJGWKCl
zUQl2LawLThcD8y98hFVGnVKRUi9G61wNtf3qz1ilzhDx9vDkN8DB+VjUYyjx6GGD5mmypHx
xK72gQnbBZPPvVOsIYbi+PIazqcUMfrldtmthURWTe93PJW5a0heuU9T2apM6AxD9XOr+F3X
PciopjJafcW9NeT9Ylui96YHGgentCYTh3vywPXbi9faWJzy/vidZ2BMG3N03ZOhyBhFA0en
h8Lathem9LKGcZulijOh/neCbkswr5MQLvNiUKT4nTx6UNhf4Qymn/8iT3CJA3PZ3W9q2aoI
aBu7DB5sY3SsJ/HUP3//AwxWafWbfehmYoTWwUnyfnNxb3Fj/XLgs/WF2vz9Rc9ffatM20b2
OBaDJV75MgS+XD6678/QSKVH9S3RWBO1+KSWmUZD0SiMxAQbFtyXMIme9WBz1ERbsFwXIl5O
BBRFToCNsTfWppRrrr/kdWjPIePbKxtf8cPJsCVA15pnStm6r+EYbm0Bn0GeMHWeqyAf0OfP
jXFEvg2CtShQ1Pt7b7oTOK7z8JBidBMD99JUFxib9sTOlFKlwl6Vjt1dyRr1V66vqszTGGPP
gcO4s/jQhNqhtFF0NxjtTFhwGBOGtxXMvbkui2LsyrKVKA0FirQxEla2kxSELG3AZ+bRXhAS
mSi5TG1OFFZ0EsDE3gPWJmZwcxGC4zgWYW8q/mXjqxwpa70h8F5H9Lp/KK2m+CsuDDSFsP9Z
rdlgGnUqfw5DP1NLDL/64+kEKZ3kyUoPCDps0koO+rk5lPfdbgOeN2R4yeX4DWQM0ocmlYhR
04y6+vT7Q84tATnDI0US/WeWNAB0p+/sDL7jdDr6RfG9DtJJfSt7sgIeJu6NBgboImn0GnMN
OVp0NypqMGiwUm+SlL/F6UluMHUbhyOpxbK01uS7C7O9YWo3BAdSipttE/3qfbSp96P/cvov
p/9y+i/XXLmf2t/rvuL8hPE6ia1U2vDDvche2GUnXlZKG9Rzv10eDztjWURuFSMrqwgZKd8U
um3BvSmw56lmuCQg89/9YGk9CMfeVzNA/NTNHWHMao2FDwsbj3qsSMXmDlzYTIyiscWG77sc
hXkijTI9FBn50iRP+MwLx/sKiRaHy0dcX9wZXhM1g8fovxM9pDbnBoKcG8epfvlZ/ZNjV+KX
/eC36iltw1H65dNVXdEn9P1P4Y5VGXvg32YibuVpjquu7CPuLPOB54w7TbjHHEStIJgN2x6j
jqZrZO9sit738WR/mKkSNTgUcb+C2Fxj4DXiutb3rbnOp+cbfOAx9zEdCUpjWXv8YNNtD17V
BzeWJSG4tTWGXc9tpt0jQvL+kXAheGCbYatwJCIJb76JUEGcjmVUNkVjcwSp2JL/RJMQ7HK2
Cxu245UBHXTke18LmHgswK3s2h+2GxvC4/5z8pISJK1pB57TZBzZ3x+mJt2w77Vu+ZovxHFn
zYXXygQUF8dhvgvBFGZEo7ihbN4KbNIT86990oJNFiGNYKfq2OQvwxr/ITlVbFEX3tjc75Vz
ko1LQ8xgrMWul9blIXJha7DY7bD2adm/gxf+n8vXEH/W/dRCrTiLpPQNHj3+zHDwUnMcFewN
E157BEdkMfqXFD/Fnj4WaNWCDe+1zxQ4Y3Oxxz9NvjlYY/16Y5AKW3Hh4D8bJ+Q2sOJZPJgF
F7uRuCzP6lxX/A6PTiyBr8AU/c9m/SZMgTi7v7/Hx0LKxydBtegLPudW/AQOJ0FtyXd8/JyH
CrH+lwH/t1pl5g0cPXkTCW/zUZ5zH0vbCeE25w5ytOCx4rKveP+pGLV02bF8fP6Sr3EW6X9T
19ezkJqe/x/FFQluvcIDPM/VePqLcGudpJPqd6EIMOGh52FZaZwmkU6KY4mS64SpD0Sykgqn
+8FGLVKi/jC8Nac1+A6DsenSPZyY2B8rbj1Dws3DWBt8QBEdazsoHN/FUnnZnrFwaL2IzrQh
LozBqi5emHopS2uWBCot564uXPB7n8DH2qaTTipe7oa/0AIDT3zQmX1BWvQIU+0FCJCzwCnH
WWHcZgSYtkIr0wBsflyoc1P+N0gnNGPu+gQ42g7HhaxaxWZfkrAMbTm2GHvlO6NWVkXqZnQw
tsCoO7qJCpKcawgyJ8rDuhSUNwYSyLPdsHUQSZiHehFiF7jBmADlUVpII3U5tzHPyx1jz2Yi
68ZIWLCZzkyq9EPkRBvw1ZzCGv0Qp+cIa/LM9r9uVvkjqsTDk40dweX6YNMzJblJknMZQ8w4
CnKVAsB/tg7tuHaYdDdftoY3tAfHeiwi8iQM0EdWrqQzdqZX/aJnXoui99E4vWU6ervyYDH4
UrN+X/xoKhy49vAbvwZH771CfnUznEKVrxEaaAqO9xo8Uc2CU/sdd4I7gMd2wNAdD5ClNaVW
HfJjt6C7tSum3vius4ZbzccwDLbgoXNIKnKSgtGabYtJUSKVfSeavFcsuMyPQ6mqYl2TiaO9
+OB03iFjKT9cCHcWD122Pm1SutT6fp1mx2o1GKk6r+5GfPQ6+oEh7zw3tsF0Vk2Ro8ZbHL8C
XsZcdNyQjJImj9cI9tMeKowpZX8iPJhqByOnOYitd/o3IK8ELu7K08IlKJTy+t9RJUYKRU9x
cIQTjPntMTdckxgj2zMfYpq9kU7DW9afMVovuY7TQZawnaQEIyWiR1jY2gxdN17EVj8TmA25
RGfRUv1dm6XRSCRnksB9MeLr742KHlQpC0Pt0/mRc+DKEqLn3pfax1mZhq0+xuDKzzbV+1Sd
H5qAIDCkDRUanI2aCzd2E/olxrOyX0PYqoGusjSFRuR+khXv9acTgeCzPLHqaVkT9wJZukgj
FZKDhhJKk1tMMSj8m8ZZpSAYDjinIAlQQGRXrhmGXmGCZBTQfawfMfo9luFxkez9LqPeT2Nn
zH1cojDmX+/zh6nLHFwOnwBbsh9G1mfMIb8/EigEz+8vnaSnihfb0Ik4FAeeSEVcSACExm6Y
ef0rs4QZ2Ue+Xx0DG4q8EleMOqIfrOlEwIWum5DQBEc1fYbHboAfmXurAfuRWqr5m5LYOXDm
M7OV0PebvBxtLQNx8lOtjr5FuD/FDgI5SEOdCy92+MPGez5uvLiJUU5dsCMuFjv72MJ91l3G
mETJYQi98/mHSvZRayf38d9YMrI7PKkyaFSaXhtvDF4Rjjdl9RGUWTjb3xTWo87h4fnVCOpg
S/ZrHuy6TEVosmYpwzcROzCjrzdsBWwYsk3h7DsCG+5+U+4V1Vm4v2cquruS84BrAc9hm3A3
/T7muZgQ4IGpO0jrCvH8/DqM83eDBYcFY/M26Bd8Fq9K6jMn5ePWaEu1KEDqzE1jnLmSys+4
t3MKAlxMYcyxgMeQtbjx7AYBEgS07io7iz4jrI8pbMaFI/bSBozp4ggBmw3TNkOw5X4ObWDJ
1uNr7PXj696naBBBAOuxd1BQTyKl+xbCetxdxXfi4lQcn+wBU6Lf7ozNZchZjopAnkTtepPC
ce/YQvT3sAKXxYVd1zk4/boMNQVPcXLJQHhRpRB5Dui+5DI+VCnBj52d+XCecQ0PwoIxpIMd
eCzNe5KNOxeJx4IxtKMDeX9Y4Nl2xMhNmiX+qPfxUA8+rEeH4/Hl9RjVyZ44DTiw8ByGTfe+
MkGr6q94uG8GerqR95+cqQ7+cxAWew2T7PgIOPhO7f7OIyJ0DvpQZV+M7Gn7QHWd5iefxNJh
neAoNIahoTGEDt4InBtG9qKXCOnE0YgGtRp9S3423cFYaz5N8lXdD2oyj6CnwBrj7xbK510A
QZ8wfFazBape70cPUysMO/NZsW8Kuu3Hm3pglyJcBvDQgt8TR+ojo4gR9mx9OwjJuaKwld4c
JNczIfrJC0UpAoqIGLWsAwQWgdj1rFjhOKQIWHw1ciy9z6XtRDdTe4y79JW5txhpi4Y1pFNm
MgHY7wgfbAazwTKbSXm26iad0FEM14Jg/oOZjprqbJWUvsChEc6w771JI0UttQ8k7R0GZ7ve
2BZX0CyH0a/eR/XfWa7/cvovp/9yfwrphNJrRakXsHaMH1zMOTCgyweYw3N0qKaNTJVL45vA
/8DbBvc5aflzbGzPgaBX42OiUoJPILpjfRBAZUYo+lhYYdCx92q6gCxauIXZMFxvQiCNVJyH
u7NcwCbOsMVX3msNpqjKuo+9s/rA04YnKzFh5Y6uo0MQVyhLM14UGwz/oGN4V0WcIdkx2D+j
J9zJHBkZm8Gt+yyceF1J6zNhI/2xKFqmb9TmxSN0diDR12VzSZ+lpoMQ/k1LUJykGuVVyjOx
rqrslzlylGOngr6GwoweS0sY8R3QZeI+JBTUaYxdWyu4OxltfFfg7qcSlHy6gxWd3DEuIveH
cEc6gCD9NGZ29cfCa1lqz5fS48jzbWFA7KN9Wu0jJuEkG/eW+RDd0QGjj2doloslGOoMB4Ln
7ExnkJ7EXy9goKmM2C+V5ODaMHO10h2TcV9Ujufr/RCghr0xbbKlaKNLb1oYr+a0dkGfsM+0
viyt+4JTfe3QbZ8SJ5SWP8Uqb2dMjBIp5qm6TBloJ6kuQ5VGmd1SPN/dG2bk/kcde8W4f0Wp
YYq48KZxspO45BXC14+Fn6s5uMQ2s2k/DGsuvyF9ilH86jI2TQiAuwUXLMr2GbgC4a/LVIIi
KBJNN/D4AThArkUFbVye5gqu9VAcztAefEfZ6xl7usJER1bapuBx/zX56rd/I9CU4MdkXRQk
ryCOeiobr3ZynVScg9sznGFsPwHXiAO5JvMo2ZcpjDBTt09BBZv0C0lpEkHrV2KN/5wcE1vU
hTc293vF32lCGIthezNwrpch6MQxVmaC/gfxwv97eUkZnm7wAc+8B7bEMX1YlP0WOtQWHNe5
iCpQZngtuD+f4J4C9NiTxui/Lj8WW3o7w2fGRszy4hD8fSfS5bhFc7HHP02+OVhj/btiN/Iw
7ickICE+FrdPBCPA3ACsNsGIK1ZmAEpe1hYWvY7hQw3x1aUeRFCbdph44D7e5FWgThGw3ARM
7z8hY0jOKCKz/CIyiosJ7rgd03pTgSUssIUu6D7rEJIL6zTO3MbwycYwx98pp4o9NoRFNvd7
mT6WjydhyzCsoz3BPckcslqhBastoyzOP4In/uFyCn0xcTHcDVvAYtTdBknwVKnH+CNLMKS+
X7tOGL3uEHZNaI/OS+41GZP8nXLNxSCltZ9o/NQi6AyiL+nGhJvTbwudStC9CbBhuWFRQmnD
yg8xOA9QBmdjaQONlCmcq9//je58K4y6QZxYBNg9OHouI4qRZrkfD4SFyzRczixGUWYElvlY
wHtFIkqqPuLceE/4b0jQWZ6ETg060wkGRq6YFZGt05nN3Bi+4+pYW3A9qVIvDWTNoIwQTxO4
L05ETf4znFsVhICB87F+dld0mbEeCwYFIGjlWTzN/T2ZTqTlqdjWyRztNzxVOomp8gZznMF2
nKFB7qhI3YD2bHMMj8jT2WfZ01XwZFtiZETj46aIIrNcyMHit0uhaDCADGIs1dXVoDzvLR6G
TkJb89aYevGjBlAiKU9H6ECiuKyIRoFYFpHNthiBmyqghrTyBbZ0NEYrjjnMuQZoaciHY9dJ
2HVf5WWgI//7Qijojs0ndmFGD1cIWeTQN2uDgSuv4H257mddRZwJvYRsuC98gEJVw7HmC86P
JmvFZQKOJn9HWVUpvqWcx2JfK7iMO0XXx6eN5qHEIeOzjZmeVk6oMqZKUGRUKRyTUZNtYWA9
jpHqVZbJpQcEJr4aqeVp4zc/HZHH1mKCnz2MW7YCz6UXpm4Kw4O3xeT6EojFdWSuG/6IiZyk
/DPiL+zA3AHEUWbYEizLjhgefAA3U743WCNYUpaBsMnu4FJpJBNEKszdCqTvD4Q5zxNzr3zS
ypSkxi8Wk2tTGTiO9oWd7zYNcpoCoJrpDI7jNNzKFcsd/hwGYCHLMCMg4MEHVNP3VY2y7DTc
3NgXNgIvLL6XI9s8iYPVkPRzN79ppBpZv3wEnvzEOHRViXndecrMFPXyvY9lokp9nul7lTYo
pypDbeLhQZYwtBmHGzniZoyXB/+9qSgoIUCU6ifvCbb5chjOt/pxUKQZXQZxxYst6GhMlPKL
SqVc9jsT+AZvxYz2fBg7DscBNSc14z3K2I0u1HWOab8O3Z+JKQaHZyJlcwdYdNtHs8ipTEjp
e/xh7kaMndcR5FDloMMmZY1x2e/MMORSNsoo8pygLZbJU8nKounI4TjiBnIlUplDtRcfRu4L
EaODiEBHEw4QwsBlHmM+exxWHuQ0AfL2FDgYmmMoMVRoYlOgAKwm9TtfpV8u/P5SAbfokmgT
YWdkiwly4LA+45dZC2Lkj/kLkRkFqJY0VjvxHUIDTHTWkVc4Go3bY6OW7FaSnKsYRhEMVUgC
UlEUJtpyGKRDamy5d2bA2dgO468qCWOVL7aiI8cK4+pButJELPcQwmcrAXioMhlWg3EpWyx3
8m1CBy45Dy9m6TyjqajAC+Q9MDKxJEqwGXrvfq6VMCYtTcZqYuBaDdyMjZSB7DgRFz5VNwkE
LknZg74WRjAN2IJ4kVgrEPf2gD/4KiWmGEqdpTfWPivXQcqjHOd8GuysJ8RJa0uRX1pL5jUS
o51kGaLEFYUQVUkUhm3B4y3oYc6F2/TLDUY76r4vAnwe2oCQo5cRGZuEJ8mxiDi0EH7mPHTa
mibXR6hsdmzwHdzhM2YLzkfF4fGtUEwj+o2R03TczpVnRKvLw4NVXSDkOmPQyqN0ObnEmAgc
Xz8HK+7kyNfdJ1yc7AKupT8WhN5ETPwjXNwwEK3dPGFtbIXRt/KVe1vtN9yc5wkTE3cM33AK
d2IfI4oYRT0tjeEy+x6dypEC/fPeJePSQg9wbUbgOJ3Kl0rbWc0gSZwZ60gM1J5YcjgCsQnk
muv7w9XBjegD7lgsz94gLY7BLEc2hM6t0WH4WoTdikFc1AnM78ADq3UwEuVkScrJMsZKiMAT
2vd7GXlOwHDsyPoW0iTkagq4f3UKM71NYdN7A6K+qYyVluMzsjjJrmcEgaM7Oo3dgrP3YnD/
3EoCShjCzH8upvgTo2bLOdyLfYBzK7vD3MhG4VSQ5EVgpCVZsy7u6DBivc57olI2X6Cei4Uf
5h64hkcJcbhDlZ0TcIk+HcfIMEa955Nsjej14D14FcLuPkbcvWOY30kAQ8thOPtFTmIm4HX4
FBfiCOqORYci6GcdvmkQXKwdYMpxwTw54Ux2f4bg2djCte8yHL0ZjXjyrNPldaTpTHbPQ+Bn
ZoXuC0Nx/VEiXTrs9rm/sGbTNXL2l+JzymUs8uTCmjhpYuVlwjLkZcLos4nvioXxTBtF9GAq
7PkyUJ9OcRvcGhy3RUgolTL0/WsTiBHeOQSpNMgo0zGF8qg5mvT5OBgeAgE4VGksuf4lKYjE
DGdz9D6cySRPB7cFx2YMLn+royMjn+0i9ovABysf5Kmk4ZVlIzEhep4qoC+tfo8j/S1hOeAY
I3qwrvgD0lTKpqU8uY/QcS5Ed/RG8IN8NQA2GUvb8GmAtf58pFrDpJNyPF3lCV7b5YrsYc1p
/5SztalNfR/dnfFz+2hT70f/5fRfTv/l9F+uuXK6Wh1xtg+ycmOcYxRukXM3GB34ZO8asQWn
TyxCv6lnERdzAXt23mD8no6UpbIRsnRnrKLPnj0BMGMZ0qXIWhLwn8Vmg019jB0w5opm0IdU
koubY6zpoJ43ogzs720B68GaqaKlNR9wjNhWLVqZo+faa0jLrW6Q3FD+LhwLe3RE4BA/4ryh
ygEdwVOR0olAldZe2NYUbcbvwrW4F3j9+hWePLiIvVtP4BV9VpcgLngAgh8XE4fYE2z0FcC8
6zyE3oxFEtEfbp36G+fTZfNQEr8M/RfG0uf9ldGOcBq2BRfvJ+JZaioe7e8H+/aaWf4k5e9w
edkEbEguBYVF5sTuwpQ5V5qFqehqqmOnbbGrQ2HnuRCXEp8hMfIwprqbocsuGbZRP3adfVVl
4vy0tuBS2UBamsBj2jlkVko1xyl6iOkEdxx2nTgPtIxTKqlA5pUl6Ok3FcdSizSdb+IyZKU/
Q/ztAxjv7oqJt/J0j6mGCjbqCB7bCePCXmt1dtJBDHx7THvILBugWkaRxs+CmKQTXg8CaFcS
cHuwV4MkAXFFLrI+fcKn+k9mMo5PcAFH0AXr41UIz0RnWtbOW4FLyzKy2jPwkKqMv9DNaTDO
1gfwEZkTY8fhYJqM3CF6MB9DVkQxoobLU0PQxcQIrnJbRv35P55H8CwVHV3nvpD3ECs7mcLC
byGORj3Fq7Q4nA3uDCHbEUMXT0J7oTl8Zx5ERMILpCVcwtpAK3DbLGHgEOKcm5hox0HrhdcR
sb4rhPxOWB1doDswUp4lnBdwUOue1RQ87j8lX5eN65McFA73uqxz6C/kwHf7Kw15mlD2eDk8
jYXoHfqGtvOKY2bCkU1sIuLg07VP1GOTRs6zFE79xtrPYo3/lpzqu6QLb2zu9/QeLRaj8v1h
9OSboNu+DFSoXZ+Wyb6MwaatYD8tWtnfP4QX/r/Ly84TghVMdQPXvCe2yfF8SWka/h5G8DTb
Yfg7TYXQVkMFjAlg5DYfj1T6r/52G0s7OyFg7X3kVFGOSCEM5ZlCmos9/mnyinlrItaozX9A
YdovtnSEsXFHbHlRIX//qWAmczqws7T6Nfb5m8J1fqxG9vcmYXr/JRkzY4L9O4DHcUC/5ccQ
ER2Hh+GbMcSBA5sR5/ClPvN6U/DJJmCOuuQiw4IRYEHkZt1FbgNyTe1Pm5wq9qj+f1Ussrnf
S6uzcHVWa5iY+mDqjguIiktC3J1DmNGGC6tRVxUBvL8aT/zT5VRb2ZO16GJjjjbTdQdfU/2e
nyTDt+eHXkd0wmNE7BsPFxbRS7uH4o3c190UTPJ3yjUXg5QRxI3As3drGBNuRr8tdBqopAOu
+TBczm7Y4UkZJN/eZSA9PRV31vrAov0KRKSmI/3JeUxytkLf0Md4lU7+//oTRPJNoip9JzqT
zTiAOJCKc+9g1tijGuwYcX401nQVolULmUHVdvxhpBaJ8DSkJzxGnsT7Kt1GbUXmZSz2lf3W
wDoIJ983nlGiOvMQUWB4CGgsUqQsEcGtObLI//RwHLr8AqK6Gnw4NgxDDmUSp3cRXl49hPMv
S5tt/P1so0DWb5fHwN5+NC59q1MZs5wRPPYW8tVeuqJH02BvpIxM19YnzcxvQmYOKnXjwcHW
YJl2R8hTLeUS6EhansKIZNkPxq7EQk2wgxwi94M7wGlgKNLLJfIU/UJwfLfjlcpzp+oIh27e
gzO3YpH8LBnRl3diAnGYteJ1wbZUuTJJjNfrIyzQytAYPLteCD5+FwnJMbiwuhcsDYzRbk2y
dicm5ewcaQOW3VhcUiu1RDtLMs5gmgeXBnDo+2E5IGh/EvLlDn2pJA83h5vDkCh0sSr1QiWi
x1jmyZZF2MjLTkkrX2FHFx74AcooW1kfhYicYAMDKrW8nCwkqczG85sHsXSED6xYLWFk0Q5D
F+3B5aQvKFMxcKnsJXatGq8ha6qSFYW6r8rs54gIXY7RvjZgtzSEqecgzN8VjsTPZQxDVSxK
xs6+luA6DMf+J0zCQXXmcQywtMKgI2/oaBTKaZsVuR3jOtmAY8SFffdFOHlpFTq3W0xnLqIN
OUdlNDFjnp9uREeuAD0OyIw4aWkCFrkZwWrMbXot04zrO2NhraVerpFNH2yOyaUP4bqsMxgg
NIDjjAc6yWrq16b7bQDIk2WAMCYO9bQGx0F9TLrto59tQ3LG7Tcq0kSKicI73MIAthOYLH66
pIrqnkKML4lU2uj1ZR8jtF2eLE/Hq7y/xTqBSjGyLw+BmYpSrn4ddpv5uJlV3eA85t8aDSsj
7ddR9MehrlGGvIhRsLafggfEuV/37TLG2lvRmaeKKCORY4WxtwuY98vpgE1kbNLyZ1jjJYDX
mme0gSDOlmUjqgcHaHIA3wjui+N1Z5chB2Rod6JkeaxQ9m/ohNnRMsCgpjwfb6J2YwRxYJv4
rkdSsQS1H08gUPAD/baywYR7+XS/tRUFeHNvO4YQcMhuxAm8U93jilNwYKQrOPR8G8DEoROG
zg3BxeQcrQSUesex0xztkSj1ZT7YVmNwWwv5io6uZzEJhvWRbmbDriqAdWlZCjb7mEDYS6lw
0e/+e4qIZYohl6msKDIijb3lAIR9qpERHK160fW9KSLkgzkuNJjzUCRp8Ez7emEIzMm9243T
nW6aAoc/HO8HUzJPLQXdsDWpqPE03xQAnnEMI+xY4HVchqgcHdlKJAW4M84aApUSU8r5jsQ4
a2uMi9QOgkpyrhDD1RQDL3zVBHNVnKWq95sXvR7+psTZMPU8AVl/jW5AgTR1tbmIGGkB04EX
ZM/q7QH4m5B5HXMen6ulijkpfrwY7mxbTL4vop0fWeHjYMdxwfSrOjLLUSThsCGwFHTG5uRi
FXLJZ4T1FaIF1w97X1crZD+fHU7e29aYRwxCpUO+Fh9P9oHQcgSu19etrp93OgJEqvm8Tw4i
jn1y1j8tZYAhVHarFqaDFedrNb1PExB4ZgS+15/PlF5zjug1pkNwWX69itSNaM9XkqM19ntK
b+U7Y06sUlei+xYSZf/xd7y+MA8dzSzgt+yGBqueluM7YlaMsoY3fT02C25zbytAa4qAGknu
uSXLDXNuq8wPea8n2JjQpBUa2E5ZB282uadZEcjWcU80WSo0EGbCbgx9jDLyk5a1hbH9dDxS
qU0ui/ppBashxxl7UFkyRTq2oPcEat6/UM9P6MfskwCnJwkI1ULYnwD0dSr3Z0AAglNaUzHT
ET7L20JI6XU6SgLIMm0JNAiLNGB0ZRjMTQcwSg7Q0Z27uoDvOBMxxVLZ/nF+AISCPgj7rJL5
L2k1vPlumHe//jyRpVc2JQZ2UplURvoc4QivmVswytUDweTsotbol9ODYOUwCRG5TLuo9ss5
BFlx0W5tLFJOjYUjzx0zrnzRsCVK4hbAlUtlripXrPusi6NhZ9oNO9MaKHEpLkTc1p5kvbfD
optZGv3K1iZ1VjEdOQ2STiTZuDTYVGsWmNqSr3iT9hJvv5WgVgP8qkVJtui3k05+9T6q/85y
/ZfTfzn9l/s3SSc0YSRqCuyFPXAoU2lH1X2/ign2HDhPvUJ0XirT70yMD9Mkw8qCN8rwgipT
TNXd/lwttwslmvgLVaruSE/wuV3xV0ZD5STycGucPdhEN2EZtkILk46YMZM4rFu1goERi3xv
j7Gq9gGdsXQTelsZyexMAx4cfYdg1qYwPPrALH8tLXuCdT7OGBZGsDJJLQqeHMHktgT76LQG
j0WyLCYFt8fAyowip+twltV+wamxk3EtRwzR/alwtBpCdIBaHfr3NUwafYKAxq+w3dca/nue
oaBMhI+PD2MKAd6tx95WYGIy7PA6lnazgXPQfqQU5CD+r+FwtfEjzovvP4Sp6B67REE6cei8
HWlltajMS8b2rlYKm7V+7A2efZIaFGW9xbusYtTI76M546SyE95d1x9dxuxDQi61dsRKZzh5
riWfM5ElqkRNdSm+PTmOCS72GHE1Rwfh5BtuLWkPnrEzJpx6iwpt9imdtXoUsRV64diHGuba
VCmjWK8X2Rgos5TQTjkK/+nYW1ESSVqegvXtuTDrF6ZjvnNwf3UXmJp2xiq1iN26L6fRz0WZ
6YYi4T5b2x7WASGIyy5G7qurWObniG6blZlmqbI5cZu6wcplAs5mVqLi7UmMdrRCj22JdE18
CrOLmuEIA46PRqCYzJZuIEjTqI3SCV6XjZuTHcHzWo7oQqWuKS2Jw2J3IzrrY7tVsYw6/NVv
9qObwAHTHxWpzGs5nm/qCE4rQ7RiOWLCuQ868W064vz+crTnWWDA8UwNuabgcf81+ZLEVfAy
Js68w7LsThT+v4TMP5WxUkO+PBUhXUxg7LUKCcUy++vdwQCYaH3/WsFhhiyDrgybNITT7Gid
Ti+tmNlPYI3/ipwKtqgLb2zu9/UEkY3t2Zp4r9x2U7zfWWfRXyjLbvRP44X/7/KK87IyE+en
uBKnbW9si4zGwSA7GNsMReiLUsY5LsusRu3JcfLSZMR3kHkRM72d0X9nkmwvlAfAsekKCdJm
Y49/mnxzsEZd/gNqT3j/d3dyRsrWocyxS/kFzdH/XBbqyhKw2I0Ny6Gn8bER21kbpvdfk2lh
SHTNcx+ZwZk3x8JGKNMfmopPNgVzVMoR7O2Wqhz1vAMhtBgu19ua21/jWKc69qgLi2zO9zJc
diisCba0TRWXlVYibZsPeM7z8FgeiPUr8cQ/Xq6Z5HFZv71hqtKvpPw1wiZ6wsFeALazrHpA
0zHJ3yP3IxikLGC8YUy4uf3qIJ0U4SGVEr5700FDKh36nXEOcFsgO4ToQ89We61EaU0Wrs7x
gkkrA/AsLWDeZhTWHr2HjCK1lEl1Zfj+/jUyc8rIC1uL7zdno12nYOIcKsbbSyswoI0pWAZc
OPZZg7vfamQHH2UEOFmh2+oIJJ6fATfimBf4h+CZPHsJlUZoR2chOYwfM5TtjycCIeA1Xt9V
ppxwFeC/YlOqJ51If87xr35waXMw6ybEPMPGjubw2fKcwcykgI9APpcRZVv/Mr3e6wcuPxAn
dJYHKEX8AlewtER5M1j5+Y8R0scSbMs+CHlcoDVyXSopx/c3L/HieTIehodgTGsuOG0XIipP
xagiG+LbE8Ph3HY2bn6vVRB9FruzGGU3dLXKlzvQmcuC5+qnMjZrxXOsb0cUVG4nbEouYThS
93WTKUTPyqUazvb8yLlwZZui76F3anNWhQ/hM+Fl7oxBa8IQmfAESVFhWN7LCsa2w3EyU+lc
yzzWF6YGZui55SG+lFZA9D4Ku4Y7gk2UFbYKwaCprfz5ZgSYEeOR44BuE9bi6N2XyNVRCkdc
loWMtBd48aLhT/pHkU4nrujNQ5yiS/aYoFUrHtotfqR41yNnu4DrNAkXP1RpOJKerfWGZcA+
umQG9W5lR8xBG7PWGLXtIqLiHuHiun6w57FhMfA8vhJnG5WGbnprH4USpkpSOtrPDGz3hQo2
db1zy7jDZnlEMlVPugOMhb2xLzpVdl+pzxB3KxSzqCwc7nNwlziISqkN0ZDKhqI9il3z/uv7
HYgLX7UT74oezYCDgTwdrEK+l3Icik8aMj6V0O+E6nj3q8mlfxAplKbyZ2vgxTKRR9CrK/sq
+4dCEZL3yycKRkQ8Eqh0fiqfB0dHw95QWSJFIW86EBe/iXXMAXEKLm0DlopSrvgdryMmjWgL
jkkHrGgocoaS39gebB3zWN8fh1zjFrkGnRLTkpwbr/OQuMoLpu3X40lZnYz8YqLqvFb+jhob
NVZKsRDII9XpFLtsO0yWR0QVP5wCu5ZC4vz/pnOslINzoq0BzIdeUt6n2l7dkmWNLlP3I1Ze
JqAkZiYcWjWl31YwHXRRZ78tyDrqsSEKX7W8z1TWh/yMhzj/10pM7ecJMyp1sKEtRpzUBIaq
Xm2HL0f3OqdKqpzuJwDXT1mTVPUcpEt/UVmY0lUdeCLcn2wHTrv1eE4bj5QyOxgWHG+sTmSS
C8XfLmCgkI/exz+ipiYTx/pYwGnaXRqgoZn/ljLyYt3Xiwiy5NIZUCoaeB+rP57DeEcZmM3z
36cxZoYsXRKQA59tLxrsU9l3OKa6smHcZhauNkScIiDHpg58tA7WUveVgJ5rvS3R84j29Ll0
uSF+O6x/XtE0Zym1F9q3ajxzXGO6Q3U24o8vw4iurjA3NmCWJxl7V2kUm2iSSWX6AtkrqH2N
TsdOwPlRl+i9Wuu1yqkzlgv7ybcVpWNk66YEsbMcwXKarSBeUufxhvYcmA8+jU9VTCVfFDsX
zkKlDiaLPuJrr7dLgc3tTOAoX1uquuuDSTbgKM4HOcjNY5b7oxXy3ZRCPktuMIjxPXwwzMyG
4NJ3sVZdLPfGCFgQx86ZrDomgM5tjWHDfYjS3wHzLrzRTCdeLydUJT7IrmeqPi5JDi4PNQNP
LVMRRV4aaiakgQnqt18vDJTVFn/TwD2VJWOlhwlc50UziJYyfbUX+GZDcTVHokLiGErWCgWG
MLP2yKIB5demnklH4lSaHqlGhixG9HR7cFTSE9P3x1FmCdGcU5nxy2E5YOCaM3iUkauRFa0i
dRM6UMa0OvAorzXKlwNaqg6ziFGWEPRWltYse7ISHlxlFAAFth3uYw6roafwqUbFwD7TD2b2
0/CIrNWq9F3oatsDoSmPsdLLCRMfFMnIwZ3N4L0mWXM9UuD+1k4wYRGd1tgOw49maI3ypRxJ
wy0Esv2RAjtz72CmCx/tVieiWAcZliJhR6/3h7mwE5bd/c7YZyR11Sj9mojQ4Xbgea9Conpq
1wZJJyJETXOAsfNkXHgtQnWdBHWipzgxvzvs2C0VRFUzryFYdvwh3haUoTT7BW5sDYLvmKt/
NunkB/ZR/XeW67+c/svpv9y/SzqROz6Is72+HCYNchJnJd98MM58qZU5wzePxrrkMi2OkWVa
S7AZOM/F4xKppj23rA2MzINwM0/SoFOzHpStLYxFcBsWDKwHYE9yAZ1JVKyF0CLDHD4j6drf
2DB7GDo7cOmgr5Zcbyy6qwTcace4mSw1vSIq/v0h9DR1wuzYEvrahRQJx5CP9jNCEZWRr0Gu
VxI3qglu1xvWPtqzpCpJJ8fJWUfhNnPha8kGi+8E/+nrsbSrnSJrnVRSiU8Rq9Dd1gkDQx4h
63si9o90g43vQoS/lRFnGsdUEhA2xo4ub6wzOEQL6cSs/rkRjKXtyF14nF/HGHtzm8Y4448j
yNoUPfdFI1UN+xFFTYQNw4lsTOynNHmwUBFi5rnBSEGKsITfwnC8r9Bio1Z/RcSidjAxdsGk
M++0Ek4UxI41XuCpOYjrbUyBit4rFRch9dxqjOvVDrYcczrbamXmYQS2n6dY2zLHshHaLkvW
SjiJWtUZpqZdseZhrtLpU/Uaf4/0gUdrR9hY2cOtrS8mnftEP4+agmQcnNgBlsYs8B26YvJf
0chWJ7zXZuPOkvaw9F6IW9+J/pW6HwPs7DH071e0zr28rRGdPUD9/ZNhAVQadBN02ngfGZmZ
yFT9fPyG4hrZ3FL4eS+hJYLCvzGw1HpHbgthIA4mf0JWVpbykxaGAaZmGHKFSQqqfrMP3UxY
aLMkuoHs4OV4d2EWvHgCdAy+hW9agjuagsf9t+TfILSXEOw2SxCrkJeV6zSW40yqWPq7w30J
NmOPidez5VifHDdpMw+XktX3ipd4l1NJ7wn12GTfU1+aiU02FWv8TXIq2KIuvLG538v2hVJ8
yXiCq4s9wTbpgs33UugxvMpknhUyDJVgUSc+/uN44f+7POP7ykycneQCFh20PRgHUjUDiGW+
LVMMvpRN+peg9OVRjHZ3x6jDaYogXir7wngbA1iMkOlGzcUe/zT55mCNyndlAM5/VXEEi6mM
a9YwtJ2AO3JMnQ4YMZU5a6l9OP3gAFgaGsN94kmkl0majOn9F2VMuv2F9HImTliashEdhN5Y
l1LeNHyyiZhjvZzZwNMaQTilcfPhoibX1P4alVPDHnVhkc3+nviD13jz4L6ImTWHJj6FdgfP
aixtc/xqPPFPl2uuX17WLxcucx7S/UrK0nF0jDf8V0QgPpTqdwiu5EiajEn+LrnmYpA05n1z
JPHzNIwJN7df7aSTmg842stSwQJtkG0nEUNMvUjFiVja1hZBl7+hpq4WOTeJk5Moa/FFdXL2
vFTDkZabdgd7x3nApVMnOJsYgO02DZeyarQaweXpoRjk0R/70kpRGBMMT3NvzDwSiYSYs1jY
Tgi3BY+JMZGBfT3MYDvyPL5QD0NShtR9fWFhYAyPxVHII6CqKHoxcYD7YderSjWDxxbGHiuJ
g7Ph+y2Kng1ntjPmxhb/UuNPpxONp/IiqERramOpZ10cATuH8biWzSTv1JfIkCkIKouKSpHW
iw8Tv790OvNkKSd54PU4pCNalTiKP17BgvY8cNwmIiyjrEmlg+gInwfT4WhojXH3CpXM8uSN
6GzTCaseZKFIXhqk8GUoevBM4LfrOfLLKhoBoKMwzsoAjjNlkf4U+DyNgM/C/mcUdflk1yrE
nZEWGuxoek2XJmN9Bw64vluRUqYWsUyMwb5mpuh1IINRU1b8/TJGEECjPpNEvfJ2ca4PhPJo
jZbGDuizaAOmtOWjDXFglkqbB8hXpO3GQAdjtDI0g9egBdgVnoBPpXU6jMifSwVLpVz9+vQ6
DiwdAR8rFlqxLOG/WkbUolJxjbEyQZddGZopJqmSRz4OGHLxK22ASIpisdjDESPPfVaWKKGd
uAK6LA21v0hyryPIhRnpJSs3MhWORkL02q9aLuYr7k13hAEBzmjgRpyNy0PMtBiBZH3dnwZ7
A0uMishH0QMKXDGnnS9NWp/yftlyJ5rm+i1BAkXIIO/k8Y81ynE0Qiaql+M0AJQplBwD2aau
CjpIyr/jzUuybzyPwjY/HrhddsnIPfXjlRMDmGOVZysy7oitLyqZ96dFXgmuyAxyk25KZ7/q
71JyX2BvoBlYdqNxOrNax/1+R/hgUwXbXdd8cLvuQQa5BmUIjbV1wLi/Q9DH0hGTb1KApYz8
YuwwA9HFzHGYdP2L/p2MabkIbgLi5HxZKjMYeN3x93vZuArvjoJFC3OMvFOoYz8i6+3GeNga
WmDo+SxF/6zWcxBOAIO0tJfIeJ+FgkqmASyKGgerpvRrYIoBpz4r+jV0nIgziRRBKgXxEXsx
3s0YRm7zcL+R9Kr0Hvn8L/QUtqSfvWoJM5lzeyQsWbqzNUhL4zDfhQW7qZqp5Khz8OF0B7Ac
ZzIyINSXlzGRv3OSgijMdeXAYfJNRlph2f57F6MtuPTeUEiVzzFth3VPZWOp/RxGnillXBXi
9T5/oqz1x8mPNbrPT9FjrOtMzifvRTh1dCzsjd2xKEak8/0tjp4BByPNtM3aWs3321jkzSX3
Og6n3lU2uCeIv4djsBmVlvW7ZrYSGoSyhMcKbY5ombIrkBOqmuQsJUbnm1PjiY5BjM4pJ5H8
Vg2YVPt8/K4le1DtV1ybTqVJ7YxpIacQEZ2EZ6kvkPpwD3oIZM+mngwm1LJfVRAnSjsTVyyI
K5WnpLXEyJu5OueIBjV5ROYGU0Yq/oaLg0wh6HsKX+pUUuyaNKDsC5Ugkzj7EoaQeddWb7f6
NQWkao6Lumb4IDPYjL9Hl8SrB7kFuhTywJP4JFfcqTIsfLVzRNVhQ5Ui4aucB/V9c3lWELKM
4U2U6xKt0Z0yOb6Kfll/PRPPVXiqOi4q5XcbvkZKVzrqhueF1c/K6d8mLHaDiZq+qn5PdCYu
E6Jf3SlQi9SRASI8l/mIK5UyCIYmHis0Sq3QxA+uLIKSdhrxqDNVLfqHqo8/2IyOAMyTKOfT
pBGdWlL2Blc3jkFne8pJRZUI7I15hxORX8ckAoWrEYEowOhMPyEsR0YwiU7lKVhHDG1VgpiM
KC6Q1+4n+/GtaXASdkaImiFVcJsYWBYjCSgpwqMFbeA8/gZyyl5hh689AfK/E6BgEVpb9sZh
LWcdtUd+vjAWdkTXE/QhRp6ujEx0pCsfthPJ+hSL8Hi5J3iucxGZp4P4STlOVhLHCbFZ1jzI
ZaZOpcpjuVAgTUvwOizE1Q+a+1jD5XUkKHt5EIOsDel3j0+M0+LoOeg2ZAOupuagorYSuelR
OLpiBNqby6PYW3LhNnQz7nyu+rNJJz+wj+q/s1z/5fRfTv/l/lXSibw0TX2UeD3x9MpQM6Vu
Qs66cxOn0bq3RrBPyWdkJJ3DFGcjmPY9iNhUGab06j2TwCkpvI95bbgwaiWLsjdiyUvr8Dtg
TXKpWuBOCDpxdOtobO91Wm055vlcjBfHRsLOkJyHgSfo+5Cdzy+xpxexk+0HI+ThV+IkJCBw
0nr4OKmkaq7JxsNd4+FjyULLFgYQuAVizsEYfFdk4yvB4yX9EfxYhNyI0bB1mIJ7Ouymkrhg
9FsYg9J4yrGrdi/EPgyV24cysoAJ3ObfR4GEimztA1PzATieWd30c0geoGVJwGpd2J1y7Krl
dRbj0sNwrOhmhw5LH6JALGWM/afPxwbKeteIPiCd4Tx+hcxCZXkkKpNK8df3yMjIxLfiGq02
AJXC/cYCL5gYu2LKucwGSzLX621mQy7hu0qWDorQu92XD5d5jzVwMTpTc3tfbCT6muj+FHgN
OM3A8nTqTSs6QWjWDese5TWpxHpja0SpW5Qh4/RkuAnbYSV5d6q/R2FdNws4jA2HtOg+xlu3
0MCjlLr0RrTnEhv1UcM2as614bBQywZTj51py/Kg+LRyxMyYYrX7WghXI2IH3C3Q/vwI1vf6
GMGOOTYYuDsRhXVS7ThGE/C4/458Hb5fnQB7QyJ/4IUW+WEKwj69z2eFY7StIQQ99yOj3kla
/gq7upjAYuRNho2i4VP4QWyyqVjj75JTxRZ14Y3N/V65Zgtxb7wNjFwXKGxY5nqVZ+unHKmv
q/9xvPD/XZ5pk7/C4eF2MGrZCobmvRSldlRb4Z2RMG9hibFRBRAl7cIAZ09MO/9eQTaks2Rf
GwsbQ0sMlxPnmos9/mnyzcEaFfix50okl1D7QQ3KczNwO2QwbFkCBOxIpYOTZMFMZK+3UO4/
VHneF4dHwZFlBIcJsuzOTcL0/nMyDeCE5L3e/6a6afhkEzHHejlG6Un5h874J69O0Nz+GpNT
xx51YZHN/Z4KnOpiYoOJkSI1/E+WYZgnx+Z+NZ74p8s1NyBf1q8s41FdaSpCR3ihx7pHyKsr
x/P17cClgipLpE3GJH+XXHMxyHrMm9cIJtzcflv8lDFD1X8kxnGjkWQtyUK4na8lBVUJYoNH
Y9fLClS82gt/HjODiMLpk/8Ay3y9MePaV9RSKdfHO8BtXjTN3qIX1KYOsB14AZV0lLc5RkQo
ryWty8HdRZ7gkANixK5QzPY0R4fVcUwmFP3QBFrTTDMVHXIQzXGBsdNM3BdJfqnxp5208hWv
VV6OtAxmCRV1osTaduboHJKmwbyUzYuMAa963cpXu8jLJEDvQ+91p1wseohp9iy4zI/TNAaJ
YlGcEooRDsYQdlmBO9+qmzUfNCnAkGyKcnZxfdQPq4G1ZEg2joZadcZf6ErVgZRHnstIJ0Zw
ms0sNyGtzsCeriYwHXBWjYxShYz9PSFgt8biR4Uaz+nLqb5EOWUSJBSGcWsWbCY9YCgwNCkn
/xPevPlAHBi1KEleDW8zX4SkVfzYO1dbhPfRp7Flem+48lqhFdcJ3SdvwImoDBSoZEn4sfI6
tSjOjMXZkFno21oAg5bGsPcbjzVH7+FVvhKIKEteAQ+umyJyWLVJ8m9hpHM3RUYKKtK/rdNE
RXkgqlGG86aODgi6Tq1HKrJpJtp0ZGackZQkYU07Y3qMWj8sL6yhHHBllJOOBcvRmqWjZCCC
GYKIM7T86Wp4snjo/rdsrQMNt/p+zYOuawUTqz+EYailEezGX0M25SCTy1uobfza+2UrnKK6
mgz4k6V7rNQyXmlpPBa6suE4KwbFUuX1651+DFk5GGZkMx735M+hIXnF74pjMMuRpbgG43ej
ZelFa7IuY6ozG/yum5FUJNFyv7L3wmLETbpWqvb5MIbDdFmqUYo4uKsLD2wTY5j2PIDXFKFG
Hl3C731c4cxT/x3VJEXRmOMsQFcCnETOdgJbxalKr1ny/HscytT6/CVFCVjbgQNjr5Wy1Kj1
9zlS+7jrW/26aqxfVuvFdHSNrF8CEA8Kx7f6uo/0OzALzkYWGpFNWp9LuSx7k3HHrXihQlyi
nNsp67zBVindpTGenCsYYsrWSiilCGIL3dmwHk3WvNo5U/RwKuy57bHheT5ebO8CnllfHFYD
4+gxlMRithMXXqsicX22Cyx6H1aQFSly2TBLJ8yKiMQKTy4cptzRvfaq3iNslB2MrYfiyOsK
YginYGMHLiyGnKUJpRry8vIYJjzNMl3qTVwQg3VdBDCyHoSDLxonSZYlLoE7X0nYYl63CA8m
28Fq1G1G5gfZ30R4MNUefC2RFPTftThL6e8JCPj+7GS4cs3QfctjosBJ0ZxWQZWi4jli6h3V
FOgSFETNIIaszFCWlayTE0JqVddQJV6G+IJnPxl384lTOnEx3Dm6S4zJ5oeSccPiBKaM+PsV
jLIyhueqp4q1RstyXTH9QpJWRV9Vx6Gjj/ja6+3WX3OR2jUlebcxyU6gyBClUMhHaVPI+Wiz
VF5urIaqqyuA9dg7Gs+RlqciVPsKGfuBrG9T2Iy7ivijQbA1aY05N79pZLypH0N9vWj6O/n1
LNT2l9pPYegjZJI66qNuhBYjcDNXIieF8xn9abunssRgtOZqPjvKODgYICDjjlAbD19GLq5R
XQ8y8I5nPZpOZ1+/HtRTHotzIjDBlo8uu2VE1Pr7a2z/VK55KptTFA5MbguukWzMDRGBZICR
QKMWe03mUfTimzKISrL9kkenWS4pe4qNHQVwnRuloS+UxM6GE3E0nYoPwxD7dlidXCYv12SP
nqGxuDDaHo6TIhR1fZVzVIf8R6vhKzSHd/e2RD/sjJAX5drvUx5tzvdeg5jYregssMPY8Czt
WalqvuF2cAfy3Htgc6xm9kBKF/yQlojbe8fAydgGoy9pnh0NkU7EuXcxx40Hx6AQXI55glef
S4lRnYMiLXuVpKYY3zLf4mNOuWJe/2TSyY/so/rvLNd/Of2X03+5f5V0QuMrRkwbizqf/Ymz
ctRdOrMrRS5cOma3zhJ0VHnaAUIKxE/XCALROHuoQAnL0bjdALldFXdKub0enbjG8FpyDU/k
OtqrzAKtpT01zoy8mwgybwETNf1XUpKOC4v9YcE2Q8egUQho1xdr72nTm8qQ9ewGdo73gIkB
AX2vyoKmKF22KDYY/sOOIONzBKY6EZ1nyhHEfRKhsroMORlJeEKwKCqzwYnh3bCYKo9KkXPU
geb0TyhWlDUUQxS/GQFWzhgb9hbl5e8QNtYZVgFbEC8SN+0cUiGd6MLuVMf+tkosL6+zkw4e
qHp7CP1sPREcQ5x8KmP/6fOxmbhjs/qu/oJr8zzBJTrhtAsNE07q7cO5zibosImZ4ZLKfDnY
TIh+Z7I0dJjaj8cR2HosWbMVeLXdH74rn2gQ/Jl6UTYil/tAaOaPDTH5TSKc0GNoZI3Qa5es
i0uLOsPadTj+iv+GrEchGORki4AVN+mSmjK8wwgGDjPwUKSuJ8ojirUQWdQbbcuajyD3zOyj
5sNx9BYYac3+qH1NysoW8Uy0Zx2nsSyiM/c15aDdmgQU6Xh+TcXj/jPyxfEEo2A3KL9aLi8V
5+HeHFdlRiDGh4NOIS+1YnhMDMlEkXUJTcQmm4Y1/j45VZtbF97Y3O8Vf698iZBOHJ0+mzpi
j05zYsNq+AVk1Ur/cbzw/11e8T0VxT/SHhzboTiY+AynJ7nQpXa2JzGJJ3T2UBYXnmPGoYtj
JyyOYJZ8VfTfbrUi+2Zzscc/Tb45WGM9zsvcC1rC2M4fsw4lIk9RwkSWpUDos5URhEoRG59t
8QGHXONwZk2TML3/nIyRFYYdj9dOCkj/iKJaadPxySZgjjI5d8y9/lJ7cMrnXJpU1bz+GpdT
xx51YZHN/Z72f3GUJfsU8sT/tqUjD04z79N+8F+NJ/7pcs1tNP7PccW8W4+wd4gXAjfHooDy
+VW/wb5ufFiPke0RTcUkf5dcczHIer9bY5hwc/v9OdKJpBI5716SF+g57m3pCguPhbhCpZFL
uoDprhbovuO+LKPHy3fIkdfUpjYU5WJ4jf1jFuKBiKpfeh0jLPjoHvpebcFk4vQYTwIaJtMp
WujsHD3tlE4FSR5uT3SE29xYVDzfgHZsE3Tb95p5AFZk4O8BFnQkpcWAv5FRLtFwCpztL4Rw
4EWFI1C7cX8Rw62Icr72SZMU/n/L+KNrzZ8dBlunyYjI0TSQJbk3McrSGN5rnynGLa3+iDMj
bMBpswTRDRBoZDWdhOh/lmkMUsbM98jV6GZqDLuh+/G0iYa5YkyVb3F8iBVY9hNxM0esuA/R
mydIZJQGicbpma3B4nXF5og4JL/83oAB/BHnxzvAmPR57bs8BT7ZGPZ3M4Gwr1I5pcvnRC1C
G44txl39zrgvqjb/CGsW7CdcxTctjP+scwMgZHtj7VNm/UnR45V07c8+xz5oLbMgM2SSERJg
TQCcB4polp9pVDrbxPBdmD/QA6aGLQmY1AHDg/fhempBs9h84uLXuHdkNcZ1tYdxy1YwcemF
aZtPI/pdUZPSPTLuMfsyBrv0x1l5CQIqE4StnAmocIq+2I7OFrKat5LSZwjxd8EglRqy9TXU
hCxHjD0cpVEq5uGJicRBb4qBF7/JsicIOBobLkUQo5h2bF4ADrypJnMfR5MfuJ1DNLLXyJTt
D3ialqe4X1m/ylSzqs+6OisCS32J09pqKI7LwYP6cTRmYMrkGjcwpcQ5tsabDbb3SrrWtQZQ
8PYA/E2UII2sX6727DPybEUm/gfwtlr1/rTLK65Bp2WtjxBX+129c5Fa+/Eb0JnHgvPUyzQr
Wx10DRToJs/I+uOj11HZe0M56qMm2qIlqy2CY2UMXWkxRWQwZgAtit+pvG90uYDpDhD4zMRM
H/LO91cSyiTFj+k00SZdQpCq9vwlFe9xYbo7OIIu2BAvu2b9faqvK00wJr7Rfo05Xgh+kE8f
/PXrquMWZlpGKemHKiFmPuSCrOQUcWSW5Yo0U0uT+S59to0uIdYmOI6ZDo3O8mAqiw6p0EUk
iiZnkCHMB59jku0k5Xi5vzdMKbKdlmwissg9cwzaswtDrQhouD5ZkUaTSYh5itWeXNj1m4YA
a3tMuJmjdP4W3cdEG3P4z+wPe64X7dTVOkZxAR6t7EiAsvZY9iBPnrKVON7PDIYF1wdbUsu1
6CKyyBeW2yIklDbwvIqfYlegOYzMemBbokhn6lEGaHekJwS2VPpzLaQqOYvaMkDTsSCL3uMx
HApNcZbW61aZ4cswe18So253U1pR1HhYcTpiiwpJRlKSjK1deWjBl9dCpUlrhmB5rmKwmKs/
n8cYOwE6bXlOv2sy8pspBpzRnfqXNhyNiYyKjiAVFyJmhTc5T5hRCHR/PHtMvpPfoA5EPe9P
JwMh1JElhs7eRhFp5Rm1ZHNWhMSNnWBi6IL5caUKkJsGlvYw97m6r+fpVNRDr2TL9piiR5ju
wNfQHZXvDbUHKUkVyr6Fsr2crNmY1T4QmPXEjqfFjHuTyfEZe0n99dT33+KYWXDkq2VtoMjZ
BBTjy50TMkcVT2NvUr+n6td/wc/EHEPDlRl66IwcFyfAgeeFFfEqdV7pPlvBqM1SJKq8PzVZ
4RhPrtVOnsWFftYccwxRJXWQe49e0Q4cAwfMkDPfFfe3O6PB/VPDwHu6Bl48WaamerKO+XBN
4FH8LRyDTNXIJeT5J6xpB2M2Vc6qXOVdysWNERYEoDyMB4cHwtJqMMK0ZFii0zETvWTaBG/Y
9D6E91R9cipT1yA7eI6bik6W3lijbsCT/bgk5S/0tRSgw7Io5IgSsdKLA+sRF2mwU3NdS5B/
m6otbg8fVwEs+h+hr6Opy2bh+gJvCKz6YEcj+5SslAwX7TY81yCc6yKd1JcUFZoPxcWvdT+k
f/7ppJPm7qP67yzXfzn9l9N/uX+TdFKVsYvgK3wEnlSWw6T0+AtEj+f4bsdL4oiofhuKsQvu
6ywtTGfFZDlgeiMZ/ujo4yvDYcFug+D4Eg3SpyivTOOcofs2ssb4e4Va9TXqPKuTaP++6PEy
4phioe3yJC0BS9X4GrUO/mZGsBx8QmsWXcVZn30Jg02FGKCS9VNam4XLkz3QYfZZJD46gAle
AhjInTsGws5Y9SgVVxb6wnNiuIZN2pDOmXMvGO2tOmPriwqCTaRiX38bYtfe00pE1vb7vKe3
cD32c8NyirGfQ+y5UWjbs74UcQXS9/eFq88ULJ/cqVlj/x1NWv0ZV+Z4gMtpjRnhH5qk58lw
C4LX7FDRhSUVeLW3O/hspSOfYc88mgGP7vvwpjIH10e0RdC1HJ22g7TmO+4u7Uj0mu7YFFvQ
ZMJJk+637juujnOB85CddHnd0uTlaGfrh+BrmcoIf6KDxix0g5GBNYLCmA5fmpS00BVsx1mM
GvvaGp15gd+bzqKrimnFLvcCmzgvrUY3kcwtL/Fi5DSHgcGptpwrQ2CmYjPV7xWVopJm43H/
DfkqvD5AbH6jRuQvfKP3ueLENWjHMYHPiht4qoKtJp4aB3sjG517pxKTkGGTlIM+TgvGUFv4
EjFPchQ4V/Owxt8lx8QWdeGNzf1e6dO4geEW2slXtbmx2BZoAZZ5f/z9uvJfwQv/3+VlhLoM
HB/lAI7NEBxMLZX1U/meLrXDtSB2aLLSDpWWxGFxayO0YLXG/Ds5DGxI2X9XbExQ9t9c7PFP
k28O1lj/rnRYFYFnL9KQ9ioDmVla8FzasSuE9bi7jCBUav/9dnEgTE0H0L6UpmB6/58y43TL
tCB77LmsBs/wenyy/2nd+GRTMUeZnB0m3c5rglxT+2sC1qmGPerCIpv7PV22ycQSw69mq+B/
lXh7bDCsyLtfn+n3V+OJf7qcVlKSYQsFyV/TRqMynbBhauuC/tsTaRyJ6vdL+Hg4mHhg2eMi
ut+mYpK/S67ZGCQddG7QKCbc3H5baDOcPof1gxm/G/7SAuJpJ5/kIWKcIzzkaVxow8JGO5tb
9HAh+o9eg+O3k5GREobxQ0LwIPk29ox1g8B1Gq59q1XptwhJm7vDc8xphVEqUxScYO63EuEx
sYjYOx7u5r7Y/IyA1OWpCPHjwYDnjcm7L+NRUjLi719G6PJhaCvkwVzAAstlBm58r1W752JE
zyDKh0qKU817LEbCKm+Y2I/HlW8/BtI2RjqhD7MJNjCwHoe7DUSi0NGXvYXgdtgg36QpoNsc
frt0ZEWgmNizXMA288fysw8R/+gyto9tA75ZALYmFWkeqjV5SIm6hYiICFzc3AOmhtYYtPMS
/f/IpCz6hft4YRpaGxvA1G8ZwuOeISUlRfFJzfiqqGdPsejCJgzE1E1HcfV+PJKT4xB1cQ/m
BFiBxfXEvBtfdZI0ZHOSj4jRlmB7rVZkwqBrkoXNxqSl+3DhTiySkhPw4PIBLOxlC7bAF8uj
lDWE6bJDF0bD1tgWAzeFIzqRrJnQOehixoHz5HB8rpYy5unubGqe+uOojkgEKo3iODsjmHhM
xO4r0UhKisH10Pnwt2DBtMd2PJWzesW5MTiw4yiuP0xEclIsboVtxvj2VnAZdRQZqnX6ylOw
vj0XwsBj+KDKJiOOkpvEUUJFTGhL76a+iVR+T8GNA8EY3tES1kMuNWtdFj+aBmdzDwyctxPh
CZ91ZtOpH1fEaGtwvVfjqZa0+bSDzcOXKBEVCpAhwNQeIw6nIK+8DN+fnsLsdjy0shyGsAdX
sL6/G9rPvoIvKs+h7vtVTLBnEcfNOXzRArqIv13AQFMjtA5OkkUnG5qiz/6H8vX3FIkPr+DA
/ABYsoTotjlRxuYk6+B7xDy05RjBuscCHLjyAAnJSYiNvIRD6yfDz9YOw1XADLpfAyF6/fUA
z0i/z5Lj8ejWWfwVPAwefAOw7AZiR5wS1FCXV30fUlIz8F1eB5mWY1H1q4sbeaZ1yIlcAm+u
Icx8p2Pn+Xt4nPwESbGRuHxkE6b5E6WXKGub5VkA6H51pPurj6ZTzfbTkLyCbXxnHKzZHlih
4mjT9jtKkXl3PAjWRmbotVuWNlCxtuh5IQrRxnDcvXuX8YmMTZf1Z6yZJYFJftmPbibkUL2o
PFRlv2NG3NOH3p2JsDMgB6UBG56rlRkWqOeffWsePDhs2PVbjuO3YpEY/wA3TmzE+I5mYFt0
x9r7SuOp/j6nPhA1CjRm35qvs1+W0BeLr39WKIOyfm21pLwrRuw8FxhZDEM4lXKPcrgOsIdr
37nYHnYTjxKSkfz4Ls7tmI4u5kbg+67A/dw6NaKSLKtMg9lrJCVI3kA55s3QZdYenIu4i9tX
TyJkeldYGPHQaeVD5GlLr0s5NomBZkTeH5bTdNzO1VEOgs5Uw0XLVq3A9liuiHKg/1YajwWu
hvTf+D20A/EUAPnmWBBs2DYIOv6GETknEcVgkTsHjlM1M6RIK9OwzYcD4YDzOmt7Ukb10eG2
MCSA4IDtd5Cs/p7Snxd4X1jDBO2m2EHQ/W/tihlVuupcf1hqK9OWcxVB5kzSVlOdpT/TKIJQ
T6Lr2A/ZisuPYhEVvhvTurjCzZYLlvsSJJJxVr8PRXchH07uzug67ygi4+Nx/8JWjGrNh2Wf
PUgtlZOEK9OxJ4API6teWHbiDuKoPfP2GexasQbn5dm+pGVPsLadMYzsBmPL5RgkxNzAgbk9
4eluDiOuGoGi6jUO9DYF264/Vp64hViypyXH3ceNU7uxbNEePFOUeylGzExH8FXS4aqT8ta1
NwbLcThCrsQgMTYChxb2gqstD4amA3FB7kiniXMCJRlC4UyhUk7zZcRHWo4yLPiOmBWjfV+m
y/nwmekEZX0r93KKIH1mrD24TpNw8VM1U47PdArJrqeWnpDoNa/3+oHvMJ1Z3oomzvJhP/UB
bRRV1f/2YcP3JK1+jyP9zcFxHoUdV2OQFBeJ0+uHwZVniT67nzEMBtpA5fFg7+yILmQ93JOv
h9Ft+DDvHoLkYmWGrDXeZN5JnzuvxyIh+ir+muGPNm5mMOIH4sSnWsb9zXik63wpQuzSgRi+
eDdOR5Dnl5yIB+E7MJE4jyx678XLcorwkYOrQRYwbj0dxx8kI+V5GjKL5GTiomjMdmbDZlgo
kr+JkP/xCcJXD4C7FQcGlqMQkSdhnE9pW33AM3VHWwshOm19rpU0LkuH2goGRg6YFJErj6oW
IXKMFVqRM0WokrVJoXe9P4WxDiZwmRqOT9VSRZlLaw4xRuO0G7k1H46hF584yHh+2KElIwql
h0Yt9gKHnIsTDz/EU9X96Xk6ssqYe6/46wUMEJpicLiW8l+6SCeUfjlSs97rf4108k/cj/7L
6b+c/svpv9zPkk5k5TBHETu1DZYmlzFsii9ng2BlKEDHmQdxbncQei06gYuH1iJ4X4ri7KoR
ZeJFSjIi1vgQx38HrLiezLA5v5ZrCa4QxWKpJwd8nyW49CofVTUVyH8XgxOLBmNs2EcGFlNP
uOZx/RSZS9Vb3ZdzGB0wCuvD7iMtqwhVdXWoEn1E0sW16G9LnF7tluBejo4SwDRhuyOMLUbh
bqHsuvnRO7B2/w0kfyhARW0tyrNTcG5BBwjtxyA8q1bNWZuC0OFOMGs9BMF7w3D29EkcO3wI
oTuDMaytOZyCDiKlWNK8s4ToYJ/uhOG+HIcUFybhzMWMRsuMN7epjn3p/nDci02ibfvwg8t/
eOwKfaD0E14+12bbMD8vMkU/QTj5hEuz2oBjaIm+IVFI11q27iO+q5R/lmEKBJQm9h/XcRDW
n72Lh1HXcXTlILgRPKcluw1mHH+ITFV8jMo8vMcfXgS7KCZ2xkrfbgr9VlNvKkTMKqLncdwx
5WQy3jLG8gV5lZKf0zOIA/b6hcfIrpHjSdHnced9uYaOV5cTiaUdCcbFskPg4oO4Gv0M6W/f
I/P1I2zpbAJupw24n6E5X59yKhR9VWceR39zLjwXXMGbompUFmTg1pYBsOWYo7W7KQw5bTHt
RBKySmsgrilDXuZT3Dq2H+HvKtXscao8JIdR4lQD/4uZBxdDQ9gRLP4qWYeP757B9pk90XXO
g2bjcf8J+W9XCMbbNHlpxUvsIraZic9GJKtiHRSp+3hv8Al+tvBSosa79zztg0qWJTk2SfB0
QftxWH8kHBF3b+P6+UPYMncg2gqsGaVrm4U1/i45NYxQF97Y3O8VunrKWnizufBdHyHDcZ4k
IuZuOELXjEMnc0MYWgVi06NcJp73D+KF//fykjI8WdcePJvBOJBawtjT6oknJi6zEJkvoWXT
TxC/m5kZeEZGsO2zDMciGu7/R7DHP02+OVijDOcn+Ed0Izg/wVVnOgrgt4ec/6dCEUawEMqf
FXlqNfrb8dF2YSTyiD+mSZje/6PMG+If4hkxZTqawzpgBmb6m4Nl4Y8FodcRnURwwviHiDi3
H6sXbkF0gRo+adkDS45EICY+Dg9vhiEkeAXO1pdObCrm2IDc0gW78bQJck3tT1VOHXvUhUU2
93sK/zs6wILgfyOxncJIif/g5KoBcOBaYcCBFwo/za/GE/90OQ3SyZO16GJjjjbTNcts0ed9
zg1MtDeAgUVvrAsn/cZH4syG4XDjWaDXjmRFdrimYpK/S645GKRyXTWOCTe33xbawGCqlpmR
/VQ6A0mTDCjRA8xwdcNsAjDLSCt9Ydt5Bx2VqRGV8eosVo3pDg8bPlhU+Y+WxIFm641+c/bj
QVaVSoqlGny9NgPenZfjYT4TZK3+cgPLejrAxIgDq/YjsPnOF0X92trcBBxZNAjtiPPByMAI
HHNn+Ayaj31RH5DzZCs6mxjCZniYGnhch68XB8PCdqKi/IQ6OFCWugN+pnYYfe5zgySJnyKd
EEB/JQEErIaf11pCgOlENoDVmFv0PH0KGwRrl+m4k6c724i46DmOz+0OZwEbHDNndBmzAZcy
SrSOozrzEHrwtKUKNIDznFiymaZjZ2eOztSDgn6nFQYOlX4zYuME9PK2h5BtgFZGXJg7dcSA
mSG49qqo8UhzOsWTMaM8B2WIvj+zBEO6tIY1ObAMDI1h6tgBA2bvQsRbzXIJUnERnp9ciD6t
zWBM1oxFm0DMPvAIX6sljGdckiRjrHcOSdWZyYbONJBxCWtH+sKRHKgGRiaw8uiNaSE38UbF
WK7OPIe53d1gbkwctSZWcPcfizWnkxVp1RQGatZZ9BcaajC2KabyYuJgtZ90gy7f0uQ1JqlG
UV5J80CKsjyIqppmkFOZDFZ5smE+7KpWkI0mpUxwg8+Gp7KahUSxS9k/DI5sWXo5jmsQVq7p
B8tWBuC79sGiE8nIVy3vIClC3HIPGJv2QejbqgbGwKIN6bcH/GGiuv5aGpLfOqLjwNnYFfGW
QYCg9qb8J6exekw3uFlwYdjKkLwLjmjXayyW7LmOl0XK90e935aGHJjatkbnAZOxMvQOMoqY
QJrGOFQ/tBNUrJQTBOLkp9omAGB1KHh2GqtG+9HjpfYzrpkDvAKCMGfzSTx4V6wgvdD98rWn
+5NlK2I6vxuSl11blvaPbT0Wt1QOLdnveuGo2u/oMmYL28LYuA3m385hjkvHvJgEHJT9Xaic
H23vW+G98bDhMBmbun4nybmJMdYtSf8WGH4jV43YUYu8pBNYGuQLJ1PiJDcWwrZtd4xffRLx
2dUM2cbmR514kpd8EstHdIGzmTEMWlL3x4LzmL14+KVCs19eDzplovp9ih7MgCNVA/VyNplP
EVLOb8GcIH942AlhbNgKhlxzuPgOxYJ9dxnAnCrAHchnRpFpBwyzERs6B308LME1NACbbwuv
wKnYdjUdJToIZ5RC93d3Hnm/zDHgeKbOLD3S2k84Ecgn989DwP43TEa2vCxQi1Y2GK3CvlZt
uVHBaGfCQ8eVjzSyQdFpLQ90B1/YG4fUSIGSnGsIMteeKlOheyQshpthC90pc6mPgTPmxpao
gXYCuC1K0IjWVBiYMTPhaDUcN9SUO7o8jElbLNeV0eUfIp3QZSXCl6CnI9GB2AI4+I7E+ksP
cLS/KUwHh9O11AsjJ8CGzOPhp7E4MKkTbEzY4Nn5YOS6y8goZd5HTXYM9s3ohdYWHBgasMC3
80af2UeQVqbMXlf66hTmBTiCxzKGmVtPzDxwD1FbO4OvhTRZm5eIIwsHyHQ0cg6whQ7wJnvw
8pNPFNkIqCx4e/34jHS4Gucwueb87rJrmrr4Y1LIJVxe3g4WKuncC26PhZWQud/Wp5wW2CjL
vuXfGq0hx2Rwj4GVoDdOfFT+XVvfElEiNvsJIeyyHo/l2d9kcszf0tcTKEkasudWgDtjrSFQ
KSNG91lwF+OsBYpsULLfMvvTdk8yfTgW+6d2gxNFuCZ6SJueU7H9VqaCFFx/f7k3RsBC2AfH
kqOwY3QHWHOJw8i6HYauOI+0YjFz3l+GYS7jWd/FrZXtwG+zVJE1R9v9MdeoCE8PL8KI7p6w
IzqpgQHpy6Urxqy9qLgefa0XxzCjqz14hmRPN1I656i9IOvWavRzFcCI6JSWbXpi2p7LODvV
GQJ5RhiGs4N6fmRfNnKZTYAz7WdNXdYZ9BO0ANt7rYJQS2eZmeNExmevIKIo3otvEZjvwYM1
cSq+ZBCJU7HFhyISH9LqbBRnX8EISw68V8YzMlWpRxho3Z+4XbEno1otUmMJ3HnU+VihpS8d
pBNah+IzSl/pP+lE/53l+i+n/3L6L/dvkU7K8WytN9hmQ3ApW62edm0uHu+bAj9HAqiTc6ul
kQCOnUZh68NceeY/qgxiJ3B06bEC3bpM+btLCO7lAOOWclkjC3gPWYnLH6s1sMBHNBY4DQ91
BL5QWNzakV3hIlRNJd8KxjY+GEXO8heiOgaB4FPUZUS++IriykqI3t3GWj9TmAUeVhC6P5yb
AV/ilGhV31crDhx7LsSp1GKt9oK0Jh/Pzq3D+O6tYWUi03FaB4zH2rNPkfcHZwn5J8f+I6WV
m9toncOoMVuKGVxDk4w2Esem8xQcDZ0HPzsTsHkO8Jt+CA/urEFXcwJw99uJF+WqeBAVdNYa
/U4RjPfLafT3mMDQb5m6DpWNUJfe9OvtrAYx8dK3uPXXfAwlTjZLEyPletb5YWbVpfXbO2vR
x56NlvTfjWDRcRy23/uMMtEzHJ3lBxt2SyU+xbaC97AtiFXTbaks1yMs2LSepxPTrM1BzK6x
8LEhmBjBUs1dumD4koOIzCxrNh73X5B/vKzp8pnHB8LCpAPWJqplt6T8KNPsdT53lkoQpVZs
ksL6LJzRsd9krPr7Lt6UiH8Ma/xdcmrYmS48rbnfy+aKygIxCKaq+CyxHwXWBHfuMw7Be2/g
RWGtZn//IF74X5CvE2XgSWapVp+QpDILKS9k2XYK7i9AO68pOJUhwve4I1gy1AeOwsb7/xHs
8U+Tp1pTsEZ6vVE4+YcaNNSqaccucfhHvsS1+X5wIDiNgSEXVh59iI8qWuGjagqm9/8pU453
4cvQv43MN2fu1gNTd9xBZrmE+CxTcW5FEHwciA7cqhVY5Nm09R+BRQejkaPiM2oMn6RaUzDH
3yWnjj3qwiKb+z3V6vIT8fesHnAl7z+lX7YNnIk9kZ8UmdkUv/+FeOL/g1xTm0SUgC29PDBo
00nsndMTrkK2TE/vMQUhEe/V/ItNwyR/l5xqaxSDrA+IaAIm3Jx+qfZT5XV+ptERwmcnYMrV
nF9eZ/RHWt3XcATZumFejGbmD7oEiJ8l2s67jZw66T80HwQYj5wBV7shOJlZ3fSXougxlnpY
IGBv89KY63/7cxulVOdETICT4xhc+FLzfzZ2Ccqeb0NXm05YEZUjL5kiQXXhZ7x9/w0ltRLU
FGXh8zdRk+pC63/7PU1SkiArEzDqKrLF+v+cmjd3T7C5Mw+8TuuRWCTR/wn5f36WOdcx3MIU
A859QbVYrFVXqXq1Hb6WymhQqYQYTHVVyDzaCwLzIFzL0f818G+0mi/hmOhkjQEnP/wwMVj7
mUbVqmwLnudqPC2X/gf1jTI8XeUJvucqrdnLGv29+CsuDDSF+bArWh1v+t9U5kpSgqS17SFw
nYN7eeJf8OxqZanYzYZq3Wd0kU5qPhxFL4ENxt8p+GH7i3a2jr6C7zV1qKtjfqoyj2PYv0Q6
+VWtqfej/3L6L6f/cvov11y5HzsvcnBt6nic/lL3yzGGStF3fMnKRlG15Beca7UoKyD9ffqC
b/llqNEakJKHuzNbK8kuLdmw770KEZ+rNLDBCmpsn7KQU1L9R+CD+t9+wZoTf8NFoieaDtQd
Ia7xGypQwdcHK56UoSR2Ltr1PPh/pVP89JzVlSP/axa+F1ZolBmQVBGn7+dPBEvLR1mtDiKO
RIya6mrUiPXfBv2/eebV73F46BDsf1Ot/5Oh/+3PW5+1ZSgoq9X/idD/9mdhtQVRWNAvGI9E
+n/W6X/7VfpXIV4lv0fpfwjXbAoGSQdErPZqFibcVGzzN5JOSvB46WjsfFn5hzwIEWIWe8Bl
wk3kqDhZpXV5uB/cAa7DDiG9TPIPHuSfcXqoF4aHfdAZRa7/TU82Oqp80gAvjLuYpbNO3Z+9
qVXg7akJcBM6ot+y44hKy4KovBpVZQX48jIWl/YuxiDvTlj8oED/AaU/rNUUvsfzxzexd7IH
eEJ/bE8t0/9n1OyzTYKyF3sx0M0VfUKe6Ezrpv/tz29lT1bCgyUHyYX9cS6rrpFnX4WMXV3A
lUfgcDqF0HX59b/94ueSdgxb957H3dhkPEmKxo3DyzHQmQeboYeRUSH9Re+xGGVf0hB3dQv6
WXPhuSpZZ7Tg//V+VfsFp/oKYTHiRpPqxGv8vjQeC135jEhN/W9qgEhFNjJS4hDx1zi48V0x
MyK7wVrFTV+j1Xgf2h08bidsjstFVZ2EGa2nRjqRSOpQXfoFUWt8wLcehavZP058oZyt7di6
o3dNAv7/SCdNuR/9l9N/Of2X03+55sr90P5eloTVo7bg+X+E7ErZRlWFWXiX8Rofcsp+yRmo
/+3/5NmXJmCRGw/tN6SioonPXVrxHteOhuNlSR2KX4Tj2M0P+h9cp//tv/l+VOfgyaNYJD4K
w+y2DhgRkaf/2Jv+N/1v+t/0vzXQxEWvERsdj/ibWxHo2BU70qv0f1L0v+l/+wkMUlr3Baeb
iAk3F9v8faQTSRkyY2LxSf8dMvrf9L/99wwoaQ1yEo5iQaAr+AZKAK4VxxadRq7EycdfFSWx
9L/9Kc+MqrFN1US2RNs+c/F3fO7/JelJ/5v+t1/V6ko+IyPtBV68eIG0jC8oayTrD01U+Poa
aS9kv0n/KNL/d+gfaAWxmzGsvaxknwGLD1vvfpj9VyQ+Vfw6YrBU/B2XhliAbdYagQvDkFos
/m/u+8UxmO3ER5fdP5Ytr/r93+gusMWkKJH+L0wdTfRgKuwNDMC17Yzpx1J/KRGxOjMMY5yN
6TTshi7zEVeqnXRCEdvD+ghkehjPC9PPZf6UA0dSkYO3L2X7nLbPq8z8Zmezo5ySYrFmxgDF
R0e2qV9jlDftfvRfTv/l9F9O/+WaK/dDZ3ONCJk/8Xv9b/rf/pRGl+3mW2PsHf0PNtL/pv9N
Q//Mv4cZ7hwYss3hOWoPkkRi/Z8U/W/63/S/6X9roFW83ImeFiwYmjigR/A1fK7Wf91C/5v+
t5/BIGWlKpuGCTcX22yh/49C/5v+N/1v/1SjS+sUZ+Pj29d49ykbxdX6n/pM/5v+N/1v+t/0
v+l/0/+m/+2X6Fm1pcjOzEBGZh6DzMsgnUgqkPPuFdLffEJexZ8JaFdl7EYXru6sAS24XbA7
Q/8jmfS/6X/T/6b/jW7Q/4/+N/1v+t/0v+l/0/+m/03/m/43/W/63/S/6X/789r/AIaJfkFN
Qc4VAAAAAElFTkSuQmCC'
	) base64Decoded asByteArray readStream! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:36'!
dejaVuSansOblique7Data
	"Created using:
	Clipboard default clipboardText:
		((FileStream oldFileNamed: 'AAFonts\DejaVu Sans Oblique 7.txt') contentsOfEntireFile substrings 
			collect: [ :each | each asNumber]) asString
	"
	^#(7 8 3 0 3 7 11 19 26 35 43 46 50 55 60 67 71 76 78 84 91 98 105 112 119 126 133 140 147 154 157 161 168 175 182 187 196 203 211 219 227 234 241 249 257 261 266 274 280 289 297 306 313 322 329 336 344 351 358 368 376 383 391 397 400 405 410 419 422 429 436 443 451 458 463 470 477 481 486 493 497 507 514 521 528 535 541 547 552 558 564 572 579 586 593 599 602 608 615 622 629 635 640 649 656 663 670 677 684 691 698 705 712 719 726 733 740 747 754 761 768 775 782 789 796 803 810 817 824 831 838 845 848 852 858 866 874 882 889 895 901 910 916 922 929 934 943 947 952 959 964 969 977 984 990 993 1000 1004 1010 1016 1027 1038 1045 1051 1058 1065 1072 1079 1086 1093 1104 1112 1119 1126 1133 1140 1144 1149 1154 1159 1167 1175 1184 1193 1202 1211 1220 1227 1236 1243 1250 1257 1264 1271 1278 1285 1292 1299 1306 1313 1320 1327 1337 1344 1351 1358 1365 1372 1376 1381 1386 1391 1398 1405 1412 1419 1426 1433 1440 1447 1454 1460 1466 1472 1478 1485 1492 1499)! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:36'!
dejaVuSansOblique7Form
	"Created using:
	Clipboard default clipboardText:
	 	((ByteArray streamContents:[:s|
			PNGReadWriter
				putForm: (Form fromFileNamed: 'AAFonts\DejaVu Sans Oblique 7.bmp')
				onStream: s]) asString base64Encoded)
	"
	^Form fromBinaryStream: (
'iVBORw0KGgoAAAANSUhEUgAABdwAAAALCAYAAABlNk9UAAA5U0lEQVR4XuV9aVBU19puft2/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'
	) base64Decoded asByteArray readStream! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:36'!
dejaVuSansOblique9Data
	"Created using:
	Clipboard default clipboardText:
		((FileStream oldFileNamed: 'AAFonts\DejaVu Sans Oblique 9.txt') contentsOfEntireFile substrings 
			collect: [ :each | each asNumber]) asString
	"
	^#(9 11 3 0 3 7 12 22 29 40 49 52 57 62 68 77 81 85 88 95 103 111 119 127 135 143 151 159 168 176 179 184 193 202 211 217 229 238 247 257 267 276 284 293 303 308 315 325 332 343 353 364 373 383 392 401 410 419 428 441 451 459 469 474 478 484 490 500 503 510 518 525 533 540 546 555 563 568 574 582 587 599 607 614 623 631 638 645 650 657 665 675 683 692 700 707 709 716 725 732 740 748 754 764 771 778 785 792 799 806 813 820 827 834 841 848 855 862 869 876 883 890 897 904 911 918 925 932 939 946 953 960 963 967 974 983 992 1000 1008 1015 1022 1032 1038 1045 1055 1059 1069 1074 1079 1088 1094 1100 1110 1119 1126 1128 1136 1141 1147 1154 1167 1180 1188 1195 1204 1213 1222 1231 1240 1249 1262 1272 1281 1290 1299 1308 1313 1318 1324 1330 1340 1350 1361 1372 1383 1394 1405 1413 1424 1433 1442 1451 1460 1468 1477 1485 1492 1499 1506 1513 1520 1527 1539 1546 1553 1560 1567 1574 1578 1584 1590 1596 1603 1612 1619 1626 1633 1641 1648 1657 1666 1673 1680 1687 1694 1703 1712 1721)! !

!StrikeFont class methodsFor: 'dejaVu font data' stamp: 'jmv 8/5/2009 22:36'!
dejaVuSansOblique9Form
	"Created using:
	Clipboard default clipboardText:
	 	((ByteArray streamContents:[:s|
			PNGReadWriter
				putForm: (Form fromFileNamed: 'AAFonts\DejaVu Sans Oblique 9.bmp')
				onStream: s]) asString base64Encoded)
	"
	^Form fromBinaryStream: (
'iVBORw0KGgoAAAANSUhEUgAABroAAAAOCAYAAABw4MzRAABUGElEQVR4XuVdd0xU2dveZLNm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'
	) base64Decoded asByteArray readStream! !

!StrikeFont class methodsFor: 'examples' stamp: ''!
example
	"Displays a line of text on the display screen at the location of the cursor.
	Example depends on the strike font file, 'TimesRoman10.strike'. existing."

	(StrikeFont new readFromStrike2: 'NewYork12.sf2')
		displayLine: 'A line of 12-pt text in New York style' at: Sensor cursorPoint
	 
	"StrikeFont example."! !

!StrikeFont class methodsFor: 'accessing' stamp: 'ar 2/3/2002 23:04'!
familyName: aName pointSize: aSize
	"Answer a font (or the default font if the name is unknown) in the specified size."

	^ ((TextStyle named: aName asSymbol) ifNil: [TextStyle default]) fontOfPointSize: aSize! !

!StrikeFont class methodsFor: 'instance creation' stamp: 'ar 2/3/2002 23:06'!
familyName: aName pointSize: aSize emphasized: emphasisCode
	"Create the font with this emphasis"

	^ (self familyName: aName pointSize: aSize) emphasized: emphasisCode! !

!StrikeFont class methodsFor: 'accessing' stamp: 'ar 11/25/2004 15:19'!
familyName: aName size: aSize
	"Answer a font (or the default font if the name is unknown) in the specified size."
	| style |
	style := TextStyle named: aName asSymbol.
	style ifNil: [^(FontSubstitutionDuringLoading forFamilyName: aName pixelSize: aSize)
			signal: 'missing font' ].
	^style fontOfSize: aSize! !

!StrikeFont class methodsFor: 'instance creation' stamp: 'tk 1/28/1999 11:31'!
familyName: aName size: aSize emphasized: emphasisCode
	"Create the font with this emphasis"

	^ (self familyName: aName size: aSize) emphasized: emphasisCode! !

!StrikeFont class methodsFor: 'accessing' stamp: 'sma 12/30/1999 13:48'!
familyNames
	^ (TextConstants select: [:each | each isKindOf: TextStyle]) keys asSortedCollection! !

!StrikeFont class methodsFor: 'instance creation' stamp: 'yo 3/18/2004 00:10'!
fixAccuISO8859From: aStrikeFont

	^aStrikeFont copy fixAccuISO8859From: aStrikeFont.
! !

!StrikeFont class methodsFor: 'instance creation' stamp: 'yo 9/16/2002 15:55'!
fixForISO8859From: aStrikeFont

	^aStrikeFont copy fixForISO8859From: aStrikeFont.
! !

!StrikeFont class methodsFor: 'font creation' stamp: 'ar 6/4/2000 22:27'!
fromHostFont: fontName size: fontSize flags: fontFlags weight: fontWeight
	"
		^StrikeFont fromHostFont: (StrikeFont hostFontFromUser)
					size: 12 flags: 0 weight: 4.
	"
	| fontHandle glyphs xTable xStart maxWidth w glyphForm ascent descent fontHeight |
	fontHandle := self primitiveCreateFont: fontName size: fontSize flags: fontFlags weight: fontWeight.
	ascent := self primitiveFontAscent: fontHandle.
	descent := self primitiveFontDescent: fontHandle.
	fontHeight := ascent + descent.
	xTable := Array new: 258.
	xStart := maxWidth := 0.
	0 to: 255 do:[:i|
		xTable at: i+1 put: xStart.
		w := self primitiveFont: fontHandle widthOfChar: i.
		w > maxWidth ifTrue:[maxWidth := w].
		xStart := xStart + w].
	xTable at: 256 put: xStart.
	xTable at: 257 put: xStart.
	xTable at: 258 put: xStart.
	glyphs := Form extent: xTable last @ fontHeight depth: 1.
	glyphForm := Form extent: maxWidth @ fontHeight depth: 1.
	0 to: 255 do:[:i|
		glyphForm fillWhite.
		self primitiveFont: fontHandle glyphOfChar: i into: glyphForm.
		xStart := xTable at: i+1.
		glyphForm displayOn: glyphs at: xStart@0.
		glyphForm displayOn: Display at: xStart@0.
	].
	self primitiveDestroyFont: fontHandle.
	^Array with: glyphs with: xTable! !

!StrikeFont class methodsFor: 'instance creation' stamp: ''!
fromStrike: fileName 
	"Read a font from disk in the old ST-80 'strike' format.
	Note: this is an old format; use strike2 format instead"

	^self new newFromStrike: fileName! !

!StrikeFont class methodsFor: 'instance creation' stamp: 'ar 1/5/2002 21:41'!
fromUser
	"StrikeFont fromUser"
	^self fromUser: TextStyle defaultFont! !

!StrikeFont class methodsFor: 'instance creation' stamp: 'nk 9/1/2004 14:29'!
fromUser: priorFont
	^self fromUser: priorFont allowKeyboard: true! !

!StrikeFont class methodsFor: 'instance creation' stamp: 'ar 7/30/2009 22:03'!
fromUser: priorFont allowKeyboard: aBoolean	"StrikeFont fromUser"
	"Present a menu of available fonts, and if one is chosen, return it.
	Otherwise return nil. Using ToolBuilder for added abstraction."
	| fontList fontMenu style active ptMenu label font builder resultBlock result item widget |
	builder := ToolBuilder default.
	fontList := StrikeFont actualFamilyNames.
	fontMenu := builder pluggableMenuSpec new.
	resultBlock := [:value| result := value].
	fontList do: [:fontName |
		style := TextStyle named: fontName.
		active := priorFont familyName sameAs: fontName.
		ptMenu := builder pluggableMenuSpec new.
		style pointSizes do: [:pt |
			label := pt printString, ' pt'.
			item := ptMenu add: label 
				target: resultBlock
				selector: #value:
				argumentList: {{fontName. pt}}.
			item checked: (active and:[pt = priorFont pointSize]).
		].
		style isTTCStyle ifTrue: [
			ptMenu add: 'new size'
				target: style 
				selector: #addNewFontSizeDialog: 
				argumentList: {{fontName. fontMenu}}.
		].
		item := fontMenu add: fontName action: nil.
		item subMenu: ptMenu.
		item checked: active.
	].
	TTFileDescription offerNonPortableFonts ifTrue:[
		fontMenu addSeparator.
		item := fontMenu add: 'More (non-portable) Fonts...' 
			target: resultBlock 
			selector: #value:
			argumentList: #(TTFileDescription).
	].
	widget := builder open: fontMenu.
	builder runModal: widget.
	result ifNil:[^nil].
	"Pick a non-portable font if requested"
	result == #TTFileDescription ifTrue:[
		^Smalltalk at: #TTFileDescription ifPresent:[:ttf| 
			ttf fontFromUser: priorFont allowKeyboard: aBoolean]].
	style := TextStyle named: result first.
	style ifNil: [^ self].
	font := style fonts detect: [:any | any pointSize = result last] ifNone: [nil].
	^ font! !

!StrikeFont class methodsFor: 'font creation' stamp: 'rbb 2/18/2005 13:21'!
hostFontFromUser
	"StrikeFont hostFontFromUser"
	| fontNames index labels |
	fontNames := self listFontNames asSortedCollection.
	labels := WriteStream on: (String new: 100).
	fontNames do:[:fn| labels nextPutAll: fn] separatedBy:[labels cr].
	index := (UIManager default chooseFrom: (labels contents substrings) 
				title: 'Choose your font').
	index = 0 ifTrue:[^nil].
	^fontNames at: index! !

!StrikeFont class methodsFor: 'font creation' stamp: 'jmv 8/6/2009 14:24'!
installDejaVu
	"Warning: Uses the methods in 'dejaVu font data' category, that will be removed soon (or are already removed) to save space."
"
StrikeFont installDejaVu
"

	TextConstants at: 'Bitmap DejaVu Sans' put: (TextStyle fontArray: 
		(Array
			with: (self createDejaVu: 7)
			with: (self createDejaVu: 9)
			with: (self createDejaVu: 12))).
	Preferences restoreDefaultFonts.
	StrikeFont limitTo16Bits.
	StrikeFont useUnderscoreIfOver1bpp! !

!StrikeFont class methodsFor: 'removing' stamp: 'jmv 3/31/2009 08:20'!
limitTo16Bits
	"Limit glyph depth to 16 bits (it is usually 16 or 32).
	
	StrikeFont limitTo16Bits
	"
	StrikeFont allInstances do: [ :f | f
		setGlyphsDepthAtMost: 16 ].! !

!StrikeFont class methodsFor: 'font creation' stamp: 'ar 6/4/2000 21:12'!
listFont: index
	<primitive:'primitiveListFont' module:'FontPlugin'>
	^nil! !

!StrikeFont class methodsFor: 'font creation' stamp: 'ar 6/4/2000 21:12'!
listFontNames
	"StrikeFont listFontNames"
	"List all the OS font names"
	| font fontNames index |
	fontNames := WriteStream on: Array new.
	index := 0.
	[font := self listFont: index.
	font == nil] whileFalse:[
		fontNames nextPut: font.
		index := index + 1].
	^fontNames contents! !

!StrikeFont class methodsFor: 'font creation' stamp: 'tak 8/3/2005 21:11'!
localeChanged
	self setupDefaultFallbackFont! !

!StrikeFont class methodsFor: 'character shapes' stamp: 'jmv 3/27/2009 09:28'!
makeControlCharsVisible
	"
	Make normally not visible characters, visible
	StrikeFont makeControlCharsVisible
	"
	self allInstances do: [ :font | font makeControlCharsVisible ]! !

!StrikeFont class methodsFor: 'character shapes' stamp: 'jmv 3/26/2009 14:51'!
makeLfInvisible
	"
	Make line feed characters invisible
	StrikeFont makeLfInvisible
	"
	self allInstances do: [ :font | font makeLfInvisible ]! !

!StrikeFont class methodsFor: 'character shapes' stamp: 'jmv 3/26/2009 14:51'!
makeLfVisible
	"
	Make line feed characters visible
	StrikeFont makeLfVisible
	"
	self allInstances do: [ :font | font makeLfVisible ]! !

!StrikeFont class methodsFor: 'character shapes' stamp: 'jmv 3/26/2009 14:53'!
makeTabInvisible
	"
	Make tab characters invisible
	StrikeFont makeTabInvisible
	"
	self allInstances do: [ :font | font makeTabInvisible ]! !

!StrikeFont class methodsFor: 'character shapes' stamp: 'jmv 3/26/2009 14:53'!
makeTabVisible
	"
	Make tab characters visible
	StrikeFont makeTabVisible
	"
	self allInstances do: [ :font | font makeTabVisible ]! !

!StrikeFont class methodsFor: 'instance creation' stamp: 'yo 8/5/2003 13:11'!
newForJapaneseFromEFontBDFFile: fileName name: aString overrideWith: otherFileName

	| n |
	n := self new.
	n readEFontBDFForJapaneseFromFile: fileName name: aString overrideWith: otherFileName.
	^ n.
! !

!StrikeFont class methodsFor: 'instance creation' stamp: 'yo 1/15/2004 16:48'!
newForKoreanFromEFontBDFFile: fileName name: aString overrideWith: otherFileName

	| n |
	n := self new.
	n readEFontBDFForKoreanFromFile: fileName name: aString overrideWith: otherFileName.
	^ n.
! !

!StrikeFont class methodsFor: 'instance creation' stamp: 'nop 1/23/2000 19:21'!
newFromBDFFile: aFileName name: aString  "StrikeFont newFromBDFFile: 'helvR12.bdf' name: 'Helvetica12'"
	"Read a font from disk in the X11 Bitmap Distribution Format."

	| n |
	n := self new.
	n readBDFFromFile: aFileName name: aString.
	^n.

	"TextConstants at: #Helvetica put: (TextStyle fontArray: {StrikeFont newFromBDFFile: 'helvR12.bdf' name: 'Helvetica12'})"
	"TextConstants at: #Lucida put: (TextStyle fontArray: {StrikeFont newFromBDFFile: 'luRS12.bdf' name: 'Lucida'})"
	"TextStyle default fontAt: 5 put: (StrikeFont new readFromStrike2: 'helv12.sf2')."

! !

!StrikeFont class methodsFor: 'instance creation' stamp: 'yo 1/19/2005 11:22'!
newFromEFontBDFFile: fileName name: aString ranges: ranges

	| n |
	n := self new.
	n readEFontBDFFromFile: fileName name: aString ranges: ranges.
	^ n.
! !

!StrikeFont class methodsFor: 'instance creation' stamp: 'yo 12/27/2002 16:57'!
newFromEFontBDFFile: aFileName name: aString startRange: start endRange: end

	| n |
	n := self new.
	n readEFontBDFFromFile: aFileName name: aString rangeFrom: start to: end.
	^n.

	"TextConstants at: #Helvetica put: (TextStyle fontArray: {StrikeFont newFromBDFFile: 'helvR12.bdf' name: 'Helvetica12'})"
	"TextConstants at: #Lucida put: (TextStyle fontArray: {StrikeFont newFromBDFFile: 'luRS12.bdf' name: 'Lucida'})"
	"TextStyle default fontAt: 5 put: (StrikeFont new readFromStrike2: 'helv12.sf2')."

! !

!StrikeFont class methodsFor: 'instance creation' stamp: 'yo 9/23/2002 16:28'!
newFromF12File: aFileName
	"StrikeFont newFromF12File: 'kaname.f12'"

	| file n |
	('*.F12' match: aFileName) ifFalse: ["self halt. " "likely incompatible"].
	file := FileStream readOnlyFileNamed: aFileName.
	file binary.
	n := self new.
	n name: (FileDirectory baseNameFor: (FileDirectory localNameFor: aFileName)).
	n readF12FromStream: file.
	^ n.
! !

!StrikeFont class methodsFor: 'instance creation' stamp: 'tak 12/20/2004 10:23'!
passwordFontSize: aSize 
	^ FixedFaceFont new passwordFont fontSize: aSize! !

!StrikeFont class methodsFor: 'font creation' stamp: 'ar 6/4/2000 21:13'!
primitiveCreateFont: fontName size: fontSize flags: fontFlags weight: fontWeight
	<primitive:'primitiveCreateFont' module:'FontPlugin'>
	^self primitiveFailed! !

!StrikeFont class methodsFor: 'font creation' stamp: 'ar 6/4/2000 21:13'!
primitiveDestroyFont: fontHandle
	<primitive:'primitiveDestroyFont' module:'FontPlugin'>
	^self primitiveFailed! !

!StrikeFont class methodsFor: 'font creation' stamp: 'ar 6/4/2000 21:14'!
primitiveFont: fontHandle glyphOfChar: charIndex into: glyphForm
	<primitive:'primitiveFontGlyphOfChar' module:'FontPlugin'>
	^self primitiveFailed! !

!StrikeFont class methodsFor: 'font creation' stamp: 'ar 6/4/2000 21:15'!
primitiveFont: fontHandle widthOfChar: charIndex
	<primitive:'primitiveFontWidthOfChar' module:'FontPlugin'>
	^self primitiveFailed! !

!StrikeFont class methodsFor: 'font creation' stamp: 'ar 6/4/2000 22:25'!
primitiveFontAscent: fontHandle
	<primitive:'primitiveFontAscent' module:'FontPlugin'>
	^self primitiveFailed! !

!StrikeFont class methodsFor: 'font creation' stamp: 'ar 6/4/2000 22:25'!
primitiveFontDescent: fontHandle
	<primitive:'primitiveFontDescent' module:'FontPlugin'>
	^self primitiveFailed! !

!StrikeFont class methodsFor: 'font creation' stamp: 'ar 6/4/2000 21:14'!
primitiveFontEncoding: fontHandle
	<primitive:'primitiveFontEncoding' module:'FontPlugin'>
	^self primitiveFailed! !

!StrikeFont class methodsFor: 'examples' stamp: 'tpr 6/10/2005 16:07'!
readStrikeFont2Family: familyName 
	"StrikeFont readStrikeFont2Family: 'Lucida'"
	^self readStrikeFont2Family: familyName fromDirectory: FileDirectory default! !

!StrikeFont class methodsFor: 'examples' stamp: 'tpr 6/10/2005 16:07'!
readStrikeFont2Family: familyName fromDirectory: aDirectory
	"StrikeFont readStrikeFont2Family: 'Lucida' fromDirectory: FileDirectory default"
	"This utility reads all available .sf2 StrikeFont files for a given family from  
	the current directory. It returns an Array, sorted by size, suitable for handing 
	to TextStyle newFontArray: ."
	"For this utility to work as is, the .sf2 files must be named 'familyNN.sf2'."
	| fileNames strikeFonts fontArray |
	fileNames := aDirectory fileNamesMatching: familyName , '##.sf2'.
	strikeFonts := fileNames collect: [:fname | StrikeFont new readFromStrike2: fname].
	strikeFonts do: [ :font | font reset ].
	strikeFonts := strikeFonts asSortedCollection: [:a :b | a height < b height].
	fontArray := strikeFonts asArray.
	^ fontArray

"TextConstants at: #Lucida put: (TextStyle fontArray: (StrikeFont 
	readStrikeFont2Family: 'Lucida'))."! !

!StrikeFont class methodsFor: 'removing' stamp: 'jmv 3/31/2009 08:19'!
saveSpace
	"Removes glyphs over 128, leaving only lower ascii.
	Also limit glyph depth to 4 bits (it is usually 16 or 32).
	This effectively turns off subpixel rendering, as glyphs will only have 16 shades of gray.
	
	StrikeFont saveSpace
	"
	StrikeFont allInstances do: [ :f | f
		stripHighGlyphs;
		setGlyphsDepthAtMost: 4 ].! !

!StrikeFont class methodsFor: 'accessing' stamp: 'tak 11/11/2004 21:14'!
setupDefaultFallbackFont
"
	StrikeFont setupDefaultFallbackFont
"

	(#(#Accuat #Accujen #Accula #Accumon #Accusf #Accushi #Accuve #Atlanta) collect: [:e | TextStyle named: e]) do: [:style |
		style fontArray do: [:e |
			e reset.
			e setupDefaultFallbackFont.
		].
	].
	TTCFont allSubInstances
		do: [:font | font reset.
			font setupDefaultFallbackFont]

! !

!StrikeFont class methodsFor: 'derivative font caching' stamp: 'tak 3/11/2005 16:27'!
shutDown  "StrikeFont shutDown"
	"Deallocate synthetically derived copies of base fonts to save space"
	self allSubInstancesDo: [:sf | sf reset].
	StrikeFontSet allSubInstancesDo: [:sf | sf reset].
	DefaultStringScanner := nil.
! !

!StrikeFont class methodsFor: 'character shapes' stamp: 'jmv 8/6/2009 14:31'!
useUnderscoreIfOver1bpp
	"Sets underscore and caret glyphs for chars 95 and 94. 
	Only for enhanced StrikeFonts, i.e. those with glyphs of more than 1bpp.
	ASCII standard glyphs"
	"
	StrikeFont useUnderscoreIfOver1bpp
	"
	self allInstances do: [ :font | font useUnderscoreIfOver1bpp ]! !

!StrikeFont methodsFor: 'Mac reader' stamp: ''!
aComment
	"To read Mac font resources.  
1) Use ResEdit in the Fonts folder in the System Folder.  Open the file of the Font you want.  (A screen font, not a TrueType outline font).
2) Open the FOND resource and scroll down to the list of sizes and resource numbers. Note the resource number of the size you want.
3) Open the NFNT resource.  Click on the number you have noted.
4) Choose 'Open Using Hex Editor' from the resource editor.
5) Copy all of the hex numbers and paste into a text editor.  Save the file into the Smalltalk folder under the name 'FontName 12 hex' (or other size).
6) Enter the fileName below and execute: 

TextStyle default fontAt: 8 put: (StrikeFont new readMacFontHex: 'fileName').

Select text and type Command-7 to change it to your new font.

(There is some problem in the ParagraphEditor with the large size of Cairo 18.  Its line heights are not the right.)
	"! !

!StrikeFont methodsFor: 'character shapes' stamp: ''!
alter: char formBlock: formBlock
	self characterFormAt: char 
		put: (formBlock value: (self characterFormAt: char))! !

!StrikeFont methodsFor: 'accessing' stamp: ''!
ascent
	"Answer the receiver's maximum extent of characters above the baseline."

	^ascent! !

!StrikeFont methodsFor: 'accessing' stamp: 'di 9/2/2000 13:06'!
ascentKern
	"Return the kern delta for ascenders."
	(emphasis noMask: 2) ifTrue: [^ 0].
	^ (self ascent-5+4)//4 max: 0  "See makeItalicGlyphs"

! !

!StrikeFont methodsFor: 'accessing' stamp: 'yo 1/6/2005 04:19'!
ascentOf: aCharacter

	(self hasGlyphOf: aCharacter) ifFalse: [
		fallbackFont ifNotNil: [
			^ fallbackFont ascentOf: aCharacter.
		].
	].
	^ self ascent.
! !

!StrikeFont methodsFor: 'accessing' stamp: 'jmv 3/19/2009 14:52'!
baseKern
	"Return the base kern value to be used for all characters."
	
	| italic |
	italic _ emphasis allMask: 2.
	
	self isSynthetic ifFalse: [
		^(italic or: [ pointSize < 9 ])
			ifTrue: [ 1 ]
			ifFalse: [ 0] ].
		
	italic ifFalse: [^ 0].
	^ ((self height-1-self ascent+4)//4 max: 0)  "See makeItalicGlyphs"
		+ (((self ascent-5+4)//4 max: 0))! !

!StrikeFont methodsFor: 'emphasis' stamp: 'yo 12/18/2003 23:30'!
bonk: glyphForm with: bonkForm
	"Bonking means to run through the glyphs clearing out black pixels
	between characters to prevent them from straying into an adjacent
	character as a result of, eg, bolding or italicizing"
	"Uses the bonkForm to erase at every character boundary in glyphs."
	| bb offset x |
	offset := bonkForm offset x.
	bb := BitBlt current toForm: glyphForm.
	bb sourceForm: bonkForm; sourceRect: bonkForm boundingBox;
		combinationRule: Form erase; destY: 0.
	x := self xTable.
	(x isMemberOf: SparseLargeTable) ifTrue: [
		x base to: x size-1 do: [:i | bb destX: (x at: i) + offset; copyBits].
	] ifFalse: [
		1 to: x size-1 do: [:i | bb destX: (x at: i) + offset; copyBits].
	].
! !

!StrikeFont methodsFor: 'building' stamp: 'jmv 8/3/2009 13:36'!
buildFromForm: allGlyphs data: data name: aString

	| x |
	pointSize _ data first.
	ascent _ data second.
	descent _ data third.
	
	minAscii _ 32.
	maxAscii _ 255.
	name _ aString.
	type _ 0.  "ignored for now"
	superscript _ ascent - descent // 3.	
	subscript _ descent - ascent // 3.	
	emphasis _ 0.

	xTable _ (Array new: 258) atAllPut: 0.
	maxWidth _ 0.
	glyphs _ allGlyphs.
	x _ 0.
	minAscii to: maxAscii+1 do: [ :i |
		x _ (data at: i-minAscii+4).
		xTable at: i+1 put: x].
	xTable at: 258 put: x.
	self reset.
	derivativeFonts _ Array new: 32! !

!StrikeFont methodsFor: 'file in/out' stamp: 'ar 5/23/2000 12:50'!
buildfontNamed: nm fromForms: forms startingAtAscii: startAscii
	ascent: a descent: d maxWid: m
	"This builds a StrikeFont instance from existing forms."

	| lastAscii width ascii charForm missingForm tempGlyphs |
	name := nm.
	ascent := 11.
	descent := 3.
	maxWidth := 16.
	pointSize := 8.
	name := (name copyWithout: Character space) ,
				(pointSize < 10
					ifTrue: ['0' , pointSize printString]
					ifFalse: [pointSize printString]).
	minAscii := 258.
	maxAscii := 0.
	superscript := ascent - descent // 3.	
	subscript := descent - ascent // 3.	
	emphasis := 0.
	type := 0.  "ignored for now"

	tempGlyphs := Form extent: (maxWidth*257) @ self height.
	xTable := (Array new: 258) atAllPut: 0.
	xTable at: 1 put: 0.

	"Read character forms and blt into tempGlyphs"
	lastAscii := -1.
	1 to: forms size do:
		[:i | charForm := forms at: i. width := charForm width.
		ascii := startAscii-1+i.
		self displayChar: ascii form: charForm.
		ascii = 256
			ifTrue: [missingForm := charForm deepCopy]
			ifFalse:
			[minAscii := minAscii min: ascii.
			maxAscii := maxAscii max: ascii.
			lastAscii+1 to: ascii-1 do: [:as | xTable at: as+2 put: (xTable at: as+1)].
			tempGlyphs copy: ((xTable at: ascii+1)@0
									extent: charForm extent)
						from: 0@0 in: charForm rule: Form over.
			xTable at: ascii+2 put: (xTable at: ascii+1) + width.
			lastAscii := ascii]].
	lastAscii+1 to: maxAscii+1 do: [:as | xTable at: as+2 put: (xTable at: as+1)].
	missingForm == nil ifFalse:
		[tempGlyphs copy: missingForm boundingBox from: missingForm
				to: (xTable at: maxAscii+2)@0 rule: Form over.
		xTable at: maxAscii+3 put: (xTable at: maxAscii+2) + missingForm width].
	glyphs := Form extent: (xTable at: maxAscii+3) @ self height.
	glyphs copy: glyphs boundingBox from: 0@0 in: tempGlyphs rule: Form over.
	xTable := xTable copyFrom: 1 to: maxAscii+3.
	characterToGlyphMap := nil.! !

!StrikeFont methodsFor: 'character shapes' stamp: ''!
characterForm: char pixelValueAt: pt put: val
	| f |
	f := self characterFormAt: char.
	f pixelAt: pt put: val.
	self characterFormAt: char put: val! !

!StrikeFont methodsFor: 'character shapes' stamp: 'yo 12/1/2003 17:01'!
characterFormAt: character 
	"Answer a Form copied out of the glyphs for the argument, character."
	| ascii leftX rightX |
	ascii := character charCode.
	(ascii between: minAscii and: maxAscii) ifFalse: [ascii := maxAscii + 1].
	leftX := xTable at: ascii + 1.
	rightX := xTable at: ascii + 2.
	leftX < 0 ifTrue: [^ glyphs copy: (0@0 corner: 0@self height)].
	^ glyphs copy: (leftX @ 0 corner: rightX @ self height)! !

!StrikeFont methodsFor: 'character shapes' stamp: 'jmv 4/22/2007 20:47'!
characterFormAt: character put: characterForm
	"Copy characterForm over the glyph for the argument, character."
	| ascii leftX rightX widthDif newGlyphs |
	ascii _ character asciiValue.
	ascii < minAscii ifTrue: [^ self error: 'Cant store characters below min ascii'].
	ascii > maxAscii ifTrue:
		[(self confirm:
'This font does not accomodate ascii values higher than ' , maxAscii printString , '.
Do you wish to extend it permanently to handle values up to ' , ascii printString)
			ifTrue: [self extendMaxAsciiTo: ascii]
			ifFalse: [^ self error: 'No change made']].
	leftX _ xTable at: ascii + 1.
	rightX _ xTable at: ascii + 2.
	widthDif _ characterForm width - (rightX - leftX).
	widthDif ~= 0 ifTrue:
		["Make new glyphs with more or less space for this char"
		newGlyphs _ Form extent: (glyphs width + widthDif) @ glyphs height depth: glyphs depth.
		newGlyphs copy: (0@0 corner: leftX@glyphs height)
			from: 0@0 in: glyphs rule: Form over.
		newGlyphs copy: ((rightX+widthDif)@0 corner: newGlyphs width@glyphs height)
			from: rightX@0 in: glyphs rule: Form over.
		glyphs _ newGlyphs.
		"adjust further entries on xTable"
		xTable _ xTable copy.
		ascii+2 to: xTable size
			do: [:i | xTable at: i put: (xTable at: i) + widthDif]].
	glyphs copy: (leftX @ 0 extent: characterForm extent)
		from: 0@0 in: characterForm rule: Form over
"
| f |  f _ TextStyle defaultFont.
f characterFormAt: $  put: (Form extent: (f widthOf: $ )+10@f height)
"! !

!StrikeFont methodsFor: 'character shapes' stamp: 'yo 12/28/2002 22:37'!
characterFormAtMulti: character 
	"Answer a Form copied out of the glyphs for the argument, character."
	| ascii leftX rightX |
	ascii := character charCode.
	(ascii between: minAscii and: maxAscii) ifFalse: [ascii := maxAscii + 1].
	leftX := xTable at: ascii + 1.
	rightX := xTable at: ascii + 2.
	^ glyphs copy: (leftX @ 0 corner: rightX @ self height)! !

!StrikeFont methodsFor: 'accessing' stamp: 'ar 5/23/2000 12:52'!
characterToGlyphMap
	^characterToGlyphMap ifNil:[characterToGlyphMap := self createCharacterToGlyphMap].! !

!StrikeFont methodsFor: 'accessing' stamp: 'ar 5/23/2000 12:52'!
characterToGlyphMap: anArray
	characterToGlyphMap := anArray.! !

!StrikeFont methodsFor: 'displaying' stamp: 'yo 5/19/2004 11:34'!
characters: anInterval in: sourceString displayAt: aPoint clippedBy: clippingRectangle rule: ruleInteger fillColor: aForm kernDelta: kernDelta on: aBitBlt
	"Simple, slow, primitive method for displaying a line of characters.
	No wrap-around is provided."
	| ascii destPoint leftX rightX sourceRect |
	destPoint := aPoint.
	anInterval do: 
		[:i |
		self flag: #yoDisplay.
		"if the char is not supported, fall back to the specified fontset."
		ascii := (sourceString at: i) charCode.
		(ascii < minAscii or: [ascii > maxAscii])
			ifTrue: [ascii := maxAscii].
		leftX := xTable at: ascii + 1.
		rightX := xTable at: ascii + 2.
		sourceRect := leftX@0 extent: (rightX-leftX) @ self height.
		aBitBlt copyFrom: sourceRect in: glyphs to: destPoint.
		destPoint := destPoint + ((rightX-leftX+kernDelta)@0).
		"destPoint printString displayAt: 0@(i*20)"].
	^ destPoint! !

!StrikeFont methodsFor: 'testing' stamp: ''!
checkCharacter: character 
	"Answer a Character that is within the ascii range of the receiver--either 
	character or the last character in the receiver."

	| ascii |  
	ascii := character asciiValue.
	((ascii < minAscii) or: [ascii > maxAscii])
			ifTrue: [^maxAscii asCharacter]
			ifFalse:	[^character]
! !

!StrikeFont methodsFor: 'private' stamp: 'yo 3/11/2005 07:38'!
createCharacterToGlyphMap
        "Private. Create the character to glyph mapping for a font that didn't have any before. This is basically equivalent to what the former setStopCondition did, only based on indexes."

        maxAscii < 256 ifTrue: [^ (1 to: 256) collect: [:i | i - 1]].
        ^ nil.
! !

!StrikeFont methodsFor: 'copying' stamp: 'BG 12/9/2004 17:27'!
deepCopy
 " there is a circular reference from the derivative fonts back to the receiver. It is therefore not possible to make a deep copy. We make a sahllow copy. The method postCopy can be used to modify the shallow copy. " 
  ^self copy! !

!StrikeFont methodsFor: 'emphasis' stamp: 'jmv 3/20/2009 10:27'!
derivativeFont: aStrikeFont at: index

	| newDeriv |
	(aStrikeFont isNil and: [ index = 0 ]) 
		ifTrue: [derivativeFonts := nil. ^ self].
	derivativeFonts ifNil: [derivativeFonts := Array new: 32].
	derivativeFonts size < 32 ifTrue: [
		newDeriv := Array new: 32.
		newDeriv replaceFrom: 1 to: derivativeFonts size with: derivativeFonts.
		derivativeFonts := newDeriv.
	].
	derivativeFonts at: index put: aStrikeFont.! !

!StrikeFont methodsFor: 'accessing' stamp: 'nk 3/15/2004 18:57'!
derivativeFonts
	^derivativeFonts copyWithout: nil! !

!StrikeFont methodsFor: 'accessing' stamp: 'jmv 3/19/2009 15:50'!
descent
	"Answer the receiver's maximum extent of characters below the baseline."

	^pointSize < 9
		ifTrue: [descent-1]
		ifFalse: [descent]! !

!StrikeFont methodsFor: 'accessing' stamp: 'di 9/2/2000 13:06'!
descentKern
	"Return the kern delta for descenders."
	(emphasis noMask: 2) ifTrue: [^ 0].
	^ (self height-1-self ascent+4)//4 max: 0  "See makeItalicGlyphs"

! !

!StrikeFont methodsFor: 'accessing' stamp: 'yo 1/6/2005 04:19'!
descentOf: aCharacter

	(self hasGlyphOf: aCharacter) ifFalse: [
		fallbackFont ifNotNil: [
			^ fallbackFont descentOf: aCharacter.
		].
	].
	^ self descent.
! !

!StrikeFont methodsFor: 'file in/out' stamp: ''!
displayChar: ascii form: charForm
	"Convenience utility used during conversion of BitFont files"
	| m bigForm |
	Display fillBlack: (0@0 extent: 20@14).
	ascii printString displayAt: 0@2.
	charForm width > 0 ifTrue:
		[m := 5.
		bigForm := charForm magnify: charForm boundingBox by: m@m.
		Display border: ((bigForm boundingBox expandBy: m) translateBy: 50@2) width: m.
		bigForm displayAt: 50@2.
		Display fillBlack: ((50@2)+((m*charForm width)@0) extent: 1@(m*self height))].! !

!StrikeFont methodsFor: 'displaying' stamp: 'yo 5/19/2004 11:36'!
displayLine: aString at: aPoint 
	"Display the characters in aString, starting at position aPoint."

	self characters: (1 to: aString size)
		in: aString
		displayAt: aPoint
		clippedBy: Display boundingBox
		rule: Form over
		fillColor: nil
		kernDelta: 0
		on: (BitBlt current toForm: Display).
! !

!StrikeFont methodsFor: 'displaying' stamp: 'ar 8/15/2009 10:55'!
displayMultiString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: baselineY

	| leftX rightX glyphInfo char destY form gfont destX destPt |
	destX := aPoint x.
	charIndex := startIndex.
	glyphInfo := Array new: 5.
	startIndex to: stopIndex do:[:charIndex|
		char := aString at: charIndex.
		(self hasGlyphOf: char) ifTrue: [
			self glyphInfoOf: char into: glyphInfo.
			form := glyphInfo at: 1.
			leftX := glyphInfo at: 2.
			rightX := glyphInfo at: 3.
			destY := glyphInfo at: 4.
			gfont := glyphInfo at: 5.
			(gfont == aBitBlt lastFont) ifFalse: [gfont installOn: aBitBlt].
			destY := baselineY - destY. 
			aBitBlt displayGlyph: form at: destX @ destY left: leftX right: rightX font: self.
			destX := destX + (rightX - leftX + kernDelta).
		] ifFalse:[
			destPt := self fallbackFont displayString: aString on: aBitBlt from: charIndex to: charIndex at: destX @ aPoint y kern: kernDelta from: self baselineY: baselineY.
			destX := destPt x.
		].
	].
	^destX @ aPoint y
! !

!StrikeFont methodsFor: 'displaying' stamp: 'ar 4/10/2005 18:06'!
displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta
	"Draw the given string from startIndex to stopIndex 
	at aPoint on the (already prepared) BitBlt."
	
	(aString isByteString) ifFalse: [^ self displayMultiString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: aPoint y + self ascent.].

	^ aBitBlt displayString: aString 
			from: startIndex 
			to: stopIndex 
			at: aPoint 
			strikeFont: self
			kern: kernDelta.! !

!StrikeFont methodsFor: 'displaying' stamp: 'ar 4/10/2005 18:06'!
displayString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: baselineY
	"Draw the given string from startIndex to stopIndex 
	at aPoint on the (already prepared) BitBlt."
	
	(aString isByteString) ifFalse:[^ self displayMultiString: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta baselineY: baselineY.].

	^ aBitBlt displayString: aString 
			from: startIndex 
			to: stopIndex 
			at: aPoint 
			strikeFont: self
			kern: kernDelta.! !

!StrikeFont methodsFor: 'displaying' stamp: 'yo 12/20/2002 18:54'!
displayStringR2L: aString on: aBitBlt from: startIndex to: stopIndex at: aPoint kern: kernDelta
	"You are screwed if you reach this method."
	self halt.
	aBitBlt displayString: aString 
			from: startIndex 
			to: stopIndex 
			at: aPoint 
			strikeFont: self
			kern: kernDelta.! !

!StrikeFont methodsFor: 'character shapes' stamp: 'btr 11/18/2002 15:00'!
edit: character 
	"Open a Bit Editor on the given character. Note that you must do an accept 
	(in the option menu of the bit editor) if you want this work. 
	Accepted edits will not take effect in the font until you leave or close the bit editor. 
	Also note that unaccepted edits will be lost when you leave or close."
	"Note that BitEditor only works in MVC currently."

	"(TextStyle default fontAt: 1) edit: $:="

	| charForm editRect scaleFactor bitEditor savedForm r |
	charForm := self characterFormAt: character.
	editRect := BitEditor locateMagnifiedView: charForm scale: (scaleFactor := 8 @ 8).
	bitEditor := BitEditor
				bitEdit: charForm
				at: editRect topLeft
				scale: scaleFactor
				remoteView: nil.
	savedForm := Form fromDisplay: (r := bitEditor displayBox
							expandBy: (0 @ 23 corner: 0 @ 0)).
	bitEditor controller startUp.
	bitEditor release.
	savedForm displayOn: Display at: r topLeft.
	self characterFormAt: character put: charForm! !

!StrikeFont methodsFor: 'emphasis' stamp: ''!
emphasis
	"Answer the integer code for synthetic bold, italic, underline, and 
	strike-out."

	^emphasis! !

!StrikeFont methodsFor: 'emphasis' stamp: ''!
emphasis: code 
	"Set the integer code for synthetic bold, itallic, underline, and strike-out, 
	where bold=1, italic=2, underlined=4, and struck out=8."

	emphasis := code! !

!StrikeFont methodsFor: 'emphasis' stamp: 'di 9/3/2000 13:22'!
emphasized: code 
	"Answer a copy of the receiver with emphasis set to include code."
	| derivative addedEmphasis base safeCode |
	code = 0 ifTrue: [^ self].
	(derivativeFonts == nil or: [derivativeFonts size = 0]) ifTrue: [^ self].
	derivative := derivativeFonts at: (safeCode := code min: derivativeFonts size).
	derivative == nil ifFalse: [^ derivative].  "Already have this style"

	"Dont have it -- derive from another with one with less emphasis"
	addedEmphasis := 1 bitShift: safeCode highBit - 1.
	base := self emphasized: safeCode - addedEmphasis.  "Order is Bold, Ital, Under, Narrow"
	addedEmphasis = 1 ifTrue:   "Compute synthetic bold version of the font"
		[derivative := (base copy ensureCleanBold name: base name , 'B') makeBoldGlyphs].
	addedEmphasis = 2 ifTrue:   "Compute synthetic italic version of the font"
		[ derivative := (base copy name: base name , 'I') makeItalicGlyphs].
	addedEmphasis = 4 ifTrue:   "Compute underlined version of the font"
		[derivative := (base copy name: base name , 'U') makeUnderlinedGlyphs].
	addedEmphasis = 8 ifTrue:   "Compute narrow version of the font"
		[derivative := (base copy name: base name , 'N') makeCondensedGlyphs].
	addedEmphasis = 16 ifTrue:   "Compute struck-out version of the font"
		[derivative := (base copy name: base name , 'X') makeStruckOutGlyphs].
	derivative emphasis: safeCode.
	derivativeFonts at: safeCode put: derivative.
	^ derivative! !

!StrikeFont methodsFor: 'character shapes' stamp: 'jmv 4/23/2007 22:30'!
ensureCleanBold 
	"This ensures that all character glyphs have at least one pixel of white space on the right
	so as not to cause artifacts in neighboring characters in bold or italic."

	| wider glyph |
	emphasis = 0 ifFalse: [^ self].
	minAscii to: maxAscii do:
		[:i | glyph _ self characterFormAt: (Character value: i).
		(glyph copy: (glyph boundingBox topRight - (1@0)
					corner: glyph boundingBox bottomRight)) isAllWhite ifFalse: [
			wider _ Form extent: (glyph width + 1)@glyph height depth: glyph depth.
			glyph depth > 1 ifTrue: [wider fillWhite].
			glyph displayOn: wider.
			self characterFormAt: (Character value: i) put: wider]].
"
StrikeFont allInstancesDo: [:f | f ensureCleanBold].
(StrikeFont familyName: 'NewYork' size: 21) ensureCleanBold.
StrikeFont shutDown.  'Flush synthetic fonts'.
"
! !

!StrikeFont methodsFor: 'character shapes' stamp: 'ar 5/23/2000 12:48'!
extendMaxAsciiTo: newMax
	"Extend the range of this font so that it can display glyphs up to newMax."

	(newMax+3) <= xTable size ifTrue: [^ self].  "No need to extend."
	xTable size = (maxAscii+3) ifFalse:
		[^ self error: 'This font is not well-formed.'].

	"Insert a bunch of zero-width characters..."
	xTable := (xTable copyFrom: 1 to: maxAscii+2) ,
			((maxAscii+1 to: newMax) collect: [:i | xTable at: maxAscii+2]) ,
			{ xTable at: maxAscii+3 }.
	maxAscii := newMax.
	self fillZeroWidthSlots.
	characterToGlyphMap := nil.! !

!StrikeFont methodsFor: 'accessing' stamp: 'ar 8/15/2009 11:25'!
fallbackFont
	"Answers the fallbackFont for the receiver. The fallback font must be some derivative of the receiver since it will not be asked to install itself properly on the target BitBlt so rendering a completely different font here is simply not possible. The default implementation uses a synthetic font that maps all characters to question marks."
	^ fallbackFont
		ifNil: [fallbackFont := FixedFaceFont new errorFont baseFont: self]! !

!StrikeFont methodsFor: 'accessing' stamp: 'yo 5/20/2004 11:01'!
fallbackFont: aFontSetOrNil

	fallbackFont := aFontSetOrNil.
! !

!StrikeFont methodsFor: 'accessing' stamp: 'jmv 8/6/2009 09:14'!
familyName
	^self name withoutJustTrailingDigits! !

!StrikeFont methodsFor: 'accessing' stamp: 'tk 6/26/1998 16:45'!
familySizeFace
	"Answer an array with familyName, a String, pointSize, an Integer, and
	faceCode, an Integer."

	^Array with: name
		with: self height
		with: emphasis

	"(1 to: 12) collect: [:x | (TextStyle default fontAt: x) familySizeFace]"! !

!StrikeFont methodsFor: 'character shapes' stamp: 'di 3/27/2000 16:10'!
fillZeroWidthSlots
	| nullGlyph |
	"Note: this is slow because it copies the font once for every replacement."

	nullGlyph := (Form extent: 1@glyphs height) fillGray.
	"Now fill the empty slots with narrow box characters."
	minAscii to: maxAscii do:
		[:i | (self widthOf: (Character value: i)) = 0 ifTrue:
			[self characterFormAt: (Character value: i) put: nullGlyph]].
! !

!StrikeFont methodsFor: 'multibyte character methods' stamp: 'yo 3/18/2004 00:10'!
fixAccuISO8859From: aStrikeFont

	| f |
	self reset.
	xTable := aStrikeFont xTable copy.
	glyphs := Form extent: aStrikeFont glyphs extent.
	maxAscii := 255.
	minAscii := 0.
	"stopConditions := nil."

	0 to: 127 do: [:i |
		f := aStrikeFont characterFormAt: (Character value: i) isoToSqueak.
		f width  = 0 ifTrue: [f := Form extent: 1@f height].
		
		self characterFormAt: (Character value: i) put: f.
	].
	128 to: 159 do: [:i |
		f := Form extent: 1@f height.
		self characterFormAt: (Character value: i) put: f.
	].
	160 to: 255 do: [:i |
		f := aStrikeFont characterFormAt: (Character value: i) isoToSqueak.
		f width  = 0 ifTrue: [f := Form extent: 1@f height].
		
		self characterFormAt: (Character value: i) put: f.
	].
		
	^ self.	
! !

!StrikeFont methodsFor: 'multibyte character methods' stamp: 'yo 11/12/2002 12:56'!
fixAscent: a andDescent: d head: h

	| bb newGlyphs |
	"(a + d) = (ascent + descent) ifTrue: ["
		ascent := a.
		descent := d.
		newGlyphs := Form extent: (glyphs width@(h + glyphs height)).
		bb := BitBlt toForm: newGlyphs.
		bb copy: (0@h extent: (glyphs extent)) from: 0@0 in: glyphs
			fillColor: nil rule: Form over.
		glyphs := newGlyphs.
	"]."
! !

!StrikeFont methodsFor: 'multibyte character methods' stamp: 'yo 9/16/2002 15:07'!
fixForISO8859From: aStrikeFont

	| fixer m mappingTable |
	fixer := StrikeFontFixer newOn: aStrikeFont.
	self reset.
	xTable := aStrikeFont xTable copy.
	glyphs := Form extent: aStrikeFont glyphs extent.
	maxAscii := 255.
	minAscii := 0.
	mappingTable := fixer mappingTable.
	"stopConditions := nil."

	0 to: 255 do: [:i |
		(m := mappingTable at: i+1) ~= nil ifTrue: [
			self characterFormAt: (Character value: i)
				put: (aStrikeFont characterFormAt: (Character value: m)).
		] ifFalse: [
			self characterFormAt: (Character value: i)
				put: (aStrikeFont characterFormAt: (Character space)).
		]
	].
	^self.	
! !

!StrikeFont methodsFor: 'Mac reader' stamp: ''!
fixKerning: extraWidth
	"Insert one pixel (extraWidth) between each character.  And add the bits for the space character"
	"Create a space character Form.  Estimate width by ascent / 2 - 1"
	| characterForm char leftX |
	characterForm := Form extent: (ascent//2 - 1) @ self height.
	self characterFormAt: $  put: characterForm.

	"Put one pixel of space after every character.  Mac fonts have no space in the bitmap."
	extraWidth <= 0 ifTrue: [^ self].
	minAscii to: maxAscii do: [:ascii |
		char := Character value: ascii.
		leftX := xTable at: ascii + 1.
		characterForm := Form extent: 
			((self widthOf: char) + extraWidth) @ self height.
		characterForm 
			copy: (characterForm boundingBox extendBy: 
				(0-extraWidth@0))
			from: leftX@0 in: glyphs rule: Form over.
		self characterFormAt: char put: characterForm.
		].	! !

!StrikeFont methodsFor: 'character shapes' stamp: 'di 4/28/2000 16:10'!
fixOneWideChars 
	"This fixes all 1-wide characters to be 2 wide with blank on the right
	so as not to cause artifacts in neighboring characters in bold or italic."
	| twoWide |
	minAscii to: maxAscii do:
		[:i | (self widthOf: (Character value: i)) = 1 ifTrue:
			[twoWide := Form extent: 2@glyphs height.
			(self characterFormAt: (Character value: i)) displayOn: twoWide at: 0@0.
			self characterFormAt: (Character value: i) put: twoWide]].
"
StrikeFont allInstancesDo: [:f | f fixOneWideChars].
StrikeFont shutDown.  'Flush synthetic fonts'.
"
! !

!StrikeFont methodsFor: 'multibyte character methods' stamp: 'yo 8/28/2002 16:35'!
fixXTable

	| newXTable val |
	xTable size >= 258 ifTrue: [
		^ self.
	].

	newXTable := Array new: 258.
	1 to: xTable size do: [:i |
		newXTable at: i put: (xTable at: i).
	].

	val := xTable at: (xTable size).
	
	xTable size + 1 to: 258 do: [:i |
		newXTable at: i put: val.
	].
	minAscii := 0.
	maxAscii := 255.
	xTable := newXTable.
! !

!StrikeFont methodsFor: 'displaying' stamp: 'BG 3/16/2005 08:27'!
fontDisplay
	"TextStyle default defaultFont fontDisplay."

	Display restoreAfter:
		[(Form extent: 440@400) displayAt: 90@90.
		 0 to: 15 do:
			[:i |
			i storeStringHex displayAt: 100 @ (20 * i + 100).
			0 to: 15 do:
				[:j |
				((16*i+j) between: 1 and: (self xTable size - 2)) ifTrue:
					[(self characterFormAt: (16 * i + j) asCharacter)
						displayAt: (20 * j + 150) @ (20 * i + 100)]]].
			'Click to continue...' asDisplayText displayAt: 100@450]! !

!StrikeFont methodsFor: 'accessing' stamp: 'ar 9/21/2000 11:53'!
fontNameWithPointSize
	^self name withoutTrailingDigits, ' ', self pointSize printString! !

!StrikeFont methodsFor: 'accessing' stamp: 'yo 1/7/2005 11:15'!
glyphInfoOf: aCharacter into: glyphInfoArray
	"Answer the width of the argument as a character in the receiver."

	| code |
	(self hasGlyphOf: aCharacter) ifFalse: [
		fallbackFont ifNotNil: [
			^ fallbackFont glyphInfoOf: aCharacter into: glyphInfoArray.
		].
		code := 0.
	] ifTrue: [
		code := aCharacter charCode.
	].
	glyphInfoArray at: 1 put: glyphs;
		at: 2 put: (xTable at: code + 1);
		at: 3 put: (xTable at: code + 2);
		at: 4 put: (self ascentOf: aCharacter);
		at: 5 put: self.
	^ glyphInfoArray.
! !

!StrikeFont methodsFor: 'accessing' stamp: 'yo 1/6/2005 04:19'!
glyphOf: aCharacter 
	"Answer the width of the argument as a character in the receiver."

	| code |
	(self hasGlyphOf: aCharacter) ifFalse: [
		fallbackFont ifNotNil: [
			^ fallbackFont glyphOf: aCharacter.
		].
		^ (Form extent: 1@self height) fillColor: Color white
	].
	code := aCharacter charCode.
	^ glyphs copy: (((xTable at: code + 1)@0) corner: (xTable at: code +2)@self height).
! !

!StrikeFont methodsFor: 'accessing' stamp: ''!
glyphs
	"Answer a Form containing the bits representing the characters of the 
	receiver."

	^glyphs! !

!StrikeFont methodsFor: 'multibyte character methods' stamp: 'yo 1/6/2005 04:18'!
hasGlyphOf: aCharacter

	| code |
	code := aCharacter charCode.
	((code between: self minAscii and: self maxAscii) not) ifTrue: [
		^ false.
	].
	(xTable at: code + 1) < 0 ifTrue: [
		^ false.
	].
	^ true.
! !

!StrikeFont methodsFor: 'accessing' stamp: ''!
height
	"Answer the height of the receiver, total of maximum extents of 
	characters above and below the baseline."

	^self ascent + self descent! !

!StrikeFont methodsFor: 'accessing' stamp: 'yo 1/6/2005 04:19'!
heightOf: aCharacter

	(self hasGlyphOf: aCharacter) ifFalse: [
		fallbackFont ifNotNil: [
			^ fallbackFont heightOf: aCharacter.
		].
	].
	^ self height.
! !

!StrikeFont methodsFor: 'displaying' stamp: 'yo 1/5/2005 13:59'!
installOn: aDisplayContext

	^aDisplayContext installStrikeFont: self.
! !

!StrikeFont methodsFor: 'displaying' stamp: 'ar 5/19/2000 15:08'!
installOn: aDisplayContext foregroundColor: foregroundColor backgroundColor: backgroundColor
	^aDisplayContext 
		installStrikeFont: self
		foregroundColor: foregroundColor 
		backgroundColor: backgroundColor! !

!StrikeFont methodsFor: 'emphasis' stamp: 'jmv 4/26/2008 14:12'!
isSynthetic
	^type = 3! !

!StrikeFont methodsFor: 'emphasis' stamp: 'jmv 4/26/2008 14:11'!
isSynthetic: aBoolean
	type _ aBoolean ifTrue: [3] ifFalse: [0]! !

!StrikeFont methodsFor: 'private' stamp: 'yo 5/20/2004 10:51'!
leftAndRighOrNilFor: char

	| code leftX |
	code := char charCode.
	((code between: self minAscii and: self maxAscii) not) ifTrue: [
		code := $? charCode.
	].
	leftX := xTable at: code + 1.
	leftX < 0 ifTrue: [
		code := $? charCode.
		leftX := xTable at: code + 1.
	].
	^ Array with: leftX with: (xTable at: code + 2).
! !

!StrikeFont methodsFor: 'accessing' stamp: ''!
lineGrid
	^ ascent + descent! !

!StrikeFont methodsFor: 'make arrows' stamp: 'sps 10/15/2003 17:06'!
makeAssignArrow
"Replace the underline character with an arrow for this font"

	| arrowForm arrowCanvas arrowY arrowLeft arrowRight arrowHeadLength |

	arrowForm _ (self characterFormAt: $_) copy.
	arrowCanvas _ arrowForm getCanvas.
	arrowCanvas fillColor: Color white.
	arrowY _ arrowForm height // 2.
	arrowLeft _ 0. 
	arrowRight _ arrowForm width - 2.
	arrowHeadLength _ (arrowRight - arrowLeft) * 2 // 5.
	"Draw the lines"
	arrowCanvas line: (arrowLeft@arrowY) to: (arrowRight@arrowY) color: Color black.
	arrowCanvas 
		line: (arrowLeft@arrowY) 
		to: ((arrowLeft + arrowHeadLength)@(arrowY - arrowHeadLength)) 
		color: Color black.
	arrowCanvas 
		line: (arrowLeft@arrowY) 
		to: ((arrowLeft + arrowHeadLength)@(arrowY + arrowHeadLength)) 
		color: Color black.

	"Replace the glyph"
	self characterFormAt: $_ put: arrowForm.

! !

!StrikeFont methodsFor: 'emphasis' stamp: 'jmv 8/3/2009 09:17'!
makeBoldGlyphs
	"Make a bold set of glyphs with same widths by ORing 1 bit to the right
		(requires at least 1 pixel of intercharacter space)"
	| g bonkForm |
	g _ glyphs deepCopy.
	bonkForm _ (Form extent: 1@16) fillBlack offset: -1@0.
	self bonk: g with: bonkForm.
	glyphs depth = 1 ifTrue: [
		g copyBits: g boundingBox from: g at: (1@0)
			clippingBox: g boundingBox rule: Form under fillColor: nil]
		ifFalse: [
			0 to: g width - 2 do: [ :x | 0 to: g height-1 do: [ :y |
				(glyphs colorAt:  x@y) = Color white ifFalse: [
					g colorAt: x+1@y put: 
						((glyphs colorAt: x+1@y) = Color white 
							ifTrue: [glyphs colorAt:  x@y]
							ifFalse: [Color black])]]]].
	glyphs _ g.
	self isSynthetic: true.
	fallbackFont ifNotNil: [
		fallbackFont _ fallbackFont emphasized: 1
	].! !

!StrikeFont methodsFor: 'character shapes' stamp: 'RAA 7/6/2000 16:50'!
makeCarriageReturnsWhite
	| crForm |
	"Some larger fonts have a gray carriage return (from the zero wide fixup) make it white so it doesn't show"

	crForm := self characterFormAt: 13 asCharacter.
	crForm fillWhite.
	self characterFormAt: 13 asCharacter put: crForm.
! !

!StrikeFont methodsFor: 'emphasis' stamp: 'jmv 8/3/2009 09:17'!
makeCondensedGlyphs
	"Make a condensed set of glyphs with same widths.
	NOTE: this has been superceded by kerning -- should not get called"
	| g newXTable x x1 w |
	g _ glyphs deepCopy.
	newXTable _ Array new: xTable size.
	newXTable at: 1 put: (x _ xTable at: 1).
	1 to: xTable size-1 do:
		[:i | x1 _ xTable at: i.  w _ (xTable at: i+1) - x1.
		w > 1 ifTrue: [w _ w-1].  "Shrink every character wider than 1"
		g copy: (x@0 extent: w@g height) from: x1@0 in: glyphs rule: Form over.
		newXTable at: i+1 put: (x _ x + w)].
	xTable _ newXTable.
	glyphs _ g.
	self isSynthetic: true.
	fallbackFont ifNotNil: [
		fallbackFont emphasized: 8
	].

"
(TextStyle default fontAt: 1) copy makeCondensedGlyphs
	displayLine: 'The quick brown fox jumps over the lazy dog'
	at: Sensor cursorPoint
"! !

!StrikeFont methodsFor: 'character shapes' stamp: 'jmv 3/27/2009 09:46'!
makeControlCharsVisible
	| glyph d|
	self characterToGlyphMap.
	glyph _ self characterFormAt: (Character space).
	glyph border: glyph boundingBox width: 1 fillColor: Color blue.
	self characterFormAt: (Character value: 133) put: glyph.
	
	"Keep tab(9), lf(10), cr(13) and space(32) transparent or whatever the user chose"
	#(0 1 2 3 4 5 6 7 8 11 12 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31)
		do: [ :ascii |
			characterToGlyphMap at: ascii + 1 put: 133 ]! !

!StrikeFont methodsFor: 'emphasis' stamp: 'jmv 8/3/2009 09:19'!
makeItalicGlyphs
	"Make an italic set of glyphs with same widths by skewing left and right.
	In the process, characters would overlap, so we widen them all first.
	"
	| extraWidth newGlyphs newXTable x newX w extraOnLeft |  
	extraOnLeft _ (self height-1-self ascent+4)//4 max: 0.
	extraWidth _ ((self ascent-5+4)//4 max: 0) + extraOnLeft.
	newGlyphs _ Form extent: (glyphs width + (maxAscii + 1 - minAscii*extraWidth)) @ glyphs height depth: glyphs depth.
	newGlyphs fillWhite.
	newXTable _ xTable copy.

	"Copy glyphs into newGlyphs with room on left and right for overlap."
	minAscii to: maxAscii+1 do:
		[:ascii | x _ xTable at: ascii+1.  w _ (xTable at: ascii+2) - x.
		newX _ newXTable at: ascii+1.
		newGlyphs copy: ((newX + extraOnLeft) @ 0 extent: w @ glyphs height)
			from: x @ 0 in: glyphs rule: Form over.
		newXTable at: ascii+2 put: newX + w + extraWidth].		
	glyphs _ newGlyphs. 
	xTable _ newXTable.
	"Slide the bitmaps left and right for synthetic italic effect."
	4 to: self ascent-1 by: 4 do:
		[:y | 		"Slide ascenders right..."
		glyphs copy: (1@0 extent: glyphs width @ (self ascent - y))
			from: 0@0 in: glyphs rule: Form over].
	self ascent to: self height-1 by: 4 do:
		[:y | 		"Slide descenders left..."
		glyphs copy: (0@y extent: glyphs width @ glyphs height)
			from: 1@y in: glyphs rule: Form over].
	self isSynthetic: true.
	fallbackFont ifNotNil: [
		fallbackFont _ fallbackFont emphasized: 2
	].

! !

!StrikeFont methodsFor: 'character shapes' stamp: 'jmv 3/26/2009 14:40'!
makeLfInvisible
	self characterToGlyphMap.
	characterToGlyphMap at: 11 put: (11 < minAscii ifFalse: [11] ifTrue: [maxAscii+1])! !

!StrikeFont methodsFor: 'character shapes' stamp: 'jmv 3/27/2009 09:19'!
makeLfVisible
	| glyph |
	self characterToGlyphMap.
	glyph _ self characterFormAt: (Character value: 163).
	glyph border: glyph boundingBox width: 1 fillColor: Color blue.
"	glyph _ glyph reverse."
	self characterFormAt: (Character value: 132) put: glyph.
	characterToGlyphMap at: 11 put: 132! !

!StrikeFont methodsFor: 'make arrows' stamp: 'sps 10/15/2003 17:06'!
makeReturnArrow
"Replace the caret character with an arrow"

	| arrowForm arrowCanvas arrowHeadLength arrowX arrowTop arrowBottom |

	arrowForm := (self characterFormAt: $^) copy.
	arrowCanvas := arrowForm getCanvas.
	arrowCanvas fillColor: Color white.

	arrowHeadLength := ((arrowForm width - 2)// 2).
	arrowX := (arrowHeadLength max: (arrowForm width // 2)).
	arrowTop := arrowForm height // 4. 
	arrowBottom := (arrowTop + (arrowForm width * 4 // 5 )).
	arrowBottom := (arrowBottom min: arrowForm height) max: (arrowForm height * 2 // 3).

	"Draw the lines"
	arrowCanvas line: (arrowX@arrowTop) to: (arrowX@arrowBottom) color: Color black.
	arrowCanvas 
		line: (arrowX@arrowTop) 
		to: ((arrowX - arrowHeadLength)@(arrowTop + arrowHeadLength)) 
		color: Color black.
	arrowCanvas 
		line: (arrowX@arrowTop) 
		to: ((arrowX + arrowHeadLength)@(arrowTop + arrowHeadLength)) 
		color: Color black.

	"Replace the glyph"
	self characterFormAt: $^ put: arrowForm.

! !

!StrikeFont methodsFor: 'emphasis' stamp: 'jmv 8/3/2009 09:20'!
makeStruckOutGlyphs
	"Make a struck-out set of glyphs with same widths"
	| g |
	g _ glyphs deepCopy.
	g fillBlack: (0 @ (self ascent - (self ascent//3)) extent: g width @ 1).
	glyphs _ g.
	self isSynthetic: true.
	fallbackFont ifNotNil: [
		fallbackFont _ fallbackFont emphasized: 16
	].
! !

!StrikeFont methodsFor: 'character shapes' stamp: 'jmv 3/26/2009 14:40'!
makeTabInvisible
	self characterToGlyphMap.
	characterToGlyphMap at: 10 put: (10 < minAscii ifFalse: [10] ifTrue:[maxAscii+1])! !

!StrikeFont methodsFor: 'character shapes' stamp: 'jmv 3/26/2009 14:41'!
makeTabVisible
	self characterToGlyphMap.
	characterToGlyphMap at: 10 put: 172! !

!StrikeFont methodsFor: 'emphasis' stamp: 'jmv 8/3/2009 09:20'!
makeUnderlinedGlyphs
	"Make an underlined set of glyphs with same widths"
	| g |
	g _ glyphs deepCopy.
	g fillBlack: (0 @ (self ascent+1) extent: g width @ 1).
	glyphs _ g.
	self isSynthetic: true.
	fallbackFont ifNotNil: [
		fallbackFont _ fallbackFont emphasized: 4
	].
! !

!StrikeFont methodsFor: 'accessing' stamp: ''!
maxAscii
	"Answer the integer that is the last Ascii character value of the receiver."

	^maxAscii! !

!StrikeFont methodsFor: 'accessing' stamp: ''!
maxWidth
	"Answer the integer that is the width of the receiver's widest character."

	^maxWidth! !

!StrikeFont methodsFor: 'accessing' stamp: ''!
minAscii
	"Answer the integer that is the first Ascii character value of the receiver."

	^minAscii! !

!StrikeFont methodsFor: 'accessing' stamp: 'ls 3/27/2000 19:54'!
name
	"Answer the receiver's name."

	^name ifNil: ['(unnamed)']! !

!StrikeFont methodsFor: 'accessing' stamp: ''!
name: aString
	"Set the receiver's name."

	name := aString! !

!StrikeFont methodsFor: 'file in/out' stamp: 'md 11/14/2003 17:25'!
newFromStrike: fileName
	"Build an instance from the strike font file name. The '.strike' extension
	is optional."

	| strike startName raster16 |
	name := fileName copyUpTo: $..	"assumes extension (if any) is '.strike'"
	strike := FileStream readOnlyFileNamed: name, '.strike.'.
	strike binary.

	"strip off direcory name if any"
	startName := name size.
	[startName > 0 and: [((name at: startName) ~= $>) & ((name at: startName) ~= $])]]
		whileTrue: [startName := startName - 1].
	name := name copyFrom: startName+1 to: name size.

	type			:=		strike nextWord.		"type is ignored now -- simplest
												assumed.  Kept here to make
												writing and consistency more
												straightforward."
	minAscii		:=		strike nextWord.
	maxAscii		:=		strike nextWord.
	maxWidth		:=		strike nextWord.
	strikeLength	:=		strike nextWord.
	ascent			:=		strike nextWord.
	descent			:=		strike nextWord.
	"xOffset			:="		strike nextWord. 	
	raster16			:=		strike nextWord.	
	superscript		:=		ascent - descent // 3.	
	subscript		:=		descent - ascent // 3.	
	emphasis		:=		0.
	glyphs			:=	Form extent: (raster16 * 16) @ (self height)  
							offset: 0@0.
		glyphs bits fromByteStream: strike.

	xTable := (Array new: maxAscii + 3) atAllPut: 0.
	(minAscii + 1 to: maxAscii + 3) do:
		[:index | xTable at: index put: strike nextWord].

	"Set up space character"
	((xTable at: (Space asciiValue + 2))  = 0 or:
			[(xTable at: (Space asciiValue + 2)) = (xTable at: (Space asciiValue + 1))])
		ifTrue:	[(Space asciiValue + 2) to: xTable size do:
					[:index | xTable at: index put: ((xTable at: index) + DefaultSpace)]].
	strike close.
	characterToGlyphMap := nil.! !

!StrikeFont methodsFor: 'file in/out' stamp: 'tk 9/28/2000 15:50'!
objectForDataStream: refStrm
	| dp |
	"I am about to be written on an object file.  Write a reference to a known Font in the other system instead.  "

	"A path to me"
	(TextConstants at: #forceFontWriting ifAbsent: [false]) ifTrue: [^ self].
		"special case for saving the default fonts on the disk.  See collectionFromFileNamed:"

	dp := DiskProxy global: #StrikeFont selector: #familyName:size:emphasized:
			args: (Array with: self familyName   with: self height
					with: self emphasis).
	refStrm replace: self with: dp.
	^ dp! !

!StrikeFont methodsFor: 'accessing' stamp: 'sw 1/18/2000 20:54'!
pointSize
	^ pointSize! !

!StrikeFont methodsFor: 'accessing' stamp: 'sma 5/5/2000 14:21'!
pointSize: anInteger
	pointSize := anInteger! !

!StrikeFont methodsFor: 'copying' stamp: 'BG 12/9/2004 17:35'!
postCopy
 " the receiver is a just created shallow copy. This method gives it the final touch. " 
 
    glyphs := glyphs copy.
    xTable := xTable copy.
    characterToGlyphMap := characterToGlyphMap copy.
 
    self reset.  " takes care of the derivative fonts "! !

!StrikeFont methodsFor: 'file in/out' stamp: 'sma 6/1/2000 09:32'!
printOn: aStream
	super printOn: aStream.
	aStream
		nextPut: $(;
		nextPutAll: self name;
		space;
		print: self height;
		nextPut: $)! !

!StrikeFont methodsFor: 'accessing' stamp: ''!
raster
	"Answer an integer that specifies the layout of the glyphs' form."

	^raster! !

!StrikeFont methodsFor: 'file in/out' stamp: 'ar 10/25/2005 00:21'!
readBDFFromFile: fileName name: aString 
	"This builds a StrikeFont instance by reading the X11 Binary 
	Distribution Format font source file.  See the BDFFontReader class
	comment."

	"StrikeFont new readBDFFromFile: 'helvR12' name: 'Helvetica12'."

	| fontReader stream |
	fontReader := BDFFontReader openFileNamed: fileName.
	stream := ReadStream on: fontReader read.
	xTable := stream next.
	glyphs := stream next.
	minAscii := stream next.
	maxAscii := stream next.
	maxWidth := stream next.
	ascent := stream next.
	descent := stream next.
	pointSize := stream next.
	name := aString.
"	xTable size <= 256 ifTrue: [self setStopConditions]."
	type := 0.	"no one see this"
	superscript := (ascent - descent) // 3.
	subscript := (descent - ascent) // 3.
	emphasis := 0.
	self reset! !

!StrikeFont methodsFor: 'file in/out' stamp: ''!
readBFHeaderFrom: f
	name := self restOfLine: 'Font name = ' from: f.
	ascent := (self restOfLine: 'Ascent = ' from: f) asNumber.
	descent := (self restOfLine: 'Descent = ' from: f) asNumber.
	maxWidth := (self restOfLine: 'Maximum width = ' from: f) asNumber.
	pointSize := (self restOfLine: 'Font size = ' from: f) asNumber.
	name := (name copyWithout: Character space) ,
				(pointSize < 10
					ifTrue: ['0' , pointSize printString]
					ifFalse: [pointSize printString]).
	minAscii := 258.
	maxAscii := 0.
	superscript := ascent - descent // 3.	
	subscript := descent - ascent // 3.	
	emphasis := 0.
	type := 0.  "ignored for now"
! !

!StrikeFont methodsFor: 'multibyte character methods' stamp: 'yo 8/28/2002 16:37'!
readCharacter: aBits from: aStream

	| pos |
	pos := 0.
	12 timesRepeat: [
		1 to: 2 do: [ :w |
			aBits byteAt: (pos+w) put: (aStream next ). 
		].
		pos := pos + 4.
	].
! !

!StrikeFont methodsFor: 'file in/out' stamp: 'yo 11/30/2003 17:08'!
readEFontBDFForJapaneseFromFile: fileName name: aString overrideWith: otherFileName

	| fontReader stream |
	fontReader := EFontBDFFontReaderForRanges readOnlyFileNamed: fileName.
	stream := ReadStream on: (fontReader readRanges: fontReader rangesForJapanese overrideWith: otherFileName otherRanges: {Array with: 8481 with: 12320} additionalOverrideRange: fontReader additionalRangesForJapanese).
	xTable := stream next.
	glyphs := stream next.
	minAscii := stream next.
	maxAscii := stream next.
	maxWidth := stream next.
	ascent := stream next.
	descent := stream next.
	pointSize := stream next.
	name := aString.
	type := 0. "no one see this"
	superscript := ascent - descent // 3.	
	subscript := descent - ascent // 3.	
	emphasis := 0.
	self reset.
! !

!StrikeFont methodsFor: 'file in/out' stamp: 'yo 1/15/2004 16:48'!
readEFontBDFForKoreanFromFile: fileName name: aString overrideWith: otherFileName

	| fontReader stream |
	fontReader := EFontBDFFontReaderForRanges readOnlyFileNamed: fileName.
	stream := ReadStream on: (fontReader readRanges: fontReader rangesForKorean overrideWith: otherFileName otherRanges: {Array with: 8481 with: 12320} additionalOverrideRange: fontReader additionalRangesForKorean).
	xTable := stream next.
	glyphs := stream next.
	minAscii := stream next.
	maxAscii := stream next.
	maxWidth := stream next.
	ascent := stream next.
	descent := stream next.
	pointSize := stream next.
	name := aString.
	type := 0. "no one see this"
	superscript := ascent - descent // 3.	
	subscript := descent - ascent // 3.	
	emphasis := 0.
	self reset.
! !

!StrikeFont methodsFor: 'file in/out' stamp: 'yo 12/28/2002 21:02'!
readEFontBDFFromFile: fileName name: aString rangeFrom: startRange to: endRange

	| fontReader stream |
	fontReader := EFontBDFFontReader readOnlyFileNamed: fileName.
	stream := ReadStream on: (fontReader readFrom: startRange to: endRange).
	xTable := stream next.
	glyphs := stream next.
	minAscii := stream next.
	maxAscii := stream next.
	maxWidth := stream next.
	ascent := stream next.
	descent := stream next.
	pointSize := stream next.
	name := aString.
	type := 0. "no one see this"
	superscript := ascent - descent // 3.	
	subscript := descent - ascent // 3.	
	emphasis := 0.
	self reset.
! !

!StrikeFont methodsFor: 'file in/out' stamp: 'yo 1/19/2005 11:22'!
readEFontBDFFromFile: fileName name: aString ranges: ranges

	| fontReader stream |
	fontReader := EFontBDFFontReaderForRanges readOnlyFileNamed: fileName.
	stream := ReadStream on: (fontReader readRanges: ranges).
	xTable := stream next.
	glyphs := stream next.
	minAscii := stream next.
	maxAscii := stream next.
	maxWidth := stream next.
	ascent := stream next.
	descent := stream next.
	pointSize := stream next.
	name := aString.
	type := 0. "no one see this"
	superscript := ascent - descent // 3.	
	subscript := descent - ascent // 3.	
	emphasis := 0.
	self reset.
! !

!StrikeFont methodsFor: 'file in/out' stamp: 'yo 9/23/2002 16:30'!
readF12FromStream: aStream

	| box blt |
	minAscii := 0.
	maxAscii := 94*94.
	ascent := 12.
	descent := 0.
	pointSize := 12.
	superscript := 0.
	subscript := 0.
	emphasis := 0.
	maxWidth := 12.
	
	box := Form extent: 12@12.
	glyphs  := Form extent: (94*94*12)@12.
	blt := BitBlt toForm: glyphs. 
	xTable := XTableForFixedFont new.
	xTable maxAscii: maxAscii + 3.
	xTable width: 12.
	1 to: 256 do:  [:index | 
		1 to: 12 do: [:i |
			aStream next.
		].
	].
	(minAscii + 1 to: 94*94) do:  [:index | 
		self readCharacter: (box bits) from: aStream.
		blt copy: ((12*(index-1))@0 extent: 12@12) from: 0@0 in: box.
	].
! !

!StrikeFont methodsFor: 'file in/out' stamp: 'ar 5/23/2000 12:50'!
readFromBitFont: fileName
	"This builds a StrikeFont instance by reading the data file format
	produced by BitFont, a widely available font conversion utility
	written by Peter DiCamillo at Brown University"
	"StrikeFont new readFromBitFont: 'Palatino10.BF' "
	| f lastAscii charLine width ascii charForm line missingForm tempGlyphs iRect p rectLine left tokens right |
	f := FileStream readOnlyFileNamed: fileName.
	self readBFHeaderFrom: f.

	"NOTE: if font has been scaled (and in any case),
	the REAL bitmap dimensions come after the header."
	self restOfLine: 'Extent information for entire font' from: f.
	"Parse the following line (including mispelling!!)"
	"Image rectange: left = -2, right = 8, bottom = -2, top = 7"
	tokens := (f upTo: Character cr)  findTokens: ' '.
	iRect := Rectangle left: (tokens at: 5) asNumber right: (tokens at: 8) asNumber
				top: (tokens at: 14) asNumber bottom: (tokens at: 11) asNumber.
	ascent := iRect top.
	descent := iRect bottom negated.
	
	tempGlyphs := Form extent: (maxWidth*257) @ self height.
	xTable := (Array new: 258) atAllPut: 0.
	xTable at: 1 put: 0.

	"Read character forms and blt into tempGlyphs"
	lastAscii := -1.
	[charLine := self restOfLine: 'Character: ' from: f.
	charLine == nil ifFalse:
		[p := f position.
		rectLine := f upTo: Character cr.
		(rectLine beginsWith: 'Image rectange: left = ')
			ifTrue: [tokens := rectLine findTokens: ' '.
					left := (tokens at: 5) asNumber. right := (tokens at: 8) asNumber]
			ifFalse: [left := right := 0. f position: p].
		width:= (self restOfLine: 'Width (final pen position) = ' from: f) asNumber - left
					max: (right-left+1).
		(charLine beginsWith: 'Missing character') ifTrue: [ascii := 256].
		('x''*' match: charLine) ifTrue:
			[ascii := Number readFrom: (charLine copyFrom: 3 to: 4) asUppercase base: 16].
		charForm := Form extent: width@self height.
		('*[all blank]' match: charLine) ifFalse:
			[self restOfLine: '  +' from: f.
			1 to: self height do:
				[:y | line := f upTo: Character cr.
				4 to: (width + 3 min: line size + iRect left - left) do:
					[:x | (line at: x - iRect left + left) = $*
						ifTrue: [charForm pixelValueAt: (x-4)@(y-1) put: 1]]]]].
	charLine == nil]
		whileFalse:
			[self displayChar: ascii form: charForm.
			ascii = 256
				ifTrue: [missingForm := charForm deepCopy]
				ifFalse:
				[minAscii := minAscii min: ascii.
				maxAscii := maxAscii max: ascii.
				lastAscii+1 to: ascii-1 do: [:a | xTable at: a+2 put: (xTable at: a+1)].
				tempGlyphs copy: ((xTable at: ascii+1)@0
										extent: charForm extent)
							from: 0@0 in: charForm rule: Form over.
				xTable at: ascii+2 put: (xTable at: ascii+1) + width.
				lastAscii := ascii]].
	f close.
	lastAscii+1 to: maxAscii+1 do: [:a | xTable at: a+2 put: (xTable at: a+1)].
	missingForm == nil ifFalse:
		[tempGlyphs copy: missingForm boundingBox from: missingForm
				to: (xTable at: maxAscii+2)@0 rule: Form over.
		xTable at: maxAscii+3 put: (xTable at: maxAscii+2) + missingForm width].
	glyphs := Form extent: (xTable at: maxAscii+3) @ self height.
	glyphs copy: glyphs boundingBox from: 0@0 in: tempGlyphs rule: Form over.
	xTable := xTable copyFrom: 1 to: maxAscii+3.
	characterToGlyphMap := nil.! !

!StrikeFont methodsFor: 'file in/out' stamp: 'sma 12/30/1999 14:20'!
readFromStrike2: fileName  "StrikeFont new readFromStrike2: 'Palatino14.sf2'"
	"Build an instance from the strike font stored in strike2 format.
	fileName is of the form: <family name><pointSize>.sf2"
	| file |
	('*.sf2' match: fileName) ifFalse: [self halt.  "likely incompatible"].
	name := fileName copyUpTo: $. .  "Drop filename extension"
	file := FileStream readOnlyFileNamed: fileName.
	file binary.
	[self readFromStrike2Stream: file] ensure: [file close]! !

!StrikeFont methodsFor: 'file in/out' stamp: 'ar 5/23/2000 12:53'!
readFromStrike2Stream: file 
	"Build an instance from the supplied binary stream on data in strike2 format"
	type := file nextInt32.  type = 2 ifFalse: [file close. self error: 'not strike2 format'].
	minAscii := file nextInt32.
	maxAscii := file nextInt32.
	maxWidth := file nextInt32.
	ascent := file nextInt32.
	descent := file nextInt32.
	pointSize := file nextInt32.
	superscript := ascent - descent // 3.	
	subscript := descent - ascent // 3.	
	emphasis := file nextInt32.
	xTable := (Array new: maxAscii + 3) atAllPut: 0.
	(minAscii + 1 to: maxAscii + 3) do:
		[:index | xTable at: index put: file nextInt32].
	glyphs := Form new readFrom: file.

	"Set up space character"
	((xTable at: (Space asciiValue + 2))  = 0 or:
			[(xTable at: (Space asciiValue + 2)) = (xTable at: (Space asciiValue + 1))])
		ifTrue:	[(Space asciiValue + 2) to: xTable size do:
					[:index | xTable at: index put: ((xTable at: index) + DefaultSpace)]].
	characterToGlyphMap := nil.! !

!StrikeFont methodsFor: 'Mac reader' stamp: 'ar 5/23/2000 12:49'!
readMacFontHex: fileName
	"Read the hex version of a Mac FONT type resource.  See the method aComment for how to prepare the input file. 4/26/96 tk"
	| file hh fRectWidth |
	name := fileName.	"Palatino 12"
	file := FileStream readOnlyFileNamed: fileName, ' hex'.

	"See Inside Macintosh page IV-42 for this record"
	"FontType := " Number readFrom: (file next: 4) base: 16.
	emphasis		:=		0.
	minAscii := Number readFrom: (file next: 4) base: 16.
	maxAscii := Number readFrom: (file next: 4) base: 16.
	maxWidth		:= Number readFrom: (file next: 4) base: 16.
	"kernMax := " Number readFrom: (file next: 4) base: 16.
	"NDescent := " Number readFrom: (file next: 4) base: 16.
	fRectWidth :=  Number readFrom: (file next: 4) base: 16.
	hh :=  Number readFrom: (file next: 4) base: 16.
	"OWTLoc := " Number readFrom: (file next: 4) base: 16.
	ascent			:= Number readFrom: (file next: 4) base: 16.
	descent			:= Number readFrom: (file next: 4) base: 16.
	"leading := " Number readFrom: (file next: 4) base: 16.
	xOffset			:=		0. 	
	raster			:= Number readFrom: (file next: 4) base: 16.

	strikeLength	:=		raster*16.
	superscript		:=		ascent - descent // 3.	
	subscript		:=		descent - ascent // 3.	
	self strikeFromHex: file width: raster height: hh.
	self xTableFromHex: file.
	file close.

	"Insert one pixel between each character.  And add space character."
	self fixKerning: (fRectWidth - maxWidth).	

	"Recompute character to glyph mapping"
	characterToGlyphMap := nil.! !

!StrikeFont methodsFor: 'emphasis' stamp: 'nk 3/15/2004 18:46'!
releaseCachedState

	self reset.! !

!StrikeFont methodsFor: 'emphasis' stamp: 'jmv 8/3/2009 09:22'!
reset
	"Reset the cache of derivative emphasized fonts"

	fallbackFont class = FixedFaceFont
		ifTrue: [fallbackFont _ nil].
		
	derivativeFonts notNil ifTrue: [
		derivativeFonts withIndexDo: [ :f :i |
			(f notNil and: [f isSynthetic]) ifTrue: [derivativeFonts at: i put: nil]]].
	"
	derivativeFonts _ Array new: 32.
	#('B' 'I' 'BI') doWithIndex:
		[:tag :index | 
		(style _ TextStyle named: self familyName) ifNotNil:
			[(font _ style fontArray
				detect: [:each | each name = (self name , tag)]
				ifNone: [nil]) ifNotNil: [derivativeFonts at: index put: font]]]
	"! !

!StrikeFont methodsFor: 'file in/out' stamp: ''!
restOfLine: leadString from: file
	"Utility method to assist reading of BitFont data files"
	| line |
	[line := file upTo: Character cr.
	line size < leadString size or: [leadString ~= (line copyFrom: 1 to: leadString size)]]
	whileTrue: [file atEnd ifTrue: [^ nil]].
	^ line copyFrom: leadString size+1 to: line size! !

!StrikeFont methodsFor: 'accessing' stamp: ''!
setGlyphs: newGlyphs
	"Replace the glyphs form.  Used to make a synthetic bold or italic font quickly."

	glyphs := newGlyphs! !

!StrikeFont methodsFor: 'building' stamp: 'jmv 10/18/2008 23:50'!
setGlyphsDepthAtMost: aNumber
	glyphs depth > aNumber ifTrue: [
		glyphs _ glyphs asFormOfDepth: aNumber ]! !

!StrikeFont methodsFor: 'emphasis' stamp: 'rej 9/29/2007 09:02'!
setupDefaultFallbackFont
	"This used to be the default textstyle, but it needs to be a StrikeFont and not a TTCFont and sometimes the default textstyle is a TTCFont.  So, we use a typical StrikeFont as the default fallback font."
	self fallbackFont: (StrikeFont defaultSized: self height).
	self reset.

! !

!StrikeFont methodsFor: 'Mac reader' stamp: ''!
strikeFromHex: file width: w height: h
	"read in just the raw strike bits from a hex file.  No spaces or returns.  W is in words (2 bytes), h in pixels." 
	| newForm theBits offsetX offsetY str num cnt |
	offsetX  := 0.
	offsetY := 0.
	offsetX > 32767 ifTrue: [offsetX := offsetX - 65536]. "stored two's-complement"
	offsetY > 32767 ifTrue: [offsetY := offsetY - 65536]. "stored two's-complement"
	newForm := Form extent: strikeLength @ h offset: offsetX @ offsetY.
	theBits := newForm bits.
	cnt := 0.		"raster may be 16 bits, but theBits width is 32" 
	1 to: theBits size do: [:i | 
		(cnt := cnt + 32) > strikeLength 
		  ifTrue: [cnt := 0.
			num := Number readFrom: (str := file next: 4) base: 16]
		  ifFalse: [
			cnt = strikeLength ifTrue: [cnt := 0].
			num := Number readFrom: (str := file next: 8) base: 16].
		theBits at: i put: num].
	glyphs := newForm.! !

!StrikeFont methodsFor: 'building' stamp: 'jmv 6/19/2008 16:55'!
stripHighGlyphs
	"Remove glyphs for characters above 128"
	| i |
	maxAscii _ 127.
	
	xTable _ xTable copyFrom: 1 to: maxAscii + 3.
	i _ xTable at: maxAscii + 1.
	xTable at: maxAscii + 2 put: i.
	xTable at: maxAscii + 3 put: i.
	glyphs _ glyphs copy: (0@0 extent: i@glyphs height).
	maxWidth _ 0.
	2 to: xTable size do: [ :ii |
		maxWidth _ maxWidth max: (xTable at: ii) - (xTable at: ii-1)-1 ].
	characterToGlyphMap _ nil.
	self reset! !

!StrikeFont methodsFor: 'accessing' stamp: ''!
subscript
	"Answer an integer that is the further vertical offset relative to the 
	baseline for positioning characters as subscripts."

	^subscript! !

!StrikeFont methodsFor: 'accessing' stamp: ''!
superscript
	"Answer an integer that is the further vertical offset relative to the 
	baseline for positioning characters as superscripts."

	^superscript! !

!StrikeFont methodsFor: 'accessing' stamp: 'nk 6/17/2003 14:26'!
textStyle
	^ TextStyle actualTextStyles detect:
		[:aStyle | aStyle fontArray includes: self] ifNone: [nil]! !

!StrikeFont methodsFor: 'character shapes' stamp: 'jmv 3/26/2009 15:17'!
useLeftArrow
	self characterToGlyphMap.
	characterToGlyphMap at: 96 put: 95.
	characterToGlyphMap at: 95 put: 94! !

!StrikeFont methodsFor: 'character shapes' stamp: 'jmv 3/26/2009 14:49'!
useUnderscore
	self characterToGlyphMap.
	characterToGlyphMap at: 96 put: 129.
	characterToGlyphMap at: 95 put: 128! !

!StrikeFont methodsFor: 'character shapes' stamp: 'jmv 8/6/2009 14:23'!
useUnderscoreIfOver1bpp

	glyphs depth = 1 ifTrue: [
		characterToGlyphMap ifNotNil: [	
			characterToGlyphMap at: 96 put: 95.
			characterToGlyphMap at: 95 put: 94 ].
		^self ].
	
	self characterToGlyphMap.
	characterToGlyphMap at: 96 put: 129.
	characterToGlyphMap at: 95 put: 128! !

!StrikeFont methodsFor: 'copying' stamp: 'tk 8/19/1998 16:15'!
veryDeepCopyWith: deepCopier
	"Return self.  I am shared.  Do not record me."! !

!StrikeFont methodsFor: 'character shapes' stamp: ''!
widen: char by: delta
	| newForm |
	^ self alter: char formBlock:  "Make a new form, wider or narrower..."
		[:charForm | newForm := Form extent: charForm extent + (delta@0).
		charForm displayOn: newForm.  "Copy this image into it"
		newForm]    "and substitute it in the font"! !

!StrikeFont methodsFor: 'accessing' stamp: 'jmv 8/3/2009 10:01'!
widthOf: aCharacter 
	"Answer the width of the argument as a character in the receiver."
	| code |
	code := aCharacter charCode.
	self characterToGlyphMap size > code ifTrue: [
		code := characterToGlyphMap at: code + 1 ].
	((code < minAscii or: [maxAscii < code]) 
		or: [(xTable at: code + 1) < 0])
			ifTrue: [^ self fallbackFont widthOf: aCharacter].
	^ (xTable at: code + 2) - (xTable at: code + 1)! !

!StrikeFont methodsFor: 'displaying' stamp: 'tak 1/11/2005 18:03'!
widthOfString: aString from: firstIndex to: lastIndex
	| resultX |
	resultX := 0.
	firstIndex to: lastIndex do:[:i | 
		resultX := resultX + (self widthOf: (aString at: i))].
	^ resultX.
! !

!StrikeFont methodsFor: 'file in/out' stamp: 'ls 3/27/2000 17:43'!
writeAsStrike2On: file
	"Write me onto a file in strike2 format.
	fileName should be of the form: <family name><pointSize>.sf2"
	file binary.
	file nextInt32Put: 2.
	file nextInt32Put: minAscii.
	file nextInt32Put: maxAscii.
	file nextInt32Put: maxWidth.
	file nextInt32Put: ascent.
	file nextInt32Put: descent.
	file nextInt32Put: pointSize.
	superscript := ascent - descent // 3.	
	subscript := descent - ascent // 3.	
	file nextInt32Put: emphasis.
	(minAscii + 1 to: maxAscii + 3) do:
		[:index | file nextInt32Put: (xTable at: index)].
	glyphs writeOn: file.
	file close.
! !

!StrikeFont methodsFor: 'file in/out' stamp: 'ls 4/11/2000 18:57'!
writeAsStrike2named: fileName
	"Write me onto a file in strike2 format.
	fileName should be of the form: <family name><pointSize>.sf2"
	| file |
	file := FileStream fileNamed: fileName.
	self writeAsStrike2On: file.
	file close.! !

!StrikeFont methodsFor: 'accessing' stamp: ''!
xTable
	"Answer an Array of the left x-coordinate of characters in glyphs."

	^xTable! !

!StrikeFont methodsFor: 'accessing' stamp: 'yo 8/28/2002 16:33'!
xTable: anObject

	xTable := anObject.
! !

!StrikeFont methodsFor: 'Mac reader' stamp: ''!
xTableFromHex: file

	| strike num str wid |
	strike := file.
	xTable := (Array new: maxAscii + 3) atAllPut: 0.
	(minAscii + 1 to: maxAscii + 3) do:
		[:index | 
			num := Number readFrom: (str := strike next: 4) base: 16. 
			xTable at: index put: num].

	1 to: xTable size - 1 do: [:ind |
		wid := (xTable at: ind+1) - (xTable at: ind).
		(wid < 0) | (wid > 40) ifTrue: [
			file close.
			self error: 'illegal character width']].
! !

Object subclass: #TextStyle
	instanceVariableNames: 'fontArray fontFamilySize lineGrid baseline alignment firstIndent restIndent rightIndent tabsArray marginTabsArray leading defaultFontIndex'
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

!TextStyle commentStamp: '<historical>' prior: 0!
A textStyle comprises the formatting information for composing and displaying a unit (usually a paragraph) of text.  Typically one makes a copy of a master textStyle (such as TextStyle default), and then that copy may get altered in the process of editing.  Bad things can happen if you do not copy first.

Each of my instances consists of...
	fontArray		An array of StrikeFonts
	fontFamilySize	unused
	lineGrid			An integer; default line spacing for paragraphs
	baseline			An integer; default baseline (dist from line top to bottom of an 'a')
	alignment		An integer; text alignment, see TextStyle alignment:
	firstIndent		An integer; indent of first line in pixels
	restIndent		An integer; indent of remaining lines in pixels
	rightIndent		An integer; indent of right margin rel to section
	tabsArray		An array of integers giving tab offsets in pixels
	marginTabsArray	An array of margin tabs
	leading			An integer giving default vertical line separation

For a concrete example, look at TextStyle default copy inspect!

!TextStyle class methodsFor: 'TextConstants access' stamp: 'nk 9/1/2004 11:00'!
actualTextStyles
	| aDict |
	"TextStyle actualTextStyles"

	"Answer dictionary whose keys are the names of styles in the system and whose values are the actual styles"

	aDict := TextConstants select: [:thang | thang isKindOf: self ].
	self defaultFamilyNames do: [ :sym | aDict removeKey: sym ].
	^ aDict! !

!TextStyle class methodsFor: 'instance creation' stamp: 'ar 8/6/2009 18:36'!
changeDefaultFontSizeBy: delta      "TextStyle changeDefaultFontSizeBy: 1"
	"This sample method recreates the default textStyle, with font 1 being a size
	larger than the smallest.  It then initializes most references in the system
	as well, although most windows will have to beclosed and reopened to get the effect."
	| allFonts |
	allFonts := TextStyle default fontArray asSortedCollection: [:a :b | a height < b height].
	TextConstants at: #DefaultTextStyle put:
		(TextStyle fontArray: ((1 to: allFonts size) collect: [:i | allFonts atWrap: i+delta])).
	#(ListParagraph PopUpMenu StandardSystemView) do:[:className|
		Smalltalk at: className ifPresent:[:aClass| aClass initialize].
	].! !

!TextStyle class methodsFor: 'user interface' stamp: 'rbb 3/1/2005 11:18'!
chooseTTCFontSize: args
	"Prompt for a point size and, if one is given, add a new font size to the font named by the first member of args. If args' length is three, send a message with the selector equal to the third of args, and the receiver equal to the second of args, passing the selected style as an argument."

	| f n style |
	f := UIManager default request: 'New Point Size' initialAnswer: '0'.
	n := f asNumber.
	style := (TextConstants at: args first) addNewFontSize: n.
	style ifNotNil: [
		args second ifNotNil: [args second perform: args third with: style].
	].
! !

!TextStyle class methodsFor: 'mime file in/out' stamp: 'nk 1/19/2004 20:36'!
collectionFromCompressedMIMEString: aString
	"aString holds a compressed, Base64 representation of a SmartRefStream storage of a TextStyle.
	Install the TextStyle."

	| this newName style heights data |
	data := (Base64MimeConverter mimeDecode: aString as: String) unzipped.
	(RWBinaryOrTextStream with: data) reset; fileIn.
	this := SmartRefStream scannedObject.

	"now install it"

	newName := this fontArray first familyName.
	this fontArray do: [:aFont | aFont familyName = newName ifFalse: [
		self error: 'All must be same family']].
	style := TextConstants at: newName asSymbol ifAbsent: [
		^ TextConstants at: newName asSymbol put: this].		"new family"
	this fontArray do: [:aFont | "add new fonts"
		heights := style fontArray collect: [:bFont | bFont height].
		(heights includes: aFont height) ifFalse: [
			style fontAt: style fontArray size + 1 put: aFont]].
! !

!TextStyle class methodsFor: 'utilities' stamp: 'nk 3/25/2004 17:55'!
decodeStyleName: styleName 
	"Given a string styleName, return a collection with: 
	 
	* [1] the probable Squeak emphasis code, which is a bit combination of: 
	1	bold 
	2	italic 
	4	underlined 
	8	narrow 
	16	strikeout 
	 
	* [2] the base style name without the modifiers (can be empty)
	* [3] the modifiers in the order they were found 
	* [4] the codes for those modifiers, in the same order
	"
	| decoder keys modifiers modifierCodes baseName styleCode matchedKey |

	decoder := self styleDecoder.

	modifiers := OrderedCollection new.
	modifierCodes := OrderedCollection new.
	keys := decoder keys asArray
				sort: [:a :b | a size > b size].
	styleCode := 0.
	baseName := styleName asString.
	[matchedKey := keys
				detect: [:k | baseName endsWith: k]
				ifNone: [].
	matchedKey notNil]
		whileTrue: [| last code | 
			last := baseName size - matchedKey size.
			last > 0
				ifTrue: [('- ' includes: (baseName at: last))
						ifTrue: [last := last - 1]].
			baseName := baseName copyFrom: 1 to: last.
			code := decoder at: matchedKey.
			styleCode := styleCode + code.
			modifiers addFirst: matchedKey.
			modifierCodes addFirst: code.
	].
	^ {styleCode. baseName. modifiers. modifierCodes }! !

!TextStyle class methodsFor: 'constants' stamp: ''!
default
	"Answer the system default text style."

	^DefaultTextStyle! !

!TextStyle class methodsFor: 'TextConstants access' stamp: 'nk 9/1/2004 10:59'!
defaultFamilyNames
	^#(DefaultTextStyle DefaultFixedTextStyle DefaultMultiStyle)! !

!TextStyle class methodsFor: 'constants' stamp: 'sw 12/6/1999 12:32'!
defaultFont
	"Answer the default system font"

	^ DefaultTextStyle defaultFont! !

!TextStyle class methodsFor: 'user interface' stamp: 'nk 9/1/2004 13:18'!
emphasisMenuForFont: font target: target selector: selector highlight: currentEmphasis
	"Offer a font emphasis menu for the given style. If one is selected, pass that font to target with a call to selector. The fonts will be displayed in that font.
	Answer nil if no derivatives exist.
	"

 	| aMenu derivs |
	derivs := font derivativeFonts.
	derivs isEmpty ifTrue: [ ^nil ].
	aMenu := MenuMorph entitled: 'emphasis' translated.
	derivs := derivs asOrderedCollection.
	derivs addFirst: font.
	derivs do: [ :df | 
			aMenu 
				add: (AbstractFont emphasisStringFor: df emphasis)
				target: target 
				selector: selector
				argument: df.
                aMenu lastItem font: df.
                df emphasis == currentEmphasis ifTrue: [aMenu lastItem color: Color blue darker]].
        ^ aMenu! !

!TextStyle class methodsFor: 'instance creation' stamp: ''!
fontArray: anArray 
	"Answer an instance of me with fonts those in the argument, anArray."

	^self new newFontArray: anArray! !

!TextStyle class methodsFor: 'TextConstants access' stamp: 'nk 7/3/2003 19:06'!
fontArrayForStyle: aName
	"Answer the fonts in the style named aName,
	or an empty Array if no such named style."

	"TextStyle fontArrayForStyle: #Atlanta"
	"TextStyle fontPointSizesFor: 'NewYork'"

	^ ((self named: aName) ifNil: [ ^#() ]) fontArray
! !

!TextStyle class methodsFor: 'user interface' stamp: 'laza 3/25/2004 23:12'!
fontMenuForStyle: styleName target: target selector: selector
	^self fontMenuForStyle: styleName target: target selector: selector highlight: nil! !

!TextStyle class methodsFor: 'user interface' stamp: 'nk 9/1/2004 13:31'!
fontMenuForStyle: styleName target: target selector: selector highlight: currentFont 
	"Offer a font menu for the given style. If one is selected, pass 
	that font to target with a  
	call to selector. The fonts will be displayed in that font."
	| aMenu |
	aMenu := MenuMorph entitled: styleName.
	(TextStyle named: styleName)
		ifNotNilDo: [:s | s isTTCStyle
				ifTrue: [aMenu
						add: 'New Size'
						target: self
						selector: #chooseTTCFontSize:
						argument: {styleName. target. selector}]].
	(self pointSizesFor: styleName)
		do: [:pointSize | 
			| font subMenu | 
			font := (self named: styleName)
						fontOfPointSize: pointSize.
			subMenu := self
						emphasisMenuForFont: font
						target: target
						selector: selector
						highlight: (currentFont
								ifNotNilDo: [:cf | (cf familyName = styleName
											and: [cf pointSize = font pointSize])
										ifTrue: [currentFont emphasis]]).
			subMenu
				ifNil: [aMenu
						add: pointSize asString , ' Point'
						target: target
						selector: selector
						argument: font]
				ifNotNil: [aMenu add: pointSize asString , ' Point' subMenu: subMenu].
			aMenu lastItem font: font.
			currentFont
				ifNotNilDo: [:cf | (cf familyName = styleName
							and: [cf pointSize = font pointSize])
						ifTrue: [aMenu lastItem color: Color blue darker]]].
	^ aMenu! !

!TextStyle class methodsFor: 'TextConstants access' stamp: 'nk 7/3/2003 18:56'!
fontPointSizesFor: aName
	"Answer the point sizes for all the fonts in the given text style"

	"TextStyle fontPointSizesFor: 'Arial'"
	"TextStyle fontPointSizesFor: 'NewYork'"

	^ (self fontArrayForStyle: aName) collect: [:f | f pointSize]
! !

!TextStyle class methodsFor: 'user interface' stamp: 'dgd 10/8/2003 18:38'!
fontSizeSummary
	"Open a text window with a simple summary of the available sizes in each of the fonts in the system."

	"TextStyle fontSizeSummary"
	| aString aList |
	aList := self knownTextStyles.
	aString := String streamContents:
		[:aStream |
			aList do: [:aStyleName |
				aStream nextPutAll:
					aStyleName, '  ',
					(self fontPointSizesFor: aStyleName) asArray storeString.
				aStream cr]].
	(StringHolder new contents: aString)
		openLabel: 'Font styles and sizes' translated! !

!TextStyle class methodsFor: 'TextConstants access' stamp: 'nk 7/3/2003 18:58'!
fontSizesFor: aName
	"Answer the pixel sizes for all the fonts in the given text style"

	"TextStyle fontSizesFor: 'Arial'"
	"TextStyle fontSizesFor: 'NewYork'"

	^ (self fontArrayForStyle: aName) collect: [:f | f height ]
! !

!TextStyle class methodsFor: 'TextConstants access' stamp: 'nk 7/3/2003 18:58'!
fontWidthsFor: aName
	"Answer the widths for all the fonts in the given text style"

	"TextStyle fontWidthsFor: 'ComicPlain'"
	^ (self fontArrayForStyle: aName) collect: [:f | f maxWidth]
! !

!TextStyle class methodsFor: 'user interface' stamp: 'sw 8/12/2004 19:08'!
importFontsFromStyleFiles
	"Import any and all of the fonts found in the default directory in files named ComicBold.style, ComicPlain.style, NewYork.style, Palatino.style, Courier.style"

	| aName |
	#('ComicBold' 'ComicPlain' 'NewYork' 'Palatino' 'Courier') do:
		[:frag |
			(TextStyle knownTextStyles includes: frag) ifFalse:
				[(FileDirectory default fileExists: (aName := frag, '.style'))
						ifTrue:
							[TextStyle default collectionFromFileNamed: aName]]].! !

!TextStyle class methodsFor: 'instance creation' stamp: ''!
initDefaultFontsAndStyle
	"This provides the system with 10 and 12-pt basal fonts.
	Bold and italic versions will be automatically generated as needed"
	| fontArray |	
	fontArray := Array new: 2.
	fontArray at: 1 put: (StrikeFont new readFromStrike2: 'NewYork10.sf2').
	fontArray at: 2 put: (StrikeFont new readFromStrike2: 'NewYork12.sf2').
	TextConstants at: #DefaultTextStyle put:
		(TextStyle fontArray: fontArray).

	"TextStyle initDefaultFontsAndStyle."! !

!TextStyle class methodsFor: 'class initialization' stamp: 'nk 3/25/2004 17:51'!
initialize
	self initializeStyleDecoder.! !

!TextStyle class methodsFor: 'class initialization' stamp: 'nk 3/25/2004 17:53'!
initializeStyleDecoder
	TextConstants at: #StyleDecoder put: nil.
	self styleDecoder.! !

!TextStyle class methodsFor: 'TextConstants access' stamp: 'nk 7/3/2003 19:00'!
knownTextStyles
	"Answer the names of the known text styles, sorted in alphabetical order"

	"TextStyle knownTextStyles"
	^ (TextConstants select: [:thang | thang isKindOf: TextStyle]) keys asSortedArray

! !

!TextStyle class methodsFor: 'TextConstants access' stamp: 'nk 9/1/2004 11:08'!
knownTextStylesWithoutDefault
	"Answer the names of the known text styles, sorted in alphabetical order without default"

	"TextStyle knownTextStylesWithoutDefault"
	| result |
	result := self knownTextStyles asOrderedCollection.
	^ result copyWithoutAll: self defaultFamilyNames

! !

!TextStyle class methodsFor: 'mime file in/out' stamp: 'nk 3/15/2004 19:37'!
looseFontsFromFamily: familyName
	"
	TextStyle looseFontsFromFamily: 'Accuny'
	TextStyle looseFontsFromFamily: 'Accujen'
	TextStyle actualTextStyles keys collect: [ :k | TextStyle looseFontsFromFamily: k ]
	"

	| looseFonts realStyle classes |
	realStyle := TextStyle named: familyName.
	classes := ((realStyle fontArray copyWithout: nil) collect: [ :f | f class ]) asSet.
	classes do: [ :cls | cls allSubInstancesDo: [ :f | f releaseCachedState ]].
	Smalltalk garbageCollect.
	looseFonts := IdentitySet new.
	classes do: [ :cls |
		looseFonts addAll: ((cls allSubInstances select: [ :ea | ea familyName = familyName ])
			reject: [ :f | realStyle fontArray anySatisfy: [ :fn | fn == f or: [ fn derivativeFonts includes: f ] ]]) ].
	^looseFonts! !

!TextStyle class methodsFor: 'user interface' stamp: 'nk 9/1/2004 10:38'!
modalMVCStyleSelectorWithTitle: title
	"MVC Only!! Presents a modal font-style choice menu, answers a TextStyle or nil."
	"TextStyle modalMVCStyleSelectorWithTitle: 'testing'"
	
	| aMenu actualStyles |
	aMenu := CustomMenu new.
	actualStyles := self actualTextStyles.
	actualStyles keysSortedSafely do: [ :styleName | | style |
		style := actualStyles at: styleName.
		aMenu add: styleName action: style
	].
	^aMenu startUpWithCaption: title.
! !

!TextStyle class methodsFor: 'user interface' stamp: 'nk 9/1/2004 10:34'!
modalStyleSelectorWithTitle: title
	"Presents a modal font-style choice menu, answers a TextStyle or nil."
	"TextStyle modalStyleSelectorWithTitle: 'testing'"
	
	| menu actualStyles |
	Smalltalk isMorphic ifFalse: [ ^self modalMVCStyleSelectorWithTitle: title ].

	menu := MenuMorph entitled: title.
	actualStyles := self actualTextStyles.
	actualStyles keysSortedSafely do: [ :styleName | | style |
		style := actualStyles at: styleName.
		menu add: styleName target: menu selector: #modalSelection: argument: style.
		menu lastItem font: (style fontOfSize: 18)
	].
	^menu invokeModal.
! !

!TextStyle class methodsFor: 'user interface' stamp: 'nk 9/1/2004 10:37'!
mvcPromptForFont: aPrompt andSendTo: aTarget withSelector: aSelector
	"MVC Only!! prompt for a font and if one is provided, send it to aTarget using a message with selector aSelector."
	| aMenu aChoice aStyle namesAndSizes aFont |
	"TextStyle mvcPromptForFont: 'Choose system font style' andSendTo: TextStyle withSelector: #setSystemFontTo:"
	aMenu := CustomMenu new.
	self actualTextStyles keysSortedSafely do:
		[:styleName |
			aMenu add: styleName action: styleName].
	aChoice := aMenu startUpWithCaption: aPrompt.
	aChoice ifNil: [^ self].
	aMenu := CustomMenu new.
	aStyle := self named: aChoice.
	(namesAndSizes := aStyle fontNamesWithPointSizes) do:
		[:aString | aMenu add: aString action: aString].
	aChoice := aMenu startUpWithCaption: nil.
	aChoice ifNil: [^ self].
	aFont := aStyle fontAt: (namesAndSizes indexOf: aChoice).
	aTarget perform: aSelector with: aFont! !

!TextStyle class methodsFor: 'constants' stamp: 'nk 7/3/2003 19:11'!
named: familyName
	"Answer the TextStyle with the given name, or nil."
	"TextStyle named: 'NewYork'"
	| textStyle |
	textStyle := TextConstants at: familyName ifAbsent: [ ^nil ].
	(textStyle isKindOf: self) ifFalse: [ ^nil ].
	^textStyle! !

!TextStyle class methodsFor: 'instance creation' stamp: ''!
new
	^ super new leading: 2! !

!TextStyle class methodsFor: 'utilities' stamp: 'nk 4/2/2004 11:26'!
pixelsPerInch
	"Answer the nominal resolution of the screen."

	^TextConstants at: #pixelsPerInch ifAbsentPut: [ 96.0 ].! !

!TextStyle class methodsFor: 'utilities' stamp: 'nk 4/2/2004 11:24'!
pixelsPerInch: aNumber
	"Set the nominal number of pixels per inch to aNumber."
	TextConstants at: #pixelsPerInch put: aNumber asFloat.
	AbstractFont allSubInstancesDo: [ :font | font pixelsPerInchChanged ].! !

!TextStyle class methodsFor: 'utilities' stamp: 'nk 4/2/2004 11:23'!
pixelsToPoints: pixels
	^pixels * 72.0 / self pixelsPerInch! !

!TextStyle class methodsFor: 'TextConstants access' stamp: 'nk 7/3/2003 19:11'!
pointSizesFor: aName
	"Answer all the point sizes for the given text style name"

	"TextStyle pointSizesFor: 'NewYork'"
	^ (self fontArrayForStyle: aName) collect: [:f | f pointSize]
! !

!TextStyle class methodsFor: 'utilities' stamp: 'nk 4/2/2004 11:22'!
pointsToPixels: points
	^points * self pixelsPerInch / 72.0! !

!TextStyle class methodsFor: 'user interface' stamp: 'laza 3/25/2004 23:12'!
promptForFont: aPrompt andSendTo: aTarget withSelector: aSelector
	self promptForFont: aPrompt andSendTo: aTarget withSelector: aSelector highlight: nil! !

!TextStyle class methodsFor: 'user interface' stamp: 'nk 9/1/2004 13:19'!
promptForFont: aPrompt andSendTo: aTarget withSelector: aSelector highlight: currentFont 
	"Morphic Only!! prompt for a font and if one is provided, send it to aTarget using a 
	message with selector aSelector."
	"TextStyle promptForFont: 'Choose system font:' andSendTo: Preferences withSelector: 
	#setSystemFontTo: "
	"Derived from a method written by Robin Gibson"
	| menu subMenu currentTextStyle |
	currentTextStyle := currentFont
				ifNotNil: [currentFont textStyleName].
	menu := MenuMorph entitled: aPrompt.
	self actualTextStyles keysSortedSafely
		do: [:styleName | 
			subMenu := self
						fontMenuForStyle: styleName
						target: aTarget
						selector: aSelector
						highlight: currentFont.
			menu add: styleName subMenu: subMenu.
			menu lastItem
				font: ((self named: styleName)
						fontOfSize: 18).
			styleName = currentTextStyle
				ifTrue: [menu lastItem color: Color blue darker]].
	menu popUpInWorld: self currentWorld! !

!TextStyle class methodsFor: 'mime file in/out' stamp: 'nk 3/15/2004 19:25'!
replaceFontsIn: oldFontArray with: newStyle
	"
	TextStyle replaceFontsIn: (TextStyle looseFontsFromFamily: #Accuny) with: (TextStyle named: #Accuny)
	"
	"Try to find corresponding fonts in newStyle and substitute them for the fonts in oldFontArray"

	newStyle fontArray do: [ :newFont | newFont releaseCachedState ].

	oldFontArray do: [ :oldFont | | newFont |
		oldFont reset.
		newFont := (newStyle fontOfPointSize: oldFont pointSize) emphasis: oldFont emphasis.
		oldFont becomeForward: newFont ].

	StringMorph allSubInstancesDo: [ :s | s layoutChanged ].
	TextMorph allSubInstancesDo: [ :s | s layoutChanged ].
	SystemWindow allInstancesDo: [ :w | [ w update: #relabel ] on: Error do: [ :ex | ] ].
	World ifNotNilDo: [ :w | w changed ].! !

!TextStyle class methodsFor: 'mime file in/out' stamp: 'nk 9/1/2004 11:03'!
replaceStyle: oldStyle with: newStyle
	"
	TextStyle replaceStyle: (TextStyle named: #AccunyOLD) with: (TextStyle named: #Accuny)
	"
	"Try to find corresponding fonts in newStyle and substitute the fonts in oldStyle for them."
	| oldKeys |
	oldKeys := Set new.
	TextConstants keysAndValuesDo: [ :k :v | v = oldStyle ifTrue: [ oldKeys add: k ]].
	oldKeys removeAllFoundIn: self defaultFamilyNames.

	self replaceFontsIn: oldStyle fontArray with: newStyle.

	oldStyle becomeForward: newStyle.
	oldKeys do: [ :k | TextConstants removeKey: k ].
! !

!TextStyle class methodsFor: 'constants' stamp: 'ar 1/27/2002 20:36'!
setDefault: aTextStyle
	"Answer the system default text style."

	DefaultTextStyle := aTextStyle.! !

!TextStyle class methodsFor: 'class initialization' stamp: 'bf 7/31/2009 18:21'!
styleDecoder
	TextConstants at: #StyleDecoder ifPresent: [ :dict | dict ifNotNil: [ ^dict ]].
	^TextConstants at: #StyleDecoder put: (
		Dictionary new at: 'Regular' put: 0;
				 at: 'Roman' put: 0;
				 at: 'Medium' put: 0;
				 at: 'Light' put: 0;
				 at: 'Normal' put: 0;
				 at: 'Plain' put: 0;
				 at: 'Book' put: 0;
				 at: 'Demi' put: 0;
				 at: 'Demibold' put: 0;
				 at: 'Semibold' put: 0;
				 at: 'SemiBold' put: 0;
				 at: 'ExtraBold' put: 1;
				 at: 'SuperBold' put: 1;
				 at: 'Negreta' put: 1;
				 at: 'B' put: 1;
				 at: 'I' put: 2;
				 at: 'U' put: 4;
				 at: 'X' put: 16;
				 at: 'N' put: 8;
				 at: 'Bold' put: 1;
				 at: 'Italic' put: 2;
				 at: 'Oblique' put: 2;
				 at: 'Narrow' put: 8;
				 at: 'Condensed' put: 8;
				 at: 'Underlined' put: 4;
				 yourself )! !

!TextStyle class methodsFor: 'mime file in/out' stamp: 'tpr 6/10/2005 18:29'!
writeSF2FamilyNamed: familyName inDirectory: directoryName toChangeSet: csName 
	"
	TextStyle writeSF2FamilyNamed: 'Accuny' inDirectory: 'AccunyCorrectedFeb252004Beta Folder' toChangeSet: 'AccunyInstall'.
	"

	|  family |
	family := OrderedCollection new.
	family addAll: (StrikeFont readStrikeFont2Family: familyName fromDirectory: (FileDirectory default fullNameFor: directoryName)) .
	family do: [:f | f reset].
	self 
		writeStyle: (TextStyle fontArray: family asArray)
		named: familyName
		toChangeSet: csName! !

!TextStyle class methodsFor: 'mime file in/out' stamp: 'nk 3/15/2004 19:46'!
writeStyle: aTextStyle named: familyName toChangeSet: csName
	"Write the text style to a change set, with a postscript that will re-load it.
	NOTE: to do TTCFonts, you have to have a working ShortPointArray endianness conversion."
	"
	TTCFont recreateCache.
	TextStyle writeStyle: (TextStyle named: #Arial) named: 'Arial' toChangeSet: 'ArialInstall'.
	
	TextStyle writeStyle: (TextStyle named: #Accuny) named: 'Accuny' toChangeSet: 'AccunyInstall2'.
	"

	| cs mimeStream |

	cs := ChangeSet basicNewNamed: csName.
	cs adoptSelector: #collectionFromCompressedMIMEString: forClass: self class.
	cs adoptSelector: #replaceStyle:with: forClass: self class.
	cs adoptSelector: #replaceFontsIn:with: forClass: self class.
	cs adoptSelector: #looseFontsFromFamily: forClass: self class.
	((aTextStyle fontArray copyWithout: nil) collect: [ :f | f class ]) asSet do: [ :cls  | 
		cs adoptSelector: #derivativeFonts forClass: cls.
		cs adoptSelector: #releaseCachedState forClass: cls ].

	cs preambleString: (String streamContents: [ :s |
		s nextPutAll: '"Change Set:		'; nextPutAll: csName; cr;
		nextPutAll: 'Date:		'; print: Date today; cr;
		nextPutAll: 'Author:		'; nextPutAll: Utilities authorName; cr; cr;
		nextPutAll: 'Installs the text style '''; nextPutAll: familyName; nextPutAll: ''''; cr;
		nextPutAll: 'from a compressed MIME encoding in the postscript."'; cr. ]).

	mimeStream := aTextStyle compressedMIMEEncodedStream.

	cs postscriptString: (String streamContents: [ :s | s
		nextPutAll: '"Postscript:'; cr;
		nextPutAll: 'Install the text style from the compressed MIME encoding, and replace the old one.';
		nextPut: $"; cr;
		nextPutAll: 'TextConstants at: #';
		nextPutAll: familyName;
		nextPutAll: ' ifPresent: [ :oldStyle | TextConstants at: #';
		nextPutAll: familyName;
		nextPutAll: 'OLD put: oldStyle. TextConstants removeKey: #';
		nextPutAll: familyName;
		nextPutAll: ' ].';
		cr;
		nextPutAll: 'TextStyle collectionFromCompressedMIMEString: ';
		cr;
		print: mimeStream contents;
		nextPut: $.; cr; cr;
		nextPutAll: 'TextConstants at: #';
		nextPutAll: familyName;
		nextPutAll: 'OLD ifPresent: [ :oldStyle | TextStyle replaceStyle: oldStyle with: (TextStyle named: ''';
		nextPutAll: familyName;
		nextPutAll: ''') ].';
		cr;
		nextPutAll: 'TextStyle replaceFontsIn: (TextStyle looseFontsFromFamily: ''';
		nextPutAll: familyName;
		nextPutAll: ''') with: (TextStyle named: ''';
		nextPutAll: familyName;
		nextPutAll: ''').';		
		cr ]).

	cs fileOut.

! !

!TextStyle methodsFor: 'comparing' stamp: 'tk 7/5/2001 22:18'!
= other

	self species == other species ifFalse: [^ false].
	1 to: self class instSize do:
		[:i | (self instVarAt: i) == (other instVarAt: i) ifFalse: [^ false]].
	^ true! !

!TextStyle methodsFor: 'fonts and font indexes' stamp: 'yo 5/7/2004 11:18'!
addLinedIfTT

	(fontArray first isKindOf: TTCFont) ifFalse: [^ self].

	fontArray do: [:f |
		f addLined.
	].
! !

!TextStyle methodsFor: 'fonts and font indexes' stamp: 'yo 3/17/2005 11:10'!
addNewFontSize: pointSize
	"Add a font in specified size to the array of fonts."
	| f d newArray t isSet fallbackStyle |
	fontArray first emphasis ~= 0 ifTrue: [
		t := TextConstants at: self fontArray first familyName asSymbol.
		t fonts first emphasis = 0 ifTrue: [
			^ t addNewFontSize: pointSize.
		].
	].

	pointSize <= 0 ifTrue: [^ nil].
	fontArray do: [:s |
		s pointSize = pointSize ifTrue: [^ s].
	].

	(isSet := fontArray first isKindOf: TTCFontSet) 
	ifTrue:[
		| fonts |
		fonts := fontArray first fontArray collect: [ :font |
			| newFont |
			(font isNil)
			ifTrue: [newFont := nil]
			ifFalse: [
				newFont := (font ttcDescription size > 256)
					ifTrue: [MultiTTCFont new initialize]
					ifFalse: [TTCFont new initialize].
				newFont ttcDescription: font ttcDescription.
				newFont pixelSize: pointSize * 96 // 72.
				font derivativeFonts notEmpty ifTrue: [font derivativeFonts do: [ :proto |
					proto ifNotNil: [
						d := proto class new initialize.
						d ttcDescription: proto ttcDescription.
						d pixelSize: newFont pixelSize.
						newFont derivativeFont: d]]].
				].
			newFont].
		f := TTCFontSet newFontArray: fonts]
	ifFalse: [
		f := TTCFont new initialize.
		f ttcDescription: fontArray first ttcDescription.
		f pointSize: pointSize.
		fontArray first derivativeFonts do: [:proto |
			proto ifNotNil: [
				d := TTCFont new initialize.
				d ttcDescription: proto ttcDescription.
				d pointSize: f pointSize.
				f derivativeFont: d.
			].
		].
	].
	isSet ifFalse: [
		fallbackStyle := TextStyle named: (fontArray first fallbackFont textStyleName).
	].
	newArray := ((fontArray copyWith: f) asSortedCollection: [:a :b | a pointSize <= b pointSize]) asArray.
	self newFontArray: newArray.
	isSet ifTrue: [
		TTCFontSet register: newArray at: newArray first familyName asSymbol.
	].
	isSet ifFalse: [
		f setupDefaultFallbackFontTo: fallbackStyle.
		f derivativeFonts do: [:g | g setupDefaultFallbackFontTo: fallbackStyle].
	].
	^ self fontOfPointSize: pointSize
! !

!TextStyle methodsFor: 'fonts and font indexes' stamp: 'rbb 3/1/2005 11:18'!
addNewFontSizeDialog: args
	"This is called from a modal menu and call back the menu with entered argument."
	| f n r |
	f := UIManager default request: 'Enter the point size' initialAnswer: '12'.
	n := f asNumber.
	r := self addNewFontSize: n.
	r ifNotNil: [
		args second ifNotNil: [args second modalSelection: {args first. n}].
	].
! !

!TextStyle methodsFor: 'accessing' stamp: ''!
alignment
	"Answer the code for the current setting of the alignment."

	^alignment! !

!TextStyle methodsFor: 'accessing' stamp: 'ar 9/21/2000 15:17'!
alignment: anInteger 
	"Set the current setting of the alignment to be anInteger:
	0=left flush, 1=right flush, 2=centered, 3=justified."

	alignment := anInteger \\ (Justified + 1)! !

!TextStyle methodsFor: 'accessing' stamp: 'ar 9/21/2000 15:16'!
alignmentSymbol
	"Answer the symbol for the current setting of the alignment."
	alignment = LeftFlush ifTrue:[^#leftFlush].
	alignment = Centered ifTrue:[^#centered].
	alignment = RightFlush ifTrue:[^#rightFlush].
	alignment = Justified ifTrue:[^#justified].
	^#leftFlush! !

!TextStyle methodsFor: 'accessing' stamp: ''!
baseline
	"Answer the distance from the top of the line to the bottom of most of the 
	characters (by convention, bottom of the letter 'A')."

	^baseline! !

!TextStyle methodsFor: 'accessing' stamp: ''!
baseline: anInteger 
	"Set the distance from the top of the line to the bottom of most of the 
	characters."

	baseline := anInteger! !

!TextStyle methodsFor: 'accessing' stamp: ''!
centered
	alignment := 2! !

!TextStyle methodsFor: 'tabs and margins' stamp: ''!
clearIndents
	"Reset all the margin (index) settings to be 0."

	self firstIndent: 0.
	self restIndent: 0.
	self rightIndent: 0! !

!TextStyle methodsFor: 'fonts and font indexes' stamp: 'tk 3/5/2002 09:57'!
collectionFromFileNamed: fileName
	"Read the file.  It is an TextStyle whose StrikeFonts are to be added to the system.  (Written by fooling SmartRefStream, so it won't write a DiskProxy!!)  These fonts will be added to the master TextSytle for this font family.  
	To write out fonts: 
		| ff | ff := ReferenceStream fileNamed: 'new fonts'.
		TextConstants at: #forceFontWriting put: true.
		ff nextPut: (TextConstants at: #AFontName).
			'do not mix font families in the TextStyle written out'.
		TextConstants at: #forceFontWriting put: false.
		ff close.

	To read: (TextStyle default collectionFromFileNamed: 'new fonts')
*** Do not remove this method *** "

	| ff this newName style heights |
	ff := ReferenceStream fileNamed: fileName.
	this := ff nextAndClose.	"Only works if file created by special code above"
	newName := this fontArray first familyName.
	this fontArray do: [:aFont | aFont familyName = newName ifFalse: [
		self error: 'All must be same family']].
	style := TextConstants at: newName asSymbol ifAbsent: [
		^ TextConstants at: newName asSymbol put: this].		"new family"
	this fontArray do: [:aFont | "add new fonts"
		heights := style fontArray collect: [:bFont | bFont height].
		(heights includes: aFont height) ifFalse: [
			style fontAt: style fontArray size + 1 put: aFont]].
! !

!TextStyle methodsFor: 'mime file in/out' stamp: 'nk 8/31/2004 09:23'!
compressedMIMEEncodedStream
	"Answer a ReadWriteStream with my compressed, stored representation as Base64"

	| s ff ffcontents s2 gzs |
	self fontArray do: [:f | f releaseCachedState].
	s := RWBinaryOrTextStream on: ''.
	ff := SmartRefStream on: s reset.
	TextConstants at: #forceFontWriting put: true.
	[ff nextPut: self] 
		ensure: [TextConstants at: #forceFontWriting put: false].
	ffcontents := s contents.
	ff close.
	s2 := RWBinaryOrTextStream on: ''.
	gzs := GZipWriteStream on: s2.
	gzs nextPutAll: ffcontents.
	gzs close.
	s2 reset.
	s := RWBinaryOrTextStream on: (ByteArray new: 10000).
	Base64MimeConverter mimeEncode: s2 to: s.
	^s
		ascii;
		reset;
		yourself! !

!TextStyle methodsFor: 'fonts and font indexes' stamp: 'sw 12/8/1999 18:02'!
consistOnlyOf: aFont
	fontArray := Array with: aFont.
	defaultFontIndex := 1! !

!TextStyle methodsFor: 'private' stamp: ''!
consolidate
	"If this style includes any fonts that are also in the default style,
	then replace them with references to the default ones."
"
	TextStyle allInstancesDo: [:s | s == TextStyle default ifFalse: [s consolidate]]
"
	| defFonts font |
	defFonts := TextStyle default fontArray.
	1 to: fontArray size do:
		[:i | font := fontArray at: i.
		1 to: defFonts size do:
			[:j | (font name asUppercase copyWithout: $ )
			= ((defFonts at: j) name asUppercase copyWithout: $ )
			ifTrue: [fontArray at: i put: (defFonts at: j)]]]! !

!TextStyle methodsFor: 'accessing' stamp: 'sw 12/6/1999 12:31'!
defaultFont
	^ fontArray at: self defaultFontIndex! !

!TextStyle methodsFor: 'default font' stamp: 'sw 12/6/1999 12:30'!
defaultFontIndex
	^ defaultFontIndex ifNil: [defaultFontIndex := 1]! !

!TextStyle methodsFor: 'default font' stamp: 'sw 12/6/1999 13:50'!
defaultFontIndex: anIndex
	defaultFontIndex := anIndex! !

!TextStyle methodsFor: 'fonts and font indexes' stamp: 'yo 6/23/2003 19:58'!
discardOtherSizes
	"This method trys to discard the fonts in non-standard size.  If the size is still in use, there will be a problem."
	| newArray |
	self isTTCStyle ifFalse: [^ self].
	newArray := fontArray select: [:s | TTCFont pointSizes includes: s pointSize].
	self newFontArray: newArray.

"(TextConstants at: #ComicSansMS) discardOtherSizes"! !

!TextStyle methodsFor: 'accessing' stamp: ''!
firstIndent
	"Answer the horizontal indenting of the first line of a paragraph in the 
	style of the receiver."

	^firstIndent! !

!TextStyle methodsFor: 'accessing' stamp: ''!
firstIndent: anInteger 
	"Set the horizontal indenting of the first line of a paragraph in the style 
	of the receiver to be the argument, anInteger."

	firstIndent := anInteger! !

!TextStyle methodsFor: 'fonts and font indexes' stamp: ''!
flushFonts
	"Clean out the fonts, an aid when snapshotting claims too many are 
	holding onto Display."

	(self confirm: 
'flushFonts is very dangerous.
Are you foolish or clever enough to proceed?')
		ifTrue: [1 to: fontArray size do: [:index | fontArray at: index put: nil]]
		ifFalse: [Transcript cr; show: 'flushFonts cancelled']

	"TextStyle default flushFonts"! !

!TextStyle methodsFor: 'private' stamp: ''!
fontArray
	"Only for writing out fonts, etc.  8/16/96 tk"
	^ fontArray! !

!TextStyle methodsFor: 'private' stamp: 'di 3/20/1999 22:31'!
fontAt: index 
	"This is private because no object outside TextStyle should depend on the 
	representation of the font family in fontArray."

	^ fontArray atPin: index! !

!TextStyle methodsFor: 'private' stamp: ''!
fontAt: index put: font
	"Automatically grow the array.  8/20/96 tk"
	index > fontArray size ifTrue: [
		fontArray := fontArray, (Array new: index - fontArray size)].
	fontArray at: index put: font! !

!TextStyle methodsFor: 'fonts and font indexes' stamp: 'sw 12/6/1999 13:54'!
fontIndexOf: aFont
	^ fontArray indexOf: aFont ifAbsent: [nil]! !

!TextStyle methodsFor: 'fonts and font indexes' stamp: 'nk 6/12/2004 16:31'!
fontIndexOfPointSize: desiredPointSize
	"Returns an index in fontArray of the font with pointSize <= desiredPointSize"
	"Leading is not inluded in the comparison"
	| bestMatch bestIndex d |
	bestMatch := 9999.  bestIndex := 1.
	1 to: fontArray size do:
		[:i | d := desiredPointSize - (fontArray at: i) pointSize.
		d = 0 ifTrue: [^ i].
		(d > 0 and: [d < bestMatch]) ifTrue: [bestIndex := i. bestMatch := d]].
	^ bestIndex! !

!TextStyle methodsFor: 'fonts and font indexes' stamp: 'di 10/11/97 09:23'!
fontIndexOfSize: desiredHeight
	"Returns an index in fontArray of the font with height <= desiredHeight"
	"Leading is not inluded in the comparison"
	| bestMatch bestIndex d |
	bestMatch := 9999.  bestIndex := 1.
	1 to: fontArray size do:
		[:i | d := desiredHeight - (fontArray at: i) height.
		d = 0 ifTrue: [^ i].
		(d > 0 and: [d < bestMatch]) ifTrue: [bestIndex := i. bestMatch := d]].
	^ bestIndex! !

!TextStyle methodsFor: 'accessing' stamp: ''!
fontNamed: fontName  "TextStyle default fontNamed: 'TimesRoman10'"
	^ fontArray detect: [:x | x name sameAs: fontName]! !

!TextStyle methodsFor: 'accessing' stamp: ''!
fontNames  "TextStyle default fontNames"
	^ fontArray collect: [:x | x name]! !

!TextStyle methodsFor: 'accessing' stamp: 'tk 6/26/1998 15:03'!
fontNamesAndSizes  "TextStyle default fontNames"
	^ fontArray collect: [:x | x name, ' ', x height printString]! !

!TextStyle methodsFor: 'accessing' stamp: 'ar 9/21/2000 11:53'!
fontNamesWithPointSizes
	^ fontArray collect:
		[:x | x fontNameWithPointSize]

  "TextStyle default fontNamesWithPointSizes"! !

!TextStyle methodsFor: 'fonts and font indexes' stamp: 'nk 6/12/2004 16:33'!
fontOfPointSize: aPointSize
	^ fontArray at: (self fontIndexOfPointSize: aPointSize)! !

!TextStyle methodsFor: 'fonts and font indexes' stamp: 'di 10/11/97 09:33'!
fontOfSize: aHeight
	"See fontIndexOfSize.
	Returns the actual font.  Leading not considered."

	^ fontArray at: (self fontIndexOfSize: aHeight)! !

!TextStyle methodsFor: 'accessing' stamp: 'ar 12/16/2001 16:58'!
fonts
	"Return a collection of fonts contained in this text style"
	^fontArray! !

!TextStyle methodsFor: 'private' stamp: ''!
gridForFont: fontIndex withLead: leadInteger 
	"Force whole style to suit one of its fonts. Assumes only one font referred
	to by runs."
	| font |
	font := self fontAt: fontIndex.
	self lineGrid: font height + leadInteger.
	self baseline: font ascent.
	self leading: leadInteger! !

!TextStyle methodsFor: 'comparing' stamp: 'ar 9/9/2003 22:04'!
hash
	"#hash is re-implemented because #= is re-implemented"
	^fontArray hash
! !

!TextStyle methodsFor: 'accessing' stamp: 'nk 6/25/2003 12:54'!
isTTCStyle

	^ fontArray first isTTCFont.
! !

!TextStyle methodsFor: 'accessing' stamp: ''!
justified
	alignment := 3! !

!TextStyle methodsFor: 'accessing' stamp: ''!
leading
	"Leading (from typographers historical use of extra lead (type metal))
	is the extra spacing above and beyond that needed just to accomodate
	the various font heights in the set."
	^ leading! !

!TextStyle methodsFor: 'accessing' stamp: ''!
leading: yDelta

	leading := yDelta! !

!TextStyle methodsFor: 'accessing' stamp: ''!
leftFlush
	alignment := 0! !

!TextStyle methodsFor: 'tabs and margins' stamp: ''!
leftMarginTabAt: marginIndex 
	"Set the 'nesting' level of left margin indents of the paragraph in the 
	style of the receiver to be the argument, marginIndex."

	(marginIndex > 0 and: [marginIndex < marginTabsArray size])
		ifTrue: [^(marginTabsArray at: marginIndex) at: 1]
		ifFalse: [^0]	
	"The marginTabsArray is an Array of tuples.  The Array is indexed according 
	to the marginIndex, the 'nesting' level of the requestor."
! !

!TextStyle methodsFor: 'accessing' stamp: ''!
lineGrid
	"Answer the relative space between lines of a paragraph in the style of 
	the receiver."

	^lineGrid! !

!TextStyle methodsFor: 'accessing' stamp: ''!
lineGrid: anInteger 
	"Set the relative space between lines of a paragraph in the style of the 
	receiver to be the argument, anInteger."

	lineGrid := anInteger! !

!TextStyle methodsFor: 'make arrows' stamp: 'sps 10/15/2003 17:09'!
makeArrows
"
TextStyle default makeArrows.
"
	fontArray do: [ :font |
		(font isKindOf: StrikeFont) ifTrue:[ 
			font 
				makeAssignArrow; 
				makeReturnArrow.
		]
	].
! !

!TextStyle methodsFor: 'private' stamp: ''!
marginTabAt: marginIndex side: sideIndex 
	"The marginTabsArray is an Array of tuples.  The Array is indexed
	according to the marginIndex, the 'nesting' level of the requestor.
	sideIndex is 1 for left, 2 for right."

	(marginIndex > 0 and: [marginIndex < marginTabsArray size])
		ifTrue: [^(marginTabsArray at: marginIndex) at: sideIndex]
		ifFalse: [^0]! !

!TextStyle methodsFor: 'private' stamp: 'tk 8/20/96'!
newFontArray: anArray
	"Currently there is no supporting protocol for changing these arrays. If an editor wishes to implement margin setting, then a copy of the default should be stored with these instance variables.  
	, Make size depend on first font."

	fontArray := anArray.
	lineGrid := (fontArray at: 1) height + leading.	"For whole family"
	baseline := (fontArray at: 1) ascent + leading.
	alignment := 0.
	firstIndent := 0.
	restIndent := 0.
	rightIndent := 0.
	tabsArray := DefaultTabsArray.
	marginTabsArray := DefaultMarginTabsArray
"
TextStyle allInstancesDo: [:ts | ts newFontArray: TextStyle default fontArray].
"! !

!TextStyle methodsFor: 'tabs and margins' stamp: ''!
nextTabXFrom: anX leftMargin: leftMargin rightMargin: rightMargin 
	"Tab stops are distances from the left margin. Set the distance into the 
	argument, anX, normalized for the paragraph's left margin."

	| normalizedX tabX |
	normalizedX := anX - leftMargin.
	1 to: tabsArray size do: 
		[:i | (tabX := tabsArray at: i) > normalizedX 
				ifTrue: [^leftMargin + tabX min: rightMargin]].
	^rightMargin! !

!TextStyle methodsFor: 'accessing' stamp: 'ar 12/16/2001 16:43'!
pointSizes
	^ fontArray collect:
		[:x | x pointSize]

  "TextStyle default fontNamesWithPointSizes"! !

!TextStyle methodsFor: 'accessing' stamp: 'yo 5/24/2004 22:52'!
printOn: aStream

	super printOn: aStream.
	(fontArray first isMemberOf: StrikeFontSet) ifTrue: [
		aStream space; nextPutAll: self defaultFont familySizeFace first; nextPutAll: '(FontSet)'
	] ifFalse: [
		aStream space; nextPutAll: self defaultFont familySizeFace first
	]
! !

!TextStyle methodsFor: 'accessing' stamp: ''!
restIndent
	"Answer the indent for all but the first line of a paragraph in the style 
	of the receiver."

	^restIndent! !

!TextStyle methodsFor: 'accessing' stamp: ''!
restIndent: anInteger 
	"Set the indent for all but the first line of a paragraph in the style of the 
	receiver to be the argument, anInteger."

	restIndent := anInteger! !

!TextStyle methodsFor: 'accessing' stamp: ''!
rightFlush
	alignment := 1! !

!TextStyle methodsFor: 'accessing' stamp: ''!
rightIndent
	"Answer the right margin indent for the lines of a paragraph in the style 
	of the receiver."

	^rightIndent! !

!TextStyle methodsFor: 'accessing' stamp: ''!
rightIndent: anInteger 
	"Answer the right margin indent for the lines of a paragraph in the style 
	of the receiver to be the argument, anInteger."

	rightIndent := anInteger! !

!TextStyle methodsFor: 'tabs and margins' stamp: ''!
rightMarginTabAt: marginIndex 
	"Set the 'nesting' level of right margin indents of the paragraph in the 
	style of the receiver to be marginIndex."

	(marginIndex > 0 and: [marginIndex < marginTabsArray size])
		ifTrue: [^(marginTabsArray at: marginIndex) at: 2]
		ifFalse: [^0]
	"The marginTabsArray is an Array of tuples.  The Array is indexed according 
	to the marginIndex, the 'nesting' level of the requestor."
! !

!TextStyle methodsFor: 'comparing' stamp: 'tk 7/5/2001 22:18'!
species

	^TextStyle! !

!TextStyle methodsFor: 'Disk I/O' stamp: 'tk 9/25/2000 10:10'!
storeDataOn: aDataStream
	"The shared arrays in tabsArray and marginTabsArray are the globals DefaultTabsArray and DefaultMarginTabsArray.  DiskProxies will be substituted for these in (Array objectForDataStream:)."

	^ super storeDataOn: aDataStream! !

!TextStyle methodsFor: 'tabs and margins' stamp: ''!
tabWidth
	"Answer the width of a tab."

	^DefaultTab! !

!TextStyle methodsFor: 'Disk I/O' stamp: 'di 11/19/1999 20:12'!
veryDeepCopyWith: deepCopier
	"All inst vars are meant to be shared"

	self == #veryDeepCopyWith:.	"to satisfy checkVariables"
	^ deepCopier references at: self ifAbsent: [
		deepCopier references at: self put: self clone].	"remember"! !

Interval subclass: #TextLineInterval
	instanceVariableNames: 'internalSpaces paddingWidth lineHeight baseline'
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'Graphics-Text'!

!TextLineInterval commentStamp: '<historical>' prior: 0!
My instances specify the starting and stopping points in a String of a composed line. The step is always 1.!

!TextLineInterval class methodsFor: 'instance creation' stamp: ''!
start: startInteger stop: stopInteger internalSpaces: spacesInteger paddingWidth: padWidthInteger
	"Answer an instance of me with the arguments as the start, stop points, 
	number of spaces in the line, and width of the padding."
	| newSelf |
	newSelf := super from: startInteger to: stopInteger by: 1.
	^newSelf internalSpaces: spacesInteger paddingWidth: padWidthInteger! !

!TextLineInterval methodsFor: 'comparing' stamp: ''!
= line

	self species = line species
		ifTrue: [^((start = line first and: [stop = line last])
				and: [internalSpaces = line internalSpaces])
				and: [paddingWidth = line paddingWidth]]
		ifFalse: [^false]! !

!TextLineInterval methodsFor: 'accessing' stamp: ''!
baseline
	^ baseline! !

!TextLineInterval methodsFor: 'accessing' stamp: ''!
internalSpaces
	"Answer the number of spaces in the line."

	^internalSpaces! !

!TextLineInterval methodsFor: 'accessing' stamp: ''!
internalSpaces: spacesInteger 
	"Set the number of spaces in the line to be spacesInteger."

	internalSpaces := spacesInteger! !

!TextLineInterval methodsFor: 'private' stamp: ''!
internalSpaces: spacesInteger paddingWidth: padWidthInteger

	internalSpaces := spacesInteger.
	paddingWidth := padWidthInteger! !

!TextLineInterval methodsFor: 'scanning' stamp: ''!
justifiedPadFor: spaceIndex 
	"Compute the width of pad for a given space in a line of justified text."

	| pad |
	internalSpaces = 0 ifTrue: [^0].
	pad := paddingWidth // internalSpaces.
	spaceIndex <= (paddingWidth \\ internalSpaces)
		ifTrue: [^pad + 1]
		ifFalse: [^pad]! !

!TextLineInterval methodsFor: 'scanning' stamp: ''!
justifiedTabDeltaFor: spaceIndex 
	"Compute the delta for a tab in a line of justified text, so tab falls 
	somewhere plausible when line is justified."

	| pad extraPad |
	internalSpaces = 0 ifTrue: [^0].
	pad := paddingWidth // internalSpaces.
	extraPad := paddingWidth \\ internalSpaces.
	spaceIndex <= extraPad
		ifTrue: [^spaceIndex * (pad + 1)]
		ifFalse: [^extraPad * (pad + 1) + (spaceIndex - extraPad * pad)]! !

!TextLineInterval methodsFor: 'accessing' stamp: ''!
lineHeight
	^ lineHeight! !

!TextLineInterval methodsFor: 'private' stamp: ''!
lineHeight: height baseline: ascent

	lineHeight := height.
	baseline := ascent! !

!TextLineInterval methodsFor: 'accessing' stamp: ''!
paddingWidth
	"Answer the amount of space to be added to the font."

	^paddingWidth! !

!TextLineInterval methodsFor: 'accessing' stamp: ''!
paddingWidth: padWidthInteger 
	"Set the amount of space to be added to the font to be padWidthInteger."

	paddingWidth := padWidthInteger! !

!TextLineInterval methodsFor: 'updating' stamp: ''!
slide: delta 
	"Change the starting and stopping points of the line by delta."

	start := start + delta.
	stop := stop + delta! !

!TextLineInterval methodsFor: 'accessing' stamp: ''!
stop: stopInteger 
	"Set the stopping point in the string of the line to be stopInteger."

	stop := stopInteger! !
Color initialize!
CornerRounder initialize!
Cursor initialize!
Form initialize!
IdentityTransform initialize!
GIFReadWriter initialize!
JPEGReadWriter initialize!
PNGReadWriter initialize!
XBMReadWriter initialize!
JPEGHuffmanTable initialize!
TextPrinter initialize!
JPEGReadStream initialize!
CharacterScanner initialize!
TextStyle initialize!
